/*
 * Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 */

#define __NVSHMEM_NUMBA_SUPPORT__
#include <nvshmem.h>
#include <nvshmemx.h>
