# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
__init__.py for nvshmem Python package
"""
import os

__package_name__ = os.getenv("PACKAGE_NAME", "nvshmem4py")

# Version is autogenerated by setuptools_scm
from .version import version, __version__, __version_tuple__
