/* Generated by Cython 3.2.8 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "language": "c++",
        "name": "nvshmem.bindings._internal.nvshmem",
        "sources": [
            "nvshmem/bindings/_internal/nvshmem.pyx"
        ]
    },
    "module_name": "nvshmem.bindings._internal.nvshmem"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_8"
#define CYTHON_HEX_VERSION 0x030208F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__nvshmem__bindings___internal__nvshmem
#define __PYX_HAVE_API__nvshmem__bindings___internal__nvshmem
/* Early includes */
#include <stdint.h>
#include "cuda_runtime_api.h"
#include "cuda_fp16.h"
#include "cuda_bf16.h"
#include "limits.h"
#include <dlfcn.h>
#include <limits.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "nvshmem/bindings/_internal/nvshmem.pyx",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */

/* "cynvshmem.pxd":143
 * 
 * # Longs
 * ctypedef long longlong             # <<<<<<<<<<<<<<
 * ctypedef signed char schar
 * 
*/
typedef long __pyx_t_7nvshmem_8bindings_9cynvshmem_longlong;

/* "cynvshmem.pxd":144
 * # Longs
 * ctypedef long longlong
 * ctypedef signed char schar             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef signed char __pyx_t_7nvshmem_8bindings_9cynvshmem_schar;

/* "cynvshmem.pxd":151
 *     int version
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'             # <<<<<<<<<<<<<<
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
*/
typedef int nvshmemx_qp_handle_t;

/* "cynvshmem.pxd":152
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
*/
typedef uint64_t nvshmemx_team_uniqueid_t;

/* "cynvshmem.pxd":153
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
*/
typedef int32_t nvshmem_team_t;

/* "cynvshmem.pxd":165
 *     nvshmemx_team_uniqueid_t uniqueid
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
*/
typedef nvshmem_team_t nvshmemx_team_t;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct nvshmemx_uniqueid_v1;
typedef struct nvshmemx_uniqueid_v1 nvshmemx_uniqueid_v1;
struct nvshmemx_uniqueid_args_v1;
typedef struct nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_v1;
struct nvshmem_team_config_v2;
typedef struct nvshmem_team_config_v2 nvshmem_team_config_v2;
struct nvshmemx_init_args_v2;
typedef struct nvshmemx_init_args_v2 nvshmemx_init_args_v2;
struct nvshmemx_init_attr_v2;
typedef struct nvshmemx_init_attr_v2 nvshmemx_init_attr_v2;

/* "cynvshmem.pxd":21
 * 
 * # enums
 * ctypedef enum nvshmemx_signal_op_t "nvshmemx_signal_op_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_SIGNAL_SET "NVSHMEM_SIGNAL_SET" = 9
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
*/
enum nvshmemx_signal_op_t {
  NVSHMEM_SIGNAL_SET = 9,
  NVSHMEM_SIGNAL_ADD = 10
};
typedef enum nvshmemx_signal_op_t nvshmemx_signal_op_t;

/* "cynvshmem.pxd":25
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
 * 
 * ctypedef enum nvshmemx_tma_policy_t "nvshmemx_tma_policy_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_TMA_DISABLE "NVSHMEMX_TMA_DISABLE" = 0
 *     NVSHMEMX_TMA_ENABLE "NVSHMEMX_TMA_ENABLE" = 1
*/
enum nvshmemx_tma_policy_t {
  NVSHMEMX_TMA_DISABLE = 0,
  NVSHMEMX_TMA_ENABLE = 1,
  NVSHMEMX_TMA_FORCE = 2,
  NVSHMEMX_TMA_POLICY_MAX = 0x7FFF
};
typedef enum nvshmemx_tma_policy_t nvshmemx_tma_policy_t;

/* "cynvshmem.pxd":31
 *     NVSHMEMX_TMA_POLICY_MAX "NVSHMEMX_TMA_POLICY_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_smem_amount_t "nvshmemx_smem_amount_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SMEM_RECOMMENDED "NVSHMEMX_SMEM_RECOMMENDED" = 0
 *     NVSHMEMX_SMEM_MINIMUM "NVSHMEMX_SMEM_MINIMUM" = 1
*/
enum nvshmemx_smem_amount_t {
  NVSHMEMX_SMEM_RECOMMENDED = 0,
  NVSHMEMX_SMEM_MINIMUM = 1,
  NVSHMEMX_SMEM_BARRIERS_ONLY = 2,
  NVSHMEMX_SMEM_AMOUNT_MAX = 0x7FFF
};
typedef enum nvshmemx_smem_amount_t nvshmemx_smem_amount_t;

/* "cynvshmem.pxd":37
 *     NVSHMEMX_SMEM_AMOUNT_MAX "NVSHMEMX_SMEM_AMOUNT_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_cmp_type_t "nvshmemx_cmp_type_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_CMP_EQ "NVSHMEM_CMP_EQ" = 0
 *     NVSHMEM_CMP_NE "NVSHMEM_CMP_NE"
*/
enum nvshmemx_cmp_type_t {
  NVSHMEM_CMP_EQ = 0,
  NVSHMEM_CMP_NE,
  NVSHMEM_CMP_GT,
  NVSHMEM_CMP_LE,
  NVSHMEM_CMP_LT,
  NVSHMEM_CMP_GE,
  NVSHMEM_CMP_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_cmp_type_t nvshmemx_cmp_type_t;

/* "cynvshmem.pxd":46
 *     NVSHMEM_CMP_SENTINEL "NVSHMEM_CMP_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_thread_support_t "nvshmemx_thread_support_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_THREAD_SINGLE "NVSHMEM_THREAD_SINGLE" = 0
 *     NVSHMEM_THREAD_FUNNELED "NVSHMEM_THREAD_FUNNELED"
*/
enum nvshmemx_thread_support_t {
  NVSHMEM_THREAD_SINGLE = 0,
  NVSHMEM_THREAD_FUNNELED,
  NVSHMEM_THREAD_SERIALIZED,
  NVSHMEM_THREAD_MULTIPLE,
  NVSHMEM_THREAD_TYPE_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_thread_support_t nvshmemx_thread_support_t;

/* "cynvshmem.pxd":53
 *     NVSHMEM_THREAD_TYPE_SENTINEL "NVSHMEM_THREAD_TYPE_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_proxy_status_t "nvshmemx_proxy_status_t":             # <<<<<<<<<<<<<<
 *     PROXY_GLOBAL_EXIT_NOT_REQUESTED "PROXY_GLOBAL_EXIT_NOT_REQUESTED" = 0
 *     PROXY_GLOBAL_EXIT_INIT "PROXY_GLOBAL_EXIT_INIT"
*/
enum nvshmemx_proxy_status_t {
  PROXY_GLOBAL_EXIT_NOT_REQUESTED = 0,
  PROXY_GLOBAL_EXIT_INIT,
  PROXY_GLOBAL_EXIT_REQUESTED,
  PROXY_GLOBAL_EXIT_FINISHED,
  PROXY_GLOBAL_EXIT_MAX_STATE = 0x7FFF
};
typedef enum nvshmemx_proxy_status_t nvshmemx_proxy_status_t;

/* "cynvshmem.pxd":60
 *     PROXY_GLOBAL_EXIT_MAX_STATE "PROXY_GLOBAL_EXIT_MAX_STATE" = 32767
 * 
 * ctypedef enum nvshmemx_init_status_t "nvshmemx_init_status_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_STATUS_NOT_INITIALIZED "NVSHMEM_STATUS_NOT_INITIALIZED" = 0
 *     NVSHMEM_STATUS_IS_BOOTSTRAPPED "NVSHMEM_STATUS_IS_BOOTSTRAPPED"
*/
enum nvshmemx_init_status_t {
  NVSHMEM_STATUS_NOT_INITIALIZED = 0,
  NVSHMEM_STATUS_IS_BOOTSTRAPPED,
  NVSHMEM_STATUS_IS_INITIALIZED,
  NVSHMEM_STATUS_LIMITED_MPG,
  NVSHMEM_STATUS_FULL_MPG,
  NVSHMEM_STATUS_INVALID = 0x7FFF
};
typedef enum nvshmemx_init_status_t nvshmemx_init_status_t;

/* "cynvshmem.pxd":68
 *     NVSHMEM_STATUS_INVALID "NVSHMEM_STATUS_INVALID" = 32767
 * 
 * ctypedef enum nvshmemx_qp_handle_index_t "nvshmemx_qp_handle_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_QP_HOST "NVSHMEMX_QP_HOST" = 0
 *     NVSHMEMX_QP_DEFAULT "NVSHMEMX_QP_DEFAULT" = 1
*/
enum nvshmemx_qp_handle_index_t {
  NVSHMEMX_QP_HOST = 0,
  NVSHMEMX_QP_DEFAULT = 1,
  NVSHMEMX_QP_ANY = 0x7FFF,
  NVSHMEMX_QP_ALL = 0x7FFF
};
typedef enum nvshmemx_qp_handle_index_t nvshmemx_qp_handle_index_t;

/* "cynvshmem.pxd":74
 *     NVSHMEMX_QP_ALL "NVSHMEMX_QP_ALL" = 32767
 * 
 * ctypedef enum nvshmem_pe_index_t "nvshmem_pe_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
*/
enum nvshmem_pe_index_t {

  /* "cynvshmem.pxd":77
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)             # <<<<<<<<<<<<<<
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":
*/
  NVSHMEM_PE_INVALID = -1L,
  NVSHMEMX_PE_ANY = (1 << 15),
  NVSHMEMX_PE_ALL = (1 << 15)
};
typedef enum nvshmem_pe_index_t nvshmem_pe_index_t;

/* "cynvshmem.pxd":79
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_TEAM_INVALID "NVSHMEM_TEAM_INVALID" = -(1)
 *     NVSHMEM_TEAM_WORLD "NVSHMEM_TEAM_WORLD" = 0
*/
enum nvshmem_team_id_t {
  NVSHMEM_TEAM_INVALID = -1L,
  NVSHMEM_TEAM_WORLD = 0,
  NVSHMEM_TEAM_WORLD_INDEX = 0,
  NVSHMEM_TEAM_SHARED = 1,
  NVSHMEM_TEAM_SHARED_INDEX = 1,
  NVSHMEMX_TEAM_NODE = 2,
  NVSHMEM_TEAM_NODE_INDEX = 2,
  NVSHMEMX_TEAM_SAME_MYPE_NODE = 3,
  NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX = 3,
  NVSHMEMI_TEAM_SAME_GPU = 4,
  NVSHMEM_TEAM_SAME_GPU_INDEX = 4,
  NVSHMEMI_TEAM_GPU_LEADERS = 5,
  NVSHMEM_TEAM_GPU_LEADERS_INDEX = 5,
  NVSHMEM_TEAM_MC_SHARED = 6,
  NVSHMEM_TEAM_MC_SHARED_INDEX = 6,
  NVSHMEM_TEAMS_MIN = 7,
  NVSHMEM_TEAM_INDEX_MAX = 0x7FFF
};
typedef enum nvshmem_team_id_t nvshmem_team_id_t;

/* "cynvshmem.pxd":98
 *     NVSHMEM_TEAM_INDEX_MAX "NVSHMEM_TEAM_INDEX_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_status "nvshmemx_status":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SUCCESS "NVSHMEMX_SUCCESS"
 *     NVSHMEMX_ERROR_INVALID_VALUE "NVSHMEMX_ERROR_INVALID_VALUE"
*/
enum nvshmemx_status {
  NVSHMEMX_SUCCESS,
  NVSHMEMX_ERROR_INVALID_VALUE,
  NVSHMEMX_ERROR_OUT_OF_MEMORY,
  NVSHMEMX_ERROR_NOT_SUPPORTED,
  NVSHMEMX_ERROR_SYMMETRY,
  NVSHMEMX_ERROR_GPU_NOT_SELECTED,
  NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED,
  NVSHMEMX_ERROR_INTERNAL,
  NVSHMEMX_ERROR_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_status nvshmemx_status;

/* "cynvshmem.pxd":109
 *     NVSHMEMX_ERROR_SENTINEL "NVSHMEMX_ERROR_SENTINEL" = 32767
 * 
 * ctypedef enum flags "flags":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_INIT_THREAD_PES "NVSHMEMX_INIT_THREAD_PES" = 1
 *     NVSHMEMX_INIT_WITH_MPI_COMM "NVSHMEMX_INIT_WITH_MPI_COMM" = (1 << 1)
*/
enum flags {

  /* "cynvshmem.pxd":114
 *     NVSHMEMX_INIT_WITH_SHMEM "NVSHMEMX_INIT_WITH_SHMEM" = (1 << 2)
 *     NVSHMEMX_INIT_WITH_UNIQUEID "NVSHMEMX_INIT_WITH_UNIQUEID" = (1 << 3)
 *     NVSHMEMX_INIT_MAX "NVSHMEMX_INIT_MAX" = (1 << 31)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  NVSHMEMX_INIT_THREAD_PES = 1,
  NVSHMEMX_INIT_WITH_MPI_COMM = (1 << 1),
  NVSHMEMX_INIT_WITH_SHMEM = (1 << 2),
  NVSHMEMX_INIT_WITH_UNIQUEID = (1 << 3),
  NVSHMEMX_INIT_MAX = (1 << 31)
};
typedef enum flags flags;

/* "cynvshmem.pxd":122
 *     ctypedef CUstream_st* cudaStream_t
 * 
 * ctypedef void* CUmodule 'CUmodule'             # <<<<<<<<<<<<<<
 * ctypedef void* CUlibrary 'CUlibrary'
 * 
*/
typedef void *CUmodule;

/* "cynvshmem.pxd":123
 * 
 * ctypedef void* CUmodule 'CUmodule'
 * ctypedef void* CUlibrary 'CUlibrary'             # <<<<<<<<<<<<<<
 * 
 * # Types for NVSHMEM Collectives
*/
typedef void *CUlibrary;

/* "cynvshmem.pxd":148
 * 
 * 
 * ctypedef struct nvshmemx_uniqueid_v1 'nvshmemx_uniqueid_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     char internal[124]
*/
struct nvshmemx_uniqueid_v1 {
  int version;
  char internal[124];
};

/* "cynvshmem.pxd":154
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
 *     int version
*/
typedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t;

/* "cynvshmem.pxd":155
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_v1* id
*/
struct nvshmemx_uniqueid_args_v1 {
  int version;
  nvshmemx_uniqueid_v1 *id;
  int myrank;
  int nranks;
};

/* "cynvshmem.pxd":160
 *     int myrank
 *     int nranks
 * ctypedef struct nvshmem_team_config_v2 'nvshmem_team_config_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     int num_contexts
*/
struct nvshmem_team_config_v2 {
  int version;
  int num_contexts;
  nvshmemx_team_uniqueid_t uniqueid;
  char padding[48];
};

/* "cynvshmem.pxd":166
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
*/
typedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t;

/* "cynvshmem.pxd":167
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
 *     int version
*/
typedef nvshmem_team_config_v2 nvshmem_team_config_t;

/* "cynvshmem.pxd":168
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_args_t uid_args
*/
struct nvshmemx_init_args_v2 {
  int version;
  nvshmemx_uniqueid_args_t uid_args;
  int cuda_device_id;
  char content[92];
};

/* "cynvshmem.pxd":173
 *     int cuda_device_id
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':
 *     int version
*/
typedef nvshmemx_init_args_v2 nvshmemx_init_args_t;

/* "cynvshmem.pxd":174
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     void* mpi_comm
*/
struct nvshmemx_init_attr_v2 {
  int version;
  void *mpi_comm;
  nvshmemx_init_args_t args;
};

/* "cynvshmem.pxd":178
 *     void* mpi_comm
 *     nvshmemx_init_args_t args
 * ctypedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t 'nvshmemx_init_attr_t'             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyRuntimeError_Check.proto */
#define __Pyx_PyExc_RuntimeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeError)

/* IncludeStringH.proto (used by decode_c_string) */
#include <string.h>

/* decode_c_string_utf16.proto (used by decode_c_string) */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ErrOccurredWithGIL.proto */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetAttrStr.proto (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* FunctionExport.proto */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig);

/* GetApiDict.proto */
static PyObject *__Pyx_ApiExport_GetApiDict(void);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectCall2Args.proto (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectLookupSpecial.proto (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto (used by FetchCommonType) */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* CallTypeTraverse.proto (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* GCCDiagnostics.proto (used by CIntToPy) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* PyObjectVectorCallKwBuilder.proto (used by CIntToPy) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."


/* Module declarations from "libc.stdint" */

/* Module declarations from "nvshmem.bindings.cynvshmem" */

/* Module declarations from "nvshmem.bindings._internal.nvshmem" */
static int __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___py_nvshmem_init;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream;
static void *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream;
static PyObject *__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs = 0;
static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(void); /*proto*/
static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(void); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointer(PyObject *, int __pyx_skip_dispatch); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "nvshmem.bindings._internal.nvshmem"
extern int __pyx_module_is_main_nvshmem__bindings___internal__nvshmem;
int __pyx_module_is_main_nvshmem__bindings___internal__nvshmem = 0;

/* Implementation of "nvshmem.bindings._internal.nvshmem" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
static const char __pyx_k_nvshmem_ptr[] = "nvshmem_ptr";
static const char __pyx_k_nvshmem_free[] = "nvshmem_free";
static const char __pyx_k_nvshmem_align[] = "nvshmem_align";
static const char __pyx_k_nvshmem_my_pe[] = "nvshmem_my_pe";
static const char __pyx_k_nvshmem_n_pes[] = "nvshmem_n_pes";
static const char __pyx_k_nvshmem_calloc[] = "nvshmem_calloc";
static const char __pyx_k_nvshmem_malloc[] = "nvshmem_malloc";
static const char __pyx_k_nvshmem_barrier[] = "nvshmem_barrier";
static const char __pyx_k_nvshmemx_mc_ptr[] = "nvshmemx_mc_ptr";
static const char __pyx_k_nvshmem_team_my_pe[] = "nvshmem_team_my_pe";
static const char __pyx_k_nvshmem_team_n_pes[] = "nvshmem_team_n_pes";
static const char __pyx_k_nvshmemx_team_init[] = "nvshmemx_team_init";
static const char __pyx_k_nvshmem_barrier_all[] = "nvshmem_barrier_all";
static const char __pyx_k_libnvshmem_host_so_3[] = "libnvshmem_host.so.3";
static const char __pyx_k_nvshmem_team_destroy[] = "nvshmem_team_destroy";
static const char __pyx_k_nvshmemx_init_status[] = "nvshmemx_init_status";
static const char __pyx_k_nvshmem_team_split_2d[] = "nvshmem_team_split_2d";
static const char __pyx_k_nvshmemx_get_uniqueid[] = "nvshmemx_get_uniqueid";
static const char __pyx_k_nvshmemx_cumodule_init[] = "nvshmemx_cumodule_init";
static const char __pyx_k_nvshmem_team_get_config[] = "nvshmem_team_get_config";
static const char __pyx_k_nvshmemx_culibrary_init[] = "nvshmemx_culibrary_init";
static const char __pyx_k_nvshmem_info_get_version[] = "nvshmem_info_get_version";
static const char __pyx_k_nvshmemx_quiet_on_stream[] = "nvshmemx_quiet_on_stream";
static const char __pyx_k_nvshmem_team_translate_pe[] = "nvshmem_team_translate_pe";
static const char __pyx_k_nvshmemx_getmem_on_stream[] = "nvshmemx_getmem_on_stream";
static const char __pyx_k_nvshmemx_hostlib_finalize[] = "nvshmemx_hostlib_finalize";
static const char __pyx_k_nvshmemx_putmem_on_stream[] = "nvshmemx_putmem_on_stream";
static const char __pyx_k_nvshmem_team_split_strided[] = "nvshmem_team_split_strided";
static const char __pyx_k_nvshmemx_barrier_on_stream[] = "nvshmemx_barrier_on_stream";
static const char __pyx_k_nvshmemx_cumodule_finalize[] = "nvshmemx_cumodule_finalize";
static const char __pyx_k_nvshmemx_hostlib_init_attr[] = "nvshmemx_hostlib_init_attr";
static const char __pyx_k_nvshmemx_team_get_uniqueid[] = "nvshmemx_team_get_uniqueid";
static const char __pyx_k_nvshmemx_culibrary_finalize[] = "nvshmemx_culibrary_finalize";
static const char __pyx_k_nvshmemx_sync_all_on_stream[] = "nvshmemx_sync_all_on_stream";
static const char __pyx_k_nvshmemx_signal_op_on_stream[] = "nvshmemx_signal_op_on_stream";
static const char __pyx_k_nvshmemx_team_sync_on_stream[] = "nvshmemx_team_sync_on_stream";
static const char __pyx_k_nvshmemx_barrier_all_on_stream[] = "nvshmemx_barrier_all_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_max_reduce_on[] = "nvshmemx_bfloat16_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_min_reduce_on[] = "nvshmemx_bfloat16_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_sum_reduce_on[] = "nvshmemx_bfloat16_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_char_max_reducescatter[] = "nvshmemx_char_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_char_min_reducescatter[] = "nvshmemx_char_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_char_sum_reducescatter[] = "nvshmemx_char_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_half_max_reducescatter[] = "nvshmemx_half_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_half_min_reducescatter[] = "nvshmemx_half_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_half_sum_reducescatter[] = "nvshmemx_half_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int8_max_reducescatter[] = "nvshmemx_int8_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int8_min_reducescatter[] = "nvshmemx_int8_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int8_sum_reducescatter[] = "nvshmemx_int8_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int_alltoall_on_stream[] = "nvshmemx_int_alltoall_on_stream";
static const char __pyx_k_nvshmemx_int_fcollect_on_stream[] = "nvshmemx_int_fcollect_on_stream";
static const char __pyx_k_nvshmemx_long_max_reducescatter[] = "nvshmemx_long_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_long_min_reducescatter[] = "nvshmemx_long_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_long_sum_reducescatter[] = "nvshmemx_long_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_longlong_max_reduce_on[] = "nvshmemx_longlong_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_longlong_min_reduce_on[] = "nvshmemx_longlong_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_longlong_sum_reduce_on[] = "nvshmemx_longlong_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_set_attr_mpi_comm_args[] = "nvshmemx_set_attr_mpi_comm_args";
static const char __pyx_k_nvshmemx_set_attr_uniqueid_args[] = "nvshmemx_set_attr_uniqueid_args";
static const char __pyx_k_nvshmemx_size_max_reducescatter[] = "nvshmemx_size_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_size_min_reducescatter[] = "nvshmemx_size_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_size_sum_reducescatter[] = "nvshmemx_size_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_alltoall_on_st[] = "nvshmemx_bfloat16_alltoall_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_broadcast_on_s[] = "nvshmemx_bfloat16_broadcast_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_fcollect_on_st[] = "nvshmemx_bfloat16_fcollect_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_max_reducescat[] = "nvshmemx_bfloat16_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_min_reducescat[] = "nvshmemx_bfloat16_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_bfloat16_sum_reducescat[] = "nvshmemx_bfloat16_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_buffer_register_symmetr[] = "nvshmemx_buffer_register_symmetric";
static const char __pyx_k_nvshmemx_buffer_unregister_symme[] = "nvshmemx_buffer_unregister_symmetric";
static const char __pyx_k_nvshmemx_char_alltoall_on_stream[] = "nvshmemx_char_alltoall_on_stream";
static const char __pyx_k_nvshmemx_char_broadcast_on_strea[] = "nvshmemx_char_broadcast_on_stream";
static const char __pyx_k_nvshmemx_char_fcollect_on_stream[] = "nvshmemx_char_fcollect_on_stream";
static const char __pyx_k_nvshmemx_char_max_reduce_on_stre[] = "nvshmemx_char_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_char_min_reduce_on_stre[] = "nvshmemx_char_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_char_sum_reduce_on_stre[] = "nvshmemx_char_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_double_alltoall_on_stre[] = "nvshmemx_double_alltoall_on_stream";
static const char __pyx_k_nvshmemx_double_broadcast_on_str[] = "nvshmemx_double_broadcast_on_stream";
static const char __pyx_k_nvshmemx_double_fcollect_on_stre[] = "nvshmemx_double_fcollect_on_stream";
static const char __pyx_k_nvshmemx_double_max_reduce_on_st[] = "nvshmemx_double_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_double_max_reducescatte[] = "nvshmemx_double_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_double_min_reduce_on_st[] = "nvshmemx_double_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_double_min_reducescatte[] = "nvshmemx_double_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_double_sum_reduce_on_st[] = "nvshmemx_double_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_double_sum_reducescatte[] = "nvshmemx_double_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_float_alltoall_on_strea[] = "nvshmemx_float_alltoall_on_stream";
static const char __pyx_k_nvshmemx_float_broadcast_on_stre[] = "nvshmemx_float_broadcast_on_stream";
static const char __pyx_k_nvshmemx_float_fcollect_on_strea[] = "nvshmemx_float_fcollect_on_stream";
static const char __pyx_k_nvshmemx_float_max_reduce_on_str[] = "nvshmemx_float_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_float_max_reducescatter[] = "nvshmemx_float_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_float_min_reduce_on_str[] = "nvshmemx_float_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_float_min_reducescatter[] = "nvshmemx_float_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_float_sum_reduce_on_str[] = "nvshmemx_float_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_float_sum_reducescatter[] = "nvshmemx_float_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_half_alltoall_on_stream[] = "nvshmemx_half_alltoall_on_stream";
static const char __pyx_k_nvshmemx_half_broadcast_on_strea[] = "nvshmemx_half_broadcast_on_stream";
static const char __pyx_k_nvshmemx_half_fcollect_on_stream[] = "nvshmemx_half_fcollect_on_stream";
static const char __pyx_k_nvshmemx_half_max_reduce_on_stre[] = "nvshmemx_half_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_half_min_reduce_on_stre[] = "nvshmemx_half_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_half_sum_reduce_on_stre[] = "nvshmemx_half_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int16_alltoall_on_strea[] = "nvshmemx_int16_alltoall_on_stream";
static const char __pyx_k_nvshmemx_int16_broadcast_on_stre[] = "nvshmemx_int16_broadcast_on_stream";
static const char __pyx_k_nvshmemx_int16_fcollect_on_strea[] = "nvshmemx_int16_fcollect_on_stream";
static const char __pyx_k_nvshmemx_int16_max_reduce_on_str[] = "nvshmemx_int16_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_int16_max_reducescatter[] = "nvshmemx_int16_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int16_min_reduce_on_str[] = "nvshmemx_int16_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_int16_min_reducescatter[] = "nvshmemx_int16_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int16_sum_reduce_on_str[] = "nvshmemx_int16_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int16_sum_reducescatter[] = "nvshmemx_int16_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int32_alltoall_on_strea[] = "nvshmemx_int32_alltoall_on_stream";
static const char __pyx_k_nvshmemx_int32_broadcast_on_stre[] = "nvshmemx_int32_broadcast_on_stream";
static const char __pyx_k_nvshmemx_int32_fcollect_on_strea[] = "nvshmemx_int32_fcollect_on_stream";
static const char __pyx_k_nvshmemx_int32_max_reduce_on_str[] = "nvshmemx_int32_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_int32_max_reducescatter[] = "nvshmemx_int32_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int32_min_reduce_on_str[] = "nvshmemx_int32_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_int32_min_reducescatter[] = "nvshmemx_int32_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int32_sum_reduce_on_str[] = "nvshmemx_int32_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int32_sum_reducescatter[] = "nvshmemx_int32_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int64_alltoall_on_strea[] = "nvshmemx_int64_alltoall_on_stream";
static const char __pyx_k_nvshmemx_int64_broadcast_on_stre[] = "nvshmemx_int64_broadcast_on_stream";
static const char __pyx_k_nvshmemx_int64_fcollect_on_strea[] = "nvshmemx_int64_fcollect_on_stream";
static const char __pyx_k_nvshmemx_int64_max_reduce_on_str[] = "nvshmemx_int64_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_int64_max_reducescatter[] = "nvshmemx_int64_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int64_min_reduce_on_str[] = "nvshmemx_int64_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_int64_min_reducescatter[] = "nvshmemx_int64_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int64_sum_reduce_on_str[] = "nvshmemx_int64_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int64_sum_reducescatter[] = "nvshmemx_int64_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int8_alltoall_on_stream[] = "nvshmemx_int8_alltoall_on_stream";
static const char __pyx_k_nvshmemx_int8_broadcast_on_strea[] = "nvshmemx_int8_broadcast_on_stream";
static const char __pyx_k_nvshmemx_int8_fcollect_on_stream[] = "nvshmemx_int8_fcollect_on_stream";
static const char __pyx_k_nvshmemx_int8_max_reduce_on_stre[] = "nvshmemx_int8_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_int8_min_reduce_on_stre[] = "nvshmemx_int8_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_int8_sum_reduce_on_stre[] = "nvshmemx_int8_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int_broadcast_on_stream[] = "nvshmemx_int_broadcast_on_stream";
static const char __pyx_k_nvshmemx_int_max_reduce_on_strea[] = "nvshmemx_int_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_int_max_reducescatter_o[] = "nvshmemx_int_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int_min_reduce_on_strea[] = "nvshmemx_int_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_int_min_reducescatter_o[] = "nvshmemx_int_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_int_sum_reduce_on_strea[] = "nvshmemx_int_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_int_sum_reducescatter_o[] = "nvshmemx_int_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_long_alltoall_on_stream[] = "nvshmemx_long_alltoall_on_stream";
static const char __pyx_k_nvshmemx_long_broadcast_on_strea[] = "nvshmemx_long_broadcast_on_stream";
static const char __pyx_k_nvshmemx_long_fcollect_on_stream[] = "nvshmemx_long_fcollect_on_stream";
static const char __pyx_k_nvshmemx_long_max_reduce_on_stre[] = "nvshmemx_long_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_long_min_reduce_on_stre[] = "nvshmemx_long_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_long_sum_reduce_on_stre[] = "nvshmemx_long_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_longlong_alltoall_on_st[] = "nvshmemx_longlong_alltoall_on_stream";
static const char __pyx_k_nvshmemx_longlong_broadcast_on_s[] = "nvshmemx_longlong_broadcast_on_stream";
static const char __pyx_k_nvshmemx_longlong_fcollect_on_st[] = "nvshmemx_longlong_fcollect_on_stream";
static const char __pyx_k_nvshmemx_longlong_max_reducescat[] = "nvshmemx_longlong_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_longlong_min_reducescat[] = "nvshmemx_longlong_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_longlong_sum_reducescat[] = "nvshmemx_longlong_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_putmem_signal_on_stream[] = "nvshmemx_putmem_signal_on_stream";
static const char __pyx_k_nvshmemx_schar_alltoall_on_strea[] = "nvshmemx_schar_alltoall_on_stream";
static const char __pyx_k_nvshmemx_schar_broadcast_on_stre[] = "nvshmemx_schar_broadcast_on_stream";
static const char __pyx_k_nvshmemx_schar_fcollect_on_strea[] = "nvshmemx_schar_fcollect_on_stream";
static const char __pyx_k_nvshmemx_schar_max_reduce_on_str[] = "nvshmemx_schar_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_schar_max_reducescatter[] = "nvshmemx_schar_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_schar_min_reduce_on_str[] = "nvshmemx_schar_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_schar_min_reducescatter[] = "nvshmemx_schar_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_schar_sum_reduce_on_str[] = "nvshmemx_schar_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_schar_sum_reducescatter[] = "nvshmemx_schar_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_short_alltoall_on_strea[] = "nvshmemx_short_alltoall_on_stream";
static const char __pyx_k_nvshmemx_short_broadcast_on_stre[] = "nvshmemx_short_broadcast_on_stream";
static const char __pyx_k_nvshmemx_short_fcollect_on_strea[] = "nvshmemx_short_fcollect_on_stream";
static const char __pyx_k_nvshmemx_short_max_reduce_on_str[] = "nvshmemx_short_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_short_max_reducescatter[] = "nvshmemx_short_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_short_min_reduce_on_str[] = "nvshmemx_short_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_short_min_reducescatter[] = "nvshmemx_short_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_short_sum_reduce_on_str[] = "nvshmemx_short_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_short_sum_reducescatter[] = "nvshmemx_short_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_signal_wait_until_on_st[] = "nvshmemx_signal_wait_until_on_stream";
static const char __pyx_k_nvshmemx_size_alltoall_on_stream[] = "nvshmemx_size_alltoall_on_stream";
static const char __pyx_k_nvshmemx_size_broadcast_on_strea[] = "nvshmemx_size_broadcast_on_stream";
static const char __pyx_k_nvshmemx_size_fcollect_on_stream[] = "nvshmemx_size_fcollect_on_stream";
static const char __pyx_k_nvshmemx_size_max_reduce_on_stre[] = "nvshmemx_size_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_size_min_reduce_on_stre[] = "nvshmemx_size_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_size_sum_reduce_on_stre[] = "nvshmemx_size_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint16_alltoall_on_stre[] = "nvshmemx_uint16_alltoall_on_stream";
static const char __pyx_k_nvshmemx_uint16_broadcast_on_str[] = "nvshmemx_uint16_broadcast_on_stream";
static const char __pyx_k_nvshmemx_uint16_fcollect_on_stre[] = "nvshmemx_uint16_fcollect_on_stream";
static const char __pyx_k_nvshmemx_uint16_max_reduce_on_st[] = "nvshmemx_uint16_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint16_max_reducescatte[] = "nvshmemx_uint16_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint16_min_reduce_on_st[] = "nvshmemx_uint16_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint16_min_reducescatte[] = "nvshmemx_uint16_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint16_sum_reduce_on_st[] = "nvshmemx_uint16_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint16_sum_reducescatte[] = "nvshmemx_uint16_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint32_alltoall_on_stre[] = "nvshmemx_uint32_alltoall_on_stream";
static const char __pyx_k_nvshmemx_uint32_broadcast_on_str[] = "nvshmemx_uint32_broadcast_on_stream";
static const char __pyx_k_nvshmemx_uint32_fcollect_on_stre[] = "nvshmemx_uint32_fcollect_on_stream";
static const char __pyx_k_nvshmemx_uint32_max_reduce_on_st[] = "nvshmemx_uint32_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint32_max_reducescatte[] = "nvshmemx_uint32_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint32_min_reduce_on_st[] = "nvshmemx_uint32_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint32_min_reducescatte[] = "nvshmemx_uint32_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint32_sum_reduce_on_st[] = "nvshmemx_uint32_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint32_sum_reducescatte[] = "nvshmemx_uint32_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint64_alltoall_on_stre[] = "nvshmemx_uint64_alltoall_on_stream";
static const char __pyx_k_nvshmemx_uint64_broadcast_on_str[] = "nvshmemx_uint64_broadcast_on_stream";
static const char __pyx_k_nvshmemx_uint64_fcollect_on_stre[] = "nvshmemx_uint64_fcollect_on_stream";
static const char __pyx_k_nvshmemx_uint64_max_reduce_on_st[] = "nvshmemx_uint64_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint64_max_reducescatte[] = "nvshmemx_uint64_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint64_min_reduce_on_st[] = "nvshmemx_uint64_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint64_min_reducescatte[] = "nvshmemx_uint64_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint64_sum_reduce_on_st[] = "nvshmemx_uint64_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint64_sum_reducescatte[] = "nvshmemx_uint64_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint8_alltoall_on_strea[] = "nvshmemx_uint8_alltoall_on_stream";
static const char __pyx_k_nvshmemx_uint8_broadcast_on_stre[] = "nvshmemx_uint8_broadcast_on_stream";
static const char __pyx_k_nvshmemx_uint8_fcollect_on_strea[] = "nvshmemx_uint8_fcollect_on_stream";
static const char __pyx_k_nvshmemx_uint8_max_reduce_on_str[] = "nvshmemx_uint8_max_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint8_max_reducescatter[] = "nvshmemx_uint8_max_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint8_min_reduce_on_str[] = "nvshmemx_uint8_min_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint8_min_reducescatter[] = "nvshmemx_uint8_min_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_uint8_sum_reduce_on_str[] = "nvshmemx_uint8_sum_reduce_on_stream";
static const char __pyx_k_nvshmemx_uint8_sum_reducescatter[] = "nvshmemx_uint8_sum_reducescatter_on_stream";
static const char __pyx_k_nvshmemx_vendor_get_version_info[] = "nvshmemx_vendor_get_version_info";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem_2_inspect_function_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_codeobj_tab[2];
  PyObject *__pyx_string_tab[464];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_Failed_to_dlopen_libnvshmem __pyx_string_tab[1]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[2]
#define __pyx_kp_u__2 __pyx_string_tab[3]
#define __pyx_kp_u_add_note __pyx_string_tab[4]
#define __pyx_kp_u_function_nvshmem_align_is_not_fo __pyx_string_tab[5]
#define __pyx_kp_u_function_nvshmem_barrier_all_is __pyx_string_tab[6]
#define __pyx_kp_u_function_nvshmem_barrier_is_not __pyx_string_tab[7]
#define __pyx_kp_u_function_nvshmem_calloc_is_not_f __pyx_string_tab[8]
#define __pyx_kp_u_function_nvshmem_free_is_not_fou __pyx_string_tab[9]
#define __pyx_kp_u_function_nvshmem_info_get_versio __pyx_string_tab[10]
#define __pyx_kp_u_function_nvshmem_malloc_is_not_f __pyx_string_tab[11]
#define __pyx_kp_u_function_nvshmem_my_pe_is_not_fo __pyx_string_tab[12]
#define __pyx_kp_u_function_nvshmem_n_pes_is_not_fo __pyx_string_tab[13]
#define __pyx_kp_u_function_nvshmem_ptr_is_not_foun __pyx_string_tab[14]
#define __pyx_kp_u_function_nvshmem_team_destroy_is __pyx_string_tab[15]
#define __pyx_kp_u_function_nvshmem_team_get_config __pyx_string_tab[16]
#define __pyx_kp_u_function_nvshmem_team_my_pe_is_n __pyx_string_tab[17]
#define __pyx_kp_u_function_nvshmem_team_n_pes_is_n __pyx_string_tab[18]
#define __pyx_kp_u_function_nvshmem_team_split_2d_i __pyx_string_tab[19]
#define __pyx_kp_u_function_nvshmem_team_split_stri __pyx_string_tab[20]
#define __pyx_kp_u_function_nvshmem_team_translate __pyx_string_tab[21]
#define __pyx_kp_u_function_nvshmemx_barrier_all_on __pyx_string_tab[22]
#define __pyx_kp_u_function_nvshmemx_barrier_on_str __pyx_string_tab[23]
#define __pyx_kp_u_function_nvshmemx_bfloat16_allto __pyx_string_tab[24]
#define __pyx_kp_u_function_nvshmemx_bfloat16_broad __pyx_string_tab[25]
#define __pyx_kp_u_function_nvshmemx_bfloat16_fcoll __pyx_string_tab[26]
#define __pyx_kp_u_function_nvshmemx_bfloat16_max_r __pyx_string_tab[27]
#define __pyx_kp_u_function_nvshmemx_bfloat16_max_r_2 __pyx_string_tab[28]
#define __pyx_kp_u_function_nvshmemx_bfloat16_min_r __pyx_string_tab[29]
#define __pyx_kp_u_function_nvshmemx_bfloat16_min_r_2 __pyx_string_tab[30]
#define __pyx_kp_u_function_nvshmemx_bfloat16_sum_r __pyx_string_tab[31]
#define __pyx_kp_u_function_nvshmemx_bfloat16_sum_r_2 __pyx_string_tab[32]
#define __pyx_kp_u_function_nvshmemx_buffer_registe __pyx_string_tab[33]
#define __pyx_kp_u_function_nvshmemx_buffer_unregis __pyx_string_tab[34]
#define __pyx_kp_u_function_nvshmemx_char_alltoall __pyx_string_tab[35]
#define __pyx_kp_u_function_nvshmemx_char_broadcast __pyx_string_tab[36]
#define __pyx_kp_u_function_nvshmemx_char_fcollect __pyx_string_tab[37]
#define __pyx_kp_u_function_nvshmemx_char_max_reduc __pyx_string_tab[38]
#define __pyx_kp_u_function_nvshmemx_char_max_reduc_2 __pyx_string_tab[39]
#define __pyx_kp_u_function_nvshmemx_char_min_reduc __pyx_string_tab[40]
#define __pyx_kp_u_function_nvshmemx_char_min_reduc_2 __pyx_string_tab[41]
#define __pyx_kp_u_function_nvshmemx_char_sum_reduc __pyx_string_tab[42]
#define __pyx_kp_u_function_nvshmemx_char_sum_reduc_2 __pyx_string_tab[43]
#define __pyx_kp_u_function_nvshmemx_culibrary_fina __pyx_string_tab[44]
#define __pyx_kp_u_function_nvshmemx_culibrary_init __pyx_string_tab[45]
#define __pyx_kp_u_function_nvshmemx_cumodule_final __pyx_string_tab[46]
#define __pyx_kp_u_function_nvshmemx_cumodule_init __pyx_string_tab[47]
#define __pyx_kp_u_function_nvshmemx_double_alltoal __pyx_string_tab[48]
#define __pyx_kp_u_function_nvshmemx_double_broadca __pyx_string_tab[49]
#define __pyx_kp_u_function_nvshmemx_double_fcollec __pyx_string_tab[50]
#define __pyx_kp_u_function_nvshmemx_double_max_red __pyx_string_tab[51]
#define __pyx_kp_u_function_nvshmemx_double_max_red_2 __pyx_string_tab[52]
#define __pyx_kp_u_function_nvshmemx_double_min_red __pyx_string_tab[53]
#define __pyx_kp_u_function_nvshmemx_double_min_red_2 __pyx_string_tab[54]
#define __pyx_kp_u_function_nvshmemx_double_sum_red __pyx_string_tab[55]
#define __pyx_kp_u_function_nvshmemx_double_sum_red_2 __pyx_string_tab[56]
#define __pyx_kp_u_function_nvshmemx_float_alltoall __pyx_string_tab[57]
#define __pyx_kp_u_function_nvshmemx_float_broadcas __pyx_string_tab[58]
#define __pyx_kp_u_function_nvshmemx_float_fcollect __pyx_string_tab[59]
#define __pyx_kp_u_function_nvshmemx_float_max_redu __pyx_string_tab[60]
#define __pyx_kp_u_function_nvshmemx_float_max_redu_2 __pyx_string_tab[61]
#define __pyx_kp_u_function_nvshmemx_float_min_redu __pyx_string_tab[62]
#define __pyx_kp_u_function_nvshmemx_float_min_redu_2 __pyx_string_tab[63]
#define __pyx_kp_u_function_nvshmemx_float_sum_redu __pyx_string_tab[64]
#define __pyx_kp_u_function_nvshmemx_float_sum_redu_2 __pyx_string_tab[65]
#define __pyx_kp_u_function_nvshmemx_get_uniqueid_i __pyx_string_tab[66]
#define __pyx_kp_u_function_nvshmemx_getmem_on_stre __pyx_string_tab[67]
#define __pyx_kp_u_function_nvshmemx_half_alltoall __pyx_string_tab[68]
#define __pyx_kp_u_function_nvshmemx_half_broadcast __pyx_string_tab[69]
#define __pyx_kp_u_function_nvshmemx_half_fcollect __pyx_string_tab[70]
#define __pyx_kp_u_function_nvshmemx_half_max_reduc __pyx_string_tab[71]
#define __pyx_kp_u_function_nvshmemx_half_max_reduc_2 __pyx_string_tab[72]
#define __pyx_kp_u_function_nvshmemx_half_min_reduc __pyx_string_tab[73]
#define __pyx_kp_u_function_nvshmemx_half_min_reduc_2 __pyx_string_tab[74]
#define __pyx_kp_u_function_nvshmemx_half_sum_reduc __pyx_string_tab[75]
#define __pyx_kp_u_function_nvshmemx_half_sum_reduc_2 __pyx_string_tab[76]
#define __pyx_kp_u_function_nvshmemx_hostlib_finali __pyx_string_tab[77]
#define __pyx_kp_u_function_nvshmemx_hostlib_init_a __pyx_string_tab[78]
#define __pyx_kp_u_function_nvshmemx_init_status_is __pyx_string_tab[79]
#define __pyx_kp_u_function_nvshmemx_int16_alltoall __pyx_string_tab[80]
#define __pyx_kp_u_function_nvshmemx_int16_broadcas __pyx_string_tab[81]
#define __pyx_kp_u_function_nvshmemx_int16_fcollect __pyx_string_tab[82]
#define __pyx_kp_u_function_nvshmemx_int16_max_redu __pyx_string_tab[83]
#define __pyx_kp_u_function_nvshmemx_int16_max_redu_2 __pyx_string_tab[84]
#define __pyx_kp_u_function_nvshmemx_int16_min_redu __pyx_string_tab[85]
#define __pyx_kp_u_function_nvshmemx_int16_min_redu_2 __pyx_string_tab[86]
#define __pyx_kp_u_function_nvshmemx_int16_sum_redu __pyx_string_tab[87]
#define __pyx_kp_u_function_nvshmemx_int16_sum_redu_2 __pyx_string_tab[88]
#define __pyx_kp_u_function_nvshmemx_int32_alltoall __pyx_string_tab[89]
#define __pyx_kp_u_function_nvshmemx_int32_broadcas __pyx_string_tab[90]
#define __pyx_kp_u_function_nvshmemx_int32_fcollect __pyx_string_tab[91]
#define __pyx_kp_u_function_nvshmemx_int32_max_redu __pyx_string_tab[92]
#define __pyx_kp_u_function_nvshmemx_int32_max_redu_2 __pyx_string_tab[93]
#define __pyx_kp_u_function_nvshmemx_int32_min_redu __pyx_string_tab[94]
#define __pyx_kp_u_function_nvshmemx_int32_min_redu_2 __pyx_string_tab[95]
#define __pyx_kp_u_function_nvshmemx_int32_sum_redu __pyx_string_tab[96]
#define __pyx_kp_u_function_nvshmemx_int32_sum_redu_2 __pyx_string_tab[97]
#define __pyx_kp_u_function_nvshmemx_int64_alltoall __pyx_string_tab[98]
#define __pyx_kp_u_function_nvshmemx_int64_broadcas __pyx_string_tab[99]
#define __pyx_kp_u_function_nvshmemx_int64_fcollect __pyx_string_tab[100]
#define __pyx_kp_u_function_nvshmemx_int64_max_redu __pyx_string_tab[101]
#define __pyx_kp_u_function_nvshmemx_int64_max_redu_2 __pyx_string_tab[102]
#define __pyx_kp_u_function_nvshmemx_int64_min_redu __pyx_string_tab[103]
#define __pyx_kp_u_function_nvshmemx_int64_min_redu_2 __pyx_string_tab[104]
#define __pyx_kp_u_function_nvshmemx_int64_sum_redu __pyx_string_tab[105]
#define __pyx_kp_u_function_nvshmemx_int64_sum_redu_2 __pyx_string_tab[106]
#define __pyx_kp_u_function_nvshmemx_int8_alltoall __pyx_string_tab[107]
#define __pyx_kp_u_function_nvshmemx_int8_broadcast __pyx_string_tab[108]
#define __pyx_kp_u_function_nvshmemx_int8_fcollect __pyx_string_tab[109]
#define __pyx_kp_u_function_nvshmemx_int8_max_reduc __pyx_string_tab[110]
#define __pyx_kp_u_function_nvshmemx_int8_max_reduc_2 __pyx_string_tab[111]
#define __pyx_kp_u_function_nvshmemx_int8_min_reduc __pyx_string_tab[112]
#define __pyx_kp_u_function_nvshmemx_int8_min_reduc_2 __pyx_string_tab[113]
#define __pyx_kp_u_function_nvshmemx_int8_sum_reduc __pyx_string_tab[114]
#define __pyx_kp_u_function_nvshmemx_int8_sum_reduc_2 __pyx_string_tab[115]
#define __pyx_kp_u_function_nvshmemx_int_alltoall_o __pyx_string_tab[116]
#define __pyx_kp_u_function_nvshmemx_int_broadcast __pyx_string_tab[117]
#define __pyx_kp_u_function_nvshmemx_int_fcollect_o __pyx_string_tab[118]
#define __pyx_kp_u_function_nvshmemx_int_max_reduce __pyx_string_tab[119]
#define __pyx_kp_u_function_nvshmemx_int_max_reduce_2 __pyx_string_tab[120]
#define __pyx_kp_u_function_nvshmemx_int_min_reduce __pyx_string_tab[121]
#define __pyx_kp_u_function_nvshmemx_int_min_reduce_2 __pyx_string_tab[122]
#define __pyx_kp_u_function_nvshmemx_int_sum_reduce __pyx_string_tab[123]
#define __pyx_kp_u_function_nvshmemx_int_sum_reduce_2 __pyx_string_tab[124]
#define __pyx_kp_u_function_nvshmemx_long_alltoall __pyx_string_tab[125]
#define __pyx_kp_u_function_nvshmemx_long_broadcast __pyx_string_tab[126]
#define __pyx_kp_u_function_nvshmemx_long_fcollect __pyx_string_tab[127]
#define __pyx_kp_u_function_nvshmemx_long_max_reduc __pyx_string_tab[128]
#define __pyx_kp_u_function_nvshmemx_long_max_reduc_2 __pyx_string_tab[129]
#define __pyx_kp_u_function_nvshmemx_long_min_reduc __pyx_string_tab[130]
#define __pyx_kp_u_function_nvshmemx_long_min_reduc_2 __pyx_string_tab[131]
#define __pyx_kp_u_function_nvshmemx_long_sum_reduc __pyx_string_tab[132]
#define __pyx_kp_u_function_nvshmemx_long_sum_reduc_2 __pyx_string_tab[133]
#define __pyx_kp_u_function_nvshmemx_longlong_allto __pyx_string_tab[134]
#define __pyx_kp_u_function_nvshmemx_longlong_broad __pyx_string_tab[135]
#define __pyx_kp_u_function_nvshmemx_longlong_fcoll __pyx_string_tab[136]
#define __pyx_kp_u_function_nvshmemx_longlong_max_r __pyx_string_tab[137]
#define __pyx_kp_u_function_nvshmemx_longlong_max_r_2 __pyx_string_tab[138]
#define __pyx_kp_u_function_nvshmemx_longlong_min_r __pyx_string_tab[139]
#define __pyx_kp_u_function_nvshmemx_longlong_min_r_2 __pyx_string_tab[140]
#define __pyx_kp_u_function_nvshmemx_longlong_sum_r __pyx_string_tab[141]
#define __pyx_kp_u_function_nvshmemx_longlong_sum_r_2 __pyx_string_tab[142]
#define __pyx_kp_u_function_nvshmemx_mc_ptr_is_not __pyx_string_tab[143]
#define __pyx_kp_u_function_nvshmemx_putmem_on_stre __pyx_string_tab[144]
#define __pyx_kp_u_function_nvshmemx_putmem_signal __pyx_string_tab[145]
#define __pyx_kp_u_function_nvshmemx_quiet_on_strea __pyx_string_tab[146]
#define __pyx_kp_u_function_nvshmemx_schar_alltoall __pyx_string_tab[147]
#define __pyx_kp_u_function_nvshmemx_schar_broadcas __pyx_string_tab[148]
#define __pyx_kp_u_function_nvshmemx_schar_fcollect __pyx_string_tab[149]
#define __pyx_kp_u_function_nvshmemx_schar_max_redu __pyx_string_tab[150]
#define __pyx_kp_u_function_nvshmemx_schar_max_redu_2 __pyx_string_tab[151]
#define __pyx_kp_u_function_nvshmemx_schar_min_redu __pyx_string_tab[152]
#define __pyx_kp_u_function_nvshmemx_schar_min_redu_2 __pyx_string_tab[153]
#define __pyx_kp_u_function_nvshmemx_schar_sum_redu __pyx_string_tab[154]
#define __pyx_kp_u_function_nvshmemx_schar_sum_redu_2 __pyx_string_tab[155]
#define __pyx_kp_u_function_nvshmemx_set_attr_mpi_c __pyx_string_tab[156]
#define __pyx_kp_u_function_nvshmemx_set_attr_uniqu __pyx_string_tab[157]
#define __pyx_kp_u_function_nvshmemx_short_alltoall __pyx_string_tab[158]
#define __pyx_kp_u_function_nvshmemx_short_broadcas __pyx_string_tab[159]
#define __pyx_kp_u_function_nvshmemx_short_fcollect __pyx_string_tab[160]
#define __pyx_kp_u_function_nvshmemx_short_max_redu __pyx_string_tab[161]
#define __pyx_kp_u_function_nvshmemx_short_max_redu_2 __pyx_string_tab[162]
#define __pyx_kp_u_function_nvshmemx_short_min_redu __pyx_string_tab[163]
#define __pyx_kp_u_function_nvshmemx_short_min_redu_2 __pyx_string_tab[164]
#define __pyx_kp_u_function_nvshmemx_short_sum_redu __pyx_string_tab[165]
#define __pyx_kp_u_function_nvshmemx_short_sum_redu_2 __pyx_string_tab[166]
#define __pyx_kp_u_function_nvshmemx_signal_op_on_s __pyx_string_tab[167]
#define __pyx_kp_u_function_nvshmemx_signal_wait_un __pyx_string_tab[168]
#define __pyx_kp_u_function_nvshmemx_size_alltoall __pyx_string_tab[169]
#define __pyx_kp_u_function_nvshmemx_size_broadcast __pyx_string_tab[170]
#define __pyx_kp_u_function_nvshmemx_size_fcollect __pyx_string_tab[171]
#define __pyx_kp_u_function_nvshmemx_size_max_reduc __pyx_string_tab[172]
#define __pyx_kp_u_function_nvshmemx_size_max_reduc_2 __pyx_string_tab[173]
#define __pyx_kp_u_function_nvshmemx_size_min_reduc __pyx_string_tab[174]
#define __pyx_kp_u_function_nvshmemx_size_min_reduc_2 __pyx_string_tab[175]
#define __pyx_kp_u_function_nvshmemx_size_sum_reduc __pyx_string_tab[176]
#define __pyx_kp_u_function_nvshmemx_size_sum_reduc_2 __pyx_string_tab[177]
#define __pyx_kp_u_function_nvshmemx_sync_all_on_st __pyx_string_tab[178]
#define __pyx_kp_u_function_nvshmemx_team_get_uniqu __pyx_string_tab[179]
#define __pyx_kp_u_function_nvshmemx_team_init_is_n __pyx_string_tab[180]
#define __pyx_kp_u_function_nvshmemx_team_sync_on_s __pyx_string_tab[181]
#define __pyx_kp_u_function_nvshmemx_uint16_alltoal __pyx_string_tab[182]
#define __pyx_kp_u_function_nvshmemx_uint16_broadca __pyx_string_tab[183]
#define __pyx_kp_u_function_nvshmemx_uint16_fcollec __pyx_string_tab[184]
#define __pyx_kp_u_function_nvshmemx_uint16_max_red __pyx_string_tab[185]
#define __pyx_kp_u_function_nvshmemx_uint16_max_red_2 __pyx_string_tab[186]
#define __pyx_kp_u_function_nvshmemx_uint16_min_red __pyx_string_tab[187]
#define __pyx_kp_u_function_nvshmemx_uint16_min_red_2 __pyx_string_tab[188]
#define __pyx_kp_u_function_nvshmemx_uint16_sum_red __pyx_string_tab[189]
#define __pyx_kp_u_function_nvshmemx_uint16_sum_red_2 __pyx_string_tab[190]
#define __pyx_kp_u_function_nvshmemx_uint32_alltoal __pyx_string_tab[191]
#define __pyx_kp_u_function_nvshmemx_uint32_broadca __pyx_string_tab[192]
#define __pyx_kp_u_function_nvshmemx_uint32_fcollec __pyx_string_tab[193]
#define __pyx_kp_u_function_nvshmemx_uint32_max_red __pyx_string_tab[194]
#define __pyx_kp_u_function_nvshmemx_uint32_max_red_2 __pyx_string_tab[195]
#define __pyx_kp_u_function_nvshmemx_uint32_min_red __pyx_string_tab[196]
#define __pyx_kp_u_function_nvshmemx_uint32_min_red_2 __pyx_string_tab[197]
#define __pyx_kp_u_function_nvshmemx_uint32_sum_red __pyx_string_tab[198]
#define __pyx_kp_u_function_nvshmemx_uint32_sum_red_2 __pyx_string_tab[199]
#define __pyx_kp_u_function_nvshmemx_uint64_alltoal __pyx_string_tab[200]
#define __pyx_kp_u_function_nvshmemx_uint64_broadca __pyx_string_tab[201]
#define __pyx_kp_u_function_nvshmemx_uint64_fcollec __pyx_string_tab[202]
#define __pyx_kp_u_function_nvshmemx_uint64_max_red __pyx_string_tab[203]
#define __pyx_kp_u_function_nvshmemx_uint64_max_red_2 __pyx_string_tab[204]
#define __pyx_kp_u_function_nvshmemx_uint64_min_red __pyx_string_tab[205]
#define __pyx_kp_u_function_nvshmemx_uint64_min_red_2 __pyx_string_tab[206]
#define __pyx_kp_u_function_nvshmemx_uint64_sum_red __pyx_string_tab[207]
#define __pyx_kp_u_function_nvshmemx_uint64_sum_red_2 __pyx_string_tab[208]
#define __pyx_kp_u_function_nvshmemx_uint8_alltoall __pyx_string_tab[209]
#define __pyx_kp_u_function_nvshmemx_uint8_broadcas __pyx_string_tab[210]
#define __pyx_kp_u_function_nvshmemx_uint8_fcollect __pyx_string_tab[211]
#define __pyx_kp_u_function_nvshmemx_uint8_max_redu __pyx_string_tab[212]
#define __pyx_kp_u_function_nvshmemx_uint8_max_redu_2 __pyx_string_tab[213]
#define __pyx_kp_u_function_nvshmemx_uint8_min_redu __pyx_string_tab[214]
#define __pyx_kp_u_function_nvshmemx_uint8_min_redu_2 __pyx_string_tab[215]
#define __pyx_kp_u_function_nvshmemx_uint8_sum_redu __pyx_string_tab[216]
#define __pyx_kp_u_function_nvshmemx_uint8_sum_redu_2 __pyx_string_tab[217]
#define __pyx_kp_u_function_nvshmemx_vendor_get_ver __pyx_string_tab[218]
#define __pyx_kp_u_nvshmem_bindings__internal_nvshm_2 __pyx_string_tab[219]
#define __pyx_n_u_FunctionNotFoundError __pyx_string_tab[220]
#define __pyx_n_u_NotSupportedError __pyx_string_tab[221]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[222]
#define __pyx_n_u_annotate __pyx_string_tab[223]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[224]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[225]
#define __pyx_n_u_doc __pyx_string_tab[226]
#define __pyx_n_u_func __pyx_string_tab[227]
#define __pyx_n_u_inspect_function_pointer __pyx_string_tab[228]
#define __pyx_n_u_inspect_function_pointers __pyx_string_tab[229]
#define __pyx_n_u_is_coroutine __pyx_string_tab[230]
#define __pyx_n_u_items __pyx_string_tab[231]
#define __pyx_n_u_main __pyx_string_tab[232]
#define __pyx_n_u_metaclass __pyx_string_tab[233]
#define __pyx_n_u_module __pyx_string_tab[234]
#define __pyx_n_u_mro_entries __pyx_string_tab[235]
#define __pyx_n_u_name __pyx_string_tab[236]
#define __pyx_n_u_name_2 __pyx_string_tab[237]
#define __pyx_n_u_nvshmem_align_2 __pyx_string_tab[238]
#define __pyx_n_u_nvshmem_barrier_2 __pyx_string_tab[239]
#define __pyx_n_u_nvshmem_barrier_all_2 __pyx_string_tab[240]
#define __pyx_n_u_nvshmem_bindings__internal_nvshm __pyx_string_tab[241]
#define __pyx_n_u_nvshmem_calloc_2 __pyx_string_tab[242]
#define __pyx_n_u_nvshmem_free_2 __pyx_string_tab[243]
#define __pyx_n_u_nvshmem_info_get_version_2 __pyx_string_tab[244]
#define __pyx_n_u_nvshmem_malloc_2 __pyx_string_tab[245]
#define __pyx_n_u_nvshmem_my_pe_2 __pyx_string_tab[246]
#define __pyx_n_u_nvshmem_n_pes_2 __pyx_string_tab[247]
#define __pyx_n_u_nvshmem_ptr_2 __pyx_string_tab[248]
#define __pyx_n_u_nvshmem_team_destroy_2 __pyx_string_tab[249]
#define __pyx_n_u_nvshmem_team_get_config_2 __pyx_string_tab[250]
#define __pyx_n_u_nvshmem_team_my_pe_2 __pyx_string_tab[251]
#define __pyx_n_u_nvshmem_team_n_pes_2 __pyx_string_tab[252]
#define __pyx_n_u_nvshmem_team_split_2d_2 __pyx_string_tab[253]
#define __pyx_n_u_nvshmem_team_split_strided_2 __pyx_string_tab[254]
#define __pyx_n_u_nvshmem_team_translate_pe_2 __pyx_string_tab[255]
#define __pyx_n_u_nvshmemx_barrier_all_on_stream_2 __pyx_string_tab[256]
#define __pyx_n_u_nvshmemx_barrier_on_stream_2 __pyx_string_tab[257]
#define __pyx_n_u_nvshmemx_bfloat16_alltoall_on __pyx_string_tab[258]
#define __pyx_n_u_nvshmemx_bfloat16_broadcast_on __pyx_string_tab[259]
#define __pyx_n_u_nvshmemx_bfloat16_fcollect_on __pyx_string_tab[260]
#define __pyx_n_u_nvshmemx_bfloat16_max_reduce_o __pyx_string_tab[261]
#define __pyx_n_u_nvshmemx_bfloat16_max_reducesc __pyx_string_tab[262]
#define __pyx_n_u_nvshmemx_bfloat16_min_reduce_o __pyx_string_tab[263]
#define __pyx_n_u_nvshmemx_bfloat16_min_reducesc __pyx_string_tab[264]
#define __pyx_n_u_nvshmemx_bfloat16_sum_reduce_o __pyx_string_tab[265]
#define __pyx_n_u_nvshmemx_bfloat16_sum_reducesc __pyx_string_tab[266]
#define __pyx_n_u_nvshmemx_buffer_register_symme __pyx_string_tab[267]
#define __pyx_n_u_nvshmemx_buffer_unregister_sym __pyx_string_tab[268]
#define __pyx_n_u_nvshmemx_char_alltoall_on_stre __pyx_string_tab[269]
#define __pyx_n_u_nvshmemx_char_broadcast_on_str __pyx_string_tab[270]
#define __pyx_n_u_nvshmemx_char_fcollect_on_stre __pyx_string_tab[271]
#define __pyx_n_u_nvshmemx_char_max_reduce_on_st __pyx_string_tab[272]
#define __pyx_n_u_nvshmemx_char_max_reducescatte __pyx_string_tab[273]
#define __pyx_n_u_nvshmemx_char_min_reduce_on_st __pyx_string_tab[274]
#define __pyx_n_u_nvshmemx_char_min_reducescatte __pyx_string_tab[275]
#define __pyx_n_u_nvshmemx_char_sum_reduce_on_st __pyx_string_tab[276]
#define __pyx_n_u_nvshmemx_char_sum_reducescatte __pyx_string_tab[277]
#define __pyx_n_u_nvshmemx_culibrary_finalize_2 __pyx_string_tab[278]
#define __pyx_n_u_nvshmemx_culibrary_init_2 __pyx_string_tab[279]
#define __pyx_n_u_nvshmemx_cumodule_finalize_2 __pyx_string_tab[280]
#define __pyx_n_u_nvshmemx_cumodule_init_2 __pyx_string_tab[281]
#define __pyx_n_u_nvshmemx_double_alltoall_on_st __pyx_string_tab[282]
#define __pyx_n_u_nvshmemx_double_broadcast_on_s __pyx_string_tab[283]
#define __pyx_n_u_nvshmemx_double_fcollect_on_st __pyx_string_tab[284]
#define __pyx_n_u_nvshmemx_double_max_reduce_on __pyx_string_tab[285]
#define __pyx_n_u_nvshmemx_double_max_reducescat __pyx_string_tab[286]
#define __pyx_n_u_nvshmemx_double_min_reduce_on __pyx_string_tab[287]
#define __pyx_n_u_nvshmemx_double_min_reducescat __pyx_string_tab[288]
#define __pyx_n_u_nvshmemx_double_sum_reduce_on __pyx_string_tab[289]
#define __pyx_n_u_nvshmemx_double_sum_reducescat __pyx_string_tab[290]
#define __pyx_n_u_nvshmemx_float_alltoall_on_str __pyx_string_tab[291]
#define __pyx_n_u_nvshmemx_float_broadcast_on_st __pyx_string_tab[292]
#define __pyx_n_u_nvshmemx_float_fcollect_on_str __pyx_string_tab[293]
#define __pyx_n_u_nvshmemx_float_max_reduce_on_s __pyx_string_tab[294]
#define __pyx_n_u_nvshmemx_float_max_reducescatt __pyx_string_tab[295]
#define __pyx_n_u_nvshmemx_float_min_reduce_on_s __pyx_string_tab[296]
#define __pyx_n_u_nvshmemx_float_min_reducescatt __pyx_string_tab[297]
#define __pyx_n_u_nvshmemx_float_sum_reduce_on_s __pyx_string_tab[298]
#define __pyx_n_u_nvshmemx_float_sum_reducescatt __pyx_string_tab[299]
#define __pyx_n_u_nvshmemx_get_uniqueid_2 __pyx_string_tab[300]
#define __pyx_n_u_nvshmemx_getmem_on_stream_2 __pyx_string_tab[301]
#define __pyx_n_u_nvshmemx_half_alltoall_on_stre __pyx_string_tab[302]
#define __pyx_n_u_nvshmemx_half_broadcast_on_str __pyx_string_tab[303]
#define __pyx_n_u_nvshmemx_half_fcollect_on_stre __pyx_string_tab[304]
#define __pyx_n_u_nvshmemx_half_max_reduce_on_st __pyx_string_tab[305]
#define __pyx_n_u_nvshmemx_half_max_reducescatte __pyx_string_tab[306]
#define __pyx_n_u_nvshmemx_half_min_reduce_on_st __pyx_string_tab[307]
#define __pyx_n_u_nvshmemx_half_min_reducescatte __pyx_string_tab[308]
#define __pyx_n_u_nvshmemx_half_sum_reduce_on_st __pyx_string_tab[309]
#define __pyx_n_u_nvshmemx_half_sum_reducescatte __pyx_string_tab[310]
#define __pyx_n_u_nvshmemx_hostlib_finalize_2 __pyx_string_tab[311]
#define __pyx_n_u_nvshmemx_hostlib_init_attr_2 __pyx_string_tab[312]
#define __pyx_n_u_nvshmemx_init_status_2 __pyx_string_tab[313]
#define __pyx_n_u_nvshmemx_int16_alltoall_on_str __pyx_string_tab[314]
#define __pyx_n_u_nvshmemx_int16_broadcast_on_st __pyx_string_tab[315]
#define __pyx_n_u_nvshmemx_int16_fcollect_on_str __pyx_string_tab[316]
#define __pyx_n_u_nvshmemx_int16_max_reduce_on_s __pyx_string_tab[317]
#define __pyx_n_u_nvshmemx_int16_max_reducescatt __pyx_string_tab[318]
#define __pyx_n_u_nvshmemx_int16_min_reduce_on_s __pyx_string_tab[319]
#define __pyx_n_u_nvshmemx_int16_min_reducescatt __pyx_string_tab[320]
#define __pyx_n_u_nvshmemx_int16_sum_reduce_on_s __pyx_string_tab[321]
#define __pyx_n_u_nvshmemx_int16_sum_reducescatt __pyx_string_tab[322]
#define __pyx_n_u_nvshmemx_int32_alltoall_on_str __pyx_string_tab[323]
#define __pyx_n_u_nvshmemx_int32_broadcast_on_st __pyx_string_tab[324]
#define __pyx_n_u_nvshmemx_int32_fcollect_on_str __pyx_string_tab[325]
#define __pyx_n_u_nvshmemx_int32_max_reduce_on_s __pyx_string_tab[326]
#define __pyx_n_u_nvshmemx_int32_max_reducescatt __pyx_string_tab[327]
#define __pyx_n_u_nvshmemx_int32_min_reduce_on_s __pyx_string_tab[328]
#define __pyx_n_u_nvshmemx_int32_min_reducescatt __pyx_string_tab[329]
#define __pyx_n_u_nvshmemx_int32_sum_reduce_on_s __pyx_string_tab[330]
#define __pyx_n_u_nvshmemx_int32_sum_reducescatt __pyx_string_tab[331]
#define __pyx_n_u_nvshmemx_int64_alltoall_on_str __pyx_string_tab[332]
#define __pyx_n_u_nvshmemx_int64_broadcast_on_st __pyx_string_tab[333]
#define __pyx_n_u_nvshmemx_int64_fcollect_on_str __pyx_string_tab[334]
#define __pyx_n_u_nvshmemx_int64_max_reduce_on_s __pyx_string_tab[335]
#define __pyx_n_u_nvshmemx_int64_max_reducescatt __pyx_string_tab[336]
#define __pyx_n_u_nvshmemx_int64_min_reduce_on_s __pyx_string_tab[337]
#define __pyx_n_u_nvshmemx_int64_min_reducescatt __pyx_string_tab[338]
#define __pyx_n_u_nvshmemx_int64_sum_reduce_on_s __pyx_string_tab[339]
#define __pyx_n_u_nvshmemx_int64_sum_reducescatt __pyx_string_tab[340]
#define __pyx_n_u_nvshmemx_int8_alltoall_on_stre __pyx_string_tab[341]
#define __pyx_n_u_nvshmemx_int8_broadcast_on_str __pyx_string_tab[342]
#define __pyx_n_u_nvshmemx_int8_fcollect_on_stre __pyx_string_tab[343]
#define __pyx_n_u_nvshmemx_int8_max_reduce_on_st __pyx_string_tab[344]
#define __pyx_n_u_nvshmemx_int8_max_reducescatte __pyx_string_tab[345]
#define __pyx_n_u_nvshmemx_int8_min_reduce_on_st __pyx_string_tab[346]
#define __pyx_n_u_nvshmemx_int8_min_reducescatte __pyx_string_tab[347]
#define __pyx_n_u_nvshmemx_int8_sum_reduce_on_st __pyx_string_tab[348]
#define __pyx_n_u_nvshmemx_int8_sum_reducescatte __pyx_string_tab[349]
#define __pyx_n_u_nvshmemx_int_alltoall_on_strea __pyx_string_tab[350]
#define __pyx_n_u_nvshmemx_int_broadcast_on_stre __pyx_string_tab[351]
#define __pyx_n_u_nvshmemx_int_fcollect_on_strea __pyx_string_tab[352]
#define __pyx_n_u_nvshmemx_int_max_reduce_on_str __pyx_string_tab[353]
#define __pyx_n_u_nvshmemx_int_max_reducescatter __pyx_string_tab[354]
#define __pyx_n_u_nvshmemx_int_min_reduce_on_str __pyx_string_tab[355]
#define __pyx_n_u_nvshmemx_int_min_reducescatter __pyx_string_tab[356]
#define __pyx_n_u_nvshmemx_int_sum_reduce_on_str __pyx_string_tab[357]
#define __pyx_n_u_nvshmemx_int_sum_reducescatter __pyx_string_tab[358]
#define __pyx_n_u_nvshmemx_long_alltoall_on_stre __pyx_string_tab[359]
#define __pyx_n_u_nvshmemx_long_broadcast_on_str __pyx_string_tab[360]
#define __pyx_n_u_nvshmemx_long_fcollect_on_stre __pyx_string_tab[361]
#define __pyx_n_u_nvshmemx_long_max_reduce_on_st __pyx_string_tab[362]
#define __pyx_n_u_nvshmemx_long_max_reducescatte __pyx_string_tab[363]
#define __pyx_n_u_nvshmemx_long_min_reduce_on_st __pyx_string_tab[364]
#define __pyx_n_u_nvshmemx_long_min_reducescatte __pyx_string_tab[365]
#define __pyx_n_u_nvshmemx_long_sum_reduce_on_st __pyx_string_tab[366]
#define __pyx_n_u_nvshmemx_long_sum_reducescatte __pyx_string_tab[367]
#define __pyx_n_u_nvshmemx_longlong_alltoall_on __pyx_string_tab[368]
#define __pyx_n_u_nvshmemx_longlong_broadcast_on __pyx_string_tab[369]
#define __pyx_n_u_nvshmemx_longlong_fcollect_on __pyx_string_tab[370]
#define __pyx_n_u_nvshmemx_longlong_max_reduce_o __pyx_string_tab[371]
#define __pyx_n_u_nvshmemx_longlong_max_reducesc __pyx_string_tab[372]
#define __pyx_n_u_nvshmemx_longlong_min_reduce_o __pyx_string_tab[373]
#define __pyx_n_u_nvshmemx_longlong_min_reducesc __pyx_string_tab[374]
#define __pyx_n_u_nvshmemx_longlong_sum_reduce_o __pyx_string_tab[375]
#define __pyx_n_u_nvshmemx_longlong_sum_reducesc __pyx_string_tab[376]
#define __pyx_n_u_nvshmemx_mc_ptr_2 __pyx_string_tab[377]
#define __pyx_n_u_nvshmemx_putmem_on_stream_2 __pyx_string_tab[378]
#define __pyx_n_u_nvshmemx_putmem_signal_on_stre __pyx_string_tab[379]
#define __pyx_n_u_nvshmemx_quiet_on_stream_2 __pyx_string_tab[380]
#define __pyx_n_u_nvshmemx_schar_alltoall_on_str __pyx_string_tab[381]
#define __pyx_n_u_nvshmemx_schar_broadcast_on_st __pyx_string_tab[382]
#define __pyx_n_u_nvshmemx_schar_fcollect_on_str __pyx_string_tab[383]
#define __pyx_n_u_nvshmemx_schar_max_reduce_on_s __pyx_string_tab[384]
#define __pyx_n_u_nvshmemx_schar_max_reducescatt __pyx_string_tab[385]
#define __pyx_n_u_nvshmemx_schar_min_reduce_on_s __pyx_string_tab[386]
#define __pyx_n_u_nvshmemx_schar_min_reducescatt __pyx_string_tab[387]
#define __pyx_n_u_nvshmemx_schar_sum_reduce_on_s __pyx_string_tab[388]
#define __pyx_n_u_nvshmemx_schar_sum_reducescatt __pyx_string_tab[389]
#define __pyx_n_u_nvshmemx_set_attr_mpi_comm_arg __pyx_string_tab[390]
#define __pyx_n_u_nvshmemx_set_attr_uniqueid_arg __pyx_string_tab[391]
#define __pyx_n_u_nvshmemx_short_alltoall_on_str __pyx_string_tab[392]
#define __pyx_n_u_nvshmemx_short_broadcast_on_st __pyx_string_tab[393]
#define __pyx_n_u_nvshmemx_short_fcollect_on_str __pyx_string_tab[394]
#define __pyx_n_u_nvshmemx_short_max_reduce_on_s __pyx_string_tab[395]
#define __pyx_n_u_nvshmemx_short_max_reducescatt __pyx_string_tab[396]
#define __pyx_n_u_nvshmemx_short_min_reduce_on_s __pyx_string_tab[397]
#define __pyx_n_u_nvshmemx_short_min_reducescatt __pyx_string_tab[398]
#define __pyx_n_u_nvshmemx_short_sum_reduce_on_s __pyx_string_tab[399]
#define __pyx_n_u_nvshmemx_short_sum_reducescatt __pyx_string_tab[400]
#define __pyx_n_u_nvshmemx_signal_op_on_stream_2 __pyx_string_tab[401]
#define __pyx_n_u_nvshmemx_signal_wait_until_on __pyx_string_tab[402]
#define __pyx_n_u_nvshmemx_size_alltoall_on_stre __pyx_string_tab[403]
#define __pyx_n_u_nvshmemx_size_broadcast_on_str __pyx_string_tab[404]
#define __pyx_n_u_nvshmemx_size_fcollect_on_stre __pyx_string_tab[405]
#define __pyx_n_u_nvshmemx_size_max_reduce_on_st __pyx_string_tab[406]
#define __pyx_n_u_nvshmemx_size_max_reducescatte __pyx_string_tab[407]
#define __pyx_n_u_nvshmemx_size_min_reduce_on_st __pyx_string_tab[408]
#define __pyx_n_u_nvshmemx_size_min_reducescatte __pyx_string_tab[409]
#define __pyx_n_u_nvshmemx_size_sum_reduce_on_st __pyx_string_tab[410]
#define __pyx_n_u_nvshmemx_size_sum_reducescatte __pyx_string_tab[411]
#define __pyx_n_u_nvshmemx_sync_all_on_stream_2 __pyx_string_tab[412]
#define __pyx_n_u_nvshmemx_team_get_uniqueid_2 __pyx_string_tab[413]
#define __pyx_n_u_nvshmemx_team_init_2 __pyx_string_tab[414]
#define __pyx_n_u_nvshmemx_team_sync_on_stream_2 __pyx_string_tab[415]
#define __pyx_n_u_nvshmemx_uint16_alltoall_on_st __pyx_string_tab[416]
#define __pyx_n_u_nvshmemx_uint16_broadcast_on_s __pyx_string_tab[417]
#define __pyx_n_u_nvshmemx_uint16_fcollect_on_st __pyx_string_tab[418]
#define __pyx_n_u_nvshmemx_uint16_max_reduce_on __pyx_string_tab[419]
#define __pyx_n_u_nvshmemx_uint16_max_reducescat __pyx_string_tab[420]
#define __pyx_n_u_nvshmemx_uint16_min_reduce_on __pyx_string_tab[421]
#define __pyx_n_u_nvshmemx_uint16_min_reducescat __pyx_string_tab[422]
#define __pyx_n_u_nvshmemx_uint16_sum_reduce_on __pyx_string_tab[423]
#define __pyx_n_u_nvshmemx_uint16_sum_reducescat __pyx_string_tab[424]
#define __pyx_n_u_nvshmemx_uint32_alltoall_on_st __pyx_string_tab[425]
#define __pyx_n_u_nvshmemx_uint32_broadcast_on_s __pyx_string_tab[426]
#define __pyx_n_u_nvshmemx_uint32_fcollect_on_st __pyx_string_tab[427]
#define __pyx_n_u_nvshmemx_uint32_max_reduce_on __pyx_string_tab[428]
#define __pyx_n_u_nvshmemx_uint32_max_reducescat __pyx_string_tab[429]
#define __pyx_n_u_nvshmemx_uint32_min_reduce_on __pyx_string_tab[430]
#define __pyx_n_u_nvshmemx_uint32_min_reducescat __pyx_string_tab[431]
#define __pyx_n_u_nvshmemx_uint32_sum_reduce_on __pyx_string_tab[432]
#define __pyx_n_u_nvshmemx_uint32_sum_reducescat __pyx_string_tab[433]
#define __pyx_n_u_nvshmemx_uint64_alltoall_on_st __pyx_string_tab[434]
#define __pyx_n_u_nvshmemx_uint64_broadcast_on_s __pyx_string_tab[435]
#define __pyx_n_u_nvshmemx_uint64_fcollect_on_st __pyx_string_tab[436]
#define __pyx_n_u_nvshmemx_uint64_max_reduce_on __pyx_string_tab[437]
#define __pyx_n_u_nvshmemx_uint64_max_reducescat __pyx_string_tab[438]
#define __pyx_n_u_nvshmemx_uint64_min_reduce_on __pyx_string_tab[439]
#define __pyx_n_u_nvshmemx_uint64_min_reducescat __pyx_string_tab[440]
#define __pyx_n_u_nvshmemx_uint64_sum_reduce_on __pyx_string_tab[441]
#define __pyx_n_u_nvshmemx_uint64_sum_reducescat __pyx_string_tab[442]
#define __pyx_n_u_nvshmemx_uint8_alltoall_on_str __pyx_string_tab[443]
#define __pyx_n_u_nvshmemx_uint8_broadcast_on_st __pyx_string_tab[444]
#define __pyx_n_u_nvshmemx_uint8_fcollect_on_str __pyx_string_tab[445]
#define __pyx_n_u_nvshmemx_uint8_max_reduce_on_s __pyx_string_tab[446]
#define __pyx_n_u_nvshmemx_uint8_max_reducescatt __pyx_string_tab[447]
#define __pyx_n_u_nvshmemx_uint8_min_reduce_on_s __pyx_string_tab[448]
#define __pyx_n_u_nvshmemx_uint8_min_reducescatt __pyx_string_tab[449]
#define __pyx_n_u_nvshmemx_uint8_sum_reduce_on_s __pyx_string_tab[450]
#define __pyx_n_u_nvshmemx_uint8_sum_reducescatt __pyx_string_tab[451]
#define __pyx_n_u_nvshmemx_vendor_get_version_in __pyx_string_tab[452]
#define __pyx_n_u_pop __pyx_string_tab[453]
#define __pyx_n_u_prepare __pyx_string_tab[454]
#define __pyx_n_u_pyx_capi __pyx_string_tab[455]
#define __pyx_n_u_qualname __pyx_string_tab[456]
#define __pyx_n_u_set_name __pyx_string_tab[457]
#define __pyx_n_u_setdefault __pyx_string_tab[458]
#define __pyx_n_u_test __pyx_string_tab[459]
#define __pyx_n_u_values __pyx_string_tab[460]
#define __pyx_kp_b_int_CUlibrary_int_CUmodule_int_i __pyx_string_tab[461]
#define __pyx_kp_b_iso88591_z_A_a_9AQ __pyx_string_tab[462]
#define __pyx_kp_b_iso88591_z_q_Q_Ja_Zq_j_j_1_1_1_z_z_j_Zq __pyx_string_tab[463]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  for (int i=0; i<2; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<464; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  for (int i=0; i<2; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<464; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "nvshmem/bindings/_internal/nvshmem.pyx":259
 * 
 * 
 * cdef void* load_library() except* nogil:             # <<<<<<<<<<<<<<
 *     cdef void* handle
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(void) {
  void *__pyx_v_handle;
  char *__pyx_v_err_msg;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[3];
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("load_library", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":261
 * cdef void* load_library() except* nogil:
 *     cdef void* handle
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)             # <<<<<<<<<<<<<<
 *     if handle == NULL:
 *         with gil:
*/
  __pyx_v_handle = dlopen(__pyx_k_libnvshmem_host_so_3, (RTLD_NOW | RTLD_GLOBAL));

  /* "nvshmem/bindings/_internal/nvshmem.pyx":262
 *     cdef void* handle
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
 *     if handle == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             err_msg = dlerror()
*/
  __pyx_t_1 = (__pyx_v_handle == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":263
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
 *     if handle == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             err_msg = dlerror()
 *             raise RuntimeError(f'Failed to dlopen libnvshmem ({err_msg.decode()})')
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":264
 *     if handle == NULL:
 *         with gil:
 *             err_msg = dlerror()             # <<<<<<<<<<<<<<
 *             raise RuntimeError(f'Failed to dlopen libnvshmem ({err_msg.decode()})')
 *     return handle
*/
          __pyx_v_err_msg = dlerror();

          /* "nvshmem/bindings/_internal/nvshmem.pyx":265
 *         with gil:
 *             err_msg = dlerror()
 *             raise RuntimeError(f'Failed to dlopen libnvshmem ({err_msg.decode()})')             # <<<<<<<<<<<<<<
 *     return handle
 * 
*/
          __pyx_t_3 = NULL;
          __pyx_t_4 = __Pyx_ssize_strlen(__pyx_v_err_msg); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 265, __pyx_L5_error)
          __pyx_t_5 = __Pyx_decode_c_string(__pyx_v_err_msg, 0, __pyx_t_4, NULL, NULL, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 265, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_Failed_to_dlopen_libnvshmem;
          __pyx_t_6[1] = __pyx_t_5;
          __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_;
          __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, 29 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 265, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_8 = 1;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
            __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 265, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":263
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
 *     if handle == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             err_msg = dlerror()
 *             raise RuntimeError(f'Failed to dlopen libnvshmem ({err_msg.decode()})')
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":262
 *     cdef void* handle
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
 *     if handle == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             err_msg = dlerror()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":266
 *             err_msg = dlerror()
 *             raise RuntimeError(f'Failed to dlopen libnvshmem ({err_msg.decode()})')
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":259
 * 
 * 
 * cdef void* load_library() except* nogil:             # <<<<<<<<<<<<<<
 *     cdef void* handle
 *     handle = dlopen("libnvshmem_host.so.3", RTLD_NOW | RTLD_GLOBAL)
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem.load_library", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":269
 * 
 * 
 * cdef int _check_or_init_nvshmem() except -1 nogil:             # <<<<<<<<<<<<<<
 *     global __py_nvshmem_init
 *     if __py_nvshmem_init:
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(void) {
  void *__pyx_v_handle;
  int __pyx_r;
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":271
 * cdef int _check_or_init_nvshmem() except -1 nogil:
 *     global __py_nvshmem_init
 *     if __py_nvshmem_init:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  if (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___py_nvshmem_init) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":272
 *     global __py_nvshmem_init
 *     if __py_nvshmem_init:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     # Load function
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "nvshmem/bindings/_internal/nvshmem.pyx":271
 * cdef int _check_or_init_nvshmem() except -1 nogil:
 *     global __py_nvshmem_init
 *     if __py_nvshmem_init:             # <<<<<<<<<<<<<<
 *         return 0
 * 
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":275
 * 
 *     # Load function
 *     cdef void* handle = NULL             # <<<<<<<<<<<<<<
 *     global __nvshmem_barrier
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')
*/
  __pyx_v_handle = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":277
 *     cdef void* handle = NULL
 *     global __nvshmem_barrier
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')             # <<<<<<<<<<<<<<
 *     if __nvshmem_barrier == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_barrier);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":278
 *     global __nvshmem_barrier
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')
 *     if __nvshmem_barrier == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":279
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')
 *     if __nvshmem_barrier == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_barrier = dlsym(handle, 'nvshmem_barrier')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":280
 *     if __nvshmem_barrier == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_barrier = dlsym(handle, 'nvshmem_barrier')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 280, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":279
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')
 *     if __nvshmem_barrier == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_barrier = dlsym(handle, 'nvshmem_barrier')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":281
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_barrier = dlsym(handle, 'nvshmem_barrier')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_barrier_all
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier = dlsym(__pyx_v_handle, __pyx_k_nvshmem_barrier);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":278
 *     global __nvshmem_barrier
 *     __nvshmem_barrier = dlsym(RTLD_DEFAULT, 'nvshmem_barrier')
 *     if __nvshmem_barrier == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":284
 * 
 *     global __nvshmem_barrier_all
 *     __nvshmem_barrier_all = dlsym(RTLD_DEFAULT, 'nvshmem_barrier_all')             # <<<<<<<<<<<<<<
 *     if __nvshmem_barrier_all == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_barrier_all);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":285
 *     global __nvshmem_barrier_all
 *     __nvshmem_barrier_all = dlsym(RTLD_DEFAULT, 'nvshmem_barrier_all')
 *     if __nvshmem_barrier_all == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":286
 *     __nvshmem_barrier_all = dlsym(RTLD_DEFAULT, 'nvshmem_barrier_all')
 *     if __nvshmem_barrier_all == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_barrier_all = dlsym(handle, 'nvshmem_barrier_all')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":287
 *     if __nvshmem_barrier_all == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_barrier_all = dlsym(handle, 'nvshmem_barrier_all')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 287, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":286
 *     __nvshmem_barrier_all = dlsym(RTLD_DEFAULT, 'nvshmem_barrier_all')
 *     if __nvshmem_barrier_all == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_barrier_all = dlsym(handle, 'nvshmem_barrier_all')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":288
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_barrier_all = dlsym(handle, 'nvshmem_barrier_all')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_init_status
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all = dlsym(__pyx_v_handle, __pyx_k_nvshmem_barrier_all);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":285
 *     global __nvshmem_barrier_all
 *     __nvshmem_barrier_all = dlsym(RTLD_DEFAULT, 'nvshmem_barrier_all')
 *     if __nvshmem_barrier_all == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":291
 * 
 *     global __nvshmemx_init_status
 *     __nvshmemx_init_status = dlsym(RTLD_DEFAULT, 'nvshmemx_init_status')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_init_status == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_init_status);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":292
 *     global __nvshmemx_init_status
 *     __nvshmemx_init_status = dlsym(RTLD_DEFAULT, 'nvshmemx_init_status')
 *     if __nvshmemx_init_status == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":293
 *     __nvshmemx_init_status = dlsym(RTLD_DEFAULT, 'nvshmemx_init_status')
 *     if __nvshmemx_init_status == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_init_status = dlsym(handle, 'nvshmemx_init_status')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":294
 *     if __nvshmemx_init_status == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_init_status = dlsym(handle, 'nvshmemx_init_status')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 294, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":293
 *     __nvshmemx_init_status = dlsym(RTLD_DEFAULT, 'nvshmemx_init_status')
 *     if __nvshmemx_init_status == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_init_status = dlsym(handle, 'nvshmemx_init_status')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":295
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_init_status = dlsym(handle, 'nvshmemx_init_status')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_my_pe
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_init_status);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":292
 *     global __nvshmemx_init_status
 *     __nvshmemx_init_status = dlsym(RTLD_DEFAULT, 'nvshmemx_init_status')
 *     if __nvshmemx_init_status == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":298
 * 
 *     global __nvshmem_my_pe
 *     __nvshmem_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_my_pe')             # <<<<<<<<<<<<<<
 *     if __nvshmem_my_pe == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_my_pe);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":299
 *     global __nvshmem_my_pe
 *     __nvshmem_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_my_pe')
 *     if __nvshmem_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":300
 *     __nvshmem_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_my_pe')
 *     if __nvshmem_my_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_my_pe = dlsym(handle, 'nvshmem_my_pe')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":301
 *     if __nvshmem_my_pe == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_my_pe = dlsym(handle, 'nvshmem_my_pe')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 301, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":300
 *     __nvshmem_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_my_pe')
 *     if __nvshmem_my_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_my_pe = dlsym(handle, 'nvshmem_my_pe')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":302
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_my_pe = dlsym(handle, 'nvshmem_my_pe')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_n_pes
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe = dlsym(__pyx_v_handle, __pyx_k_nvshmem_my_pe);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":299
 *     global __nvshmem_my_pe
 *     __nvshmem_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_my_pe')
 *     if __nvshmem_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":305
 * 
 *     global __nvshmem_n_pes
 *     __nvshmem_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_n_pes')             # <<<<<<<<<<<<<<
 *     if __nvshmem_n_pes == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_n_pes);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":306
 *     global __nvshmem_n_pes
 *     __nvshmem_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_n_pes')
 *     if __nvshmem_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":307
 *     __nvshmem_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_n_pes')
 *     if __nvshmem_n_pes == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_n_pes = dlsym(handle, 'nvshmem_n_pes')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":308
 *     if __nvshmem_n_pes == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_n_pes = dlsym(handle, 'nvshmem_n_pes')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 308, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":307
 *     __nvshmem_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_n_pes')
 *     if __nvshmem_n_pes == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_n_pes = dlsym(handle, 'nvshmem_n_pes')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":309
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_n_pes = dlsym(handle, 'nvshmem_n_pes')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_info_get_version
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes = dlsym(__pyx_v_handle, __pyx_k_nvshmem_n_pes);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":306
 *     global __nvshmem_n_pes
 *     __nvshmem_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_n_pes')
 *     if __nvshmem_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":312
 * 
 *     global __nvshmem_info_get_version
 *     __nvshmem_info_get_version = dlsym(RTLD_DEFAULT, 'nvshmem_info_get_version')             # <<<<<<<<<<<<<<
 *     if __nvshmem_info_get_version == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_info_get_version);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":313
 *     global __nvshmem_info_get_version
 *     __nvshmem_info_get_version = dlsym(RTLD_DEFAULT, 'nvshmem_info_get_version')
 *     if __nvshmem_info_get_version == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":314
 *     __nvshmem_info_get_version = dlsym(RTLD_DEFAULT, 'nvshmem_info_get_version')
 *     if __nvshmem_info_get_version == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_info_get_version = dlsym(handle, 'nvshmem_info_get_version')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":315
 *     if __nvshmem_info_get_version == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_info_get_version = dlsym(handle, 'nvshmem_info_get_version')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 315, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":314
 *     __nvshmem_info_get_version = dlsym(RTLD_DEFAULT, 'nvshmem_info_get_version')
 *     if __nvshmem_info_get_version == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_info_get_version = dlsym(handle, 'nvshmem_info_get_version')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":316
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_info_get_version = dlsym(handle, 'nvshmem_info_get_version')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_vendor_get_version_info
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version = dlsym(__pyx_v_handle, __pyx_k_nvshmem_info_get_version);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":313
 *     global __nvshmem_info_get_version
 *     __nvshmem_info_get_version = dlsym(RTLD_DEFAULT, 'nvshmem_info_get_version')
 *     if __nvshmem_info_get_version == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":319
 * 
 *     global __nvshmemx_vendor_get_version_info
 *     __nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, 'nvshmemx_vendor_get_version_info')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_vendor_get_version_info);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":320
 *     global __nvshmemx_vendor_get_version_info
 *     __nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, 'nvshmemx_vendor_get_version_info')
 *     if __nvshmemx_vendor_get_version_info == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":321
 *     __nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, 'nvshmemx_vendor_get_version_info')
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_vendor_get_version_info = dlsym(handle, 'nvshmemx_vendor_get_version_info')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":322
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_vendor_get_version_info = dlsym(handle, 'nvshmemx_vendor_get_version_info')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 322, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":321
 *     __nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, 'nvshmemx_vendor_get_version_info')
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_vendor_get_version_info = dlsym(handle, 'nvshmemx_vendor_get_version_info')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":323
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_vendor_get_version_info = dlsym(handle, 'nvshmemx_vendor_get_version_info')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_malloc
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_vendor_get_version_info);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":320
 *     global __nvshmemx_vendor_get_version_info
 *     __nvshmemx_vendor_get_version_info = dlsym(RTLD_DEFAULT, 'nvshmemx_vendor_get_version_info')
 *     if __nvshmemx_vendor_get_version_info == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":326
 * 
 *     global __nvshmem_malloc
 *     __nvshmem_malloc = dlsym(RTLD_DEFAULT, 'nvshmem_malloc')             # <<<<<<<<<<<<<<
 *     if __nvshmem_malloc == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_malloc);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":327
 *     global __nvshmem_malloc
 *     __nvshmem_malloc = dlsym(RTLD_DEFAULT, 'nvshmem_malloc')
 *     if __nvshmem_malloc == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":328
 *     __nvshmem_malloc = dlsym(RTLD_DEFAULT, 'nvshmem_malloc')
 *     if __nvshmem_malloc == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_malloc = dlsym(handle, 'nvshmem_malloc')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":329
 *     if __nvshmem_malloc == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_malloc = dlsym(handle, 'nvshmem_malloc')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 329, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":328
 *     __nvshmem_malloc = dlsym(RTLD_DEFAULT, 'nvshmem_malloc')
 *     if __nvshmem_malloc == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_malloc = dlsym(handle, 'nvshmem_malloc')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":330
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_malloc = dlsym(handle, 'nvshmem_malloc')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_calloc
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc = dlsym(__pyx_v_handle, __pyx_k_nvshmem_malloc);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":327
 *     global __nvshmem_malloc
 *     __nvshmem_malloc = dlsym(RTLD_DEFAULT, 'nvshmem_malloc')
 *     if __nvshmem_malloc == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":333
 * 
 *     global __nvshmem_calloc
 *     __nvshmem_calloc = dlsym(RTLD_DEFAULT, 'nvshmem_calloc')             # <<<<<<<<<<<<<<
 *     if __nvshmem_calloc == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_calloc);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":334
 *     global __nvshmem_calloc
 *     __nvshmem_calloc = dlsym(RTLD_DEFAULT, 'nvshmem_calloc')
 *     if __nvshmem_calloc == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":335
 *     __nvshmem_calloc = dlsym(RTLD_DEFAULT, 'nvshmem_calloc')
 *     if __nvshmem_calloc == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_calloc = dlsym(handle, 'nvshmem_calloc')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":336
 *     if __nvshmem_calloc == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_calloc = dlsym(handle, 'nvshmem_calloc')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 336, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":335
 *     __nvshmem_calloc = dlsym(RTLD_DEFAULT, 'nvshmem_calloc')
 *     if __nvshmem_calloc == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_calloc = dlsym(handle, 'nvshmem_calloc')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":337
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_calloc = dlsym(handle, 'nvshmem_calloc')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_align
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc = dlsym(__pyx_v_handle, __pyx_k_nvshmem_calloc);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":334
 *     global __nvshmem_calloc
 *     __nvshmem_calloc = dlsym(RTLD_DEFAULT, 'nvshmem_calloc')
 *     if __nvshmem_calloc == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":340
 * 
 *     global __nvshmem_align
 *     __nvshmem_align = dlsym(RTLD_DEFAULT, 'nvshmem_align')             # <<<<<<<<<<<<<<
 *     if __nvshmem_align == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_align);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":341
 *     global __nvshmem_align
 *     __nvshmem_align = dlsym(RTLD_DEFAULT, 'nvshmem_align')
 *     if __nvshmem_align == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":342
 *     __nvshmem_align = dlsym(RTLD_DEFAULT, 'nvshmem_align')
 *     if __nvshmem_align == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_align = dlsym(handle, 'nvshmem_align')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":343
 *     if __nvshmem_align == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_align = dlsym(handle, 'nvshmem_align')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 343, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":342
 *     __nvshmem_align = dlsym(RTLD_DEFAULT, 'nvshmem_align')
 *     if __nvshmem_align == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_align = dlsym(handle, 'nvshmem_align')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":344
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_align = dlsym(handle, 'nvshmem_align')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_free
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align = dlsym(__pyx_v_handle, __pyx_k_nvshmem_align);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":341
 *     global __nvshmem_align
 *     __nvshmem_align = dlsym(RTLD_DEFAULT, 'nvshmem_align')
 *     if __nvshmem_align == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":347
 * 
 *     global __nvshmem_free
 *     __nvshmem_free = dlsym(RTLD_DEFAULT, 'nvshmem_free')             # <<<<<<<<<<<<<<
 *     if __nvshmem_free == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_free);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":348
 *     global __nvshmem_free
 *     __nvshmem_free = dlsym(RTLD_DEFAULT, 'nvshmem_free')
 *     if __nvshmem_free == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":349
 *     __nvshmem_free = dlsym(RTLD_DEFAULT, 'nvshmem_free')
 *     if __nvshmem_free == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_free = dlsym(handle, 'nvshmem_free')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":350
 *     if __nvshmem_free == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_free = dlsym(handle, 'nvshmem_free')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 350, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":349
 *     __nvshmem_free = dlsym(RTLD_DEFAULT, 'nvshmem_free')
 *     if __nvshmem_free == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_free = dlsym(handle, 'nvshmem_free')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":351
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_free = dlsym(handle, 'nvshmem_free')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_ptr
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free = dlsym(__pyx_v_handle, __pyx_k_nvshmem_free);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":348
 *     global __nvshmem_free
 *     __nvshmem_free = dlsym(RTLD_DEFAULT, 'nvshmem_free')
 *     if __nvshmem_free == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":354
 * 
 *     global __nvshmem_ptr
 *     __nvshmem_ptr = dlsym(RTLD_DEFAULT, 'nvshmem_ptr')             # <<<<<<<<<<<<<<
 *     if __nvshmem_ptr == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_ptr);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":355
 *     global __nvshmem_ptr
 *     __nvshmem_ptr = dlsym(RTLD_DEFAULT, 'nvshmem_ptr')
 *     if __nvshmem_ptr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":356
 *     __nvshmem_ptr = dlsym(RTLD_DEFAULT, 'nvshmem_ptr')
 *     if __nvshmem_ptr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_ptr = dlsym(handle, 'nvshmem_ptr')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":357
 *     if __nvshmem_ptr == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_ptr = dlsym(handle, 'nvshmem_ptr')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 357, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":356
 *     __nvshmem_ptr = dlsym(RTLD_DEFAULT, 'nvshmem_ptr')
 *     if __nvshmem_ptr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_ptr = dlsym(handle, 'nvshmem_ptr')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":358
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_ptr = dlsym(handle, 'nvshmem_ptr')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_mc_ptr
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr = dlsym(__pyx_v_handle, __pyx_k_nvshmem_ptr);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":355
 *     global __nvshmem_ptr
 *     __nvshmem_ptr = dlsym(RTLD_DEFAULT, 'nvshmem_ptr')
 *     if __nvshmem_ptr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":361
 * 
 *     global __nvshmemx_mc_ptr
 *     __nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, 'nvshmemx_mc_ptr')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_mc_ptr == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_mc_ptr);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":362
 *     global __nvshmemx_mc_ptr
 *     __nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, 'nvshmemx_mc_ptr')
 *     if __nvshmemx_mc_ptr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":363
 *     __nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, 'nvshmemx_mc_ptr')
 *     if __nvshmemx_mc_ptr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_mc_ptr = dlsym(handle, 'nvshmemx_mc_ptr')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":364
 *     if __nvshmemx_mc_ptr == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_mc_ptr = dlsym(handle, 'nvshmemx_mc_ptr')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 364, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":363
 *     __nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, 'nvshmemx_mc_ptr')
 *     if __nvshmemx_mc_ptr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_mc_ptr = dlsym(handle, 'nvshmemx_mc_ptr')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":365
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_mc_ptr = dlsym(handle, 'nvshmemx_mc_ptr')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_my_pe
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_mc_ptr);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":362
 *     global __nvshmemx_mc_ptr
 *     __nvshmemx_mc_ptr = dlsym(RTLD_DEFAULT, 'nvshmemx_mc_ptr')
 *     if __nvshmemx_mc_ptr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":368
 * 
 *     global __nvshmem_team_my_pe
 *     __nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_my_pe')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_my_pe == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_my_pe);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":369
 *     global __nvshmem_team_my_pe
 *     __nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_my_pe')
 *     if __nvshmem_team_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":370
 *     __nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_my_pe')
 *     if __nvshmem_team_my_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_my_pe = dlsym(handle, 'nvshmem_team_my_pe')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":371
 *     if __nvshmem_team_my_pe == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_my_pe = dlsym(handle, 'nvshmem_team_my_pe')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 371, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":370
 *     __nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_my_pe')
 *     if __nvshmem_team_my_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_my_pe = dlsym(handle, 'nvshmem_team_my_pe')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":372
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_my_pe = dlsym(handle, 'nvshmem_team_my_pe')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_n_pes
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_my_pe);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":369
 *     global __nvshmem_team_my_pe
 *     __nvshmem_team_my_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_my_pe')
 *     if __nvshmem_team_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":375
 * 
 *     global __nvshmem_team_n_pes
 *     __nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_team_n_pes')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_n_pes == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_n_pes);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":376
 *     global __nvshmem_team_n_pes
 *     __nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_team_n_pes')
 *     if __nvshmem_team_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":377
 *     __nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_team_n_pes')
 *     if __nvshmem_team_n_pes == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_n_pes = dlsym(handle, 'nvshmem_team_n_pes')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":378
 *     if __nvshmem_team_n_pes == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_n_pes = dlsym(handle, 'nvshmem_team_n_pes')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 378, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":377
 *     __nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_team_n_pes')
 *     if __nvshmem_team_n_pes == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_n_pes = dlsym(handle, 'nvshmem_team_n_pes')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":379
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_n_pes = dlsym(handle, 'nvshmem_team_n_pes')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_get_config
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_n_pes);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":376
 *     global __nvshmem_team_n_pes
 *     __nvshmem_team_n_pes = dlsym(RTLD_DEFAULT, 'nvshmem_team_n_pes')
 *     if __nvshmem_team_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":382
 * 
 *     global __nvshmem_team_get_config
 *     __nvshmem_team_get_config = dlsym(RTLD_DEFAULT, 'nvshmem_team_get_config')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_get_config == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_get_config);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":383
 *     global __nvshmem_team_get_config
 *     __nvshmem_team_get_config = dlsym(RTLD_DEFAULT, 'nvshmem_team_get_config')
 *     if __nvshmem_team_get_config == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":384
 *     __nvshmem_team_get_config = dlsym(RTLD_DEFAULT, 'nvshmem_team_get_config')
 *     if __nvshmem_team_get_config == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_get_config = dlsym(handle, 'nvshmem_team_get_config')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":385
 *     if __nvshmem_team_get_config == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_get_config = dlsym(handle, 'nvshmem_team_get_config')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 385, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":384
 *     __nvshmem_team_get_config = dlsym(RTLD_DEFAULT, 'nvshmem_team_get_config')
 *     if __nvshmem_team_get_config == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_get_config = dlsym(handle, 'nvshmem_team_get_config')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":386
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_get_config = dlsym(handle, 'nvshmem_team_get_config')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_translate_pe
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_get_config);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":383
 *     global __nvshmem_team_get_config
 *     __nvshmem_team_get_config = dlsym(RTLD_DEFAULT, 'nvshmem_team_get_config')
 *     if __nvshmem_team_get_config == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":389
 * 
 *     global __nvshmem_team_translate_pe
 *     __nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_translate_pe')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_translate_pe == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_translate_pe);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":390
 *     global __nvshmem_team_translate_pe
 *     __nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_translate_pe')
 *     if __nvshmem_team_translate_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":391
 *     __nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_translate_pe')
 *     if __nvshmem_team_translate_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_translate_pe = dlsym(handle, 'nvshmem_team_translate_pe')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":392
 *     if __nvshmem_team_translate_pe == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_translate_pe = dlsym(handle, 'nvshmem_team_translate_pe')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 392, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":391
 *     __nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_translate_pe')
 *     if __nvshmem_team_translate_pe == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_translate_pe = dlsym(handle, 'nvshmem_team_translate_pe')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":393
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_translate_pe = dlsym(handle, 'nvshmem_team_translate_pe')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_split_strided
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_translate_pe);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":390
 *     global __nvshmem_team_translate_pe
 *     __nvshmem_team_translate_pe = dlsym(RTLD_DEFAULT, 'nvshmem_team_translate_pe')
 *     if __nvshmem_team_translate_pe == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":396
 * 
 *     global __nvshmem_team_split_strided
 *     __nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_strided')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_split_strided == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_split_strided);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":397
 *     global __nvshmem_team_split_strided
 *     __nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_strided')
 *     if __nvshmem_team_split_strided == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":398
 *     __nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_strided')
 *     if __nvshmem_team_split_strided == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_split_strided = dlsym(handle, 'nvshmem_team_split_strided')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":399
 *     if __nvshmem_team_split_strided == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_split_strided = dlsym(handle, 'nvshmem_team_split_strided')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 399, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":398
 *     __nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_strided')
 *     if __nvshmem_team_split_strided == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_split_strided = dlsym(handle, 'nvshmem_team_split_strided')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":400
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_split_strided = dlsym(handle, 'nvshmem_team_split_strided')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_get_uniqueid
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_split_strided);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":397
 *     global __nvshmem_team_split_strided
 *     __nvshmem_team_split_strided = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_strided')
 *     if __nvshmem_team_split_strided == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":403
 * 
 *     global __nvshmemx_team_get_uniqueid
 *     __nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_team_get_uniqueid')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_team_get_uniqueid);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":404
 *     global __nvshmemx_team_get_uniqueid
 *     __nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_team_get_uniqueid')
 *     if __nvshmemx_team_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":405
 *     __nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_team_get_uniqueid')
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_get_uniqueid = dlsym(handle, 'nvshmemx_team_get_uniqueid')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":406
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_team_get_uniqueid = dlsym(handle, 'nvshmemx_team_get_uniqueid')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 406, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":405
 *     __nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_team_get_uniqueid')
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_get_uniqueid = dlsym(handle, 'nvshmemx_team_get_uniqueid')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":407
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_team_get_uniqueid = dlsym(handle, 'nvshmemx_team_get_uniqueid')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_init
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_team_get_uniqueid);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":404
 *     global __nvshmemx_team_get_uniqueid
 *     __nvshmemx_team_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_team_get_uniqueid')
 *     if __nvshmemx_team_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":410
 * 
 *     global __nvshmemx_team_init
 *     __nvshmemx_team_init = dlsym(RTLD_DEFAULT, 'nvshmemx_team_init')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_init == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_team_init);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":411
 *     global __nvshmemx_team_init
 *     __nvshmemx_team_init = dlsym(RTLD_DEFAULT, 'nvshmemx_team_init')
 *     if __nvshmemx_team_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":412
 *     __nvshmemx_team_init = dlsym(RTLD_DEFAULT, 'nvshmemx_team_init')
 *     if __nvshmemx_team_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_init = dlsym(handle, 'nvshmemx_team_init')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":413
 *     if __nvshmemx_team_init == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_team_init = dlsym(handle, 'nvshmemx_team_init')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 413, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":412
 *     __nvshmemx_team_init = dlsym(RTLD_DEFAULT, 'nvshmemx_team_init')
 *     if __nvshmemx_team_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_init = dlsym(handle, 'nvshmemx_team_init')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":414
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_team_init = dlsym(handle, 'nvshmemx_team_init')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_split_2d
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_team_init);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":411
 *     global __nvshmemx_team_init
 *     __nvshmemx_team_init = dlsym(RTLD_DEFAULT, 'nvshmemx_team_init')
 *     if __nvshmemx_team_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":417
 * 
 *     global __nvshmem_team_split_2d
 *     __nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_2d')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_split_2d == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_split_2d);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":418
 *     global __nvshmem_team_split_2d
 *     __nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_2d')
 *     if __nvshmem_team_split_2d == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":419
 *     __nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_2d')
 *     if __nvshmem_team_split_2d == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_split_2d = dlsym(handle, 'nvshmem_team_split_2d')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":420
 *     if __nvshmem_team_split_2d == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_split_2d = dlsym(handle, 'nvshmem_team_split_2d')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 420, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":419
 *     __nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_2d')
 *     if __nvshmem_team_split_2d == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_split_2d = dlsym(handle, 'nvshmem_team_split_2d')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":421
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_split_2d = dlsym(handle, 'nvshmem_team_split_2d')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_destroy
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_split_2d);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":418
 *     global __nvshmem_team_split_2d
 *     __nvshmem_team_split_2d = dlsym(RTLD_DEFAULT, 'nvshmem_team_split_2d')
 *     if __nvshmem_team_split_2d == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":424
 * 
 *     global __nvshmem_team_destroy
 *     __nvshmem_team_destroy = dlsym(RTLD_DEFAULT, 'nvshmem_team_destroy')             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_destroy == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy = dlsym(RTLD_DEFAULT, __pyx_k_nvshmem_team_destroy);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":425
 *     global __nvshmem_team_destroy
 *     __nvshmem_team_destroy = dlsym(RTLD_DEFAULT, 'nvshmem_team_destroy')
 *     if __nvshmem_team_destroy == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":426
 *     __nvshmem_team_destroy = dlsym(RTLD_DEFAULT, 'nvshmem_team_destroy')
 *     if __nvshmem_team_destroy == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_destroy = dlsym(handle, 'nvshmem_team_destroy')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":427
 *     if __nvshmem_team_destroy == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmem_team_destroy = dlsym(handle, 'nvshmem_team_destroy')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 427, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":426
 *     __nvshmem_team_destroy = dlsym(RTLD_DEFAULT, 'nvshmem_team_destroy')
 *     if __nvshmem_team_destroy == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmem_team_destroy = dlsym(handle, 'nvshmem_team_destroy')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":428
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmem_team_destroy = dlsym(handle, 'nvshmem_team_destroy')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy = dlsym(__pyx_v_handle, __pyx_k_nvshmem_team_destroy);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":425
 *     global __nvshmem_team_destroy
 *     __nvshmem_team_destroy = dlsym(RTLD_DEFAULT, 'nvshmem_team_destroy')
 *     if __nvshmem_team_destroy == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":431
 * 
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     __nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_alltoall_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":432
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     __nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_alltoall_on_stream')
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":433
 *     __nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_alltoall_on_stream')
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_alltoall_on_stream = dlsym(handle, 'nvshmemx_bfloat16_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":434
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_alltoall_on_stream = dlsym(handle, 'nvshmemx_bfloat16_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 434, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":433
 *     __nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_alltoall_on_stream')
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_alltoall_on_stream = dlsym(handle, 'nvshmemx_bfloat16_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":435
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_alltoall_on_stream = dlsym(handle, 'nvshmemx_bfloat16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_alltoall_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":432
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     __nvshmemx_bfloat16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_alltoall_on_stream')
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":438
 * 
 *     global __nvshmemx_half_alltoall_on_stream
 *     __nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":439
 *     global __nvshmemx_half_alltoall_on_stream
 *     __nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_alltoall_on_stream')
 *     if __nvshmemx_half_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":440
 *     __nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_alltoall_on_stream')
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_alltoall_on_stream = dlsym(handle, 'nvshmemx_half_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":441
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_alltoall_on_stream = dlsym(handle, 'nvshmemx_half_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 441, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":440
 *     __nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_alltoall_on_stream')
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_alltoall_on_stream = dlsym(handle, 'nvshmemx_half_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":442
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_alltoall_on_stream = dlsym(handle, 'nvshmemx_half_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":439
 *     global __nvshmemx_half_alltoall_on_stream
 *     __nvshmemx_half_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_alltoall_on_stream')
 *     if __nvshmemx_half_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":445
 * 
 *     global __nvshmemx_float_alltoall_on_stream
 *     __nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":446
 *     global __nvshmemx_float_alltoall_on_stream
 *     __nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_alltoall_on_stream')
 *     if __nvshmemx_float_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":447
 *     __nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_alltoall_on_stream')
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_alltoall_on_stream = dlsym(handle, 'nvshmemx_float_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":448
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_alltoall_on_stream = dlsym(handle, 'nvshmemx_float_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 448, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":447
 *     __nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_alltoall_on_stream')
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_alltoall_on_stream = dlsym(handle, 'nvshmemx_float_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":449
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_alltoall_on_stream = dlsym(handle, 'nvshmemx_float_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":446
 *     global __nvshmemx_float_alltoall_on_stream
 *     __nvshmemx_float_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_alltoall_on_stream')
 *     if __nvshmemx_float_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":452
 * 
 *     global __nvshmemx_double_alltoall_on_stream
 *     __nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_alltoall_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":453
 *     global __nvshmemx_double_alltoall_on_stream
 *     __nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_alltoall_on_stream')
 *     if __nvshmemx_double_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":454
 *     __nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_alltoall_on_stream')
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_alltoall_on_stream = dlsym(handle, 'nvshmemx_double_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":455
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_alltoall_on_stream = dlsym(handle, 'nvshmemx_double_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 455, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":454
 *     __nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_alltoall_on_stream')
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_alltoall_on_stream = dlsym(handle, 'nvshmemx_double_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":456
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_alltoall_on_stream = dlsym(handle, 'nvshmemx_double_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_alltoall_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":453
 *     global __nvshmemx_double_alltoall_on_stream
 *     __nvshmemx_double_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_alltoall_on_stream')
 *     if __nvshmemx_double_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":459
 * 
 *     global __nvshmemx_char_alltoall_on_stream
 *     __nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":460
 *     global __nvshmemx_char_alltoall_on_stream
 *     __nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_alltoall_on_stream')
 *     if __nvshmemx_char_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":461
 *     __nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_alltoall_on_stream')
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_alltoall_on_stream = dlsym(handle, 'nvshmemx_char_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":462
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_alltoall_on_stream = dlsym(handle, 'nvshmemx_char_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 462, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":461
 *     __nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_alltoall_on_stream')
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_alltoall_on_stream = dlsym(handle, 'nvshmemx_char_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":463
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_alltoall_on_stream = dlsym(handle, 'nvshmemx_char_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":460
 *     global __nvshmemx_char_alltoall_on_stream
 *     __nvshmemx_char_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_alltoall_on_stream')
 *     if __nvshmemx_char_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":466
 * 
 *     global __nvshmemx_short_alltoall_on_stream
 *     __nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":467
 *     global __nvshmemx_short_alltoall_on_stream
 *     __nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_alltoall_on_stream')
 *     if __nvshmemx_short_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":468
 *     __nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_alltoall_on_stream')
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_alltoall_on_stream = dlsym(handle, 'nvshmemx_short_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":469
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_alltoall_on_stream = dlsym(handle, 'nvshmemx_short_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 469, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":468
 *     __nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_alltoall_on_stream')
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_alltoall_on_stream = dlsym(handle, 'nvshmemx_short_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":470
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_alltoall_on_stream = dlsym(handle, 'nvshmemx_short_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":467
 *     global __nvshmemx_short_alltoall_on_stream
 *     __nvshmemx_short_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_alltoall_on_stream')
 *     if __nvshmemx_short_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":473
 * 
 *     global __nvshmemx_schar_alltoall_on_stream
 *     __nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":474
 *     global __nvshmemx_schar_alltoall_on_stream
 *     __nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_alltoall_on_stream')
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":475
 *     __nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_alltoall_on_stream')
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_alltoall_on_stream = dlsym(handle, 'nvshmemx_schar_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":476
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_alltoall_on_stream = dlsym(handle, 'nvshmemx_schar_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 476, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":475
 *     __nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_alltoall_on_stream')
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_alltoall_on_stream = dlsym(handle, 'nvshmemx_schar_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":477
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_alltoall_on_stream = dlsym(handle, 'nvshmemx_schar_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":474
 *     global __nvshmemx_schar_alltoall_on_stream
 *     __nvshmemx_schar_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_alltoall_on_stream')
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":480
 * 
 *     global __nvshmemx_int_alltoall_on_stream
 *     __nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":481
 *     global __nvshmemx_int_alltoall_on_stream
 *     __nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_alltoall_on_stream')
 *     if __nvshmemx_int_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":482
 *     __nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_alltoall_on_stream')
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_alltoall_on_stream = dlsym(handle, 'nvshmemx_int_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":483
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_alltoall_on_stream = dlsym(handle, 'nvshmemx_int_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 483, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":482
 *     __nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_alltoall_on_stream')
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_alltoall_on_stream = dlsym(handle, 'nvshmemx_int_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":484
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_alltoall_on_stream = dlsym(handle, 'nvshmemx_int_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":481
 *     global __nvshmemx_int_alltoall_on_stream
 *     __nvshmemx_int_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_alltoall_on_stream')
 *     if __nvshmemx_int_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":487
 * 
 *     global __nvshmemx_long_alltoall_on_stream
 *     __nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":488
 *     global __nvshmemx_long_alltoall_on_stream
 *     __nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_alltoall_on_stream')
 *     if __nvshmemx_long_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":489
 *     __nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_alltoall_on_stream')
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_alltoall_on_stream = dlsym(handle, 'nvshmemx_long_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":490
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_alltoall_on_stream = dlsym(handle, 'nvshmemx_long_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 490, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":489
 *     __nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_alltoall_on_stream')
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_alltoall_on_stream = dlsym(handle, 'nvshmemx_long_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":491
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_alltoall_on_stream = dlsym(handle, 'nvshmemx_long_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":488
 *     global __nvshmemx_long_alltoall_on_stream
 *     __nvshmemx_long_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_alltoall_on_stream')
 *     if __nvshmemx_long_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":494
 * 
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     __nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_alltoall_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":495
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     __nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_alltoall_on_stream')
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":496
 *     __nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_alltoall_on_stream')
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_alltoall_on_stream = dlsym(handle, 'nvshmemx_longlong_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":497
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_alltoall_on_stream = dlsym(handle, 'nvshmemx_longlong_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 497, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":496
 *     __nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_alltoall_on_stream')
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_alltoall_on_stream = dlsym(handle, 'nvshmemx_longlong_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":498
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_alltoall_on_stream = dlsym(handle, 'nvshmemx_longlong_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_alltoall_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":495
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     __nvshmemx_longlong_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_alltoall_on_stream')
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":501
 * 
 *     global __nvshmemx_int8_alltoall_on_stream
 *     __nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":502
 *     global __nvshmemx_int8_alltoall_on_stream
 *     __nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_alltoall_on_stream')
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":503
 *     __nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_alltoall_on_stream')
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_alltoall_on_stream = dlsym(handle, 'nvshmemx_int8_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":504
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_alltoall_on_stream = dlsym(handle, 'nvshmemx_int8_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 504, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":503
 *     __nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_alltoall_on_stream')
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_alltoall_on_stream = dlsym(handle, 'nvshmemx_int8_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":505
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_alltoall_on_stream = dlsym(handle, 'nvshmemx_int8_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":502
 *     global __nvshmemx_int8_alltoall_on_stream
 *     __nvshmemx_int8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_alltoall_on_stream')
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":508
 * 
 *     global __nvshmemx_int16_alltoall_on_stream
 *     __nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":509
 *     global __nvshmemx_int16_alltoall_on_stream
 *     __nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_alltoall_on_stream')
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":510
 *     __nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_alltoall_on_stream')
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_alltoall_on_stream = dlsym(handle, 'nvshmemx_int16_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":511
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_alltoall_on_stream = dlsym(handle, 'nvshmemx_int16_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 511, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":510
 *     __nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_alltoall_on_stream')
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_alltoall_on_stream = dlsym(handle, 'nvshmemx_int16_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":512
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_alltoall_on_stream = dlsym(handle, 'nvshmemx_int16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":509
 *     global __nvshmemx_int16_alltoall_on_stream
 *     __nvshmemx_int16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_alltoall_on_stream')
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":515
 * 
 *     global __nvshmemx_int32_alltoall_on_stream
 *     __nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":516
 *     global __nvshmemx_int32_alltoall_on_stream
 *     __nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_alltoall_on_stream')
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":517
 *     __nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_alltoall_on_stream')
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_alltoall_on_stream = dlsym(handle, 'nvshmemx_int32_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":518
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_alltoall_on_stream = dlsym(handle, 'nvshmemx_int32_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 518, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":517
 *     __nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_alltoall_on_stream')
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_alltoall_on_stream = dlsym(handle, 'nvshmemx_int32_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":519
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_alltoall_on_stream = dlsym(handle, 'nvshmemx_int32_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":516
 *     global __nvshmemx_int32_alltoall_on_stream
 *     __nvshmemx_int32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_alltoall_on_stream')
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":522
 * 
 *     global __nvshmemx_int64_alltoall_on_stream
 *     __nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":523
 *     global __nvshmemx_int64_alltoall_on_stream
 *     __nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_alltoall_on_stream')
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":524
 *     __nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_alltoall_on_stream')
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_alltoall_on_stream = dlsym(handle, 'nvshmemx_int64_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":525
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_alltoall_on_stream = dlsym(handle, 'nvshmemx_int64_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 525, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":524
 *     __nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_alltoall_on_stream')
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_alltoall_on_stream = dlsym(handle, 'nvshmemx_int64_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":526
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_alltoall_on_stream = dlsym(handle, 'nvshmemx_int64_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":523
 *     global __nvshmemx_int64_alltoall_on_stream
 *     __nvshmemx_int64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_alltoall_on_stream')
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":529
 * 
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     __nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_alltoall_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":530
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     __nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_alltoall_on_stream')
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":531
 *     __nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_alltoall_on_stream')
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint8_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":532
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint8_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 532, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":531
 *     __nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_alltoall_on_stream')
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint8_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":533
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint8_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_alltoall_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":530
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     __nvshmemx_uint8_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_alltoall_on_stream')
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":536
 * 
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     __nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_alltoall_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":537
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     __nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_alltoall_on_stream')
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":538
 *     __nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_alltoall_on_stream')
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint16_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":539
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint16_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 539, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":538
 *     __nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_alltoall_on_stream')
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint16_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":540
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint16_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_alltoall_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":537
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     __nvshmemx_uint16_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_alltoall_on_stream')
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":543
 * 
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     __nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_alltoall_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":544
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     __nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_alltoall_on_stream')
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":545
 *     __nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_alltoall_on_stream')
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint32_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":546
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint32_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 546, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":545
 *     __nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_alltoall_on_stream')
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint32_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":547
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint32_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_alltoall_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":544
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     __nvshmemx_uint32_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_alltoall_on_stream')
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":550
 * 
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     __nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_alltoall_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":551
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     __nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_alltoall_on_stream')
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":552
 *     __nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_alltoall_on_stream')
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint64_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":553
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint64_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 553, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":552
 *     __nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_alltoall_on_stream')
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint64_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":554
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_alltoall_on_stream = dlsym(handle, 'nvshmemx_uint64_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_alltoall_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_alltoall_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":551
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     __nvshmemx_uint64_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_alltoall_on_stream')
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":557
 * 
 *     global __nvshmemx_size_alltoall_on_stream
 *     __nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_alltoall_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_alltoall_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":558
 *     global __nvshmemx_size_alltoall_on_stream
 *     __nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_alltoall_on_stream')
 *     if __nvshmemx_size_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":559
 *     __nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_alltoall_on_stream')
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_alltoall_on_stream = dlsym(handle, 'nvshmemx_size_alltoall_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":560
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_alltoall_on_stream = dlsym(handle, 'nvshmemx_size_alltoall_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 560, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":559
 *     __nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_alltoall_on_stream')
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_alltoall_on_stream = dlsym(handle, 'nvshmemx_size_alltoall_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":561
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_alltoall_on_stream = dlsym(handle, 'nvshmemx_size_alltoall_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_barrier_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_alltoall_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":558
 *     global __nvshmemx_size_alltoall_on_stream
 *     __nvshmemx_size_alltoall_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_alltoall_on_stream')
 *     if __nvshmemx_size_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":564
 * 
 *     global __nvshmemx_barrier_on_stream
 *     __nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_barrier_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":565
 *     global __nvshmemx_barrier_on_stream
 *     __nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_on_stream')
 *     if __nvshmemx_barrier_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":566
 *     __nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_on_stream')
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_barrier_on_stream = dlsym(handle, 'nvshmemx_barrier_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":567
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_barrier_on_stream = dlsym(handle, 'nvshmemx_barrier_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 567, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":566
 *     __nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_on_stream')
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_barrier_on_stream = dlsym(handle, 'nvshmemx_barrier_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":568
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_barrier_on_stream = dlsym(handle, 'nvshmemx_barrier_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_barrier_all_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_barrier_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":565
 *     global __nvshmemx_barrier_on_stream
 *     __nvshmemx_barrier_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_on_stream')
 *     if __nvshmemx_barrier_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":571
 * 
 *     global __nvshmemx_barrier_all_on_stream
 *     __nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_all_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_barrier_all_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":572
 *     global __nvshmemx_barrier_all_on_stream
 *     __nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_all_on_stream')
 *     if __nvshmemx_barrier_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":573
 *     __nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_all_on_stream')
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_barrier_all_on_stream = dlsym(handle, 'nvshmemx_barrier_all_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":574
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_barrier_all_on_stream = dlsym(handle, 'nvshmemx_barrier_all_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 574, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":573
 *     __nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_all_on_stream')
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_barrier_all_on_stream = dlsym(handle, 'nvshmemx_barrier_all_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":575
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_barrier_all_on_stream = dlsym(handle, 'nvshmemx_barrier_all_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_sync_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_barrier_all_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":572
 *     global __nvshmemx_barrier_all_on_stream
 *     __nvshmemx_barrier_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_barrier_all_on_stream')
 *     if __nvshmemx_barrier_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":578
 * 
 *     global __nvshmemx_team_sync_on_stream
 *     __nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_team_sync_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_team_sync_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":579
 *     global __nvshmemx_team_sync_on_stream
 *     __nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_team_sync_on_stream')
 *     if __nvshmemx_team_sync_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":580
 *     __nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_team_sync_on_stream')
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_sync_on_stream = dlsym(handle, 'nvshmemx_team_sync_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":581
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_team_sync_on_stream = dlsym(handle, 'nvshmemx_team_sync_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 581, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":580
 *     __nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_team_sync_on_stream')
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_team_sync_on_stream = dlsym(handle, 'nvshmemx_team_sync_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":582
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_team_sync_on_stream = dlsym(handle, 'nvshmemx_team_sync_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_sync_all_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_team_sync_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":579
 *     global __nvshmemx_team_sync_on_stream
 *     __nvshmemx_team_sync_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_team_sync_on_stream')
 *     if __nvshmemx_team_sync_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":585
 * 
 *     global __nvshmemx_sync_all_on_stream
 *     __nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_sync_all_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_sync_all_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":586
 *     global __nvshmemx_sync_all_on_stream
 *     __nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_sync_all_on_stream')
 *     if __nvshmemx_sync_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":587
 *     __nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_sync_all_on_stream')
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_sync_all_on_stream = dlsym(handle, 'nvshmemx_sync_all_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":588
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_sync_all_on_stream = dlsym(handle, 'nvshmemx_sync_all_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 588, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":587
 *     __nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_sync_all_on_stream')
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_sync_all_on_stream = dlsym(handle, 'nvshmemx_sync_all_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":589
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_sync_all_on_stream = dlsym(handle, 'nvshmemx_sync_all_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_sync_all_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":586
 *     global __nvshmemx_sync_all_on_stream
 *     __nvshmemx_sync_all_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_sync_all_on_stream')
 *     if __nvshmemx_sync_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":592
 * 
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     __nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_broadcast_on_s);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":593
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     __nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_broadcast_on_stream')
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":594
 *     __nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_broadcast_on_stream')
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_broadcast_on_stream = dlsym(handle, 'nvshmemx_bfloat16_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":595
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_broadcast_on_stream = dlsym(handle, 'nvshmemx_bfloat16_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 595, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":594
 *     __nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_broadcast_on_stream')
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_broadcast_on_stream = dlsym(handle, 'nvshmemx_bfloat16_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":596
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_broadcast_on_stream = dlsym(handle, 'nvshmemx_bfloat16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_broadcast_on_s);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":593
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     __nvshmemx_bfloat16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_broadcast_on_stream')
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":599
 * 
 *     global __nvshmemx_half_broadcast_on_stream
 *     __nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_broadcast_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":600
 *     global __nvshmemx_half_broadcast_on_stream
 *     __nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_broadcast_on_stream')
 *     if __nvshmemx_half_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":601
 *     __nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_broadcast_on_stream')
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_broadcast_on_stream = dlsym(handle, 'nvshmemx_half_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":602
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_broadcast_on_stream = dlsym(handle, 'nvshmemx_half_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 602, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":601
 *     __nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_broadcast_on_stream')
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_broadcast_on_stream = dlsym(handle, 'nvshmemx_half_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":603
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_broadcast_on_stream = dlsym(handle, 'nvshmemx_half_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_broadcast_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":600
 *     global __nvshmemx_half_broadcast_on_stream
 *     __nvshmemx_half_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_broadcast_on_stream')
 *     if __nvshmemx_half_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":606
 * 
 *     global __nvshmemx_float_broadcast_on_stream
 *     __nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":607
 *     global __nvshmemx_float_broadcast_on_stream
 *     __nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_broadcast_on_stream')
 *     if __nvshmemx_float_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":608
 *     __nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_broadcast_on_stream')
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_broadcast_on_stream = dlsym(handle, 'nvshmemx_float_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":609
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_broadcast_on_stream = dlsym(handle, 'nvshmemx_float_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 609, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":608
 *     __nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_broadcast_on_stream')
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_broadcast_on_stream = dlsym(handle, 'nvshmemx_float_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":610
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_broadcast_on_stream = dlsym(handle, 'nvshmemx_float_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":607
 *     global __nvshmemx_float_broadcast_on_stream
 *     __nvshmemx_float_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_broadcast_on_stream')
 *     if __nvshmemx_float_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":613
 * 
 *     global __nvshmemx_double_broadcast_on_stream
 *     __nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_broadcast_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":614
 *     global __nvshmemx_double_broadcast_on_stream
 *     __nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_broadcast_on_stream')
 *     if __nvshmemx_double_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":615
 *     __nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_broadcast_on_stream')
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_broadcast_on_stream = dlsym(handle, 'nvshmemx_double_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":616
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_broadcast_on_stream = dlsym(handle, 'nvshmemx_double_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 616, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":615
 *     __nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_broadcast_on_stream')
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_broadcast_on_stream = dlsym(handle, 'nvshmemx_double_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":617
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_broadcast_on_stream = dlsym(handle, 'nvshmemx_double_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_broadcast_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":614
 *     global __nvshmemx_double_broadcast_on_stream
 *     __nvshmemx_double_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_broadcast_on_stream')
 *     if __nvshmemx_double_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":620
 * 
 *     global __nvshmemx_char_broadcast_on_stream
 *     __nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_broadcast_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":621
 *     global __nvshmemx_char_broadcast_on_stream
 *     __nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_broadcast_on_stream')
 *     if __nvshmemx_char_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":622
 *     __nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_broadcast_on_stream')
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_broadcast_on_stream = dlsym(handle, 'nvshmemx_char_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":623
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_broadcast_on_stream = dlsym(handle, 'nvshmemx_char_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 623, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":622
 *     __nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_broadcast_on_stream')
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_broadcast_on_stream = dlsym(handle, 'nvshmemx_char_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":624
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_broadcast_on_stream = dlsym(handle, 'nvshmemx_char_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_broadcast_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":621
 *     global __nvshmemx_char_broadcast_on_stream
 *     __nvshmemx_char_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_broadcast_on_stream')
 *     if __nvshmemx_char_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":627
 * 
 *     global __nvshmemx_short_broadcast_on_stream
 *     __nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":628
 *     global __nvshmemx_short_broadcast_on_stream
 *     __nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_broadcast_on_stream')
 *     if __nvshmemx_short_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":629
 *     __nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_broadcast_on_stream')
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_broadcast_on_stream = dlsym(handle, 'nvshmemx_short_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":630
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_broadcast_on_stream = dlsym(handle, 'nvshmemx_short_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 630, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":629
 *     __nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_broadcast_on_stream')
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_broadcast_on_stream = dlsym(handle, 'nvshmemx_short_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":631
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_broadcast_on_stream = dlsym(handle, 'nvshmemx_short_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":628
 *     global __nvshmemx_short_broadcast_on_stream
 *     __nvshmemx_short_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_broadcast_on_stream')
 *     if __nvshmemx_short_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":634
 * 
 *     global __nvshmemx_schar_broadcast_on_stream
 *     __nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":635
 *     global __nvshmemx_schar_broadcast_on_stream
 *     __nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_broadcast_on_stream')
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":636
 *     __nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_broadcast_on_stream')
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_broadcast_on_stream = dlsym(handle, 'nvshmemx_schar_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":637
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_broadcast_on_stream = dlsym(handle, 'nvshmemx_schar_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 637, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":636
 *     __nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_broadcast_on_stream')
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_broadcast_on_stream = dlsym(handle, 'nvshmemx_schar_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":638
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_broadcast_on_stream = dlsym(handle, 'nvshmemx_schar_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":635
 *     global __nvshmemx_schar_broadcast_on_stream
 *     __nvshmemx_schar_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_broadcast_on_stream')
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":641
 * 
 *     global __nvshmemx_int_broadcast_on_stream
 *     __nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_broadcast_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":642
 *     global __nvshmemx_int_broadcast_on_stream
 *     __nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_broadcast_on_stream')
 *     if __nvshmemx_int_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":643
 *     __nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_broadcast_on_stream')
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_broadcast_on_stream = dlsym(handle, 'nvshmemx_int_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":644
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_broadcast_on_stream = dlsym(handle, 'nvshmemx_int_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 644, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":643
 *     __nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_broadcast_on_stream')
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_broadcast_on_stream = dlsym(handle, 'nvshmemx_int_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":645
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_broadcast_on_stream = dlsym(handle, 'nvshmemx_int_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_broadcast_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":642
 *     global __nvshmemx_int_broadcast_on_stream
 *     __nvshmemx_int_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_broadcast_on_stream')
 *     if __nvshmemx_int_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":648
 * 
 *     global __nvshmemx_long_broadcast_on_stream
 *     __nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_broadcast_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":649
 *     global __nvshmemx_long_broadcast_on_stream
 *     __nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_broadcast_on_stream')
 *     if __nvshmemx_long_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":650
 *     __nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_broadcast_on_stream')
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_broadcast_on_stream = dlsym(handle, 'nvshmemx_long_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":651
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_broadcast_on_stream = dlsym(handle, 'nvshmemx_long_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 651, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":650
 *     __nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_broadcast_on_stream')
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_broadcast_on_stream = dlsym(handle, 'nvshmemx_long_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":652
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_broadcast_on_stream = dlsym(handle, 'nvshmemx_long_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_broadcast_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":649
 *     global __nvshmemx_long_broadcast_on_stream
 *     __nvshmemx_long_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_broadcast_on_stream')
 *     if __nvshmemx_long_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":655
 * 
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     __nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_broadcast_on_s);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":656
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     __nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_broadcast_on_stream')
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":657
 *     __nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_broadcast_on_stream')
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_broadcast_on_stream = dlsym(handle, 'nvshmemx_longlong_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":658
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_broadcast_on_stream = dlsym(handle, 'nvshmemx_longlong_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 658, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":657
 *     __nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_broadcast_on_stream')
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_broadcast_on_stream = dlsym(handle, 'nvshmemx_longlong_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":659
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_broadcast_on_stream = dlsym(handle, 'nvshmemx_longlong_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_broadcast_on_s);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":656
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     __nvshmemx_longlong_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_broadcast_on_stream')
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":662
 * 
 *     global __nvshmemx_int8_broadcast_on_stream
 *     __nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_broadcast_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":663
 *     global __nvshmemx_int8_broadcast_on_stream
 *     __nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_broadcast_on_stream')
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":664
 *     __nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_broadcast_on_stream')
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_broadcast_on_stream = dlsym(handle, 'nvshmemx_int8_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":665
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_broadcast_on_stream = dlsym(handle, 'nvshmemx_int8_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 665, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":664
 *     __nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_broadcast_on_stream')
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_broadcast_on_stream = dlsym(handle, 'nvshmemx_int8_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":666
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_broadcast_on_stream = dlsym(handle, 'nvshmemx_int8_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_broadcast_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":663
 *     global __nvshmemx_int8_broadcast_on_stream
 *     __nvshmemx_int8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_broadcast_on_stream')
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":669
 * 
 *     global __nvshmemx_int16_broadcast_on_stream
 *     __nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":670
 *     global __nvshmemx_int16_broadcast_on_stream
 *     __nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_broadcast_on_stream')
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":671
 *     __nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_broadcast_on_stream')
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_broadcast_on_stream = dlsym(handle, 'nvshmemx_int16_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":672
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_broadcast_on_stream = dlsym(handle, 'nvshmemx_int16_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 672, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":671
 *     __nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_broadcast_on_stream')
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_broadcast_on_stream = dlsym(handle, 'nvshmemx_int16_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":673
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_broadcast_on_stream = dlsym(handle, 'nvshmemx_int16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":670
 *     global __nvshmemx_int16_broadcast_on_stream
 *     __nvshmemx_int16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_broadcast_on_stream')
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":676
 * 
 *     global __nvshmemx_int32_broadcast_on_stream
 *     __nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":677
 *     global __nvshmemx_int32_broadcast_on_stream
 *     __nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_broadcast_on_stream')
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":678
 *     __nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_broadcast_on_stream')
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_broadcast_on_stream = dlsym(handle, 'nvshmemx_int32_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":679
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_broadcast_on_stream = dlsym(handle, 'nvshmemx_int32_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 679, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":678
 *     __nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_broadcast_on_stream')
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_broadcast_on_stream = dlsym(handle, 'nvshmemx_int32_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":680
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_broadcast_on_stream = dlsym(handle, 'nvshmemx_int32_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":677
 *     global __nvshmemx_int32_broadcast_on_stream
 *     __nvshmemx_int32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_broadcast_on_stream')
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":683
 * 
 *     global __nvshmemx_int64_broadcast_on_stream
 *     __nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":684
 *     global __nvshmemx_int64_broadcast_on_stream
 *     __nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_broadcast_on_stream')
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":685
 *     __nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_broadcast_on_stream')
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_broadcast_on_stream = dlsym(handle, 'nvshmemx_int64_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":686
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_broadcast_on_stream = dlsym(handle, 'nvshmemx_int64_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 686, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":685
 *     __nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_broadcast_on_stream')
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_broadcast_on_stream = dlsym(handle, 'nvshmemx_int64_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":687
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_broadcast_on_stream = dlsym(handle, 'nvshmemx_int64_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":684
 *     global __nvshmemx_int64_broadcast_on_stream
 *     __nvshmemx_int64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_broadcast_on_stream')
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":690
 * 
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     __nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_broadcast_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":691
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     __nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_broadcast_on_stream')
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":692
 *     __nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_broadcast_on_stream')
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint8_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":693
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint8_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 693, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":692
 *     __nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_broadcast_on_stream')
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint8_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":694
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint8_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_broadcast_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":691
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     __nvshmemx_uint8_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_broadcast_on_stream')
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":697
 * 
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     __nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_broadcast_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":698
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     __nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_broadcast_on_stream')
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":699
 *     __nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_broadcast_on_stream')
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint16_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":700
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint16_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 700, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":699
 *     __nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_broadcast_on_stream')
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint16_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":701
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint16_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_broadcast_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":698
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     __nvshmemx_uint16_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_broadcast_on_stream')
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":704
 * 
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     __nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_broadcast_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":705
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     __nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_broadcast_on_stream')
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":706
 *     __nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_broadcast_on_stream')
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint32_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":707
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint32_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 707, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":706
 *     __nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_broadcast_on_stream')
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint32_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":708
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint32_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_broadcast_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":705
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     __nvshmemx_uint32_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_broadcast_on_stream')
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":711
 * 
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     __nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_broadcast_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":712
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     __nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_broadcast_on_stream')
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":713
 *     __nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_broadcast_on_stream')
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint64_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":714
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint64_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 714, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":713
 *     __nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_broadcast_on_stream')
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint64_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":715
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_broadcast_on_stream = dlsym(handle, 'nvshmemx_uint64_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_broadcast_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_broadcast_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":712
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     __nvshmemx_uint64_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_broadcast_on_stream')
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":718
 * 
 *     global __nvshmemx_size_broadcast_on_stream
 *     __nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_broadcast_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_broadcast_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":719
 *     global __nvshmemx_size_broadcast_on_stream
 *     __nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_broadcast_on_stream')
 *     if __nvshmemx_size_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":720
 *     __nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_broadcast_on_stream')
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_broadcast_on_stream = dlsym(handle, 'nvshmemx_size_broadcast_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":721
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_broadcast_on_stream = dlsym(handle, 'nvshmemx_size_broadcast_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 721, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":720
 *     __nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_broadcast_on_stream')
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_broadcast_on_stream = dlsym(handle, 'nvshmemx_size_broadcast_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":722
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_broadcast_on_stream = dlsym(handle, 'nvshmemx_size_broadcast_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_broadcast_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":719
 *     global __nvshmemx_size_broadcast_on_stream
 *     __nvshmemx_size_broadcast_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_broadcast_on_stream')
 *     if __nvshmemx_size_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":725
 * 
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     __nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_fcollect_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":726
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     __nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_fcollect_on_stream')
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":727
 *     __nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_fcollect_on_stream')
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_fcollect_on_stream = dlsym(handle, 'nvshmemx_bfloat16_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":728
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_fcollect_on_stream = dlsym(handle, 'nvshmemx_bfloat16_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 728, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":727
 *     __nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_fcollect_on_stream')
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_fcollect_on_stream = dlsym(handle, 'nvshmemx_bfloat16_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":729
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_fcollect_on_stream = dlsym(handle, 'nvshmemx_bfloat16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_fcollect_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":726
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     __nvshmemx_bfloat16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_fcollect_on_stream')
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":732
 * 
 *     global __nvshmemx_half_fcollect_on_stream
 *     __nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":733
 *     global __nvshmemx_half_fcollect_on_stream
 *     __nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_fcollect_on_stream')
 *     if __nvshmemx_half_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":734
 *     __nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_fcollect_on_stream')
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_fcollect_on_stream = dlsym(handle, 'nvshmemx_half_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":735
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_fcollect_on_stream = dlsym(handle, 'nvshmemx_half_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 735, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":734
 *     __nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_fcollect_on_stream')
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_fcollect_on_stream = dlsym(handle, 'nvshmemx_half_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":736
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_fcollect_on_stream = dlsym(handle, 'nvshmemx_half_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":733
 *     global __nvshmemx_half_fcollect_on_stream
 *     __nvshmemx_half_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_fcollect_on_stream')
 *     if __nvshmemx_half_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":739
 * 
 *     global __nvshmemx_float_fcollect_on_stream
 *     __nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":740
 *     global __nvshmemx_float_fcollect_on_stream
 *     __nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_fcollect_on_stream')
 *     if __nvshmemx_float_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":741
 *     __nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_fcollect_on_stream')
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_fcollect_on_stream = dlsym(handle, 'nvshmemx_float_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":742
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_fcollect_on_stream = dlsym(handle, 'nvshmemx_float_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 742, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":741
 *     __nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_fcollect_on_stream')
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_fcollect_on_stream = dlsym(handle, 'nvshmemx_float_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":743
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_fcollect_on_stream = dlsym(handle, 'nvshmemx_float_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":740
 *     global __nvshmemx_float_fcollect_on_stream
 *     __nvshmemx_float_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_fcollect_on_stream')
 *     if __nvshmemx_float_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":746
 * 
 *     global __nvshmemx_double_fcollect_on_stream
 *     __nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_fcollect_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":747
 *     global __nvshmemx_double_fcollect_on_stream
 *     __nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_fcollect_on_stream')
 *     if __nvshmemx_double_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":748
 *     __nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_fcollect_on_stream')
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_fcollect_on_stream = dlsym(handle, 'nvshmemx_double_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":749
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_fcollect_on_stream = dlsym(handle, 'nvshmemx_double_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 749, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":748
 *     __nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_fcollect_on_stream')
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_fcollect_on_stream = dlsym(handle, 'nvshmemx_double_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":750
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_fcollect_on_stream = dlsym(handle, 'nvshmemx_double_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_fcollect_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":747
 *     global __nvshmemx_double_fcollect_on_stream
 *     __nvshmemx_double_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_fcollect_on_stream')
 *     if __nvshmemx_double_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":753
 * 
 *     global __nvshmemx_char_fcollect_on_stream
 *     __nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":754
 *     global __nvshmemx_char_fcollect_on_stream
 *     __nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_fcollect_on_stream')
 *     if __nvshmemx_char_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":755
 *     __nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_fcollect_on_stream')
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_fcollect_on_stream = dlsym(handle, 'nvshmemx_char_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":756
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_fcollect_on_stream = dlsym(handle, 'nvshmemx_char_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 756, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":755
 *     __nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_fcollect_on_stream')
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_fcollect_on_stream = dlsym(handle, 'nvshmemx_char_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":757
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_fcollect_on_stream = dlsym(handle, 'nvshmemx_char_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":754
 *     global __nvshmemx_char_fcollect_on_stream
 *     __nvshmemx_char_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_fcollect_on_stream')
 *     if __nvshmemx_char_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":760
 * 
 *     global __nvshmemx_short_fcollect_on_stream
 *     __nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":761
 *     global __nvshmemx_short_fcollect_on_stream
 *     __nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_fcollect_on_stream')
 *     if __nvshmemx_short_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":762
 *     __nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_fcollect_on_stream')
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_fcollect_on_stream = dlsym(handle, 'nvshmemx_short_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":763
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_fcollect_on_stream = dlsym(handle, 'nvshmemx_short_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 763, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":762
 *     __nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_fcollect_on_stream')
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_fcollect_on_stream = dlsym(handle, 'nvshmemx_short_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":764
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_fcollect_on_stream = dlsym(handle, 'nvshmemx_short_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":761
 *     global __nvshmemx_short_fcollect_on_stream
 *     __nvshmemx_short_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_fcollect_on_stream')
 *     if __nvshmemx_short_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":767
 * 
 *     global __nvshmemx_schar_fcollect_on_stream
 *     __nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":768
 *     global __nvshmemx_schar_fcollect_on_stream
 *     __nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_fcollect_on_stream')
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":769
 *     __nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_fcollect_on_stream')
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_fcollect_on_stream = dlsym(handle, 'nvshmemx_schar_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":770
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_fcollect_on_stream = dlsym(handle, 'nvshmemx_schar_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 770, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":769
 *     __nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_fcollect_on_stream')
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_fcollect_on_stream = dlsym(handle, 'nvshmemx_schar_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":771
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_fcollect_on_stream = dlsym(handle, 'nvshmemx_schar_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":768
 *     global __nvshmemx_schar_fcollect_on_stream
 *     __nvshmemx_schar_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_fcollect_on_stream')
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":774
 * 
 *     global __nvshmemx_int_fcollect_on_stream
 *     __nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":775
 *     global __nvshmemx_int_fcollect_on_stream
 *     __nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_fcollect_on_stream')
 *     if __nvshmemx_int_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":776
 *     __nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_fcollect_on_stream')
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_fcollect_on_stream = dlsym(handle, 'nvshmemx_int_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":777
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_fcollect_on_stream = dlsym(handle, 'nvshmemx_int_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 777, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":776
 *     __nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_fcollect_on_stream')
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_fcollect_on_stream = dlsym(handle, 'nvshmemx_int_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":778
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_fcollect_on_stream = dlsym(handle, 'nvshmemx_int_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":775
 *     global __nvshmemx_int_fcollect_on_stream
 *     __nvshmemx_int_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_fcollect_on_stream')
 *     if __nvshmemx_int_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":781
 * 
 *     global __nvshmemx_long_fcollect_on_stream
 *     __nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":782
 *     global __nvshmemx_long_fcollect_on_stream
 *     __nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_fcollect_on_stream')
 *     if __nvshmemx_long_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":783
 *     __nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_fcollect_on_stream')
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_fcollect_on_stream = dlsym(handle, 'nvshmemx_long_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":784
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_fcollect_on_stream = dlsym(handle, 'nvshmemx_long_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 784, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":783
 *     __nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_fcollect_on_stream')
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_fcollect_on_stream = dlsym(handle, 'nvshmemx_long_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":785
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_fcollect_on_stream = dlsym(handle, 'nvshmemx_long_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":782
 *     global __nvshmemx_long_fcollect_on_stream
 *     __nvshmemx_long_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_fcollect_on_stream')
 *     if __nvshmemx_long_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":788
 * 
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     __nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_fcollect_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":789
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     __nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_fcollect_on_stream')
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":790
 *     __nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_fcollect_on_stream')
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_fcollect_on_stream = dlsym(handle, 'nvshmemx_longlong_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":791
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_fcollect_on_stream = dlsym(handle, 'nvshmemx_longlong_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 791, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":790
 *     __nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_fcollect_on_stream')
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_fcollect_on_stream = dlsym(handle, 'nvshmemx_longlong_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":792
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_fcollect_on_stream = dlsym(handle, 'nvshmemx_longlong_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_fcollect_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":789
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     __nvshmemx_longlong_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_fcollect_on_stream')
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":795
 * 
 *     global __nvshmemx_int8_fcollect_on_stream
 *     __nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":796
 *     global __nvshmemx_int8_fcollect_on_stream
 *     __nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_fcollect_on_stream')
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":797
 *     __nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_fcollect_on_stream')
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_fcollect_on_stream = dlsym(handle, 'nvshmemx_int8_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":798
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_fcollect_on_stream = dlsym(handle, 'nvshmemx_int8_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 798, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":797
 *     __nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_fcollect_on_stream')
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_fcollect_on_stream = dlsym(handle, 'nvshmemx_int8_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":799
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_fcollect_on_stream = dlsym(handle, 'nvshmemx_int8_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":796
 *     global __nvshmemx_int8_fcollect_on_stream
 *     __nvshmemx_int8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_fcollect_on_stream')
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":802
 * 
 *     global __nvshmemx_int16_fcollect_on_stream
 *     __nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":803
 *     global __nvshmemx_int16_fcollect_on_stream
 *     __nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_fcollect_on_stream')
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":804
 *     __nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_fcollect_on_stream')
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_fcollect_on_stream = dlsym(handle, 'nvshmemx_int16_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":805
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_fcollect_on_stream = dlsym(handle, 'nvshmemx_int16_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 805, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":804
 *     __nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_fcollect_on_stream')
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_fcollect_on_stream = dlsym(handle, 'nvshmemx_int16_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":806
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_fcollect_on_stream = dlsym(handle, 'nvshmemx_int16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":803
 *     global __nvshmemx_int16_fcollect_on_stream
 *     __nvshmemx_int16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_fcollect_on_stream')
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":809
 * 
 *     global __nvshmemx_int32_fcollect_on_stream
 *     __nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":810
 *     global __nvshmemx_int32_fcollect_on_stream
 *     __nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_fcollect_on_stream')
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":811
 *     __nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_fcollect_on_stream')
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_fcollect_on_stream = dlsym(handle, 'nvshmemx_int32_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":812
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_fcollect_on_stream = dlsym(handle, 'nvshmemx_int32_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 812, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":811
 *     __nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_fcollect_on_stream')
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_fcollect_on_stream = dlsym(handle, 'nvshmemx_int32_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":813
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_fcollect_on_stream = dlsym(handle, 'nvshmemx_int32_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":810
 *     global __nvshmemx_int32_fcollect_on_stream
 *     __nvshmemx_int32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_fcollect_on_stream')
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":816
 * 
 *     global __nvshmemx_int64_fcollect_on_stream
 *     __nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":817
 *     global __nvshmemx_int64_fcollect_on_stream
 *     __nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_fcollect_on_stream')
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":818
 *     __nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_fcollect_on_stream')
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_fcollect_on_stream = dlsym(handle, 'nvshmemx_int64_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":819
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_fcollect_on_stream = dlsym(handle, 'nvshmemx_int64_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 819, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":818
 *     __nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_fcollect_on_stream')
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_fcollect_on_stream = dlsym(handle, 'nvshmemx_int64_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":820
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_fcollect_on_stream = dlsym(handle, 'nvshmemx_int64_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":817
 *     global __nvshmemx_int64_fcollect_on_stream
 *     __nvshmemx_int64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_fcollect_on_stream')
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":823
 * 
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     __nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_fcollect_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":824
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     __nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_fcollect_on_stream')
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":825
 *     __nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_fcollect_on_stream')
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint8_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":826
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint8_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 826, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":825
 *     __nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_fcollect_on_stream')
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint8_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":827
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint8_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_fcollect_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":824
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     __nvshmemx_uint8_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_fcollect_on_stream')
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":830
 * 
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     __nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_fcollect_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":831
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     __nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_fcollect_on_stream')
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":832
 *     __nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_fcollect_on_stream')
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint16_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":833
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint16_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 833, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":832
 *     __nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_fcollect_on_stream')
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint16_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":834
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint16_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_fcollect_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":831
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     __nvshmemx_uint16_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_fcollect_on_stream')
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":837
 * 
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     __nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_fcollect_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":838
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     __nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_fcollect_on_stream')
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":839
 *     __nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_fcollect_on_stream')
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint32_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":840
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint32_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 840, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":839
 *     __nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_fcollect_on_stream')
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint32_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":841
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint32_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_fcollect_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":838
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     __nvshmemx_uint32_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_fcollect_on_stream')
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":844
 * 
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     __nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_fcollect_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":845
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     __nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_fcollect_on_stream')
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":846
 *     __nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_fcollect_on_stream')
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint64_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":847
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint64_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 847, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":846
 *     __nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_fcollect_on_stream')
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint64_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":848
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_fcollect_on_stream = dlsym(handle, 'nvshmemx_uint64_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_fcollect_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_fcollect_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":845
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     __nvshmemx_uint64_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_fcollect_on_stream')
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":851
 * 
 *     global __nvshmemx_size_fcollect_on_stream
 *     __nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_fcollect_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_fcollect_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":852
 *     global __nvshmemx_size_fcollect_on_stream
 *     __nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_fcollect_on_stream')
 *     if __nvshmemx_size_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":853
 *     __nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_fcollect_on_stream')
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_fcollect_on_stream = dlsym(handle, 'nvshmemx_size_fcollect_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":854
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_fcollect_on_stream = dlsym(handle, 'nvshmemx_size_fcollect_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 854, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":853
 *     __nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_fcollect_on_stream')
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_fcollect_on_stream = dlsym(handle, 'nvshmemx_size_fcollect_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":855
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_fcollect_on_stream = dlsym(handle, 'nvshmemx_size_fcollect_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_fcollect_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":852
 *     global __nvshmemx_size_fcollect_on_stream
 *     __nvshmemx_size_fcollect_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_fcollect_on_stream')
 *     if __nvshmemx_size_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":858
 * 
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     __nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_max_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":859
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     __nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reduce_on_stream')
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":860
 *     __nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reduce_on_stream')
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":861
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 861, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":860
 *     __nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reduce_on_stream')
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":862
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_max_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":859
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     __nvshmemx_int8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reduce_on_stream')
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":865
 * 
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     __nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":866
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     __nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reduce_on_stream')
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":867
 *     __nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reduce_on_stream')
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":868
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 868, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":867
 *     __nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reduce_on_stream')
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":869
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":866
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     __nvshmemx_int16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reduce_on_stream')
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":872
 * 
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     __nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":873
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     __nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reduce_on_stream')
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":874
 *     __nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reduce_on_stream')
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":875
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 875, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":874
 *     __nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reduce_on_stream')
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":876
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":873
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     __nvshmemx_int32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reduce_on_stream')
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":879
 * 
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     __nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":880
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     __nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reduce_on_stream')
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":881
 *     __nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reduce_on_stream')
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":882
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 882, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":881
 *     __nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reduce_on_stream')
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":883
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":880
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     __nvshmemx_int64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reduce_on_stream')
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":886
 * 
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     __nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":887
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     __nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reduce_on_stream')
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":888
 *     __nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reduce_on_stream')
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":889
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 889, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":888
 *     __nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reduce_on_stream')
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":890
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":887
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     __nvshmemx_uint8_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reduce_on_stream')
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":893
 * 
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     __nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_max_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":894
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     __nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reduce_on_stream')
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":895
 *     __nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reduce_on_stream')
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":896
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 896, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":895
 *     __nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reduce_on_stream')
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":897
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_max_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":894
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     __nvshmemx_uint16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reduce_on_stream')
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":900
 * 
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     __nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_max_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":901
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     __nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reduce_on_stream')
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":902
 *     __nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reduce_on_stream')
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":903
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 903, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":902
 *     __nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reduce_on_stream')
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":904
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_max_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":901
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     __nvshmemx_uint32_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reduce_on_stream')
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":907
 * 
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     __nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_max_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":908
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     __nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reduce_on_stream')
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":909
 *     __nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reduce_on_stream')
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":910
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 910, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":909
 *     __nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reduce_on_stream')
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":911
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_max_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":908
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     __nvshmemx_uint64_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reduce_on_stream')
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":914
 * 
 *     global __nvshmemx_size_max_reduce_on_stream
 *     __nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_max_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":915
 *     global __nvshmemx_size_max_reduce_on_stream
 *     __nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reduce_on_stream')
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":916
 *     __nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reduce_on_stream')
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_max_reduce_on_stream = dlsym(handle, 'nvshmemx_size_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":917
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_max_reduce_on_stream = dlsym(handle, 'nvshmemx_size_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 917, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":916
 *     __nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reduce_on_stream')
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_max_reduce_on_stream = dlsym(handle, 'nvshmemx_size_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":918
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_max_reduce_on_stream = dlsym(handle, 'nvshmemx_size_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_max_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":915
 *     global __nvshmemx_size_max_reduce_on_stream
 *     __nvshmemx_size_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reduce_on_stream')
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":921
 * 
 *     global __nvshmemx_char_max_reduce_on_stream
 *     __nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_max_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":922
 *     global __nvshmemx_char_max_reduce_on_stream
 *     __nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reduce_on_stream')
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":923
 *     __nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reduce_on_stream')
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_max_reduce_on_stream = dlsym(handle, 'nvshmemx_char_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":924
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_max_reduce_on_stream = dlsym(handle, 'nvshmemx_char_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 924, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":923
 *     __nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reduce_on_stream')
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_max_reduce_on_stream = dlsym(handle, 'nvshmemx_char_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":925
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_max_reduce_on_stream = dlsym(handle, 'nvshmemx_char_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_max_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":922
 *     global __nvshmemx_char_max_reduce_on_stream
 *     __nvshmemx_char_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reduce_on_stream')
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":928
 * 
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     __nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":929
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     __nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reduce_on_stream')
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":930
 *     __nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reduce_on_stream')
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_max_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":931
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_max_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 931, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":930
 *     __nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reduce_on_stream')
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_max_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":932
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_max_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":929
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     __nvshmemx_schar_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reduce_on_stream')
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":935
 * 
 *     global __nvshmemx_short_max_reduce_on_stream
 *     __nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":936
 *     global __nvshmemx_short_max_reduce_on_stream
 *     __nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reduce_on_stream')
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":937
 *     __nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reduce_on_stream')
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_max_reduce_on_stream = dlsym(handle, 'nvshmemx_short_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":938
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_max_reduce_on_stream = dlsym(handle, 'nvshmemx_short_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 938, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":937
 *     __nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reduce_on_stream')
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_max_reduce_on_stream = dlsym(handle, 'nvshmemx_short_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":939
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_max_reduce_on_stream = dlsym(handle, 'nvshmemx_short_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":936
 *     global __nvshmemx_short_max_reduce_on_stream
 *     __nvshmemx_short_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reduce_on_stream')
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":942
 * 
 *     global __nvshmemx_int_max_reduce_on_stream
 *     __nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_max_reduce_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":943
 *     global __nvshmemx_int_max_reduce_on_stream
 *     __nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reduce_on_stream')
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":944
 *     __nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reduce_on_stream')
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":945
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 945, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":944
 *     __nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reduce_on_stream')
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":946
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_max_reduce_on_stream = dlsym(handle, 'nvshmemx_int_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_max_reduce_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":943
 *     global __nvshmemx_int_max_reduce_on_stream
 *     __nvshmemx_int_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reduce_on_stream')
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":949
 * 
 *     global __nvshmemx_long_max_reduce_on_stream
 *     __nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_max_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":950
 *     global __nvshmemx_long_max_reduce_on_stream
 *     __nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reduce_on_stream')
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":951
 *     __nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reduce_on_stream')
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_max_reduce_on_stream = dlsym(handle, 'nvshmemx_long_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":952
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_max_reduce_on_stream = dlsym(handle, 'nvshmemx_long_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 952, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":951
 *     __nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reduce_on_stream')
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_max_reduce_on_stream = dlsym(handle, 'nvshmemx_long_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":953
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_max_reduce_on_stream = dlsym(handle, 'nvshmemx_long_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_max_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":950
 *     global __nvshmemx_long_max_reduce_on_stream
 *     __nvshmemx_long_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reduce_on_stream')
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":956
 * 
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     __nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_max_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":957
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     __nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reduce_on_stream')
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":958
 *     __nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reduce_on_stream')
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":959
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_max_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 959, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":958
 *     __nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reduce_on_stream')
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":960
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_max_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":957
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     __nvshmemx_longlong_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reduce_on_stream')
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":963
 * 
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_max_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":964
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reduce_on_stream')
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":965
 *     __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reduce_on_stream')
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":966
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 966, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":965
 *     __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reduce_on_stream')
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":967
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_max_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":964
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     __nvshmemx_bfloat16_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reduce_on_stream')
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":970
 * 
 *     global __nvshmemx_half_max_reduce_on_stream
 *     __nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_max_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":971
 *     global __nvshmemx_half_max_reduce_on_stream
 *     __nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reduce_on_stream')
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":972
 *     __nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reduce_on_stream')
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_max_reduce_on_stream = dlsym(handle, 'nvshmemx_half_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":973
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_max_reduce_on_stream = dlsym(handle, 'nvshmemx_half_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 973, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":972
 *     __nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reduce_on_stream')
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_max_reduce_on_stream = dlsym(handle, 'nvshmemx_half_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":974
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_max_reduce_on_stream = dlsym(handle, 'nvshmemx_half_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_max_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":971
 *     global __nvshmemx_half_max_reduce_on_stream
 *     __nvshmemx_half_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reduce_on_stream')
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":977
 * 
 *     global __nvshmemx_float_max_reduce_on_stream
 *     __nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_max_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":978
 *     global __nvshmemx_float_max_reduce_on_stream
 *     __nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reduce_on_stream')
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":979
 *     __nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reduce_on_stream')
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_max_reduce_on_stream = dlsym(handle, 'nvshmemx_float_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":980
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_max_reduce_on_stream = dlsym(handle, 'nvshmemx_float_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 980, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":979
 *     __nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reduce_on_stream')
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_max_reduce_on_stream = dlsym(handle, 'nvshmemx_float_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":981
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_max_reduce_on_stream = dlsym(handle, 'nvshmemx_float_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_max_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_max_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":978
 *     global __nvshmemx_float_max_reduce_on_stream
 *     __nvshmemx_float_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reduce_on_stream')
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":984
 * 
 *     global __nvshmemx_double_max_reduce_on_stream
 *     __nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_max_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":985
 *     global __nvshmemx_double_max_reduce_on_stream
 *     __nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reduce_on_stream')
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":986
 *     __nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reduce_on_stream')
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_max_reduce_on_stream = dlsym(handle, 'nvshmemx_double_max_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":987
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_max_reduce_on_stream = dlsym(handle, 'nvshmemx_double_max_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 987, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":986
 *     __nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reduce_on_stream')
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_max_reduce_on_stream = dlsym(handle, 'nvshmemx_double_max_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":988
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_max_reduce_on_stream = dlsym(handle, 'nvshmemx_double_max_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_max_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":985
 *     global __nvshmemx_double_max_reduce_on_stream
 *     __nvshmemx_double_max_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reduce_on_stream')
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":991
 * 
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     __nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_min_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":992
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     __nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reduce_on_stream')
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":993
 *     __nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reduce_on_stream')
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":994
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 994, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":993
 *     __nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reduce_on_stream')
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":995
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_min_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":992
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     __nvshmemx_int8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reduce_on_stream')
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":998
 * 
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     __nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":999
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     __nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reduce_on_stream')
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1000
 *     __nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reduce_on_stream')
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1001
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1001, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1000
 *     __nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reduce_on_stream')
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1002
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":999
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     __nvshmemx_int16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reduce_on_stream')
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1005
 * 
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     __nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1006
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     __nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reduce_on_stream')
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1007
 *     __nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reduce_on_stream')
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1008
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1008, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1007
 *     __nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reduce_on_stream')
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1009
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1006
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     __nvshmemx_int32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reduce_on_stream')
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1012
 * 
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     __nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1013
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     __nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reduce_on_stream')
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1014
 *     __nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reduce_on_stream')
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1015
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1015, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1014
 *     __nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reduce_on_stream')
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1016
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1013
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     __nvshmemx_int64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reduce_on_stream')
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1019
 * 
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     __nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1020
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     __nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reduce_on_stream')
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1021
 *     __nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reduce_on_stream')
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1022
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1022, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1021
 *     __nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reduce_on_stream')
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1023
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1020
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     __nvshmemx_uint8_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reduce_on_stream')
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1026
 * 
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     __nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_min_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1027
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     __nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reduce_on_stream')
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1028
 *     __nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reduce_on_stream')
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1029
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1029, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1028
 *     __nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reduce_on_stream')
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1030
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_min_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1027
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     __nvshmemx_uint16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reduce_on_stream')
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1033
 * 
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     __nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_min_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1034
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     __nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reduce_on_stream')
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1035
 *     __nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reduce_on_stream')
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1036
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1036, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1035
 *     __nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reduce_on_stream')
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1037
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_min_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1034
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     __nvshmemx_uint32_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reduce_on_stream')
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1040
 * 
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     __nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_min_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1041
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     __nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reduce_on_stream')
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1042
 *     __nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reduce_on_stream')
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1043
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1043, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1042
 *     __nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reduce_on_stream')
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1044
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_min_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1041
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     __nvshmemx_uint64_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reduce_on_stream')
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1047
 * 
 *     global __nvshmemx_size_min_reduce_on_stream
 *     __nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_min_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1048
 *     global __nvshmemx_size_min_reduce_on_stream
 *     __nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reduce_on_stream')
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1049
 *     __nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reduce_on_stream')
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_min_reduce_on_stream = dlsym(handle, 'nvshmemx_size_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1050
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_min_reduce_on_stream = dlsym(handle, 'nvshmemx_size_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1050, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1049
 *     __nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reduce_on_stream')
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_min_reduce_on_stream = dlsym(handle, 'nvshmemx_size_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1051
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_min_reduce_on_stream = dlsym(handle, 'nvshmemx_size_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_min_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1048
 *     global __nvshmemx_size_min_reduce_on_stream
 *     __nvshmemx_size_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reduce_on_stream')
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1054
 * 
 *     global __nvshmemx_char_min_reduce_on_stream
 *     __nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_min_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1055
 *     global __nvshmemx_char_min_reduce_on_stream
 *     __nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reduce_on_stream')
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1056
 *     __nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reduce_on_stream')
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_min_reduce_on_stream = dlsym(handle, 'nvshmemx_char_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1057
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_min_reduce_on_stream = dlsym(handle, 'nvshmemx_char_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1057, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1056
 *     __nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reduce_on_stream')
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_min_reduce_on_stream = dlsym(handle, 'nvshmemx_char_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1058
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_min_reduce_on_stream = dlsym(handle, 'nvshmemx_char_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_min_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1055
 *     global __nvshmemx_char_min_reduce_on_stream
 *     __nvshmemx_char_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reduce_on_stream')
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1061
 * 
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     __nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1062
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     __nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reduce_on_stream')
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1063
 *     __nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reduce_on_stream')
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_min_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1064
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_min_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1064, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1063
 *     __nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reduce_on_stream')
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_min_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1065
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_min_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1062
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     __nvshmemx_schar_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reduce_on_stream')
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1068
 * 
 *     global __nvshmemx_short_min_reduce_on_stream
 *     __nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1069
 *     global __nvshmemx_short_min_reduce_on_stream
 *     __nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reduce_on_stream')
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1070
 *     __nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reduce_on_stream')
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_min_reduce_on_stream = dlsym(handle, 'nvshmemx_short_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1071
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_min_reduce_on_stream = dlsym(handle, 'nvshmemx_short_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1071, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1070
 *     __nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reduce_on_stream')
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_min_reduce_on_stream = dlsym(handle, 'nvshmemx_short_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1072
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_min_reduce_on_stream = dlsym(handle, 'nvshmemx_short_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1069
 *     global __nvshmemx_short_min_reduce_on_stream
 *     __nvshmemx_short_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reduce_on_stream')
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1075
 * 
 *     global __nvshmemx_int_min_reduce_on_stream
 *     __nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_min_reduce_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1076
 *     global __nvshmemx_int_min_reduce_on_stream
 *     __nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reduce_on_stream')
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1077
 *     __nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reduce_on_stream')
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1078
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1078, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1077
 *     __nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reduce_on_stream')
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1079
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_min_reduce_on_stream = dlsym(handle, 'nvshmemx_int_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_min_reduce_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1076
 *     global __nvshmemx_int_min_reduce_on_stream
 *     __nvshmemx_int_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reduce_on_stream')
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1082
 * 
 *     global __nvshmemx_long_min_reduce_on_stream
 *     __nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_min_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1083
 *     global __nvshmemx_long_min_reduce_on_stream
 *     __nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reduce_on_stream')
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1084
 *     __nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reduce_on_stream')
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_min_reduce_on_stream = dlsym(handle, 'nvshmemx_long_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1085
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_min_reduce_on_stream = dlsym(handle, 'nvshmemx_long_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1085, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1084
 *     __nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reduce_on_stream')
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_min_reduce_on_stream = dlsym(handle, 'nvshmemx_long_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1086
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_min_reduce_on_stream = dlsym(handle, 'nvshmemx_long_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_min_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1083
 *     global __nvshmemx_long_min_reduce_on_stream
 *     __nvshmemx_long_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reduce_on_stream')
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1089
 * 
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     __nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_min_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1090
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     __nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reduce_on_stream')
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1091
 *     __nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reduce_on_stream')
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1092
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_min_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1092, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1091
 *     __nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reduce_on_stream')
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1093
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_min_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1090
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     __nvshmemx_longlong_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reduce_on_stream')
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1096
 * 
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_min_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1097
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reduce_on_stream')
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1098
 *     __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reduce_on_stream')
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1099
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1099, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1098
 *     __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reduce_on_stream')
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1100
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_min_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1097
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     __nvshmemx_bfloat16_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reduce_on_stream')
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1103
 * 
 *     global __nvshmemx_half_min_reduce_on_stream
 *     __nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_min_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1104
 *     global __nvshmemx_half_min_reduce_on_stream
 *     __nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reduce_on_stream')
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1105
 *     __nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reduce_on_stream')
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_min_reduce_on_stream = dlsym(handle, 'nvshmemx_half_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1106
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_min_reduce_on_stream = dlsym(handle, 'nvshmemx_half_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1106, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1105
 *     __nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reduce_on_stream')
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_min_reduce_on_stream = dlsym(handle, 'nvshmemx_half_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1107
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_min_reduce_on_stream = dlsym(handle, 'nvshmemx_half_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_min_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1104
 *     global __nvshmemx_half_min_reduce_on_stream
 *     __nvshmemx_half_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reduce_on_stream')
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1110
 * 
 *     global __nvshmemx_float_min_reduce_on_stream
 *     __nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_min_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1111
 *     global __nvshmemx_float_min_reduce_on_stream
 *     __nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reduce_on_stream')
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1112
 *     __nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reduce_on_stream')
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_min_reduce_on_stream = dlsym(handle, 'nvshmemx_float_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1113
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_min_reduce_on_stream = dlsym(handle, 'nvshmemx_float_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1113, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1112
 *     __nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reduce_on_stream')
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_min_reduce_on_stream = dlsym(handle, 'nvshmemx_float_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1114
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_min_reduce_on_stream = dlsym(handle, 'nvshmemx_float_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_min_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_min_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1111
 *     global __nvshmemx_float_min_reduce_on_stream
 *     __nvshmemx_float_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reduce_on_stream')
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1117
 * 
 *     global __nvshmemx_double_min_reduce_on_stream
 *     __nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_min_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1118
 *     global __nvshmemx_double_min_reduce_on_stream
 *     __nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reduce_on_stream')
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1119
 *     __nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reduce_on_stream')
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_min_reduce_on_stream = dlsym(handle, 'nvshmemx_double_min_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1120
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_min_reduce_on_stream = dlsym(handle, 'nvshmemx_double_min_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1120, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1119
 *     __nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reduce_on_stream')
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_min_reduce_on_stream = dlsym(handle, 'nvshmemx_double_min_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1121
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_min_reduce_on_stream = dlsym(handle, 'nvshmemx_double_min_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_min_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1118
 *     global __nvshmemx_double_min_reduce_on_stream
 *     __nvshmemx_double_min_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reduce_on_stream')
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1124
 * 
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     __nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_sum_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1125
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     __nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reduce_on_stream')
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1126
 *     __nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reduce_on_stream')
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1127
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1127, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1126
 *     __nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reduce_on_stream')
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1128
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_sum_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1125
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     __nvshmemx_int8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reduce_on_stream')
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1131
 * 
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     __nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1132
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     __nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reduce_on_stream')
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1133
 *     __nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reduce_on_stream')
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1134
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1134, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1133
 *     __nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reduce_on_stream')
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1135
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1132
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     __nvshmemx_int16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reduce_on_stream')
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1138
 * 
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     __nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1139
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     __nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reduce_on_stream')
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1140
 *     __nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reduce_on_stream')
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1141
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1141, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1140
 *     __nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reduce_on_stream')
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1142
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1139
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     __nvshmemx_int32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reduce_on_stream')
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1145
 * 
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     __nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1146
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     __nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reduce_on_stream')
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1147
 *     __nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reduce_on_stream')
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1148
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1148, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1147
 *     __nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reduce_on_stream')
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1149
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1146
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     __nvshmemx_int64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reduce_on_stream')
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1152
 * 
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     __nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1153
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     __nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reduce_on_stream')
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1154
 *     __nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reduce_on_stream')
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1155
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1155, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1154
 *     __nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reduce_on_stream')
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1156
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1153
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     __nvshmemx_uint8_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reduce_on_stream')
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1159
 * 
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     __nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_sum_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1160
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     __nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reduce_on_stream')
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1161
 *     __nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reduce_on_stream')
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1162
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1162, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1161
 *     __nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reduce_on_stream')
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1163
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_sum_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1160
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     __nvshmemx_uint16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reduce_on_stream')
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1166
 * 
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     __nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_sum_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1167
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     __nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reduce_on_stream')
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1168
 *     __nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reduce_on_stream')
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1169
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1169, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1168
 *     __nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reduce_on_stream')
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1170
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_sum_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1167
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     __nvshmemx_uint32_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reduce_on_stream')
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1173
 * 
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     __nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_sum_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1174
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     __nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reduce_on_stream')
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1175
 *     __nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reduce_on_stream')
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1176
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1176, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1175
 *     __nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reduce_on_stream')
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1177
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_sum_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1174
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     __nvshmemx_uint64_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reduce_on_stream')
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1180
 * 
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     __nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_sum_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1181
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     __nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reduce_on_stream')
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1182
 *     __nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reduce_on_stream')
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_size_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1183
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_size_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1183, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1182
 *     __nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reduce_on_stream')
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_size_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1184
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_size_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_sum_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1181
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     __nvshmemx_size_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reduce_on_stream')
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1187
 * 
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     __nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_sum_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1188
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     __nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reduce_on_stream')
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1189
 *     __nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reduce_on_stream')
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_char_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1190
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_char_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1190, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1189
 *     __nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reduce_on_stream')
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_char_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1191
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_char_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_sum_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1188
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     __nvshmemx_char_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reduce_on_stream')
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1194
 * 
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     __nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1195
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     __nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reduce_on_stream')
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1196
 *     __nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reduce_on_stream')
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1197
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1197, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1196
 *     __nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reduce_on_stream')
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1198
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1195
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     __nvshmemx_schar_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reduce_on_stream')
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1201
 * 
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     __nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1202
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     __nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reduce_on_stream')
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1203
 *     __nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reduce_on_stream')
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_short_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1204
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_short_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1204, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1203
 *     __nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reduce_on_stream')
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_short_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1205
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_short_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1202
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     __nvshmemx_short_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reduce_on_stream')
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1208
 * 
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     __nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_sum_reduce_on_strea);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1209
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     __nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reduce_on_stream')
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1210
 *     __nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reduce_on_stream')
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1211
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1211, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1210
 *     __nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reduce_on_stream')
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1212
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_int_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_sum_reduce_on_strea);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1209
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     __nvshmemx_int_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reduce_on_stream')
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1215
 * 
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     __nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_sum_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1216
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     __nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reduce_on_stream')
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1217
 *     __nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reduce_on_stream')
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_long_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1218
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_long_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1218, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1217
 *     __nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reduce_on_stream')
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_long_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1219
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_long_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_sum_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1216
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     __nvshmemx_long_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reduce_on_stream')
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1222
 * 
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     __nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_sum_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1223
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     __nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reduce_on_stream')
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1224
 *     __nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reduce_on_stream')
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1225
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1225, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1224
 *     __nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reduce_on_stream')
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1226
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_sum_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1223
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     __nvshmemx_longlong_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reduce_on_stream')
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1229
 * 
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_sum_reduce_on);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1230
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reduce_on_stream')
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1231
 *     __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reduce_on_stream')
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1232
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1232, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1231
 *     __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reduce_on_stream')
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1233
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_sum_reduce_on);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1230
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     __nvshmemx_bfloat16_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reduce_on_stream')
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1236
 * 
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     __nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_sum_reduce_on_stre);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1237
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     __nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reduce_on_stream')
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1238
 *     __nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reduce_on_stream')
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_half_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1239
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_half_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1239, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1238
 *     __nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reduce_on_stream')
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_half_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1240
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_half_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_sum_reduce_on_stre);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1237
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     __nvshmemx_half_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reduce_on_stream')
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1243
 * 
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     __nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_sum_reduce_on_str);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1244
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     __nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reduce_on_stream')
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1245
 *     __nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reduce_on_stream')
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_float_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1246
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_float_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1246, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1245
 *     __nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reduce_on_stream')
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_float_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1247
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_float_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_sum_reduce_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_sum_reduce_on_str);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1244
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     __nvshmemx_float_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reduce_on_stream')
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1250
 * 
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     __nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_sum_reduce_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1251
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     __nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reduce_on_stream')
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1252
 *     __nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reduce_on_stream')
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_double_sum_reduce_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1253
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_double_sum_reduce_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1253, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1252
 *     __nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reduce_on_stream')
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_double_sum_reduce_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1254
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_sum_reduce_on_stream = dlsym(handle, 'nvshmemx_double_sum_reduce_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_sum_reduce_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1251
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     __nvshmemx_double_sum_reduce_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reduce_on_stream')
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1257
 * 
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     __nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1258
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     __nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reducescatter_on_stream')
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1259
 *     __nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reducescatter_on_stream')
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1260
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1260, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1259
 *     __nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reducescatter_on_stream')
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1261
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1258
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     __nvshmemx_int8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_max_reducescatter_on_stream')
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1264
 * 
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     __nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1265
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     __nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reducescatter_on_stream')
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1266
 *     __nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reducescatter_on_stream')
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1267
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1267, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1266
 *     __nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reducescatter_on_stream')
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1268
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1265
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     __nvshmemx_int16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_max_reducescatter_on_stream')
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1271
 * 
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     __nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1272
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     __nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reducescatter_on_stream')
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1273
 *     __nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reducescatter_on_stream')
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1274
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1274, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1273
 *     __nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reducescatter_on_stream')
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1275
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1272
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     __nvshmemx_int32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_max_reducescatter_on_stream')
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1278
 * 
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     __nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1279
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     __nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reducescatter_on_stream')
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1280
 *     __nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reducescatter_on_stream')
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1281
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1281, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1280
 *     __nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reducescatter_on_stream')
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1282
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1279
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     __nvshmemx_int64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_max_reducescatter_on_stream')
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1285
 * 
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1286
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reducescatter_on_stream')
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1287
 *     __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reducescatter_on_stream')
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1288
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1288, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1287
 *     __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reducescatter_on_stream')
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1289
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1286
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     __nvshmemx_uint8_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_max_reducescatter_on_stream')
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1292
 * 
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_max_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1293
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reducescatter_on_stream')
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1294
 *     __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reducescatter_on_stream')
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1295
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1295, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1294
 *     __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reducescatter_on_stream')
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1296
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_max_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1293
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     __nvshmemx_uint16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_max_reducescatter_on_stream')
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1299
 * 
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_max_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1300
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reducescatter_on_stream')
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1301
 *     __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reducescatter_on_stream')
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1302
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1302, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1301
 *     __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reducescatter_on_stream')
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1303
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_max_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1300
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     __nvshmemx_uint32_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_max_reducescatter_on_stream')
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1306
 * 
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_max_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1307
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reducescatter_on_stream')
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1308
 *     __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reducescatter_on_stream')
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1309
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1309, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1308
 *     __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reducescatter_on_stream')
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1310
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_max_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1307
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     __nvshmemx_uint64_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_max_reducescatter_on_stream')
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1313
 * 
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     __nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1314
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     __nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reducescatter_on_stream')
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1315
 *     __nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reducescatter_on_stream')
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1316
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1316, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1315
 *     __nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reducescatter_on_stream')
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1317
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1314
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     __nvshmemx_size_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_max_reducescatter_on_stream')
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1320
 * 
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     __nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1321
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     __nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reducescatter_on_stream')
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1322
 *     __nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reducescatter_on_stream')
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1323
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1323, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1322
 *     __nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reducescatter_on_stream')
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1324
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1321
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     __nvshmemx_char_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_max_reducescatter_on_stream')
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1327
 * 
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     __nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1328
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     __nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reducescatter_on_stream')
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1329
 *     __nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reducescatter_on_stream')
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1330
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1330, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1329
 *     __nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reducescatter_on_stream')
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1331
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1328
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     __nvshmemx_schar_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_max_reducescatter_on_stream')
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1334
 * 
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     __nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1335
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     __nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reducescatter_on_stream')
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1336
 *     __nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reducescatter_on_stream')
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1337
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1337, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1336
 *     __nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reducescatter_on_stream')
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1338
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1335
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     __nvshmemx_short_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_max_reducescatter_on_stream')
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1341
 * 
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     __nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_max_reducescatter_o);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1342
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     __nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reducescatter_on_stream')
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1343
 *     __nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reducescatter_on_stream')
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1344
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1344, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1343
 *     __nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reducescatter_on_stream')
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1345
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_max_reducescatter_o);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1342
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     __nvshmemx_int_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_max_reducescatter_on_stream')
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1348
 * 
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     __nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1349
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     __nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reducescatter_on_stream')
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1350
 *     __nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reducescatter_on_stream')
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1351
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1351, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1350
 *     __nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reducescatter_on_stream')
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1352
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1349
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     __nvshmemx_long_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_max_reducescatter_on_stream')
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1355
 * 
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_max_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1356
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reducescatter_on_stream')
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1357
 *     __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reducescatter_on_stream')
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1358
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1358, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1357
 *     __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reducescatter_on_stream')
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1359
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_max_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1356
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     __nvshmemx_longlong_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_max_reducescatter_on_stream')
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1362
 * 
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_max_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1363
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1364
 *     __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1365
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1365, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1364
 *     __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1366
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_max_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1363
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     __nvshmemx_bfloat16_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_max_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1369
 * 
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     __nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1370
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     __nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reducescatter_on_stream')
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1371
 *     __nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reducescatter_on_stream')
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1372
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1372, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1371
 *     __nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reducescatter_on_stream')
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1373
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1370
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     __nvshmemx_half_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_max_reducescatter_on_stream')
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1376
 * 
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     __nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_max_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1377
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     __nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reducescatter_on_stream')
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1378
 *     __nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reducescatter_on_stream')
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1379
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1379, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1378
 *     __nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reducescatter_on_stream')
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1380
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_max_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_max_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1377
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     __nvshmemx_float_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_max_reducescatter_on_stream')
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1383
 * 
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     __nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_max_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1384
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     __nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reducescatter_on_stream')
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1385
 *     __nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reducescatter_on_stream')
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_max_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1386
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_max_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1386, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1385
 *     __nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reducescatter_on_stream')
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_max_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1387
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_max_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_max_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_max_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1384
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     __nvshmemx_double_max_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_max_reducescatter_on_stream')
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1390
 * 
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     __nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1391
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     __nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reducescatter_on_stream')
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1392
 *     __nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reducescatter_on_stream')
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1393
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1393, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1392
 *     __nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reducescatter_on_stream')
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1394
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1391
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     __nvshmemx_int8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_min_reducescatter_on_stream')
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1397
 * 
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     __nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1398
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     __nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reducescatter_on_stream')
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1399
 *     __nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reducescatter_on_stream')
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1400
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1400, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1399
 *     __nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reducescatter_on_stream')
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1401
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1398
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     __nvshmemx_int16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_min_reducescatter_on_stream')
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1404
 * 
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     __nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1405
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     __nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reducescatter_on_stream')
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1406
 *     __nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reducescatter_on_stream')
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1407
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1407, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1406
 *     __nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reducescatter_on_stream')
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1408
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1405
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     __nvshmemx_int32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_min_reducescatter_on_stream')
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1411
 * 
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     __nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1412
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     __nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reducescatter_on_stream')
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1413
 *     __nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reducescatter_on_stream')
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1414
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1414, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1413
 *     __nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reducescatter_on_stream')
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1415
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1412
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     __nvshmemx_int64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_min_reducescatter_on_stream')
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1418
 * 
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1419
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reducescatter_on_stream')
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1420
 *     __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reducescatter_on_stream')
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1421
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1421, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1420
 *     __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reducescatter_on_stream')
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1422
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1419
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     __nvshmemx_uint8_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_min_reducescatter_on_stream')
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1425
 * 
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_min_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1426
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reducescatter_on_stream')
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1427
 *     __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reducescatter_on_stream')
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1428
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1428, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1427
 *     __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reducescatter_on_stream')
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1429
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_min_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1426
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     __nvshmemx_uint16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_min_reducescatter_on_stream')
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1432
 * 
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_min_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1433
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reducescatter_on_stream')
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1434
 *     __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reducescatter_on_stream')
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1435
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1435, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1434
 *     __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reducescatter_on_stream')
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1436
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_min_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1433
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     __nvshmemx_uint32_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_min_reducescatter_on_stream')
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1439
 * 
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_min_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1440
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reducescatter_on_stream')
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1441
 *     __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reducescatter_on_stream')
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1442
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1442, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1441
 *     __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reducescatter_on_stream')
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1443
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_min_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1440
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     __nvshmemx_uint64_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_min_reducescatter_on_stream')
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1446
 * 
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     __nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1447
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     __nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reducescatter_on_stream')
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1448
 *     __nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reducescatter_on_stream')
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1449
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1449, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1448
 *     __nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reducescatter_on_stream')
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1450
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1447
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     __nvshmemx_size_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_min_reducescatter_on_stream')
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1453
 * 
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     __nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1454
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     __nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reducescatter_on_stream')
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1455
 *     __nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reducescatter_on_stream')
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1456
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1456, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1455
 *     __nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reducescatter_on_stream')
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1457
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1454
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     __nvshmemx_char_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_min_reducescatter_on_stream')
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1460
 * 
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     __nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1461
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     __nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reducescatter_on_stream')
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1462
 *     __nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reducescatter_on_stream')
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1463
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1463, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1462
 *     __nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reducescatter_on_stream')
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1464
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1461
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     __nvshmemx_schar_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_min_reducescatter_on_stream')
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1467
 * 
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     __nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1468
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     __nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reducescatter_on_stream')
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1469
 *     __nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reducescatter_on_stream')
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1470
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1470, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1469
 *     __nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reducescatter_on_stream')
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1471
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1468
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     __nvshmemx_short_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_min_reducescatter_on_stream')
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1474
 * 
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     __nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_min_reducescatter_o);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1475
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     __nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reducescatter_on_stream')
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1476
 *     __nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reducescatter_on_stream')
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1477
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1477, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1476
 *     __nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reducescatter_on_stream')
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1478
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_min_reducescatter_o);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1475
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     __nvshmemx_int_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_min_reducescatter_on_stream')
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1481
 * 
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     __nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1482
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     __nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reducescatter_on_stream')
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1483
 *     __nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reducescatter_on_stream')
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1484
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1484, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1483
 *     __nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reducescatter_on_stream')
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1485
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1482
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     __nvshmemx_long_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_min_reducescatter_on_stream')
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1488
 * 
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_min_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1489
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reducescatter_on_stream')
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1490
 *     __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reducescatter_on_stream')
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1491
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1491, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1490
 *     __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reducescatter_on_stream')
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1492
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_min_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1489
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     __nvshmemx_longlong_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_min_reducescatter_on_stream')
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1495
 * 
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_min_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1496
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1497
 *     __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1498
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1498, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1497
 *     __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1499
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_min_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1496
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     __nvshmemx_bfloat16_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_min_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1502
 * 
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     __nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1503
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     __nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reducescatter_on_stream')
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1504
 *     __nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reducescatter_on_stream')
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1505
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1505, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1504
 *     __nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reducescatter_on_stream')
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1506
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1503
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     __nvshmemx_half_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_min_reducescatter_on_stream')
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1509
 * 
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     __nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_min_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1510
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     __nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reducescatter_on_stream')
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1511
 *     __nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reducescatter_on_stream')
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1512
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1512, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1511
 *     __nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reducescatter_on_stream')
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1513
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_min_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_min_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1510
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     __nvshmemx_float_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_min_reducescatter_on_stream')
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1516
 * 
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     __nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_min_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1517
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     __nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reducescatter_on_stream')
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1518
 *     __nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reducescatter_on_stream')
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_min_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1519
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_min_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1519, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1518
 *     __nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reducescatter_on_stream')
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_min_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1520
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_min_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_min_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_min_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1517
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     __nvshmemx_double_min_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_min_reducescatter_on_stream')
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1523
 * 
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int8_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1524
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reducescatter_on_stream')
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1525
 *     __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reducescatter_on_stream')
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1526
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1526, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1525
 *     __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reducescatter_on_stream')
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1527
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int8_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int8_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1524
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     __nvshmemx_int8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int8_sum_reducescatter_on_stream')
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1530
 * 
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int16_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1531
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reducescatter_on_stream')
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1532
 *     __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reducescatter_on_stream')
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1533
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1533, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1532
 *     __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reducescatter_on_stream')
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1534
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int16_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1531
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     __nvshmemx_int16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int16_sum_reducescatter_on_stream')
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1537
 * 
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int32_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1538
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reducescatter_on_stream')
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1539
 *     __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reducescatter_on_stream')
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1540
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1540, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1539
 *     __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reducescatter_on_stream')
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1541
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int32_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int32_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1538
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     __nvshmemx_int32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int32_sum_reducescatter_on_stream')
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1544
 * 
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int64_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1545
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reducescatter_on_stream')
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1546
 *     __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reducescatter_on_stream')
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1547
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1547, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1546
 *     __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reducescatter_on_stream')
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1548
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int64_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int64_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1545
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     __nvshmemx_int64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int64_sum_reducescatter_on_stream')
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1551
 * 
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint8_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1552
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1553
 *     __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1554
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1554, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1553
 *     __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1555
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint8_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint8_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1552
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     __nvshmemx_uint8_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint8_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1558
 * 
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint16_sum_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1559
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1560
 *     __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1561
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1561, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1560
 *     __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1562
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint16_sum_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1559
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     __nvshmemx_uint16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint16_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1565
 * 
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint32_sum_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1566
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1567
 *     __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1568
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1568, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1567
 *     __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1569
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint32_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint32_sum_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1566
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     __nvshmemx_uint32_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint32_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1572
 * 
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_uint64_sum_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1573
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1574
 *     __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1575
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1575, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1574
 *     __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1576
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_uint64_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_uint64_sum_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1573
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     __nvshmemx_uint64_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_uint64_sum_reducescatter_on_stream')
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1579
 * 
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     __nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_size_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1580
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     __nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reducescatter_on_stream')
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1581
 *     __nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reducescatter_on_stream')
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1582
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_size_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1582, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1581
 *     __nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reducescatter_on_stream')
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_size_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1583
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_size_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_size_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_size_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1580
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     __nvshmemx_size_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_size_sum_reducescatter_on_stream')
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1586
 * 
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     __nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_char_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1587
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     __nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reducescatter_on_stream')
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1588
 *     __nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reducescatter_on_stream')
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1589
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_char_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1589, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1588
 *     __nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reducescatter_on_stream')
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_char_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1590
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_char_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_char_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_char_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1587
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     __nvshmemx_char_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_char_sum_reducescatter_on_stream')
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1593
 * 
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_schar_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1594
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reducescatter_on_stream')
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1595
 *     __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reducescatter_on_stream')
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1596
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1596, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1595
 *     __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reducescatter_on_stream')
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1597
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_schar_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_schar_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1594
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     __nvshmemx_schar_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_schar_sum_reducescatter_on_stream')
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1600
 * 
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     __nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_short_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1601
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     __nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reducescatter_on_stream')
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1602
 *     __nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reducescatter_on_stream')
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1603
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_short_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1603, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1602
 *     __nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reducescatter_on_stream')
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_short_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1604
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_short_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_short_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_short_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1601
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     __nvshmemx_short_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_short_sum_reducescatter_on_stream')
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1607
 * 
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     __nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_int_sum_reducescatter_o);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1608
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     __nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reducescatter_on_stream')
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1609
 *     __nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reducescatter_on_stream')
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1610
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_int_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1610, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1609
 *     __nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reducescatter_on_stream')
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_int_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1611
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_int_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_int_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_int_sum_reducescatter_o);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1608
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     __nvshmemx_int_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_int_sum_reducescatter_on_stream')
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1614
 * 
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     __nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_long_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1615
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     __nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reducescatter_on_stream')
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1616
 *     __nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reducescatter_on_stream')
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1617
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_long_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1617, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1616
 *     __nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reducescatter_on_stream')
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_long_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1618
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_long_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_long_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_long_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1615
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     __nvshmemx_long_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_long_sum_reducescatter_on_stream')
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1621
 * 
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_longlong_sum_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1622
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reducescatter_on_stream')
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1623
 *     __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reducescatter_on_stream')
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1624
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1624, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1623
 *     __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reducescatter_on_stream')
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1625
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_longlong_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_longlong_sum_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1622
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     __nvshmemx_longlong_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_longlong_sum_reducescatter_on_stream')
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1628
 * 
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_bfloat16_sum_reducescat);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1629
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1630
 *     __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1631
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1631, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1630
 *     __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1632
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_bfloat16_sum_reducescat);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1629
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     __nvshmemx_bfloat16_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_bfloat16_sum_reducescatter_on_stream')
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1635
 * 
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     __nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_half_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1636
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     __nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reducescatter_on_stream')
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1637
 *     __nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reducescatter_on_stream')
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1638
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_half_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1638, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1637
 *     __nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reducescatter_on_stream')
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_half_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1639
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_half_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_half_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_half_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1636
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     __nvshmemx_half_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_half_sum_reducescatter_on_stream')
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1642
 * 
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     __nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_float_sum_reducescatter);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1643
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     __nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reducescatter_on_stream')
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1644
 *     __nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reducescatter_on_stream')
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1645
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_float_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1645, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1644
 *     __nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reducescatter_on_stream')
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_float_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1646
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_float_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_float_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_sum_reducescatter_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_float_sum_reducescatter);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1643
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     __nvshmemx_float_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_float_sum_reducescatter_on_stream')
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1649
 * 
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     __nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_double_sum_reducescatte);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1650
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     __nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reducescatter_on_stream')
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1651
 *     __nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reducescatter_on_stream')
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_sum_reducescatter_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1652
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_double_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_sum_reducescatter_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1652, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1651
 *     __nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reducescatter_on_stream')
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_double_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_sum_reducescatter_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1653
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_double_sum_reducescatter_on_stream = dlsym(handle, 'nvshmemx_double_sum_reducescatter_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_hostlib_init_attr
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_double_sum_reducescatte);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1650
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     __nvshmemx_double_sum_reducescatter_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_double_sum_reducescatter_on_stream')
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1656
 * 
 *     global __nvshmemx_hostlib_init_attr
 *     __nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_init_attr')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_hostlib_init_attr);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1657
 *     global __nvshmemx_hostlib_init_attr
 *     __nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_init_attr')
 *     if __nvshmemx_hostlib_init_attr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1658
 *     __nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_init_attr')
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_hostlib_init_attr = dlsym(handle, 'nvshmemx_hostlib_init_attr')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1659
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_hostlib_init_attr = dlsym(handle, 'nvshmemx_hostlib_init_attr')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1659, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1658
 *     __nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_init_attr')
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_hostlib_init_attr = dlsym(handle, 'nvshmemx_hostlib_init_attr')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1660
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_hostlib_init_attr = dlsym(handle, 'nvshmemx_hostlib_init_attr')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_hostlib_finalize
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_hostlib_init_attr);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1657
 *     global __nvshmemx_hostlib_init_attr
 *     __nvshmemx_hostlib_init_attr = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_init_attr')
 *     if __nvshmemx_hostlib_init_attr == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1663
 * 
 *     global __nvshmemx_hostlib_finalize
 *     __nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_finalize')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_hostlib_finalize);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1664
 *     global __nvshmemx_hostlib_finalize
 *     __nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_finalize')
 *     if __nvshmemx_hostlib_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1665
 *     __nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_finalize')
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_hostlib_finalize = dlsym(handle, 'nvshmemx_hostlib_finalize')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1666
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_hostlib_finalize = dlsym(handle, 'nvshmemx_hostlib_finalize')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1666, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1665
 *     __nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_finalize')
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_hostlib_finalize = dlsym(handle, 'nvshmemx_hostlib_finalize')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1667
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_hostlib_finalize = dlsym(handle, 'nvshmemx_hostlib_finalize')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_set_attr_uniqueid_args
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_hostlib_finalize);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1664
 *     global __nvshmemx_hostlib_finalize
 *     __nvshmemx_hostlib_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_hostlib_finalize')
 *     if __nvshmemx_hostlib_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1670
 * 
 *     global __nvshmemx_set_attr_uniqueid_args
 *     __nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_uniqueid_args')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_set_attr_uniqueid_args);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1671
 *     global __nvshmemx_set_attr_uniqueid_args
 *     __nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_uniqueid_args')
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1672
 *     __nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_uniqueid_args')
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_set_attr_uniqueid_args = dlsym(handle, 'nvshmemx_set_attr_uniqueid_args')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1673
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_set_attr_uniqueid_args = dlsym(handle, 'nvshmemx_set_attr_uniqueid_args')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1673, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1672
 *     __nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_uniqueid_args')
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_set_attr_uniqueid_args = dlsym(handle, 'nvshmemx_set_attr_uniqueid_args')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1674
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_set_attr_uniqueid_args = dlsym(handle, 'nvshmemx_set_attr_uniqueid_args')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_set_attr_mpi_comm_args
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_set_attr_uniqueid_args);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1671
 *     global __nvshmemx_set_attr_uniqueid_args
 *     __nvshmemx_set_attr_uniqueid_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_uniqueid_args')
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1677
 * 
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     __nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_mpi_comm_args')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_set_attr_mpi_comm_args);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1678
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     __nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_mpi_comm_args')
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1679
 *     __nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_mpi_comm_args')
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_set_attr_mpi_comm_args = dlsym(handle, 'nvshmemx_set_attr_mpi_comm_args')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1680
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_set_attr_mpi_comm_args = dlsym(handle, 'nvshmemx_set_attr_mpi_comm_args')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1680, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1679
 *     __nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_mpi_comm_args')
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_set_attr_mpi_comm_args = dlsym(handle, 'nvshmemx_set_attr_mpi_comm_args')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1681
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_set_attr_mpi_comm_args = dlsym(handle, 'nvshmemx_set_attr_mpi_comm_args')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_get_uniqueid
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_set_attr_mpi_comm_args);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1678
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     __nvshmemx_set_attr_mpi_comm_args = dlsym(RTLD_DEFAULT, 'nvshmemx_set_attr_mpi_comm_args')
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1684
 * 
 *     global __nvshmemx_get_uniqueid
 *     __nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_get_uniqueid')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_get_uniqueid == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_get_uniqueid);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1685
 *     global __nvshmemx_get_uniqueid
 *     __nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_get_uniqueid')
 *     if __nvshmemx_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1686
 *     __nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_get_uniqueid')
 *     if __nvshmemx_get_uniqueid == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_get_uniqueid = dlsym(handle, 'nvshmemx_get_uniqueid')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1687
 *     if __nvshmemx_get_uniqueid == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_get_uniqueid = dlsym(handle, 'nvshmemx_get_uniqueid')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1687, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1686
 *     __nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_get_uniqueid')
 *     if __nvshmemx_get_uniqueid == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_get_uniqueid = dlsym(handle, 'nvshmemx_get_uniqueid')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1688
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_get_uniqueid = dlsym(handle, 'nvshmemx_get_uniqueid')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_cumodule_init
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_get_uniqueid);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1685
 *     global __nvshmemx_get_uniqueid
 *     __nvshmemx_get_uniqueid = dlsym(RTLD_DEFAULT, 'nvshmemx_get_uniqueid')
 *     if __nvshmemx_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1691
 * 
 *     global __nvshmemx_cumodule_init
 *     __nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_init')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_cumodule_init == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_cumodule_init);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1692
 *     global __nvshmemx_cumodule_init
 *     __nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_init')
 *     if __nvshmemx_cumodule_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1693
 *     __nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_init')
 *     if __nvshmemx_cumodule_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_cumodule_init = dlsym(handle, 'nvshmemx_cumodule_init')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1694
 *     if __nvshmemx_cumodule_init == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_cumodule_init = dlsym(handle, 'nvshmemx_cumodule_init')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1694, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1693
 *     __nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_init')
 *     if __nvshmemx_cumodule_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_cumodule_init = dlsym(handle, 'nvshmemx_cumodule_init')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1695
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_cumodule_init = dlsym(handle, 'nvshmemx_cumodule_init')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_cumodule_finalize
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_cumodule_init);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1692
 *     global __nvshmemx_cumodule_init
 *     __nvshmemx_cumodule_init = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_init')
 *     if __nvshmemx_cumodule_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1698
 * 
 *     global __nvshmemx_cumodule_finalize
 *     __nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_finalize')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_cumodule_finalize);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1699
 *     global __nvshmemx_cumodule_finalize
 *     __nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_finalize')
 *     if __nvshmemx_cumodule_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1700
 *     __nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_finalize')
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_cumodule_finalize = dlsym(handle, 'nvshmemx_cumodule_finalize')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1701
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_cumodule_finalize = dlsym(handle, 'nvshmemx_cumodule_finalize')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1701, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1700
 *     __nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_finalize')
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_cumodule_finalize = dlsym(handle, 'nvshmemx_cumodule_finalize')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1702
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_cumodule_finalize = dlsym(handle, 'nvshmemx_cumodule_finalize')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_buffer_register_symmetric
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_cumodule_finalize);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1699
 *     global __nvshmemx_cumodule_finalize
 *     __nvshmemx_cumodule_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_cumodule_finalize')
 *     if __nvshmemx_cumodule_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1705
 * 
 *     global __nvshmemx_buffer_register_symmetric
 *     __nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_register_symmetric')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_buffer_register_symmetr);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1706
 *     global __nvshmemx_buffer_register_symmetric
 *     __nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_register_symmetric')
 *     if __nvshmemx_buffer_register_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1707
 *     __nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_register_symmetric')
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_buffer_register_symmetric = dlsym(handle, 'nvshmemx_buffer_register_symmetric')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1708
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_buffer_register_symmetric = dlsym(handle, 'nvshmemx_buffer_register_symmetric')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1708, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1707
 *     __nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_register_symmetric')
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_buffer_register_symmetric = dlsym(handle, 'nvshmemx_buffer_register_symmetric')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1709
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_buffer_register_symmetric = dlsym(handle, 'nvshmemx_buffer_register_symmetric')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_buffer_unregister_symmetric
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_buffer_register_symmetr);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1706
 *     global __nvshmemx_buffer_register_symmetric
 *     __nvshmemx_buffer_register_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_register_symmetric')
 *     if __nvshmemx_buffer_register_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1712
 * 
 *     global __nvshmemx_buffer_unregister_symmetric
 *     __nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_unregister_symmetric')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_buffer_unregister_symme);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1713
 *     global __nvshmemx_buffer_unregister_symmetric
 *     __nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_unregister_symmetric')
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1714
 *     __nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_unregister_symmetric')
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_buffer_unregister_symmetric = dlsym(handle, 'nvshmemx_buffer_unregister_symmetric')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1715
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_buffer_unregister_symmetric = dlsym(handle, 'nvshmemx_buffer_unregister_symmetric')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1715, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1714
 *     __nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_unregister_symmetric')
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_buffer_unregister_symmetric = dlsym(handle, 'nvshmemx_buffer_unregister_symmetric')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1716
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_buffer_unregister_symmetric = dlsym(handle, 'nvshmemx_buffer_unregister_symmetric')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_culibrary_init
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_buffer_unregister_symme);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1713
 *     global __nvshmemx_buffer_unregister_symmetric
 *     __nvshmemx_buffer_unregister_symmetric = dlsym(RTLD_DEFAULT, 'nvshmemx_buffer_unregister_symmetric')
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1719
 * 
 *     global __nvshmemx_culibrary_init
 *     __nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_init')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_culibrary_init == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_culibrary_init);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1720
 *     global __nvshmemx_culibrary_init
 *     __nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_init')
 *     if __nvshmemx_culibrary_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1721
 *     __nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_init')
 *     if __nvshmemx_culibrary_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_culibrary_init = dlsym(handle, 'nvshmemx_culibrary_init')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1722
 *     if __nvshmemx_culibrary_init == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_culibrary_init = dlsym(handle, 'nvshmemx_culibrary_init')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1722, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1721
 *     __nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_init')
 *     if __nvshmemx_culibrary_init == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_culibrary_init = dlsym(handle, 'nvshmemx_culibrary_init')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1723
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_culibrary_init = dlsym(handle, 'nvshmemx_culibrary_init')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_culibrary_finalize
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_culibrary_init);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1720
 *     global __nvshmemx_culibrary_init
 *     __nvshmemx_culibrary_init = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_init')
 *     if __nvshmemx_culibrary_init == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1726
 * 
 *     global __nvshmemx_culibrary_finalize
 *     __nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_finalize')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_culibrary_finalize);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1727
 *     global __nvshmemx_culibrary_finalize
 *     __nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_finalize')
 *     if __nvshmemx_culibrary_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1728
 *     __nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_finalize')
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_culibrary_finalize = dlsym(handle, 'nvshmemx_culibrary_finalize')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1729
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_culibrary_finalize = dlsym(handle, 'nvshmemx_culibrary_finalize')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1729, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1728
 *     __nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_finalize')
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_culibrary_finalize = dlsym(handle, 'nvshmemx_culibrary_finalize')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1730
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_culibrary_finalize = dlsym(handle, 'nvshmemx_culibrary_finalize')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_putmem_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_culibrary_finalize);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1727
 *     global __nvshmemx_culibrary_finalize
 *     __nvshmemx_culibrary_finalize = dlsym(RTLD_DEFAULT, 'nvshmemx_culibrary_finalize')
 *     if __nvshmemx_culibrary_finalize == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1733
 * 
 *     global __nvshmemx_putmem_on_stream
 *     __nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_putmem_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1734
 *     global __nvshmemx_putmem_on_stream
 *     __nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_on_stream')
 *     if __nvshmemx_putmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1735
 *     __nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_on_stream')
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_putmem_on_stream = dlsym(handle, 'nvshmemx_putmem_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1736
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_putmem_on_stream = dlsym(handle, 'nvshmemx_putmem_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1736, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1735
 *     __nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_on_stream')
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_putmem_on_stream = dlsym(handle, 'nvshmemx_putmem_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1737
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_putmem_on_stream = dlsym(handle, 'nvshmemx_putmem_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_putmem_signal_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_putmem_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1734
 *     global __nvshmemx_putmem_on_stream
 *     __nvshmemx_putmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_on_stream')
 *     if __nvshmemx_putmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1740
 * 
 *     global __nvshmemx_putmem_signal_on_stream
 *     __nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_signal_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_putmem_signal_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1741
 *     global __nvshmemx_putmem_signal_on_stream
 *     __nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_signal_on_stream')
 *     if __nvshmemx_putmem_signal_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1742
 *     __nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_signal_on_stream')
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_putmem_signal_on_stream = dlsym(handle, 'nvshmemx_putmem_signal_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1743
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_putmem_signal_on_stream = dlsym(handle, 'nvshmemx_putmem_signal_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1743, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1742
 *     __nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_signal_on_stream')
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_putmem_signal_on_stream = dlsym(handle, 'nvshmemx_putmem_signal_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1744
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_putmem_signal_on_stream = dlsym(handle, 'nvshmemx_putmem_signal_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_getmem_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_putmem_signal_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1741
 *     global __nvshmemx_putmem_signal_on_stream
 *     __nvshmemx_putmem_signal_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_putmem_signal_on_stream')
 *     if __nvshmemx_putmem_signal_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1747
 * 
 *     global __nvshmemx_getmem_on_stream
 *     __nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_getmem_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_getmem_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1748
 *     global __nvshmemx_getmem_on_stream
 *     __nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_getmem_on_stream')
 *     if __nvshmemx_getmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1749
 *     __nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_getmem_on_stream')
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_getmem_on_stream = dlsym(handle, 'nvshmemx_getmem_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1750
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_getmem_on_stream = dlsym(handle, 'nvshmemx_getmem_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1750, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1749
 *     __nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_getmem_on_stream')
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_getmem_on_stream = dlsym(handle, 'nvshmemx_getmem_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1751
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_getmem_on_stream = dlsym(handle, 'nvshmemx_getmem_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_quiet_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_getmem_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1748
 *     global __nvshmemx_getmem_on_stream
 *     __nvshmemx_getmem_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_getmem_on_stream')
 *     if __nvshmemx_getmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1754
 * 
 *     global __nvshmemx_quiet_on_stream
 *     __nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_quiet_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_quiet_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1755
 *     global __nvshmemx_quiet_on_stream
 *     __nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_quiet_on_stream')
 *     if __nvshmemx_quiet_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1756
 *     __nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_quiet_on_stream')
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_quiet_on_stream = dlsym(handle, 'nvshmemx_quiet_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1757
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_quiet_on_stream = dlsym(handle, 'nvshmemx_quiet_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1757, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1756
 *     __nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_quiet_on_stream')
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_quiet_on_stream = dlsym(handle, 'nvshmemx_quiet_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1758
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_quiet_on_stream = dlsym(handle, 'nvshmemx_quiet_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_signal_op_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_quiet_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1755
 *     global __nvshmemx_quiet_on_stream
 *     __nvshmemx_quiet_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_quiet_on_stream')
 *     if __nvshmemx_quiet_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1761
 * 
 *     global __nvshmemx_signal_op_on_stream
 *     __nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_op_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_signal_op_on_stream);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1762
 *     global __nvshmemx_signal_op_on_stream
 *     __nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_op_on_stream')
 *     if __nvshmemx_signal_op_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1763
 *     __nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_op_on_stream')
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_signal_op_on_stream = dlsym(handle, 'nvshmemx_signal_op_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1764
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_signal_op_on_stream = dlsym(handle, 'nvshmemx_signal_op_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1764, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1763
 *     __nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_op_on_stream')
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_signal_op_on_stream = dlsym(handle, 'nvshmemx_signal_op_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1765
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_signal_op_on_stream = dlsym(handle, 'nvshmemx_signal_op_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_signal_wait_until_on_stream
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_signal_op_on_stream);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1762
 *     global __nvshmemx_signal_op_on_stream
 *     __nvshmemx_signal_op_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_op_on_stream')
 *     if __nvshmemx_signal_op_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1768
 * 
 *     global __nvshmemx_signal_wait_until_on_stream
 *     __nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_wait_until_on_stream')             # <<<<<<<<<<<<<<
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         if handle == NULL:
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, __pyx_k_nvshmemx_signal_wait_until_on_st);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1769
 *     global __nvshmemx_signal_wait_until_on_stream
 *     __nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_wait_until_on_stream')
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream == NULL);
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1770
 *     __nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_wait_until_on_stream')
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_signal_wait_until_on_stream = dlsym(handle, 'nvshmemx_signal_wait_until_on_stream')
*/
    __pyx_t_1 = (__pyx_v_handle == NULL);
    if (__pyx_t_1) {

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1771
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         if handle == NULL:
 *             handle = load_library()             # <<<<<<<<<<<<<<
 *         __nvshmemx_signal_wait_until_on_stream = dlsym(handle, 'nvshmemx_signal_wait_until_on_stream')
 * 
*/
      __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem_load_library(); if (unlikely(__pyx_t_2 == ((void *)NULL) && __Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1771, __pyx_L1_error)
      __pyx_v_handle = __pyx_t_2;

      /* "nvshmem/bindings/_internal/nvshmem.pyx":1770
 *     __nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_wait_until_on_stream')
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         if handle == NULL:             # <<<<<<<<<<<<<<
 *             handle = load_library()
 *         __nvshmemx_signal_wait_until_on_stream = dlsym(handle, 'nvshmemx_signal_wait_until_on_stream')
*/
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1772
 *         if handle == NULL:
 *             handle = load_library()
 *         __nvshmemx_signal_wait_until_on_stream = dlsym(handle, 'nvshmemx_signal_wait_until_on_stream')             # <<<<<<<<<<<<<<
 * 
 *     __py_nvshmem_init = True
*/
    __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream = dlsym(__pyx_v_handle, __pyx_k_nvshmemx_signal_wait_until_on_st);

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1769
 *     global __nvshmemx_signal_wait_until_on_stream
 *     __nvshmemx_signal_wait_until_on_stream = dlsym(RTLD_DEFAULT, 'nvshmemx_signal_wait_until_on_stream')
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         if handle == NULL:
 *             handle = load_library()
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1774
 *         __nvshmemx_signal_wait_until_on_stream = dlsym(handle, 'nvshmemx_signal_wait_until_on_stream')
 * 
 *     __py_nvshmem_init = True             # <<<<<<<<<<<<<<
 *     return 0
 * 
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___py_nvshmem_init = 1;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1775
 * 
 *     __py_nvshmem_init = True
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":269
 * 
 * 
 * cdef int _check_or_init_nvshmem() except -1 nogil:             # <<<<<<<<<<<<<<
 *     global __py_nvshmem_init
 *     if __py_nvshmem_init:
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._check_or_init_nvshmem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":1781
 * 
 * 
 * cpdef dict _inspect_function_pointers():             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is not None:
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_inspect_function_pointers", 0);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1783
 * cpdef dict _inspect_function_pointers():
 *     global func_ptrs
 *     if func_ptrs is not None:             # <<<<<<<<<<<<<<
 *         return func_ptrs
 * 
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1784
 *     global func_ptrs
 *     if func_ptrs is not None:
 *         return func_ptrs             # <<<<<<<<<<<<<<
 * 
 *     _check_or_init_nvshmem()
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs);
    __pyx_r = __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs;
    goto __pyx_L0;

    /* "nvshmem/bindings/_internal/nvshmem.pyx":1783
 * cpdef dict _inspect_function_pointers():
 *     global func_ptrs
 *     if func_ptrs is not None:             # <<<<<<<<<<<<<<
 *         return func_ptrs
 * 
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1786
 *         return func_ptrs
 * 
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     cdef dict data = {}
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1786, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1787
 * 
 *     _check_or_init_nvshmem()
 *     cdef dict data = {}             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_barrier
*/
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_data = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1790
 * 
 *     global __nvshmem_barrier
 *     data["__nvshmem_barrier"] = <intptr_t>__nvshmem_barrier             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_barrier_all
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_barrier_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1793
 * 
 *     global __nvshmem_barrier_all
 *     data["__nvshmem_barrier_all"] = <intptr_t>__nvshmem_barrier_all             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_init_status
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_barrier_all_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1796
 * 
 *     global __nvshmemx_init_status
 *     data["__nvshmemx_init_status"] = <intptr_t>__nvshmemx_init_status             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_my_pe
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_init_status_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1796, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1799
 * 
 *     global __nvshmem_my_pe
 *     data["__nvshmem_my_pe"] = <intptr_t>__nvshmem_my_pe             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_n_pes
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_my_pe_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1802
 * 
 *     global __nvshmem_n_pes
 *     data["__nvshmem_n_pes"] = <intptr_t>__nvshmem_n_pes             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_info_get_version
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_n_pes_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1805
 * 
 *     global __nvshmem_info_get_version
 *     data["__nvshmem_info_get_version"] = <intptr_t>__nvshmem_info_get_version             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_vendor_get_version_info
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_info_get_version_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1808
 * 
 *     global __nvshmemx_vendor_get_version_info
 *     data["__nvshmemx_vendor_get_version_info"] = <intptr_t>__nvshmemx_vendor_get_version_info             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_malloc
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_vendor_get_version_in, __pyx_t_3) < 0))) __PYX_ERR(0, 1808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1811
 * 
 *     global __nvshmem_malloc
 *     data["__nvshmem_malloc"] = <intptr_t>__nvshmem_malloc             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_calloc
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_malloc_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1814
 * 
 *     global __nvshmem_calloc
 *     data["__nvshmem_calloc"] = <intptr_t>__nvshmem_calloc             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_align
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_calloc_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1817
 * 
 *     global __nvshmem_align
 *     data["__nvshmem_align"] = <intptr_t>__nvshmem_align             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_free
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_align_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1820
 * 
 *     global __nvshmem_free
 *     data["__nvshmem_free"] = <intptr_t>__nvshmem_free             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_ptr
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_free_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1820, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1823
 * 
 *     global __nvshmem_ptr
 *     data["__nvshmem_ptr"] = <intptr_t>__nvshmem_ptr             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_mc_ptr
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_ptr_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1826
 * 
 *     global __nvshmemx_mc_ptr
 *     data["__nvshmemx_mc_ptr"] = <intptr_t>__nvshmemx_mc_ptr             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_my_pe
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_mc_ptr_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1829
 * 
 *     global __nvshmem_team_my_pe
 *     data["__nvshmem_team_my_pe"] = <intptr_t>__nvshmem_team_my_pe             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_n_pes
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_my_pe_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1832
 * 
 *     global __nvshmem_team_n_pes
 *     data["__nvshmem_team_n_pes"] = <intptr_t>__nvshmem_team_n_pes             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_get_config
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_n_pes_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1835
 * 
 *     global __nvshmem_team_get_config
 *     data["__nvshmem_team_get_config"] = <intptr_t>__nvshmem_team_get_config             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_translate_pe
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_get_config_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1838
 * 
 *     global __nvshmem_team_translate_pe
 *     data["__nvshmem_team_translate_pe"] = <intptr_t>__nvshmem_team_translate_pe             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_split_strided
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_translate_pe_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1841
 * 
 *     global __nvshmem_team_split_strided
 *     data["__nvshmem_team_split_strided"] = <intptr_t>__nvshmem_team_split_strided             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_get_uniqueid
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_split_strided_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1844
 * 
 *     global __nvshmemx_team_get_uniqueid
 *     data["__nvshmemx_team_get_uniqueid"] = <intptr_t>__nvshmemx_team_get_uniqueid             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_init
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_team_get_uniqueid_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1847
 * 
 *     global __nvshmemx_team_init
 *     data["__nvshmemx_team_init"] = <intptr_t>__nvshmemx_team_init             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_split_2d
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_team_init_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1850
 * 
 *     global __nvshmem_team_split_2d
 *     data["__nvshmem_team_split_2d"] = <intptr_t>__nvshmem_team_split_2d             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmem_team_destroy
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_split_2d_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1853
 * 
 *     global __nvshmem_team_destroy
 *     data["__nvshmem_team_destroy"] = <intptr_t>__nvshmem_team_destroy             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmem_team_destroy_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1856
 * 
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     data["__nvshmemx_bfloat16_alltoall_on_stream"] = <intptr_t>__nvshmemx_bfloat16_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_alltoall_on, __pyx_t_3) < 0))) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1859
 * 
 *     global __nvshmemx_half_alltoall_on_stream
 *     data["__nvshmemx_half_alltoall_on_stream"] = <intptr_t>__nvshmemx_half_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_alltoall_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1862
 * 
 *     global __nvshmemx_float_alltoall_on_stream
 *     data["__nvshmemx_float_alltoall_on_stream"] = <intptr_t>__nvshmemx_float_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1865
 * 
 *     global __nvshmemx_double_alltoall_on_stream
 *     data["__nvshmemx_double_alltoall_on_stream"] = <intptr_t>__nvshmemx_double_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_alltoall_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1868
 * 
 *     global __nvshmemx_char_alltoall_on_stream
 *     data["__nvshmemx_char_alltoall_on_stream"] = <intptr_t>__nvshmemx_char_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_alltoall_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1871
 * 
 *     global __nvshmemx_short_alltoall_on_stream
 *     data["__nvshmemx_short_alltoall_on_stream"] = <intptr_t>__nvshmemx_short_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1874
 * 
 *     global __nvshmemx_schar_alltoall_on_stream
 *     data["__nvshmemx_schar_alltoall_on_stream"] = <intptr_t>__nvshmemx_schar_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1877
 * 
 *     global __nvshmemx_int_alltoall_on_stream
 *     data["__nvshmemx_int_alltoall_on_stream"] = <intptr_t>__nvshmemx_int_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_alltoall_on_strea, __pyx_t_3) < 0))) __PYX_ERR(0, 1877, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1880
 * 
 *     global __nvshmemx_long_alltoall_on_stream
 *     data["__nvshmemx_long_alltoall_on_stream"] = <intptr_t>__nvshmemx_long_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_alltoall_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1883
 * 
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     data["__nvshmemx_longlong_alltoall_on_stream"] = <intptr_t>__nvshmemx_longlong_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_alltoall_on, __pyx_t_3) < 0))) __PYX_ERR(0, 1883, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1886
 * 
 *     global __nvshmemx_int8_alltoall_on_stream
 *     data["__nvshmemx_int8_alltoall_on_stream"] = <intptr_t>__nvshmemx_int8_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_alltoall_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1889
 * 
 *     global __nvshmemx_int16_alltoall_on_stream
 *     data["__nvshmemx_int16_alltoall_on_stream"] = <intptr_t>__nvshmemx_int16_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1892
 * 
 *     global __nvshmemx_int32_alltoall_on_stream
 *     data["__nvshmemx_int32_alltoall_on_stream"] = <intptr_t>__nvshmemx_int32_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1895
 * 
 *     global __nvshmemx_int64_alltoall_on_stream
 *     data["__nvshmemx_int64_alltoall_on_stream"] = <intptr_t>__nvshmemx_int64_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1898
 * 
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     data["__nvshmemx_uint8_alltoall_on_stream"] = <intptr_t>__nvshmemx_uint8_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_alltoall_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1901
 * 
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     data["__nvshmemx_uint16_alltoall_on_stream"] = <intptr_t>__nvshmemx_uint16_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_alltoall_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1904
 * 
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     data["__nvshmemx_uint32_alltoall_on_stream"] = <intptr_t>__nvshmemx_uint32_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_alltoall_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1907
 * 
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     data["__nvshmemx_uint64_alltoall_on_stream"] = <intptr_t>__nvshmemx_uint64_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_alltoall_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_alltoall_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1910
 * 
 *     global __nvshmemx_size_alltoall_on_stream
 *     data["__nvshmemx_size_alltoall_on_stream"] = <intptr_t>__nvshmemx_size_alltoall_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_barrier_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_alltoall_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1913
 * 
 *     global __nvshmemx_barrier_on_stream
 *     data["__nvshmemx_barrier_on_stream"] = <intptr_t>__nvshmemx_barrier_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_barrier_all_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_barrier_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1916
 * 
 *     global __nvshmemx_barrier_all_on_stream
 *     data["__nvshmemx_barrier_all_on_stream"] = <intptr_t>__nvshmemx_barrier_all_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_team_sync_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_barrier_all_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1919
 * 
 *     global __nvshmemx_team_sync_on_stream
 *     data["__nvshmemx_team_sync_on_stream"] = <intptr_t>__nvshmemx_team_sync_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_sync_all_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_team_sync_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1922
 * 
 *     global __nvshmemx_sync_all_on_stream
 *     data["__nvshmemx_sync_all_on_stream"] = <intptr_t>__nvshmemx_sync_all_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_sync_all_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 1922, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1925
 * 
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     data["__nvshmemx_bfloat16_broadcast_on_stream"] = <intptr_t>__nvshmemx_bfloat16_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_broadcast_on, __pyx_t_3) < 0))) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1928
 * 
 *     global __nvshmemx_half_broadcast_on_stream
 *     data["__nvshmemx_half_broadcast_on_stream"] = <intptr_t>__nvshmemx_half_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_broadcast_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1931
 * 
 *     global __nvshmemx_float_broadcast_on_stream
 *     data["__nvshmemx_float_broadcast_on_stream"] = <intptr_t>__nvshmemx_float_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1931, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1934
 * 
 *     global __nvshmemx_double_broadcast_on_stream
 *     data["__nvshmemx_double_broadcast_on_stream"] = <intptr_t>__nvshmemx_double_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_broadcast_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1937
 * 
 *     global __nvshmemx_char_broadcast_on_stream
 *     data["__nvshmemx_char_broadcast_on_stream"] = <intptr_t>__nvshmemx_char_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_broadcast_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1937, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1940
 * 
 *     global __nvshmemx_short_broadcast_on_stream
 *     data["__nvshmemx_short_broadcast_on_stream"] = <intptr_t>__nvshmemx_short_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1943
 * 
 *     global __nvshmemx_schar_broadcast_on_stream
 *     data["__nvshmemx_schar_broadcast_on_stream"] = <intptr_t>__nvshmemx_schar_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1946
 * 
 *     global __nvshmemx_int_broadcast_on_stream
 *     data["__nvshmemx_int_broadcast_on_stream"] = <intptr_t>__nvshmemx_int_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_broadcast_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1949
 * 
 *     global __nvshmemx_long_broadcast_on_stream
 *     data["__nvshmemx_long_broadcast_on_stream"] = <intptr_t>__nvshmemx_long_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_broadcast_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1952
 * 
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     data["__nvshmemx_longlong_broadcast_on_stream"] = <intptr_t>__nvshmemx_longlong_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_broadcast_on, __pyx_t_3) < 0))) __PYX_ERR(0, 1952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1955
 * 
 *     global __nvshmemx_int8_broadcast_on_stream
 *     data["__nvshmemx_int8_broadcast_on_stream"] = <intptr_t>__nvshmemx_int8_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_broadcast_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1958
 * 
 *     global __nvshmemx_int16_broadcast_on_stream
 *     data["__nvshmemx_int16_broadcast_on_stream"] = <intptr_t>__nvshmemx_int16_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1961
 * 
 *     global __nvshmemx_int32_broadcast_on_stream
 *     data["__nvshmemx_int32_broadcast_on_stream"] = <intptr_t>__nvshmemx_int32_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1964
 * 
 *     global __nvshmemx_int64_broadcast_on_stream
 *     data["__nvshmemx_int64_broadcast_on_stream"] = <intptr_t>__nvshmemx_int64_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1967
 * 
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     data["__nvshmemx_uint8_broadcast_on_stream"] = <intptr_t>__nvshmemx_uint8_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_broadcast_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1970
 * 
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     data["__nvshmemx_uint16_broadcast_on_stream"] = <intptr_t>__nvshmemx_uint16_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_broadcast_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 1970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1973
 * 
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     data["__nvshmemx_uint32_broadcast_on_stream"] = <intptr_t>__nvshmemx_uint32_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_broadcast_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1976
 * 
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     data["__nvshmemx_uint64_broadcast_on_stream"] = <intptr_t>__nvshmemx_uint64_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_broadcast_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_broadcast_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 1976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1979
 * 
 *     global __nvshmemx_size_broadcast_on_stream
 *     data["__nvshmemx_size_broadcast_on_stream"] = <intptr_t>__nvshmemx_size_broadcast_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_broadcast_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1982
 * 
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     data["__nvshmemx_bfloat16_fcollect_on_stream"] = <intptr_t>__nvshmemx_bfloat16_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_fcollect_on, __pyx_t_3) < 0))) __PYX_ERR(0, 1982, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1985
 * 
 *     global __nvshmemx_half_fcollect_on_stream
 *     data["__nvshmemx_half_fcollect_on_stream"] = <intptr_t>__nvshmemx_half_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_fcollect_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1988
 * 
 *     global __nvshmemx_float_fcollect_on_stream
 *     data["__nvshmemx_float_fcollect_on_stream"] = <intptr_t>__nvshmemx_float_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1991
 * 
 *     global __nvshmemx_double_fcollect_on_stream
 *     data["__nvshmemx_double_fcollect_on_stream"] = <intptr_t>__nvshmemx_double_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_fcollect_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 1991, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1994
 * 
 *     global __nvshmemx_char_fcollect_on_stream
 *     data["__nvshmemx_char_fcollect_on_stream"] = <intptr_t>__nvshmemx_char_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_fcollect_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 1994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1997
 * 
 *     global __nvshmemx_short_fcollect_on_stream
 *     data["__nvshmemx_short_fcollect_on_stream"] = <intptr_t>__nvshmemx_short_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2000
 * 
 *     global __nvshmemx_schar_fcollect_on_stream
 *     data["__nvshmemx_schar_fcollect_on_stream"] = <intptr_t>__nvshmemx_schar_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2000, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2003
 * 
 *     global __nvshmemx_int_fcollect_on_stream
 *     data["__nvshmemx_int_fcollect_on_stream"] = <intptr_t>__nvshmemx_int_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_fcollect_on_strea, __pyx_t_3) < 0))) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2006
 * 
 *     global __nvshmemx_long_fcollect_on_stream
 *     data["__nvshmemx_long_fcollect_on_stream"] = <intptr_t>__nvshmemx_long_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_fcollect_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 2006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2009
 * 
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     data["__nvshmemx_longlong_fcollect_on_stream"] = <intptr_t>__nvshmemx_longlong_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_fcollect_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2012
 * 
 *     global __nvshmemx_int8_fcollect_on_stream
 *     data["__nvshmemx_int8_fcollect_on_stream"] = <intptr_t>__nvshmemx_int8_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_fcollect_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2015
 * 
 *     global __nvshmemx_int16_fcollect_on_stream
 *     data["__nvshmemx_int16_fcollect_on_stream"] = <intptr_t>__nvshmemx_int16_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2018
 * 
 *     global __nvshmemx_int32_fcollect_on_stream
 *     data["__nvshmemx_int32_fcollect_on_stream"] = <intptr_t>__nvshmemx_int32_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2021
 * 
 *     global __nvshmemx_int64_fcollect_on_stream
 *     data["__nvshmemx_int64_fcollect_on_stream"] = <intptr_t>__nvshmemx_int64_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2024
 * 
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     data["__nvshmemx_uint8_fcollect_on_stream"] = <intptr_t>__nvshmemx_uint8_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_fcollect_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2027
 * 
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     data["__nvshmemx_uint16_fcollect_on_stream"] = <intptr_t>__nvshmemx_uint16_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_fcollect_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2030
 * 
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     data["__nvshmemx_uint32_fcollect_on_stream"] = <intptr_t>__nvshmemx_uint32_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_fcollect_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2033
 * 
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     data["__nvshmemx_uint64_fcollect_on_stream"] = <intptr_t>__nvshmemx_uint64_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_fcollect_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_fcollect_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2036
 * 
 *     global __nvshmemx_size_fcollect_on_stream
 *     data["__nvshmemx_size_fcollect_on_stream"] = <intptr_t>__nvshmemx_size_fcollect_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_fcollect_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 2036, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2039
 * 
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     data["__nvshmemx_int8_max_reduce_on_stream"] = <intptr_t>__nvshmemx_int8_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_max_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2042
 * 
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     data["__nvshmemx_int16_max_reduce_on_stream"] = <intptr_t>__nvshmemx_int16_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2042, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2045
 * 
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     data["__nvshmemx_int32_max_reduce_on_stream"] = <intptr_t>__nvshmemx_int32_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2045, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2048
 * 
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     data["__nvshmemx_int64_max_reduce_on_stream"] = <intptr_t>__nvshmemx_int64_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2051
 * 
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     data["__nvshmemx_uint8_max_reduce_on_stream"] = <intptr_t>__nvshmemx_uint8_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2054
 * 
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     data["__nvshmemx_uint16_max_reduce_on_stream"] = <intptr_t>__nvshmemx_uint16_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_max_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2057
 * 
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     data["__nvshmemx_uint32_max_reduce_on_stream"] = <intptr_t>__nvshmemx_uint32_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_max_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2057, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2060
 * 
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     data["__nvshmemx_uint64_max_reduce_on_stream"] = <intptr_t>__nvshmemx_uint64_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_max_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2063
 * 
 *     global __nvshmemx_size_max_reduce_on_stream
 *     data["__nvshmemx_size_max_reduce_on_stream"] = <intptr_t>__nvshmemx_size_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_max_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2066
 * 
 *     global __nvshmemx_char_max_reduce_on_stream
 *     data["__nvshmemx_char_max_reduce_on_stream"] = <intptr_t>__nvshmemx_char_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_max_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2069
 * 
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     data["__nvshmemx_schar_max_reduce_on_stream"] = <intptr_t>__nvshmemx_schar_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2072
 * 
 *     global __nvshmemx_short_max_reduce_on_stream
 *     data["__nvshmemx_short_max_reduce_on_stream"] = <intptr_t>__nvshmemx_short_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2075
 * 
 *     global __nvshmemx_int_max_reduce_on_stream
 *     data["__nvshmemx_int_max_reduce_on_stream"] = <intptr_t>__nvshmemx_int_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_max_reduce_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2075, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2078
 * 
 *     global __nvshmemx_long_max_reduce_on_stream
 *     data["__nvshmemx_long_max_reduce_on_stream"] = <intptr_t>__nvshmemx_long_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_max_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2081
 * 
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     data["__nvshmemx_longlong_max_reduce_on_stream"] = <intptr_t>__nvshmemx_longlong_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_max_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2084
 * 
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     data["__nvshmemx_bfloat16_max_reduce_on_stream"] = <intptr_t>__nvshmemx_bfloat16_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_max_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2087
 * 
 *     global __nvshmemx_half_max_reduce_on_stream
 *     data["__nvshmemx_half_max_reduce_on_stream"] = <intptr_t>__nvshmemx_half_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_max_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2090
 * 
 *     global __nvshmemx_float_max_reduce_on_stream
 *     data["__nvshmemx_float_max_reduce_on_stream"] = <intptr_t>__nvshmemx_float_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_max_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_max_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2093
 * 
 *     global __nvshmemx_double_max_reduce_on_stream
 *     data["__nvshmemx_double_max_reduce_on_stream"] = <intptr_t>__nvshmemx_double_max_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_max_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2096
 * 
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     data["__nvshmemx_int8_min_reduce_on_stream"] = <intptr_t>__nvshmemx_int8_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_min_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2096, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2099
 * 
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     data["__nvshmemx_int16_min_reduce_on_stream"] = <intptr_t>__nvshmemx_int16_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2102
 * 
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     data["__nvshmemx_int32_min_reduce_on_stream"] = <intptr_t>__nvshmemx_int32_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2105
 * 
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     data["__nvshmemx_int64_min_reduce_on_stream"] = <intptr_t>__nvshmemx_int64_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2108
 * 
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     data["__nvshmemx_uint8_min_reduce_on_stream"] = <intptr_t>__nvshmemx_uint8_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2111
 * 
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     data["__nvshmemx_uint16_min_reduce_on_stream"] = <intptr_t>__nvshmemx_uint16_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_min_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2114
 * 
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     data["__nvshmemx_uint32_min_reduce_on_stream"] = <intptr_t>__nvshmemx_uint32_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_min_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2117
 * 
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     data["__nvshmemx_uint64_min_reduce_on_stream"] = <intptr_t>__nvshmemx_uint64_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_min_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2120
 * 
 *     global __nvshmemx_size_min_reduce_on_stream
 *     data["__nvshmemx_size_min_reduce_on_stream"] = <intptr_t>__nvshmemx_size_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_min_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2123
 * 
 *     global __nvshmemx_char_min_reduce_on_stream
 *     data["__nvshmemx_char_min_reduce_on_stream"] = <intptr_t>__nvshmemx_char_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_min_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2126
 * 
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     data["__nvshmemx_schar_min_reduce_on_stream"] = <intptr_t>__nvshmemx_schar_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2129
 * 
 *     global __nvshmemx_short_min_reduce_on_stream
 *     data["__nvshmemx_short_min_reduce_on_stream"] = <intptr_t>__nvshmemx_short_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2132
 * 
 *     global __nvshmemx_int_min_reduce_on_stream
 *     data["__nvshmemx_int_min_reduce_on_stream"] = <intptr_t>__nvshmemx_int_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_min_reduce_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2135
 * 
 *     global __nvshmemx_long_min_reduce_on_stream
 *     data["__nvshmemx_long_min_reduce_on_stream"] = <intptr_t>__nvshmemx_long_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_min_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2138
 * 
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     data["__nvshmemx_longlong_min_reduce_on_stream"] = <intptr_t>__nvshmemx_longlong_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_min_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2141
 * 
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     data["__nvshmemx_bfloat16_min_reduce_on_stream"] = <intptr_t>__nvshmemx_bfloat16_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_min_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2144
 * 
 *     global __nvshmemx_half_min_reduce_on_stream
 *     data["__nvshmemx_half_min_reduce_on_stream"] = <intptr_t>__nvshmemx_half_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_min_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2147
 * 
 *     global __nvshmemx_float_min_reduce_on_stream
 *     data["__nvshmemx_float_min_reduce_on_stream"] = <intptr_t>__nvshmemx_float_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_min_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_min_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2150
 * 
 *     global __nvshmemx_double_min_reduce_on_stream
 *     data["__nvshmemx_double_min_reduce_on_stream"] = <intptr_t>__nvshmemx_double_min_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_min_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2153
 * 
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     data["__nvshmemx_int8_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_int8_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_sum_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2156
 * 
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     data["__nvshmemx_int16_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_int16_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2159
 * 
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     data["__nvshmemx_int32_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_int32_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2162
 * 
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     data["__nvshmemx_int64_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_int64_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2165
 * 
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     data["__nvshmemx_uint8_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_uint8_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2168
 * 
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     data["__nvshmemx_uint16_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_uint16_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_sum_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2171
 * 
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     data["__nvshmemx_uint32_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_uint32_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_sum_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2174
 * 
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     data["__nvshmemx_uint64_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_uint64_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_sum_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2177
 * 
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     data["__nvshmemx_size_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_size_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_sum_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2180
 * 
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     data["__nvshmemx_char_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_char_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_sum_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2183
 * 
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     data["__nvshmemx_schar_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_schar_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2186
 * 
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     data["__nvshmemx_short_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_short_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2189
 * 
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     data["__nvshmemx_int_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_int_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_sum_reduce_on_str, __pyx_t_3) < 0))) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2192
 * 
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     data["__nvshmemx_long_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_long_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_sum_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2195
 * 
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     data["__nvshmemx_longlong_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_longlong_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_sum_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2198
 * 
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     data["__nvshmemx_bfloat16_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_bfloat16_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_sum_reduce_o, __pyx_t_3) < 0))) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2201
 * 
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     data["__nvshmemx_half_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_half_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_sum_reduce_on_st, __pyx_t_3) < 0))) __PYX_ERR(0, 2201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2204
 * 
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     data["__nvshmemx_float_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_float_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_sum_reduce_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_sum_reduce_on_s, __pyx_t_3) < 0))) __PYX_ERR(0, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2207
 * 
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     data["__nvshmemx_double_sum_reduce_on_stream"] = <intptr_t>__nvshmemx_double_sum_reduce_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_sum_reduce_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2210
 * 
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     data["__nvshmemx_int8_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int8_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_max_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2213
 * 
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     data["__nvshmemx_int16_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int16_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2216
 * 
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     data["__nvshmemx_int32_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int32_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2219
 * 
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     data["__nvshmemx_int64_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int64_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2222
 * 
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     data["__nvshmemx_uint8_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint8_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2225
 * 
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     data["__nvshmemx_uint16_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint16_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_max_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2228
 * 
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     data["__nvshmemx_uint32_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint32_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_max_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2231
 * 
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     data["__nvshmemx_uint64_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint64_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_max_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2234
 * 
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     data["__nvshmemx_size_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_size_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_max_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2237
 * 
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     data["__nvshmemx_char_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_char_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_max_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2240
 * 
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     data["__nvshmemx_schar_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_schar_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2243
 * 
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     data["__nvshmemx_short_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_short_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2246
 * 
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     data["__nvshmemx_int_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_max_reducescatter, __pyx_t_3) < 0))) __PYX_ERR(0, 2246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2249
 * 
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     data["__nvshmemx_long_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_long_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_max_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2252
 * 
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     data["__nvshmemx_longlong_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_longlong_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_max_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2255
 * 
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     data["__nvshmemx_bfloat16_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_bfloat16_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_max_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2258
 * 
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     data["__nvshmemx_half_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_half_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_max_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2261
 * 
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     data["__nvshmemx_float_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_float_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_max_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_max_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2264
 * 
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     data["__nvshmemx_double_max_reducescatter_on_stream"] = <intptr_t>__nvshmemx_double_max_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_max_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2267
 * 
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     data["__nvshmemx_int8_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int8_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_min_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2270
 * 
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     data["__nvshmemx_int16_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int16_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2273
 * 
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     data["__nvshmemx_int32_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int32_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2276
 * 
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     data["__nvshmemx_int64_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int64_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2279
 * 
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     data["__nvshmemx_uint8_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint8_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2282
 * 
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     data["__nvshmemx_uint16_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint16_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_min_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2285
 * 
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     data["__nvshmemx_uint32_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint32_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_min_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2288
 * 
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     data["__nvshmemx_uint64_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint64_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_min_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2291
 * 
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     data["__nvshmemx_size_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_size_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_min_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2294
 * 
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     data["__nvshmemx_char_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_char_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_min_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2297
 * 
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     data["__nvshmemx_schar_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_schar_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2300
 * 
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     data["__nvshmemx_short_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_short_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2303
 * 
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     data["__nvshmemx_int_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_min_reducescatter, __pyx_t_3) < 0))) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2306
 * 
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     data["__nvshmemx_long_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_long_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_min_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2309
 * 
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     data["__nvshmemx_longlong_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_longlong_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_min_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2312
 * 
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     data["__nvshmemx_bfloat16_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_bfloat16_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_min_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2312, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2315
 * 
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     data["__nvshmemx_half_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_half_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_min_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2318
 * 
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     data["__nvshmemx_float_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_float_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_min_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_min_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2321
 * 
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     data["__nvshmemx_double_min_reducescatter_on_stream"] = <intptr_t>__nvshmemx_double_min_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_min_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2324
 * 
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     data["__nvshmemx_int8_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int8_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int8_sum_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2327
 * 
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     data["__nvshmemx_int16_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int16_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int16_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2330
 * 
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     data["__nvshmemx_int32_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int32_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int32_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2333
 * 
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     data["__nvshmemx_int64_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int64_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int64_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2336
 * 
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     data["__nvshmemx_uint8_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint8_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint8_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2339
 * 
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     data["__nvshmemx_uint16_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint16_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint16_sum_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2342
 * 
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     data["__nvshmemx_uint32_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint32_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint32_sum_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2345
 * 
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     data["__nvshmemx_uint64_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_uint64_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_size_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_uint64_sum_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2348
 * 
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     data["__nvshmemx_size_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_size_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_char_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_size_sum_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2351
 * 
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     data["__nvshmemx_char_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_char_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_char_sum_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2354
 * 
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     data["__nvshmemx_schar_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_schar_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_short_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_schar_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2357
 * 
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     data["__nvshmemx_short_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_short_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_int_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_short_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2360
 * 
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     data["__nvshmemx_int_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_int_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_long_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_int_sum_reducescatter, __pyx_t_3) < 0))) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2363
 * 
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     data["__nvshmemx_long_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_long_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_long_sum_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2366
 * 
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     data["__nvshmemx_longlong_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_longlong_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_longlong_sum_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2369
 * 
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     data["__nvshmemx_bfloat16_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_bfloat16_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_half_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_bfloat16_sum_reducesc, __pyx_t_3) < 0))) __PYX_ERR(0, 2369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2372
 * 
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     data["__nvshmemx_half_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_half_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_float_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_half_sum_reducescatte, __pyx_t_3) < 0))) __PYX_ERR(0, 2372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2375
 * 
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     data["__nvshmemx_float_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_float_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_double_sum_reducescatter_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_float_sum_reducescatt, __pyx_t_3) < 0))) __PYX_ERR(0, 2375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2378
 * 
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     data["__nvshmemx_double_sum_reducescatter_on_stream"] = <intptr_t>__nvshmemx_double_sum_reducescatter_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_hostlib_init_attr
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_double_sum_reducescat, __pyx_t_3) < 0))) __PYX_ERR(0, 2378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2381
 * 
 *     global __nvshmemx_hostlib_init_attr
 *     data["__nvshmemx_hostlib_init_attr"] = <intptr_t>__nvshmemx_hostlib_init_attr             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_hostlib_finalize
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_hostlib_init_attr_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2384
 * 
 *     global __nvshmemx_hostlib_finalize
 *     data["__nvshmemx_hostlib_finalize"] = <intptr_t>__nvshmemx_hostlib_finalize             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_set_attr_uniqueid_args
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_hostlib_finalize_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2387
 * 
 *     global __nvshmemx_set_attr_uniqueid_args
 *     data["__nvshmemx_set_attr_uniqueid_args"] = <intptr_t>__nvshmemx_set_attr_uniqueid_args             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_set_attr_mpi_comm_args
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_set_attr_uniqueid_arg, __pyx_t_3) < 0))) __PYX_ERR(0, 2387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2390
 * 
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     data["__nvshmemx_set_attr_mpi_comm_args"] = <intptr_t>__nvshmemx_set_attr_mpi_comm_args             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_get_uniqueid
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_set_attr_mpi_comm_arg, __pyx_t_3) < 0))) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2393
 * 
 *     global __nvshmemx_get_uniqueid
 *     data["__nvshmemx_get_uniqueid"] = <intptr_t>__nvshmemx_get_uniqueid             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_cumodule_init
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_get_uniqueid_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2396
 * 
 *     global __nvshmemx_cumodule_init
 *     data["__nvshmemx_cumodule_init"] = <intptr_t>__nvshmemx_cumodule_init             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_cumodule_finalize
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_cumodule_init_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2399
 * 
 *     global __nvshmemx_cumodule_finalize
 *     data["__nvshmemx_cumodule_finalize"] = <intptr_t>__nvshmemx_cumodule_finalize             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_buffer_register_symmetric
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_cumodule_finalize_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2402
 * 
 *     global __nvshmemx_buffer_register_symmetric
 *     data["__nvshmemx_buffer_register_symmetric"] = <intptr_t>__nvshmemx_buffer_register_symmetric             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_buffer_unregister_symmetric
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_buffer_register_symme, __pyx_t_3) < 0))) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2405
 * 
 *     global __nvshmemx_buffer_unregister_symmetric
 *     data["__nvshmemx_buffer_unregister_symmetric"] = <intptr_t>__nvshmemx_buffer_unregister_symmetric             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_culibrary_init
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_buffer_unregister_sym, __pyx_t_3) < 0))) __PYX_ERR(0, 2405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2408
 * 
 *     global __nvshmemx_culibrary_init
 *     data["__nvshmemx_culibrary_init"] = <intptr_t>__nvshmemx_culibrary_init             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_culibrary_finalize
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_culibrary_init_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2411
 * 
 *     global __nvshmemx_culibrary_finalize
 *     data["__nvshmemx_culibrary_finalize"] = <intptr_t>__nvshmemx_culibrary_finalize             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_putmem_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_culibrary_finalize_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2414
 * 
 *     global __nvshmemx_putmem_on_stream
 *     data["__nvshmemx_putmem_on_stream"] = <intptr_t>__nvshmemx_putmem_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_putmem_signal_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_putmem_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2417
 * 
 *     global __nvshmemx_putmem_signal_on_stream
 *     data["__nvshmemx_putmem_signal_on_stream"] = <intptr_t>__nvshmemx_putmem_signal_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_getmem_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_putmem_signal_on_stre, __pyx_t_3) < 0))) __PYX_ERR(0, 2417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2420
 * 
 *     global __nvshmemx_getmem_on_stream
 *     data["__nvshmemx_getmem_on_stream"] = <intptr_t>__nvshmemx_getmem_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_quiet_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_getmem_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2423
 * 
 *     global __nvshmemx_quiet_on_stream
 *     data["__nvshmemx_quiet_on_stream"] = <intptr_t>__nvshmemx_quiet_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_signal_op_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_quiet_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2426
 * 
 *     global __nvshmemx_signal_op_on_stream
 *     data["__nvshmemx_signal_op_on_stream"] = <intptr_t>__nvshmemx_signal_op_on_stream             # <<<<<<<<<<<<<<
 * 
 *     global __nvshmemx_signal_wait_until_on_stream
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_signal_op_on_stream_2, __pyx_t_3) < 0))) __PYX_ERR(0, 2426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2429
 * 
 *     global __nvshmemx_signal_wait_until_on_stream
 *     data["__nvshmemx_signal_wait_until_on_stream"] = <intptr_t>__nvshmemx_signal_wait_until_on_stream             # <<<<<<<<<<<<<<
 * 
 *     func_ptrs = data
*/
  __pyx_t_3 = PyLong_FromSsize_t(((intptr_t)__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_nvshmemx_signal_wait_until_on, __pyx_t_3) < 0))) __PYX_ERR(0, 2429, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2431
 *     data["__nvshmemx_signal_wait_until_on_stream"] = <intptr_t>__nvshmemx_signal_wait_until_on_stream
 * 
 *     func_ptrs = data             # <<<<<<<<<<<<<<
 *     return data
 * 
*/
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_XGOTREF(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs);
  __Pyx_DECREF_SET(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs, __pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2432
 * 
 *     func_ptrs = data
 *     return data             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1781
 * 
 * 
 * cpdef dict _inspect_function_pointers():             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is not None:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._inspect_function_pointers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers, "_inspect_function_pointers() -> dict");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers = {"_inspect_function_pointers", (PyCFunction)__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers};
static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_inspect_function_pointers (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_inspect_function_pointers", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._inspect_function_pointers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2435
 * 
 * 
 * cpdef _inspect_function_pointer(str name):             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is None:
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointer(PyObject *__pyx_v_name, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_inspect_function_pointer", 0);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2437
 * cpdef _inspect_function_pointer(str name):
 *     global func_ptrs
 *     if func_ptrs is None:             # <<<<<<<<<<<<<<
 *         func_ptrs = _inspect_function_pointers()
 *     return func_ptrs[name]
*/
  __pyx_t_1 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2438
 *     global func_ptrs
 *     if func_ptrs is None:
 *         func_ptrs = _inspect_function_pointers()             # <<<<<<<<<<<<<<
 *     return func_ptrs[name]
 * 
*/
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointers(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs);
    __Pyx_DECREF_SET(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs, ((PyObject*)__pyx_t_2));
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2437
 * cpdef _inspect_function_pointer(str name):
 *     global func_ptrs
 *     if func_ptrs is None:             # <<<<<<<<<<<<<<
 *         func_ptrs = _inspect_function_pointers()
 *     return func_ptrs[name]
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2439
 *     if func_ptrs is None:
 *         func_ptrs = _inspect_function_pointers()
 *     return func_ptrs[name]             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2439, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2435
 * 
 * 
 * cpdef _inspect_function_pointer(str name):             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is None:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._inspect_function_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_9_internal_7nvshmem_2_inspect_function_pointer, "_inspect_function_pointer(str name)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer = {"_inspect_function_pointer", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_9_internal_7nvshmem_2_inspect_function_pointer};
static PyObject *__pyx_pw_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_inspect_function_pointer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_name,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 2435, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 2435, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_inspect_function_pointer", 0) < (0)) __PYX_ERR(0, 2435, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_inspect_function_pointer", 1, 1, 1, i); __PYX_ERR(0, 2435, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 2435, __pyx_L3_error)
    }
    __pyx_v_name = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_inspect_function_pointer", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2435, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._inspect_function_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 2435, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem_2_inspect_function_pointer(__pyx_self, __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_9_internal_7nvshmem_2_inspect_function_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_inspect_function_pointer", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__inspect_function_pointer(__pyx_v_name, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._inspect_function_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2446
 * ###############################################################################
 * 
 * cdef int _nvshmem_barrier(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_barrier
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_barrier", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2448
 * cdef int _nvshmem_barrier(nvshmem_team_t team) except* nogil:
 *     global __nvshmem_barrier
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_barrier == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2448, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2449
 *     global __nvshmem_barrier
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2450
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_barrier)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2451
 *     if __nvshmem_barrier == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_barrier)(
 *         team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2451, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_barrier_is_not};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2451, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2451, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2450
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_barrier)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2449
 *     global __nvshmem_barrier
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2452
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_barrier)(             # <<<<<<<<<<<<<<
 *         team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier)(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2452, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2446
 * ###############################################################################
 * 
 * cdef int _nvshmem_barrier(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_barrier
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_barrier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2456
 * 
 * 
 * cdef void _nvshmem_barrier_all() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_barrier_all
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier_all(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_barrier_all", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2458
 * cdef void _nvshmem_barrier_all() except* nogil:
 *     global __nvshmem_barrier_all
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_barrier_all == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2458, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2459
 *     global __nvshmem_barrier_all
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier_all == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2460
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier_all == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")
 *     (<void (*)() nogil>__nvshmem_barrier_all)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2461
 *     if __nvshmem_barrier_all == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)() nogil>__nvshmem_barrier_all)(
 *         )
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2461, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_barrier_all_is};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2461, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2461, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2460
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier_all == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")
 *     (<void (*)() nogil>__nvshmem_barrier_all)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2459
 *     global __nvshmem_barrier_all
 *     _check_or_init_nvshmem()
 *     if __nvshmem_barrier_all == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2462
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_barrier_all is not found")
 *     (<void (*)() nogil>__nvshmem_barrier_all)(             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  ((void (*)(void))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all)(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2462, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2456
 * 
 * 
 * cdef void _nvshmem_barrier_all() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_barrier_all
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_barrier_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2466
 * 
 * 
 * cdef int _nvshmemx_init_status() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_init_status
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_init_status(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_init_status", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2468
 * cdef int _nvshmemx_init_status() except* nogil:
 *     global __nvshmemx_init_status
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_init_status == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2468, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2469
 *     global __nvshmemx_init_status
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_init_status == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_init_status == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")
 *     return (<int (*)() nogil>__nvshmemx_init_status)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2471
 *     if __nvshmemx_init_status == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)() nogil>__nvshmemx_init_status)(
 *         )
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2471, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_init_status_is};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2471, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2471, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_init_status == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")
 *     return (<int (*)() nogil>__nvshmemx_init_status)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2469
 *     global __nvshmemx_init_status
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_init_status == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2472
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_init_status is not found")
 *     return (<int (*)() nogil>__nvshmemx_init_status)(             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_1 = ((int (*)(void))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status)(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2472, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2466
 * 
 * 
 * cdef int _nvshmemx_init_status() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_init_status
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_init_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2476
 * 
 * 
 * cdef int _nvshmem_my_pe() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_my_pe
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_my_pe(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_my_pe", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2478
 * cdef int _nvshmem_my_pe() except* nogil:
 *     global __nvshmem_my_pe
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_my_pe == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2478, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2479
 *     global __nvshmem_my_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2480
 *     _check_or_init_nvshmem()
 *     if __nvshmem_my_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")
 *     return (<int (*)() nogil>__nvshmem_my_pe)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2481
 *     if __nvshmem_my_pe == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)() nogil>__nvshmem_my_pe)(
 *         )
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2481, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_my_pe_is_not_fo};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2481, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2481, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2480
 *     _check_or_init_nvshmem()
 *     if __nvshmem_my_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")
 *     return (<int (*)() nogil>__nvshmem_my_pe)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2479
 *     global __nvshmem_my_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2482
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_my_pe is not found")
 *     return (<int (*)() nogil>__nvshmem_my_pe)(             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_1 = ((int (*)(void))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe)(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2482, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2476
 * 
 * 
 * cdef int _nvshmem_my_pe() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_my_pe
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2486
 * 
 * 
 * cdef int _nvshmem_n_pes() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_n_pes
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_n_pes(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_n_pes", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2488
 * cdef int _nvshmem_n_pes() except* nogil:
 *     global __nvshmem_n_pes
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_n_pes == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2488, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2489
 *     global __nvshmem_n_pes
 *     _check_or_init_nvshmem()
 *     if __nvshmem_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2490
 *     _check_or_init_nvshmem()
 *     if __nvshmem_n_pes == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")
 *     return (<int (*)() nogil>__nvshmem_n_pes)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2491
 *     if __nvshmem_n_pes == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)() nogil>__nvshmem_n_pes)(
 *         )
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2491, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_n_pes_is_not_fo};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2491, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2490
 *     _check_or_init_nvshmem()
 *     if __nvshmem_n_pes == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")
 *     return (<int (*)() nogil>__nvshmem_n_pes)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2489
 *     global __nvshmem_n_pes
 *     _check_or_init_nvshmem()
 *     if __nvshmem_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2492
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_n_pes is not found")
 *     return (<int (*)() nogil>__nvshmem_n_pes)(             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_1 = ((int (*)(void))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes)(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2492, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2486
 * 
 * 
 * cdef int _nvshmem_n_pes() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_n_pes
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2496
 * 
 * 
 * cdef void _nvshmem_info_get_version(int* major, int* minor) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_info_get_version
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_info_get_version(int *__pyx_v_major, int *__pyx_v_minor) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_info_get_version", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2498
 * cdef void _nvshmem_info_get_version(int* major, int* minor) except* nogil:
 *     global __nvshmem_info_get_version
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_info_get_version == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2498, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2499
 *     global __nvshmem_info_get_version
 *     _check_or_init_nvshmem()
 *     if __nvshmem_info_get_version == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2500
 *     _check_or_init_nvshmem()
 *     if __nvshmem_info_get_version == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")
 *     (<void (*)(int*, int*) nogil>__nvshmem_info_get_version)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2501
 *     if __nvshmem_info_get_version == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(int*, int*) nogil>__nvshmem_info_get_version)(
 *         major, minor)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2501, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_info_get_versio};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2501, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2501, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2500
 *     _check_or_init_nvshmem()
 *     if __nvshmem_info_get_version == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")
 *     (<void (*)(int*, int*) nogil>__nvshmem_info_get_version)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2499
 *     global __nvshmem_info_get_version
 *     _check_or_init_nvshmem()
 *     if __nvshmem_info_get_version == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2502
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_info_get_version is not found")
 *     (<void (*)(int*, int*) nogil>__nvshmem_info_get_version)(             # <<<<<<<<<<<<<<
 *         major, minor)
 * 
*/
  ((void (*)(int *, int *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version)(__pyx_v_major, __pyx_v_minor); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2502, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2496
 * 
 * 
 * cdef void _nvshmem_info_get_version(int* major, int* minor) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_info_get_version
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_info_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2506
 * 
 * 
 * cdef void _nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_vendor_get_version_info
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_vendor_get_version_info(int *__pyx_v_major, int *__pyx_v_minor, int *__pyx_v_patch) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_vendor_get_version_info", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2508
 * cdef void _nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:
 *     global __nvshmemx_vendor_get_version_info
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2508, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2509
 *     global __nvshmemx_vendor_get_version_info
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_vendor_get_version_info == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")
 *     (<void (*)(int*, int*, int*) nogil>__nvshmemx_vendor_get_version_info)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2511
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(int*, int*, int*) nogil>__nvshmemx_vendor_get_version_info)(
 *         major, minor, patch)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2511, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_vendor_get_ver};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2511, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2511, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_vendor_get_version_info == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")
 *     (<void (*)(int*, int*, int*) nogil>__nvshmemx_vendor_get_version_info)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2509
 *     global __nvshmemx_vendor_get_version_info
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_vendor_get_version_info == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2512
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_vendor_get_version_info is not found")
 *     (<void (*)(int*, int*, int*) nogil>__nvshmemx_vendor_get_version_info)(             # <<<<<<<<<<<<<<
 *         major, minor, patch)
 * 
*/
  ((void (*)(int *, int *, int *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info)(__pyx_v_major, __pyx_v_minor, __pyx_v_patch); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2512, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2506
 * 
 * 
 * cdef void _nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_vendor_get_version_info
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_vendor_get_version_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2516
 * 
 * 
 * cdef void* _nvshmem_malloc(size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_malloc
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_malloc(size_t __pyx_v_size) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_malloc", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2518
 * cdef void* _nvshmem_malloc(size_t size) except* nogil:
 *     global __nvshmem_malloc
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_malloc == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2518, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2519
 *     global __nvshmem_malloc
 *     _check_or_init_nvshmem()
 *     if __nvshmem_malloc == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2520
 *     _check_or_init_nvshmem()
 *     if __nvshmem_malloc == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")
 *     return (<void* (*)(size_t) nogil>__nvshmem_malloc)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2521
 *     if __nvshmem_malloc == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(size_t) nogil>__nvshmem_malloc)(
 *         size)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2521, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_malloc_is_not_f};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2521, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2521, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2520
 *     _check_or_init_nvshmem()
 *     if __nvshmem_malloc == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")
 *     return (<void* (*)(size_t) nogil>__nvshmem_malloc)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2519
 *     global __nvshmem_malloc
 *     _check_or_init_nvshmem()
 *     if __nvshmem_malloc == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2522
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_malloc is not found")
 *     return (<void* (*)(size_t) nogil>__nvshmem_malloc)(             # <<<<<<<<<<<<<<
 *         size)
 * 
*/
  __pyx_t_7 = ((void *(*)(size_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc)(__pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2522, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2516
 * 
 * 
 * cdef void* _nvshmem_malloc(size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_malloc
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2526
 * 
 * 
 * cdef void* _nvshmem_calloc(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_calloc
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_calloc(size_t __pyx_v_count, size_t __pyx_v_size) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_calloc", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2528
 * cdef void* _nvshmem_calloc(size_t count, size_t size) except* nogil:
 *     global __nvshmem_calloc
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_calloc == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2528, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2529
 *     global __nvshmem_calloc
 *     _check_or_init_nvshmem()
 *     if __nvshmem_calloc == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2530
 *     _check_or_init_nvshmem()
 *     if __nvshmem_calloc == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_calloc)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2531
 *     if __nvshmem_calloc == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_calloc)(
 *         count, size)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2531, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_calloc_is_not_f};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2531, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2531, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2530
 *     _check_or_init_nvshmem()
 *     if __nvshmem_calloc == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_calloc)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2529
 *     global __nvshmem_calloc
 *     _check_or_init_nvshmem()
 *     if __nvshmem_calloc == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2532
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_calloc is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_calloc)(             # <<<<<<<<<<<<<<
 *         count, size)
 * 
*/
  __pyx_t_7 = ((void *(*)(size_t, size_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc)(__pyx_v_count, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2532, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2526
 * 
 * 
 * cdef void* _nvshmem_calloc(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_calloc
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2536
 * 
 * 
 * cdef void* _nvshmem_align(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_align
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_align(size_t __pyx_v_count, size_t __pyx_v_size) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_align", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2538
 * cdef void* _nvshmem_align(size_t count, size_t size) except* nogil:
 *     global __nvshmem_align
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_align == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2538, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2539
 *     global __nvshmem_align
 *     _check_or_init_nvshmem()
 *     if __nvshmem_align == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_align is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2540
 *     _check_or_init_nvshmem()
 *     if __nvshmem_align == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_align is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_align)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2541
 *     if __nvshmem_align == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_align is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_align)(
 *         count, size)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2541, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_align_is_not_fo};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2541, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2541, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2540
 *     _check_or_init_nvshmem()
 *     if __nvshmem_align == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_align is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_align)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2539
 *     global __nvshmem_align
 *     _check_or_init_nvshmem()
 *     if __nvshmem_align == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_align is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2542
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_align is not found")
 *     return (<void* (*)(size_t, size_t) nogil>__nvshmem_align)(             # <<<<<<<<<<<<<<
 *         count, size)
 * 
*/
  __pyx_t_7 = ((void *(*)(size_t, size_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align)(__pyx_v_count, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2542, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2536
 * 
 * 
 * cdef void* _nvshmem_align(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_align
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2546
 * 
 * 
 * cdef void _nvshmem_free(void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_free
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_free(void *__pyx_v_ptr) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_free", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2548
 * cdef void _nvshmem_free(void* ptr) except* nogil:
 *     global __nvshmem_free
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_free == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2548, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2549
 *     global __nvshmem_free
 *     _check_or_init_nvshmem()
 *     if __nvshmem_free == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_free is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2550
 *     _check_or_init_nvshmem()
 *     if __nvshmem_free == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_free is not found")
 *     (<void (*)(void*) nogil>__nvshmem_free)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2551
 *     if __nvshmem_free == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_free is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(void*) nogil>__nvshmem_free)(
 *         ptr)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2551, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_free_is_not_fou};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2551, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2551, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2550
 *     _check_or_init_nvshmem()
 *     if __nvshmem_free == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_free is not found")
 *     (<void (*)(void*) nogil>__nvshmem_free)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2549
 *     global __nvshmem_free
 *     _check_or_init_nvshmem()
 *     if __nvshmem_free == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_free is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2552
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_free is not found")
 *     (<void (*)(void*) nogil>__nvshmem_free)(             # <<<<<<<<<<<<<<
 *         ptr)
 * 
*/
  ((void (*)(void *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free)(__pyx_v_ptr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2552, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2546
 * 
 * 
 * cdef void _nvshmem_free(void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_free
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_free", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2556
 * 
 * 
 * cdef void* _nvshmem_ptr(const void* dest, int pe) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_ptr
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_ptr(void const *__pyx_v_dest, int __pyx_v_pe) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_ptr", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2558
 * cdef void* _nvshmem_ptr(const void* dest, int pe) except* nogil:
 *     global __nvshmem_ptr
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_ptr == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2558, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2559
 *     global __nvshmem_ptr
 *     _check_or_init_nvshmem()
 *     if __nvshmem_ptr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2560
 *     _check_or_init_nvshmem()
 *     if __nvshmem_ptr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")
 *     return (<void* (*)(const void*, int) nogil>__nvshmem_ptr)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2561
 *     if __nvshmem_ptr == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(const void*, int) nogil>__nvshmem_ptr)(
 *         dest, pe)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2561, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_ptr_is_not_foun};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2561, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2561, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2560
 *     _check_or_init_nvshmem()
 *     if __nvshmem_ptr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")
 *     return (<void* (*)(const void*, int) nogil>__nvshmem_ptr)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2559
 *     global __nvshmem_ptr
 *     _check_or_init_nvshmem()
 *     if __nvshmem_ptr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2562
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_ptr is not found")
 *     return (<void* (*)(const void*, int) nogil>__nvshmem_ptr)(             # <<<<<<<<<<<<<<
 *         dest, pe)
 * 
*/
  __pyx_t_7 = ((void *(*)(void const *, int))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr)(__pyx_v_dest, __pyx_v_pe); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2562, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2556
 * 
 * 
 * cdef void* _nvshmem_ptr(const void* dest, int pe) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_ptr
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2566
 * 
 * 
 * cdef void* _nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_mc_ptr
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_mc_ptr(nvshmem_team_t __pyx_v_team, void const *__pyx_v_ptr) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_mc_ptr", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2568
 * cdef void* _nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:
 *     global __nvshmemx_mc_ptr
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_mc_ptr == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2568, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2569
 *     global __nvshmemx_mc_ptr
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_mc_ptr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_mc_ptr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")
 *     return (<void* (*)(nvshmem_team_t, const void*) nogil>__nvshmemx_mc_ptr)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2571
 *     if __nvshmemx_mc_ptr == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(nvshmem_team_t, const void*) nogil>__nvshmemx_mc_ptr)(
 *         team, ptr)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2571, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_mc_ptr_is_not};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2571, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2571, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_mc_ptr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")
 *     return (<void* (*)(nvshmem_team_t, const void*) nogil>__nvshmemx_mc_ptr)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2569
 *     global __nvshmemx_mc_ptr
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_mc_ptr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2572
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_mc_ptr is not found")
 *     return (<void* (*)(nvshmem_team_t, const void*) nogil>__nvshmemx_mc_ptr)(             # <<<<<<<<<<<<<<
 *         team, ptr)
 * 
*/
  __pyx_t_7 = ((void *(*)(nvshmem_team_t, void const *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr)(__pyx_v_team, __pyx_v_ptr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2572, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2566
 * 
 * 
 * cdef void* _nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_mc_ptr
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_mc_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2576
 * 
 * 
 * cdef int _nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_my_pe
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_my_pe(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_my_pe", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2578
 * cdef int _nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:
 *     global __nvshmem_team_my_pe
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_my_pe == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2578, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2579
 *     global __nvshmem_team_my_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2580
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_my_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_my_pe)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2581
 *     if __nvshmem_team_my_pe == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_my_pe)(
 *         team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2581, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_my_pe_is_n};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2581, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2581, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2580
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_my_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_my_pe)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2579
 *     global __nvshmem_team_my_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_my_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2582
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_my_pe is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_my_pe)(             # <<<<<<<<<<<<<<
 *         team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe)(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2582, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2576
 * 
 * 
 * cdef int _nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_my_pe
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2586
 * 
 * 
 * cdef int _nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_n_pes
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_n_pes(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_n_pes", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2588
 * cdef int _nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:
 *     global __nvshmem_team_n_pes
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_n_pes == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2588, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2589
 *     global __nvshmem_team_n_pes
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2590
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_n_pes == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_n_pes)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2591
 *     if __nvshmem_team_n_pes == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_n_pes)(
 *         team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2591, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_n_pes_is_n};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2591, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2591, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2590
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_n_pes == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_n_pes)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2589
 *     global __nvshmem_team_n_pes
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_n_pes == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2592
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_n_pes is not found")
 *     return (<int (*)(nvshmem_team_t) nogil>__nvshmem_team_n_pes)(             # <<<<<<<<<<<<<<
 *         team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes)(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2592, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2586
 * 
 * 
 * cdef int _nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_n_pes
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2596
 * 
 * 
 * cdef void _nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_get_config
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_get_config(nvshmem_team_t __pyx_v_team, nvshmem_team_config_t *__pyx_v_config) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_get_config", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2598
 * cdef void _nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:
 *     global __nvshmem_team_get_config
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_get_config == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2598, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2599
 *     global __nvshmem_team_get_config
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_get_config == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2600
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_get_config == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")
 *     (<void (*)(nvshmem_team_t, nvshmem_team_config_t*) nogil>__nvshmem_team_get_config)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2601
 *     if __nvshmem_team_get_config == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(nvshmem_team_t, nvshmem_team_config_t*) nogil>__nvshmem_team_get_config)(
 *         team, config)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2601, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_get_config};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2601, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2601, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2600
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_get_config == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")
 *     (<void (*)(nvshmem_team_t, nvshmem_team_config_t*) nogil>__nvshmem_team_get_config)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2599
 *     global __nvshmem_team_get_config
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_get_config == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2602
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_get_config is not found")
 *     (<void (*)(nvshmem_team_t, nvshmem_team_config_t*) nogil>__nvshmem_team_get_config)(             # <<<<<<<<<<<<<<
 *         team, config)
 * 
*/
  ((void (*)(nvshmem_team_t, nvshmem_team_config_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config)(__pyx_v_team, __pyx_v_config); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2602, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2596
 * 
 * 
 * cdef void _nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_get_config
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_get_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2606
 * 
 * 
 * cdef int _nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_translate_pe
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_translate_pe(nvshmem_team_t __pyx_v_src_team, int __pyx_v_src_pe, nvshmem_team_t __pyx_v_dest_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_translate_pe", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2608
 * cdef int _nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:
 *     global __nvshmem_team_translate_pe
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_translate_pe == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2608, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2609
 *     global __nvshmem_team_translate_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_translate_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2610
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_translate_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")
 *     return (<int (*)(nvshmem_team_t, int, nvshmem_team_t) nogil>__nvshmem_team_translate_pe)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2611
 *     if __nvshmem_team_translate_pe == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int, nvshmem_team_t) nogil>__nvshmem_team_translate_pe)(
 *         src_team, src_pe, dest_team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2611, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_translate};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2611, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2611, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2610
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_translate_pe == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")
 *     return (<int (*)(nvshmem_team_t, int, nvshmem_team_t) nogil>__nvshmem_team_translate_pe)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2609
 *     global __nvshmem_team_translate_pe
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_translate_pe == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2612
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_translate_pe is not found")
 *     return (<int (*)(nvshmem_team_t, int, nvshmem_team_t) nogil>__nvshmem_team_translate_pe)(             # <<<<<<<<<<<<<<
 *         src_team, src_pe, dest_team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int, nvshmem_team_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe)(__pyx_v_src_team, __pyx_v_src_pe, __pyx_v_dest_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2612, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2606
 * 
 * 
 * cdef int _nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_translate_pe
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_translate_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2616
 * 
 * 
 * cdef int _nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_split_strided
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_strided(nvshmem_team_t __pyx_v_parent_team, int __pyx_v_PE_start, int __pyx_v_PE_stride, int __pyx_v_PE_size, nvshmem_team_config_t const *__pyx_v_config, long __pyx_v_config_mask, nvshmem_team_t *__pyx_v_new_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_split_strided", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2618
 * cdef int _nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:
 *     global __nvshmem_team_split_strided
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_split_strided == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2618, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2619
 *     global __nvshmem_team_split_strided
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_strided == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2620
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_strided == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")
 *     return (<int (*)(nvshmem_team_t, int, int, int, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_strided)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2621
 *     if __nvshmem_team_split_strided == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int, int, int, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_strided)(
 *         parent_team, PE_start, PE_stride, PE_size, config, config_mask, new_team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2621, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_split_stri};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2621, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2621, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2620
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_strided == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")
 *     return (<int (*)(nvshmem_team_t, int, int, int, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_strided)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2619
 *     global __nvshmem_team_split_strided
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_strided == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2622
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_strided is not found")
 *     return (<int (*)(nvshmem_team_t, int, int, int, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_strided)(             # <<<<<<<<<<<<<<
 *         parent_team, PE_start, PE_stride, PE_size, config, config_mask, new_team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided)(__pyx_v_parent_team, __pyx_v_PE_start, __pyx_v_PE_stride, __pyx_v_PE_size, __pyx_v_config, __pyx_v_config_mask, __pyx_v_new_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2622, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2616
 * 
 * 
 * cdef int _nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_split_strided
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_split_strided", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2626
 * 
 * 
 * cdef int _nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_get_uniqueid
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t *__pyx_v_uniqueid) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_team_get_uniqueid", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2628
 * cdef int _nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:
 *     global __nvshmemx_team_get_uniqueid
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2628, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2629
 *     global __nvshmemx_team_get_uniqueid
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2630
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_team_uniqueid_t*) nogil>__nvshmemx_team_get_uniqueid)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2631
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmemx_team_uniqueid_t*) nogil>__nvshmemx_team_get_uniqueid)(
 *         uniqueid)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2631, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_team_get_uniqu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2631, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2631, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2630
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_get_uniqueid == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_team_uniqueid_t*) nogil>__nvshmemx_team_get_uniqueid)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2629
 *     global __nvshmemx_team_get_uniqueid
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2632
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_team_uniqueid_t*) nogil>__nvshmemx_team_get_uniqueid)(             # <<<<<<<<<<<<<<
 *         uniqueid)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmemx_team_uniqueid_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid)(__pyx_v_uniqueid); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2632, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2626
 * 
 * 
 * cdef int _nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_get_uniqueid
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_team_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2636
 * 
 * 
 * cdef int _nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_init
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_init(nvshmem_team_t *__pyx_v_team, nvshmem_team_config_t *__pyx_v_config, long __pyx_v_config_mask, int __pyx_v_npes, int __pyx_v_pe_idx_in_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_team_init", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2638
 * cdef int _nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:
 *     global __nvshmemx_team_init
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_init == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2638, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2639
 *     global __nvshmemx_team_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2640
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")
 *     return (<int (*)(nvshmem_team_t*, nvshmem_team_config_t*, long, int, int) nogil>__nvshmemx_team_init)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2641
 *     if __nvshmemx_team_init == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t*, nvshmem_team_config_t*, long, int, int) nogil>__nvshmemx_team_init)(
 *         team, config, config_mask, npes, pe_idx_in_team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2641, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_team_init_is_n};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2641, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2641, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2640
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")
 *     return (<int (*)(nvshmem_team_t*, nvshmem_team_config_t*, long, int, int) nogil>__nvshmemx_team_init)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2639
 *     global __nvshmemx_team_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2642
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_init is not found")
 *     return (<int (*)(nvshmem_team_t*, nvshmem_team_config_t*, long, int, int) nogil>__nvshmemx_team_init)(             # <<<<<<<<<<<<<<
 *         team, config, config_mask, npes, pe_idx_in_team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t *, nvshmem_team_config_t *, long, int, int))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init)(__pyx_v_team, __pyx_v_config, __pyx_v_config_mask, __pyx_v_npes, __pyx_v_pe_idx_in_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2642, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2636
 * 
 * 
 * cdef int _nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_init
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_team_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2646
 * 
 * 
 * cdef int _nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_split_2d
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_2d(nvshmem_team_t __pyx_v_parent_team, int __pyx_v_xrange, nvshmem_team_config_t const *__pyx_v_xaxis_config, long __pyx_v_xaxis_mask, nvshmem_team_t *__pyx_v_xaxis_team, nvshmem_team_config_t const *__pyx_v_yaxis_config, long __pyx_v_yaxis_mask, nvshmem_team_t *__pyx_v_yaxis_team) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_split_2d", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2648
 * cdef int _nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:
 *     global __nvshmem_team_split_2d
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_split_2d == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2648, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2649
 *     global __nvshmem_team_split_2d
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_2d == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2650
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_2d == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")
 *     return (<int (*)(nvshmem_team_t, int, const nvshmem_team_config_t*, long, nvshmem_team_t*, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_2d)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2651
 *     if __nvshmem_team_split_2d == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int, const nvshmem_team_config_t*, long, nvshmem_team_t*, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_2d)(
 *         parent_team, xrange, xaxis_config, xaxis_mask, xaxis_team, yaxis_config, yaxis_mask, yaxis_team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2651, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_split_2d_i};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2651, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2651, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2650
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_2d == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")
 *     return (<int (*)(nvshmem_team_t, int, const nvshmem_team_config_t*, long, nvshmem_team_t*, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_2d)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2649
 *     global __nvshmem_team_split_2d
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_split_2d == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2652
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_split_2d is not found")
 *     return (<int (*)(nvshmem_team_t, int, const nvshmem_team_config_t*, long, nvshmem_team_t*, const nvshmem_team_config_t*, long, nvshmem_team_t*) nogil>__nvshmem_team_split_2d)(             # <<<<<<<<<<<<<<
 *         parent_team, xrange, xaxis_config, xaxis_mask, xaxis_team, yaxis_config, yaxis_mask, yaxis_team)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d)(__pyx_v_parent_team, __pyx_v_xrange, __pyx_v_xaxis_config, __pyx_v_xaxis_mask, __pyx_v_xaxis_team, __pyx_v_yaxis_config, __pyx_v_yaxis_mask, __pyx_v_yaxis_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2652, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2646
 * 
 * 
 * cdef int _nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_split_2d
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_split_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2656
 * 
 * 
 * cdef void _nvshmem_team_destroy(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_destroy
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_destroy(nvshmem_team_t __pyx_v_team) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmem_team_destroy", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2658
 * cdef void _nvshmem_team_destroy(nvshmem_team_t team) except* nogil:
 *     global __nvshmem_team_destroy
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmem_team_destroy == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2658, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2659
 *     global __nvshmem_team_destroy
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_destroy == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2660
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_destroy == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")
 *     (<void (*)(nvshmem_team_t) nogil>__nvshmem_team_destroy)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2661
 *     if __nvshmem_team_destroy == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(nvshmem_team_t) nogil>__nvshmem_team_destroy)(
 *         team)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2661, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmem_team_destroy_is};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2661, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2661, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2660
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_destroy == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")
 *     (<void (*)(nvshmem_team_t) nogil>__nvshmem_team_destroy)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2659
 *     global __nvshmem_team_destroy
 *     _check_or_init_nvshmem()
 *     if __nvshmem_team_destroy == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2662
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmem_team_destroy is not found")
 *     (<void (*)(nvshmem_team_t) nogil>__nvshmem_team_destroy)(             # <<<<<<<<<<<<<<
 *         team)
 * 
*/
  ((void (*)(nvshmem_team_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy)(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2662, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2656
 * 
 * 
 * cdef void _nvshmem_team_destroy(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmem_team_destroy
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmem_team_destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2666
 * 
 * 
 * cdef int _nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2668
 * cdef int _nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2668, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2669
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2670
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2671
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2671, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_allto};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2671, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2671, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2670
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2669
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2672
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2672, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2666
 * 
 * 
 * cdef int _nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2676
 * 
 * 
 * cdef int _nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_alltoall_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2678
 * cdef int _nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2678, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2679
 *     global __nvshmemx_half_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2680
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2681
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2681, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2681, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2681, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2680
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2679
 *     global __nvshmemx_half_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2682
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2682, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2676
 * 
 * 
 * cdef int _nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2686
 * 
 * 
 * cdef int _nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_alltoall_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2688
 * cdef int _nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2688, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2689
 *     global __nvshmemx_float_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2690
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2691
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2691, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2691, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2691, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2690
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2689
 *     global __nvshmemx_float_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2692
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2692, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2686
 * 
 * 
 * cdef int _nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2696
 * 
 * 
 * cdef int _nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_alltoall_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2698
 * cdef int _nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2698, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2699
 *     global __nvshmemx_double_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2700
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2701
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2701, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_alltoal};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2701, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2701, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2700
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2699
 *     global __nvshmemx_double_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2702
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2702, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2696
 * 
 * 
 * cdef int _nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2706
 * 
 * 
 * cdef int _nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_alltoall_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2708
 * cdef int _nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2708, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2709
 *     global __nvshmemx_char_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2710
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2711
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2711, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2711, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2711, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2710
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2709
 *     global __nvshmemx_char_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2712
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2712, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2706
 * 
 * 
 * cdef int _nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2716
 * 
 * 
 * cdef int _nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_alltoall_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2718
 * cdef int _nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2718, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2719
 *     global __nvshmemx_short_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2720
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2721
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2721, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2721, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2721, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2720
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2719
 *     global __nvshmemx_short_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2722
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2722, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2716
 * 
 * 
 * cdef int _nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2726
 * 
 * 
 * cdef int _nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_alltoall_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2728
 * cdef int _nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2728, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2729
 *     global __nvshmemx_schar_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2730
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2731
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2731, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2731, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2731, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2730
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2729
 *     global __nvshmemx_schar_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2732
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2732, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2726
 * 
 * 
 * cdef int _nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2736
 * 
 * 
 * cdef int _nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2738
 * cdef int _nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2738, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2739
 *     global __nvshmemx_int_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2740
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2741
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2741, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_alltoall_o};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2741, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2741, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2740
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2739
 *     global __nvshmemx_int_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2742
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2742, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2736
 * 
 * 
 * cdef int _nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2746
 * 
 * 
 * cdef int _nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_alltoall_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2748
 * cdef int _nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2748, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2749
 *     global __nvshmemx_long_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2750
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2751
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2751, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2751, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2751, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2750
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2749
 *     global __nvshmemx_long_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2752
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2752, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2746
 * 
 * 
 * cdef int _nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2756
 * 
 * 
 * cdef int _nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2758
 * cdef int _nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2758, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2759
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2760
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2761
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2761, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_allto};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2761, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2761, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2760
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2759
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2762
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2762, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2756
 * 
 * 
 * cdef int _nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2766
 * 
 * 
 * cdef int _nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2768
 * cdef int _nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2768, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2769
 *     global __nvshmemx_int8_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2770
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2771
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2771, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2771, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2771, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2770
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2769
 *     global __nvshmemx_int8_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2772
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2772, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2766
 * 
 * 
 * cdef int _nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2776
 * 
 * 
 * cdef int _nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2778
 * cdef int _nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2778, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2779
 *     global __nvshmemx_int16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2780
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2781
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2781, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2781, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2781, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2780
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2779
 *     global __nvshmemx_int16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2782
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2782, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2776
 * 
 * 
 * cdef int _nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2786
 * 
 * 
 * cdef int _nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2788
 * cdef int _nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2788, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2789
 *     global __nvshmemx_int32_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2790
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2791
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2791, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2791, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2791, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2790
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2789
 *     global __nvshmemx_int32_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2792
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2792, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2786
 * 
 * 
 * cdef int _nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2796
 * 
 * 
 * cdef int _nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2798
 * cdef int _nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2798, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2799
 *     global __nvshmemx_int64_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2800
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2801
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2801, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2801, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2801, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2800
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2799
 *     global __nvshmemx_int64_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2802
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2802, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2796
 * 
 * 
 * cdef int _nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2806
 * 
 * 
 * cdef int _nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2808
 * cdef int _nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2808, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2809
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2810
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2811
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2811, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2811, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2810
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2809
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2812
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2812, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2806
 * 
 * 
 * cdef int _nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2816
 * 
 * 
 * cdef int _nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2818
 * cdef int _nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2818, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2819
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2820
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2821
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2821, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_alltoal};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2821, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2821, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2820
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2819
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2822
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2822, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2816
 * 
 * 
 * cdef int _nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2826
 * 
 * 
 * cdef int _nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2828
 * cdef int _nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2828, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2829
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2830
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2831
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2831, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_alltoal};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2831, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2831, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2830
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2829
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2832
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2832, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2826
 * 
 * 
 * cdef int _nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2836
 * 
 * 
 * cdef int _nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2838
 * cdef int _nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2838, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2839
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2840
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2841
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2841, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_alltoal};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2841, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2841, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2840
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2839
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2842
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2842, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2836
 * 
 * 
 * cdef int _nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2846
 * 
 * 
 * cdef int _nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_alltoall_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_alltoall_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2848
 * cdef int _nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_alltoall_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2848, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2849
 *     global __nvshmemx_size_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2850
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_alltoall_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2851
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_alltoall_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2851, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_alltoall};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2851, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2851, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2850
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_alltoall_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_alltoall_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2849
 *     global __nvshmemx_size_alltoall_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_alltoall_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2852
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_alltoall_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_alltoall_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2852, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2846
 * 
 * 
 * cdef int _nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_alltoall_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2856
 * 
 * 
 * cdef int _nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_barrier_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_on_stream(nvshmem_team_t __pyx_v_team, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_barrier_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2858
 * cdef int _nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_barrier_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2858, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2859
 *     global __nvshmemx_barrier_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2860
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_barrier_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2861
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_barrier_on_stream)(
 *         team, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2861, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_barrier_on_str};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2861, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2861, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2860
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_barrier_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2859
 *     global __nvshmemx_barrier_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2862
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_barrier_on_stream)(             # <<<<<<<<<<<<<<
 *         team, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream)(__pyx_v_team, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2862, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2856
 * 
 * 
 * cdef int _nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_barrier_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_barrier_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2866
 * 
 * 
 * cdef void _nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_barrier_all_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_all_on_stream(cudaStream_t __pyx_v_stream) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_barrier_all_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2868
 * cdef void _nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:
 *     global __nvshmemx_barrier_all_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2868, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2869
 *     global __nvshmemx_barrier_all_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2870
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_barrier_all_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2871
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_barrier_all_on_stream)(
 *         stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2871, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_barrier_all_on};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2871, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2871, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2870
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_all_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_barrier_all_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2869
 *     global __nvshmemx_barrier_all_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_barrier_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2872
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_barrier_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_barrier_all_on_stream)(             # <<<<<<<<<<<<<<
 *         stream)
 * 
*/
  ((void (*)(cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream)(__pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2872, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2866
 * 
 * 
 * cdef void _nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_barrier_all_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_barrier_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2876
 * 
 * 
 * cdef int _nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_sync_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_sync_on_stream(nvshmem_team_t __pyx_v_team, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_team_sync_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2878
 * cdef int _nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_team_sync_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2878, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2879
 *     global __nvshmemx_team_sync_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_sync_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2880
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_team_sync_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2881
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_team_sync_on_stream)(
 *         team, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2881, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_team_sync_on_s};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2881, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2881, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2880
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_sync_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_team_sync_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2879
 *     global __nvshmemx_team_sync_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_team_sync_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2882
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_team_sync_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, cudaStream_t) nogil>__nvshmemx_team_sync_on_stream)(             # <<<<<<<<<<<<<<
 *         team, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream)(__pyx_v_team, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2882, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2876
 * 
 * 
 * cdef int _nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_team_sync_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_team_sync_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2886
 * 
 * 
 * cdef void _nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_sync_all_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_sync_all_on_stream(cudaStream_t __pyx_v_stream) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_sync_all_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2888
 * cdef void _nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:
 *     global __nvshmemx_sync_all_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2888, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2889
 *     global __nvshmemx_sync_all_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_sync_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2890
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_sync_all_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2891
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_sync_all_on_stream)(
 *         stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2891, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_sync_all_on_st};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2891, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2890
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_sync_all_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_sync_all_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2889
 *     global __nvshmemx_sync_all_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_sync_all_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2892
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_sync_all_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_sync_all_on_stream)(             # <<<<<<<<<<<<<<
 *         stream)
 * 
*/
  ((void (*)(cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream)(__pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2892, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2886
 * 
 * 
 * cdef void _nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_sync_all_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_sync_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2896
 * 
 * 
 * cdef int _nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2898
 * cdef int _nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2898, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2899
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2900
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, int, cudaStream_t) nogil>__nvshmemx_bfloat16_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2901
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, int, cudaStream_t) nogil>__nvshmemx_bfloat16_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2901, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_broad};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2901, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2901, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2900
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, int, cudaStream_t) nogil>__nvshmemx_bfloat16_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2899
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2902
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, int, cudaStream_t) nogil>__nvshmemx_bfloat16_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2902, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2896
 * 
 * 
 * cdef int _nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2906
 * 
 * 
 * cdef int _nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_broadcast_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2908
 * cdef int _nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2908, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2909
 *     global __nvshmemx_half_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2910
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, int, cudaStream_t) nogil>__nvshmemx_half_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2911
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, int, cudaStream_t) nogil>__nvshmemx_half_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2911, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2911, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2911, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2910
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, int, cudaStream_t) nogil>__nvshmemx_half_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2909
 *     global __nvshmemx_half_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2912
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, int, cudaStream_t) nogil>__nvshmemx_half_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2912, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2906
 * 
 * 
 * cdef int _nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2916
 * 
 * 
 * cdef int _nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_broadcast_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2918
 * cdef int _nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2918, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2919
 *     global __nvshmemx_float_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2920
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, int, cudaStream_t) nogil>__nvshmemx_float_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2921
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, int, cudaStream_t) nogil>__nvshmemx_float_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2921, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2921, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2921, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2920
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, int, cudaStream_t) nogil>__nvshmemx_float_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2919
 *     global __nvshmemx_float_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2922
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, int, cudaStream_t) nogil>__nvshmemx_float_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2922, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2916
 * 
 * 
 * cdef int _nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2926
 * 
 * 
 * cdef int _nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_broadcast_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2928
 * cdef int _nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2928, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2929
 *     global __nvshmemx_double_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2930
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, int, cudaStream_t) nogil>__nvshmemx_double_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2931
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, int, cudaStream_t) nogil>__nvshmemx_double_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2931, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_broadca};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2931, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2931, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2930
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, int, cudaStream_t) nogil>__nvshmemx_double_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2929
 *     global __nvshmemx_double_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2932
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, int, cudaStream_t) nogil>__nvshmemx_double_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2932, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2926
 * 
 * 
 * cdef int _nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2936
 * 
 * 
 * cdef int _nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_broadcast_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2938
 * cdef int _nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2938, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2939
 *     global __nvshmemx_char_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2940
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, int, cudaStream_t) nogil>__nvshmemx_char_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2941
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, int, cudaStream_t) nogil>__nvshmemx_char_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2941, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2941, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2941, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2940
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, int, cudaStream_t) nogil>__nvshmemx_char_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2939
 *     global __nvshmemx_char_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2942
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, int, cudaStream_t) nogil>__nvshmemx_char_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2942, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2936
 * 
 * 
 * cdef int _nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2946
 * 
 * 
 * cdef int _nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_broadcast_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2948
 * cdef int _nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2948, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2949
 *     global __nvshmemx_short_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2950
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, int, cudaStream_t) nogil>__nvshmemx_short_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2951
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, int, cudaStream_t) nogil>__nvshmemx_short_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2951, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2951, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2951, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2950
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, int, cudaStream_t) nogil>__nvshmemx_short_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2949
 *     global __nvshmemx_short_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2952
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, int, cudaStream_t) nogil>__nvshmemx_short_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2952, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2946
 * 
 * 
 * cdef int _nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2956
 * 
 * 
 * cdef int _nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_broadcast_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2958
 * cdef int _nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2958, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2959
 *     global __nvshmemx_schar_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2960
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, int, cudaStream_t) nogil>__nvshmemx_schar_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2961
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, int, cudaStream_t) nogil>__nvshmemx_schar_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2961, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2961, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2961, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2960
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, int, cudaStream_t) nogil>__nvshmemx_schar_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2959
 *     global __nvshmemx_schar_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2962
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, int, cudaStream_t) nogil>__nvshmemx_schar_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2962, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2956
 * 
 * 
 * cdef int _nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2966
 * 
 * 
 * cdef int _nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2968
 * cdef int _nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2968, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2969
 *     global __nvshmemx_int_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2970
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, int, cudaStream_t) nogil>__nvshmemx_int_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2971
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, int, cudaStream_t) nogil>__nvshmemx_int_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2971, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2971, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2971, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2970
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, int, cudaStream_t) nogil>__nvshmemx_int_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2969
 *     global __nvshmemx_int_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2972
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, int, cudaStream_t) nogil>__nvshmemx_int_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2972, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2966
 * 
 * 
 * cdef int _nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2976
 * 
 * 
 * cdef int _nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_broadcast_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2978
 * cdef int _nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2978, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2979
 *     global __nvshmemx_long_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2980
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, int, cudaStream_t) nogil>__nvshmemx_long_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2981
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, int, cudaStream_t) nogil>__nvshmemx_long_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2981, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2981, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2981, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2980
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, int, cudaStream_t) nogil>__nvshmemx_long_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2979
 *     global __nvshmemx_long_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2982
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, int, cudaStream_t) nogil>__nvshmemx_long_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2982, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2976
 * 
 * 
 * cdef int _nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2986
 * 
 * 
 * cdef int _nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2988
 * cdef int _nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2988, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2989
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2990
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, int, cudaStream_t) nogil>__nvshmemx_longlong_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":2991
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, int, cudaStream_t) nogil>__nvshmemx_longlong_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2991, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_broad};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2991, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 2991, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":2990
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, int, cudaStream_t) nogil>__nvshmemx_longlong_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2989
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2992
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, int, cudaStream_t) nogil>__nvshmemx_longlong_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 2992, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2986
 * 
 * 
 * cdef int _nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":2996
 * 
 * 
 * cdef int _nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2998
 * cdef int _nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2998, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2999
 *     global __nvshmemx_int8_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3000
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int8_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3001
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int8_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3001, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3001, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3001, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3000
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int8_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":2999
 *     global __nvshmemx_int8_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3002
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int8_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3002, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2996
 * 
 * 
 * cdef int _nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3006
 * 
 * 
 * cdef int _nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3008
 * cdef int _nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3008, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3009
 *     global __nvshmemx_int16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3010
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int16_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3011
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int16_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3011, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3011, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3011, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3010
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int16_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3009
 *     global __nvshmemx_int16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3012
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int16_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3012, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3006
 * 
 * 
 * cdef int _nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3016
 * 
 * 
 * cdef int _nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3018
 * cdef int _nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3018, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3019
 *     global __nvshmemx_int32_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3020
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int32_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3021
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int32_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3021, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3021, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3021, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3020
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int32_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3019
 *     global __nvshmemx_int32_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3022
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int32_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3022, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3016
 * 
 * 
 * cdef int _nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3026
 * 
 * 
 * cdef int _nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3028
 * cdef int _nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3028, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3029
 *     global __nvshmemx_int64_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3030
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int64_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3031
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int64_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3031, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3031, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3031, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3030
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int64_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3029
 *     global __nvshmemx_int64_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3032
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_int64_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3032, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3026
 * 
 * 
 * cdef int _nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3036
 * 
 * 
 * cdef int _nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3038
 * cdef int _nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3038, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3039
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3040
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint8_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3041
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint8_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3041, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_broadcas};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3041, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3041, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3040
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint8_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3039
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3042
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint8_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3042, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3036
 * 
 * 
 * cdef int _nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3046
 * 
 * 
 * cdef int _nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3048
 * cdef int _nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3048, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3049
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3050
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint16_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3051
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint16_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3051, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_broadca};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3051, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3051, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3050
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint16_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3049
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3052
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint16_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3052, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3046
 * 
 * 
 * cdef int _nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3056
 * 
 * 
 * cdef int _nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3058
 * cdef int _nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3058, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3059
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3060
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint32_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3061
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint32_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3061, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_broadca};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3061, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3061, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3060
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint32_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3059
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3062
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint32_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3062, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3056
 * 
 * 
 * cdef int _nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3066
 * 
 * 
 * cdef int _nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3068
 * cdef int _nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3068, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3069
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3070
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint64_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3071
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint64_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3071, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_broadca};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3071, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3071, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3070
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint64_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3069
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3072
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_uint64_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3072, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3066
 * 
 * 
 * cdef int _nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3076
 * 
 * 
 * cdef int _nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_broadcast_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_broadcast_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3078
 * cdef int _nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_broadcast_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3078, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3079
 *     global __nvshmemx_size_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3080
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_size_broadcast_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3081
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_size_broadcast_on_stream)(
 *         team, dest, src, nelem, PE_root, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3081, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_broadcast};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3081, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3081, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3080
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_broadcast_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_size_broadcast_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3079
 *     global __nvshmemx_size_broadcast_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_broadcast_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3082
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_broadcast_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, int, cudaStream_t) nogil>__nvshmemx_size_broadcast_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, PE_root, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3082, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3076
 * 
 * 
 * cdef int _nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_broadcast_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3086
 * 
 * 
 * cdef int _nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3088
 * cdef int _nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3088, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3089
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3090
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3091
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3091, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_fcoll};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3091, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3090
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3089
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3092
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3092, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3086
 * 
 * 
 * cdef int _nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3096
 * 
 * 
 * cdef int _nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_fcollect_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3098
 * cdef int _nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3098, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3099
 *     global __nvshmemx_half_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3100
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3101
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3101, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3101, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3101, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3100
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3099
 *     global __nvshmemx_half_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3102
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3102, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3096
 * 
 * 
 * cdef int _nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3106
 * 
 * 
 * cdef int _nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_fcollect_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3108
 * cdef int _nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3108, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3109
 *     global __nvshmemx_float_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3110
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3111
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3111, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3111, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3111, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3110
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3109
 *     global __nvshmemx_float_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3112
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3112, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3106
 * 
 * 
 * cdef int _nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3116
 * 
 * 
 * cdef int _nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_fcollect_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3118
 * cdef int _nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3118, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3119
 *     global __nvshmemx_double_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3120
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3121
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3121, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_fcollec};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3121, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3121, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3120
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3119
 *     global __nvshmemx_double_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3122
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3122, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3116
 * 
 * 
 * cdef int _nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3126
 * 
 * 
 * cdef int _nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_fcollect_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3128
 * cdef int _nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3128, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3129
 *     global __nvshmemx_char_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3130
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3131
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3131, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3131, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3131, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3130
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3129
 *     global __nvshmemx_char_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3132
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3132, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3126
 * 
 * 
 * cdef int _nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3136
 * 
 * 
 * cdef int _nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_fcollect_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3138
 * cdef int _nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3138, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3139
 *     global __nvshmemx_short_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3140
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3141
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3141, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3141, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3141, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3140
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3139
 *     global __nvshmemx_short_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3142
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3142, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3136
 * 
 * 
 * cdef int _nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3146
 * 
 * 
 * cdef int _nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_fcollect_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3148
 * cdef int _nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3148, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3149
 *     global __nvshmemx_schar_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3150
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3151
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3151, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3151, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3151, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3150
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3149
 *     global __nvshmemx_schar_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3152
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3152, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3146
 * 
 * 
 * cdef int _nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3156
 * 
 * 
 * cdef int _nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3158
 * cdef int _nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3158, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3159
 *     global __nvshmemx_int_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3160
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3161
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3161, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_fcollect_o};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3161, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3161, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3160
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3159
 *     global __nvshmemx_int_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3162
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3162, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3156
 * 
 * 
 * cdef int _nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3166
 * 
 * 
 * cdef int _nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_fcollect_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3168
 * cdef int _nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3168, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3169
 *     global __nvshmemx_long_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3170
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3171
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3171, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3171, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3171, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3170
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3169
 *     global __nvshmemx_long_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3172
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3172, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3166
 * 
 * 
 * cdef int _nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3176
 * 
 * 
 * cdef int _nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3178
 * cdef int _nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3178, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3179
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3180
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3181
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3181, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_fcoll};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3181, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3181, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3180
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3179
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3182
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3182, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3176
 * 
 * 
 * cdef int _nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3186
 * 
 * 
 * cdef int _nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3188
 * cdef int _nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3188, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3189
 *     global __nvshmemx_int8_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3190
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3191
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3191, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3191, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3191, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3190
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3189
 *     global __nvshmemx_int8_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3192
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3192, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3186
 * 
 * 
 * cdef int _nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3196
 * 
 * 
 * cdef int _nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3198
 * cdef int _nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3198, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3199
 *     global __nvshmemx_int16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3200
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3201
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3201, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3201, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3201, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3200
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3199
 *     global __nvshmemx_int16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3202
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3202, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3196
 * 
 * 
 * cdef int _nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3206
 * 
 * 
 * cdef int _nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3208
 * cdef int _nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3208, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3209
 *     global __nvshmemx_int32_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3210
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3211
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3211, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3211, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3211, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3210
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3209
 *     global __nvshmemx_int32_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3212
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3212, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3206
 * 
 * 
 * cdef int _nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3216
 * 
 * 
 * cdef int _nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3218
 * cdef int _nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3218, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3219
 *     global __nvshmemx_int64_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3220
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3221
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3221, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3221, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3221, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3220
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3219
 *     global __nvshmemx_int64_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3222
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3222, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3216
 * 
 * 
 * cdef int _nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3226
 * 
 * 
 * cdef int _nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3228
 * cdef int _nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3228, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3229
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3230
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3231
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3231, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3231, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3231, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3230
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3229
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3232
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3232, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3226
 * 
 * 
 * cdef int _nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3236
 * 
 * 
 * cdef int _nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3238
 * cdef int _nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3238, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3239
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3240
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3241
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3241, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_fcollec};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3241, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3241, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3240
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3239
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3242
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3242, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3236
 * 
 * 
 * cdef int _nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3246
 * 
 * 
 * cdef int _nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3248
 * cdef int _nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3248, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3249
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3250
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3251
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3251, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_fcollec};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3251, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3251, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3250
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3249
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3252
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3252, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3246
 * 
 * 
 * cdef int _nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3256
 * 
 * 
 * cdef int _nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3258
 * cdef int _nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3258, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3259
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3260
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3261
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3261, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_fcollec};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3261, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3261, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3260
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3259
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3262
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3262, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3256
 * 
 * 
 * cdef int _nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3266
 * 
 * 
 * cdef int _nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_fcollect_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_fcollect_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3268
 * cdef int _nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_fcollect_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3268, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3269
 *     global __nvshmemx_size_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3270
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_fcollect_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3271
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_fcollect_on_stream)(
 *         team, dest, src, nelem, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3271, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_fcollect};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3271, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3271, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3270
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_fcollect_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_fcollect_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3269
 *     global __nvshmemx_size_fcollect_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_fcollect_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3272
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_fcollect_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_fcollect_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nelem, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3272, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3266
 * 
 * 
 * cdef int _nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_fcollect_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3276
 * 
 * 
 * cdef int _nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3278
 * cdef int _nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3278, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3279
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3280
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3281
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3281, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_max_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3281, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3281, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3280
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3279
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3282
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3282, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3276
 * 
 * 
 * cdef int _nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3286
 * 
 * 
 * cdef int _nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3288
 * cdef int _nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3288, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3289
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3290
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3291
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3291, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3291, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3291, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3290
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3289
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3292
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3292, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3286
 * 
 * 
 * cdef int _nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3296
 * 
 * 
 * cdef int _nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3298
 * cdef int _nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3298, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3299
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3300
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3301
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3301, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3301, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3301, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3300
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3299
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3302
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3302, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3296
 * 
 * 
 * cdef int _nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3306
 * 
 * 
 * cdef int _nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3308
 * cdef int _nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3308, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3309
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3310
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3311
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3311, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3311, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3311, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3310
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3309
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3312
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3312, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3306
 * 
 * 
 * cdef int _nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3316
 * 
 * 
 * cdef int _nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3318
 * cdef int _nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3318, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3319
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3320
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3321
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3321, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3321, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3321, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3320
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3319
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3322
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3322, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3316
 * 
 * 
 * cdef int _nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3326
 * 
 * 
 * cdef int _nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3328
 * cdef int _nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3328, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3329
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3330
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3331
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3331, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_max_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3331, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3331, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3330
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3329
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3332
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3332, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3326
 * 
 * 
 * cdef int _nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3336
 * 
 * 
 * cdef int _nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3338
 * cdef int _nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3338, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3339
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3340
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3341
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3341, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_max_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3341, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3341, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3340
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3339
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3342
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3342, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3336
 * 
 * 
 * cdef int _nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3346
 * 
 * 
 * cdef int _nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3348
 * cdef int _nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3348, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3349
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3350
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3351
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3351, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_max_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3351, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3351, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3350
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3349
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3352
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3352, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3346
 * 
 * 
 * cdef int _nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3356
 * 
 * 
 * cdef int _nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3358
 * cdef int _nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3358, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3359
 *     global __nvshmemx_size_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3360
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3361
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3361, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_max_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3361, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3361, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3360
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3359
 *     global __nvshmemx_size_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3362
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3362, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3356
 * 
 * 
 * cdef int _nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3366
 * 
 * 
 * cdef int _nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3368
 * cdef int _nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3368, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3369
 *     global __nvshmemx_char_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3370
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3371
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3371, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_max_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3371, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3371, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3370
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3369
 *     global __nvshmemx_char_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3372
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3372, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3366
 * 
 * 
 * cdef int _nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3376
 * 
 * 
 * cdef int _nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3378
 * cdef int _nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3378, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3379
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3380
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3381
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3381, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3381, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3381, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3380
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3379
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3382
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3382, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3376
 * 
 * 
 * cdef int _nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3386
 * 
 * 
 * cdef int _nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3388
 * cdef int _nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3388, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3389
 *     global __nvshmemx_short_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3390
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3391
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3391, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3391, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3391, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3390
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3389
 *     global __nvshmemx_short_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3392
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3392, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3386
 * 
 * 
 * cdef int _nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3396
 * 
 * 
 * cdef int _nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3398
 * cdef int _nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3398, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3399
 *     global __nvshmemx_int_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3400
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3401
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3401, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_max_reduce};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3401, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3401, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3400
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3399
 *     global __nvshmemx_int_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3402
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3402, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3396
 * 
 * 
 * cdef int _nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3406
 * 
 * 
 * cdef int _nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3408
 * cdef int _nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3408, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3409
 *     global __nvshmemx_long_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3410
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3411
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3411, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_max_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3411, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3411, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3410
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3409
 *     global __nvshmemx_long_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3412
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3412, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3406
 * 
 * 
 * cdef int _nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3416
 * 
 * 
 * cdef int _nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3418
 * cdef int _nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3418, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3419
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3420
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3421
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3421, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_max_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3421, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3421, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3420
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3419
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3422
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3422, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3416
 * 
 * 
 * cdef int _nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3426
 * 
 * 
 * cdef int _nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3428
 * cdef int _nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3428, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3429
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3430
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3431
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3431, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_max_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3431, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3431, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3430
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3429
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3432
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3432, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3426
 * 
 * 
 * cdef int _nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3436
 * 
 * 
 * cdef int _nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3438
 * cdef int _nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3438, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3439
 *     global __nvshmemx_half_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3440
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3441
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3441, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_max_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3441, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3441, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3440
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3439
 *     global __nvshmemx_half_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3442
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3442, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3436
 * 
 * 
 * cdef int _nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3446
 * 
 * 
 * cdef int _nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3448
 * cdef int _nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3448, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3449
 *     global __nvshmemx_float_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3450
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3451
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3451, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_max_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3451, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3451, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3450
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3449
 *     global __nvshmemx_float_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3452
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3452, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3446
 * 
 * 
 * cdef int _nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3456
 * 
 * 
 * cdef int _nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_max_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3458
 * cdef int _nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_max_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3458, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3459
 *     global __nvshmemx_double_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3460
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3461
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3461, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_max_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3461, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3461, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3460
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3459
 *     global __nvshmemx_double_max_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3462
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3462, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3456
 * 
 * 
 * cdef int _nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_max_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3466
 * 
 * 
 * cdef int _nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3468
 * cdef int _nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3468, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3469
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3471
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3471, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_min_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3471, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3471, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3469
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3472
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3472, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3466
 * 
 * 
 * cdef int _nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3476
 * 
 * 
 * cdef int _nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3478
 * cdef int _nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3478, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3479
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3480
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3481
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3481, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3481, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3481, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3480
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3479
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3482
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3482, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3476
 * 
 * 
 * cdef int _nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3486
 * 
 * 
 * cdef int _nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3488
 * cdef int _nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3488, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3489
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3490
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3491
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3491, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3491, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3491, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3490
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3489
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3492
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3492, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3486
 * 
 * 
 * cdef int _nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3496
 * 
 * 
 * cdef int _nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3498
 * cdef int _nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3498, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3499
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3500
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3501
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3501, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3501, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3501, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3500
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3499
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3502
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3502, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3496
 * 
 * 
 * cdef int _nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3506
 * 
 * 
 * cdef int _nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3508
 * cdef int _nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3508, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3509
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3511
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3511, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3511, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3511, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3509
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3512
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3512, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3506
 * 
 * 
 * cdef int _nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3516
 * 
 * 
 * cdef int _nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3518
 * cdef int _nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3518, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3519
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3520
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3521
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3521, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_min_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3521, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3521, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3520
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3519
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3522
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3522, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3516
 * 
 * 
 * cdef int _nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3526
 * 
 * 
 * cdef int _nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3528
 * cdef int _nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3528, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3529
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3530
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3531
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3531, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_min_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3531, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3531, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3530
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3529
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3532
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3532, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3526
 * 
 * 
 * cdef int _nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3536
 * 
 * 
 * cdef int _nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3538
 * cdef int _nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3538, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3539
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3540
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3541
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3541, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_min_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3541, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3541, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3540
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3539
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3542
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3542, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3536
 * 
 * 
 * cdef int _nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3546
 * 
 * 
 * cdef int _nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3548
 * cdef int _nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3548, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3549
 *     global __nvshmemx_size_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3550
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3551
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3551, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_min_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3551, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3551, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3550
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3549
 *     global __nvshmemx_size_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3552
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3552, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3546
 * 
 * 
 * cdef int _nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3556
 * 
 * 
 * cdef int _nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3558
 * cdef int _nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3558, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3559
 *     global __nvshmemx_char_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3560
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3561
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3561, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_min_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3561, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3561, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3560
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3559
 *     global __nvshmemx_char_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3562
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3562, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3556
 * 
 * 
 * cdef int _nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3566
 * 
 * 
 * cdef int _nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3568
 * cdef int _nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3568, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3569
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3571
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3571, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3571, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3571, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3569
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3572
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3572, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3566
 * 
 * 
 * cdef int _nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3576
 * 
 * 
 * cdef int _nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3578
 * cdef int _nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3578, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3579
 *     global __nvshmemx_short_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3580
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3581
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3581, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3581, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3581, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3580
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3579
 *     global __nvshmemx_short_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3582
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3582, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3576
 * 
 * 
 * cdef int _nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3586
 * 
 * 
 * cdef int _nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3588
 * cdef int _nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3588, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3589
 *     global __nvshmemx_int_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3590
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3591
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3591, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_min_reduce};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3591, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3591, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3590
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3589
 *     global __nvshmemx_int_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3592
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3592, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3586
 * 
 * 
 * cdef int _nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3596
 * 
 * 
 * cdef int _nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3598
 * cdef int _nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3598, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3599
 *     global __nvshmemx_long_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3600
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3601
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3601, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_min_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3601, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3601, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3600
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3599
 *     global __nvshmemx_long_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3602
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3602, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3596
 * 
 * 
 * cdef int _nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3606
 * 
 * 
 * cdef int _nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3608
 * cdef int _nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3608, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3609
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3610
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3611
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3611, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_min_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3611, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3611, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3610
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3609
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3612
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3612, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3606
 * 
 * 
 * cdef int _nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3616
 * 
 * 
 * cdef int _nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3618
 * cdef int _nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3618, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3619
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3620
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3621
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3621, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_min_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3621, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3621, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3620
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3619
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3622
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3622, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3616
 * 
 * 
 * cdef int _nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3626
 * 
 * 
 * cdef int _nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3628
 * cdef int _nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3628, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3629
 *     global __nvshmemx_half_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3630
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3631
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3631, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_min_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3631, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3631, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3630
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3629
 *     global __nvshmemx_half_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3632
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3632, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3626
 * 
 * 
 * cdef int _nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3636
 * 
 * 
 * cdef int _nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3638
 * cdef int _nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3638, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3639
 *     global __nvshmemx_float_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3640
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3641
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3641, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_min_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3641, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3641, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3640
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3639
 *     global __nvshmemx_float_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3642
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3642, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3636
 * 
 * 
 * cdef int _nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3646
 * 
 * 
 * cdef int _nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_min_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3648
 * cdef int _nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_min_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3648, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3649
 *     global __nvshmemx_double_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3650
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3651
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3651, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_min_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3651, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3651, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3650
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3649
 *     global __nvshmemx_double_min_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3652
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3652, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3646
 * 
 * 
 * cdef int _nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_min_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3656
 * 
 * 
 * cdef int _nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3658
 * cdef int _nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3658, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3659
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3660
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3661
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3661, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_sum_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3661, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3661, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3660
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3659
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3662
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3662, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3656
 * 
 * 
 * cdef int _nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3666
 * 
 * 
 * cdef int _nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3668
 * cdef int _nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3668, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3669
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3670
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3671
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3671, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3671, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3671, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3670
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3669
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3672
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3672, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3666
 * 
 * 
 * cdef int _nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3676
 * 
 * 
 * cdef int _nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3678
 * cdef int _nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3678, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3679
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3680
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3681
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3681, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3681, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3681, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3680
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3679
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3682
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3682, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3676
 * 
 * 
 * cdef int _nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3686
 * 
 * 
 * cdef int _nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3688
 * cdef int _nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3688, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3689
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3690
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3691
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3691, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3691, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3691, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3690
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3689
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3692
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3692, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3686
 * 
 * 
 * cdef int _nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3696
 * 
 * 
 * cdef int _nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3698
 * cdef int _nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3698, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3699
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3700
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3701
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3701, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3701, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3701, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3700
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3699
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3702
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3702, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3696
 * 
 * 
 * cdef int _nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3706
 * 
 * 
 * cdef int _nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3708
 * cdef int _nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3708, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3709
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3710
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3711
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3711, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_sum_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3711, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3711, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3710
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3709
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3712
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3712, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3706
 * 
 * 
 * cdef int _nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3716
 * 
 * 
 * cdef int _nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3718
 * cdef int _nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3718, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3719
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3720
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3721
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3721, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_sum_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3721, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3721, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3720
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3719
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3722
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3722, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3716
 * 
 * 
 * cdef int _nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3726
 * 
 * 
 * cdef int _nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3728
 * cdef int _nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3728, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3729
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3730
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3731
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3731, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_sum_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3731, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3731, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3730
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3729
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3732
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3732, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3726
 * 
 * 
 * cdef int _nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3736
 * 
 * 
 * cdef int _nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3738
 * cdef int _nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3738, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3739
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3740
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3741
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3741, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_sum_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3741, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3741, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3740
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3739
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3742
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3742, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3736
 * 
 * 
 * cdef int _nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3746
 * 
 * 
 * cdef int _nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3748
 * cdef int _nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3748, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3749
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3750
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3751
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3751, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_sum_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3751, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3751, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3750
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3749
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3752
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3752, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3746
 * 
 * 
 * cdef int _nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3756
 * 
 * 
 * cdef int _nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3758
 * cdef int _nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3758, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3759
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3760
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3761
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3761, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3761, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3761, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3760
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3759
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3762
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3762, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3756
 * 
 * 
 * cdef int _nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3766
 * 
 * 
 * cdef int _nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3768
 * cdef int _nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3768, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3769
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3770
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3771
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3771, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3771, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3771, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3770
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3769
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3772
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3772, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3766
 * 
 * 
 * cdef int _nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3776
 * 
 * 
 * cdef int _nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3778
 * cdef int _nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3778, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3779
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3780
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3781
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3781, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_sum_reduce};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3781, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3781, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3780
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3779
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3782
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3782, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3776
 * 
 * 
 * cdef int _nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3786
 * 
 * 
 * cdef int _nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3788
 * cdef int _nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3788, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3789
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3790
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3791
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3791, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_sum_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3791, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3791, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3790
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3789
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3792
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3792, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3786
 * 
 * 
 * cdef int _nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3796
 * 
 * 
 * cdef int _nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3798
 * cdef int _nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3798, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3799
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3800
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3801
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3801, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_sum_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3801, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3801, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3800
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3799
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3802
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3802, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3796
 * 
 * 
 * cdef int _nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3806
 * 
 * 
 * cdef int _nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3808
 * cdef int _nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3808, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3809
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3810
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3811
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3811, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_sum_r};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3811, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3811, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3810
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3809
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3812
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3812, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3806
 * 
 * 
 * cdef int _nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3816
 * 
 * 
 * cdef int _nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3818
 * cdef int _nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3818, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3819
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3820
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3821
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3821, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_sum_reduc};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3821, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3821, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3820
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3819
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3822
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3822, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3816
 * 
 * 
 * cdef int _nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3826
 * 
 * 
 * cdef int _nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3828
 * cdef int _nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3828, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3829
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3830
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3831
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3831, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_sum_redu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3831, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3831, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3830
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3829
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3832
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3832, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3826
 * 
 * 
 * cdef int _nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3836
 * 
 * 
 * cdef int _nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_sum_reduce_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3838
 * cdef int _nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3838, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3839
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3840
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reduce_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3841
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reduce_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3841, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_sum_red};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3841, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3841, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3840
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reduce_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3839
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reduce_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3842
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reduce_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reduce_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3842, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3836
 * 
 * 
 * cdef int _nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_sum_reduce_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3846
 * 
 * 
 * cdef int _nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3848
 * cdef int _nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3848, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3849
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3850
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3851
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3851, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_max_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3851, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3850
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3849
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3852
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3852, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3846
 * 
 * 
 * cdef int _nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3856
 * 
 * 
 * cdef int _nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3858
 * cdef int _nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3858, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3859
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3860
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3861
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3861, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3861, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3861, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3860
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3859
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3862
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3862, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3856
 * 
 * 
 * cdef int _nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3866
 * 
 * 
 * cdef int _nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3868
 * cdef int _nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3868, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3869
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3870
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3871
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3871, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3871, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3871, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3870
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3869
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3872
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3872, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3866
 * 
 * 
 * cdef int _nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3876
 * 
 * 
 * cdef int _nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3878
 * cdef int _nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3878, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3879
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3880
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3881
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3881, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3881, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3881, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3880
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3879
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3882
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3882, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3876
 * 
 * 
 * cdef int _nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3886
 * 
 * 
 * cdef int _nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3888
 * cdef int _nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3888, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3889
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3890
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3891
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3891, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3891, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3891, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3890
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3889
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3892
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3892, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3886
 * 
 * 
 * cdef int _nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3896
 * 
 * 
 * cdef int _nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3898
 * cdef int _nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3898, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3899
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3900
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3901
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3901, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_max_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3901, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3901, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3900
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3899
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3902
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3902, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3896
 * 
 * 
 * cdef int _nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3906
 * 
 * 
 * cdef int _nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3908
 * cdef int _nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3908, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3909
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3910
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3911
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3911, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_max_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3911, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3911, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3910
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3909
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3912
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3912, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3906
 * 
 * 
 * cdef int _nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3916
 * 
 * 
 * cdef int _nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3918
 * cdef int _nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3918, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3919
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3920
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3921
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3921, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_max_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3921, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3921, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3920
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3919
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3922
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3922, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3916
 * 
 * 
 * cdef int _nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3926
 * 
 * 
 * cdef int _nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3928
 * cdef int _nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3928, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3929
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3930
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3931
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3931, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_max_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3931, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3931, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3930
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3929
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3932
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3932, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3926
 * 
 * 
 * cdef int _nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3936
 * 
 * 
 * cdef int _nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3938
 * cdef int _nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3938, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3939
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3940
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3941
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3941, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_max_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3941, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3941, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3940
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3939
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3942
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3942, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3936
 * 
 * 
 * cdef int _nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3946
 * 
 * 
 * cdef int _nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3948
 * cdef int _nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3948, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3949
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3950
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3951
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3951, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3951, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3951, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3950
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3949
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3952
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3952, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3946
 * 
 * 
 * cdef int _nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3956
 * 
 * 
 * cdef int _nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3958
 * cdef int _nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3958, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3959
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3960
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3961
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3961, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3961, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3961, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3960
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3959
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3962
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3962, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3956
 * 
 * 
 * cdef int _nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3966
 * 
 * 
 * cdef int _nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3968
 * cdef int _nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3968, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3969
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3970
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3971
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3971, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_max_reduce_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3971, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3971, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3970
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3969
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3972
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3972, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3966
 * 
 * 
 * cdef int _nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3976
 * 
 * 
 * cdef int _nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3978
 * cdef int _nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3978, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3979
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3980
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3981
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3981, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_max_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3981, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3981, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3980
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3979
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3982
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3982, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3976
 * 
 * 
 * cdef int _nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3986
 * 
 * 
 * cdef int _nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3988
 * cdef int _nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3988, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3989
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3990
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":3991
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3991, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_max_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3991, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 3991, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":3990
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3989
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3992
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 3992, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3986
 * 
 * 
 * cdef int _nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":3996
 * 
 * 
 * cdef int _nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3998
 * cdef int _nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 3998, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3999
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4000
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4001
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4001, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_max_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4001, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4001, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4000
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":3999
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4002
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4002, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":3996
 * 
 * 
 * cdef int _nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4006
 * 
 * 
 * cdef int _nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4008
 * cdef int _nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4008, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4009
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4010
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4011
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4011, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_max_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4011, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4011, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4010
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4009
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4012
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4012, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4006
 * 
 * 
 * cdef int _nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4016
 * 
 * 
 * cdef int _nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4018
 * cdef int _nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4018, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4019
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4020
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4021
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4021, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_max_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4021, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4021, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4020
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4019
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4022
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4022, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4016
 * 
 * 
 * cdef int _nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4026
 * 
 * 
 * cdef int _nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_max_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4028
 * cdef int _nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4028, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4029
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4030
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4031
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4031, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_max_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4031, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4031, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4030
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4029
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_max_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4032
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_max_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_max_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4032, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4026
 * 
 * 
 * cdef int _nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_max_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4036
 * 
 * 
 * cdef int _nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4038
 * cdef int _nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4038, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4039
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4040
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4041
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4041, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_min_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4041, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4041, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4040
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4039
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4042
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4042, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4036
 * 
 * 
 * cdef int _nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4046
 * 
 * 
 * cdef int _nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4048
 * cdef int _nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4048, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4049
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4050
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4051
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4051, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4051, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4051, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4050
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4049
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4052
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4052, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4046
 * 
 * 
 * cdef int _nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4056
 * 
 * 
 * cdef int _nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4058
 * cdef int _nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4058, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4059
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4060
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4061
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4061, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4061, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4061, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4060
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4059
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4062
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4062, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4056
 * 
 * 
 * cdef int _nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4066
 * 
 * 
 * cdef int _nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4068
 * cdef int _nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4068, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4069
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4070
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4071
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4071, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4071, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4071, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4070
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4069
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4072
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4072, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4066
 * 
 * 
 * cdef int _nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4076
 * 
 * 
 * cdef int _nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4078
 * cdef int _nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4078, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4079
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4080
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4081
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4081, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4081, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4081, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4080
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4079
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4082
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4082, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4076
 * 
 * 
 * cdef int _nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4086
 * 
 * 
 * cdef int _nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4088
 * cdef int _nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4088, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4089
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4090
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4091
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4091, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_min_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4091, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4091, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4090
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4089
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4092
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4092, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4086
 * 
 * 
 * cdef int _nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4096
 * 
 * 
 * cdef int _nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4098
 * cdef int _nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4098, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4099
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4100
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4101
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4101, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_min_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4101, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4101, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4100
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4099
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4102
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4102, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4096
 * 
 * 
 * cdef int _nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4106
 * 
 * 
 * cdef int _nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4108
 * cdef int _nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4108, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4109
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4110
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4111
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4111, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_min_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4111, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4111, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4110
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4109
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4112
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4112, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4106
 * 
 * 
 * cdef int _nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4116
 * 
 * 
 * cdef int _nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4118
 * cdef int _nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4118, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4119
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4120
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4121
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4121, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_min_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4121, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4121, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4120
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4119
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4122
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4122, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4116
 * 
 * 
 * cdef int _nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4126
 * 
 * 
 * cdef int _nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4128
 * cdef int _nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4128, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4129
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4130
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4131
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4131, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_min_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4131, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4131, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4130
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4129
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4132
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4132, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4126
 * 
 * 
 * cdef int _nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4136
 * 
 * 
 * cdef int _nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4138
 * cdef int _nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4138, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4139
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4140
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4141
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4141, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4141, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4141, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4140
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4139
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4142
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4142, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4136
 * 
 * 
 * cdef int _nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4146
 * 
 * 
 * cdef int _nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4148
 * cdef int _nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4148, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4149
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4150
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4151
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4151, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4151, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4151, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4150
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4149
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4152
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4152, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4146
 * 
 * 
 * cdef int _nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4156
 * 
 * 
 * cdef int _nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4158
 * cdef int _nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4158, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4159
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4160
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4161
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4161, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_min_reduce_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4161, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4161, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4160
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4159
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4162
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4162, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4156
 * 
 * 
 * cdef int _nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4166
 * 
 * 
 * cdef int _nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4168
 * cdef int _nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4168, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4169
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4170
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4171
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4171, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_min_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4171, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4171, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4170
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4169
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4172
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4172, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4166
 * 
 * 
 * cdef int _nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4176
 * 
 * 
 * cdef int _nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4178
 * cdef int _nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4178, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4179
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4180
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4181
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4181, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_min_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4181, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4181, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4180
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4179
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4182
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4182, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4176
 * 
 * 
 * cdef int _nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4186
 * 
 * 
 * cdef int _nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4188
 * cdef int _nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4188, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4189
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4190
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4191
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4191, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_min_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4191, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4191, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4190
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4189
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4192
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4192, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4186
 * 
 * 
 * cdef int _nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4196
 * 
 * 
 * cdef int _nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4198
 * cdef int _nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4198, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4199
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4200
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4201
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4201, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_min_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4201, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4201, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4200
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4199
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4202
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4202, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4196
 * 
 * 
 * cdef int _nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4206
 * 
 * 
 * cdef int _nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4208
 * cdef int _nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4208, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4209
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4210
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4211
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4211, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_min_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4211, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4211, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4210
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4209
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4212
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4212, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4206
 * 
 * 
 * cdef int _nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4216
 * 
 * 
 * cdef int _nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_min_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4218
 * cdef int _nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4218, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4219
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4220
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4221
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4221, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_min_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4221, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4221, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4220
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4219
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_min_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4222
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_min_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_min_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4222, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4216
 * 
 * 
 * cdef int _nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_min_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4226
 * 
 * 
 * cdef int _nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int8_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4228
 * cdef int _nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4228, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4229
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4230
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4231
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4231, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int8_sum_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4231, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4231, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4230
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4229
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4232
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int8_t*, const int8_t*, size_t, cudaStream_t) nogil>__nvshmemx_int8_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4232, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4226
 * 
 * 
 * cdef int _nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4236
 * 
 * 
 * cdef int _nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int16_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4238
 * cdef int _nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4238, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4239
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4240
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4241
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4241, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int16_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4241, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4241, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4240
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4239
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4242
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int16_t*, const int16_t*, size_t, cudaStream_t) nogil>__nvshmemx_int16_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4242, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4236
 * 
 * 
 * cdef int _nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4246
 * 
 * 
 * cdef int _nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int32_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4248
 * cdef int _nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4248, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4249
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4250
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4251
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4251, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int32_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4251, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4251, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4250
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4249
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4252
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int32_t*, const int32_t*, size_t, cudaStream_t) nogil>__nvshmemx_int32_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4252, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4246
 * 
 * 
 * cdef int _nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4256
 * 
 * 
 * cdef int _nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int64_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4258
 * cdef int _nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4258, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4259
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4260
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4261
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4261, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int64_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4261, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4261, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4260
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4259
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4262
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int64_t*, const int64_t*, size_t, cudaStream_t) nogil>__nvshmemx_int64_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4262, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4256
 * 
 * 
 * cdef int _nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4266
 * 
 * 
 * cdef int _nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint8_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4268
 * cdef int _nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4268, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4269
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4270
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4271
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4271, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint8_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4271, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4270
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4269
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint8_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4272
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint8_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint8_t*, const uint8_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint8_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4272, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4266
 * 
 * 
 * cdef int _nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint8_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4276
 * 
 * 
 * cdef int _nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint16_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4278
 * cdef int _nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4278, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4279
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4280
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4281
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4281, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint16_sum_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4281, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4281, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4280
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4279
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4282
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint16_t*, const uint16_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint16_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4282, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4276
 * 
 * 
 * cdef int _nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4286
 * 
 * 
 * cdef int _nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint32_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4288
 * cdef int _nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4288, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4289
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4290
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4291
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4291, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint32_sum_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4291, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4291, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4290
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4289
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint32_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4292
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint32_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint32_t*, const uint32_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint32_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4292, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4286
 * 
 * 
 * cdef int _nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint32_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4296
 * 
 * 
 * cdef int _nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_uint64_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4298
 * cdef int _nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4298, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4299
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4300
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4301
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4301, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_uint64_sum_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4301, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4301, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4300
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4299
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_uint64_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4302
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_uint64_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, uint64_t*, const uint64_t*, size_t, cudaStream_t) nogil>__nvshmemx_uint64_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4302, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4296
 * 
 * 
 * cdef int _nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_uint64_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_uint64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4306
 * 
 * 
 * cdef int _nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_size_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4308
 * cdef int _nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4308, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4309
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4310
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4311
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4311, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_size_sum_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4311, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4311, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4310
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4309
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_size_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4312
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_size_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, size_t*, const size_t*, size_t, cudaStream_t) nogil>__nvshmemx_size_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4312, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4306
 * 
 * 
 * cdef int _nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_size_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_size_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4316
 * 
 * 
 * cdef int _nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_char_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4318
 * cdef int _nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4318, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4319
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4320
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4321
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4321, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_char_sum_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4321, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4321, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4320
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4319
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_char_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4322
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_char_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, char*, const char*, size_t, cudaStream_t) nogil>__nvshmemx_char_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4322, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4316
 * 
 * 
 * cdef int _nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_char_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_char_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4326
 * 
 * 
 * cdef int _nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_schar_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4328
 * cdef int _nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4328, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4329
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4330
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4331
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4331, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_schar_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4331, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4331, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4330
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4329
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_schar_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4332
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_schar_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, signed char*, const signed char*, size_t, cudaStream_t) nogil>__nvshmemx_schar_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4332, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4326
 * 
 * 
 * cdef int _nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_schar_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_schar_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4336
 * 
 * 
 * cdef int _nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_short_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4338
 * cdef int _nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4338, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4339
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4340
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4341
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4341, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_short_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4341, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4341, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4340
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4339
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_short_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4342
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_short_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, short*, const short*, size_t, cudaStream_t) nogil>__nvshmemx_short_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4342, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4336
 * 
 * 
 * cdef int _nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_short_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_short_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4346
 * 
 * 
 * cdef int _nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_int_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4348
 * cdef int _nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4348, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4349
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4350
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4351
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4351, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_int_sum_reduce_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4351, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4351, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4350
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4349
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_int_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4352
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_int_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, int*, const int*, size_t, cudaStream_t) nogil>__nvshmemx_int_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4352, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4346
 * 
 * 
 * cdef int _nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_int_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_int_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4356
 * 
 * 
 * cdef int _nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_long_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4358
 * cdef int _nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4358, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4359
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4360
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4361
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4361, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_long_sum_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4361, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4361, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4360
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4359
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_long_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4362
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_long_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long*, const long*, size_t, cudaStream_t) nogil>__nvshmemx_long_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4362, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4356
 * 
 * 
 * cdef int _nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_long_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_long_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4366
 * 
 * 
 * cdef int _nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_longlong_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4368
 * cdef int _nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4368, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4369
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4370
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4371
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4371, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_longlong_sum_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4371, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4371, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4370
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4369
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_longlong_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4372
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_longlong_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, long long*, const long long*, size_t, cudaStream_t) nogil>__nvshmemx_longlong_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4372, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4366
 * 
 * 
 * cdef int _nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_longlong_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_longlong_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4376
 * 
 * 
 * cdef int _nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_bfloat16_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4378
 * cdef int _nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4378, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4379
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4380
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4381
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4381, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_bfloat16_sum_r_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4381, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4381, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4380
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4379
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_bfloat16_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4382
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_bfloat16_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, __nv_bfloat16*, const __nv_bfloat16*, size_t, cudaStream_t) nogil>__nvshmemx_bfloat16_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4382, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4376
 * 
 * 
 * cdef int _nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_bfloat16_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_bfloat16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4386
 * 
 * 
 * cdef int _nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_half_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4388
 * cdef int _nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4388, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4389
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4390
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4391
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4391, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_half_sum_reduc_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4391, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4391, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4390
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4389
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_half_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4392
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_half_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, half*, const half*, size_t, cudaStream_t) nogil>__nvshmemx_half_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4392, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4386
 * 
 * 
 * cdef int _nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_half_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_half_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4396
 * 
 * 
 * cdef int _nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_float_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4398
 * cdef int _nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4398, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4399
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4400
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4401
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4401, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_float_sum_redu_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4401, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4401, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4400
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4399
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_float_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4402
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_float_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, float*, const float*, size_t, cudaStream_t) nogil>__nvshmemx_float_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4402, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4396
 * 
 * 
 * cdef int _nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_float_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_float_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4406
 * 
 * 
 * cdef int _nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_double_sum_reducescatter_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4408
 * cdef int _nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4408, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4409
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4410
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reducescatter_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4411
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reducescatter_on_stream)(
 *         team, dest, src, nreduce, stream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4411, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_double_sum_red_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4411, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4411, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4410
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reducescatter_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4409
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_double_sum_reducescatter_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4412
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_double_sum_reducescatter_on_stream is not found")
 *     return (<int (*)(nvshmem_team_t, double*, const double*, size_t, cudaStream_t) nogil>__nvshmemx_double_sum_reducescatter_on_stream)(             # <<<<<<<<<<<<<<
 *         team, dest, src, nreduce, stream)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream)(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4412, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4406
 * 
 * 
 * cdef int _nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_double_sum_reducescatter_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_double_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4416
 * 
 * 
 * cdef int _nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_hostlib_init_attr
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_init_attr(unsigned int __pyx_v_flags, nvshmemx_init_attr_t *__pyx_v_attr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_hostlib_init_attr", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4418
 * cdef int _nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:
 *     global __nvshmemx_hostlib_init_attr
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4418, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4419
 *     global __nvshmemx_hostlib_init_attr
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_init_attr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4420
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")
 *     return (<int (*)(unsigned int, nvshmemx_init_attr_t*) nogil>__nvshmemx_hostlib_init_attr)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4421
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(unsigned int, nvshmemx_init_attr_t*) nogil>__nvshmemx_hostlib_init_attr)(
 *         flags, attr)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4421, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_hostlib_init_a};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4421, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4421, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4420
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_init_attr == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")
 *     return (<int (*)(unsigned int, nvshmemx_init_attr_t*) nogil>__nvshmemx_hostlib_init_attr)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4419
 *     global __nvshmemx_hostlib_init_attr
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_init_attr == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4422
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_init_attr is not found")
 *     return (<int (*)(unsigned int, nvshmemx_init_attr_t*) nogil>__nvshmemx_hostlib_init_attr)(             # <<<<<<<<<<<<<<
 *         flags, attr)
 * 
*/
  __pyx_t_1 = ((int (*)(unsigned int, nvshmemx_init_attr_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr)(__pyx_v_flags, __pyx_v_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4422, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4416
 * 
 * 
 * cdef int _nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_hostlib_init_attr
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_hostlib_init_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4426
 * 
 * 
 * cdef void _nvshmemx_hostlib_finalize() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_hostlib_finalize
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_finalize(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_hostlib_finalize", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4428
 * cdef void _nvshmemx_hostlib_finalize() except* nogil:
 *     global __nvshmemx_hostlib_finalize
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4428, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4429
 *     global __nvshmemx_hostlib_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4430
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")
 *     (<void (*)() nogil>__nvshmemx_hostlib_finalize)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4431
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)() nogil>__nvshmemx_hostlib_finalize)(
 *         )
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4431, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_hostlib_finali};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4431, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4431, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4430
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")
 *     (<void (*)() nogil>__nvshmemx_hostlib_finalize)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4429
 *     global __nvshmemx_hostlib_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_hostlib_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4432
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_hostlib_finalize is not found")
 *     (<void (*)() nogil>__nvshmemx_hostlib_finalize)(             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  ((void (*)(void))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize)(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4432, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4426
 * 
 * 
 * cdef void _nvshmemx_hostlib_finalize() except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_hostlib_finalize
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_hostlib_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4436
 * 
 * 
 * cdef int _nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_set_attr_uniqueid_args
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_uniqueid_args(int const __pyx_v_myrank, int const __pyx_v_nranks, nvshmemx_uniqueid_t const *__pyx_v_uniqueid, nvshmemx_init_attr_t *__pyx_v_attr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_set_attr_uniqueid_args", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4438
 * cdef int _nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:
 *     global __nvshmemx_set_attr_uniqueid_args
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4438, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4439
 *     global __nvshmemx_set_attr_uniqueid_args
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4440
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")
 *     return (<int (*)(const int, const int, const nvshmemx_uniqueid_t*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_uniqueid_args)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4441
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(const int, const int, const nvshmemx_uniqueid_t*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_uniqueid_args)(
 *         myrank, nranks, uniqueid, attr)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4441, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_set_attr_uniqu};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4441, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4441, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4440
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")
 *     return (<int (*)(const int, const int, const nvshmemx_uniqueid_t*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_uniqueid_args)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4439
 *     global __nvshmemx_set_attr_uniqueid_args
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_uniqueid_args == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4442
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_uniqueid_args is not found")
 *     return (<int (*)(const int, const int, const nvshmemx_uniqueid_t*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_uniqueid_args)(             # <<<<<<<<<<<<<<
 *         myrank, nranks, uniqueid, attr)
 * 
*/
  __pyx_t_1 = ((int (*)(int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args)(__pyx_v_myrank, __pyx_v_nranks, __pyx_v_uniqueid, __pyx_v_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4442, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4436
 * 
 * 
 * cdef int _nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_set_attr_uniqueid_args
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_set_attr_uniqueid_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4446
 * 
 * 
 * cdef int _nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_mpi_comm_args(void *__pyx_v_mpi_comm, nvshmemx_init_attr_t *__pyx_v_nvshmem_attr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_set_attr_mpi_comm_args", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4448
 * cdef int _nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4448, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4449
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4450
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")
 *     return (<int (*)(void*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_mpi_comm_args)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4451
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(void*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_mpi_comm_args)(
 *         mpi_comm, nvshmem_attr)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4451, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_set_attr_mpi_c};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4451, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4451, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4450
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")
 *     return (<int (*)(void*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_mpi_comm_args)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4449
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_set_attr_mpi_comm_args == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4452
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_set_attr_mpi_comm_args is not found")
 *     return (<int (*)(void*, nvshmemx_init_attr_t*) nogil>__nvshmemx_set_attr_mpi_comm_args)(             # <<<<<<<<<<<<<<
 *         mpi_comm, nvshmem_attr)
 * 
*/
  __pyx_t_1 = ((int (*)(void *, nvshmemx_init_attr_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args)(__pyx_v_mpi_comm, __pyx_v_nvshmem_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4452, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4446
 * 
 * 
 * cdef int _nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_set_attr_mpi_comm_args
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_set_attr_mpi_comm_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4456
 * 
 * 
 * cdef int _nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_get_uniqueid
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_get_uniqueid(nvshmemx_uniqueid_t *__pyx_v_uniqueid) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_get_uniqueid", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4458
 * cdef int _nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:
 *     global __nvshmemx_get_uniqueid
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_get_uniqueid == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4458, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4459
 *     global __nvshmemx_get_uniqueid
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4460
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_get_uniqueid == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_uniqueid_t*) nogil>__nvshmemx_get_uniqueid)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4461
 *     if __nvshmemx_get_uniqueid == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(nvshmemx_uniqueid_t*) nogil>__nvshmemx_get_uniqueid)(
 *         uniqueid)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4461, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_get_uniqueid_i};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4461, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4461, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4460
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_get_uniqueid == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_uniqueid_t*) nogil>__nvshmemx_get_uniqueid)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4459
 *     global __nvshmemx_get_uniqueid
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_get_uniqueid == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4462
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_get_uniqueid is not found")
 *     return (<int (*)(nvshmemx_uniqueid_t*) nogil>__nvshmemx_get_uniqueid)(             # <<<<<<<<<<<<<<
 *         uniqueid)
 * 
*/
  __pyx_t_1 = ((int (*)(nvshmemx_uniqueid_t *))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid)(__pyx_v_uniqueid); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4462, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4456
 * 
 * 
 * cdef int _nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_get_uniqueid
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4466
 * 
 * 
 * cdef int _nvshmemx_cumodule_init(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_cumodule_init
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_init(CUmodule __pyx_v_module) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_cumodule_init", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4468
 * cdef int _nvshmemx_cumodule_init(CUmodule module) except* nogil:
 *     global __nvshmemx_cumodule_init
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_cumodule_init == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4468, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4469
 *     global __nvshmemx_cumodule_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_init)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4471
 *     if __nvshmemx_cumodule_init == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_init)(
 *         module)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4471, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_cumodule_init};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4471, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4471, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4470
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_init)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4469
 *     global __nvshmemx_cumodule_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4472
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_init is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_init)(             # <<<<<<<<<<<<<<
 *         module)
 * 
*/
  __pyx_t_1 = ((int (*)(CUmodule))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init)(__pyx_v_module); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4472, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4466
 * 
 * 
 * cdef int _nvshmemx_cumodule_init(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_cumodule_init
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_cumodule_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4476
 * 
 * 
 * cdef int _nvshmemx_cumodule_finalize(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_cumodule_finalize
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_finalize(CUmodule __pyx_v_module) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_cumodule_finalize", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4478
 * cdef int _nvshmemx_cumodule_finalize(CUmodule module) except* nogil:
 *     global __nvshmemx_cumodule_finalize
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4478, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4479
 *     global __nvshmemx_cumodule_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4480
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_finalize)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4481
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_finalize)(
 *         module)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4481, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_cumodule_final};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4481, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4481, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4480
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_finalize)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4479
 *     global __nvshmemx_cumodule_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_cumodule_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4482
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_cumodule_finalize is not found")
 *     return (<int (*)(CUmodule) nogil>__nvshmemx_cumodule_finalize)(             # <<<<<<<<<<<<<<
 *         module)
 * 
*/
  __pyx_t_1 = ((int (*)(CUmodule))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize)(__pyx_v_module); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4482, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4476
 * 
 * 
 * cdef int _nvshmemx_cumodule_finalize(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_cumodule_finalize
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_cumodule_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4486
 * 
 * 
 * cdef void* _nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_buffer_register_symmetric
 *     _check_or_init_nvshmem()
*/

static void *__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_register_symmetric(void *__pyx_v_buf_ptr, size_t __pyx_v_size, int __pyx_v_flags) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_buffer_register_symmetric", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4488
 * cdef void* _nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:
 *     global __nvshmemx_buffer_register_symmetric
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4488, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4489
 *     global __nvshmemx_buffer_register_symmetric
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_register_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4490
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")
 *     return (<void* (*)(void*, size_t, int) nogil>__nvshmemx_buffer_register_symmetric)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4491
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")             # <<<<<<<<<<<<<<
 *     return (<void* (*)(void*, size_t, int) nogil>__nvshmemx_buffer_register_symmetric)(
 *         buf_ptr, size, flags)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4491, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_buffer_registe};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4491, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4491, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4490
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_register_symmetric == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")
 *     return (<void* (*)(void*, size_t, int) nogil>__nvshmemx_buffer_register_symmetric)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4489
 *     global __nvshmemx_buffer_register_symmetric
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_register_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4492
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_register_symmetric is not found")
 *     return (<void* (*)(void*, size_t, int) nogil>__nvshmemx_buffer_register_symmetric)(             # <<<<<<<<<<<<<<
 *         buf_ptr, size, flags)
 * 
*/
  __pyx_t_7 = ((void *(*)(void *, size_t, int))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric)(__pyx_v_buf_ptr, __pyx_v_size, __pyx_v_flags); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4492, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4486
 * 
 * 
 * cdef void* _nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_buffer_register_symmetric
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_buffer_register_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4496
 * 
 * 
 * cdef int _nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_buffer_unregister_symmetric
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_unregister_symmetric(void *__pyx_v_mmap_ptr, size_t __pyx_v_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_buffer_unregister_symmetric", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4498
 * cdef int _nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:
 *     global __nvshmemx_buffer_unregister_symmetric
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4498, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4499
 *     global __nvshmemx_buffer_unregister_symmetric
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4500
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")
 *     return (<int (*)(void*, size_t) nogil>__nvshmemx_buffer_unregister_symmetric)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4501
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(void*, size_t) nogil>__nvshmemx_buffer_unregister_symmetric)(
 *         mmap_ptr, size)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4501, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_buffer_unregis};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4501, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4500
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")
 *     return (<int (*)(void*, size_t) nogil>__nvshmemx_buffer_unregister_symmetric)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4499
 *     global __nvshmemx_buffer_unregister_symmetric
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_buffer_unregister_symmetric == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4502
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_buffer_unregister_symmetric is not found")
 *     return (<int (*)(void*, size_t) nogil>__nvshmemx_buffer_unregister_symmetric)(             # <<<<<<<<<<<<<<
 *         mmap_ptr, size)
 * 
*/
  __pyx_t_1 = ((int (*)(void *, size_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric)(__pyx_v_mmap_ptr, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4502, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4496
 * 
 * 
 * cdef int _nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_buffer_unregister_symmetric
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_buffer_unregister_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4506
 * 
 * 
 * cdef int _nvshmemx_culibrary_init(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_culibrary_init
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_init(CUlibrary __pyx_v_library) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_culibrary_init", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4508
 * cdef int _nvshmemx_culibrary_init(CUlibrary library) except* nogil:
 *     global __nvshmemx_culibrary_init
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_culibrary_init == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4508, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4509
 *     global __nvshmemx_culibrary_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_init)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4511
 *     if __nvshmemx_culibrary_init == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_init)(
 *         library)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4511, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_culibrary_init};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4511, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4511, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4510
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_init == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_init)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4509
 *     global __nvshmemx_culibrary_init
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_init == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4512
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_init is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_init)(             # <<<<<<<<<<<<<<
 *         library)
 * 
*/
  __pyx_t_1 = ((int (*)(CUlibrary))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init)(__pyx_v_library); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4512, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4506
 * 
 * 
 * cdef int _nvshmemx_culibrary_init(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_culibrary_init
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_culibrary_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4516
 * 
 * 
 * cdef int _nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_culibrary_finalize
 *     _check_or_init_nvshmem()
*/

static int __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_finalize(CUlibrary __pyx_v_library) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_culibrary_finalize", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4518
 * cdef int _nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:
 *     global __nvshmemx_culibrary_finalize
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4518, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4519
 *     global __nvshmemx_culibrary_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4520
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_finalize)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4521
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")             # <<<<<<<<<<<<<<
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_finalize)(
 *         library)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4521, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_culibrary_fina};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4521, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4521, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4520
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_finalize == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_finalize)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4519
 *     global __nvshmemx_culibrary_finalize
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_culibrary_finalize == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4522
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_culibrary_finalize is not found")
 *     return (<int (*)(CUlibrary) nogil>__nvshmemx_culibrary_finalize)(             # <<<<<<<<<<<<<<
 *         library)
 * 
*/
  __pyx_t_1 = ((int (*)(CUlibrary))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize)(__pyx_v_library); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4522, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4516
 * 
 * 
 * cdef int _nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_culibrary_finalize
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_culibrary_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
  return __pyx_r;
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4526
 * 
 * 
 * cdef void _nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_putmem_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_putmem_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4528
 * cdef void _nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:
 *     global __nvshmemx_putmem_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4528, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4529
 *     global __nvshmemx_putmem_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4530
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_putmem_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4531
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_putmem_on_stream)(
 *         dest, source, bytes, pe, cstrm)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4531, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_putmem_on_stre};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4531, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4531, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4530
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_putmem_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4529
 *     global __nvshmemx_putmem_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4532
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_putmem_on_stream)(             # <<<<<<<<<<<<<<
 *         dest, source, bytes, pe, cstrm)
 * 
*/
  ((void (*)(void *, void const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream)(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4532, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4526
 * 
 * 
 * cdef void _nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_putmem_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_putmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4536
 * 
 * 
 * cdef void _nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_putmem_signal_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_signal_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, uint64_t *__pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_putmem_signal_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4538
 * cdef void _nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:
 *     global __nvshmemx_putmem_signal_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4538, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4539
 *     global __nvshmemx_putmem_signal_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_signal_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4540
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_putmem_signal_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4541
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(void*, const void*, size_t, uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_putmem_signal_on_stream)(
 *         dest, source, bytes, sig_addr, signal, sig_op, pe, cstrm)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4541, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_putmem_signal};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4541, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4541, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4540
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_signal_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_putmem_signal_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4539
 *     global __nvshmemx_putmem_signal_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_putmem_signal_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4542
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_putmem_signal_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_putmem_signal_on_stream)(             # <<<<<<<<<<<<<<
 *         dest, source, bytes, sig_addr, signal, sig_op, pe, cstrm)
 * 
*/
  ((void (*)(void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream)(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4542, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4536
 * 
 * 
 * cdef void _nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_putmem_signal_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_putmem_signal_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4546
 * 
 * 
 * cdef void _nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_getmem_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_getmem_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_getmem_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4548
 * cdef void _nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:
 *     global __nvshmemx_getmem_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4548, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4549
 *     global __nvshmemx_getmem_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_getmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4550
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_getmem_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4551
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_getmem_on_stream)(
 *         dest, source, bytes, pe, cstrm)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4551, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_getmem_on_stre};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4551, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4551, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4550
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_getmem_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_getmem_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4549
 *     global __nvshmemx_getmem_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_getmem_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4552
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_getmem_on_stream is not found")
 *     (<void (*)(void*, const void*, size_t, int, cudaStream_t) nogil>__nvshmemx_getmem_on_stream)(             # <<<<<<<<<<<<<<
 *         dest, source, bytes, pe, cstrm)
 * 
*/
  ((void (*)(void *, void const *, size_t, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream)(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4552, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4546
 * 
 * 
 * cdef void _nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_getmem_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_getmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4556
 * 
 * 
 * cdef void _nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_quiet_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_quiet_on_stream(cudaStream_t __pyx_v_cstrm) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_quiet_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4558
 * cdef void _nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:
 *     global __nvshmemx_quiet_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4558, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4559
 *     global __nvshmemx_quiet_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_quiet_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4560
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_quiet_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4561
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_quiet_on_stream)(
 *         cstrm)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4561, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_quiet_on_strea};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4561, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4561, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4560
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_quiet_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_quiet_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4559
 *     global __nvshmemx_quiet_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_quiet_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4562
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_quiet_on_stream is not found")
 *     (<void (*)(cudaStream_t) nogil>__nvshmemx_quiet_on_stream)(             # <<<<<<<<<<<<<<
 *         cstrm)
 * 
*/
  ((void (*)(cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream)(__pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4562, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4556
 * 
 * 
 * cdef void _nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_quiet_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_quiet_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4566
 * 
 * 
 * cdef void _nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_signal_op_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_op_on_stream(uint64_t *__pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_signal_op_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4568
 * cdef void _nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:
 *     global __nvshmemx_signal_op_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4568, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4569
 *     global __nvshmemx_signal_op_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_op_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")
 *     (<void (*)(uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_signal_op_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4571
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_signal_op_on_stream)(
 *         sig_addr, signal, sig_op, pe, cstrm)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4571, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_signal_op_on_s};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4571, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4571, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4570
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_op_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")
 *     (<void (*)(uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_signal_op_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4569
 *     global __nvshmemx_signal_op_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_op_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4572
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_op_on_stream is not found")
 *     (<void (*)(uint64_t*, uint64_t, int, int, cudaStream_t) nogil>__nvshmemx_signal_op_on_stream)(             # <<<<<<<<<<<<<<
 *         sig_addr, signal, sig_op, pe, cstrm)
 * 
*/
  ((void (*)(uint64_t *, uint64_t, int, int, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream)(__pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4572, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4566
 * 
 * 
 * cdef void _nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_signal_op_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_signal_op_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "nvshmem/bindings/_internal/nvshmem.pyx":4576
 * 
 * 
 * cdef void _nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_signal_wait_until_on_stream
 *     _check_or_init_nvshmem()
*/

static void __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_wait_until_on_stream(uint64_t *__pyx_v_sig_addr, int __pyx_v_cmp, uint64_t __pyx_v_cmp_value, cudaStream_t __pyx_v_cstream) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;
  __Pyx_RefNannySetupContext("_nvshmemx_signal_wait_until_on_stream", 1);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4578
 * cdef void _nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:
 *     global __nvshmemx_signal_wait_until_on_stream
 *     _check_or_init_nvshmem()             # <<<<<<<<<<<<<<
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         with gil:
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__check_or_init_nvshmem(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4578, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4579
 *     global __nvshmemx_signal_wait_until_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")
*/
  __pyx_t_2 = (__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4580
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")
 *     (<void (*)(uint64_t*, int, uint64_t, cudaStream_t) nogil>__nvshmemx_signal_wait_until_on_stream)(
*/
    {
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        /*try:*/ {

          /* "nvshmem/bindings/_internal/nvshmem.pyx":4581
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")             # <<<<<<<<<<<<<<
 *     (<void (*)(uint64_t*, int, uint64_t, cudaStream_t) nogil>__nvshmemx_signal_wait_until_on_stream)(
 *         sig_addr, cmp, cmp_value, cstream)
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4581, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_function_nvshmemx_signal_wait_un};
            __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4581, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 4581, __pyx_L5_error)
        }

        /* "nvshmem/bindings/_internal/nvshmem.pyx":4580
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")
 *     (<void (*)(uint64_t*, int, uint64_t, cudaStream_t) nogil>__nvshmemx_signal_wait_until_on_stream)(
*/
        /*finally:*/ {
          __pyx_L5_error: {
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            goto __pyx_L1_error;
          }
        }
    }

    /* "nvshmem/bindings/_internal/nvshmem.pyx":4579
 *     global __nvshmemx_signal_wait_until_on_stream
 *     _check_or_init_nvshmem()
 *     if __nvshmemx_signal_wait_until_on_stream == NULL:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")
*/
  }

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4582
 *         with gil:
 *             raise FunctionNotFoundError("function nvshmemx_signal_wait_until_on_stream is not found")
 *     (<void (*)(uint64_t*, int, uint64_t, cudaStream_t) nogil>__nvshmemx_signal_wait_until_on_stream)(             # <<<<<<<<<<<<<<
 *         sig_addr, cmp, cmp_value, cstream)
*/
  ((void (*)(uint64_t *, int, uint64_t, cudaStream_t))__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream)(__pyx_v_sig_addr, __pyx_v_cmp, __pyx_v_cmp_value, __pyx_v_cstream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 4582, __pyx_L1_error)

  /* "nvshmem/bindings/_internal/nvshmem.pyx":4576
 * 
 * 
 * cdef void _nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:             # <<<<<<<<<<<<<<
 *     global __nvshmemx_signal_wait_until_on_stream
 *     _check_or_init_nvshmem()
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings._internal.nvshmem._nvshmemx_signal_wait_until_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  {
    __pyx_t_1 = __Pyx_ApiExport_GetApiDict(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_export_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_int_CUlibrary_int_CUmodule_int_i);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_export_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_export_name = __pyx_export_signature + 4132;
    void (*const __pyx_export_pointers[])(void) = {(void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_uniqueid_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_sync_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_strided, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_2d, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_translate_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_init_attr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_mpi_comm_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_unregister_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_init_status, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_quiet_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_sync_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_info_get_version, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_vendor_get_version_info, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_destroy, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_get_config, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_wait_until_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_op_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_free, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_getmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_signal_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier_all, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_mc_ptr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_malloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_align, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_calloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_register_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_ptr, (void (*)(void)) NULL};
    void (*const *__pyx_export_pointer)(void) = __pyx_export_pointers;
    const char *__pyx_export_current_signature = __pyx_export_signature;
    while (*__pyx_export_pointer) {
      if (__Pyx_ExportFunction(__pyx_t_1, __pyx_export_name, *__pyx_export_pointer, __pyx_export_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_export_pointer;
      __pyx_export_name = strchr(__pyx_export_name, '\0') + 1;
      __pyx_export_signature = strchr(__pyx_export_signature, '\0') + 1;
      if (*__pyx_export_signature != '\0') __pyx_export_current_signature = __pyx_export_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_nvshmem(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_nvshmem},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "nvshmem",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_nvshmem(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_nvshmem(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_nvshmem(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'nvshmem' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "nvshmem" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_nvshmem", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_nvshmem__bindings___internal__nvshmem) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name_2, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "nvshmem.bindings._internal.nvshmem")) {
      if (unlikely((PyDict_SetItemString(modules, "nvshmem.bindings._internal.nvshmem", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_export_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  (void)__Pyx_modinit_type_import_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "nvshmem/bindings/_internal/nvshmem.pyx":9
 * from libc.stdint cimport intptr_t
 * 
 * class FunctionNotFoundError(RuntimeError): pass             # <<<<<<<<<<<<<<
 * 
 * class NotSupportedError(RuntimeError): pass
*/
  __pyx_t_2 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeError)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_3, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings__internal_nvshm, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 9, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError, __pyx_t_3, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_FunctionNotFoundError, __pyx_t_2) < (0)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":11
 * class FunctionNotFoundError(RuntimeError): pass
 * 
 * class NotSupportedError(RuntimeError): pass             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_RuntimeError)))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_mstate_global->__pyx_n_u_NotSupportedError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings__internal_nvshm, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 11, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_t_4, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NotSupportedError, __pyx_t_3) < (0)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":41
 * ###############################################################################
 * 
 * cdef bint __py_nvshmem_init = False             # <<<<<<<<<<<<<<
 * 
 * cdef void* __nvshmem_barrier = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___py_nvshmem_init = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":43
 * cdef bint __py_nvshmem_init = False
 * 
 * cdef void* __nvshmem_barrier = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_barrier_all = NULL
 * cdef void* __nvshmemx_init_status = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":44
 * 
 * cdef void* __nvshmem_barrier = NULL
 * cdef void* __nvshmem_barrier_all = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_init_status = NULL
 * cdef void* __nvshmem_my_pe = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_barrier_all = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":45
 * cdef void* __nvshmem_barrier = NULL
 * cdef void* __nvshmem_barrier_all = NULL
 * cdef void* __nvshmemx_init_status = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_my_pe = NULL
 * cdef void* __nvshmem_n_pes = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_init_status = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":46
 * cdef void* __nvshmem_barrier_all = NULL
 * cdef void* __nvshmemx_init_status = NULL
 * cdef void* __nvshmem_my_pe = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_n_pes = NULL
 * cdef void* __nvshmem_info_get_version = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_my_pe = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":47
 * cdef void* __nvshmemx_init_status = NULL
 * cdef void* __nvshmem_my_pe = NULL
 * cdef void* __nvshmem_n_pes = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_info_get_version = NULL
 * cdef void* __nvshmemx_vendor_get_version_info = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_n_pes = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":48
 * cdef void* __nvshmem_my_pe = NULL
 * cdef void* __nvshmem_n_pes = NULL
 * cdef void* __nvshmem_info_get_version = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_vendor_get_version_info = NULL
 * cdef void* __nvshmem_malloc = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_info_get_version = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":49
 * cdef void* __nvshmem_n_pes = NULL
 * cdef void* __nvshmem_info_get_version = NULL
 * cdef void* __nvshmemx_vendor_get_version_info = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_malloc = NULL
 * cdef void* __nvshmem_calloc = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_vendor_get_version_info = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":50
 * cdef void* __nvshmem_info_get_version = NULL
 * cdef void* __nvshmemx_vendor_get_version_info = NULL
 * cdef void* __nvshmem_malloc = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_calloc = NULL
 * cdef void* __nvshmem_align = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_malloc = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":51
 * cdef void* __nvshmemx_vendor_get_version_info = NULL
 * cdef void* __nvshmem_malloc = NULL
 * cdef void* __nvshmem_calloc = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_align = NULL
 * cdef void* __nvshmem_free = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_calloc = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":52
 * cdef void* __nvshmem_malloc = NULL
 * cdef void* __nvshmem_calloc = NULL
 * cdef void* __nvshmem_align = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_free = NULL
 * cdef void* __nvshmem_ptr = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_align = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":53
 * cdef void* __nvshmem_calloc = NULL
 * cdef void* __nvshmem_align = NULL
 * cdef void* __nvshmem_free = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_ptr = NULL
 * cdef void* __nvshmemx_mc_ptr = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_free = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":54
 * cdef void* __nvshmem_align = NULL
 * cdef void* __nvshmem_free = NULL
 * cdef void* __nvshmem_ptr = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_mc_ptr = NULL
 * cdef void* __nvshmem_team_my_pe = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_ptr = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":55
 * cdef void* __nvshmem_free = NULL
 * cdef void* __nvshmem_ptr = NULL
 * cdef void* __nvshmemx_mc_ptr = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_my_pe = NULL
 * cdef void* __nvshmem_team_n_pes = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_mc_ptr = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":56
 * cdef void* __nvshmem_ptr = NULL
 * cdef void* __nvshmemx_mc_ptr = NULL
 * cdef void* __nvshmem_team_my_pe = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_n_pes = NULL
 * cdef void* __nvshmem_team_get_config = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_my_pe = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":57
 * cdef void* __nvshmemx_mc_ptr = NULL
 * cdef void* __nvshmem_team_my_pe = NULL
 * cdef void* __nvshmem_team_n_pes = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_get_config = NULL
 * cdef void* __nvshmem_team_translate_pe = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_n_pes = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":58
 * cdef void* __nvshmem_team_my_pe = NULL
 * cdef void* __nvshmem_team_n_pes = NULL
 * cdef void* __nvshmem_team_get_config = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_translate_pe = NULL
 * cdef void* __nvshmem_team_split_strided = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_get_config = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":59
 * cdef void* __nvshmem_team_n_pes = NULL
 * cdef void* __nvshmem_team_get_config = NULL
 * cdef void* __nvshmem_team_translate_pe = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_split_strided = NULL
 * cdef void* __nvshmemx_team_get_uniqueid = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_translate_pe = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":60
 * cdef void* __nvshmem_team_get_config = NULL
 * cdef void* __nvshmem_team_translate_pe = NULL
 * cdef void* __nvshmem_team_split_strided = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_team_get_uniqueid = NULL
 * cdef void* __nvshmemx_team_init = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_strided = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":61
 * cdef void* __nvshmem_team_translate_pe = NULL
 * cdef void* __nvshmem_team_split_strided = NULL
 * cdef void* __nvshmemx_team_get_uniqueid = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_team_init = NULL
 * cdef void* __nvshmem_team_split_2d = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_get_uniqueid = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":62
 * cdef void* __nvshmem_team_split_strided = NULL
 * cdef void* __nvshmemx_team_get_uniqueid = NULL
 * cdef void* __nvshmemx_team_init = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_split_2d = NULL
 * cdef void* __nvshmem_team_destroy = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_init = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":63
 * cdef void* __nvshmemx_team_get_uniqueid = NULL
 * cdef void* __nvshmemx_team_init = NULL
 * cdef void* __nvshmem_team_split_2d = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmem_team_destroy = NULL
 * cdef void* __nvshmemx_bfloat16_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_split_2d = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":64
 * cdef void* __nvshmemx_team_init = NULL
 * cdef void* __nvshmem_team_split_2d = NULL
 * cdef void* __nvshmem_team_destroy = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_half_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmem_team_destroy = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":65
 * cdef void* __nvshmem_team_split_2d = NULL
 * cdef void* __nvshmem_team_destroy = NULL
 * cdef void* __nvshmemx_bfloat16_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_float_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":66
 * cdef void* __nvshmem_team_destroy = NULL
 * cdef void* __nvshmemx_bfloat16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_half_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_double_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":67
 * cdef void* __nvshmemx_bfloat16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_half_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_float_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_char_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":68
 * cdef void* __nvshmemx_half_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_float_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_double_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_short_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":69
 * cdef void* __nvshmemx_float_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_double_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_char_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_schar_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":70
 * cdef void* __nvshmemx_double_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_char_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_short_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":71
 * cdef void* __nvshmemx_char_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_short_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_schar_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_long_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":72
 * cdef void* __nvshmemx_short_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_schar_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_longlong_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":73
 * cdef void* __nvshmemx_schar_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_long_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int8_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":74
 * cdef void* __nvshmemx_int_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_long_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_longlong_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int16_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":75
 * cdef void* __nvshmemx_long_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_longlong_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int8_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int32_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":76
 * cdef void* __nvshmemx_longlong_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int16_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int64_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":77
 * cdef void* __nvshmemx_int8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int32_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint8_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":78
 * cdef void* __nvshmemx_int16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int64_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint16_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":79
 * cdef void* __nvshmemx_int32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_int64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint8_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint32_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":80
 * cdef void* __nvshmemx_int64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint16_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint64_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":81
 * cdef void* __nvshmemx_uint8_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint32_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_size_alltoall_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":82
 * cdef void* __nvshmemx_uint16_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint64_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_barrier_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":83
 * cdef void* __nvshmemx_uint32_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_uint64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_size_alltoall_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_barrier_on_stream = NULL
 * cdef void* __nvshmemx_barrier_all_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_alltoall_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":84
 * cdef void* __nvshmemx_uint64_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_size_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_barrier_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_barrier_all_on_stream = NULL
 * cdef void* __nvshmemx_team_sync_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":85
 * cdef void* __nvshmemx_size_alltoall_on_stream = NULL
 * cdef void* __nvshmemx_barrier_on_stream = NULL
 * cdef void* __nvshmemx_barrier_all_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_team_sync_on_stream = NULL
 * cdef void* __nvshmemx_sync_all_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_barrier_all_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":86
 * cdef void* __nvshmemx_barrier_on_stream = NULL
 * cdef void* __nvshmemx_barrier_all_on_stream = NULL
 * cdef void* __nvshmemx_team_sync_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_sync_all_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_team_sync_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":87
 * cdef void* __nvshmemx_barrier_all_on_stream = NULL
 * cdef void* __nvshmemx_team_sync_on_stream = NULL
 * cdef void* __nvshmemx_sync_all_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_half_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_sync_all_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":88
 * cdef void* __nvshmemx_team_sync_on_stream = NULL
 * cdef void* __nvshmemx_sync_all_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_float_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":89
 * cdef void* __nvshmemx_sync_all_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_half_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_double_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":90
 * cdef void* __nvshmemx_bfloat16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_half_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_float_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_char_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":91
 * cdef void* __nvshmemx_half_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_float_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_double_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_short_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":92
 * cdef void* __nvshmemx_float_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_double_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_char_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_schar_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":93
 * cdef void* __nvshmemx_double_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_char_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_short_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":94
 * cdef void* __nvshmemx_char_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_short_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_schar_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_long_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":95
 * cdef void* __nvshmemx_short_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_schar_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_longlong_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":96
 * cdef void* __nvshmemx_schar_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_long_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int8_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":97
 * cdef void* __nvshmemx_int_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_long_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_longlong_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int16_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":98
 * cdef void* __nvshmemx_long_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_longlong_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int8_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int32_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":99
 * cdef void* __nvshmemx_longlong_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int16_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int64_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":100
 * cdef void* __nvshmemx_int8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int32_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint8_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":101
 * cdef void* __nvshmemx_int16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int64_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint16_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":102
 * cdef void* __nvshmemx_int32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_int64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint8_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint32_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":103
 * cdef void* __nvshmemx_int64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint16_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint64_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":104
 * cdef void* __nvshmemx_uint8_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint32_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_size_broadcast_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":105
 * cdef void* __nvshmemx_uint16_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint64_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":106
 * cdef void* __nvshmemx_uint32_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_uint64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_size_broadcast_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_half_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_broadcast_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":107
 * cdef void* __nvshmemx_uint64_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_size_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_float_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":108
 * cdef void* __nvshmemx_size_broadcast_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_half_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_double_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":109
 * cdef void* __nvshmemx_bfloat16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_half_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_float_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_char_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":110
 * cdef void* __nvshmemx_half_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_float_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_double_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_short_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":111
 * cdef void* __nvshmemx_float_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_double_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_char_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_schar_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":112
 * cdef void* __nvshmemx_double_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_char_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_short_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":113
 * cdef void* __nvshmemx_char_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_short_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_schar_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_long_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":114
 * cdef void* __nvshmemx_short_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_schar_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_longlong_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":115
 * cdef void* __nvshmemx_schar_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_long_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":116
 * cdef void* __nvshmemx_int_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_long_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_longlong_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int16_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":117
 * cdef void* __nvshmemx_long_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_longlong_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int32_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":118
 * cdef void* __nvshmemx_longlong_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int16_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int64_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":119
 * cdef void* __nvshmemx_int8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int32_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint8_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":120
 * cdef void* __nvshmemx_int16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int64_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint16_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":121
 * cdef void* __nvshmemx_int32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint8_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint32_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":122
 * cdef void* __nvshmemx_int64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint16_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint64_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":123
 * cdef void* __nvshmemx_uint8_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint32_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_size_fcollect_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":124
 * cdef void* __nvshmemx_uint16_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint64_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":125
 * cdef void* __nvshmemx_uint32_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_uint64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_size_fcollect_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_fcollect_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":126
 * cdef void* __nvshmemx_uint64_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_size_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":127
 * cdef void* __nvshmemx_size_fcollect_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":128
 * cdef void* __nvshmemx_int8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":129
 * cdef void* __nvshmemx_int16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":130
 * cdef void* __nvshmemx_int32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":131
 * cdef void* __nvshmemx_int64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":132
 * cdef void* __nvshmemx_uint8_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":133
 * cdef void* __nvshmemx_uint16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":134
 * cdef void* __nvshmemx_uint32_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":135
 * cdef void* __nvshmemx_uint64_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":136
 * cdef void* __nvshmemx_size_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":137
 * cdef void* __nvshmemx_char_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":138
 * cdef void* __nvshmemx_schar_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":139
 * cdef void* __nvshmemx_short_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":140
 * cdef void* __nvshmemx_int_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":141
 * cdef void* __nvshmemx_long_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":142
 * cdef void* __nvshmemx_longlong_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":143
 * cdef void* __nvshmemx_bfloat16_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":144
 * cdef void* __nvshmemx_half_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":145
 * cdef void* __nvshmemx_float_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":146
 * cdef void* __nvshmemx_double_max_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":147
 * cdef void* __nvshmemx_int8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":148
 * cdef void* __nvshmemx_int16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":149
 * cdef void* __nvshmemx_int32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":150
 * cdef void* __nvshmemx_int64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":151
 * cdef void* __nvshmemx_uint8_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":152
 * cdef void* __nvshmemx_uint16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":153
 * cdef void* __nvshmemx_uint32_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":154
 * cdef void* __nvshmemx_uint64_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":155
 * cdef void* __nvshmemx_size_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":156
 * cdef void* __nvshmemx_char_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":157
 * cdef void* __nvshmemx_schar_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":158
 * cdef void* __nvshmemx_short_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":159
 * cdef void* __nvshmemx_int_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":160
 * cdef void* __nvshmemx_long_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":161
 * cdef void* __nvshmemx_longlong_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":162
 * cdef void* __nvshmemx_bfloat16_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":163
 * cdef void* __nvshmemx_half_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":164
 * cdef void* __nvshmemx_float_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":165
 * cdef void* __nvshmemx_double_min_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":166
 * cdef void* __nvshmemx_int8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":167
 * cdef void* __nvshmemx_int16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":168
 * cdef void* __nvshmemx_int32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":169
 * cdef void* __nvshmemx_int64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":170
 * cdef void* __nvshmemx_uint8_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":171
 * cdef void* __nvshmemx_uint16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":172
 * cdef void* __nvshmemx_uint32_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":173
 * cdef void* __nvshmemx_uint64_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":174
 * cdef void* __nvshmemx_size_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":175
 * cdef void* __nvshmemx_char_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":176
 * cdef void* __nvshmemx_schar_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":177
 * cdef void* __nvshmemx_short_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":178
 * cdef void* __nvshmemx_int_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":179
 * cdef void* __nvshmemx_long_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":180
 * cdef void* __nvshmemx_longlong_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reduce_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":181
 * cdef void* __nvshmemx_bfloat16_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":182
 * cdef void* __nvshmemx_half_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reduce_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reduce_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":183
 * cdef void* __nvshmemx_float_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":184
 * cdef void* __nvshmemx_double_sum_reduce_on_stream = NULL
 * cdef void* __nvshmemx_int8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":185
 * cdef void* __nvshmemx_int8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":186
 * cdef void* __nvshmemx_int16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":187
 * cdef void* __nvshmemx_int32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":188
 * cdef void* __nvshmemx_int64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":189
 * cdef void* __nvshmemx_uint8_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":190
 * cdef void* __nvshmemx_uint16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":191
 * cdef void* __nvshmemx_uint32_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":192
 * cdef void* __nvshmemx_uint64_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":193
 * cdef void* __nvshmemx_size_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":194
 * cdef void* __nvshmemx_char_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":195
 * cdef void* __nvshmemx_schar_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":196
 * cdef void* __nvshmemx_short_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":197
 * cdef void* __nvshmemx_int_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":198
 * cdef void* __nvshmemx_long_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":199
 * cdef void* __nvshmemx_longlong_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":200
 * cdef void* __nvshmemx_bfloat16_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":201
 * cdef void* __nvshmemx_half_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_max_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":202
 * cdef void* __nvshmemx_float_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":203
 * cdef void* __nvshmemx_double_max_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":204
 * cdef void* __nvshmemx_int8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":205
 * cdef void* __nvshmemx_int16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":206
 * cdef void* __nvshmemx_int32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":207
 * cdef void* __nvshmemx_int64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":208
 * cdef void* __nvshmemx_uint8_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":209
 * cdef void* __nvshmemx_uint16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":210
 * cdef void* __nvshmemx_uint32_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":211
 * cdef void* __nvshmemx_uint64_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":212
 * cdef void* __nvshmemx_size_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":213
 * cdef void* __nvshmemx_char_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":214
 * cdef void* __nvshmemx_schar_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":215
 * cdef void* __nvshmemx_short_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":216
 * cdef void* __nvshmemx_int_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":217
 * cdef void* __nvshmemx_long_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":218
 * cdef void* __nvshmemx_longlong_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":219
 * cdef void* __nvshmemx_bfloat16_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":220
 * cdef void* __nvshmemx_half_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_min_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":221
 * cdef void* __nvshmemx_float_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int8_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":222
 * cdef void* __nvshmemx_double_min_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int16_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":223
 * cdef void* __nvshmemx_int8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int32_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":224
 * cdef void* __nvshmemx_int16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int64_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":225
 * cdef void* __nvshmemx_int32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint8_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":226
 * cdef void* __nvshmemx_int64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint16_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":227
 * cdef void* __nvshmemx_uint8_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_uint64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint32_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":228
 * cdef void* __nvshmemx_uint16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_size_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_uint64_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":229
 * cdef void* __nvshmemx_uint32_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_uint64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_char_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_size_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":230
 * cdef void* __nvshmemx_uint64_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_size_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_schar_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_char_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":231
 * cdef void* __nvshmemx_size_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_char_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_short_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_schar_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":232
 * cdef void* __nvshmemx_char_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_schar_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_int_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_short_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":233
 * cdef void* __nvshmemx_schar_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_short_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_long_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_int_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":234
 * cdef void* __nvshmemx_short_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_int_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_longlong_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_long_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":235
 * cdef void* __nvshmemx_int_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_long_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_longlong_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":236
 * cdef void* __nvshmemx_long_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_longlong_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_half_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":237
 * cdef void* __nvshmemx_longlong_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_float_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reducescatter_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_half_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":238
 * cdef void* __nvshmemx_bfloat16_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_half_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_double_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_hostlib_init_attr = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_float_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":239
 * cdef void* __nvshmemx_half_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_float_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reducescatter_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_hostlib_init_attr = NULL
 * cdef void* __nvshmemx_hostlib_finalize = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_double_sum_reducescatter_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":240
 * cdef void* __nvshmemx_float_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_double_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_hostlib_init_attr = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_hostlib_finalize = NULL
 * cdef void* __nvshmemx_set_attr_uniqueid_args = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_init_attr = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":241
 * cdef void* __nvshmemx_double_sum_reducescatter_on_stream = NULL
 * cdef void* __nvshmemx_hostlib_init_attr = NULL
 * cdef void* __nvshmemx_hostlib_finalize = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_set_attr_uniqueid_args = NULL
 * cdef void* __nvshmemx_set_attr_mpi_comm_args = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_hostlib_finalize = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":242
 * cdef void* __nvshmemx_hostlib_init_attr = NULL
 * cdef void* __nvshmemx_hostlib_finalize = NULL
 * cdef void* __nvshmemx_set_attr_uniqueid_args = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_set_attr_mpi_comm_args = NULL
 * cdef void* __nvshmemx_get_uniqueid = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_uniqueid_args = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":243
 * cdef void* __nvshmemx_hostlib_finalize = NULL
 * cdef void* __nvshmemx_set_attr_uniqueid_args = NULL
 * cdef void* __nvshmemx_set_attr_mpi_comm_args = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_get_uniqueid = NULL
 * cdef void* __nvshmemx_cumodule_init = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_set_attr_mpi_comm_args = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":244
 * cdef void* __nvshmemx_set_attr_uniqueid_args = NULL
 * cdef void* __nvshmemx_set_attr_mpi_comm_args = NULL
 * cdef void* __nvshmemx_get_uniqueid = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_cumodule_init = NULL
 * cdef void* __nvshmemx_cumodule_finalize = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_get_uniqueid = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":245
 * cdef void* __nvshmemx_set_attr_mpi_comm_args = NULL
 * cdef void* __nvshmemx_get_uniqueid = NULL
 * cdef void* __nvshmemx_cumodule_init = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_cumodule_finalize = NULL
 * cdef void* __nvshmemx_buffer_register_symmetric = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_init = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":246
 * cdef void* __nvshmemx_get_uniqueid = NULL
 * cdef void* __nvshmemx_cumodule_init = NULL
 * cdef void* __nvshmemx_cumodule_finalize = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_buffer_register_symmetric = NULL
 * cdef void* __nvshmemx_buffer_unregister_symmetric = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_cumodule_finalize = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":247
 * cdef void* __nvshmemx_cumodule_init = NULL
 * cdef void* __nvshmemx_cumodule_finalize = NULL
 * cdef void* __nvshmemx_buffer_register_symmetric = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_buffer_unregister_symmetric = NULL
 * cdef void* __nvshmemx_culibrary_init = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_register_symmetric = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":248
 * cdef void* __nvshmemx_cumodule_finalize = NULL
 * cdef void* __nvshmemx_buffer_register_symmetric = NULL
 * cdef void* __nvshmemx_buffer_unregister_symmetric = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_culibrary_init = NULL
 * cdef void* __nvshmemx_culibrary_finalize = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_buffer_unregister_symmetric = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":249
 * cdef void* __nvshmemx_buffer_register_symmetric = NULL
 * cdef void* __nvshmemx_buffer_unregister_symmetric = NULL
 * cdef void* __nvshmemx_culibrary_init = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_culibrary_finalize = NULL
 * cdef void* __nvshmemx_putmem_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_init = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":250
 * cdef void* __nvshmemx_buffer_unregister_symmetric = NULL
 * cdef void* __nvshmemx_culibrary_init = NULL
 * cdef void* __nvshmemx_culibrary_finalize = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_putmem_on_stream = NULL
 * cdef void* __nvshmemx_putmem_signal_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_culibrary_finalize = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":251
 * cdef void* __nvshmemx_culibrary_init = NULL
 * cdef void* __nvshmemx_culibrary_finalize = NULL
 * cdef void* __nvshmemx_putmem_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_putmem_signal_on_stream = NULL
 * cdef void* __nvshmemx_getmem_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":252
 * cdef void* __nvshmemx_culibrary_finalize = NULL
 * cdef void* __nvshmemx_putmem_on_stream = NULL
 * cdef void* __nvshmemx_putmem_signal_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_getmem_on_stream = NULL
 * cdef void* __nvshmemx_quiet_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_putmem_signal_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":253
 * cdef void* __nvshmemx_putmem_on_stream = NULL
 * cdef void* __nvshmemx_putmem_signal_on_stream = NULL
 * cdef void* __nvshmemx_getmem_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_quiet_on_stream = NULL
 * cdef void* __nvshmemx_signal_op_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_getmem_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":254
 * cdef void* __nvshmemx_putmem_signal_on_stream = NULL
 * cdef void* __nvshmemx_getmem_on_stream = NULL
 * cdef void* __nvshmemx_quiet_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_signal_op_on_stream = NULL
 * cdef void* __nvshmemx_signal_wait_until_on_stream = NULL
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_quiet_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":255
 * cdef void* __nvshmemx_getmem_on_stream = NULL
 * cdef void* __nvshmemx_quiet_on_stream = NULL
 * cdef void* __nvshmemx_signal_op_on_stream = NULL             # <<<<<<<<<<<<<<
 * cdef void* __nvshmemx_signal_wait_until_on_stream = NULL
 * 
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_op_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":256
 * cdef void* __nvshmemx_quiet_on_stream = NULL
 * cdef void* __nvshmemx_signal_op_on_stream = NULL
 * cdef void* __nvshmemx_signal_wait_until_on_stream = NULL             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_v_7nvshmem_8bindings_9_internal_7nvshmem___nvshmemx_signal_wait_until_on_stream = NULL;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1778
 * 
 * 
 * cdef dict func_ptrs = None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs);
  __Pyx_DECREF_SET(__pyx_v_7nvshmem_8bindings_9_internal_7nvshmem_func_ptrs, ((PyObject*)Py_None));
  __Pyx_GIVEREF(Py_None);

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1781
 * 
 * 
 * cpdef dict _inspect_function_pointers():             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is not None:
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_9_internal_7nvshmem_1_inspect_function_pointers, 0, __pyx_mstate_global->__pyx_n_u_inspect_function_pointers, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings__internal_nvshm, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_inspect_function_pointers, __pyx_t_4) < (0)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":2435
 * 
 * 
 * cpdef _inspect_function_pointer(str name):             # <<<<<<<<<<<<<<
 *     global func_ptrs
 *     if func_ptrs is None:
*/
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_9_internal_7nvshmem_3_inspect_function_pointer, 0, __pyx_mstate_global->__pyx_n_u_inspect_function_pointer, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings__internal_nvshm, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_4);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_inspect_function_pointer, __pyx_t_4) < (0)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "nvshmem/bindings/_internal/nvshmem.pyx":1
 * # Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.             # <<<<<<<<<<<<<<
 * #
 * # SPDX-License-Identifier: Apache-2.0
*/
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_4) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init nvshmem.bindings._internal.nvshmem", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init nvshmem.bindings._internal.nvshmem");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 14; } index[] = {{1},{29},{179},{1},{8},{35},{41},{37},{36},{34},{46},{36},{35},{35},{33},{42},{45},{40},{40},{43},{48},{47},{52},{48},{58},{59},{58},{60},{67},{60},{67},{60},{67},{56},{58},{54},{55},{54},{56},{63},{56},{63},{56},{63},{49},{45},{48},{44},{56},{57},{56},{58},{65},{58},{65},{58},{65},{55},{56},{55},{57},{64},{57},{64},{57},{64},{43},{47},{54},{55},{54},{56},{63},{56},{63},{56},{63},{47},{48},{42},{55},{56},{55},{57},{64},{57},{64},{57},{64},{55},{56},{55},{57},{64},{57},{64},{57},{64},{55},{56},{55},{57},{64},{57},{64},{57},{64},{54},{55},{54},{56},{63},{56},{63},{56},{63},{53},{54},{53},{55},{62},{55},{62},{55},{62},{54},{55},{54},{56},{63},{56},{63},{56},{63},{58},{59},{58},{60},{67},{60},{67},{60},{67},{37},{47},{54},{46},{55},{56},{55},{57},{64},{57},{64},{57},{64},{53},{53},{55},{56},{55},{57},{64},{57},{64},{57},{64},{50},{58},{54},{55},{54},{56},{63},{56},{63},{56},{63},{49},{48},{40},{50},{56},{57},{56},{58},{65},{58},{65},{58},{65},{56},{57},{56},{58},{65},{58},{65},{58},{65},{56},{57},{56},{58},{65},{58},{65},{58},{65},{55},{56},{55},{57},{64},{57},{64},{57},{64},{54},{38},{21},{17},{20},{12},{18},{18},{7},{8},{25},{26},{13},{5},{8},{13},{10},{15},{4},{8},{15},{17},{21},{34},{16},{14},{26},{16},{15},{15},{13},{22},{25},{20},{20},{23},{28},{27},{32},{28},{38},{39},{38},{40},{47},{40},{47},{40},{47},{36},{38},{34},{35},{34},{36},{43},{36},{43},{36},{43},{29},{25},{28},{24},{36},{37},{36},{38},{45},{38},{45},{38},{45},{35},{36},{35},{37},{44},{37},{44},{37},{44},{23},{27},{34},{35},{34},{36},{43},{36},{43},{36},{43},{27},{28},{22},{35},{36},{35},{37},{44},{37},{44},{37},{44},{35},{36},{35},{37},{44},{37},{44},{37},{44},{35},{36},{35},{37},{44},{37},{44},{37},{44},{34},{35},{34},{36},{43},{36},{43},{36},{43},{33},{34},{33},{35},{42},{35},{42},{35},{42},{34},{35},{34},{36},{43},{36},{43},{36},{43},{38},{39},{38},{40},{47},{40},{47},{40},{47},{17},{27},{34},{26},{35},{36},{35},{37},{44},{37},{44},{37},{44},{33},{33},{35},{36},{35},{37},{44},{37},{44},{37},{44},{30},{38},{34},{35},{34},{36},{43},{36},{43},{36},{43},{29},{28},{20},{30},{36},{37},{36},{38},{45},{38},{45},{38},{45},{36},{37},{36},{38},{45},{38},{45},{38},{45},{36},{37},{36},{38},{45},{38},{45},{38},{45},{35},{36},{35},{37},{44},{37},{44},{37},{44},{34},{3},{11},{12},{12},{12},{10},{8},{6},{11868},{31},{3249}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (3151 bytes) */
const char* const cstring = "(\265/\375`\272\211-b\000V\377\3024 w:@!l\201a!\354z\325cH\224\240\204;\033\253L\"\333\226r#\ni;\255,U\236\367\341iq\227\362\244\270\234\376\002;\343\257\030R\004\023\026\247\000\246\000\240\000[\364\277\317a\347\272:\327\335V\355\304\327\207\372}\325V\253\323\375j\322R\032%\245\261\257\317{\243*\253\331jp5il(k\326\374m\355,\213.)\3074\316\370\320|\351wf0cU'\327}W;?\276\2555f\300w\014\253\263\036G\316+\314\266\203Z\rLa\347\272\207u>\235Y\0055\013\261c\034\251Te\014\361K\017\362\013\263\032!\t\n\211\207\216\214\016\254\327_\316\373\030v\324?^\367\014\342\t\253e}\007\003?\262\036\372[\034\r\270\371\306\302CU5)\254\316Y\232A\3730s:\343\314F\375\201Wufu\252>\256\315\252\017\315\334\262v\256\246\363\177F\"\256\377Z\026^\335\330\333feO\356?\366;~\237\271\025\353j\342\314q-\267U\265\024\205\304Dd\004\347\217\217\357\177l\325\244\217\255F#+\233\236[\263f\247\252l3\2479\263\333\317:\3355i\254S\353\356\306\253\2324\337\310\2225\313y\337\267\255\303x\256\374\366uM\313[\306u^\377\217\022\273\314\231\337.\212\233tJ\255\324\036G\231E{U\366MB>Z\252U\333E\376\221\342\032\357{\226\211\364O\312\341\325\324\252\225\235\275\217=\353\035\360q\265\022\000`\303\007F\003~\235\3303\215C\335\217\375\273\237\346\250y\326\357\230\306\334\352\356\236\\\224]\324q\335\335~#\347\227\353\346,\327\335\335\335\335\335]swww\367}\256{\353N\245\272}\213\265\033\373\274_\275\355\3058R\031\337\273\353\214[7\336\374g\343\331\330r\315\341:k\326\272\273\245\315\272\205dT\271\214\367\273\316\366:\251\304^\347g{\255L\331\031\273\275[s\236/r\234~9c\212:\235n\377k9.\367\\{\257\314\314[\234\231\242\216K}\320X\257\177\235\001!G\007\001\261\023R\202\302\347ze\234\226\202`-!J\017\260\224\250d\003c\331QR\201\255\274(\241`*\nJ\256\357$\302l>\222\223R\320g\022?\033\257\243\365\001y\036\n0\227\367A\0013\370@v`*\337\310\013\014\365\211\264`@_\311\351\013za|*\036\307\313\207\362P\237\201\247\261\343\251\274*\342\313&\t\242\365\002J\212P\360\\I\023\260\207/$\344{\3606N^P\222D\313\003z]D\317\346eA\341\031\274`\001\323\351""\210\240i\007\251! F\221\364\020\320\324:B \210\235\222\034v\220^\220R\216\016\002b\321K101\020\304^\220\026\220\\F\003\0011\201\200\330\311HrV\004\245\201\202\331\264C5\243P2\257\253\211Bbr\277'fY\322\340\344\334\364\217\316\006\n\006\205I\250\303]\214 hDD$)H*\351\343\361\"00l\026G\222\310A}C\0001\030H\014\202\204\200\260\0104\020\\\004\002\301@ \260\010\004\202\000\341\2000b\014!\221fa9\273\263z\342V\320\tM\322\375\226\301\0232\313h\206\r\017\001\377k\356^<\025Y\0203\370\327V\3218I\224\006\001\3616.!\363{\377\326\317\364>\212h\240\370\370\327\036\004\344\333H\0025\207x\002oQP\023b2\301p_\253\"\026y@\367*\013o\000\r\022(,\225\210\n\202\003\336T\3513\023\311\001\211+\261\013\371F\025R*\224\373\267\004\026hMtA@\326\007\221sx\215\274\302*\310\r\362\317?\271\305\014is\326\210\010\n\371\025\267\221\003\246\177\223H\362pu\217\\\026\365#\363\374\005\233\344K\301\246\330&-\3235qS\005\243+\031\241\362\002ym\263\241\217n\224\242\263y\177\240*\217[\354\254\030\245\267\\:\022\030)\321S\0169P\220|\006\254D\205yi\366=A\266\242\016&\362\222\203lx\335O\314\311\211\033\222M\320\177\320&\216\321\232\221D\255\216\004\302\351\252\337\362\210\006\272\2421}/:B`\353\204\016'\025\033\322\376S\324#\232,\364\312\261\213\310\001&w\017\305|^\276N\251\324\300\243\240_\301\205\3326\200\305\237\327R\022\327\207i\3219\364k\231-\22401Y\363\232\301\244\253\347U\266\n\347N\277\320U\325,%\226\206J\320\355\243\3620\306\213Y\220_\314\001\257\214A\n\331=\034\206\321\304\036\327\214+\34480JV4\226*IKO\350\276\317E\026`\302h\371\210\367\023$S\374\244D]Jj!?\306.\346\357\253\362:r$m\225\240[G\332\376\030\274H\201\275\020\014\375j\036$\250a\1770\205Vt\274\314p\024\001p/&\007FX\354\317'\024\352+\030e\346\020\260\344\211\177\314 SSI*<\272=3\331\037\257-\216\317\233b8\237B\215\343E{\t7\242\016\202H\350<\r\265\232\013uT`\351\311\317\220\360\253B\030J,\r\225\240\333G\ru\214\027\243 \211\235\0076\010\346RS\247\365\377\021}N.\355\236\214`\357\250Q\177\272\326\323\032\255\264g\322\217v\276\222~\320\336\205\002\200""\273U\375x\010D^\215\013\332\232\0240\252\236:\327\201W4\027\312i\3620y\0232\34386I\017q\026\014|\267G\t\017\367|\352\032x\255\027\256\326\302PB\277d\261\024\333\245tc\2670^\007\367\"\350\346&\204\322\303\305\2147\025\311,\006\245o\322x<C\236\315\301b\024\024{\207\303f\271\207\313\236\274\032^\304v=R\256\365xt\252\376\353^\233\235KE\217n:\263\357\211\334O)\350 D4\301]5e\023\323W\231\343\233_\303\276\010TX$\341\351\227Z9}\"\303\376\371f\036\316\204\202`>d5Z\010.\007\314\013\376\343|H\242kY%\207z9\360\211\371SN\"S0\275\0058\275\216E`\367\253\221\233\220\315eZ\257BWEA\244\367\223\266\356\372#\374x}\010Ob\333\t\353\t\355\361~A\353\261\301\030*\335\230\306\336\032\262\001\017I\023\033J\321_\242\202\033\311\230\204F$w(\025\375\037\261\013pV|kL\327\327\3742\346\343\036\233\212{\211\235\335\031\246P\206]\330A4 \235\320\214x\027\255\210\376\2521R\207\250\301\256,\357\273\327\202@\357\\\364\035\352\211\203>h\336\277H\241}-\027\241\343\372\215\232\270\375GU\372u\260\374\032\321,\224\236\310\273j\327\336\226\206\264\342\r\200q\361N\014N\310\210\237^\373\274G\256\342^\302qTY\177\217\300\330\205\310\274\214s$\357U\273\352\377\021y\315v\236\311\254(E\023\342e\247\215\362\013&c\000\377~ezs DjXD\274\234\343\342\267\2662\307'\2266\"\233\326Q}\312\225\001\014N\013\265\353=\241\020\275\370?\223\217\254\214\014Z\271\025\010V\316[\016\240V\374\033\320\363\313\360g\311\255\374\026qX@\230Aa\320\234\267\245LX N\246)\234\261\223W\001B\007%~\r\260\360\360A%Z\366\353}\"\222\002\000k\306\016J\n\360\001\341+\246\020\366`\305B\240D\216O\355(\212'\216\007\251\265Yp\311\352\006\351J\320\231\323\027Vi\364\300\322\3408ns \313\343\372\200cn\247\251<&\214\021\250q;s\226hE6m\242\024\331\244\373\237\225\342\031LREIZ9cH\332\344\213\371\323\310\003A\353vj\306\024u\024\240\273\306\0217laX8\232\226\007\233\023\212}\210A\364\234^'\026\355\301\250\034\342s5\021+!\303H\302I\023\311F\031\034K*u\022\375\327?$\332\211\331]\231[Qp\265\227R\336e\227\236b3\243\212^Y9\020\353N\004\r\370\265\205l~\211\356R\035{!_)\277""\005\253\224\034d\206\324\240I\244\340\231J~\276\324.\317Hb\241\232r\315\006\264A'\246\342\022S\251Ij\335\301j*\323Q\n[r\274\321\007\313a\316\3035$\306!\272\327\254h\r\342\255\014A\3260\370i#fZ\340\207\022\r\231\216\372#\335\247\003'\372\246\221\264\302\223'\202\201\014\322\231\014\343\232U\014\305^{\242S\017t+\024\342T|\224=\212\262\005\334I\267L\341y\357\212yj\330\250\036\342\241(\272\004\036\272\025]_\216\3408$\226\2610dVf\020\264z`\325\022PD\n\342\211\\D9\234c\217\371\000\351\360\245!\324\257\342Yd\216\016\0062S\0261\267t\333\306\346e7\201\344\nS*\033\213\355\001G\263\313\205\231\217\347\263\332rBCsA\022\322\222vb\210\232$uqm1\2009\260\266\254`V\032{\227\226\253n8\216\351>\246\277\201pJ\234G\363%4\271V\206\262\2635\352\\.e\242\313(\330\351\370D\306\272\316\017\216<\331\216a\266\273\367\003\003^\261\342\351\341T\000\022\331X\"\352\365\317\025\370e\317\256\314T\215\257.\364\251\312\036i[\t$?4\335\357p&\216\177\211\006\017\005\273\235X\225\277S\341\276p\010\345\374\330ID\263\000\213\010TNT\231\350\021\026\344\014\213\035\240\224\203\374\225r\222\331a\331OL\375\317[\003\035H\356*\n\337E\214\036\214D\344\237\205\314\226#\317\207\214\210\353a {g\3039c+8\t9Q!g\342\345\276\360\224D\226\237\031N)pC\261\2415\000\340\014\237\002[\006j(D\241\263I;\236d\2713\367&\004fC%\203\201\"\036t\004\337\301'\271\303\356\307\335\202\224\226l-\221|\331\2674Edq\276t\311\031\355\033f\346\210\\Yn,\226\233A\324\360\212pi\232\310\315f\222R\302\222\r\223v3\240_\210\340Y\264\256\373\324c\205)\021\033\316\207y\260\217i\231\037\232\257y5\253H\355\330\344\217>\325(\262S\256\021\210!2\363P\230\323\240\317\035\201. ?$'<_4\236J\014\262\350\227\241\354{V2\374q\253\3321Xm\020\216lK\215\3646[\021\216\322\355'\220+\251\210jW\231\235\245U\350\036\004\354\243|\321\370\025\002\377\204\300G<9f\250\177m8\370\005\3701\341\327\037d\367\224\nDE\017d\004*\357\340F\r$\343*0\255I\200r\311F7q\347\224%\361\277\375bM=\313\237\024X\2070\203\275\240\013\335\301\322;\306\226!L\376G\010\r\222\000\362\000}\023l\347\352""\371`\276\035lx*[\334I\216\346\034\352o\2020\240\007\031\n9\031\364\331\321A]\343)\320\235{\234\212\001\377\222\033M\347`V\201\343\230B\243/\355\326\236\3434\";\372'\267(\353`\372N\335*\305\250\326=\254S\314<\262\276{\241Q!{G[\253\341|\270F5\032\233d4\374\333\234Ygp\003\000\216#S/8\t\251B\302\342\345\276\360\224D\206\237\031\246\3655\320\356\332\r\254o9\250\212\311\236\347\231\252\366\260\374\365U4!r8\351\272\324\214\366\3152sDN\2267\232x\216\035\372g\363\254\001\347\241p\252\"\352-@\006Jh\335{K$\204\374H\266D\342\363\255\222hD\246\3148\265\356\264Y\265k\310<\237\273\221\026\317\225Q\r\233\t*%\302\220\230\313\337\271\000\212s\373\3325\022Z\030s\224\364\211\312+6\221Fr\262\307\nRT\212Z\361\374\204\344W\2222\247\t/e]a\304\004\300\353f\002`\232\022%\301\2520v^\030\334\325\022wQ\254r\020}\002\303d\371\033\277\005\006\264b\345\026\225h\205B\027\203\354\351\030X\320r\263p\001\334\005=]\244\002\320\327\025";
    PyObject *data = __Pyx_DecompressString(cstring, 3151, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2968 bytes) */
const char* const cstring = "BZh91AY&SY\215\301#\027\000\006Z\377\377f\370\356\227\357\377\377\375\257\265\3720\277\377\377\360@@@@@@@@@@@@@\000`\023?\037$\217\251\265\330\361\326\203\300\000\003\340=\307\232\207\254\314\032`-ci`\000l\200Y\344\267E\361\237}\202J\224\320\006\364OTh\000\007\250\000\000\003@\007\250\006\203G\251\264\324\003@\365\004FL\364\252\223\323P\362\243\023\000\001\244`L\000\000\000\000M0\000\000\025\237\252\222\237\245\032\232\001\246\203 \r\r\003A\221\220\320\0002d\000\014\200\000\tO\325MJ\n\236F\243O\022=C@\032\000\000\000\000\000\000\000\000\000\"Rd\322\032ja#&\243SOL\322\021\372\241\350A\204h\000\030\004\003&&\0323*\010\222!4\322L\232O\312h\236\215\010\r\006\200\032\000\r\000\001\241\240\017P\000t\224\"\204b\204$\204b\207\273\366}\035\270w/\361\317\253\361\275\376\\g\333BeT\206\230k\226\020\306\251\016\345\261\275\321\032\200T\007V4'n)\204D6|\334\303BF@\032Z)d\220\223\224Q\036S\350\372~\263\306O\264\362\017Y\330(\214\021FI\261C\240E\034\010(\217`\242=fF\240\251\347L\014;6\016\334\257|%\313\243w\274\275\016#wE\246@Y\000!$! \301$\371\316\321\306b\\\345\"\275\".P\300\205E\t\025a\036\257\253\323\365\000\355gFD\2227\014\363\327\235o\0239X\227\275\353{^%\345b^\031(\235\357\317\246\000TQ\030\212Z m\357\314\274\031\341Jk\257Nu\256\242\234\024\271\020\246\311(B\360Kj\306\312\255@h\210\227\302\221\r\232\366\001\375d(\215\204\262A\201\004X\305\020$\253\345k\252;03\327\256\325\221\226J\233\375\321\326\3429\374Dn\007\017f\242#\352\302\020d$\271x\253/9\352\274\\\265v\252L\261D \276\267m\335d,\r\003AM\025\262bdB\020\204E!\010B\"]\256U\236\026g\022I$\222\300s\0347\227$\222,\262I&\"H\272\277b\364<\347\227\311\315O\307Za\212zZ\321R\210\355t \235\367\370\212\275M\2051\025b\257\277\332\352\002\005\224M\301\200 D\024\240:\326\025vY\364^t\242\272*\350\211EC\301\341*\324\220\023r\277\270\350\356\364wL\312\227\244\242\004\253\366\264u\002t\025\260\021\000\346\026\033b\002\031h\253,\312\373\004\343\240\004\212\355D\346\376\200\244\356\033\203\207\204\314\314C@2{\001\312\025\272\350s\240\035""\215\275\377]\375\345\374Z\376r\306\323\236\032\263\360\263\306\365X\341\324\354^\375R\265)\221\020\255\371(B\360Kj\306\312\255@h\210\227\302\221\017\265CE\251\213\265\357{\351^gQ\217\036\227\3719Y:\000\212\232X\002S'@\021SK\000Jd\350\0036L\267\275\356\351\232\321f\376m{\333\360\315{Y\255\0314rh\341\364tr\345\221BI<\340X\252%2t\001\0254\260\004\246\312\231^\367\276\241\330\300^\323\223\202o\2504t\242\217\213^y\345\245v\355\333m\004\350\002*i`\tL\235\000EM,\001)\223\240\017\342~\256\365U[\024y\216d\n$\233\316\215\201\014v\2302j\264\225UZ\234\013j\007\265\254\314\315\375\3713\303Z\315i-\25142\347\306l\331\247;'@\021SK\000Jd\350\002*i`\tL\235\000EM,\001)\223\240\010\251\245\200%2t\001\0254\260\004\246N\2002[\005\032\030IN\203'@\021SK\000Jd\350\002*i`\tL\235\000EM,\001)\223\240\010\251\245\200%2t\001\0254\260\004\246N\216\356\356\356\334\360Vxn\377/+\346\300\000$\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\342y\235\231\335\316y\3329\315\346\350\363s\256\272\321\316o7G9\274\335\r\335\321\316o7G9\274\335\034\346\363ts\233\315\321\316o7Cwtz\213\256\272\347:\321\316o7F\356\363\233\243\227\r\346\350\3477\233\243\234\336n\216sy\2727\215\267Cv\335\r\333t7m\320\353\256\256\272\r\333t7m\320\335\267Cv\335\r\333t7m\320\335\267C\256\272.\272\r\333t;z\272\272\3507\265\325\327A\273n\206\355\272\033\266\350i\266\350n\333\241\273n\206\355\272\035\275uu\320n\333\241\273n\206\355\272\033\266\350n\333\241\273n\206\355\272\035u\321u\320n\333\241\333\325\325\327A\275\256\256\272\r\333t7m\320\335\267Cs\313j'\243V%\370\2233;jE.\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000P\000\000\000\000\000\000\002d\000\000\000\000\000\000\000\000\000\000\000\000\000\nS'\004\002\321G\006 \020u6U@#\273\034\260\341]\373\367\316vN\200\"\246\226\000\224\311\320\004T\322\300\022\231:\000\212\232X\002S'@\021SK\000Jd""\350\002*i`\tL\235\000S[%\262Y\356\331\331\333MI\223;&\300\332\rMML\324\325e33\215\357{\332\326\265\357{\337%L $\201 \305\th\005\"Q@\225\006As\030\014`\2437\267\254\313\002\216\301\330\336\305\034\033`\320w9\202\226\355\327\231\005\377\277>\276\2675<\212\273Z!\034\201\221dV@\221Y\002a\264\222R\013\350j\351\363D\004L\203W\235\177gs\262\302\257\231a\336\r\227\016.(-\271\013\300W\205\233\323\025\0211\006\217F\361\005\351\360\317\017\337\352A\245\024\351\036]`\363\030QW\247\246\252\265\002\252\257o\206\004a9\241B\203\243`\363P^\225\0212Q\023\252(\211\250\013VI$\222I$\222@\000\000\000\003\273\262\363[\343\312\314\354\314\261\000f\335\232(\263R\n4,\241\342\2008\203\207\177\020\032\004\tR\214h\242\252\322\311\275\035\340SxE\247vy\271\252#D\250\320R\216\241DN\000;\263\376-U\307\306\027\273w>\236\034\351\267\244\354w\367\347~oo\000\000\031\347\253\200\000\000zY\226Ywf\350\001eg~ffUme\024$\022@d`\003[\226Ak\327\307$\225\327\226\030\240\242[\355\302\265\036PmJ\3274\025\221\021\020\327\311\310\357\277\022\361`\251 2 \262 \274\354\030\367(\3429\007\022c\016\n\"m\322\3727_E\2314\007\223\n4\037\311\304\355\303\307\332\242\214\342\202\360\035\243\240\030!\316\034Uh\243\200\340(\215\036\253\356DG\003v!\"H(\304\241\013\215\224l\201\264(\206A\202\217\017\021V<Q\001I\025p\036P\344V\343\343\233\016\0020\204\002\200\033\235\340b\352\027[\346Z\224\355\010.wC\211\226%Go\003\225\227\203\307\227\323\004\001@\000\000\026z\273\330\212++\263\304\000\000\000\000\000\000\034\220p\000\000\000\000\000\000\034\220p\000\000\000\000\000\000\027%\005p\000\000\000\000\000\000\034\220s2\317A<}\367\2073,\261L\362Vx{\325\021\2778\324t`1\340\362\345\202\032\201C\017Q\360\236\003\244t\271\233\000:0B\241QB\241Phl\346\356\016\022!\006\010Y\005\307\025\\\307=\336\\\206\316\276a\240\210r\240\261\000<\036\266\314\314Y\337W\223\310\022\000\000\000\000\000\0032\317/o\203\305\335\225\231\276{,\263#,\263%2\3132g\233\031\\\251\213\030\305\000\006\006\002\013\032\001\030\212\0202AX\034\200\007\000\3060@D\356\\l \272>!""\341mcf\232\305\021\n\016\227\322T\274\024FoT\265\006\003\013\010,U@$\343.\205\220\2772\031\350\261\005\034\340\2428\0236\262N\n5\201\006\004\210\211\013\003\0041*(\217A\252\301V\306 \014\n\2537;\373+\313\341\364\374~\237g\235\352\271\304DDD\023!\021\021\013\322\014\263.q\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\014\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\034q\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\021\034\347\017G\321\217?\307\270\312\314\334\254\315\004\003\\6\201\253\254\013\304t\001]}\010*\216\305T\002X\014\220R\0006\253\206y\233\034L\225o\007\266(\217\236BN\346*\277{P\242:\223\2349\267h\207\232V\201\252\035\034\\,5-\311\321\202<\\F\320l\362u\371:\221\n\203P\364\270\370g4\n\246s\tL\033\010Y\252\001\263H\nP0hc\003\221j\326Q\0044\320Q\023k\305D\223o\016C\220\233\201\231\233\264Anl0q\007\262d\267\000,\267\333\325\300\233\364Z\305\205\021\265\371\364\\\255,\343s\035wQ\023}Ft\365\033\367\036|\344D\005,\024\010@\341\313P\335F\343\324\200$\230%\207|(\001\306\220[:\017\000\322\252\001\214ph\241U\303\007\024F\313\t\245\313%XH\302fQ@$\022\027\240\227\240\274KZ\202\321(\324(\216.U\001\251\004d\021\200\222\020\t\021!\030De,\303;r\344DDDDDDDDDDDDE\226Vv|\031\264pf6X/\277\000\030Q\202&&+u\322\n\305kP/\245kP\000\000\000\005\326\214)\232L\314\322ffA\001\316\\\000\000\000\003\234\340\000\000\000\000s\234\000\013R-k\000\000\265\"\350\230\214:_\005\251L\021\021H\245\253[a\345\301DMCG,\303\273\203\237\003$\024u\314\207md\2202qAwi$$\210h\031*\240\037\340)\250h4\216:\364\021S^Zv\\\242\257xQ\022\340\245A\333\334\0322\3529\004Q\201\024yhH4Q\201p.:\365\210-\007\241Q\036\3643TG%uGv\366\247 \356\013r\3049\3002\036I\010`\210\n]\332tU\t\024\0218\004\026(\243\267Cm\256\254\252\201U\nj\024\271\206\263k\\j\020\216S\030}\367\304\311:H\004""\204)\347_f\221N\207\357\204>4b\244\243.\2175\351\352I\356\177A\304\262\311\247q\354D\024\321\031\370\251\034.\260\201\207\32024&\023\304\213\223\311$\246\005\004\t\344}\210\035!\036<\203\267\245\332\031+\232>TL\305:\301\252\256\304\036\301\330!\304\330:\234\225\272%\301\260\371\005\021\372\341\345\001F~\316\312u{\264\354w}uqK\265\336b\035\350\212\231\253\336pH\361X\233E\342\322\315v\303Cb\224\273yw\\\255\354\352`\307n8\255\333\2737\254vS\010\203\010\302\305\253\032\221\201)K6\03280\256.-\233\014\030\300\263\010\306\256\r\217\376.\344\212p\241!\033\202F.";
    PyObject *data = __Pyx_DecompressString(cstring, 2968, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (3784 bytes) */
const char* const cstring = "x\332\355\234\277\217\334\306\025\307\277\007(\210\343\024\211\021\007\001\204\374\240\220\037\226.\362)\262\025e\255\004\t\014\307\n\"\004\212\024#E\324\020\274]\356\035\225]rEr\224\273\253\\\272T\251\362\312-\267\274ReJ\227*\375\247dH.o\311!9\357\275\241\322I\200\366v8\357\315<\276yKr\336|\2067\356\007\321\"\234yy\342\315\026\311*\214\275Et\030?\317\216\227\341\322\273\3760\311C/?\016r\357\263\323\3748\211\275(\363f\241\226\010\323 \017\027\247^\226\247\3214\017\323B(\366\036}\376\350\303;\223;^\020\317\2744|\032N\363\314\313\324\341t\021dY\230y\311\334;T\321\"\217b/?]\205\331\201\367\327\271w\232(/\016+\013VZ\256\251\220\037k{\2620/\276x\037\004q\234\344A\036%\261\257\325\243\370\350\003o\026\245\272\223\350yXh\337\017\026Yx\360\247`6\363\265`8W\361\264\020\366\266g\343\007\213\350\250<\003]\353\315\023\025\317:\"\207A\232Fa\252E\027,A\273\320T7\223L\3552\3634\014\355\022Q<O\374\2430\367\237\207i\026%\304\031,\031}.O\375\025\321i\254%2\273\310*'N?\017\203\245?\013u\210$\247\014\311\342\024\247I<\217\216\030\302\214S(\345\030\347Q\312e\253E\224\373\037\315\330\242E\340\317B\216|\236\006q\266\320?\027\312\342\223f\370\371:\312u\037\272\001\246\016[~\276H\202\374\366\335\242\223<\021\365Tk\036\246I0\233\006Y.W\235O\223\305B\377d\345\232\313\340\304O\303\231\232\206ct\263i\220\347.\336ZF\261{\367\227\272\316\335gj\351\334\375NW\334\275\232\317\265x\032\036EY\241\227\235.\227aq\305\347\351\251X\2549=\016Ryd\226Z\362\250,\325\304\021Yj9D\243\241'\035\213J]\036\205\206\236S\267\016\321g\350\211\273U\372!#\r\322S\177\036\305\372\306}\026\262\025\2428\312I\341e2S\213\220\337\370V\236\321\366,Q\207ZR\034\303[=y\024o\025\305q\274\325s\210\344\216\246tt\353\006\344\321\334\321t\354\332!\242;\232\322\256\313\353\261<0*5y\\Tz\342\260\250\324\034\242\302Tts\217CL\230\212n\035;D\204\251(\355\270x\342Uq\364L\205\321\214\026-\036(\271-\037\007\213\271<\322J-y\240\225j\3428+\265\034\302\314\320\223\372\274R\227\007\231\241\347\324\255C\210\031z\342n\223,\327\367E\356}\256\026/ns""\276\356)%\344K\271L\317\311UFJ:M9*5yHVz\342\230\254\324\034\202\322T\224\216\323V_\036\226\246\242[\307\016\201i*:t\374\361GN\001\241\325\234\002B\353\271\004\204Vs\013\210\226\242\233{\334\002\242\245\350\326\261[@\264\024\035:\276{\307) \264\232S@h=\227\200\320jn\001\321Rts\217[@\264\024\335:v\013\210\226\242C\307\023\247x\230\270\205\303\304)\032&\216\3010\031\027\013\023\307P\230\214\213\204\211c L\306\305\201S\030\270E\201S\0208\306\300\270\020p\214\200q\001\3408\376#\206\177\221\304G\362\361/\265\344\001P\252\211#\240\324r\010\001C\317\3113\016A`\3509u\353\020\006\206\236K\267\356\2610\"\036\334cbL\\\274\201\330\030\023\037o F\306\304\311\310XYN\311\325\311\023\177\245D)\225\255x\026\035\351\2514[\353\231\212B~\320dn\313\037\231\343\372G\346\266\000\222\271\256\200d#\227@2\3275\220l\344\"H\346\272\n\222\215\\\006\311\302*\017\343/W\221?M\226K?H\2172\256R\235[d)\035'\251\303sV\245\346\020z\245\236<\364J5\227\3203\024\305\003Q\351;\204\236\241\350\326\261K\350\031\212\342\216\267\027\272\225T\343?AT\244\265\363H\020E\321\231\303\212Y\251\345\020z\205\232<\362\n-\227\300k\353\311G\341\314i\235\314\320s\352\326%\350\332z\342nO\343\251\010~\271\004\207\230\313(\245<c\021\267\002}\nk\270\226(\307\014\267rMq+\307\034\267rNr\253\261Yn\345\234\346Vc\363\334\3129\321\255\306f\272\225c\252[\271\346\272\225c\262[9g\273\325\330t\267r\316w\253\261\to\345\234\361VcS\336\3121\347\255\\\223\336\3121\353\255\234\323\336jl\336[9'\276\325\330\314\267rN}\253\261\271o\345\226\374V\216\331o\345\226\376V\256\371o52\001\256\\3\340jd\n\\\271\346\300\325\310$\370\3630\236%i\023\003/\271\360\226\326V\370\326a\024\317\242\370(\273Ud^\303T?\223\337\332V\035\254NO\356o\333~\230\344\367\013\255\317\3234Iu\341\013\265Z\351\031CX\035\360\375G\247'\372\377\237#\035\017\017\303\223\374\037\341\334\367\267\004~\250\277\025\017GQr0M\322D\345Q\254Oh\241?u\217\005\357<\r\017\203\351\277}\177\226L}\375\2578\035\375'\212\263U\021^\365\331\371\253\2444p\260\"\363\243\314\277\354!\312\303e\346\353\250""\321}\350\177\3130\017\312}\002e\241\302\363\212oi\342\207q\236F\241\256\210\203eX\177\352\277\315}\000\273b\rP\373=\233\000j\277\325.=\270t\351A-\355\267i\377]\271 \373w%\223\342\337\325,\r\275\022m\337\025K\202}W\\\345\rK\233\204\275qtG\323\033\025F\363;J\3368V\023\361\275\207\267\364\273Q\327$\335/\253\006\250\366\236\372\336\272aZ\275O\252\347\322\327'\326\275\322\365I\365]\330\354r\235_u\257x\317e\313.\307j\266\357\242d\227\2636;D~we\372(\357\206\324@J\323\224\260\217\334@\242\322\224 F\214J?vD\355#E%\025MQb\204\250TaS\264CG\367V\226x\226?L=\367\325\031:\2034sW\306>\202\203\224rW\206\030E\232>\356\021\266\217$M\025w\205\211\321\244i\341\206\360\020\031\334\021\261\373x\210\370\355\210\020\036&I\336\256\254\335\277$\241\333\221%\274K\222\267\r\331fz\250}\270\265\374\323\250\032\240gM\t\373`\014P\261\246\0041\024\024\355\332\021\265\017\004E\261\232\242\3040PtjS\324 Q{\252.\251\323F]\2030m\035%\037\t\006\323j\035\031\353\030\r'\313,B6?\014\247\300,Bt\203\304@\2219\254\266lo\272\252#B:\2677\t\325\021\241\235kO-ueI\347\332\023F\035Y\332\271\3664P[\2667\343\323\021!\235\333\233\307\351\210\320\316\265gg\272\262\244s\3559\227\216,\355\\{&\245-;!};\241];!=;a8v\302\367\353\204\341\326\t\337\253\023\206S'|\237\222.\245=J:\224\341O\276;\031\336\344;\223\341K\246+\007\350%S\302\356\314\001\032\311\224 \334I\321E\035Q\273C)Z\310\024%\\J\321?\206(\317\257L\337\362\374\313\365\261\320\317\\_\013\375\315\365\271\300\357\025a\3258`\322T\335*\223\234jH\030\224T\243&\243\323\007\031#\177\220\321\t\204\214\223A\310\004)\204\214\223C\310\004I\204\214\223E\310\004i\204~\272\250O\240E\0225\005\006\250\241\216\0101<\0034PG\204\032\036\212\362\351\312\022\303C\321;\035Yjx(*\247)\333%p\272\265}\264MK\352\214\310\336\014Q4\246\210}t\206\350\230a\031\373\231\237\2219\033\212p1E\251\201!\310\225\246h\207RiTv\210\024\263\316H\260\365\220&\215Z\305\230\351*\316TW1\346\272\2125\331U\222\331\256bMw\225d\276\253X\023^%\231\361*\306\224Wq\346\274\2121\351U\254Y\257\222L{\025k\336\253$\023_\305""\232\371*\311\324W1\346\276\2123\371U\214\331\257bM\177\225d\376\253X\023`%\231\001+\326\024XI\346\300\212\236\004+\306,X\321\323`\305\231\007+\301DXqf\302J0\025V\234\271\260\022L\206\007V\341W\372\026\355\257\322p\025\244\345B\363\352\364\304\237\006\372\231J\377{\246\202E\275\000]<NU\337\365\267Y8\017\324B\337\n\362P\217\201\377<X\2500\323\266x\327?\373\347v\001\347\006\260=P-\316\324\345\342c\232\304Y\356\335\364\232\337w\347T\337{\352\272\375\233\306\246\375\362\251N\037\277Q\265\330^\306m\210WG\266\013\311eE1I(\273-?z\365\265\235}\207oz\217\376\345\377\355\357\017\377R~\024m\265\312\227\226\226\267b-=U\263\340\213j\010\312&\313\177o\240\335\322\366v\343\275\255\026\003\177\271nZ4\320>0\302^i\313\\\213\213g\377B\257\374;\302>{;lk\332\375\365\312T\313dE\353\333o#\254\246\333\342Z^\016D\241X}\031a\023\325\022\327\242b\245\245\320+\377\216\260\307\336\016\327\232\342\340~\363\332\343d\213\255\025\201%\315\217\376\253\326e\353\325\265\313\274\330\331\232vj\360\377m\207\305\031\372\231\267vj\371u\334\360\220\255\t\206I?'\326m\225_\307YF\266&\260L?[\325m\225_\307YF\266&\260lr\331\324d\264]D[\\\253\212\320\255Cx\214E\366v\270\326\224\351\224R\261\3742\302\036\252%\266E\321Q\034\316.\357\240\315\342\030\353\004\255\362--\344w\232\343\354\243\332\342Z\245\032\227\035\365\006\256b\274\366$\326\325\227\036\365\006\256d\274\366$\326\325\227\037\365\006\256f\274\366$\326Mv\215M\306\333F\265FX\266M\2765&K\373\246@O\235\212\267\277\276\346}\271\177V\365<\211f\344\324\253\026\252\214n\034,\316\277\254\274nz\245:\332x|\322mu\217\031U\346\023D\337\321\301i_-\256\3327\270]X\230~\356\212\357$w\217k}:\225?\332\245\233^\371\2273\300,=W\273\352\346\367;~k\366s\243\026\252\207\264U\334\215t}\274\035\001\333iu\263\356\322\376\262\306\206\201b\010\003\205\005\003\305\000\006\nj!\010=\251m\230\030=z\360s\364\340\347`-\352\202\265Z\013\346j-d\253\265`\256\326B\266Z\013\346j-d\253\265\340\255\177\203\005\333\203E\321\203I\321CF\321\203I\321CF\321\203I\321CF\321\203\267/\001$\"\017\222\177""\007\203\177\007\237\177\007\203\177\007\237\177\007\203\177\007\237\177\007\275c\000\226M%\260\256\271\201\201\274\203A\263\203E\263CB\263\203E\263CB\263\203E\263CB\263\203\263\037\0004\362\016\032g\007\007g\207\000g\007\007g\207\000g\007\007g\207\000g\007c#\000H\200\035$\235\016\006\235\016>\235\016\006\235\016>\235\016\006\235\016>\235\016\232\347\007\005B\202\302\034Ac\216`c\216\2401G\2601G\320\230#\330\230#H \024\226}\204\350\337z\210\341\355\205\2407\003\200\006\375\301\001\375!\000\375\301\001\375!\000\375\301\001\375!\000\375\301\330\"\001z3\000h\320\037\034\320\037\002\320\037\034\320\037\002\320\037\034\320\037\002\320\037\214-\022\2407\003\200\006\375\301\001\375!\000\375\301\001\375!\000\375\301\001\375!\000\375\301\330\"\001r3\000H\322\037\014\322\037|\322\037\014\322\037|\322\037\014\322\037|\322\037\364\336\010\220\004;H<\035\014<\035|<\035\014<\035|<\035\014<\035|<\0354\320\017\032\241\005\215\307\202\203\307B\200\307\202\203\307B\200\307\202\203\307B\200\307\202\001\026\203\306\307A\243\341\340\240\341\020\240\341\340\240\341\020\240\341\340\240\341\020\240\341`@\365 \331b\220\3400\030\3400\370\3400\030\3400\370\3400\030\3400\370\3400h\324\032\014\340\027\014\226\027,\226\027\022\226\027,\226\027\022\226\027,\226\027\022\226\027\034\032\032\014\340\027\014\226\027,\226\027\022\226\027,\226\027\022\226\027,\226\027\022\226\027\034\032\032\014\340\027\014\226\027,\226\027\022\226\027,\226\027\022\226\027,\226\027\022\226\027\034\032\0324\360\013\032\346\005\007\346\205\000\346\005\007\346\205\000\346\005\007\346\205\000\346\005\003\203\206e7\007\372\337+\001\313\333\014@m\270\002\353\225:h\277 \n\355\027D\241\367\225\t ^\302\204\301-x\260mw\301\340[\255@\262\322\350}y\025\206^^\005\326\206'X7M\241\365J.\014\277\002\004\303\333\031AngD\337\373\3030\374\002\014\230\233)a\274\021\014\355\027\226\301x\321\030\030/kB\363\265a_\356\275\276\362\255\327W\276\375\345\331\213\037\274\370\364\365;\357\177\375\376\301:x}\345\273_}\362\342\323\027\217\033\325?z""\271\367\372\235\357}\365\354\233+W_^{}\345\207/\036\177ST\275\363\325\336\327\337\361\316\337]_\333\025\177q\376`\035\354\212\277<\177\262~\266+\376\364\345\323\365\336`\361\306\372\352\346\366\256x{s\365\242Q\374\331\313\263\365{\203E\243\251\237\274|r\336\350\367\307/\037\234\007\2036\377\374\374\336\372\361`\361\372\372\335MCx\177\275\277\371tW\374\365\372\336\346\361`\321h\352W\347O7{\203\316\371\355\346\311\305\263\301\323\377h\263\177\321\350\367\343\315\275\213\307\\a\243\370\233\315\273\027\327\006uEf\210\212\206\315\366\242\321\257\341\330[\353\263\213\306\350\177\270~\262i\330|s\375`\323\030\356\273\233\247\257\366\270V\335\331<\270\010\034O\301\356+\221\031\242\242a\263\275h\364\3736\352,g4\306\317F\3218A{q\314p\333O\377w\233\263W\357\r\026\355\035\331\215|\353\234\267\316a\236\302\275\213\375W\215\216~\177q\357\325c\307\342\037.\036\274\n\270E\243_\221\031\237\\\\}u{P\367\217\027O\377\2737X\264wd7\362\255s\336:\207y\n\3063\222\361\244j\334\177\215\242\361h\372\301\371\331\346\275\301\226\215\037\273\361\3536\036\230\215\2071\303*\343\206k\324\0323\002\3431o\333\357\367_\354\025\223\226\333\377\003#/\004\260";
    PyObject *data = __Pyx_DecompressString(cstring, 3784, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (35514 bytes) */
const char* const bytes = ")Failed to dlopen libnvshmem (Note that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.?add_notefunction nvshmem_align is not foundfunction nvshmem_barrier_all is not foundfunction nvshmem_barrier is not foundfunction nvshmem_calloc is not foundfunction nvshmem_free is not foundfunction nvshmem_info_get_version is not foundfunction nvshmem_malloc is not foundfunction nvshmem_my_pe is not foundfunction nvshmem_n_pes is not foundfunction nvshmem_ptr is not foundfunction nvshmem_team_destroy is not foundfunction nvshmem_team_get_config is not foundfunction nvshmem_team_my_pe is not foundfunction nvshmem_team_n_pes is not foundfunction nvshmem_team_split_2d is not foundfunction nvshmem_team_split_strided is not foundfunction nvshmem_team_translate_pe is not foundfunction nvshmemx_barrier_all_on_stream is not foundfunction nvshmemx_barrier_on_stream is not foundfunction nvshmemx_bfloat16_alltoall_on_stream is not foundfunction nvshmemx_bfloat16_broadcast_on_stream is not foundfunction nvshmemx_bfloat16_fcollect_on_stream is not foundfunction nvshmemx_bfloat16_max_reduce_on_stream is not foundfunction nvshmemx_bfloat16_max_reducescatter_on_stream is not foundfunction nvshmemx_bfloat16_min_reduce_on_stream is not foundfunction nvshmemx_bfloat16_min_reducescatter_on_stream is not foundfunction nvshmemx_bfloat16_sum_reduce_on_stream is not foundfunction nvshmemx_bfloat16_sum_reducescatter_on_stream is not foundfunction nvshmemx_buffer_register_symmetric is not foundfunction nvshmemx_buffer_unregister_symmetric is not foundfunction nvshmemx_char_alltoall_on_stream is not foundfunction nvshmemx_char_broadcast_on_stream is not foundfunction nvshmemx_char_fcollect_on_stream is not foundfunction nvshmemx_char_max_reduce_on_stream is not foundfunction nvshmemx_char_max_reducescatter_on_stream is not foundfunction nvshmemx_char_min_reduce_on_stream is not foundf""unction nvshmemx_char_min_reducescatter_on_stream is not foundfunction nvshmemx_char_sum_reduce_on_stream is not foundfunction nvshmemx_char_sum_reducescatter_on_stream is not foundfunction nvshmemx_culibrary_finalize is not foundfunction nvshmemx_culibrary_init is not foundfunction nvshmemx_cumodule_finalize is not foundfunction nvshmemx_cumodule_init is not foundfunction nvshmemx_double_alltoall_on_stream is not foundfunction nvshmemx_double_broadcast_on_stream is not foundfunction nvshmemx_double_fcollect_on_stream is not foundfunction nvshmemx_double_max_reduce_on_stream is not foundfunction nvshmemx_double_max_reducescatter_on_stream is not foundfunction nvshmemx_double_min_reduce_on_stream is not foundfunction nvshmemx_double_min_reducescatter_on_stream is not foundfunction nvshmemx_double_sum_reduce_on_stream is not foundfunction nvshmemx_double_sum_reducescatter_on_stream is not foundfunction nvshmemx_float_alltoall_on_stream is not foundfunction nvshmemx_float_broadcast_on_stream is not foundfunction nvshmemx_float_fcollect_on_stream is not foundfunction nvshmemx_float_max_reduce_on_stream is not foundfunction nvshmemx_float_max_reducescatter_on_stream is not foundfunction nvshmemx_float_min_reduce_on_stream is not foundfunction nvshmemx_float_min_reducescatter_on_stream is not foundfunction nvshmemx_float_sum_reduce_on_stream is not foundfunction nvshmemx_float_sum_reducescatter_on_stream is not foundfunction nvshmemx_get_uniqueid is not foundfunction nvshmemx_getmem_on_stream is not foundfunction nvshmemx_half_alltoall_on_stream is not foundfunction nvshmemx_half_broadcast_on_stream is not foundfunction nvshmemx_half_fcollect_on_stream is not foundfunction nvshmemx_half_max_reduce_on_stream is not foundfunction nvshmemx_half_max_reducescatter_on_stream is not foundfunction nvshmemx_half_min_reduce_on_stream is not foundfunction nvshmemx_half_min_reducescatter_on_stream is not foundfunction nvshmemx_half_sum_reduce_on_stream is not foundfunction nvshmemx_h""alf_sum_reducescatter_on_stream is not foundfunction nvshmemx_hostlib_finalize is not foundfunction nvshmemx_hostlib_init_attr is not foundfunction nvshmemx_init_status is not foundfunction nvshmemx_int16_alltoall_on_stream is not foundfunction nvshmemx_int16_broadcast_on_stream is not foundfunction nvshmemx_int16_fcollect_on_stream is not foundfunction nvshmemx_int16_max_reduce_on_stream is not foundfunction nvshmemx_int16_max_reducescatter_on_stream is not foundfunction nvshmemx_int16_min_reduce_on_stream is not foundfunction nvshmemx_int16_min_reducescatter_on_stream is not foundfunction nvshmemx_int16_sum_reduce_on_stream is not foundfunction nvshmemx_int16_sum_reducescatter_on_stream is not foundfunction nvshmemx_int32_alltoall_on_stream is not foundfunction nvshmemx_int32_broadcast_on_stream is not foundfunction nvshmemx_int32_fcollect_on_stream is not foundfunction nvshmemx_int32_max_reduce_on_stream is not foundfunction nvshmemx_int32_max_reducescatter_on_stream is not foundfunction nvshmemx_int32_min_reduce_on_stream is not foundfunction nvshmemx_int32_min_reducescatter_on_stream is not foundfunction nvshmemx_int32_sum_reduce_on_stream is not foundfunction nvshmemx_int32_sum_reducescatter_on_stream is not foundfunction nvshmemx_int64_alltoall_on_stream is not foundfunction nvshmemx_int64_broadcast_on_stream is not foundfunction nvshmemx_int64_fcollect_on_stream is not foundfunction nvshmemx_int64_max_reduce_on_stream is not foundfunction nvshmemx_int64_max_reducescatter_on_stream is not foundfunction nvshmemx_int64_min_reduce_on_stream is not foundfunction nvshmemx_int64_min_reducescatter_on_stream is not foundfunction nvshmemx_int64_sum_reduce_on_stream is not foundfunction nvshmemx_int64_sum_reducescatter_on_stream is not foundfunction nvshmemx_int8_alltoall_on_stream is not foundfunction nvshmemx_int8_broadcast_on_stream is not foundfunction nvshmemx_int8_fcollect_on_stream is not foundfunction nvshmemx_int8_max_reduce_on_stream is not foundfunction nvsh""memx_int8_max_reducescatter_on_stream is not foundfunction nvshmemx_int8_min_reduce_on_stream is not foundfunction nvshmemx_int8_min_reducescatter_on_stream is not foundfunction nvshmemx_int8_sum_reduce_on_stream is not foundfunction nvshmemx_int8_sum_reducescatter_on_stream is not foundfunction nvshmemx_int_alltoall_on_stream is not foundfunction nvshmemx_int_broadcast_on_stream is not foundfunction nvshmemx_int_fcollect_on_stream is not foundfunction nvshmemx_int_max_reduce_on_stream is not foundfunction nvshmemx_int_max_reducescatter_on_stream is not foundfunction nvshmemx_int_min_reduce_on_stream is not foundfunction nvshmemx_int_min_reducescatter_on_stream is not foundfunction nvshmemx_int_sum_reduce_on_stream is not foundfunction nvshmemx_int_sum_reducescatter_on_stream is not foundfunction nvshmemx_long_alltoall_on_stream is not foundfunction nvshmemx_long_broadcast_on_stream is not foundfunction nvshmemx_long_fcollect_on_stream is not foundfunction nvshmemx_long_max_reduce_on_stream is not foundfunction nvshmemx_long_max_reducescatter_on_stream is not foundfunction nvshmemx_long_min_reduce_on_stream is not foundfunction nvshmemx_long_min_reducescatter_on_stream is not foundfunction nvshmemx_long_sum_reduce_on_stream is not foundfunction nvshmemx_long_sum_reducescatter_on_stream is not foundfunction nvshmemx_longlong_alltoall_on_stream is not foundfunction nvshmemx_longlong_broadcast_on_stream is not foundfunction nvshmemx_longlong_fcollect_on_stream is not foundfunction nvshmemx_longlong_max_reduce_on_stream is not foundfunction nvshmemx_longlong_max_reducescatter_on_stream is not foundfunction nvshmemx_longlong_min_reduce_on_stream is not foundfunction nvshmemx_longlong_min_reducescatter_on_stream is not foundfunction nvshmemx_longlong_sum_reduce_on_stream is not foundfunction nvshmemx_longlong_sum_reducescatter_on_stream is not foundfunction nvshmemx_mc_ptr is not foundfunction nvshmemx_putmem_on_stream is not foundfunction nvshmemx_putmem_signal_on_stream"" is not foundfunction nvshmemx_quiet_on_stream is not foundfunction nvshmemx_schar_alltoall_on_stream is not foundfunction nvshmemx_schar_broadcast_on_stream is not foundfunction nvshmemx_schar_fcollect_on_stream is not foundfunction nvshmemx_schar_max_reduce_on_stream is not foundfunction nvshmemx_schar_max_reducescatter_on_stream is not foundfunction nvshmemx_schar_min_reduce_on_stream is not foundfunction nvshmemx_schar_min_reducescatter_on_stream is not foundfunction nvshmemx_schar_sum_reduce_on_stream is not foundfunction nvshmemx_schar_sum_reducescatter_on_stream is not foundfunction nvshmemx_set_attr_mpi_comm_args is not foundfunction nvshmemx_set_attr_uniqueid_args is not foundfunction nvshmemx_short_alltoall_on_stream is not foundfunction nvshmemx_short_broadcast_on_stream is not foundfunction nvshmemx_short_fcollect_on_stream is not foundfunction nvshmemx_short_max_reduce_on_stream is not foundfunction nvshmemx_short_max_reducescatter_on_stream is not foundfunction nvshmemx_short_min_reduce_on_stream is not foundfunction nvshmemx_short_min_reducescatter_on_stream is not foundfunction nvshmemx_short_sum_reduce_on_stream is not foundfunction nvshmemx_short_sum_reducescatter_on_stream is not foundfunction nvshmemx_signal_op_on_stream is not foundfunction nvshmemx_signal_wait_until_on_stream is not foundfunction nvshmemx_size_alltoall_on_stream is not foundfunction nvshmemx_size_broadcast_on_stream is not foundfunction nvshmemx_size_fcollect_on_stream is not foundfunction nvshmemx_size_max_reduce_on_stream is not foundfunction nvshmemx_size_max_reducescatter_on_stream is not foundfunction nvshmemx_size_min_reduce_on_stream is not foundfunction nvshmemx_size_min_reducescatter_on_stream is not foundfunction nvshmemx_size_sum_reduce_on_stream is not foundfunction nvshmemx_size_sum_reducescatter_on_stream is not foundfunction nvshmemx_sync_all_on_stream is not foundfunction nvshmemx_team_get_uniqueid is not foundfunction nvshmemx_team_init is not foundfunction nvs""hmemx_team_sync_on_stream is not foundfunction nvshmemx_uint16_alltoall_on_stream is not foundfunction nvshmemx_uint16_broadcast_on_stream is not foundfunction nvshmemx_uint16_fcollect_on_stream is not foundfunction nvshmemx_uint16_max_reduce_on_stream is not foundfunction nvshmemx_uint16_max_reducescatter_on_stream is not foundfunction nvshmemx_uint16_min_reduce_on_stream is not foundfunction nvshmemx_uint16_min_reducescatter_on_stream is not foundfunction nvshmemx_uint16_sum_reduce_on_stream is not foundfunction nvshmemx_uint16_sum_reducescatter_on_stream is not foundfunction nvshmemx_uint32_alltoall_on_stream is not foundfunction nvshmemx_uint32_broadcast_on_stream is not foundfunction nvshmemx_uint32_fcollect_on_stream is not foundfunction nvshmemx_uint32_max_reduce_on_stream is not foundfunction nvshmemx_uint32_max_reducescatter_on_stream is not foundfunction nvshmemx_uint32_min_reduce_on_stream is not foundfunction nvshmemx_uint32_min_reducescatter_on_stream is not foundfunction nvshmemx_uint32_sum_reduce_on_stream is not foundfunction nvshmemx_uint32_sum_reducescatter_on_stream is not foundfunction nvshmemx_uint64_alltoall_on_stream is not foundfunction nvshmemx_uint64_broadcast_on_stream is not foundfunction nvshmemx_uint64_fcollect_on_stream is not foundfunction nvshmemx_uint64_max_reduce_on_stream is not foundfunction nvshmemx_uint64_max_reducescatter_on_stream is not foundfunction nvshmemx_uint64_min_reduce_on_stream is not foundfunction nvshmemx_uint64_min_reducescatter_on_stream is not foundfunction nvshmemx_uint64_sum_reduce_on_stream is not foundfunction nvshmemx_uint64_sum_reducescatter_on_stream is not foundfunction nvshmemx_uint8_alltoall_on_stream is not foundfunction nvshmemx_uint8_broadcast_on_stream is not foundfunction nvshmemx_uint8_fcollect_on_stream is not foundfunction nvshmemx_uint8_max_reduce_on_stream is not foundfunction nvshmemx_uint8_max_reducescatter_on_stream is not foundfunction nvshmemx_uint8_min_reduce_on_stream is not foundfunc""tion nvshmemx_uint8_min_reducescatter_on_stream is not foundfunction nvshmemx_uint8_sum_reduce_on_stream is not foundfunction nvshmemx_uint8_sum_reducescatter_on_stream is not foundfunction nvshmemx_vendor_get_version_info is not foundnvshmem/bindings/_internal/nvshmem.pyxFunctionNotFoundErrorNotSupportedError__Pyx_PyDict_NextRef__annotate__asyncio.coroutinescline_in_traceback__doc____func___inspect_function_pointer_inspect_function_pointers_is_coroutineitems__main____metaclass____module____mro_entries__name__name____nvshmem_align__nvshmem_barrier__nvshmem_barrier_allnvshmem.bindings._internal.nvshmem__nvshmem_calloc__nvshmem_free__nvshmem_info_get_version__nvshmem_malloc__nvshmem_my_pe__nvshmem_n_pes__nvshmem_ptr__nvshmem_team_destroy__nvshmem_team_get_config__nvshmem_team_my_pe__nvshmem_team_n_pes__nvshmem_team_split_2d__nvshmem_team_split_strided__nvshmem_team_translate_pe__nvshmemx_barrier_all_on_stream__nvshmemx_barrier_on_stream__nvshmemx_bfloat16_alltoall_on_stream__nvshmemx_bfloat16_broadcast_on_stream__nvshmemx_bfloat16_fcollect_on_stream__nvshmemx_bfloat16_max_reduce_on_stream__nvshmemx_bfloat16_max_reducescatter_on_stream__nvshmemx_bfloat16_min_reduce_on_stream__nvshmemx_bfloat16_min_reducescatter_on_stream__nvshmemx_bfloat16_sum_reduce_on_stream__nvshmemx_bfloat16_sum_reducescatter_on_stream__nvshmemx_buffer_register_symmetric__nvshmemx_buffer_unregister_symmetric__nvshmemx_char_alltoall_on_stream__nvshmemx_char_broadcast_on_stream__nvshmemx_char_fcollect_on_stream__nvshmemx_char_max_reduce_on_stream__nvshmemx_char_max_reducescatter_on_stream__nvshmemx_char_min_reduce_on_stream__nvshmemx_char_min_reducescatter_on_stream__nvshmemx_char_sum_reduce_on_stream__nvshmemx_char_sum_reducescatter_on_stream__nvshmemx_culibrary_finalize__nvshmemx_culibrary_init__nvshmemx_cumodule_finalize__nvshmemx_cumodule_init__nvshmemx_double_alltoall_on_stream__nvshmemx_double_broadcast_on_stream__nvshmemx_double_fcollect_on_stream__nvshmemx_double_max_reduce_on_stream__nvshmem""x_double_max_reducescatter_on_stream__nvshmemx_double_min_reduce_on_stream__nvshmemx_double_min_reducescatter_on_stream__nvshmemx_double_sum_reduce_on_stream__nvshmemx_double_sum_reducescatter_on_stream__nvshmemx_float_alltoall_on_stream__nvshmemx_float_broadcast_on_stream__nvshmemx_float_fcollect_on_stream__nvshmemx_float_max_reduce_on_stream__nvshmemx_float_max_reducescatter_on_stream__nvshmemx_float_min_reduce_on_stream__nvshmemx_float_min_reducescatter_on_stream__nvshmemx_float_sum_reduce_on_stream__nvshmemx_float_sum_reducescatter_on_stream__nvshmemx_get_uniqueid__nvshmemx_getmem_on_stream__nvshmemx_half_alltoall_on_stream__nvshmemx_half_broadcast_on_stream__nvshmemx_half_fcollect_on_stream__nvshmemx_half_max_reduce_on_stream__nvshmemx_half_max_reducescatter_on_stream__nvshmemx_half_min_reduce_on_stream__nvshmemx_half_min_reducescatter_on_stream__nvshmemx_half_sum_reduce_on_stream__nvshmemx_half_sum_reducescatter_on_stream__nvshmemx_hostlib_finalize__nvshmemx_hostlib_init_attr__nvshmemx_init_status__nvshmemx_int16_alltoall_on_stream__nvshmemx_int16_broadcast_on_stream__nvshmemx_int16_fcollect_on_stream__nvshmemx_int16_max_reduce_on_stream__nvshmemx_int16_max_reducescatter_on_stream__nvshmemx_int16_min_reduce_on_stream__nvshmemx_int16_min_reducescatter_on_stream__nvshmemx_int16_sum_reduce_on_stream__nvshmemx_int16_sum_reducescatter_on_stream__nvshmemx_int32_alltoall_on_stream__nvshmemx_int32_broadcast_on_stream__nvshmemx_int32_fcollect_on_stream__nvshmemx_int32_max_reduce_on_stream__nvshmemx_int32_max_reducescatter_on_stream__nvshmemx_int32_min_reduce_on_stream__nvshmemx_int32_min_reducescatter_on_stream__nvshmemx_int32_sum_reduce_on_stream__nvshmemx_int32_sum_reducescatter_on_stream__nvshmemx_int64_alltoall_on_stream__nvshmemx_int64_broadcast_on_stream__nvshmemx_int64_fcollect_on_stream__nvshmemx_int64_max_reduce_on_stream__nvshmemx_int64_max_reducescatter_on_stream__nvshmemx_int64_min_reduce_on_stream__nvshmemx_int64_min_reducescatter_on_stream__nvshmemx_int64""_sum_reduce_on_stream__nvshmemx_int64_sum_reducescatter_on_stream__nvshmemx_int8_alltoall_on_stream__nvshmemx_int8_broadcast_on_stream__nvshmemx_int8_fcollect_on_stream__nvshmemx_int8_max_reduce_on_stream__nvshmemx_int8_max_reducescatter_on_stream__nvshmemx_int8_min_reduce_on_stream__nvshmemx_int8_min_reducescatter_on_stream__nvshmemx_int8_sum_reduce_on_stream__nvshmemx_int8_sum_reducescatter_on_stream__nvshmemx_int_alltoall_on_stream__nvshmemx_int_broadcast_on_stream__nvshmemx_int_fcollect_on_stream__nvshmemx_int_max_reduce_on_stream__nvshmemx_int_max_reducescatter_on_stream__nvshmemx_int_min_reduce_on_stream__nvshmemx_int_min_reducescatter_on_stream__nvshmemx_int_sum_reduce_on_stream__nvshmemx_int_sum_reducescatter_on_stream__nvshmemx_long_alltoall_on_stream__nvshmemx_long_broadcast_on_stream__nvshmemx_long_fcollect_on_stream__nvshmemx_long_max_reduce_on_stream__nvshmemx_long_max_reducescatter_on_stream__nvshmemx_long_min_reduce_on_stream__nvshmemx_long_min_reducescatter_on_stream__nvshmemx_long_sum_reduce_on_stream__nvshmemx_long_sum_reducescatter_on_stream__nvshmemx_longlong_alltoall_on_stream__nvshmemx_longlong_broadcast_on_stream__nvshmemx_longlong_fcollect_on_stream__nvshmemx_longlong_max_reduce_on_stream__nvshmemx_longlong_max_reducescatter_on_stream__nvshmemx_longlong_min_reduce_on_stream__nvshmemx_longlong_min_reducescatter_on_stream__nvshmemx_longlong_sum_reduce_on_stream__nvshmemx_longlong_sum_reducescatter_on_stream__nvshmemx_mc_ptr__nvshmemx_putmem_on_stream__nvshmemx_putmem_signal_on_stream__nvshmemx_quiet_on_stream__nvshmemx_schar_alltoall_on_stream__nvshmemx_schar_broadcast_on_stream__nvshmemx_schar_fcollect_on_stream__nvshmemx_schar_max_reduce_on_stream__nvshmemx_schar_max_reducescatter_on_stream__nvshmemx_schar_min_reduce_on_stream__nvshmemx_schar_min_reducescatter_on_stream__nvshmemx_schar_sum_reduce_on_stream__nvshmemx_schar_sum_reducescatter_on_stream__nvshmemx_set_attr_mpi_comm_args__nvshmemx_set_attr_uniqueid_args__nvshmemx_short_alltoall_on_""stream__nvshmemx_short_broadcast_on_stream__nvshmemx_short_fcollect_on_stream__nvshmemx_short_max_reduce_on_stream__nvshmemx_short_max_reducescatter_on_stream__nvshmemx_short_min_reduce_on_stream__nvshmemx_short_min_reducescatter_on_stream__nvshmemx_short_sum_reduce_on_stream__nvshmemx_short_sum_reducescatter_on_stream__nvshmemx_signal_op_on_stream__nvshmemx_signal_wait_until_on_stream__nvshmemx_size_alltoall_on_stream__nvshmemx_size_broadcast_on_stream__nvshmemx_size_fcollect_on_stream__nvshmemx_size_max_reduce_on_stream__nvshmemx_size_max_reducescatter_on_stream__nvshmemx_size_min_reduce_on_stream__nvshmemx_size_min_reducescatter_on_stream__nvshmemx_size_sum_reduce_on_stream__nvshmemx_size_sum_reducescatter_on_stream__nvshmemx_sync_all_on_stream__nvshmemx_team_get_uniqueid__nvshmemx_team_init__nvshmemx_team_sync_on_stream__nvshmemx_uint16_alltoall_on_stream__nvshmemx_uint16_broadcast_on_stream__nvshmemx_uint16_fcollect_on_stream__nvshmemx_uint16_max_reduce_on_stream__nvshmemx_uint16_max_reducescatter_on_stream__nvshmemx_uint16_min_reduce_on_stream__nvshmemx_uint16_min_reducescatter_on_stream__nvshmemx_uint16_sum_reduce_on_stream__nvshmemx_uint16_sum_reducescatter_on_stream__nvshmemx_uint32_alltoall_on_stream__nvshmemx_uint32_broadcast_on_stream__nvshmemx_uint32_fcollect_on_stream__nvshmemx_uint32_max_reduce_on_stream__nvshmemx_uint32_max_reducescatter_on_stream__nvshmemx_uint32_min_reduce_on_stream__nvshmemx_uint32_min_reducescatter_on_stream__nvshmemx_uint32_sum_reduce_on_stream__nvshmemx_uint32_sum_reducescatter_on_stream__nvshmemx_uint64_alltoall_on_stream__nvshmemx_uint64_broadcast_on_stream__nvshmemx_uint64_fcollect_on_stream__nvshmemx_uint64_max_reduce_on_stream__nvshmemx_uint64_max_reducescatter_on_stream__nvshmemx_uint64_min_reduce_on_stream__nvshmemx_uint64_min_reducescatter_on_stream__nvshmemx_uint64_sum_reduce_on_stream__nvshmemx_uint64_sum_reducescatter_on_stream__nvshmemx_uint8_alltoall_on_stream__nvshmemx_uint8_broadcast_on_stream__nvshmemx_uint8_fco""llect_on_stream__nvshmemx_uint8_max_reduce_on_stream__nvshmemx_uint8_max_reducescatter_on_stream__nvshmemx_uint8_min_reduce_on_stream__nvshmemx_uint8_min_reducescatter_on_stream__nvshmemx_uint8_sum_reduce_on_stream__nvshmemx_uint8_sum_reducescatter_on_stream__nvshmemx_vendor_get_version_infopop__prepare____pyx_capi____qualname____set_name__setdefault__test__valuesint (CUlibrary)\000\000int (CUmodule)\000\000int (int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *)\000int (nvshmem_team_t *, nvshmem_team_config_t *, long, int, int)\000int (nvshmem_team_t)\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, char *, char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, cudaStream_t)\000\000int (nvshmem_team_t, double *, double const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, float *, float const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, half *, half const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int *, int const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshm""em_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_t)\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, long *, long const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, short *, short const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t,"" uint64_t *, uint64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t)\000int (nvshmemx_team_uniqueid_t *)\000int (nvshmemx_uniqueid_t *)\000int (unsigned int, nvshmemx_init_attr_t *)\000int (void *, nvshmemx_init_attr_t *)\000int (void *, size_t)\000int (void)\000\000\000void (cudaStream_t)\000\000\000void (int *, int *)\000void (int *, int *, int *)\000void (nvshmem_team_t)\000void (nvshmem_team_t, nvshmem_team_config_t *)\000void (uint64_t *, int, uint64_t, cudaStream_t)\000void (uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void *)\000void (void *, void const *, size_t, int, cudaStream_t)\000\000void (void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void)\000\000void *(nvshmem_team_t, void const *)\000void *(size_t)\000void *(size_t, size_t)\000\000void *(void *, size_t, int)\000void *(void const *, int)\000_nvshmemx_culibrary_finalize\000_nvshmemx_culibrary_init\000_nvshmemx_cumodule_finalize\000_nvshmemx_cumodule_init\000_nvshmemx_set_attr_uniqueid_args\000_nvshmemx_team_init\000_nvshmem_barrier\000_nvshmem_team_my_pe\000_nvshmem_team_n_pes\000_nvshmemx_longlong_alltoall_on_stream\000_nvshmemx_longlong_fcollect_on_stream\000_nvshmemx_longlong_max_reduce_on_stream\000_nvshmemx_longlong_max_reducescatter_on_stream\000_nvshmemx_longlong_min_reduce_on_stream\000_nvshmemx_longlong_min_reducescatter_on_stream\000_nvshmemx_longlong_sum_reduce_on_stream\000_nvshmemx_longlong_sum_reducescatter_on_stream\000_nvshmemx_longlong_broadcast_on_stream\000_nvshmemx_bfloat16_alltoall_on_stream\000_nvshmemx_bfloat16_fcollect_on_stream\000_nvshmemx_bfloat16_max_reduce_on_stream\000_nvshmemx_bfloat16_max_reducescatter_on_stream\000_nvshmemx_bfloat16_min_reduce_on_stream""\000_nvshmemx_bfloat16_min_reducescatter_on_stream\000_nvshmemx_bfloat16_sum_reduce_on_stream\000_nvshmemx_bfloat16_sum_reducescatter_on_stream\000_nvshmemx_bfloat16_broadcast_on_stream\000_nvshmemx_char_alltoall_on_stream\000_nvshmemx_char_fcollect_on_stream\000_nvshmemx_char_max_reduce_on_stream\000_nvshmemx_char_max_reducescatter_on_stream\000_nvshmemx_char_min_reduce_on_stream\000_nvshmemx_char_min_reducescatter_on_stream\000_nvshmemx_char_sum_reduce_on_stream\000_nvshmemx_char_sum_reducescatter_on_stream\000_nvshmemx_char_broadcast_on_stream\000_nvshmemx_barrier_on_stream\000_nvshmemx_team_sync_on_stream\000_nvshmemx_double_alltoall_on_stream\000_nvshmemx_double_fcollect_on_stream\000_nvshmemx_double_max_reduce_on_stream\000_nvshmemx_double_max_reducescatter_on_stream\000_nvshmemx_double_min_reduce_on_stream\000_nvshmemx_double_min_reducescatter_on_stream\000_nvshmemx_double_sum_reduce_on_stream\000_nvshmemx_double_sum_reducescatter_on_stream\000_nvshmemx_double_broadcast_on_stream\000_nvshmemx_float_alltoall_on_stream\000_nvshmemx_float_fcollect_on_stream\000_nvshmemx_float_max_reduce_on_stream\000_nvshmemx_float_max_reducescatter_on_stream\000_nvshmemx_float_min_reduce_on_stream\000_nvshmemx_float_min_reducescatter_on_stream\000_nvshmemx_float_sum_reduce_on_stream\000_nvshmemx_float_sum_reducescatter_on_stream\000_nvshmemx_float_broadcast_on_stream\000_nvshmemx_half_alltoall_on_stream\000_nvshmemx_half_fcollect_on_stream\000_nvshmemx_half_max_reduce_on_stream\000_nvshmemx_half_max_reducescatter_on_stream\000_nvshmemx_half_min_reduce_on_stream\000_nvshmemx_half_min_reducescatter_on_stream\000_nvshmemx_half_sum_reduce_on_stream\000_nvshmemx_half_sum_reducescatter_on_stream\000_nvshmemx_half_broadcast_on_stream\000_nvshmemx_int_alltoall_on_stream\000_nvshmemx_int_fcollect_on_stream\000_nvshmemx_int_max_reduce_on_stream\000_nvshmemx_int_max_reducescatter_on_stream\000_nvshmemx_int_min_reduce_on_stream\000_nvshmemx_int_min_reducescatter_on_stream\000_nvshmemx_int_""sum_reduce_on_stream\000_nvshmemx_int_sum_reducescatter_on_stream\000_nvshmemx_int_broadcast_on_stream\000_nvshmem_team_split_strided\000_nvshmem_team_split_2d\000_nvshmem_team_translate_pe\000_nvshmemx_int16_alltoall_on_stream\000_nvshmemx_int16_fcollect_on_stream\000_nvshmemx_int16_max_reduce_on_stream\000_nvshmemx_int16_max_reducescatter_on_stream\000_nvshmemx_int16_min_reduce_on_stream\000_nvshmemx_int16_min_reducescatter_on_stream\000_nvshmemx_int16_sum_reduce_on_stream\000_nvshmemx_int16_sum_reducescatter_on_stream\000_nvshmemx_int16_broadcast_on_stream\000_nvshmemx_int32_alltoall_on_stream\000_nvshmemx_int32_fcollect_on_stream\000_nvshmemx_int32_max_reduce_on_stream\000_nvshmemx_int32_max_reducescatter_on_stream\000_nvshmemx_int32_min_reduce_on_stream\000_nvshmemx_int32_min_reducescatter_on_stream\000_nvshmemx_int32_sum_reduce_on_stream\000_nvshmemx_int32_sum_reducescatter_on_stream\000_nvshmemx_int32_broadcast_on_stream\000_nvshmemx_int64_alltoall_on_stream\000_nvshmemx_int64_fcollect_on_stream\000_nvshmemx_int64_max_reduce_on_stream\000_nvshmemx_int64_max_reducescatter_on_stream\000_nvshmemx_int64_min_reduce_on_stream\000_nvshmemx_int64_min_reducescatter_on_stream\000_nvshmemx_int64_sum_reduce_on_stream\000_nvshmemx_int64_sum_reducescatter_on_stream\000_nvshmemx_int64_broadcast_on_stream\000_nvshmemx_int8_alltoall_on_stream\000_nvshmemx_int8_fcollect_on_stream\000_nvshmemx_int8_max_reduce_on_stream\000_nvshmemx_int8_max_reducescatter_on_stream\000_nvshmemx_int8_min_reduce_on_stream\000_nvshmemx_int8_min_reducescatter_on_stream\000_nvshmemx_int8_sum_reduce_on_stream\000_nvshmemx_int8_sum_reducescatter_on_stream\000_nvshmemx_int8_broadcast_on_stream\000_nvshmemx_long_alltoall_on_stream\000_nvshmemx_long_fcollect_on_stream\000_nvshmemx_long_max_reduce_on_stream\000_nvshmemx_long_max_reducescatter_on_stream\000_nvshmemx_long_min_reduce_on_stream\000_nvshmemx_long_min_reducescatter_on_stream\000_nvshmemx_long_sum_reduce_on_stream\000_nvshmemx_long_sum_reducescat""ter_on_stream\000_nvshmemx_long_broadcast_on_stream\000_nvshmemx_short_alltoall_on_stream\000_nvshmemx_short_fcollect_on_stream\000_nvshmemx_short_max_reduce_on_stream\000_nvshmemx_short_max_reducescatter_on_stream\000_nvshmemx_short_min_reduce_on_stream\000_nvshmemx_short_min_reducescatter_on_stream\000_nvshmemx_short_sum_reduce_on_stream\000_nvshmemx_short_sum_reducescatter_on_stream\000_nvshmemx_short_broadcast_on_stream\000_nvshmemx_schar_alltoall_on_stream\000_nvshmemx_schar_fcollect_on_stream\000_nvshmemx_schar_max_reduce_on_stream\000_nvshmemx_schar_max_reducescatter_on_stream\000_nvshmemx_schar_min_reduce_on_stream\000_nvshmemx_schar_min_reducescatter_on_stream\000_nvshmemx_schar_sum_reduce_on_stream\000_nvshmemx_schar_sum_reducescatter_on_stream\000_nvshmemx_schar_broadcast_on_stream\000_nvshmemx_size_alltoall_on_stream\000_nvshmemx_size_fcollect_on_stream\000_nvshmemx_size_max_reduce_on_stream\000_nvshmemx_size_max_reducescatter_on_stream\000_nvshmemx_size_min_reduce_on_stream\000_nvshmemx_size_min_reducescatter_on_stream\000_nvshmemx_size_sum_reduce_on_stream\000_nvshmemx_size_sum_reducescatter_on_stream\000_nvshmemx_size_broadcast_on_stream\000_nvshmemx_uint16_alltoall_on_stream\000_nvshmemx_uint16_fcollect_on_stream\000_nvshmemx_uint16_max_reduce_on_stream\000_nvshmemx_uint16_max_reducescatter_on_stream\000_nvshmemx_uint16_min_reduce_on_stream\000_nvshmemx_uint16_min_reducescatter_on_stream\000_nvshmemx_uint16_sum_reduce_on_stream\000_nvshmemx_uint16_sum_reducescatter_on_stream\000_nvshmemx_uint16_broadcast_on_stream\000_nvshmemx_uint32_alltoall_on_stream\000_nvshmemx_uint32_fcollect_on_stream\000_nvshmemx_uint32_max_reduce_on_stream\000_nvshmemx_uint32_max_reducescatter_on_stream\000_nvshmemx_uint32_min_reduce_on_stream\000_nvshmemx_uint32_min_reducescatter_on_stream\000_nvshmemx_uint32_sum_reduce_on_stream\000_nvshmemx_uint32_sum_reducescatter_on_stream\000_nvshmemx_uint32_broadcast_on_stream\000_nvshmemx_uint64_alltoall_on_stream\000_nvshmemx_uint64_""fcollect_on_stream\000_nvshmemx_uint64_max_reduce_on_stream\000_nvshmemx_uint64_max_reducescatter_on_stream\000_nvshmemx_uint64_min_reduce_on_stream\000_nvshmemx_uint64_min_reducescatter_on_stream\000_nvshmemx_uint64_sum_reduce_on_stream\000_nvshmemx_uint64_sum_reducescatter_on_stream\000_nvshmemx_uint64_broadcast_on_stream\000_nvshmemx_uint8_alltoall_on_stream\000_nvshmemx_uint8_fcollect_on_stream\000_nvshmemx_uint8_max_reduce_on_stream\000_nvshmemx_uint8_max_reducescatter_on_stream\000_nvshmemx_uint8_min_reduce_on_stream\000_nvshmemx_uint8_min_reducescatter_on_stream\000_nvshmemx_uint8_sum_reduce_on_stream\000_nvshmemx_uint8_sum_reducescatter_on_stream\000_nvshmemx_uint8_broadcast_on_stream\000_nvshmemx_team_get_uniqueid\000_nvshmemx_get_uniqueid\000_nvshmemx_hostlib_init_attr\000_nvshmemx_set_attr_mpi_comm_args\000_nvshmemx_buffer_unregister_symmetric\000_nvshmem_my_pe\000_nvshmem_n_pes\000_nvshmemx_init_status\000_nvshmemx_barrier_all_on_stream\000_nvshmemx_quiet_on_stream\000_nvshmemx_sync_all_on_stream\000_nvshmem_info_get_version\000_nvshmemx_vendor_get_version_info\000_nvshmem_team_destroy\000_nvshmem_team_get_config\000_nvshmemx_signal_wait_until_on_stream\000_nvshmemx_signal_op_on_stream\000_nvshmem_free\000_nvshmemx_getmem_on_stream\000_nvshmemx_putmem_on_stream\000_nvshmemx_putmem_signal_on_stream\000_nvshmem_barrier_all\000_nvshmemx_hostlib_finalize\000_nvshmemx_mc_ptr\000_nvshmem_malloc\000_nvshmem_align\000_nvshmem_calloc\000_nvshmemx_buffer_register_symmetric\000_nvshmem_ptr\200\001\330\004\005\330\004\007\200z\220\023\220A\330\010\024\320\024.\250a\330\004\013\2109\220A\220Q\200\001\330\004\005\330\004\007\200z\220\027\230\001\330\010\017\210q\340\004\032\230!\330\004\025\220Q\340\004\005\330\004\010\210\001\320\t \240\n\250!\340\004\005\330\004\010\210\001\320\t$\240J\250a\340\004\005\330\004\010\210\001\320\t%\240Z\250q\340\004\005\330\004\010\210\001\320\t\036\230j\250\001\340\004\005\330\004\010\210\001\320\t\036\230j\250\001\340\004\005\330""\004\010\210\001\320\t)\250\032\2601\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t\037\230z\250\021\340\004\005\330\004\010\210\001\320\t\037\230z\250\021\340\004\005\330\004\010\210\001\320\t\036\230j\250\001\340\004\005\330\004\010\210\001\320\t\035\230Z\240q\340\004\005\330\004\010\210\001\320\t\034\230J\240a\340\004\005\330\004\010\210\001\320\t \240\n\250!\340\004\005\330\004\010\210\001\320\t#\240:\250Q\340\004\005\330\004\010\210\001\320\t#\240:\250Q\340\004\005\330\004\010\210\001\320\t(\250\n\260!\340\004\005\330\004\010\210\001\320\t*\250*\260A\340\004\005\330\004\010\210\001\320\t+\250:\260Q\340\004\005\330\004\010\210\001\320\t+\250:\260Q\340\004\005\330\004\010\210\001\320\t#\240:\250Q\340\004\005\330\004\010\210\001\320\t&\240j\260\001\340\004\005\330\004\010\210\001\320\t%\240Z\250q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t0\260\n\270!\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t+\250:\260Q\340\004\005\330\004\010\210\001\320\t/\250z\270\021\340\004\005\330\004\010\210\001\320\t-\250Z\260q\340\004\005\330\004\010\210\001\320\t,\250J\260a""\340\004\005\330\004\010\210\001\320\t6\260j\300\001\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t6\260j\300\001\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t0\260\n\270!\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005""\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320""\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t2\260*\270A\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t7\260z\300\021\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t4\260J\270a\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t9\270\032\3001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004""\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t9\270\032\3001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t9\270\032\3001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t>\270j\310\001\340\004\005\330\004\010\210\001\320\t:\270*\300A\340\004\005\330\004\010\210\001\320\t;\270:\300Q\340\004\005\330\004\010\210\001\320\t<\270J\300a\340\004\005\330\004\010\210\001\320\t+\250:\260Q\340\004\005\330\004\010\210\001\320\t*\250*\260A\340\004\005\330\004\010\210\001\320\t0\260\n\270!\340\004\005\330\004\010\210\001\320\t0\260\n\270!\340\004\005\330\004\010\210\001\320\t&\240j\260\001\340\004\005\330\004\010\210\001\320\t'\240z\260\021\340\004\005\330\004\010\210\001\320\t+\250:\260Q\340\004\005\330\004\010\210\001\320\t3\260:\270Q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\005\330\004""\010\210\001\320\t(\250\n\260!\340\004\005\330\004\010\210\001\320\t,\250J\260a\340\004\005\330\004\010\210\001\320\t*\250*\260A\340\004\005\330\004\010\210\001\320\t1\260\032\2701\340\004\005\330\004\010\210\001\320\t*\250*\260A\340\004\005\330\004\010\210\001\320\t)\250\032\2601\340\004\005\330\004\010\210\001\320\t-\250Z\260q\340\004\005\330\004\010\210\001\320\t5\260Z\270q\340\004\020\220\001\330\004\013\2101";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 461; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 220) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 461; i < 464; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 464; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 461;
      for (Py_ssize_t i=0; i<3; ++i) {
        #if PY_VERSION_HEX >= 0x030F0000
        PyUnstable_SetImmortal(table[i]);
        #elif CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        if ((PY_SSIZE_T_MAX <= _Py_IMMORTAL_REFCNT_LOCAL)) break;
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], ((Py_ssize_t)_Py_IMMORTAL_REFCNT_LOCAL + 1));
        }
        #else
        if ((PY_SSIZE_T_MAX < _Py_IMMORTAL_INITIAL_REFCNT)) break;
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 1;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 1;
    unsigned int flags : 10;
    unsigned int first_line : 12;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1781};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings__internal_nvshm_2, __pyx_mstate->__pyx_n_u_inspect_function_pointers, __pyx_mstate->__pyx_kp_b_iso88591_z_q_Q_Ja_Zq_j_j_1_1_1_z_z_j_Zq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 2435};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_name};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings__internal_nvshm_2, __pyx_mstate->__pyx_n_u_inspect_function_pointer, __pyx_mstate->__pyx_kp_b_iso88591_z_A_a_9AQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        Py_INCREF(values[i]);
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyErrFetchRestore (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* ErrOccurredWithGIL */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void) {
  int err;
  PyGILState_STATE _save = PyGILState_Ensure();
  err = !!PyErr_Occurred();
  PyGILState_Release(_save);
  return err;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        Py_INCREF(src[i]);
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetAttrStr (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* FunctionExport */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig) {
    PyObject *cobj;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(api_dict, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(cobj);
    return -1;
}

/* GetApiDict */
static PyObject *__Pyx_ApiExport_GetApiDict(void) {
    PyObject *d;
    if (__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_capi, &d) == -1)
        return NULL;
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_pyx_capi, d) < 0)
            goto bad;
    }
    return d;
bad:
    Py_XDECREF(d);
    return NULL;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base = NULL, *meth, *new_base, *result, *new_bases = NULL;
#if CYTHON_ASSUME_SAFE_SIZE
    size_bases = PyTuple_GET_SIZE(bases);
#else
    size_bases = PyTuple_Size(bases);
    if (size_bases < 0) return NULL;
#endif
    for (i = 0; i < size_bases; i++) {
#if CYTHON_AVOID_BORROWED_REFS
        Py_CLEAR(base);
#endif
#if CYTHON_ASSUME_SAFE_MACROS
        base = PyTuple_GET_ITEM(bases, i);
#else
        base = PyTuple_GetItem(bases, i);
        if (!base) goto error;
#endif
#if CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(base);
#endif
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_mstate_global->__pyx_n_u_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                PyObject *base_from_list;
#if CYTHON_ASSUME_SAFE_MACROS
                base_from_list = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base_from_list);
                Py_INCREF(base_from_list);
#else
                base_from_list = PyTuple_GetItem(bases, j);
                if (!base_from_list) goto error;
                Py_INCREF(base_from_list);
                if (PyList_SetItem(new_bases, j, base_from_list) < 0) goto error;
#endif
            }
        }
#if CYTHON_ASSUME_SAFE_SIZE
        j = PyList_GET_SIZE(new_bases);
#else
        j = PyList_Size(new_bases);
        if (j < 0) goto error;
#endif
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return result;
error:
    Py_XDECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_SIZE
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
        metaclass = &PyType_Type;
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectLookupSpecial (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_mstate_global->__pyx_n_u_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_mstate_global->__pyx_n_u_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
    Py_XDECREF(owned_metaclass);
    return result;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType (used by FetchCommonType) */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* CallTypeTraverse (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int __Pyx_CyFunction_set_annotate_in_dict_if_exists(PyObject *op_in, PyObject *value);
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    if (unlikely(__Pyx_CyFunction_set_annotate_in_dict_if_exists((PyObject*) op, Py_None) < 0)) return -1;
    return 0;
}
static int
__Pyx_CyFunction_get_dict_if_exists(PyObject *op_in, PyObject **dict) {
    /* Return 1 if the function dict exists, 0 otherwise.  This cannot fail:
     * _PyObject_GetDictPtr() may clear errors internally, but never reports them. */
#if CYTHON_COMPILING_IN_PYPY
    *dict = PyObject_GenericGetDict(op_in, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030C0000
    *dict = ((__pyx_CyFunctionObject*) op_in)->func_dict;
#else
    PyObject **dictptr = _PyObject_GetDictPtr(op_in);
    *dict = likely(dictptr) ? *dictptr : NULL;
#endif
    return *dict ? 1 : 0;
}
static int
__Pyx_CyFunction_get_annotate_from_dict_if_exists(PyObject *op_in, PyObject **annotate) {
    PyObject *dict;
    int dict_found;
    *annotate = NULL;
    dict_found = __Pyx_CyFunction_get_dict_if_exists(op_in, &dict);
    if (!dict_found) return 0;
    return __Pyx_PyDict_GetItemRef(dict, __pyx_mstate_global->__pyx_n_u_annotate, annotate);
}
static int
__Pyx_CyFunction_set_annotate_in_dict_if_exists(PyObject *op_in, PyObject *value) {
    PyObject *dict;
    int dict_found;
    dict_found = __Pyx_CyFunction_get_dict_if_exists(op_in, &dict);
    if (!dict_found) return 0;
    return PyDict_SetItem(dict, __pyx_mstate_global->__pyx_n_u_annotate, value);
}
static int
__Pyx_CyFunction_set_annotate_in_dict(PyObject *op_in, PyObject *value) {
    PyObject *dict;
    int result;
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = __Pyx_CyFunction_get_dict((__pyx_CyFunctionObject*) op_in, NULL);
#else
    dict = PyObject_GenericGetDict(op_in, NULL);
#endif
    if (unlikely(!dict)) return -1;
    result = PyDict_SetItem(dict, __pyx_mstate_global->__pyx_n_u_annotate, value);
    Py_DECREF(dict);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(PyObject *op_in, void *context) {
    PyObject *annotate = NULL;
    PyObject *result = NULL;
    __pyx_CyFunctionObject *op = (__pyx_CyFunctionObject*) op_in;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op_in);
    result = __Pyx_XNewRef(op->func_annotations);
    __Pyx_END_CRITICAL_SECTION();
    if (result) return result;
    if (unlikely(__Pyx_CyFunction_get_annotate_from_dict_if_exists(op_in, &annotate) < 0)) return NULL;
    if (!annotate || annotate == Py_None) {
        Py_XDECREF(annotate);
        __Pyx_BEGIN_CRITICAL_SECTION(op_in);
        result = __Pyx_CyFunction_get_annotations_locked(op);
        __Pyx_END_CRITICAL_SECTION();
        return result;
    }
    PyObject *format = PyLong_FromLong(1L);  // annotationlib.Format.VALUE
    if (likely(format)) {
        result = __Pyx_PyObject_CallOneArg(annotate, format);
        Py_DECREF(format);
    }
    Py_DECREF(annotate);
    if (unlikely(!result)) return NULL;
    if (unlikely(!PyDict_Check(result))) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ must return a dict");
        Py_DECREF(result);
        return NULL;
    }
    __Pyx_BEGIN_CRITICAL_SECTION(op_in);
    __Pyx_Py_XDECREF_SET(op->func_annotations, __Pyx_NewRef(result));
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *__Pyx_CyFunction_annotate_impl(PyObject *self, PyObject *args) {
    CYTHON_UNUSED_VAR(args);
    if (unlikely(!self)) {
        PyErr_SetString(PyExc_SystemError, "cython __annotate__ called without 'self' argument");
    }
    Py_XINCREF(self);
    return self;
}
static PyMethodDef __Pyx_CyFunction_annotate_method = {
    "__annotate__",
    (PyCFunction)(void (*)(void))__Pyx_CyFunction_annotate_impl,
    METH_VARARGS,
    "Placeholder __annotate__ function to allow 'functools.wraps' to work "
    "on Cython functions."
};
static PyObject *
__Pyx_CyFunction_get_annotate(PyObject *op_in, void *context) {
    PyObject *annotate = NULL;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(__Pyx_CyFunction_get_annotate_from_dict_if_exists(op_in, &annotate) < 0)) return NULL;
    if (annotate) return annotate;
    PyObject *annotations = __Pyx_CyFunction_get_annotations(op_in, NULL);
    if (unlikely(!annotations)) return NULL;
    PyObject *method = PyCFunction_New(
        &__Pyx_CyFunction_annotate_method,
        annotations);
    Py_DECREF(annotations);
    return method;
}
static int
__Pyx_CyFunction_set_annotate(PyObject *op_in, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ cannot be deleted");
        return -1;
    }
    if (unlikely(value != Py_None && !PyCallable_Check(value))) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ must be callable or None");
        return -1;
    }
    if (value != Py_None) {
        __pyx_CyFunctionObject *op = (__pyx_CyFunctionObject*) op_in;
        __Pyx_BEGIN_CRITICAL_SECTION(op_in);
        Py_CLEAR(op->func_annotations);
        __Pyx_END_CRITICAL_SECTION();
    }
    return __Pyx_CyFunction_set_annotate_in_dict(op_in, value);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"__annotate__", (getter)__Pyx_CyFunction_get_annotate, (setter)__Pyx_CyFunction_set_annotate, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_annotations);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__2);
    }
    goto done;
}
#endif

/* PyObjectVectorCallKwBuilder (used by CIntToPy) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    Py_INCREF(key);
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPyVerify (used by CIntFromPy) */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
