/* Generated by Cython 3.2.8 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "language": "c++",
        "name": "nvshmem.bindings.nvshmem",
        "sources": [
            "nvshmem/bindings/nvshmem.pyx"
        ]
    },
    "module_name": "nvshmem.bindings.nvshmem"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_8"
#define CYTHON_HEX_VERSION 0x030208F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__nvshmem__bindings__nvshmem
#define __PYX_HAVE_API__nvshmem__bindings__nvshmem
/* Early includes */
#include <stdint.h>
#include "cuda_runtime_api.h"
#include "cuda_fp16.h"
#include "cuda_bf16.h"
#include "limits.h"
#include <string.h>
#include <stdio.h>

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE PyObject *
    __Pyx_CAPI_PyList_GetItemRef(PyObject *list, Py_ssize_t index)
    {
        PyObject *item = PyList_GetItem(list, index);
        Py_XINCREF(item);
        return item;
    }
    #else
    #define __Pyx_CAPI_PyList_GetItemRef PyList_GetItemRef
    #endif

    #if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Extend(PyObject *list, PyObject *iterable)
    {
        return PyList_SetSlice(list, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, iterable);
    }

    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Clear(PyObject *list)
    {
        return PyList_SetSlice(list, 0, PY_SSIZE_T_MAX, NULL);
    }
    #else
    #define __Pyx_CAPI_PyList_Extend PyList_Extend
    #define __Pyx_CAPI_PyList_Clear PyList_Clear
    #endif
    

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    
#include <stddef.h>

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result)
    {
        int res;
        PyObject *key_obj = PyUnicode_FromString(key);
        if (key_obj == NULL) {
            *result = NULL;
            return -1;
        }
        res = __Pyx_PyDict_GetItemRef(mp, key_obj, result);
        Py_DECREF(key_obj);
        return res;
    }
    #else
    #define __Pyx_CAPI_PyDict_GetItemStringRef PyDict_GetItemStringRef
    #endif
    #if PY_VERSION_HEX < 0x030d0000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030F0000)
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_SetDefaultRef(PyObject *d, PyObject *key, PyObject *default_value,
                        PyObject **result)
    {
        PyObject *value;
        if (__Pyx_PyDict_GetItemRef(d, key, &value) < 0) {
            // get error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (value != NULL) {
            // present
            if (result) {
                *result = value;
            }
            else {
                Py_DECREF(value);
            }
            return 1;
        }

        // missing: set the item
        if (PyDict_SetItem(d, key, default_value) < 0) {
            // set error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (result) {
            Py_INCREF(default_value);
            *result = default_value;
        }
        return 0;
    }
    #else
    #define __Pyx_CAPI_PyDict_SetDefaultRef PyDict_SetDefaultRef
    #endif
    

    #if PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int __Pyx_PyWeakref_GetRef(PyObject *ref, PyObject **pobj)
    {
        PyObject *obj = PyWeakref_GetObject(ref);
        if (obj == NULL) {
            // SystemError if ref is NULL
            *pobj = NULL;
            return -1;
        }
        if (obj == Py_None) {
            *pobj = NULL;
            return 0;
        }
        Py_INCREF(obj);
        *pobj = obj;
        return 1;
    }
    #else
    #define __Pyx_PyWeakref_GetRef PyWeakref_GetRef
    #endif
    
#include "pythread.h"

    #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "nvshmem/bindings/nvshmem.pyx",
  "cpython/contextvars.pxd",
  "(tree fragment)",
  "cpython/type.pxd",
  "cpython/bool.pxd",
  "cpython/complex.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "cynvshmem.pxd":143
 * 
 * # Longs
 * ctypedef long longlong             # <<<<<<<<<<<<<<
 * ctypedef signed char schar
 * 
*/
typedef long __pyx_t_7nvshmem_8bindings_9cynvshmem_longlong;

/* "cynvshmem.pxd":144
 * # Longs
 * ctypedef long longlong
 * ctypedef signed char schar             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef signed char __pyx_t_7nvshmem_8bindings_9cynvshmem_schar;

/* "cynvshmem.pxd":151
 *     int version
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'             # <<<<<<<<<<<<<<
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
*/
typedef int nvshmemx_qp_handle_t;

/* "cynvshmem.pxd":152
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
*/
typedef uint64_t nvshmemx_team_uniqueid_t;

/* "cynvshmem.pxd":153
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
*/
typedef int32_t nvshmem_team_t;

/* "cynvshmem.pxd":165
 *     nvshmemx_team_uniqueid_t uniqueid
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
*/
typedef nvshmem_team_t nvshmemx_team_t;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid;
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId;
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid;
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId;
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr;
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig;
struct nvshmemx_uniqueid_v1;
typedef struct nvshmemx_uniqueid_v1 nvshmemx_uniqueid_v1;
struct nvshmemx_uniqueid_args_v1;
typedef struct nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_v1;
struct nvshmem_team_config_v2;
typedef struct nvshmem_team_config_v2 nvshmem_team_config_v2;
struct nvshmemx_init_args_v2;
typedef struct nvshmemx_init_args_v2 nvshmemx_init_args_v2;
struct nvshmemx_init_attr_v2;
typedef struct nvshmemx_init_attr_v2 nvshmemx_init_attr_v2;

/* "cynvshmem.pxd":21
 * 
 * # enums
 * ctypedef enum nvshmemx_signal_op_t "nvshmemx_signal_op_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_SIGNAL_SET "NVSHMEM_SIGNAL_SET" = 9
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
*/
enum nvshmemx_signal_op_t {
  NVSHMEM_SIGNAL_SET = 9,
  NVSHMEM_SIGNAL_ADD = 10
};
typedef enum nvshmemx_signal_op_t nvshmemx_signal_op_t;

/* "cynvshmem.pxd":25
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
 * 
 * ctypedef enum nvshmemx_tma_policy_t "nvshmemx_tma_policy_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_TMA_DISABLE "NVSHMEMX_TMA_DISABLE" = 0
 *     NVSHMEMX_TMA_ENABLE "NVSHMEMX_TMA_ENABLE" = 1
*/
enum nvshmemx_tma_policy_t {
  NVSHMEMX_TMA_DISABLE = 0,
  NVSHMEMX_TMA_ENABLE = 1,
  NVSHMEMX_TMA_FORCE = 2,
  NVSHMEMX_TMA_POLICY_MAX = 0x7FFF
};
typedef enum nvshmemx_tma_policy_t nvshmemx_tma_policy_t;

/* "cynvshmem.pxd":31
 *     NVSHMEMX_TMA_POLICY_MAX "NVSHMEMX_TMA_POLICY_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_smem_amount_t "nvshmemx_smem_amount_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SMEM_RECOMMENDED "NVSHMEMX_SMEM_RECOMMENDED" = 0
 *     NVSHMEMX_SMEM_MINIMUM "NVSHMEMX_SMEM_MINIMUM" = 1
*/
enum nvshmemx_smem_amount_t {
  NVSHMEMX_SMEM_RECOMMENDED = 0,
  NVSHMEMX_SMEM_MINIMUM = 1,
  NVSHMEMX_SMEM_BARRIERS_ONLY = 2,
  NVSHMEMX_SMEM_AMOUNT_MAX = 0x7FFF
};
typedef enum nvshmemx_smem_amount_t nvshmemx_smem_amount_t;

/* "cynvshmem.pxd":37
 *     NVSHMEMX_SMEM_AMOUNT_MAX "NVSHMEMX_SMEM_AMOUNT_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_cmp_type_t "nvshmemx_cmp_type_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_CMP_EQ "NVSHMEM_CMP_EQ" = 0
 *     NVSHMEM_CMP_NE "NVSHMEM_CMP_NE"
*/
enum nvshmemx_cmp_type_t {
  NVSHMEM_CMP_EQ = 0,
  NVSHMEM_CMP_NE,
  NVSHMEM_CMP_GT,
  NVSHMEM_CMP_LE,
  NVSHMEM_CMP_LT,
  NVSHMEM_CMP_GE,
  NVSHMEM_CMP_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_cmp_type_t nvshmemx_cmp_type_t;

/* "cynvshmem.pxd":46
 *     NVSHMEM_CMP_SENTINEL "NVSHMEM_CMP_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_thread_support_t "nvshmemx_thread_support_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_THREAD_SINGLE "NVSHMEM_THREAD_SINGLE" = 0
 *     NVSHMEM_THREAD_FUNNELED "NVSHMEM_THREAD_FUNNELED"
*/
enum nvshmemx_thread_support_t {
  NVSHMEM_THREAD_SINGLE = 0,
  NVSHMEM_THREAD_FUNNELED,
  NVSHMEM_THREAD_SERIALIZED,
  NVSHMEM_THREAD_MULTIPLE,
  NVSHMEM_THREAD_TYPE_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_thread_support_t nvshmemx_thread_support_t;

/* "cynvshmem.pxd":53
 *     NVSHMEM_THREAD_TYPE_SENTINEL "NVSHMEM_THREAD_TYPE_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_proxy_status_t "nvshmemx_proxy_status_t":             # <<<<<<<<<<<<<<
 *     PROXY_GLOBAL_EXIT_NOT_REQUESTED "PROXY_GLOBAL_EXIT_NOT_REQUESTED" = 0
 *     PROXY_GLOBAL_EXIT_INIT "PROXY_GLOBAL_EXIT_INIT"
*/
enum nvshmemx_proxy_status_t {
  PROXY_GLOBAL_EXIT_NOT_REQUESTED = 0,
  PROXY_GLOBAL_EXIT_INIT,
  PROXY_GLOBAL_EXIT_REQUESTED,
  PROXY_GLOBAL_EXIT_FINISHED,
  PROXY_GLOBAL_EXIT_MAX_STATE = 0x7FFF
};
typedef enum nvshmemx_proxy_status_t nvshmemx_proxy_status_t;

/* "cynvshmem.pxd":60
 *     PROXY_GLOBAL_EXIT_MAX_STATE "PROXY_GLOBAL_EXIT_MAX_STATE" = 32767
 * 
 * ctypedef enum nvshmemx_init_status_t "nvshmemx_init_status_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_STATUS_NOT_INITIALIZED "NVSHMEM_STATUS_NOT_INITIALIZED" = 0
 *     NVSHMEM_STATUS_IS_BOOTSTRAPPED "NVSHMEM_STATUS_IS_BOOTSTRAPPED"
*/
enum nvshmemx_init_status_t {
  NVSHMEM_STATUS_NOT_INITIALIZED = 0,
  NVSHMEM_STATUS_IS_BOOTSTRAPPED,
  NVSHMEM_STATUS_IS_INITIALIZED,
  NVSHMEM_STATUS_LIMITED_MPG,
  NVSHMEM_STATUS_FULL_MPG,
  NVSHMEM_STATUS_INVALID = 0x7FFF
};
typedef enum nvshmemx_init_status_t nvshmemx_init_status_t;

/* "cynvshmem.pxd":68
 *     NVSHMEM_STATUS_INVALID "NVSHMEM_STATUS_INVALID" = 32767
 * 
 * ctypedef enum nvshmemx_qp_handle_index_t "nvshmemx_qp_handle_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_QP_HOST "NVSHMEMX_QP_HOST" = 0
 *     NVSHMEMX_QP_DEFAULT "NVSHMEMX_QP_DEFAULT" = 1
*/
enum nvshmemx_qp_handle_index_t {
  NVSHMEMX_QP_HOST = 0,
  NVSHMEMX_QP_DEFAULT = 1,
  NVSHMEMX_QP_ANY = 0x7FFF,
  NVSHMEMX_QP_ALL = 0x7FFF
};
typedef enum nvshmemx_qp_handle_index_t nvshmemx_qp_handle_index_t;

/* "cynvshmem.pxd":74
 *     NVSHMEMX_QP_ALL "NVSHMEMX_QP_ALL" = 32767
 * 
 * ctypedef enum nvshmem_pe_index_t "nvshmem_pe_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
*/
enum nvshmem_pe_index_t {

  /* "cynvshmem.pxd":77
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)             # <<<<<<<<<<<<<<
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":
*/
  NVSHMEM_PE_INVALID = -1L,
  NVSHMEMX_PE_ANY = (1 << 15),
  NVSHMEMX_PE_ALL = (1 << 15)
};
typedef enum nvshmem_pe_index_t nvshmem_pe_index_t;

/* "cynvshmem.pxd":79
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_TEAM_INVALID "NVSHMEM_TEAM_INVALID" = -(1)
 *     NVSHMEM_TEAM_WORLD "NVSHMEM_TEAM_WORLD" = 0
*/
enum nvshmem_team_id_t {
  NVSHMEM_TEAM_INVALID = -1L,
  NVSHMEM_TEAM_WORLD = 0,
  NVSHMEM_TEAM_WORLD_INDEX = 0,
  NVSHMEM_TEAM_SHARED = 1,
  NVSHMEM_TEAM_SHARED_INDEX = 1,
  NVSHMEMX_TEAM_NODE = 2,
  NVSHMEM_TEAM_NODE_INDEX = 2,
  NVSHMEMX_TEAM_SAME_MYPE_NODE = 3,
  NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX = 3,
  NVSHMEMI_TEAM_SAME_GPU = 4,
  NVSHMEM_TEAM_SAME_GPU_INDEX = 4,
  NVSHMEMI_TEAM_GPU_LEADERS = 5,
  NVSHMEM_TEAM_GPU_LEADERS_INDEX = 5,
  NVSHMEM_TEAM_MC_SHARED = 6,
  NVSHMEM_TEAM_MC_SHARED_INDEX = 6,
  NVSHMEM_TEAMS_MIN = 7,
  NVSHMEM_TEAM_INDEX_MAX = 0x7FFF
};
typedef enum nvshmem_team_id_t nvshmem_team_id_t;

/* "cynvshmem.pxd":98
 *     NVSHMEM_TEAM_INDEX_MAX "NVSHMEM_TEAM_INDEX_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_status "nvshmemx_status":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SUCCESS "NVSHMEMX_SUCCESS"
 *     NVSHMEMX_ERROR_INVALID_VALUE "NVSHMEMX_ERROR_INVALID_VALUE"
*/
enum nvshmemx_status {
  NVSHMEMX_SUCCESS,
  NVSHMEMX_ERROR_INVALID_VALUE,
  NVSHMEMX_ERROR_OUT_OF_MEMORY,
  NVSHMEMX_ERROR_NOT_SUPPORTED,
  NVSHMEMX_ERROR_SYMMETRY,
  NVSHMEMX_ERROR_GPU_NOT_SELECTED,
  NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED,
  NVSHMEMX_ERROR_INTERNAL,
  NVSHMEMX_ERROR_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_status nvshmemx_status;

/* "cynvshmem.pxd":109
 *     NVSHMEMX_ERROR_SENTINEL "NVSHMEMX_ERROR_SENTINEL" = 32767
 * 
 * ctypedef enum flags "flags":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_INIT_THREAD_PES "NVSHMEMX_INIT_THREAD_PES" = 1
 *     NVSHMEMX_INIT_WITH_MPI_COMM "NVSHMEMX_INIT_WITH_MPI_COMM" = (1 << 1)
*/
enum flags {

  /* "cynvshmem.pxd":114
 *     NVSHMEMX_INIT_WITH_SHMEM "NVSHMEMX_INIT_WITH_SHMEM" = (1 << 2)
 *     NVSHMEMX_INIT_WITH_UNIQUEID "NVSHMEMX_INIT_WITH_UNIQUEID" = (1 << 3)
 *     NVSHMEMX_INIT_MAX "NVSHMEMX_INIT_MAX" = (1 << 31)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  NVSHMEMX_INIT_THREAD_PES = 1,
  NVSHMEMX_INIT_WITH_MPI_COMM = (1 << 1),
  NVSHMEMX_INIT_WITH_SHMEM = (1 << 2),
  NVSHMEMX_INIT_WITH_UNIQUEID = (1 << 3),
  NVSHMEMX_INIT_MAX = (1 << 31)
};
typedef enum flags flags;

/* "cynvshmem.pxd":122
 *     ctypedef CUstream_st* cudaStream_t
 * 
 * ctypedef void* CUmodule 'CUmodule'             # <<<<<<<<<<<<<<
 * ctypedef void* CUlibrary 'CUlibrary'
 * 
*/
typedef void *CUmodule;

/* "cynvshmem.pxd":123
 * 
 * ctypedef void* CUmodule 'CUmodule'
 * ctypedef void* CUlibrary 'CUlibrary'             # <<<<<<<<<<<<<<
 * 
 * # Types for NVSHMEM Collectives
*/
typedef void *CUlibrary;

/* "cynvshmem.pxd":148
 * 
 * 
 * ctypedef struct nvshmemx_uniqueid_v1 'nvshmemx_uniqueid_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     char internal[124]
*/
struct nvshmemx_uniqueid_v1 {
  int version;
  char internal[124];
};

/* "cynvshmem.pxd":154
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
 *     int version
*/
typedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t;

/* "cynvshmem.pxd":155
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_v1* id
*/
struct nvshmemx_uniqueid_args_v1 {
  int version;
  nvshmemx_uniqueid_v1 *id;
  int myrank;
  int nranks;
};

/* "cynvshmem.pxd":160
 *     int myrank
 *     int nranks
 * ctypedef struct nvshmem_team_config_v2 'nvshmem_team_config_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     int num_contexts
*/
struct nvshmem_team_config_v2 {
  int version;
  int num_contexts;
  nvshmemx_team_uniqueid_t uniqueid;
  char padding[48];
};

/* "cynvshmem.pxd":166
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
*/
typedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t;

/* "cynvshmem.pxd":167
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
 *     int version
*/
typedef nvshmem_team_config_v2 nvshmem_team_config_t;

/* "cynvshmem.pxd":168
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_args_t uid_args
*/
struct nvshmemx_init_args_v2 {
  int version;
  nvshmemx_uniqueid_args_t uid_args;
  int cuda_device_id;
  char content[92];
};

/* "cynvshmem.pxd":173
 *     int cuda_device_id
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':
 *     int version
*/
typedef nvshmemx_init_args_v2 nvshmemx_init_args_t;

/* "cynvshmem.pxd":174
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     void* mpi_comm
*/
struct nvshmemx_init_attr_v2 {
  int version;
  void *mpi_comm;
  nvshmemx_init_args_t args;
};

/* "cynvshmem.pxd":178
 *     void* mpi_comm
 *     nvshmemx_init_args_t args
 * ctypedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t 'nvshmemx_init_attr_t'             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":116
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
*/
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":134
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
*/
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};

/* "nvshmem/bindings/nvshmem.pxd":17
 * ###############################################################################
 * 
 * ctypedef nvshmemx_uniqueid_args_v1 uniqueid_args             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_team_config_v2 team_config
 * ctypedef nvshmemx_init_args_v2 init_args
*/
typedef nvshmemx_uniqueid_args_v1 __pyx_t_7nvshmem_8bindings_7nvshmem_uniqueid_args;

/* "nvshmem/bindings/nvshmem.pxd":18
 * 
 * ctypedef nvshmemx_uniqueid_args_v1 uniqueid_args
 * ctypedef nvshmem_team_config_v2 team_config             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_init_args_v2 init_args
 * ctypedef nvshmemx_init_attr_v2 init_attr
*/
typedef nvshmem_team_config_v2 __pyx_t_7nvshmem_8bindings_7nvshmem_team_config;

/* "nvshmem/bindings/nvshmem.pxd":19
 * ctypedef nvshmemx_uniqueid_args_v1 uniqueid_args
 * ctypedef nvshmem_team_config_v2 team_config
 * ctypedef nvshmemx_init_args_v2 init_args             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_init_attr_v2 init_attr
 * 
*/
typedef nvshmemx_init_args_v2 __pyx_t_7nvshmem_8bindings_7nvshmem_init_args;

/* "nvshmem/bindings/nvshmem.pxd":20
 * ctypedef nvshmem_team_config_v2 team_config
 * ctypedef nvshmemx_init_args_v2 init_args
 * ctypedef nvshmemx_init_attr_v2 init_attr             # <<<<<<<<<<<<<<
 * 
 * ctypedef cudaStream_t Stream
*/
typedef nvshmemx_init_attr_v2 __pyx_t_7nvshmem_8bindings_7nvshmem_init_attr;

/* "nvshmem/bindings/nvshmem.pxd":22
 * ctypedef nvshmemx_init_attr_v2 init_attr
 * 
 * ctypedef cudaStream_t Stream             # <<<<<<<<<<<<<<
 * ctypedef CUmodule Module
 * ctypedef CUlibrary Library
*/
typedef cudaStream_t __pyx_t_7nvshmem_8bindings_7nvshmem_Stream;

/* "nvshmem/bindings/nvshmem.pxd":23
 * 
 * ctypedef cudaStream_t Stream
 * ctypedef CUmodule Module             # <<<<<<<<<<<<<<
 * ctypedef CUlibrary Library
 * 
*/
typedef CUmodule __pyx_t_7nvshmem_8bindings_7nvshmem_Module;

/* "nvshmem/bindings/nvshmem.pxd":24
 * ctypedef cudaStream_t Stream
 * ctypedef CUmodule Module
 * ctypedef CUlibrary Library             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef CUlibrary __pyx_t_7nvshmem_8bindings_7nvshmem_Library;

/* "nvshmem/bindings/nvshmem.pxd":32
 * ###############################################################################
 * 
 * ctypedef nvshmemx_signal_op_t _Signal_op             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_tma_policy_t _Tma_policy
 * ctypedef nvshmemx_smem_amount_t _Smem_amount
*/
typedef nvshmemx_signal_op_t __pyx_t_7nvshmem_8bindings_7nvshmem__Signal_op;

/* "nvshmem/bindings/nvshmem.pxd":33
 * 
 * ctypedef nvshmemx_signal_op_t _Signal_op
 * ctypedef nvshmemx_tma_policy_t _Tma_policy             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_smem_amount_t _Smem_amount
 * ctypedef nvshmemx_cmp_type_t _Cmp_type
*/
typedef nvshmemx_tma_policy_t __pyx_t_7nvshmem_8bindings_7nvshmem__Tma_policy;

/* "nvshmem/bindings/nvshmem.pxd":34
 * ctypedef nvshmemx_signal_op_t _Signal_op
 * ctypedef nvshmemx_tma_policy_t _Tma_policy
 * ctypedef nvshmemx_smem_amount_t _Smem_amount             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_cmp_type_t _Cmp_type
 * ctypedef nvshmemx_thread_support_t _Thread_support
*/
typedef nvshmemx_smem_amount_t __pyx_t_7nvshmem_8bindings_7nvshmem__Smem_amount;

/* "nvshmem/bindings/nvshmem.pxd":35
 * ctypedef nvshmemx_tma_policy_t _Tma_policy
 * ctypedef nvshmemx_smem_amount_t _Smem_amount
 * ctypedef nvshmemx_cmp_type_t _Cmp_type             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_thread_support_t _Thread_support
 * ctypedef nvshmemx_proxy_status_t _Proxy_status
*/
typedef nvshmemx_cmp_type_t __pyx_t_7nvshmem_8bindings_7nvshmem__Cmp_type;

/* "nvshmem/bindings/nvshmem.pxd":36
 * ctypedef nvshmemx_smem_amount_t _Smem_amount
 * ctypedef nvshmemx_cmp_type_t _Cmp_type
 * ctypedef nvshmemx_thread_support_t _Thread_support             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_proxy_status_t _Proxy_status
 * ctypedef nvshmemx_init_status_t _Init_status
*/
typedef nvshmemx_thread_support_t __pyx_t_7nvshmem_8bindings_7nvshmem__Thread_support;

/* "nvshmem/bindings/nvshmem.pxd":37
 * ctypedef nvshmemx_cmp_type_t _Cmp_type
 * ctypedef nvshmemx_thread_support_t _Thread_support
 * ctypedef nvshmemx_proxy_status_t _Proxy_status             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_init_status_t _Init_status
 * ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index
*/
typedef nvshmemx_proxy_status_t __pyx_t_7nvshmem_8bindings_7nvshmem__Proxy_status;

/* "nvshmem/bindings/nvshmem.pxd":38
 * ctypedef nvshmemx_thread_support_t _Thread_support
 * ctypedef nvshmemx_proxy_status_t _Proxy_status
 * ctypedef nvshmemx_init_status_t _Init_status             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index
 * ctypedef nvshmem_pe_index_t _Pe_index
*/
typedef nvshmemx_init_status_t __pyx_t_7nvshmem_8bindings_7nvshmem__Init_status;

/* "nvshmem/bindings/nvshmem.pxd":39
 * ctypedef nvshmemx_proxy_status_t _Proxy_status
 * ctypedef nvshmemx_init_status_t _Init_status
 * ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_pe_index_t _Pe_index
 * ctypedef nvshmem_team_id_t _Team_id
*/
typedef nvshmemx_qp_handle_index_t __pyx_t_7nvshmem_8bindings_7nvshmem__Qp_handle_index;

/* "nvshmem/bindings/nvshmem.pxd":40
 * ctypedef nvshmemx_init_status_t _Init_status
 * ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index
 * ctypedef nvshmem_pe_index_t _Pe_index             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_team_id_t _Team_id
 * ctypedef nvshmemx_status _Status
*/
typedef nvshmem_pe_index_t __pyx_t_7nvshmem_8bindings_7nvshmem__Pe_index;

/* "nvshmem/bindings/nvshmem.pxd":41
 * ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index
 * ctypedef nvshmem_pe_index_t _Pe_index
 * ctypedef nvshmem_team_id_t _Team_id             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_status _Status
 * ctypedef flags _Flags
*/
typedef nvshmem_team_id_t __pyx_t_7nvshmem_8bindings_7nvshmem__Team_id;

/* "nvshmem/bindings/nvshmem.pxd":42
 * ctypedef nvshmem_pe_index_t _Pe_index
 * ctypedef nvshmem_team_id_t _Team_id
 * ctypedef nvshmemx_status _Status             # <<<<<<<<<<<<<<
 * ctypedef flags _Flags
 * 
*/
typedef nvshmemx_status __pyx_t_7nvshmem_8bindings_7nvshmem__Status;

/* "nvshmem/bindings/nvshmem.pxd":43
 * ctypedef nvshmem_team_id_t _Team_id
 * ctypedef nvshmemx_status _Status
 * ctypedef flags _Flags             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef flags __pyx_t_7nvshmem_8bindings_7nvshmem__Flags;

/* "nvshmem/bindings/nvshmem.pyx":33
 * 
 * 
 * cdef class uniqueid:             # <<<<<<<<<<<<<<
 *     """Empty-initialize an array of `nvshmemx_uniqueid_v1`.
 * 
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid {
  PyObject_HEAD
  PyObject *_data;
};


/* "nvshmem/bindings/nvshmem.pyx":150
 * 
 * 
 * cdef class UniqueId(uniqueid): pass             # <<<<<<<<<<<<<<
 * 
 * # Simple wrapper for nvshmemx_team_uniqueid_t typedef
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid __pyx_base;
};


/* "nvshmem/bindings/nvshmem.pyx":154
 * # Simple wrapper for nvshmemx_team_uniqueid_t typedef
 * # This is required because Cybind doesn't generate wrappers for classes that are raw typedefs.
 * cdef class team_uniqueid:             # <<<<<<<<<<<<<<
 *     """Simple wrapper for nvshmemx_team_uniqueid_t (typedef uint64_t)."""
 * 
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid {
  PyObject_HEAD
  PyObject *_data;
};


/* "nvshmem/bindings/nvshmem.pyx":187
 *         return f"team_uniqueid(value={self.value})"
 * 
 * cdef class TeamUniqueId(team_uniqueid): pass             # <<<<<<<<<<<<<<
 * 
 * # POD wrapper for nvshmemx_init_attr_t. cybind can't generate this automatically
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid __pyx_base;
};


/* "nvshmem/bindings/nvshmem.pyx":201
 * 
 * 
 * cdef class InitAttr:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr {
  PyObject_HEAD
  PyObject *_data;
};


/* "nvshmem/bindings/nvshmem.pyx":249
 * 
 * 
 * cdef class TeamConfig:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
*/
struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig {
  PyObject_HEAD
  PyObject *_data;
};

/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_size_t(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_size_t(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_size_t(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* PyNumber_Hex.proto */
#define __Pyx_PyNumber_Hex(obj) PyNumber_ToBase((obj), 16)

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIndexError_Check.proto */
#define __Pyx_PyExc_IndexError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_IndexError)

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* pyint_simplify.proto */
static CYTHON_INLINE int __Pyx_PyInt_FromNumber(PyObject **number_var, const char *argname, int accept_none);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ErrOccurredWithGIL.proto */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void);

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* FunctionExport.proto */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig);

/* GetApiDict.proto */
static PyObject *__Pyx_ApiExport_GetApiDict(void);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_8
#define __PYX_HAVE_RT_ImportType_proto_3_2_8
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_8(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_8(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_8 {
   __Pyx_ImportType_CheckSize_Error_3_2_8 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_8 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_8 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_8(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_8 check_size);
#endif

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_8(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* PyObjectCall2Args.proto (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectLookupSpecial.proto (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_signal_op_t(nvshmemx_signal_op_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_tma_policy_t(nvshmemx_tma_policy_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_smem_amount_t(nvshmemx_smem_amount_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_cmp_type_t(nvshmemx_cmp_type_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_thread_support_t(nvshmemx_thread_support_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_proxy_status_t(nvshmemx_proxy_status_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_init_status_t(nvshmemx_init_status_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(nvshmemx_qp_handle_index_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmem_pe_index_t(nvshmem_pe_index_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmem_team_id_t(nvshmem_team_id_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_status(nvshmemx_status value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_flags(flags value);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
#endif

/* Module declarations from "libc.stdint" */

/* Module declarations from "nvshmem.bindings.cynvshmem" */
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier)(nvshmem_team_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier_all)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_init_status)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_my_pe)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_n_pes)(void); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_info_get_version)(int *, int *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_vendor_get_version_info)(int *, int *, int *); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_malloc)(size_t); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_calloc)(size_t, size_t); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_align)(size_t, size_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_free)(void *); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_ptr)(void const *, int); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_mc_ptr)(nvshmem_team_t, void const *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_my_pe)(nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_n_pes)(nvshmem_team_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_get_config)(nvshmem_team_t, nvshmem_team_config_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_translate_pe)(nvshmem_team_t, int, nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_strided)(nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_get_uniqueid)(nvshmemx_team_uniqueid_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_init)(nvshmem_team_t *, nvshmem_team_config_t *, long, int, int); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_2d)(nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_destroy)(nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_alltoall_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_alltoall_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_alltoall_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_alltoall_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_alltoall_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_alltoall_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_alltoall_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_alltoall_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_alltoall_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_alltoall_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_alltoall_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_alltoall_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_alltoall_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_alltoall_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_alltoall_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_alltoall_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_alltoall_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_alltoall_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_alltoall_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_on_stream)(nvshmem_team_t, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_all_on_stream)(cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_sync_on_stream)(nvshmem_team_t, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_sync_all_on_stream)(cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_broadcast_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_broadcast_on_stream)(nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_broadcast_on_stream)(nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_broadcast_on_stream)(nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_broadcast_on_stream)(nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_broadcast_on_stream)(nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_broadcast_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_broadcast_on_stream)(nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_broadcast_on_stream)(nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_broadcast_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_broadcast_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_broadcast_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_broadcast_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_broadcast_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_broadcast_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_broadcast_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_broadcast_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_broadcast_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_broadcast_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_fcollect_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_fcollect_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_fcollect_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_fcollect_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_fcollect_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_fcollect_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_fcollect_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_fcollect_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_fcollect_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_fcollect_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_fcollect_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_fcollect_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_fcollect_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_fcollect_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_fcollect_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_fcollect_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_fcollect_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_fcollect_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_fcollect_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_init_attr)(unsigned int, nvshmemx_init_attr_t *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_finalize)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_uniqueid_args)(int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_mpi_comm_args)(void *, nvshmemx_init_attr_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_get_uniqueid)(nvshmemx_uniqueid_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_init)(CUmodule); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_finalize)(CUmodule); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_register_symmetric)(void *, size_t, int); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_unregister_symmetric)(void *, size_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_init)(CUlibrary); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_finalize)(CUlibrary); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_on_stream)(void *, void const *, size_t, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_signal_on_stream)(void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_getmem_on_stream)(void *, void const *, size_t, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_quiet_on_stream)(cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_op_on_stream)(uint64_t *, uint64_t, int, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_wait_until_on_stream)(uint64_t *, int, uint64_t, cudaStream_t); /*proto*/

/* Module declarations from "cython" */

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.memoryview" */

/* Module declarations from "nvshmem.bindings.nvshmem" */
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_barrier(int32_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all(int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_init_status(int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_my_pe(int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_n_pes(int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_info_get_version(intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_vendor_get_version_info(intptr_t, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_malloc(size_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_calloc(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_align(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_free(intptr_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_ptr(intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_mc_ptr(int32_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_my_pe(int32_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_n_pes(int32_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_team_get_config(int32_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_translate_pe(int32_t, int, int32_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_strided(int32_t, int, int, int, intptr_t, long, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_get_uniqueid(intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_init(intptr_t, intptr_t, long, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_2d(int32_t, int, intptr_t, long, intptr_t, intptr_t, long, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_team_destroy(int32_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_alltoall_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_barrier_on_stream(int32_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all_on_stream(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_sync_on_stream(int32_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_sync_all_on_stream(intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_broadcast_on_stream(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_fcollect_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reduce_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reducescatter_on_stream(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_init_attr(unsigned int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_finalize(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_uniqueid_args(int, int, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_mpi_comm_args(intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_get_uniqueid(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_init(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_finalize(intptr_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_register_symmetric(intptr_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_unregister_symmetric(intptr_t, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_init(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_finalize(intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_on_stream(intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_signal_on_stream(intptr_t, intptr_t, size_t, intptr_t, uint64_t, int, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_getmem_on_stream(intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_quiet_on_stream(intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_signal_op_on_stream(intptr_t, uint64_t, int, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_signal_wait_until_on_stream(intptr_t, int, uint64_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_check_status(int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_uniqueid__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *, PyObject *); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_UniqueId__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *, PyObject *); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_team_uniqueid__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *, PyObject *); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamUniqueId__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *, PyObject *); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_InitAttr__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *, PyObject *); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamConfig__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "nvshmem.bindings.nvshmem"
extern int __pyx_module_is_main_nvshmem__bindings__nvshmem;
int __pyx_module_is_main_nvshmem__bindings__nvshmem = 0;

/* Implementation of "nvshmem.bindings.nvshmem" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
static const char __pyx_k_data_2[] = "_data";
/* #### Code section: decls ### */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_2__repr__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_4__int__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_6__len__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_8__eq__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_10__getitem__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_12__setitem__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data(PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr(intptr_t __pyx_v_ptr, size_t __pyx_v_size, int __pyx_v_readonly); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_2__str__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_4__repr__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_status); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_2barrier(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_4barrier_all(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_6init_status(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8my_pe(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10n_pes(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12info_get_version(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_major, intptr_t __pyx_v_minor); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_14vendor_get_version_info(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_major, intptr_t __pyx_v_minor, intptr_t __pyx_v_patch); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_16malloc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_18calloc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_count, size_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_20align(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_count, size_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_22free(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_ptr); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_24ptr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, int __pyx_v_pe); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_26mc_ptr(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_ptr); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_28team_my_pe(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_30team_n_pes(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_32team_get_config(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_config); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_34team_translate_pe(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_src_team, int __pyx_v_src_pe, int32_t __pyx_v_dest_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_36team_split_strided(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_parent_team, int __pyx_v_pe_start, int __pyx_v_pe_stride, int __pyx_v_pe_size, intptr_t __pyx_v_config, long __pyx_v_config_mask, intptr_t __pyx_v_new_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_38team_get_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_uniqueid); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_40team_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_team, intptr_t __pyx_v_config, long __pyx_v_config_mask, int __pyx_v_npes, int __pyx_v_pe_idx_in_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_42team_split_2d(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_parent_team, int __pyx_v_xrange, intptr_t __pyx_v_xaxis_config, long __pyx_v_xaxis_mask, intptr_t __pyx_v_xaxis_team, intptr_t __pyx_v_yaxis_config, long __pyx_v_yaxis_mask, intptr_t __pyx_v_yaxis_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_44team_destroy(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_46bfloat16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_48half_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_50float_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_52double_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_54char_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_56short_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_58schar_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_60int_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_62long_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_64longlong_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_66int8_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_68int16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_70int32_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_72int64_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_74uint8_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_76uint16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_78uint32_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_80uint64_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_82size_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_84barrier_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_86barrier_all_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_88team_sync_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_90sync_all_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_92bfloat16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_94half_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_96float_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_98double_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_100char_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_102short_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_104schar_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_106int_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_108long_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_110longlong_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_112int8_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_114int16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_116int32_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_118int64_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_120uint8_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_122uint16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_124uint32_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_126uint64_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_128size_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_130bfloat16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_132half_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_134float_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_136double_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_138char_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_140short_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_142schar_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_144int_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_146long_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_148longlong_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_150int8_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_152int16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_154int32_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_156int64_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_158uint8_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_160uint16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_162uint32_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_164uint64_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_166size_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_168int8_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_170int16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_172int32_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_174int64_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_176uint8_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_178uint16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_180uint32_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_182uint64_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_184size_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_186char_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_188schar_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_190short_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_192int_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_194long_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_196longlong_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_198bfloat16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_200half_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_202float_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_204double_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_206int8_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_208int16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_210int32_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_212int64_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_214uint8_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_216uint16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_218uint32_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_220uint64_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_222size_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_224char_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_226schar_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_228short_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_230int_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_232long_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_234longlong_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_236bfloat16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_238half_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_240float_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_242double_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_244int8_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_246int16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_248int32_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_250int64_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_252uint8_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_254uint16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_256uint32_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_258uint64_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_260size_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_262char_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_264schar_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_266short_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_268int_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_270long_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_272longlong_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_274bfloat16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_276half_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_278float_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_280double_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_282int8_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_284int16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_286int32_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_288int64_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_290uint8_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_292uint16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_294uint32_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_296uint64_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_298size_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_300char_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_302schar_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_304short_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_306int_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_308long_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_310longlong_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_312bfloat16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_314half_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_316float_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_318double_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_320int8_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_322int16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_324int32_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_326int64_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_328uint8_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_330uint16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_332uint32_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_334uint64_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_336size_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_338char_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_340schar_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_342short_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_344int_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_346long_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_348longlong_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_350bfloat16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_352half_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_354float_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_356double_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_358int8_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_360int16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_362int32_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_364int64_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_366uint8_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_368uint16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_370uint32_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_372uint64_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_374size_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_376char_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_378schar_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_380short_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_382int_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_384long_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_386longlong_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_388bfloat16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_390half_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_392float_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_394double_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_396hostlib_init_attr(CYTHON_UNUSED PyObject *__pyx_self, unsigned int __pyx_v_flags, intptr_t __pyx_v_attr); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_398hostlib_finalize(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_400set_attr_uniqueid_args(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_myrank, int __pyx_v_nranks, intptr_t __pyx_v_uniqueid, intptr_t __pyx_v_attr); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_402set_attr_mpi_comm_args(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_mpi_comm, intptr_t __pyx_v_nvshmem_attr); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_404get_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_uniqueid); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_406cumodule_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_module); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_408cumodule_finalize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_module); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_410buffer_register_symmetric(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_buf_ptr, size_t __pyx_v_size, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_412buffer_unregister_symmetric(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_mmap_ptr, size_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_414culibrary_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_library); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_416culibrary_finalize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_library); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_418putmem_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_420putmem_signal_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_422getmem_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_424quiet_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_cstrm); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_426signal_op_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_428signal_wait_until_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_sig_addr, int __pyx_v_cmp, uint64_t __pyx_v_cmp_value, intptr_t __pyx_v_cstream); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_430__pyx_unpickle_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_432__pyx_unpickle_UniqueId(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_434__pyx_unpickle_team_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_436__pyx_unpickle_TeamUniqueId(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_438__pyx_unpickle_InitAttr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_440__pyx_unpickle_TeamConfig(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_InitAttr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_TeamConfig(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr;
  PyObject *__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr;
  PyTypeObject *__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_tuple[3];
  PyObject *__pyx_codeobj_tab[237];
  PyObject *__pyx_string_tab[598];
  PyObject *__pyx_number_tab[5];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_None __pyx_string_tab[1]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[2]
#define __pyx_kp_u_See_flags __pyx_string_tab[3]
#define __pyx_kp_u_See_nvshmem_pe_index_t __pyx_string_tab[4]
#define __pyx_kp_u_See_nvshmem_team_id_t __pyx_string_tab[5]
#define __pyx_kp_u_See_nvshmemx_cmp_type_t __pyx_string_tab[6]
#define __pyx_kp_u_See_nvshmemx_init_status_t __pyx_string_tab[7]
#define __pyx_kp_u_See_nvshmemx_proxy_status_t __pyx_string_tab[8]
#define __pyx_kp_u_See_nvshmemx_qp_handle_index_t __pyx_string_tab[9]
#define __pyx_kp_u_See_nvshmemx_signal_op_t __pyx_string_tab[10]
#define __pyx_kp_u_See_nvshmemx_smem_amount_t __pyx_string_tab[11]
#define __pyx_kp_u_See_nvshmemx_status __pyx_string_tab[12]
#define __pyx_kp_u_See_nvshmemx_thread_support_t __pyx_string_tab[13]
#define __pyx_kp_u_See_nvshmemx_tma_policy_t __pyx_string_tab[14]
#define __pyx_kp_u_Status_code __pyx_string_tab[15]
#define __pyx_kp_u__2 __pyx_string_tab[16]
#define __pyx_kp_u__3 __pyx_string_tab[17]
#define __pyx_kp_u__4 __pyx_string_tab[18]
#define __pyx_kp_u__5 __pyx_string_tab[19]
#define __pyx_kp_u_add_note __pyx_string_tab[20]
#define __pyx_kp_u_data_argument_must_be_a_NumPy_nd __pyx_string_tab[21]
#define __pyx_kp_u_data_array_must_be_1D __pyx_string_tab[22]
#define __pyx_kp_u_data_array_must_be_of_dtype_uniq __pyx_string_tab[23]
#define __pyx_kp_u_disable __pyx_string_tab[24]
#define __pyx_kp_u_enable __pyx_string_tab[25]
#define __pyx_kp_u_gc __pyx_string_tab[26]
#define __pyx_kp_u_index_is_out_of_bounds __pyx_string_tab[27]
#define __pyx_kp_u_int_argument_must_be_a_bytes_lik __pyx_string_tab[28]
#define __pyx_kp_u_isenabled __pyx_string_tab[29]
#define __pyx_kp_u_itemsize_2 __pyx_string_tab[30]
#define __pyx_kp_u_mismatches_struct_size __pyx_string_tab[31]
#define __pyx_kp_u_nvshmem_bindings_nvshmem_pyx __pyx_string_tab[32]
#define __pyx_kp_u_object_at __pyx_string_tab[33]
#define __pyx_kp_u_ptr_must_not_be_null_0 __pyx_string_tab[34]
#define __pyx_kp_u_team_uniqueid __pyx_string_tab[35]
#define __pyx_kp_u_team_uniqueid_value __pyx_string_tab[36]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[37]
#define __pyx_kp_u_uniqueid_Array __pyx_string_tab[38]
#define __pyx_kp_u_uniqueid_object_at __pyx_string_tab[39]
#define __pyx_n_u_CMP_EQ __pyx_string_tab[40]
#define __pyx_n_u_CMP_GE __pyx_string_tab[41]
#define __pyx_n_u_CMP_GT __pyx_string_tab[42]
#define __pyx_n_u_CMP_LE __pyx_string_tab[43]
#define __pyx_n_u_CMP_LT __pyx_string_tab[44]
#define __pyx_n_u_CMP_NE __pyx_string_tab[45]
#define __pyx_n_u_CMP_SENTINEL __pyx_string_tab[46]
#define __pyx_n_u_Cmp_type __pyx_string_tab[47]
#define __pyx_n_u_ERROR_COLLECTIVE_LAUNCH_FAILED __pyx_string_tab[48]
#define __pyx_n_u_ERROR_GPU_NOT_SELECTED __pyx_string_tab[49]
#define __pyx_n_u_ERROR_INTERNAL __pyx_string_tab[50]
#define __pyx_n_u_ERROR_INVALID_VALUE __pyx_string_tab[51]
#define __pyx_n_u_ERROR_NOT_SUPPORTED __pyx_string_tab[52]
#define __pyx_n_u_ERROR_OUT_OF_MEMORY __pyx_string_tab[53]
#define __pyx_n_u_ERROR_SENTINEL __pyx_string_tab[54]
#define __pyx_n_u_ERROR_SYMMETRY __pyx_string_tab[55]
#define __pyx_n_u_Flags __pyx_string_tab[56]
#define __pyx_n_u_INIT_MAX __pyx_string_tab[57]
#define __pyx_n_u_INIT_THREAD_PES __pyx_string_tab[58]
#define __pyx_n_u_INIT_WITH_MPI_COMM __pyx_string_tab[59]
#define __pyx_n_u_INIT_WITH_SHMEM __pyx_string_tab[60]
#define __pyx_n_u_INIT_WITH_UNIQUEID __pyx_string_tab[61]
#define __pyx_n_u_InitAttr __pyx_string_tab[62]
#define __pyx_n_u_InitAttr___reduce_cython __pyx_string_tab[63]
#define __pyx_n_u_InitAttr___setstate_cython __pyx_string_tab[64]
#define __pyx_n_u_Init_status __pyx_string_tab[65]
#define __pyx_n_u_IntEnum __pyx_string_tab[66]
#define __pyx_n_u_IntEnum_2 __pyx_string_tab[67]
#define __pyx_n_u_MAX __pyx_string_tab[68]
#define __pyx_n_u_NVSHMEMError __pyx_string_tab[69]
#define __pyx_n_u_NVSHMEMError___init __pyx_string_tab[70]
#define __pyx_n_u_NVSHMEMError___reduce __pyx_string_tab[71]
#define __pyx_n_u_PE_ALL __pyx_string_tab[72]
#define __pyx_n_u_PE_ANY __pyx_string_tab[73]
#define __pyx_n_u_PE_INVALID __pyx_string_tab[74]
#define __pyx_n_u_PROXY_GLOBAL_EXIT_FINISHED __pyx_string_tab[75]
#define __pyx_n_u_PROXY_GLOBAL_EXIT_INIT __pyx_string_tab[76]
#define __pyx_n_u_PROXY_GLOBAL_EXIT_MAX_STATE __pyx_string_tab[77]
#define __pyx_n_u_PROXY_GLOBAL_EXIT_NOT_REQUESTED __pyx_string_tab[78]
#define __pyx_n_u_PROXY_GLOBAL_EXIT_REQUESTED __pyx_string_tab[79]
#define __pyx_n_u_Pe_index __pyx_string_tab[80]
#define __pyx_n_u_Proxy_status __pyx_string_tab[81]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[82]
#define __pyx_n_u_QP_ALL __pyx_string_tab[83]
#define __pyx_n_u_QP_ANY __pyx_string_tab[84]
#define __pyx_n_u_QP_DEFAULT __pyx_string_tab[85]
#define __pyx_n_u_QP_HOST __pyx_string_tab[86]
#define __pyx_n_u_Qp_handle_index __pyx_string_tab[87]
#define __pyx_n_u_SIGNAL_ADD __pyx_string_tab[88]
#define __pyx_n_u_SIGNAL_SET __pyx_string_tab[89]
#define __pyx_n_u_SMEM_BARRIERS_ONLY __pyx_string_tab[90]
#define __pyx_n_u_SMEM_MINIMUM __pyx_string_tab[91]
#define __pyx_n_u_SMEM_RECOMMENDED __pyx_string_tab[92]
#define __pyx_n_u_STATUS_FULL_MPG __pyx_string_tab[93]
#define __pyx_n_u_STATUS_INVALID __pyx_string_tab[94]
#define __pyx_n_u_STATUS_IS_BOOTSTRAPPED __pyx_string_tab[95]
#define __pyx_n_u_STATUS_IS_INITIALIZED __pyx_string_tab[96]
#define __pyx_n_u_STATUS_LIMITED_MPG __pyx_string_tab[97]
#define __pyx_n_u_STATUS_NOT_INITIALIZED __pyx_string_tab[98]
#define __pyx_n_u_SUCCESS __pyx_string_tab[99]
#define __pyx_n_u_Signal_op __pyx_string_tab[100]
#define __pyx_n_u_Smem_amount __pyx_string_tab[101]
#define __pyx_n_u_Status __pyx_string_tab[102]
#define __pyx_n_u_TEAMS_MIN __pyx_string_tab[103]
#define __pyx_n_u_TEAM_GPU_LEADERS __pyx_string_tab[104]
#define __pyx_n_u_TEAM_GPU_LEADERS_INDEX __pyx_string_tab[105]
#define __pyx_n_u_TEAM_INDEX_MAX __pyx_string_tab[106]
#define __pyx_n_u_TEAM_INVALID __pyx_string_tab[107]
#define __pyx_n_u_TEAM_MC_SHARED __pyx_string_tab[108]
#define __pyx_n_u_TEAM_MC_SHARED_INDEX __pyx_string_tab[109]
#define __pyx_n_u_TEAM_NODE __pyx_string_tab[110]
#define __pyx_n_u_TEAM_NODE_INDEX __pyx_string_tab[111]
#define __pyx_n_u_TEAM_SAME_GPU __pyx_string_tab[112]
#define __pyx_n_u_TEAM_SAME_GPU_INDEX __pyx_string_tab[113]
#define __pyx_n_u_TEAM_SAME_MYPE_NODE __pyx_string_tab[114]
#define __pyx_n_u_TEAM_SAME_MYPE_NODE_INDEX __pyx_string_tab[115]
#define __pyx_n_u_TEAM_SHARED __pyx_string_tab[116]
#define __pyx_n_u_TEAM_SHARED_INDEX __pyx_string_tab[117]
#define __pyx_n_u_TEAM_WORLD __pyx_string_tab[118]
#define __pyx_n_u_TEAM_WORLD_INDEX __pyx_string_tab[119]
#define __pyx_n_u_THREAD_FUNNELED __pyx_string_tab[120]
#define __pyx_n_u_THREAD_MULTIPLE __pyx_string_tab[121]
#define __pyx_n_u_THREAD_SERIALIZED __pyx_string_tab[122]
#define __pyx_n_u_THREAD_SINGLE __pyx_string_tab[123]
#define __pyx_n_u_THREAD_TYPE_SENTINEL __pyx_string_tab[124]
#define __pyx_n_u_TMA_DISABLE __pyx_string_tab[125]
#define __pyx_n_u_TMA_ENABLE __pyx_string_tab[126]
#define __pyx_n_u_TMA_FORCE __pyx_string_tab[127]
#define __pyx_n_u_TeamConfig __pyx_string_tab[128]
#define __pyx_n_u_TeamConfig___reduce_cython __pyx_string_tab[129]
#define __pyx_n_u_TeamConfig___setstate_cython __pyx_string_tab[130]
#define __pyx_n_u_TeamUniqueId __pyx_string_tab[131]
#define __pyx_n_u_TeamUniqueId___reduce_cython __pyx_string_tab[132]
#define __pyx_n_u_TeamUniqueId___setstate_cython __pyx_string_tab[133]
#define __pyx_n_u_Team_id __pyx_string_tab[134]
#define __pyx_n_u_Thread_support __pyx_string_tab[135]
#define __pyx_n_u_Tma_policy __pyx_string_tab[136]
#define __pyx_n_u_UniqueId __pyx_string_tab[137]
#define __pyx_n_u_UniqueId___reduce_cython __pyx_string_tab[138]
#define __pyx_n_u_UniqueId___setstate_cython __pyx_string_tab[139]
#define __pyx_n_u_align __pyx_string_tab[140]
#define __pyx_n_u_all __pyx_string_tab[141]
#define __pyx_n_u_annotate __pyx_string_tab[142]
#define __pyx_n_u_args __pyx_string_tab[143]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[144]
#define __pyx_n_u_attr __pyx_string_tab[145]
#define __pyx_n_u_barrier __pyx_string_tab[146]
#define __pyx_n_u_barrier_all __pyx_string_tab[147]
#define __pyx_n_u_barrier_all_on_stream __pyx_string_tab[148]
#define __pyx_n_u_barrier_on_stream __pyx_string_tab[149]
#define __pyx_n_u_bfloat16_alltoall_on_stream __pyx_string_tab[150]
#define __pyx_n_u_bfloat16_broadcast_on_stream __pyx_string_tab[151]
#define __pyx_n_u_bfloat16_fcollect_on_stream __pyx_string_tab[152]
#define __pyx_n_u_bfloat16_max_reduce_on_stream __pyx_string_tab[153]
#define __pyx_n_u_bfloat16_max_reducescatter_on_st __pyx_string_tab[154]
#define __pyx_n_u_bfloat16_min_reduce_on_stream __pyx_string_tab[155]
#define __pyx_n_u_bfloat16_min_reducescatter_on_st __pyx_string_tab[156]
#define __pyx_n_u_bfloat16_sum_reduce_on_stream __pyx_string_tab[157]
#define __pyx_n_u_bfloat16_sum_reducescatter_on_st __pyx_string_tab[158]
#define __pyx_n_u_buf __pyx_string_tab[159]
#define __pyx_n_u_buf_ptr __pyx_string_tab[160]
#define __pyx_n_u_buffer __pyx_string_tab[161]
#define __pyx_n_u_buffer_register_symmetric __pyx_string_tab[162]
#define __pyx_n_u_buffer_unregister_symmetric __pyx_string_tab[163]
#define __pyx_n_u_bytes __pyx_string_tab[164]
#define __pyx_n_u_calloc __pyx_string_tab[165]
#define __pyx_n_u_char_alltoall_on_stream __pyx_string_tab[166]
#define __pyx_n_u_char_broadcast_on_stream __pyx_string_tab[167]
#define __pyx_n_u_char_fcollect_on_stream __pyx_string_tab[168]
#define __pyx_n_u_char_max_reduce_on_stream __pyx_string_tab[169]
#define __pyx_n_u_char_max_reducescatter_on_stream __pyx_string_tab[170]
#define __pyx_n_u_char_min_reduce_on_stream __pyx_string_tab[171]
#define __pyx_n_u_char_min_reducescatter_on_stream __pyx_string_tab[172]
#define __pyx_n_u_char_sum_reduce_on_stream __pyx_string_tab[173]
#define __pyx_n_u_char_sum_reducescatter_on_stream __pyx_string_tab[174]
#define __pyx_n_u_check_status __pyx_string_tab[175]
#define __pyx_n_u_class_getitem __pyx_string_tab[176]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[177]
#define __pyx_n_u_cmp __pyx_string_tab[178]
#define __pyx_n_u_cmp_value __pyx_string_tab[179]
#define __pyx_n_u_config __pyx_string_tab[180]
#define __pyx_n_u_config_mask __pyx_string_tab[181]
#define __pyx_n_u_count __pyx_string_tab[182]
#define __pyx_n_u_cstream __pyx_string_tab[183]
#define __pyx_n_u_cstrm __pyx_string_tab[184]
#define __pyx_n_u_ctypes __pyx_string_tab[185]
#define __pyx_n_u_culibrary_finalize __pyx_string_tab[186]
#define __pyx_n_u_culibrary_init __pyx_string_tab[187]
#define __pyx_n_u_cumodule_finalize __pyx_string_tab[188]
#define __pyx_n_u_cumodule_init __pyx_string_tab[189]
#define __pyx_n_u_data __pyx_string_tab[190]
#define __pyx_n_u_data_2 __pyx_string_tab[191]
#define __pyx_n_u_dest __pyx_string_tab[192]
#define __pyx_n_u_dest_team __pyx_string_tab[193]
#define __pyx_n_u_dict __pyx_string_tab[194]
#define __pyx_n_u_dict_2 __pyx_string_tab[195]
#define __pyx_n_u_doc __pyx_string_tab[196]
#define __pyx_n_u_double_alltoall_on_stream __pyx_string_tab[197]
#define __pyx_n_u_double_broadcast_on_stream __pyx_string_tab[198]
#define __pyx_n_u_double_fcollect_on_stream __pyx_string_tab[199]
#define __pyx_n_u_double_max_reduce_on_stream __pyx_string_tab[200]
#define __pyx_n_u_double_max_reducescatter_on_stre __pyx_string_tab[201]
#define __pyx_n_u_double_min_reduce_on_stream __pyx_string_tab[202]
#define __pyx_n_u_double_min_reducescatter_on_stre __pyx_string_tab[203]
#define __pyx_n_u_double_sum_reduce_on_stream __pyx_string_tab[204]
#define __pyx_n_u_double_sum_reducescatter_on_stre __pyx_string_tab[205]
#define __pyx_n_u_dtype __pyx_string_tab[206]
#define __pyx_n_u_empty __pyx_string_tab[207]
#define __pyx_n_u_enum __pyx_string_tab[208]
#define __pyx_n_u_err __pyx_string_tab[209]
#define __pyx_n_u_flag __pyx_string_tab[210]
#define __pyx_n_u_flags __pyx_string_tab[211]
#define __pyx_n_u_float_alltoall_on_stream __pyx_string_tab[212]
#define __pyx_n_u_float_broadcast_on_stream __pyx_string_tab[213]
#define __pyx_n_u_float_fcollect_on_stream __pyx_string_tab[214]
#define __pyx_n_u_float_max_reduce_on_stream __pyx_string_tab[215]
#define __pyx_n_u_float_max_reducescatter_on_strea __pyx_string_tab[216]
#define __pyx_n_u_float_min_reduce_on_stream __pyx_string_tab[217]
#define __pyx_n_u_float_min_reducescatter_on_strea __pyx_string_tab[218]
#define __pyx_n_u_float_sum_reduce_on_stream __pyx_string_tab[219]
#define __pyx_n_u_float_sum_reducescatter_on_strea __pyx_string_tab[220]
#define __pyx_n_u_free __pyx_string_tab[221]
#define __pyx_n_u_from_data __pyx_string_tab[222]
#define __pyx_n_u_from_ptr __pyx_string_tab[223]
#define __pyx_n_u_func __pyx_string_tab[224]
#define __pyx_n_u_get_uniqueid __pyx_string_tab[225]
#define __pyx_n_u_getmem_on_stream __pyx_string_tab[226]
#define __pyx_n_u_getstate __pyx_string_tab[227]
#define __pyx_n_u_half_alltoall_on_stream __pyx_string_tab[228]
#define __pyx_n_u_half_broadcast_on_stream __pyx_string_tab[229]
#define __pyx_n_u_half_fcollect_on_stream __pyx_string_tab[230]
#define __pyx_n_u_half_max_reduce_on_stream __pyx_string_tab[231]
#define __pyx_n_u_half_max_reducescatter_on_stream __pyx_string_tab[232]
#define __pyx_n_u_half_min_reduce_on_stream __pyx_string_tab[233]
#define __pyx_n_u_half_min_reducescatter_on_stream __pyx_string_tab[234]
#define __pyx_n_u_half_sum_reduce_on_stream __pyx_string_tab[235]
#define __pyx_n_u_half_sum_reducescatter_on_stream __pyx_string_tab[236]
#define __pyx_n_u_hostlib_finalize __pyx_string_tab[237]
#define __pyx_n_u_hostlib_init_attr __pyx_string_tab[238]
#define __pyx_n_u_id __pyx_string_tab[239]
#define __pyx_n_u_info_get_version __pyx_string_tab[240]
#define __pyx_n_u_init __pyx_string_tab[241]
#define __pyx_n_u_init_attr_dtype __pyx_string_tab[242]
#define __pyx_n_u_init_status __pyx_string_tab[243]
#define __pyx_n_u_int16_alltoall_on_stream __pyx_string_tab[244]
#define __pyx_n_u_int16_broadcast_on_stream __pyx_string_tab[245]
#define __pyx_n_u_int16_fcollect_on_stream __pyx_string_tab[246]
#define __pyx_n_u_int16_max_reduce_on_stream __pyx_string_tab[247]
#define __pyx_n_u_int16_max_reducescatter_on_strea __pyx_string_tab[248]
#define __pyx_n_u_int16_min_reduce_on_stream __pyx_string_tab[249]
#define __pyx_n_u_int16_min_reducescatter_on_strea __pyx_string_tab[250]
#define __pyx_n_u_int16_sum_reduce_on_stream __pyx_string_tab[251]
#define __pyx_n_u_int16_sum_reducescatter_on_strea __pyx_string_tab[252]
#define __pyx_n_u_int32 __pyx_string_tab[253]
#define __pyx_n_u_int32_alltoall_on_stream __pyx_string_tab[254]
#define __pyx_n_u_int32_broadcast_on_stream __pyx_string_tab[255]
#define __pyx_n_u_int32_fcollect_on_stream __pyx_string_tab[256]
#define __pyx_n_u_int32_max_reduce_on_stream __pyx_string_tab[257]
#define __pyx_n_u_int32_max_reducescatter_on_strea __pyx_string_tab[258]
#define __pyx_n_u_int32_min_reduce_on_stream __pyx_string_tab[259]
#define __pyx_n_u_int32_min_reducescatter_on_strea __pyx_string_tab[260]
#define __pyx_n_u_int32_sum_reduce_on_stream __pyx_string_tab[261]
#define __pyx_n_u_int32_sum_reducescatter_on_strea __pyx_string_tab[262]
#define __pyx_n_u_int64_alltoall_on_stream __pyx_string_tab[263]
#define __pyx_n_u_int64_broadcast_on_stream __pyx_string_tab[264]
#define __pyx_n_u_int64_fcollect_on_stream __pyx_string_tab[265]
#define __pyx_n_u_int64_max_reduce_on_stream __pyx_string_tab[266]
#define __pyx_n_u_int64_max_reducescatter_on_strea __pyx_string_tab[267]
#define __pyx_n_u_int64_min_reduce_on_stream __pyx_string_tab[268]
#define __pyx_n_u_int64_min_reducescatter_on_strea __pyx_string_tab[269]
#define __pyx_n_u_int64_sum_reduce_on_stream __pyx_string_tab[270]
#define __pyx_n_u_int64_sum_reducescatter_on_strea __pyx_string_tab[271]
#define __pyx_n_u_int8 __pyx_string_tab[272]
#define __pyx_n_u_int8_alltoall_on_stream __pyx_string_tab[273]
#define __pyx_n_u_int8_broadcast_on_stream __pyx_string_tab[274]
#define __pyx_n_u_int8_fcollect_on_stream __pyx_string_tab[275]
#define __pyx_n_u_int8_max_reduce_on_stream __pyx_string_tab[276]
#define __pyx_n_u_int8_max_reducescatter_on_stream __pyx_string_tab[277]
#define __pyx_n_u_int8_min_reduce_on_stream __pyx_string_tab[278]
#define __pyx_n_u_int8_min_reducescatter_on_stream __pyx_string_tab[279]
#define __pyx_n_u_int8_sum_reduce_on_stream __pyx_string_tab[280]
#define __pyx_n_u_int8_sum_reducescatter_on_stream __pyx_string_tab[281]
#define __pyx_n_u_int_alltoall_on_stream __pyx_string_tab[282]
#define __pyx_n_u_int_broadcast_on_stream __pyx_string_tab[283]
#define __pyx_n_u_int_fcollect_on_stream __pyx_string_tab[284]
#define __pyx_n_u_int_max_reduce_on_stream __pyx_string_tab[285]
#define __pyx_n_u_int_max_reducescatter_on_stream __pyx_string_tab[286]
#define __pyx_n_u_int_min_reduce_on_stream __pyx_string_tab[287]
#define __pyx_n_u_int_min_reducescatter_on_stream __pyx_string_tab[288]
#define __pyx_n_u_int_sum_reduce_on_stream __pyx_string_tab[289]
#define __pyx_n_u_int_sum_reducescatter_on_stream __pyx_string_tab[290]
#define __pyx_n_u_internal __pyx_string_tab[291]
#define __pyx_n_u_intp __pyx_string_tab[292]
#define __pyx_n_u_is_coroutine __pyx_string_tab[293]
#define __pyx_n_u_items __pyx_string_tab[294]
#define __pyx_n_u_itemsize __pyx_string_tab[295]
#define __pyx_n_u_library __pyx_string_tab[296]
#define __pyx_n_u_long_alltoall_on_stream __pyx_string_tab[297]
#define __pyx_n_u_long_broadcast_on_stream __pyx_string_tab[298]
#define __pyx_n_u_long_fcollect_on_stream __pyx_string_tab[299]
#define __pyx_n_u_long_max_reduce_on_stream __pyx_string_tab[300]
#define __pyx_n_u_long_max_reducescatter_on_stream __pyx_string_tab[301]
#define __pyx_n_u_long_min_reduce_on_stream __pyx_string_tab[302]
#define __pyx_n_u_long_min_reducescatter_on_stream __pyx_string_tab[303]
#define __pyx_n_u_long_sum_reduce_on_stream __pyx_string_tab[304]
#define __pyx_n_u_long_sum_reducescatter_on_stream __pyx_string_tab[305]
#define __pyx_n_u_longlong_alltoall_on_stream __pyx_string_tab[306]
#define __pyx_n_u_longlong_broadcast_on_stream __pyx_string_tab[307]
#define __pyx_n_u_longlong_fcollect_on_stream __pyx_string_tab[308]
#define __pyx_n_u_longlong_max_reduce_on_stream __pyx_string_tab[309]
#define __pyx_n_u_longlong_max_reducescatter_on_st __pyx_string_tab[310]
#define __pyx_n_u_longlong_min_reduce_on_stream __pyx_string_tab[311]
#define __pyx_n_u_longlong_min_reducescatter_on_st __pyx_string_tab[312]
#define __pyx_n_u_longlong_sum_reduce_on_stream __pyx_string_tab[313]
#define __pyx_n_u_longlong_sum_reducescatter_on_st __pyx_string_tab[314]
#define __pyx_n_u_main __pyx_string_tab[315]
#define __pyx_n_u_major __pyx_string_tab[316]
#define __pyx_n_u_malloc __pyx_string_tab[317]
#define __pyx_n_u_mc_ptr __pyx_string_tab[318]
#define __pyx_n_u_metaclass __pyx_string_tab[319]
#define __pyx_n_u_minor __pyx_string_tab[320]
#define __pyx_n_u_mmap_ptr __pyx_string_tab[321]
#define __pyx_n_u_module __pyx_string_tab[322]
#define __pyx_n_u_module_2 __pyx_string_tab[323]
#define __pyx_n_u_mpi_comm __pyx_string_tab[324]
#define __pyx_n_u_mro_entries __pyx_string_tab[325]
#define __pyx_n_u_my_pe __pyx_string_tab[326]
#define __pyx_n_u_myrank __pyx_string_tab[327]
#define __pyx_n_u_n_pes __pyx_string_tab[328]
#define __pyx_n_u_name __pyx_string_tab[329]
#define __pyx_n_u_ndarray __pyx_string_tab[330]
#define __pyx_n_u_ndim __pyx_string_tab[331]
#define __pyx_n_u_nelem __pyx_string_tab[332]
#define __pyx_n_u_new __pyx_string_tab[333]
#define __pyx_n_u_new_team __pyx_string_tab[334]
#define __pyx_n_u_npes __pyx_string_tab[335]
#define __pyx_n_u_nranks __pyx_string_tab[336]
#define __pyx_n_u_nreduce __pyx_string_tab[337]
#define __pyx_n_u_num_contexts __pyx_string_tab[338]
#define __pyx_n_u_numpy __pyx_string_tab[339]
#define __pyx_n_u_numpy_2 __pyx_string_tab[340]
#define __pyx_n_u_nvshmem_attr __pyx_string_tab[341]
#define __pyx_n_u_nvshmem_bindings_nvshmem __pyx_string_tab[342]
#define __pyx_n_u_obj __pyx_string_tab[343]
#define __pyx_n_u_padding __pyx_string_tab[344]
#define __pyx_n_u_parent_team __pyx_string_tab[345]
#define __pyx_n_u_patch __pyx_string_tab[346]
#define __pyx_n_u_pe __pyx_string_tab[347]
#define __pyx_n_u_pe_idx_in_team __pyx_string_tab[348]
#define __pyx_n_u_pe_root __pyx_string_tab[349]
#define __pyx_n_u_pe_size __pyx_string_tab[350]
#define __pyx_n_u_pe_start __pyx_string_tab[351]
#define __pyx_n_u_pe_stride __pyx_string_tab[352]
#define __pyx_n_u_pop __pyx_string_tab[353]
#define __pyx_n_u_prepare __pyx_string_tab[354]
#define __pyx_n_u_ptr __pyx_string_tab[355]
#define __pyx_n_u_putmem_on_stream __pyx_string_tab[356]
#define __pyx_n_u_putmem_signal_on_stream __pyx_string_tab[357]
#define __pyx_n_u_pyx_capi __pyx_string_tab[358]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[359]
#define __pyx_n_u_pyx_result __pyx_string_tab[360]
#define __pyx_n_u_pyx_state __pyx_string_tab[361]
#define __pyx_n_u_pyx_type __pyx_string_tab[362]
#define __pyx_n_u_pyx_unpickle_InitAttr __pyx_string_tab[363]
#define __pyx_n_u_pyx_unpickle_TeamConfig __pyx_string_tab[364]
#define __pyx_n_u_pyx_unpickle_TeamUniqueId __pyx_string_tab[365]
#define __pyx_n_u_pyx_unpickle_UniqueId __pyx_string_tab[366]
#define __pyx_n_u_pyx_unpickle_team_uniqueid __pyx_string_tab[367]
#define __pyx_n_u_pyx_unpickle_uniqueid __pyx_string_tab[368]
#define __pyx_n_u_qualname __pyx_string_tab[369]
#define __pyx_n_u_quiet_on_stream __pyx_string_tab[370]
#define __pyx_n_u_readonly __pyx_string_tab[371]
#define __pyx_n_u_recarray __pyx_string_tab[372]
#define __pyx_n_u_reduce __pyx_string_tab[373]
#define __pyx_n_u_reduce_cython __pyx_string_tab[374]
#define __pyx_n_u_reduce_ex __pyx_string_tab[375]
#define __pyx_n_u_schar_alltoall_on_stream __pyx_string_tab[376]
#define __pyx_n_u_schar_broadcast_on_stream __pyx_string_tab[377]
#define __pyx_n_u_schar_fcollect_on_stream __pyx_string_tab[378]
#define __pyx_n_u_schar_max_reduce_on_stream __pyx_string_tab[379]
#define __pyx_n_u_schar_max_reducescatter_on_strea __pyx_string_tab[380]
#define __pyx_n_u_schar_min_reduce_on_stream __pyx_string_tab[381]
#define __pyx_n_u_schar_min_reducescatter_on_strea __pyx_string_tab[382]
#define __pyx_n_u_schar_sum_reduce_on_stream __pyx_string_tab[383]
#define __pyx_n_u_schar_sum_reducescatter_on_strea __pyx_string_tab[384]
#define __pyx_n_u_self __pyx_string_tab[385]
#define __pyx_n_u_set_attr_mpi_comm_args __pyx_string_tab[386]
#define __pyx_n_u_set_attr_uniqueid_args __pyx_string_tab[387]
#define __pyx_n_u_set_name __pyx_string_tab[388]
#define __pyx_n_u_setdefault __pyx_string_tab[389]
#define __pyx_n_u_setstate __pyx_string_tab[390]
#define __pyx_n_u_setstate_cython __pyx_string_tab[391]
#define __pyx_n_u_short_alltoall_on_stream __pyx_string_tab[392]
#define __pyx_n_u_short_broadcast_on_stream __pyx_string_tab[393]
#define __pyx_n_u_short_fcollect_on_stream __pyx_string_tab[394]
#define __pyx_n_u_short_max_reduce_on_stream __pyx_string_tab[395]
#define __pyx_n_u_short_max_reducescatter_on_strea __pyx_string_tab[396]
#define __pyx_n_u_short_min_reduce_on_stream __pyx_string_tab[397]
#define __pyx_n_u_short_min_reducescatter_on_strea __pyx_string_tab[398]
#define __pyx_n_u_short_sum_reduce_on_stream __pyx_string_tab[399]
#define __pyx_n_u_short_sum_reducescatter_on_strea __pyx_string_tab[400]
#define __pyx_n_u_sig_addr __pyx_string_tab[401]
#define __pyx_n_u_sig_op __pyx_string_tab[402]
#define __pyx_n_u_signal __pyx_string_tab[403]
#define __pyx_n_u_signal_op_on_stream __pyx_string_tab[404]
#define __pyx_n_u_signal_wait_until_on_stream __pyx_string_tab[405]
#define __pyx_n_u_size __pyx_string_tab[406]
#define __pyx_n_u_size_alltoall_on_stream __pyx_string_tab[407]
#define __pyx_n_u_size_broadcast_on_stream __pyx_string_tab[408]
#define __pyx_n_u_size_fcollect_on_stream __pyx_string_tab[409]
#define __pyx_n_u_size_max_reduce_on_stream __pyx_string_tab[410]
#define __pyx_n_u_size_max_reducescatter_on_stream __pyx_string_tab[411]
#define __pyx_n_u_size_min_reduce_on_stream __pyx_string_tab[412]
#define __pyx_n_u_size_min_reducescatter_on_stream __pyx_string_tab[413]
#define __pyx_n_u_size_sum_reduce_on_stream __pyx_string_tab[414]
#define __pyx_n_u_size_sum_reducescatter_on_stream __pyx_string_tab[415]
#define __pyx_n_u_source __pyx_string_tab[416]
#define __pyx_n_u_src __pyx_string_tab[417]
#define __pyx_n_u_src_pe __pyx_string_tab[418]
#define __pyx_n_u_src_team __pyx_string_tab[419]
#define __pyx_n_u_state __pyx_string_tab[420]
#define __pyx_n_u_staticmethod __pyx_string_tab[421]
#define __pyx_n_u_status __pyx_string_tab[422]
#define __pyx_n_u_stream __pyx_string_tab[423]
#define __pyx_n_u_super __pyx_string_tab[424]
#define __pyx_n_u_sync_all_on_stream __pyx_string_tab[425]
#define __pyx_n_u_team __pyx_string_tab[426]
#define __pyx_n_u_team_config_dtype __pyx_string_tab[427]
#define __pyx_n_u_team_destroy __pyx_string_tab[428]
#define __pyx_n_u_team_get_config __pyx_string_tab[429]
#define __pyx_n_u_team_get_uniqueid __pyx_string_tab[430]
#define __pyx_n_u_team_init __pyx_string_tab[431]
#define __pyx_n_u_team_my_pe __pyx_string_tab[432]
#define __pyx_n_u_team_n_pes __pyx_string_tab[433]
#define __pyx_n_u_team_split_2d __pyx_string_tab[434]
#define __pyx_n_u_team_split_strided __pyx_string_tab[435]
#define __pyx_n_u_team_sync_on_stream __pyx_string_tab[436]
#define __pyx_n_u_team_translate_pe __pyx_string_tab[437]
#define __pyx_n_u_team_uniqueid_2 __pyx_string_tab[438]
#define __pyx_n_u_team_uniqueid___reduce_cython __pyx_string_tab[439]
#define __pyx_n_u_team_uniqueid___setstate_cython __pyx_string_tab[440]
#define __pyx_n_u_test __pyx_string_tab[441]
#define __pyx_n_u_uint16_alltoall_on_stream __pyx_string_tab[442]
#define __pyx_n_u_uint16_broadcast_on_stream __pyx_string_tab[443]
#define __pyx_n_u_uint16_fcollect_on_stream __pyx_string_tab[444]
#define __pyx_n_u_uint16_max_reduce_on_stream __pyx_string_tab[445]
#define __pyx_n_u_uint16_max_reducescatter_on_stre __pyx_string_tab[446]
#define __pyx_n_u_uint16_min_reduce_on_stream __pyx_string_tab[447]
#define __pyx_n_u_uint16_min_reducescatter_on_stre __pyx_string_tab[448]
#define __pyx_n_u_uint16_sum_reduce_on_stream __pyx_string_tab[449]
#define __pyx_n_u_uint16_sum_reducescatter_on_stre __pyx_string_tab[450]
#define __pyx_n_u_uint32_alltoall_on_stream __pyx_string_tab[451]
#define __pyx_n_u_uint32_broadcast_on_stream __pyx_string_tab[452]
#define __pyx_n_u_uint32_fcollect_on_stream __pyx_string_tab[453]
#define __pyx_n_u_uint32_max_reduce_on_stream __pyx_string_tab[454]
#define __pyx_n_u_uint32_max_reducescatter_on_stre __pyx_string_tab[455]
#define __pyx_n_u_uint32_min_reduce_on_stream __pyx_string_tab[456]
#define __pyx_n_u_uint32_min_reducescatter_on_stre __pyx_string_tab[457]
#define __pyx_n_u_uint32_sum_reduce_on_stream __pyx_string_tab[458]
#define __pyx_n_u_uint32_sum_reducescatter_on_stre __pyx_string_tab[459]
#define __pyx_n_u_uint64 __pyx_string_tab[460]
#define __pyx_n_u_uint64_alltoall_on_stream __pyx_string_tab[461]
#define __pyx_n_u_uint64_broadcast_on_stream __pyx_string_tab[462]
#define __pyx_n_u_uint64_fcollect_on_stream __pyx_string_tab[463]
#define __pyx_n_u_uint64_max_reduce_on_stream __pyx_string_tab[464]
#define __pyx_n_u_uint64_max_reducescatter_on_stre __pyx_string_tab[465]
#define __pyx_n_u_uint64_min_reduce_on_stream __pyx_string_tab[466]
#define __pyx_n_u_uint64_min_reducescatter_on_stre __pyx_string_tab[467]
#define __pyx_n_u_uint64_sum_reduce_on_stream __pyx_string_tab[468]
#define __pyx_n_u_uint64_sum_reducescatter_on_stre __pyx_string_tab[469]
#define __pyx_n_u_uint8_alltoall_on_stream __pyx_string_tab[470]
#define __pyx_n_u_uint8_broadcast_on_stream __pyx_string_tab[471]
#define __pyx_n_u_uint8_fcollect_on_stream __pyx_string_tab[472]
#define __pyx_n_u_uint8_max_reduce_on_stream __pyx_string_tab[473]
#define __pyx_n_u_uint8_max_reducescatter_on_strea __pyx_string_tab[474]
#define __pyx_n_u_uint8_min_reduce_on_stream __pyx_string_tab[475]
#define __pyx_n_u_uint8_min_reducescatter_on_strea __pyx_string_tab[476]
#define __pyx_n_u_uint8_sum_reduce_on_stream __pyx_string_tab[477]
#define __pyx_n_u_uint8_sum_reducescatter_on_strea __pyx_string_tab[478]
#define __pyx_n_u_uniqueid __pyx_string_tab[479]
#define __pyx_n_u_uniqueid___reduce_cython __pyx_string_tab[480]
#define __pyx_n_u_uniqueid___setstate_cython __pyx_string_tab[481]
#define __pyx_n_u_uniqueid_dtype __pyx_string_tab[482]
#define __pyx_n_u_uniqueid_from_data __pyx_string_tab[483]
#define __pyx_n_u_uniqueid_from_ptr __pyx_string_tab[484]
#define __pyx_n_u_update __pyx_string_tab[485]
#define __pyx_n_u_use_setstate __pyx_string_tab[486]
#define __pyx_n_u_value __pyx_string_tab[487]
#define __pyx_n_u_values __pyx_string_tab[488]
#define __pyx_n_u_vendor_get_version_info __pyx_string_tab[489]
#define __pyx_n_u_version __pyx_string_tab[490]
#define __pyx_n_u_view __pyx_string_tab[491]
#define __pyx_n_u_xaxis_config __pyx_string_tab[492]
#define __pyx_n_u_xaxis_mask __pyx_string_tab[493]
#define __pyx_n_u_xaxis_team __pyx_string_tab[494]
#define __pyx_n_u_xrange __pyx_string_tab[495]
#define __pyx_n_u_yaxis_config __pyx_string_tab[496]
#define __pyx_n_u_yaxis_mask __pyx_string_tab[497]
#define __pyx_n_u_yaxis_team __pyx_string_tab[498]
#define __pyx_kp_b_PyObject_int_int_intptr_t_intptr __pyx_string_tab[499]
#define __pyx_kp_b_int_CUlibrary_int_CUmodule_int_i __pyx_string_tab[500]
#define __pyx_kp_b_iso88591_0_1A_vVS__ddkksst __pyx_string_tab[501]
#define __pyx_kp_b_iso88591_0_BYYZ __pyx_string_tab[502]
#define __pyx_kp_b_iso88591_0_HKxWss_U_U_V __pyx_string_tab[503]
#define __pyx_kp_b_iso88591_1 __pyx_string_tab[504]
#define __pyx_kp_b_iso88591_1F __pyx_string_tab[505]
#define __pyx_kp_b_iso88591_1G3J __pyx_string_tab[506]
#define __pyx_kp_b_iso88591_1_1A __pyx_string_tab[507]
#define __pyx_kp_b_iso88591_1_2 __pyx_string_tab[508]
#define __pyx_kp_b_iso88591_1_2B_QWWffkkrrzz __pyx_string_tab[509]
#define __pyx_kp_b_iso88591_1_2B_fT_eelluu __pyx_string_tab[510]
#define __pyx_kp_b_iso88591_1_2B_vUbbggnnvvw __pyx_string_tab[511]
#define __pyx_kp_b_iso88591_1_5V_SS_hhnno __pyx_string_tab[512]
#define __pyx_kp_b_iso88591_1_F __pyx_string_tab[513]
#define __pyx_kp_b_iso88591_1_F2J __pyx_string_tab[514]
#define __pyx_kp_b_iso88591_1_Gq __pyx_string_tab[515]
#define __pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F __pyx_string_tab[516]
#define __pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J __pyx_string_tab[517]
#define __pyx_kp_b_iso88591_1_LFR_ffzz_I_I_Q_Q_R __pyx_string_tab[518]
#define __pyx_kp_b_iso88591_1_M_S_g_h_F_F_N_N_l_l_m __pyx_string_tab[519]
#define __pyx_kp_b_iso88591_2 __pyx_string_tab[520]
#define __pyx_kp_b_iso88591_2_3C6_PVVcchhoox_y_A_A_B __pyx_string_tab[521]
#define __pyx_kp_b_iso88591_2_3C6_QWWeejjqqyyz __pyx_string_tab[522]
#define __pyx_kp_b_iso88591_2_3C6_RXXggllss_E_E_F __pyx_string_tab[523]
#define __pyx_kp_b_iso88591_2_3C6_SYYiinnuu __pyx_string_tab[524]
#define __pyx_kp_b_iso88591_2_3C6_W_qqvv_F_F_G __pyx_string_tab[525]
#define __pyx_kp_b_iso88591_2_3C6_vUaaffoowwx __pyx_string_tab[526]
#define __pyx_kp_b_iso88591_2_fA __pyx_string_tab[527]
#define __pyx_kp_b_iso88591_314DF_QWWddiirrzz __pyx_string_tab[528]
#define __pyx_kp_b_iso88591_314DF_RXXffkkrr_D_D_E __pyx_string_tab[529]
#define __pyx_kp_b_iso88591_314DF_SYYhhmmtt __pyx_string_tab[530]
#define __pyx_kp_b_iso88591_314DF_SYYhhmmvv __pyx_string_tab[531]
#define __pyx_kp_b_iso88591_314DF_TZZjjoovv_H_H_I __pyx_string_tab[532]
#define __pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B __pyx_string_tab[533]
#define __pyx_kp_b_iso88591_314DF_X_rrww_H_H_P_P_Q __pyx_string_tab[534]
#define __pyx_kp_b_iso88591_4A5EV8SYYgglluu __pyx_string_tab[535]
#define __pyx_kp_b_iso88591_4A5EV9TZZiinnuu_G_G_H __pyx_string_tab[536]
#define __pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C __pyx_string_tab[537]
#define __pyx_kp_b_iso88591_4A5EV_V_mmrry_z_C_C_K_K_L __pyx_string_tab[538]
#define __pyx_kp_b_iso88591_4A5EV_Y__ssx_y_B_B_J_J_K __pyx_string_tab[539]
#define __pyx_kp_b_iso88591_4A_7 __pyx_string_tab[540]
#define __pyx_kp_b_iso88591_5Q6FfIU_jjoox_y_A_A_B __pyx_string_tab[541]
#define __pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F __pyx_string_tab[542]
#define __pyx_kp_b_iso88591_5Q6FfLX_ppuu_E_E_F __pyx_string_tab[543]
#define __pyx_kp_b_iso88591_5Q6FfL_bbxx_E_E_M_M_N __pyx_string_tab[544]
#define __pyx_kp_b_iso88591_6 __pyx_string_tab[545]
#define __pyx_kp_b_iso88591_6a7GvM_ccyy_F_F_O_O_W_W_X __pyx_string_tab[546]
#define __pyx_kp_b_iso88591_6a7Gv_Y__qqvv_G_G_O_O_P __pyx_string_tab[547]
#define __pyx_kp_b_iso88591_6a_QgQ __pyx_string_tab[548]
#define __pyx_kp_b_iso88591_7q8H_N_ddzz_I_I_Q_Q_R __pyx_string_tab[549]
#define __pyx_kp_b_iso88591_7q8H_lZ_rrw_x_A_A_I_I_J __pyx_string_tab[550]
#define __pyx_kp_b_iso88591_7q_y_a __pyx_string_tab[551]
#define __pyx_kp_b_iso88591_81 __pyx_string_tab[552]
#define __pyx_kp_b_iso88591_8_7 __pyx_string_tab[553]
#define __pyx_kp_b_iso88591_8_awa __pyx_string_tab[554]
#define __pyx_kp_b_iso88591_9_1G1 __pyx_string_tab[555]
#define __pyx_kp_b_iso88591_9_J_PVV_hhmmvv __pyx_string_tab[556]
#define __pyx_kp_b_iso88591_AV7 __pyx_string_tab[557]
#define __pyx_kp_b_iso88591_AV7_vQ __pyx_string_tab[558]
#define __pyx_kp_b_iso88591_AWF_xwkYcckkssww_A __pyx_string_tab[559]
#define __pyx_kp_b_iso88591_AXQ __pyx_string_tab[560]
#define __pyx_kp_b_iso88591_A_5V __pyx_string_tab[561]
#define __pyx_kp_b_iso88591_A_AXT __pyx_string_tab[562]
#define __pyx_kp_b_iso88591_A_A_4s_AQ_HHAQ_G_q_1_5_5RvQ_vXRx __pyx_string_tab[563]
#define __pyx_kp_b_iso88591_A_HHAQ_4z_z_q_1A_4vS_AQ_4wc_AQ_9 __pyx_string_tab[564]
#define __pyx_kp_b_iso88591_A_Ja_Qa_QnE_1 __pyx_string_tab[565]
#define __pyx_kp_b_iso88591_A_MVS_bbqqvv_H_H_I __pyx_string_tab[566]
#define __pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M __pyx_string_tab[567]
#define __pyx_kp_b_iso88591_A_Q_a __pyx_string_tab[568]
#define __pyx_kp_b_iso88591_EQ_q_8_ha __pyx_string_tab[569]
#define __pyx_kp_b_iso88591_Je_hVW __pyx_string_tab[570]
#define __pyx_kp_b_iso88591_Jhhd_RS __pyx_string_tab[571]
#define __pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C __pyx_string_tab[572]
#define __pyx_kp_b_iso88591_K6QZZ_oott_F_F_G __pyx_string_tab[573]
#define __pyx_kp_b_iso88591_Na_q_z __pyx_string_tab[574]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[575]
#define __pyx_kp_b_iso88591_QgV_t8ST __pyx_string_tab[576]
#define __pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_4q_4q __pyx_string_tab[577]
#define __pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_Qg_q_Qg __pyx_string_tab[578]
#define __pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_T_G1_T_A __pyx_string_tab[579]
#define __pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_d_7_WA_d_7 __pyx_string_tab[580]
#define __pyx_kp_b_iso88591_VVW_A_5ZxGWWX __pyx_string_tab[581]
#define __pyx_kp_b_iso88591__6 __pyx_string_tab[582]
#define __pyx_kp_b_iso88591__7 __pyx_string_tab[583]
#define __pyx_kp_b_iso88591_a_7_HLjjx_y_E_E_V_V_b_b_N_N_Z_Z __pyx_string_tab[584]
#define __pyx_kp_b_iso88591_a_OvUaaggyy_H_H_P_P_Q __pyx_string_tab[585]
#define __pyx_kp_b_iso88591_a_OvUeek_l_B_B_G_G_P_P_X_X_Y __pyx_string_tab[586]
#define __pyx_kp_b_iso88591_a_Q __pyx_string_tab[587]
#define __pyx_kp_b_iso88591_avQ __pyx_string_tab[588]
#define __pyx_kp_b_iso88591_q __pyx_string_tab[589]
#define __pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1 __pyx_string_tab[590]
#define __pyx_kp_b_iso88591_q_0_kQR_881A_7_nA_1 __pyx_string_tab[591]
#define __pyx_kp_b_iso88591_q_0_kQR_XQa_7_A_1 __pyx_string_tab[592]
#define __pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1 __pyx_string_tab[593]
#define __pyx_kp_b_iso88591_q_2 __pyx_string_tab[594]
#define __pyx_kp_b_iso88591_q_3 __pyx_string_tab[595]
#define __pyx_kp_b_iso88591_q_4 __pyx_string_tab[596]
#define __pyx_kp_b_iso88591_wc_l_1 __pyx_string_tab[597]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_48 __pyx_number_tab[2]
#define __pyx_int_124 __pyx_number_tab[3]
#define __pyx_int_175497610 __pyx_number_tab[4]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig);
  Py_CLEAR(clear_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig);
  for (int i=0; i<3; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<237; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<598; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<5; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig);
  Py_VISIT(traverse_module_state->__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig);
  for (int i=0; i<3; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<237; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<598; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<5; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "cpython/complex.pxd":20
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
*/

#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":23
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         # unavailable in limited API
*/
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":20
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/complex.pxd":26
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
*/

#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":29
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
*/
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":26
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/contextvars.pxd":115
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {

  /* "cpython/contextvars.pxd":116
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
*/
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":121
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
*/
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":122
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
*/
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 122, __pyx_L1_error)

  /* "cpython/contextvars.pxd":123
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
*/
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":125
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
*/
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":123
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
*/
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":128
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
*/
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":129
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
*/
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":130
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":115
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/contextvars.pxd":133
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value_no_default(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {

  /* "cpython/contextvars.pxd":134
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
*/
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":140
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
*/
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":141
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
*/
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 141, __pyx_L1_error)

  /* "cpython/contextvars.pxd":143
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":144
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
*/
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":145
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":133
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value_no_default(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "nvshmem/bindings/nvshmem.pyx":48
 *         readonly object _data
 * 
 *     def __init__(self, size=1):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(size, dtype=uniqueid_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_size = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 48, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 48, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 48, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 48, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    }
    __pyx_v_size = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 48, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid___init__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_v_arr = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8[4];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":49
 * 
 *     def __init__(self, size=1):
 *         arr = _numpy.empty(size, dtype=uniqueid_dtype)             # <<<<<<<<<<<<<<
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_uniqueid_v1), \
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_uniqueid_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_v_size};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, __pyx_t_3, __pyx_t_6, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 49, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":50
 *     def __init__(self, size=1):
 *         arr = _numpy.empty(size, dtype=uniqueid_dtype)
 *         self._data = arr.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 *         assert self._data.itemsize == sizeof(nvshmemx_uniqueid_v1), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_uniqueid_v1)}"
*/
  __pyx_t_4 = __pyx_v_arr;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_data);
  __Pyx_DECREF(__pyx_v_self->_data);
  __pyx_v_self->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":51
 *         arr = _numpy.empty(size, dtype=uniqueid_dtype)
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_uniqueid_v1), \             # <<<<<<<<<<<<<<
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_uniqueid_v1)}"
 * 
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyLong_FromSize_t((sizeof(nvshmemx_uniqueid_v1))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) {

      /* "nvshmem/bindings/nvshmem.pyx":52
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_uniqueid_v1), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_uniqueid_v1)}"             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_From_size_t((sizeof(nvshmemx_uniqueid_v1)), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8[0] = __pyx_mstate_global->__pyx_kp_u_itemsize_2;
      __pyx_t_8[1] = __pyx_t_3;
      __pyx_t_8[2] = __pyx_mstate_global->__pyx_kp_u_mismatches_struct_size;
      __pyx_t_8[3] = __pyx_t_4;
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_8, 4, 9 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 24 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3));
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_1, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 51, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 51, __pyx_L1_error)
  #endif

  /* "nvshmem/bindings/nvshmem.pyx":48
 *         readonly object _data
 * 
 *     def __init__(self, size=1):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(size, dtype=uniqueid_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":54
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_uniqueid_v1)}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self._data.size > 1:
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_2__repr__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_2__repr__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6[7];
  PyObject *__pyx_t_7[5];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":55
 * 
 *     def __repr__(self):
 *         if self._data.size > 1:             # <<<<<<<<<<<<<<
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"
 *         else:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "nvshmem/bindings/nvshmem.pyx":56
 *     def __repr__(self):
 *         if self._data.size > 1:
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"             # <<<<<<<<<<<<<<
 *         else:
 *             return f"<{__name__}.uniqueid object at {hex(id(self))}>"
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Hex(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyUnicode_Unicode(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6[0] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_6[1] = __pyx_t_1;
    __pyx_t_6[2] = __pyx_mstate_global->__pyx_kp_u_uniqueid_Array;
    __pyx_t_6[3] = __pyx_t_4;
    __pyx_t_6[4] = __pyx_mstate_global->__pyx_kp_u_object_at;
    __pyx_t_6[5] = __pyx_t_2;
    __pyx_t_6[6] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_6, 7, 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 16 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":55
 * 
 *     def __repr__(self):
 *         if self._data.size > 1:             # <<<<<<<<<<<<<<
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"
 *         else:
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":58
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"
 *         else:
 *             return f"<{__name__}.uniqueid object at {hex(id(self))}>"             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyNumber_Hex(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7[0] = __pyx_mstate_global->__pyx_kp_u_;
    __pyx_t_7[1] = __pyx_t_2;
    __pyx_t_7[2] = __pyx_mstate_global->__pyx_kp_u_uniqueid_object_at;
    __pyx_t_7[3] = __pyx_t_5;
    __pyx_t_7[4] = __pyx_mstate_global->__pyx_kp_u__2;
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_7, 5, 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 20 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "nvshmem/bindings/nvshmem.pyx":54
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_uniqueid_v1)}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self._data.size > 1:
 *             return f"<{__name__}.uniqueid_Array_{self._data.size} object at {hex(id(self))}>"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":60
 *             return f"<{__name__}.uniqueid object at {hex(id(self))}>"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":63
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
 *         return self._data.ctypes.data             # <<<<<<<<<<<<<<
 * 
 *     def __int__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":60
 *             return f"<{__name__}.uniqueid object at {hex(id(self))}>"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.ptr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":65
 *         return self._data.ctypes.data
 * 
 *     def __int__(self):             # <<<<<<<<<<<<<<
 *         if self._data.size > 1:
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5__int__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5__int__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__int__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_4__int__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_4__int__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__int__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":66
 * 
 *     def __int__(self):
 *         if self._data.size > 1:             # <<<<<<<<<<<<<<
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "
 *                             "To get the pointer address of an array, use .ptr")
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "nvshmem/bindings/nvshmem.pyx":67
 *     def __int__(self):
 *         if self._data.size > 1:
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "             # <<<<<<<<<<<<<<
 *                             "To get the pointer address of an array, use .ptr")
 *         return self._data.ctypes.data
*/
    __pyx_t_1 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_int_argument_must_be_a_bytes_lik};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 67, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":66
 * 
 *     def __int__(self):
 *         if self._data.size > 1:             # <<<<<<<<<<<<<<
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "
 *                             "To get the pointer address of an array, use .ptr")
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":69
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "
 *                             "To get the pointer address of an array, use .ptr")
 *         return self._data.ctypes.data             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":65
 *         return self._data.ctypes.data
 * 
 *     def __int__(self):             # <<<<<<<<<<<<<<
 *         if self._data.size > 1:
 *             raise TypeError("int() argument must be a bytes-like object of size 1. "
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__int__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":71
 *         return self._data.ctypes.data
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._data.size
 * 
*/

/* Python wrapper */
static Py_ssize_t __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_6__len__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_6__len__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":72
 * 
 *     def __len__(self):
 *         return self._data.size             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":71
 *         return self._data.ctypes.data
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._data.size
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":74
 *         return self._data.size
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if not isinstance(other, uniqueid):
 *             return False
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_8__eq__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_8__eq__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":75
 * 
 *     def __eq__(self, other):
 *         if not isinstance(other, uniqueid):             # <<<<<<<<<<<<<<
 *             return False
 *         if self._data.size != other._data.size:
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "nvshmem/bindings/nvshmem.pyx":76
 *     def __eq__(self, other):
 *         if not isinstance(other, uniqueid):
 *             return False             # <<<<<<<<<<<<<<
 *         if self._data.size != other._data.size:
 *             return False
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":75
 * 
 *     def __eq__(self, other):
 *         if not isinstance(other, uniqueid):             # <<<<<<<<<<<<<<
 *             return False
 *         if self._data.size != other._data.size:
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":77
 *         if not isinstance(other, uniqueid):
 *             return False
 *         if self._data.size != other._data.size:             # <<<<<<<<<<<<<<
 *             return False
 *         if self._data.dtype != other._data.dtype:
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_mstate_global->__pyx_n_u_data_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "nvshmem/bindings/nvshmem.pyx":78
 *             return False
 *         if self._data.size != other._data.size:
 *             return False             # <<<<<<<<<<<<<<
 *         if self._data.dtype != other._data.dtype:
 *             return False
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":77
 *         if not isinstance(other, uniqueid):
 *             return False
 *         if self._data.size != other._data.size:             # <<<<<<<<<<<<<<
 *             return False
 *         if self._data.dtype != other._data.dtype:
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":79
 *         if self._data.size != other._data.size:
 *             return False
 *         if self._data.dtype != other._data.dtype:             # <<<<<<<<<<<<<<
 *             return False
 *         return bool((self._data == other._data).all())
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_mstate_global->__pyx_n_u_data_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "nvshmem/bindings/nvshmem.pyx":80
 *             return False
 *         if self._data.dtype != other._data.dtype:
 *             return False             # <<<<<<<<<<<<<<
 *         return bool((self._data == other._data).all())
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":79
 *         if self._data.size != other._data.size:
 *             return False
 *         if self._data.dtype != other._data.dtype:             # <<<<<<<<<<<<<<
 *             return False
 *         return bool((self._data == other._data).all())
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":81
 *         if self._data.dtype != other._data.dtype:
 *             return False
 *         return bool((self._data == other._data).all())             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_mstate_global->__pyx_n_u_data_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_self->_data, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __pyx_t_6;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_all, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":74
 *         return self._data.size
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if not isinstance(other, uniqueid):
 *             return False
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":83
 *         return bool((self._data == other._data).all())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":86
 *     def version(self):
 *         """version (~_numpy.int32): """
 *         if self._data.size == 1:             # <<<<<<<<<<<<<<
 *             return int(self._data.version[0])
 *         return self._data.version
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nvshmem/bindings/nvshmem.pyx":87
 *         """version (~_numpy.int32): """
 *         if self._data.size == 1:
 *             return int(self._data.version[0])             # <<<<<<<<<<<<<<
 *         return self._data.version
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":86
 *     def version(self):
 *         """version (~_numpy.int32): """
 *         if self._data.size == 1:             # <<<<<<<<<<<<<<
 *             return int(self._data.version[0])
 *         return self._data.version
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":88
 *         if self._data.size == 1:
 *             return int(self._data.version[0])
 *         return self._data.version             # <<<<<<<<<<<<<<
 * 
 *     @version.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":83
 *         return bool((self._data == other._data).all())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":90
 *         return self._data.version
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":92
 *     @version.setter
 *     def version(self, val):
 *         self._data.version = val             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version, __pyx_v_val) < (0)) __PYX_ERR(0, 92, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":90
 *         return self._data.version
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.version.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":94
 *         self._data.version = val
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(key, int):
 *             size = self._data.size
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_10__getitem__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_10__getitem__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);
  __Pyx_INCREF(__pyx_v_key);

  /* "nvshmem/bindings/nvshmem.pyx":95
 * 
 *     def __getitem__(self, key):
 *         if isinstance(key, int):             # <<<<<<<<<<<<<<
 *             size = self._data.size
 *             if key >= size or key <= -(size+1):
*/
  __pyx_t_1 = PyLong_Check(__pyx_v_key); 
  if (__pyx_t_1) {

    /* "nvshmem/bindings/nvshmem.pyx":96
 *     def __getitem__(self, key):
 *         if isinstance(key, int):
 *             size = self._data.size             # <<<<<<<<<<<<<<
 *             if key >= size or key <= -(size+1):
 *                 raise IndexError("index is out of bounds")
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_size = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "nvshmem/bindings/nvshmem.pyx":97
 *         if isinstance(key, int):
 *             size = self._data.size
 *             if key >= size or key <= -(size+1):             # <<<<<<<<<<<<<<
 *                 raise IndexError("index is out of bounds")
 *             if key < 0:
*/
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_key, __pyx_v_size, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_v_size, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Negative(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_key, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (unlikely(__pyx_t_1)) {

      /* "nvshmem/bindings/nvshmem.pyx":98
 *             size = self._data.size
 *             if key >= size or key <= -(size+1):
 *                 raise IndexError("index is out of bounds")             # <<<<<<<<<<<<<<
 *             if key < 0:
 *                 key += size
*/
      __pyx_t_4 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_index_is_out_of_bounds};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_IndexError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 98, __pyx_L1_error)

      /* "nvshmem/bindings/nvshmem.pyx":97
 *         if isinstance(key, int):
 *             size = self._data.size
 *             if key >= size or key <= -(size+1):             # <<<<<<<<<<<<<<
 *                 raise IndexError("index is out of bounds")
 *             if key < 0:
*/
    }

    /* "nvshmem/bindings/nvshmem.pyx":99
 *             if key >= size or key <= -(size+1):
 *                 raise IndexError("index is out of bounds")
 *             if key < 0:             # <<<<<<<<<<<<<<
 *                 key += size
 *             return uniqueid.from_data(self._data[key:key+1])
*/
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_key, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "nvshmem/bindings/nvshmem.pyx":100
 *                 raise IndexError("index is out of bounds")
 *             if key < 0:
 *                 key += size             # <<<<<<<<<<<<<<
 *             return uniqueid.from_data(self._data[key:key+1])
 *         out = self._data[key]
*/
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_key, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "nvshmem/bindings/nvshmem.pyx":99
 *             if key >= size or key <= -(size+1):
 *                 raise IndexError("index is out of bounds")
 *             if key < 0:             # <<<<<<<<<<<<<<
 *                 key += size
 *             return uniqueid.from_data(self._data[key:key+1])
*/
    }

    /* "nvshmem/bindings/nvshmem.pyx":101
 *             if key < 0:
 *                 key += size
 *             return uniqueid.from_data(self._data[key:key+1])             # <<<<<<<<<<<<<<
 *         out = self._data[key]
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyLong_AddObjC(__pyx_v_key, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_self->_data, 0, 0, &__pyx_v_key, &__pyx_t_6, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_from_data, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":95
 * 
 *     def __getitem__(self, key):
 *         if isinstance(key, int):             # <<<<<<<<<<<<<<
 *             size = self._data.size
 *             if key >= size or key <= -(size+1):
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":102
 *                 key += size
 *             return uniqueid.from_data(self._data[key:key+1])
 *         out = self._data[key]             # <<<<<<<<<<<<<<
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:
 *             return uniqueid.from_data(out)
*/
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_self->_data, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_out = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":103
 *             return uniqueid.from_data(self._data[key:key+1])
 *         out = self._data[key]
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:             # <<<<<<<<<<<<<<
 *             return uniqueid.from_data(out)
 *         return out
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_out, __pyx_t_7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_uniqueid_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_7, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "nvshmem/bindings/nvshmem.pyx":104
 *         out = self._data[key]
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:
 *             return uniqueid.from_data(out)             # <<<<<<<<<<<<<<
 *         return out
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_out};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_from_data, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "nvshmem/bindings/nvshmem.pyx":103
 *             return uniqueid.from_data(self._data[key:key+1])
 *         out = self._data[key]
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:             # <<<<<<<<<<<<<<
 *             return uniqueid.from_data(out)
 *         return out
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":105
 *         if isinstance(out, _numpy.recarray) and out.dtype == uniqueid_dtype:
 *             return uniqueid.from_data(out)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, key, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":94
 *         self._data.version = val
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(key, int):
 *             size = self._data.size
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":107
 *         return out
 * 
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         self._data[key] = val
 * 
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_13__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_13__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_12__setitem__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_12__setitem__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":108
 * 
 *     def __setitem__(self, key, val):
 *         self._data[key] = val             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
*/
  if (unlikely((PyObject_SetItem(__pyx_v_self->_data, __pyx_v_key, __pyx_v_val) < 0))) __PYX_ERR(0, 108, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":107
 *         return out
 * 
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         self._data[key] = val
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":110
 *         self._data[key] = val
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_data(data):
 *         """Create an uniqueid instance wrapping the given NumPy array.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data(CYTHON_UNUSED PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data, "uniqueid.from_data(data)\n\nCreate an uniqueid instance wrapping the given NumPy array.\n\nArgs:\n    data (_numpy.ndarray): a 1D array of dtype `uniqueid_dtype` holding the data.");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data = {"from_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data(CYTHON_UNUSED PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 110, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "from_data", 0) < (0)) __PYX_ERR(0, 110, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("from_data", 1, 1, 1, i); __PYX_ERR(0, 110, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 110, __pyx_L3_error)
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_data", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 110, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.from_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data(__pyx_v_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data(PyObject *__pyx_v_data) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_obj = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_data", 0);

  /* "nvshmem/bindings/nvshmem.pyx":117
 *             data (_numpy.ndarray): a 1D array of dtype `uniqueid_dtype` holding the data.
 *         """
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)             # <<<<<<<<<<<<<<
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):
 *             raise TypeError("data argument must be a NumPy ndarray")
*/
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_obj = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":118
 *         """
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):             # <<<<<<<<<<<<<<
 *             raise TypeError("data argument must be a NumPy ndarray")
 *         if data.ndim != 1:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ndarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_data, __pyx_t_2); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_data, __pyx_t_3); 
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (unlikely(__pyx_t_5)) {

    /* "nvshmem/bindings/nvshmem.pyx":119
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):
 *             raise TypeError("data argument must be a NumPy ndarray")             # <<<<<<<<<<<<<<
 *         if data.ndim != 1:
 *             raise ValueError("data array must be 1D")
*/
    __pyx_t_3 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_data_argument_must_be_a_NumPy_nd};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_TypeError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 119, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":118
 *         """
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):             # <<<<<<<<<<<<<<
 *             raise TypeError("data argument must be a NumPy ndarray")
 *         if data.ndim != 1:
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":120
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):
 *             raise TypeError("data argument must be a NumPy ndarray")
 *         if data.ndim != 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("data array must be 1D")
 *         if data.dtype != uniqueid_dtype:
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = (__Pyx_PyLong_BoolNeObjC(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "nvshmem/bindings/nvshmem.pyx":121
 *             raise TypeError("data argument must be a NumPy ndarray")
 *         if data.ndim != 1:
 *             raise ValueError("data array must be 1D")             # <<<<<<<<<<<<<<
 *         if data.dtype != uniqueid_dtype:
 *             raise ValueError("data array must be of dtype uniqueid_dtype")
*/
    __pyx_t_3 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_data_array_must_be_1D};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 121, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":120
 *         if not isinstance(data, (_numpy.ndarray, _numpy.recarray)):
 *             raise TypeError("data argument must be a NumPy ndarray")
 *         if data.ndim != 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("data array must be 1D")
 *         if data.dtype != uniqueid_dtype:
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":122
 *         if data.ndim != 1:
 *             raise ValueError("data array must be 1D")
 *         if data.dtype != uniqueid_dtype:             # <<<<<<<<<<<<<<
 *             raise ValueError("data array must be of dtype uniqueid_dtype")
 *         obj._data = data.view(_numpy.recarray)
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_uniqueid_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "nvshmem/bindings/nvshmem.pyx":123
 *             raise ValueError("data array must be 1D")
 *         if data.dtype != uniqueid_dtype:
 *             raise ValueError("data array must be of dtype uniqueid_dtype")             # <<<<<<<<<<<<<<
 *         obj._data = data.view(_numpy.recarray)
 * 
*/
    __pyx_t_3 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_data_array_must_be_of_dtype_uniq};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 123, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":122
 *         if data.ndim != 1:
 *             raise ValueError("data array must be 1D")
 *         if data.dtype != uniqueid_dtype:             # <<<<<<<<<<<<<<
 *             raise ValueError("data array must be of dtype uniqueid_dtype")
 *         obj._data = data.view(_numpy.recarray)
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":124
 *         if data.dtype != uniqueid_dtype:
 *             raise ValueError("data array must be of dtype uniqueid_dtype")
 *         obj._data = data.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 * 
 *         return obj
*/
  __pyx_t_3 = __pyx_v_data;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_obj->_data);
  __Pyx_DECREF(__pyx_v_obj->_data);
  __pyx_v_obj->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":126
 *         obj._data = data.view(_numpy.recarray)
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_obj);
  __pyx_r = ((PyObject *)__pyx_v_obj);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":110
 *         self._data[key] = val
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_data(data):
 *         """Create an uniqueid instance wrapping the given NumPy array.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.from_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":128
 *         return obj
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):
 *         """Create an uniqueid instance wrapping the given pointer.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr(CYTHON_UNUSED PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr, "uniqueid.from_ptr(intptr_t ptr, size_t size=1, bool readonly=False)\n\nCreate an uniqueid instance wrapping the given pointer.\n\nArgs:\n    ptr (intptr_t): pointer address as Python :py:`int` to the data.\n    size (int): number of structs, default=1.\n    readonly (bool): whether the data is read-only (to the user). default is `False`.");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr = {"from_ptr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr(CYTHON_UNUSED PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_ptr;
  size_t __pyx_v_size;
  int __pyx_v_readonly;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_ptr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ptr,&__pyx_mstate_global->__pyx_n_u_size,&__pyx_mstate_global->__pyx_n_u_readonly,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 128, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "from_ptr", 0) < (0)) __PYX_ERR(0, 128, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("from_ptr", 0, 1, 3, i); __PYX_ERR(0, 128, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 128, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ptr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_ptr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_size = __Pyx_PyLong_As_size_t(values[1]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)1);
    }
    if (values[2]) {
      __pyx_v_readonly = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_readonly == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
    } else {

      /* "nvshmem/bindings/nvshmem.pyx":129
 * 
 *     @staticmethod
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):             # <<<<<<<<<<<<<<
 *         """Create an uniqueid instance wrapping the given pointer.
 * 
*/
      __pyx_v_readonly = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_ptr", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 128, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr(__pyx_v_ptr, __pyx_v_size, __pyx_v_readonly);

  /* "nvshmem/bindings/nvshmem.pyx":128
 *         return obj
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):
 *         """Create an uniqueid instance wrapping the given pointer.
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr(intptr_t __pyx_v_ptr, size_t __pyx_v_size, int __pyx_v_readonly) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_obj = 0;
  PyObject *__pyx_v_flag = 0;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 0);

  /* "nvshmem/bindings/nvshmem.pyx":137
 *             readonly (bool): whether the data is read-only (to the user). default is `False`.
 *         """
 *         if ptr == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("ptr must not be null (0)")
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
*/
  __pyx_t_1 = (__pyx_v_ptr == 0);
  if (unlikely(__pyx_t_1)) {

    /* "nvshmem/bindings/nvshmem.pyx":138
 *         """
 *         if ptr == 0:
 *             raise ValueError("ptr must not be null (0)")             # <<<<<<<<<<<<<<
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         cdef flag = _buffer.PyBUF_READ if readonly else _buffer.PyBUF_WRITE
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_ptr_must_not_be_null_0};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 138, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":137
 *             readonly (bool): whether the data is read-only (to the user). default is `False`.
 *         """
 *         if ptr == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("ptr must not be null (0)")
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":139
 *         if ptr == 0:
 *             raise ValueError("ptr must not be null (0)")
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)             # <<<<<<<<<<<<<<
 *         cdef flag = _buffer.PyBUF_READ if readonly else _buffer.PyBUF_WRITE
 *         cdef object buf = PyMemoryView_FromMemory(
*/
  __pyx_t_2 = ((PyObject *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid(((PyTypeObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid), __pyx_mstate_global->__pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __pyx_v_obj = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":140
 *             raise ValueError("ptr must not be null (0)")
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         cdef flag = _buffer.PyBUF_READ if readonly else _buffer.PyBUF_WRITE             # <<<<<<<<<<<<<<
 *         cdef object buf = PyMemoryView_FromMemory(
 *             <char*>ptr, sizeof(nvshmemx_uniqueid_v1) * size, flag)
*/
  if (__pyx_v_readonly) {
    __pyx_t_3 = __Pyx_PyLong_From___pyx_anon_enum(PyBUF_READ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyLong_From___pyx_anon_enum(PyBUF_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_flag = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":142
 *         cdef flag = _buffer.PyBUF_READ if readonly else _buffer.PyBUF_WRITE
 *         cdef object buf = PyMemoryView_FromMemory(
 *             <char*>ptr, sizeof(nvshmemx_uniqueid_v1) * size, flag)             # <<<<<<<<<<<<<<
 *         data = _numpy.ndarray((size,), buffer=buf,
 *                               dtype=uniqueid_dtype)
*/
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_flag); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":141
 *         cdef uniqueid obj = uniqueid.__new__(uniqueid)
 *         cdef flag = _buffer.PyBUF_READ if readonly else _buffer.PyBUF_WRITE
 *         cdef object buf = PyMemoryView_FromMemory(             # <<<<<<<<<<<<<<
 *             <char*>ptr, sizeof(nvshmemx_uniqueid_v1) * size, flag)
 *         data = _numpy.ndarray((size,), buffer=buf,
*/
  __pyx_t_2 = PyMemoryView_FromMemory(((char *)__pyx_v_ptr), ((sizeof(nvshmemx_uniqueid_v1)) * __pyx_v_size), __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_buf = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":143
 *         cdef object buf = PyMemoryView_FromMemory(
 *             <char*>ptr, sizeof(nvshmemx_uniqueid_v1) * size, flag)
 *         data = _numpy.ndarray((size,), buffer=buf,             # <<<<<<<<<<<<<<
 *                               dtype=uniqueid_dtype)
 *         obj._data = data.view(_numpy.recarray)
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ndarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyLong_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6) != (0)) __PYX_ERR(0, 143, __pyx_L1_error);
  __pyx_t_6 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":144
 *             <char*>ptr, sizeof(nvshmemx_uniqueid_v1) * size, flag)
 *         data = _numpy.ndarray((size,), buffer=buf,
 *                               dtype=uniqueid_dtype)             # <<<<<<<<<<<<<<
 *         obj._data = data.view(_numpy.recarray)
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_uniqueid_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_3, __pyx_t_8};
    __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_buffer, __pyx_v_buf, __pyx_t_9, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 143, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, __pyx_t_6, __pyx_t_9, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 143, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":145
 *         data = _numpy.ndarray((size,), buffer=buf,
 *                               dtype=uniqueid_dtype)
 *         obj._data = data.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 * 
 *         return obj
*/
  __pyx_t_7 = __pyx_v_data;
  __Pyx_INCREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_obj->_data);
  __Pyx_DECREF(__pyx_v_obj->_data);
  __pyx_v_obj->_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":147
 *         obj._data = data.view(_numpy.recarray)
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_obj);
  __pyx_r = ((PyObject *)__pyx_v_obj);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":128
 *         return obj
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):
 *         """Create an uniqueid instance wrapping the given pointer.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_flag);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":46
 *     """
 *     cdef:
 *         readonly object _data             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, size=1):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_data);
  __pyx_r = __pyx_v_self->_data;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__, "uniqueid.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_data);
  __Pyx_GIVEREF(__pyx_v_self->_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_uniqueid); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_uniqueid__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_uniqueid); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_uniqueid__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__, "uniqueid.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_uniqueid__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_uniqueid__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_uniqueid__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uniqueid.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__, "UniqueId.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._data);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base._data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->__pyx_base._data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_UniqueId); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UniqueId__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_UniqueId); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.UniqueId.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UniqueId__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__, "UniqueId.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.UniqueId.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UniqueId__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_UniqueId__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UniqueId__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.UniqueId.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":160
 *         readonly object _data
 * 
 *     def __init__(self, value: int = 0):             # <<<<<<<<<<<<<<
 *         """Initialize with an optional value."""
 *         arr = _numpy.empty(1, dtype=_numpy.uint64)
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__, "Initialize with an optional value.");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__;
#endif
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 160, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 160, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 160, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_int_0));
    }
    if (__Pyx_PyInt_FromNumber(&values[0], "value", 0) < (0)) __PYX_ERR(0, 160, __pyx_L3_error)
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_arr = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":162
 *     def __init__(self, value: int = 0):
 *         """Initialize with an optional value."""
 *         arr = _numpy.empty(1, dtype=_numpy.uint64)             # <<<<<<<<<<<<<<
 *         self._data = arr.view(_numpy.recarray)
 *         self._data[0] = value
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_uint64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_int_1};
    __pyx_t_3 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, __pyx_t_5, __pyx_t_3, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 162, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":163
 *         """Initialize with an optional value."""
 *         arr = _numpy.empty(1, dtype=_numpy.uint64)
 *         self._data = arr.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 *         self._data[0] = value
 * 
*/
  __pyx_t_4 = __pyx_v_arr;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_data);
  __Pyx_DECREF(__pyx_v_self->_data);
  __pyx_v_self->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":164
 *         arr = _numpy.empty(1, dtype=_numpy.uint64)
 *         self._data = arr.view(_numpy.recarray)
 *         self._data[0] = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  if (unlikely((__Pyx_SetItemInt(__pyx_v_self->_data, 0, __pyx_v_value, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 164, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":160
 *         readonly object _data
 * 
 *     def __init__(self, value: int = 0):             # <<<<<<<<<<<<<<
 *         """Initialize with an optional value."""
 *         arr = _numpy.empty(1, dtype=_numpy.uint64)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":166
 *         self._data[0] = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":169
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
 *         return self._data.ctypes.data             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":166
 *         self._data[0] = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.ptr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":171
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self) -> int:
 *         """Get the value as Python int."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":174
 *     def value(self) -> int:
 *         """Get the value as Python int."""
 *         return int(self._data[0])             # <<<<<<<<<<<<<<
 * 
 *     @value.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_data, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":171
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self) -> int:
 *         """Get the value as Python int."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":176
 *         return int(self._data[0])
 * 
 *     @value.setter             # <<<<<<<<<<<<<<
 *     def value(self, val: int):
 *         """Set the value."""
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), (&PyLong_Type), 0, "val", 2))) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self), ((PyObject*)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":179
 *     def value(self, val: int):
 *         """Set the value."""
 *         self._data[0] = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self) -> str:
*/
  if (unlikely((__Pyx_SetItemInt(__pyx_v_self->_data, 0, __pyx_v_val, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":176
 *         return int(self._data[0])
 * 
 *     @value.setter             # <<<<<<<<<<<<<<
 *     def value(self, val: int):
 *         """Set the value."""
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.value.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":181
 *         self._data[0] = val
 * 
 *     def __str__(self) -> str:             # <<<<<<<<<<<<<<
 *         return f"team_uniqueid({self.value})"
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_2__str__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_2__str__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3[3];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":182
 * 
 *     def __str__(self) -> str:
 *         return f"team_uniqueid({self.value})"             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self) -> str:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3[0] = __pyx_mstate_global->__pyx_kp_u_team_uniqueid;
  __pyx_t_3[1] = __pyx_t_2;
  __pyx_t_3[2] = __pyx_mstate_global->__pyx_kp_u__3;
  __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, 14 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":181
 *         self._data[0] = val
 * 
 *     def __str__(self) -> str:             # <<<<<<<<<<<<<<
 *         return f"team_uniqueid({self.value})"
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":184
 *         return f"team_uniqueid({self.value})"
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return f"team_uniqueid(value={self.value})"
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_4__repr__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_4__repr__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3[3];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":185
 * 
 *     def __repr__(self) -> str:
 *         return f"team_uniqueid(value={self.value})"             # <<<<<<<<<<<<<<
 * 
 * cdef class TeamUniqueId(team_uniqueid): pass
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3[0] = __pyx_mstate_global->__pyx_kp_u_team_uniqueid_value;
  __pyx_t_3[1] = __pyx_t_2;
  __pyx_t_3[2] = __pyx_mstate_global->__pyx_kp_u__3;
  __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, 20 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":184
 *         return f"team_uniqueid({self.value})"
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return f"team_uniqueid(value={self.value})"
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":158
 * 
 *     cdef:
 *         readonly object _data             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, value: int = 0):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_data);
  __pyx_r = __pyx_v_self->_data;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__, "team_uniqueid.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_data);
  __Pyx_GIVEREF(__pyx_v_self->_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_team_uniqueid); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_team_uniqueid__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_team_uniqueid); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_team_uniqueid__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__, "team_uniqueid.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_team_uniqueid__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_team_uniqueid__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_team_uniqueid__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_uniqueid.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__, "TeamUniqueId.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._data);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base._data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->__pyx_base._data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamUniqueId); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TeamUniqueId__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamUniqueId); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamUniqueId.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamUniqueId__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__, "TeamUniqueId.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamUniqueId.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TeamUniqueId__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamUniqueId__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamUniqueId__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamUniqueId.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":206
 *         readonly object _data
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(1, dtype=init_attr_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__init__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr___init__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_v_arr = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8[4];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":207
 * 
 *     def __init__(self):
 *         arr = _numpy.empty(1, dtype=init_attr_dtype)             # <<<<<<<<<<<<<<
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_init_attr_t), \
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_init_attr_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_int_1};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, __pyx_t_3, __pyx_t_6, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 207, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":208
 *     def __init__(self):
 *         arr = _numpy.empty(1, dtype=init_attr_dtype)
 *         self._data = arr.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 *         assert self._data.itemsize == sizeof(nvshmemx_init_attr_t), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_init_attr_t)}"
*/
  __pyx_t_4 = __pyx_v_arr;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_data);
  __Pyx_DECREF(__pyx_v_self->_data);
  __pyx_v_self->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":209
 *         arr = _numpy.empty(1, dtype=init_attr_dtype)
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_init_attr_t), \             # <<<<<<<<<<<<<<
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_init_attr_t)}"
 * 
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyLong_FromSize_t((sizeof(nvshmemx_init_attr_t))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) {

      /* "nvshmem/bindings/nvshmem.pyx":210
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmemx_init_attr_t), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_init_attr_t)}"             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_From_size_t((sizeof(nvshmemx_init_attr_t)), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8[0] = __pyx_mstate_global->__pyx_kp_u_itemsize_2;
      __pyx_t_8[1] = __pyx_t_3;
      __pyx_t_8[2] = __pyx_mstate_global->__pyx_kp_u_mismatches_struct_size;
      __pyx_t_8[3] = __pyx_t_4;
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_8, 4, 9 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 24 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3));
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_1, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 209, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 209, __pyx_L1_error)
  #endif

  /* "nvshmem/bindings/nvshmem.pyx":206
 *         readonly object _data
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(1, dtype=init_attr_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":212
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_init_attr_t)}"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":215
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
 *         return self._data.ctypes.data             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":212
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmemx_init_attr_t)}"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.ptr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":217
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":220
 *     def version(self):
 *         """version (~_numpy.int32): """
 *         return int(self._data.version[0])             # <<<<<<<<<<<<<<
 * 
 *     @version.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":217
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":222
 *         return int(self._data.version[0])
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":224
 *     @version.setter
 *     def version(self, val):
 *         self._data.version = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version, __pyx_v_val) < (0)) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":222
 *         return int(self._data.version[0])
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.version.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":226
 *         self._data.version = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mpi_comm(self):
 *         """mpi_comm (~_numpy.intp): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":229
 *     def mpi_comm(self):
 *         """mpi_comm (~_numpy.intp): """
 *         return int(self._data.mpi_comm[0])             # <<<<<<<<<<<<<<
 * 
 *     @mpi_comm.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_mpi_comm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":226
 *         self._data.version = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mpi_comm(self):
 *         """mpi_comm (~_numpy.intp): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.mpi_comm.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":231
 *         return int(self._data.mpi_comm[0])
 * 
 *     @mpi_comm.setter             # <<<<<<<<<<<<<<
 *     def mpi_comm(self, val):
 *         self._data.mpi_comm = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":233
 *     @mpi_comm.setter
 *     def mpi_comm(self, val):
 *         self._data.mpi_comm = val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_mpi_comm, __pyx_v_val) < (0)) __PYX_ERR(0, 233, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":231
 *         return int(self._data.mpi_comm[0])
 * 
 *     @mpi_comm.setter             # <<<<<<<<<<<<<<
 *     def mpi_comm(self, val):
 *         self._data.mpi_comm = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.mpi_comm.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":204
 * 
 *     cdef:
 *         readonly object _data             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_data);
  __pyx_r = __pyx_v_self->_data;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__, "InitAttr.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_data);
  __Pyx_GIVEREF(__pyx_v_self->_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InitAttr); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InitAttr__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InitAttr); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InitAttr__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__, "InitAttr.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InitAttr__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_InitAttr__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InitAttr__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.InitAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":254
 *         readonly object _data
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(1, dtype=team_config_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__init__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig___init__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig___init__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_v_arr = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8[4];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":255
 * 
 *     def __init__(self):
 *         arr = _numpy.empty(1, dtype=team_config_dtype)             # <<<<<<<<<<<<<<
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmem_team_config_t), \
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_team_config_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_int_1};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, __pyx_t_3, __pyx_t_6, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 255, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":256
 *     def __init__(self):
 *         arr = _numpy.empty(1, dtype=team_config_dtype)
 *         self._data = arr.view(_numpy.recarray)             # <<<<<<<<<<<<<<
 *         assert self._data.itemsize == sizeof(nvshmem_team_config_t), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmem_team_config_t)}"
*/
  __pyx_t_4 = __pyx_v_arr;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_recarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_view, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_data);
  __Pyx_DECREF(__pyx_v_self->_data);
  __pyx_v_self->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":257
 *         arr = _numpy.empty(1, dtype=team_config_dtype)
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmem_team_config_t), \             # <<<<<<<<<<<<<<
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmem_team_config_t)}"
 * 
*/
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyLong_FromSize_t((sizeof(nvshmem_team_config_t))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) {

      /* "nvshmem/bindings/nvshmem.pyx":258
 *         self._data = arr.view(_numpy.recarray)
 *         assert self._data.itemsize == sizeof(nvshmem_team_config_t), \
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmem_team_config_t)}"             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_From_size_t((sizeof(nvshmem_team_config_t)), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8[0] = __pyx_mstate_global->__pyx_kp_u_itemsize_2;
      __pyx_t_8[1] = __pyx_t_3;
      __pyx_t_8[2] = __pyx_mstate_global->__pyx_kp_u_mismatches_struct_size;
      __pyx_t_8[3] = __pyx_t_4;
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_8, 4, 9 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 24 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3));
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_1, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 257, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 257, __pyx_L1_error)
  #endif

  /* "nvshmem/bindings/nvshmem.pyx":254
 *         readonly object _data
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         arr = _numpy.empty(1, dtype=team_config_dtype)
 *         self._data = arr.view(_numpy.recarray)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_arr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":260
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmem_team_config_t)}"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":263
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
 *         return self._data.ctypes.data             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":260
 *             f"itemsize {self._data.itemsize} mismatches struct size {sizeof(nvshmem_team_config_t)}"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptr(self):
 *         """Get the pointer address to the data as Python :py:`int`."""
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.ptr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":265
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":268
 *     def version(self):
 *         """version (~_numpy.int32): """
 *         return int(self._data.version[0])             # <<<<<<<<<<<<<<
 * 
 *     @version.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":265
 *         return self._data.ctypes.data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def version(self):
 *         """version (~_numpy.int32): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":270
 *         return int(self._data.version[0])
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":272
 *     @version.setter
 *     def version(self, val):
 *         self._data.version = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_version, __pyx_v_val) < (0)) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":270
 *         return int(self._data.version[0])
 * 
 *     @version.setter             # <<<<<<<<<<<<<<
 *     def version(self, val):
 *         self._data.version = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.version.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":274
 *         self._data.version = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def num_contexts(self):
 *         """num_contexts (~_numpy.int32): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":277
 *     def num_contexts(self):
 *         """num_contexts (~_numpy.int32): """
 *         return int(self._data.num_contexts[0])             # <<<<<<<<<<<<<<
 * 
 *     @num_contexts.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_num_contexts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":274
 *         self._data.version = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def num_contexts(self):
 *         """num_contexts (~_numpy.int32): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.num_contexts.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":279
 *         return int(self._data.num_contexts[0])
 * 
 *     @num_contexts.setter             # <<<<<<<<<<<<<<
 *     def num_contexts(self, val):
 *         self._data.num_contexts = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":281
 *     @num_contexts.setter
 *     def num_contexts(self, val):
 *         self._data.num_contexts = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_num_contexts, __pyx_v_val) < (0)) __PYX_ERR(0, 281, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":279
 *         return int(self._data.num_contexts[0])
 * 
 *     @num_contexts.setter             # <<<<<<<<<<<<<<
 *     def num_contexts(self, val):
 *         self._data.num_contexts = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.num_contexts.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":283
 *         self._data.num_contexts = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def uniqueid(self):
 *         """uniqueid (~_numpy.uint64): """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":286
 *     def uniqueid(self):
 *         """uniqueid (~_numpy.uint64): """
 *         return int(self._data.uniqueid[0])             # <<<<<<<<<<<<<<
 * 
 *     @uniqueid.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_uniqueid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":283
 *         self._data.num_contexts = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def uniqueid(self):
 *         """uniqueid (~_numpy.uint64): """
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.uniqueid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":288
 *         return int(self._data.uniqueid[0])
 * 
 *     @uniqueid.setter             # <<<<<<<<<<<<<<
 *     def uniqueid(self, val):
 *         self._data.uniqueid = val
*/

/* Python wrapper */
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_2__set__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_2__set__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":290
 *     @uniqueid.setter
 *     def uniqueid(self, val):
 *         self._data.uniqueid = val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_data, __pyx_mstate_global->__pyx_n_u_uniqueid, __pyx_v_val) < (0)) __PYX_ERR(0, 290, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":288
 *         return int(self._data.uniqueid[0])
 * 
 *     @uniqueid.setter             # <<<<<<<<<<<<<<
 *     def uniqueid(self, val):
 *         self._data.uniqueid = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.uniqueid.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":252
 * 
 *     cdef:
 *         readonly object _data             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data___get__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data___get__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_data);
  __pyx_r = __pyx_v_self->_data;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__, "TeamConfig.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_data);
  __Pyx_GIVEREF(__pyx_v_self->_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_data) != (0)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._data is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, None), state
*/
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_data != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamConfig); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, None), state
 *     else:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TeamConfig__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamConfig); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_175497610);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_175497610);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_175497610) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamConfig__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__, "TeamConfig.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TeamConfig__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(2, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamConfig__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamConfig__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.TeamConfig.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":411
 * class NVSHMEMError(Exception):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         cdef str err = f"Status code {status}"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_12NVSHMEMError___init__, "NVSHMEMError.__init__(self, status)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12NVSHMEMError___init__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_status = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_status,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 411, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 411, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 411, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 411, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 411, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 411, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 411, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_status = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 411, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.NVSHMEMError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError___init__(__pyx_self, __pyx_v_self, __pyx_v_status);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_status) {
  PyObject *__pyx_v_err = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":412
 * 
 *     def __init__(self, status):
 *         self.status = status             # <<<<<<<<<<<<<<
 *         cdef str err = f"Status code {status}"
 *         super(NVSHMEMError, self).__init__(err)
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_status, __pyx_v_status) < (0)) __PYX_ERR(0, 412, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":413
 *     def __init__(self, status):
 *         self.status = status
 *         cdef str err = f"Status code {status}"             # <<<<<<<<<<<<<<
 *         super(NVSHMEMError, self).__init__(err)
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_status, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Status_code, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_err = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":414
 *         self.status = status
 *         cdef str err = f"Status code {status}"
 *         super(NVSHMEMError, self).__init__(err)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_NVSHMEMError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_5, __pyx_v_self};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_1 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_err};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":411
 * class NVSHMEMError(Exception):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         cdef str err = f"Status code {status}"
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.NVSHMEMError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":416
 *         super(NVSHMEMError, self).__init__(err)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_3__reduce__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_2__reduce__, "NVSHMEMError.__reduce__(self)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_3__reduce__ = {"__reduce__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_3__reduce__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_2__reduce__};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_3__reduce__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 416, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__reduce__", 0) < (0)) __PYX_ERR(0, 416, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 1, 1, i); __PYX_ERR(0, 416, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 416, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__reduce__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 416, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.NVSHMEMError.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_2__reduce__(__pyx_self, __pyx_v_self);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "nvshmem/bindings/nvshmem.pyx":417
 * 
 *     def __reduce__(self):
 *         return (type(self), (self.status,))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1) != (0)) __PYX_ERR(0, 417, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self))) != (0)) __PYX_ERR(0, 417, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 417, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":416
 *         super(NVSHMEMError, self).__init__(err)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.NVSHMEMError.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":420
 * 
 * 
 * @cython.profile(False)             # <<<<<<<<<<<<<<
 * cpdef inline check_status(int status):
 *     if status != 0:
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_1check_status(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_check_status(int __pyx_v_status, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_status", 0);

  /* "nvshmem/bindings/nvshmem.pyx":422
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise NVSHMEMError(status)
 * 
*/
  __pyx_t_1 = (__pyx_v_status != 0);
  if (unlikely(__pyx_t_1)) {

    /* "nvshmem/bindings/nvshmem.pyx":423
 * cpdef inline check_status(int status):
 *     if status != 0:
 *         raise NVSHMEMError(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_NVSHMEMError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_status); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 423, __pyx_L1_error)

    /* "nvshmem/bindings/nvshmem.pyx":422
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise NVSHMEMError(status)
 * 
*/
  }

  /* "nvshmem/bindings/nvshmem.pyx":420
 * 
 * 
 * @cython.profile(False)             # <<<<<<<<<<<<<<
 * cpdef inline check_status(int status):
 *     if status != 0:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_1check_status(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_check_status, "check_status(int status)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_1check_status = {"check_status", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_1check_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_check_status};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_1check_status(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_status;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_status (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_status,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 420, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "check_status", 0) < (0)) __PYX_ERR(0, 420, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("check_status", 1, 1, 1, i); __PYX_ERR(0, 420, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
    }
    __pyx_v_status = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 421, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("check_status", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 420, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_check_status(__pyx_self, __pyx_v_status);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_status", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":430
 * ###############################################################################
 * 
 * cpdef barrier(int32_t team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_3barrier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_barrier(int32_t __pyx_v_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier", 0);

  /* "nvshmem/bindings/nvshmem.pyx":431
 * 
 * cpdef barrier(int32_t team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":432
 * cpdef barrier(int32_t team):
 *     with nogil:
 *         status = nvshmem_barrier(<nvshmem_team_t>team)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier(((nvshmem_team_t)__pyx_v_team)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 432, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":431
 * 
 * cpdef barrier(int32_t team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":433
 *     with nogil:
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":430
 * ###############################################################################
 * 
 * cpdef barrier(int32_t team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_3barrier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_2barrier, "barrier(int32_t team)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_3barrier = {"barrier", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_3barrier, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_2barrier};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_3barrier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("barrier (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 430, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 430, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "barrier", 0) < (0)) __PYX_ERR(0, 430, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("barrier", 1, 1, 1, i); __PYX_ERR(0, 430, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 430, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("barrier", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 430, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_2barrier(__pyx_self, __pyx_v_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_2barrier(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_barrier(__pyx_v_team, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":436
 * 
 * 
 * cpdef void barrier_all() except*:             # <<<<<<<<<<<<<<
 *     nvshmem_barrier_all()
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_5barrier_all(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all(CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":437
 * 
 * cpdef void barrier_all() except*:
 *     nvshmem_barrier_all()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier_all(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 437, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":436
 * 
 * 
 * cpdef void barrier_all() except*:             # <<<<<<<<<<<<<<
 *     nvshmem_barrier_all()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_5barrier_all(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_4barrier_all, "barrier_all() -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_5barrier_all = {"barrier_all", (PyCFunction)__pyx_pw_7nvshmem_8bindings_7nvshmem_5barrier_all, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_7nvshmem_4barrier_all};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_5barrier_all(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("barrier_all (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_4barrier_all(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_4barrier_all(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier_all", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all(1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":440
 * 
 * 
 * cpdef int init_status() except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_init_status()
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_7init_status(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_init_status(CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":441
 * 
 * cpdef int init_status() except? 0:
 *     return nvshmemx_init_status()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_init_status(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":440
 * 
 * 
 * cpdef int init_status() except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_init_status()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.init_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_7init_status(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_6init_status, "init_status() -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_7init_status = {"init_status", (PyCFunction)__pyx_pw_7nvshmem_8bindings_7nvshmem_7init_status, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_7nvshmem_6init_status};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_7init_status(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_status (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_6init_status(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_6init_status(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_status", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_init_status(1); if (unlikely(__pyx_t_1 == ((int)0) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.init_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":444
 * 
 * 
 * cpdef int my_pe() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_my_pe()
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_9my_pe(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_my_pe(CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":445
 * 
 * cpdef int my_pe() except? -1:
 *     return nvshmem_my_pe()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_my_pe(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":444
 * 
 * 
 * cpdef int my_pe() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_my_pe()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_9my_pe(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_8my_pe, "my_pe() -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_9my_pe = {"my_pe", (PyCFunction)__pyx_pw_7nvshmem_8bindings_7nvshmem_9my_pe, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8my_pe};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_9my_pe(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("my_pe (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_8my_pe(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_8my_pe(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("my_pe", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_my_pe(1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":448
 * 
 * 
 * cpdef int n_pes() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_n_pes()
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_11n_pes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_n_pes(CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":449
 * 
 * cpdef int n_pes() except? -1:
 *     return nvshmem_n_pes()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_n_pes(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":448
 * 
 * 
 * cpdef int n_pes() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_n_pes()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_11n_pes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_10n_pes, "n_pes() -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_11n_pes = {"n_pes", (PyCFunction)__pyx_pw_7nvshmem_8bindings_7nvshmem_11n_pes, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_7nvshmem_10n_pes};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_11n_pes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("n_pes (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_10n_pes(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_10n_pes(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("n_pes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_n_pes(1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 448, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":452
 * 
 * 
 * cpdef void info_get_version(intptr_t major, intptr_t minor) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_info_get_version(<int*>major, <int*>minor)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13info_get_version(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_info_get_version(intptr_t __pyx_v_major, intptr_t __pyx_v_minor, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":453
 * 
 * cpdef void info_get_version(intptr_t major, intptr_t minor) except*:
 *     nvshmem_info_get_version(<int*>major, <int*>minor)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_info_get_version(((int *)__pyx_v_major), ((int *)__pyx_v_minor)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 453, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":452
 * 
 * 
 * cpdef void info_get_version(intptr_t major, intptr_t minor) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_info_get_version(<int*>major, <int*>minor)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.info_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13info_get_version(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_12info_get_version, "info_get_version(intptr_t major, intptr_t minor) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_13info_get_version = {"info_get_version", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_13info_get_version, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12info_get_version};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_13info_get_version(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_major;
  intptr_t __pyx_v_minor;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("info_get_version (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_major,&__pyx_mstate_global->__pyx_n_u_minor,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 452, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "info_get_version", 0) < (0)) __PYX_ERR(0, 452, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("info_get_version", 1, 2, 2, i); __PYX_ERR(0, 452, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 452, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 452, __pyx_L3_error)
    }
    __pyx_v_major = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_major == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L3_error)
    __pyx_v_minor = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_minor == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("info_get_version", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 452, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.info_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_12info_get_version(__pyx_self, __pyx_v_major, __pyx_v_minor);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_12info_get_version(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_major, intptr_t __pyx_v_minor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("info_get_version", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_info_get_version(__pyx_v_major, __pyx_v_minor, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.info_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":456
 * 
 * 
 * cpdef void vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_vendor_get_version_info(<int*>major, <int*>minor, <int*>patch)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_vendor_get_version_info(intptr_t __pyx_v_major, intptr_t __pyx_v_minor, intptr_t __pyx_v_patch, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":457
 * 
 * cpdef void vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) except*:
 *     nvshmemx_vendor_get_version_info(<int*>major, <int*>minor, <int*>patch)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_vendor_get_version_info(((int *)__pyx_v_major), ((int *)__pyx_v_minor), ((int *)__pyx_v_patch)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 457, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":456
 * 
 * 
 * cpdef void vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_vendor_get_version_info(<int*>major, <int*>minor, <int*>patch)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.vendor_get_version_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_14vendor_get_version_info, "vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info = {"vendor_get_version_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_14vendor_get_version_info};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_major;
  intptr_t __pyx_v_minor;
  intptr_t __pyx_v_patch;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vendor_get_version_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_major,&__pyx_mstate_global->__pyx_n_u_minor,&__pyx_mstate_global->__pyx_n_u_patch,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 456, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 456, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 456, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 456, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "vendor_get_version_info", 0) < (0)) __PYX_ERR(0, 456, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("vendor_get_version_info", 1, 3, 3, i); __PYX_ERR(0, 456, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 456, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 456, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 456, __pyx_L3_error)
    }
    __pyx_v_major = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_major == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L3_error)
    __pyx_v_minor = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_minor == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L3_error)
    __pyx_v_patch = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_patch == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("vendor_get_version_info", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 456, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.vendor_get_version_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_14vendor_get_version_info(__pyx_self, __pyx_v_major, __pyx_v_minor, __pyx_v_patch);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_14vendor_get_version_info(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_major, intptr_t __pyx_v_minor, intptr_t __pyx_v_patch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vendor_get_version_info", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_vendor_get_version_info(__pyx_v_major, __pyx_v_minor, __pyx_v_patch, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.vendor_get_version_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":460
 * 
 * 
 * cpdef intptr_t malloc(size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_malloc(size)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_17malloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_malloc(size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":461
 * 
 * cpdef intptr_t malloc(size_t size) except? 0:
 *     return <intptr_t>nvshmem_malloc(size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_malloc(__pyx_v_size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":460
 * 
 * 
 * cpdef intptr_t malloc(size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_malloc(size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_17malloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_16malloc, "malloc(size_t size) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_17malloc = {"malloc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_17malloc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_16malloc};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_17malloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  size_t __pyx_v_size;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("malloc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 460, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 460, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "malloc", 0) < (0)) __PYX_ERR(0, 460, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("malloc", 1, 1, 1, i); __PYX_ERR(0, 460, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 460, __pyx_L3_error)
    }
    __pyx_v_size = __Pyx_PyLong_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 460, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("malloc", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 460, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_16malloc(__pyx_self, __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_16malloc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("malloc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_malloc(__pyx_v_size, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 460, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":464
 * 
 * 
 * cpdef intptr_t calloc(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_calloc(count, size)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_19calloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_calloc(size_t __pyx_v_count, size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":465
 * 
 * cpdef intptr_t calloc(size_t count, size_t size) except? 0:
 *     return <intptr_t>nvshmem_calloc(count, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_calloc(__pyx_v_count, __pyx_v_size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":464
 * 
 * 
 * cpdef intptr_t calloc(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_calloc(count, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_19calloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_18calloc, "calloc(size_t count, size_t size) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_19calloc = {"calloc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_19calloc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_18calloc};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_19calloc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  size_t __pyx_v_count;
  size_t __pyx_v_size;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calloc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_count,&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 464, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "calloc", 0) < (0)) __PYX_ERR(0, 464, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("calloc", 1, 2, 2, i); __PYX_ERR(0, 464, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 464, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 464, __pyx_L3_error)
    }
    __pyx_v_count = __Pyx_PyLong_As_size_t(values[0]); if (unlikely((__pyx_v_count == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyLong_As_size_t(values[1]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calloc", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 464, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_18calloc(__pyx_self, __pyx_v_count, __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_18calloc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_count, size_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calloc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_calloc(__pyx_v_count, __pyx_v_size, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":468
 * 
 * 
 * cpdef intptr_t align(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_align(count, size)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_21align(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_align(size_t __pyx_v_count, size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":469
 * 
 * cpdef intptr_t align(size_t count, size_t size) except? 0:
 *     return <intptr_t>nvshmem_align(count, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_align(__pyx_v_count, __pyx_v_size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":468
 * 
 * 
 * cpdef intptr_t align(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_align(count, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_21align(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_20align, "align(size_t count, size_t size) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_21align = {"align", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_21align, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_20align};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_21align(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  size_t __pyx_v_count;
  size_t __pyx_v_size;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("align (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_count,&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 468, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 468, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 468, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "align", 0) < (0)) __PYX_ERR(0, 468, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("align", 1, 2, 2, i); __PYX_ERR(0, 468, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 468, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 468, __pyx_L3_error)
    }
    __pyx_v_count = __Pyx_PyLong_As_size_t(values[0]); if (unlikely((__pyx_v_count == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyLong_As_size_t(values[1]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("align", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 468, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_20align(__pyx_self, __pyx_v_count, __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_20align(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_count, size_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("align", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_align(__pyx_v_count, __pyx_v_size, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":472
 * 
 * 
 * cpdef void free(intptr_t ptr) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_free(<void*>ptr)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_23free(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_free(intptr_t __pyx_v_ptr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":473
 * 
 * cpdef void free(intptr_t ptr) except*:
 *     nvshmem_free(<void*>ptr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_free(((void *)__pyx_v_ptr)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":472
 * 
 * 
 * cpdef void free(intptr_t ptr) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_free(<void*>ptr)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.free", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_23free(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_22free, "free(intptr_t ptr) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_23free = {"free", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_23free, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_22free};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_23free(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_ptr;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("free (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ptr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 472, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 472, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "free", 0) < (0)) __PYX_ERR(0, 472, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("free", 1, 1, 1, i); __PYX_ERR(0, 472, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 472, __pyx_L3_error)
    }
    __pyx_v_ptr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_ptr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 472, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("free", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 472, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.free", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_22free(__pyx_self, __pyx_v_ptr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_22free(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("free", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_free(__pyx_v_ptr, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 472, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.free", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":476
 * 
 * 
 * cpdef intptr_t ptr(intptr_t dest, int pe) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_ptr(<const void*>dest, pe)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_25ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_ptr(intptr_t __pyx_v_dest, int __pyx_v_pe, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":477
 * 
 * cpdef intptr_t ptr(intptr_t dest, int pe) except? 0:
 *     return <intptr_t>nvshmem_ptr(<const void*>dest, pe)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_ptr(((void const *)__pyx_v_dest), __pyx_v_pe); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":476
 * 
 * 
 * cpdef intptr_t ptr(intptr_t dest, int pe) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_ptr(<const void*>dest, pe)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_25ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_24ptr, "ptr(intptr_t dest, int pe) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_25ptr = {"ptr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_25ptr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_24ptr};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_25ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_dest;
  int __pyx_v_pe;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ptr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_pe,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 476, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "ptr", 0) < (0)) __PYX_ERR(0, 476, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("ptr", 1, 2, 2, i); __PYX_ERR(0, 476, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 476, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 476, __pyx_L3_error)
    }
    __pyx_v_dest = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L3_error)
    __pyx_v_pe = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ptr", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 476, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_24ptr(__pyx_self, __pyx_v_dest, __pyx_v_pe);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_24ptr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, int __pyx_v_pe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ptr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_ptr(__pyx_v_dest, __pyx_v_pe, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":480
 * 
 * 
 * cpdef intptr_t mc_ptr(int32_t team, intptr_t ptr) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_mc_ptr(<nvshmem_team_t>team, <const void*>ptr)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_27mc_ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_mc_ptr(int32_t __pyx_v_team, intptr_t __pyx_v_ptr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":481
 * 
 * cpdef intptr_t mc_ptr(int32_t team, intptr_t ptr) except? 0:
 *     return <intptr_t>nvshmemx_mc_ptr(<nvshmem_team_t>team, <const void*>ptr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_mc_ptr(((nvshmem_team_t)__pyx_v_team), ((void const *)__pyx_v_ptr)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":480
 * 
 * 
 * cpdef intptr_t mc_ptr(int32_t team, intptr_t ptr) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_mc_ptr(<nvshmem_team_t>team, <const void*>ptr)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.mc_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_27mc_ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_26mc_ptr, "mc_ptr(int32_t team, intptr_t ptr) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_27mc_ptr = {"mc_ptr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_27mc_ptr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_26mc_ptr};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_27mc_ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_ptr;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mc_ptr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_ptr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 480, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 480, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 480, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "mc_ptr", 0) < (0)) __PYX_ERR(0, 480, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("mc_ptr", 1, 2, 2, i); __PYX_ERR(0, 480, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 480, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 480, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L3_error)
    __pyx_v_ptr = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_ptr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mc_ptr", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 480, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.mc_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_26mc_ptr(__pyx_self, __pyx_v_team, __pyx_v_ptr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_26mc_ptr(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mc_ptr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_mc_ptr(__pyx_v_team, __pyx_v_ptr, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.mc_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":484
 * 
 * 
 * cpdef int team_my_pe(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_my_pe(<nvshmem_team_t>team)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_29team_my_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_my_pe(int32_t __pyx_v_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":485
 * 
 * cpdef int team_my_pe(int32_t team) except? -1:
 *     return nvshmem_team_my_pe(<nvshmem_team_t>team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_my_pe(((nvshmem_team_t)__pyx_v_team)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":484
 * 
 * 
 * cpdef int team_my_pe(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_my_pe(<nvshmem_team_t>team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_29team_my_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_28team_my_pe, "team_my_pe(int32_t team) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_29team_my_pe = {"team_my_pe", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_29team_my_pe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_28team_my_pe};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_29team_my_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_my_pe (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 484, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 484, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_my_pe", 0) < (0)) __PYX_ERR(0, 484, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_my_pe", 1, 1, 1, i); __PYX_ERR(0, 484, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 484, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_my_pe", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 484, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_28team_my_pe(__pyx_self, __pyx_v_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_28team_my_pe(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_my_pe", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_my_pe(__pyx_v_team, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":488
 * 
 * 
 * cpdef int team_n_pes(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_n_pes(<nvshmem_team_t>team)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_31team_n_pes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_n_pes(int32_t __pyx_v_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":489
 * 
 * cpdef int team_n_pes(int32_t team) except? -1:
 *     return nvshmem_team_n_pes(<nvshmem_team_t>team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_n_pes(((nvshmem_team_t)__pyx_v_team)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 489, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":488
 * 
 * 
 * cpdef int team_n_pes(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_n_pes(<nvshmem_team_t>team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_31team_n_pes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_30team_n_pes, "team_n_pes(int32_t team) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_31team_n_pes = {"team_n_pes", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_31team_n_pes, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_30team_n_pes};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_31team_n_pes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_n_pes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 488, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_n_pes", 0) < (0)) __PYX_ERR(0, 488, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_n_pes", 1, 1, 1, i); __PYX_ERR(0, 488, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 488, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_n_pes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 488, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_30team_n_pes(__pyx_self, __pyx_v_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_30team_n_pes(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_n_pes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_n_pes(__pyx_v_team, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":492
 * 
 * 
 * cpdef void team_get_config(int32_t team, intptr_t config) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_get_config(<nvshmem_team_t>team, <nvshmem_team_config_t*>config)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_33team_get_config(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_team_get_config(int32_t __pyx_v_team, intptr_t __pyx_v_config, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":493
 * 
 * cpdef void team_get_config(int32_t team, intptr_t config) except*:
 *     nvshmem_team_get_config(<nvshmem_team_t>team, <nvshmem_team_config_t*>config)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_get_config(((nvshmem_team_t)__pyx_v_team), ((nvshmem_team_config_t *)__pyx_v_config)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":492
 * 
 * 
 * cpdef void team_get_config(int32_t team, intptr_t config) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_get_config(<nvshmem_team_t>team, <nvshmem_team_config_t*>config)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_33team_get_config(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_32team_get_config, "team_get_config(int32_t team, intptr_t config) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_33team_get_config = {"team_get_config", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_33team_get_config, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_32team_get_config};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_33team_get_config(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_config;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_get_config (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_config,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 492, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 492, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 492, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_get_config", 0) < (0)) __PYX_ERR(0, 492, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_get_config", 1, 2, 2, i); __PYX_ERR(0, 492, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 492, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 492, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L3_error)
    __pyx_v_config = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_config == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_get_config", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 492, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_32team_get_config(__pyx_self, __pyx_v_team, __pyx_v_config);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_32team_get_config(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_config) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_get_config", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_team_get_config(__pyx_v_team, __pyx_v_config, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":496
 * 
 * 
 * cpdef int team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_translate_pe(<nvshmem_team_t>src_team, src_pe, <nvshmem_team_t>dest_team)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_35team_translate_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_translate_pe(int32_t __pyx_v_src_team, int __pyx_v_src_pe, int32_t __pyx_v_dest_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":497
 * 
 * cpdef int team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) except? -1:
 *     return nvshmem_team_translate_pe(<nvshmem_team_t>src_team, src_pe, <nvshmem_team_t>dest_team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_translate_pe(((nvshmem_team_t)__pyx_v_src_team), __pyx_v_src_pe, ((nvshmem_team_t)__pyx_v_dest_team)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":496
 * 
 * 
 * cpdef int team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_translate_pe(<nvshmem_team_t>src_team, src_pe, <nvshmem_team_t>dest_team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_translate_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_35team_translate_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_34team_translate_pe, "team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_35team_translate_pe = {"team_translate_pe", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_35team_translate_pe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_34team_translate_pe};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_35team_translate_pe(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_src_team;
  int __pyx_v_src_pe;
  int32_t __pyx_v_dest_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_translate_pe (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src_team,&__pyx_mstate_global->__pyx_n_u_src_pe,&__pyx_mstate_global->__pyx_n_u_dest_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 496, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 496, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 496, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 496, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_translate_pe", 0) < (0)) __PYX_ERR(0, 496, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_translate_pe", 1, 3, 3, i); __PYX_ERR(0, 496, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 496, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 496, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 496, __pyx_L3_error)
    }
    __pyx_v_src_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_src_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 496, __pyx_L3_error)
    __pyx_v_src_pe = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_src_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 496, __pyx_L3_error)
    __pyx_v_dest_team = __Pyx_PyLong_As_int32_t(values[2]); if (unlikely((__pyx_v_dest_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 496, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_translate_pe", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 496, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_translate_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_34team_translate_pe(__pyx_self, __pyx_v_src_team, __pyx_v_src_pe, __pyx_v_dest_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_34team_translate_pe(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_src_team, int __pyx_v_src_pe, int32_t __pyx_v_dest_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_translate_pe", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_translate_pe(__pyx_v_src_team, __pyx_v_src_pe, __pyx_v_dest_team, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 496, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_translate_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":500
 * 
 * 
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_37team_split_strided(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_strided(int32_t __pyx_v_parent_team, int __pyx_v_pe_start, int __pyx_v_pe_stride, int __pyx_v_pe_size, intptr_t __pyx_v_config, long __pyx_v_config_mask, intptr_t __pyx_v_new_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_split_strided", 0);

  /* "nvshmem/bindings/nvshmem.pyx":501
 * 
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":502
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):
 *     with nogil:
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_strided(((nvshmem_team_t)__pyx_v_parent_team), __pyx_v_pe_start, __pyx_v_pe_stride, __pyx_v_pe_size, ((nvshmem_team_config_t const *)__pyx_v_config), __pyx_v_config_mask, ((nvshmem_team_t *)__pyx_v_new_team)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 502, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":501
 * 
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":503
 *     with nogil:
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":500
 * 
 * 
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_strided", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_37team_split_strided(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_36team_split_strided, "team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_37team_split_strided = {"team_split_strided", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_37team_split_strided, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_36team_split_strided};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_37team_split_strided(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_parent_team;
  int __pyx_v_pe_start;
  int __pyx_v_pe_stride;
  int __pyx_v_pe_size;
  intptr_t __pyx_v_config;
  long __pyx_v_config_mask;
  intptr_t __pyx_v_new_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_split_strided (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_parent_team,&__pyx_mstate_global->__pyx_n_u_pe_start,&__pyx_mstate_global->__pyx_n_u_pe_stride,&__pyx_mstate_global->__pyx_n_u_pe_size,&__pyx_mstate_global->__pyx_n_u_config,&__pyx_mstate_global->__pyx_n_u_config_mask,&__pyx_mstate_global->__pyx_n_u_new_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 500, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_split_strided", 0) < (0)) __PYX_ERR(0, 500, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 7; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_split_strided", 1, 7, 7, i); __PYX_ERR(0, 500, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 7)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 500, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 500, __pyx_L3_error)
    }
    __pyx_v_parent_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_parent_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_pe_start = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_pe_start == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_pe_stride = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_pe_stride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_pe_size = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_pe_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_config = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_config == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_config_mask = __Pyx_PyLong_As_long(values[5]); if (unlikely((__pyx_v_config_mask == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_new_team = PyLong_AsSsize_t(values[6]); if (unlikely((__pyx_v_new_team == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_split_strided", 1, 7, 7, __pyx_nargs); __PYX_ERR(0, 500, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_strided", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_36team_split_strided(__pyx_self, __pyx_v_parent_team, __pyx_v_pe_start, __pyx_v_pe_stride, __pyx_v_pe_size, __pyx_v_config, __pyx_v_config_mask, __pyx_v_new_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_36team_split_strided(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_parent_team, int __pyx_v_pe_start, int __pyx_v_pe_stride, int __pyx_v_pe_size, intptr_t __pyx_v_config, long __pyx_v_config_mask, intptr_t __pyx_v_new_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_split_strided", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_split_strided(__pyx_v_parent_team, __pyx_v_pe_start, __pyx_v_pe_stride, __pyx_v_pe_size, __pyx_v_config, __pyx_v_config_mask, __pyx_v_new_team, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_strided", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":506
 * 
 * 
 * cpdef team_get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_get_uniqueid(intptr_t __pyx_v_uniqueid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_get_uniqueid", 0);

  /* "nvshmem/bindings/nvshmem.pyx":507
 * 
 * cpdef team_get_uniqueid(intptr_t uniqueid):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":508
 * cpdef team_get_uniqueid(intptr_t uniqueid):
 *     with nogil:
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_get_uniqueid(((nvshmemx_team_uniqueid_t *)__pyx_v_uniqueid)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 508, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":507
 * 
 * cpdef team_get_uniqueid(intptr_t uniqueid):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":509
 *     with nogil:
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":506
 * 
 * 
 * cpdef team_get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_38team_get_uniqueid, "team_get_uniqueid(intptr_t uniqueid)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid = {"team_get_uniqueid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_38team_get_uniqueid};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_uniqueid;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_get_uniqueid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_uniqueid,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 506, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_get_uniqueid", 0) < (0)) __PYX_ERR(0, 506, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_get_uniqueid", 1, 1, 1, i); __PYX_ERR(0, 506, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 506, __pyx_L3_error)
    }
    __pyx_v_uniqueid = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_uniqueid == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 506, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_get_uniqueid", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 506, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_38team_get_uniqueid(__pyx_self, __pyx_v_uniqueid);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_38team_get_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_uniqueid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_get_uniqueid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_get_uniqueid(__pyx_v_uniqueid, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":512
 * 
 * 
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_41team_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_init(intptr_t __pyx_v_team, intptr_t __pyx_v_config, long __pyx_v_config_mask, int __pyx_v_npes, int __pyx_v_pe_idx_in_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_init", 0);

  /* "nvshmem/bindings/nvshmem.pyx":513
 * 
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":514
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):
 *     with nogil:
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_init(((nvshmem_team_t *)__pyx_v_team), ((nvshmem_team_config_t *)__pyx_v_config), __pyx_v_config_mask, __pyx_v_npes, __pyx_v_pe_idx_in_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 514, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":513
 * 
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":515
 *     with nogil:
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":512
 * 
 * 
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_41team_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_40team_init, "team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_41team_init = {"team_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_41team_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_40team_init};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_41team_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_team;
  intptr_t __pyx_v_config;
  long __pyx_v_config_mask;
  int __pyx_v_npes;
  int __pyx_v_pe_idx_in_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_init (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_config,&__pyx_mstate_global->__pyx_n_u_config_mask,&__pyx_mstate_global->__pyx_n_u_npes,&__pyx_mstate_global->__pyx_n_u_pe_idx_in_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 512, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_init", 0) < (0)) __PYX_ERR(0, 512, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_init", 1, 5, 5, i); __PYX_ERR(0, 512, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 512, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 512, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 512, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 512, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 512, __pyx_L3_error)
    }
    __pyx_v_team = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_team == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
    __pyx_v_config = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_config == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
    __pyx_v_config_mask = __Pyx_PyLong_As_long(values[2]); if (unlikely((__pyx_v_config_mask == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
    __pyx_v_npes = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_npes == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
    __pyx_v_pe_idx_in_team = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_idx_in_team == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_init", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 512, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_40team_init(__pyx_self, __pyx_v_team, __pyx_v_config, __pyx_v_config_mask, __pyx_v_npes, __pyx_v_pe_idx_in_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_40team_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_team, intptr_t __pyx_v_config, long __pyx_v_config_mask, int __pyx_v_npes, int __pyx_v_pe_idx_in_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_init", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_init(__pyx_v_team, __pyx_v_config, __pyx_v_config_mask, __pyx_v_npes, __pyx_v_pe_idx_in_team, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":518
 * 
 * 
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_43team_split_2d(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_2d(int32_t __pyx_v_parent_team, int __pyx_v_xrange, intptr_t __pyx_v_xaxis_config, long __pyx_v_xaxis_mask, intptr_t __pyx_v_xaxis_team, intptr_t __pyx_v_yaxis_config, long __pyx_v_yaxis_mask, intptr_t __pyx_v_yaxis_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_split_2d", 0);

  /* "nvshmem/bindings/nvshmem.pyx":519
 * 
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":520
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):
 *     with nogil:
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_2d(((nvshmem_team_t)__pyx_v_parent_team), __pyx_v_xrange, ((nvshmem_team_config_t const *)__pyx_v_xaxis_config), __pyx_v_xaxis_mask, ((nvshmem_team_t *)__pyx_v_xaxis_team), ((nvshmem_team_config_t const *)__pyx_v_yaxis_config), __pyx_v_yaxis_mask, ((nvshmem_team_t *)__pyx_v_yaxis_team)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 520, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":519
 * 
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":521
 *     with nogil:
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":518
 * 
 * 
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_43team_split_2d(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_42team_split_2d, "team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_43team_split_2d = {"team_split_2d", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_43team_split_2d, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_42team_split_2d};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_43team_split_2d(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_parent_team;
  int __pyx_v_xrange;
  intptr_t __pyx_v_xaxis_config;
  long __pyx_v_xaxis_mask;
  intptr_t __pyx_v_xaxis_team;
  intptr_t __pyx_v_yaxis_config;
  long __pyx_v_yaxis_mask;
  intptr_t __pyx_v_yaxis_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_split_2d (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_parent_team,&__pyx_mstate_global->__pyx_n_u_xrange,&__pyx_mstate_global->__pyx_n_u_xaxis_config,&__pyx_mstate_global->__pyx_n_u_xaxis_mask,&__pyx_mstate_global->__pyx_n_u_xaxis_team,&__pyx_mstate_global->__pyx_n_u_yaxis_config,&__pyx_mstate_global->__pyx_n_u_yaxis_mask,&__pyx_mstate_global->__pyx_n_u_yaxis_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 518, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_split_2d", 0) < (0)) __PYX_ERR(0, 518, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 8; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_split_2d", 1, 8, 8, i); __PYX_ERR(0, 518, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 8)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 518, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 518, __pyx_L3_error)
    }
    __pyx_v_parent_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_parent_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_xrange = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_xrange == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_xaxis_config = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_xaxis_config == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_xaxis_mask = __Pyx_PyLong_As_long(values[3]); if (unlikely((__pyx_v_xaxis_mask == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_xaxis_team = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_xaxis_team == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_yaxis_config = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_yaxis_config == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_yaxis_mask = __Pyx_PyLong_As_long(values[6]); if (unlikely((__pyx_v_yaxis_mask == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_yaxis_team = PyLong_AsSsize_t(values[7]); if (unlikely((__pyx_v_yaxis_team == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_split_2d", 1, 8, 8, __pyx_nargs); __PYX_ERR(0, 518, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_42team_split_2d(__pyx_self, __pyx_v_parent_team, __pyx_v_xrange, __pyx_v_xaxis_config, __pyx_v_xaxis_mask, __pyx_v_xaxis_team, __pyx_v_yaxis_config, __pyx_v_yaxis_mask, __pyx_v_yaxis_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_42team_split_2d(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_parent_team, int __pyx_v_xrange, intptr_t __pyx_v_xaxis_config, long __pyx_v_xaxis_mask, intptr_t __pyx_v_xaxis_team, intptr_t __pyx_v_yaxis_config, long __pyx_v_yaxis_mask, intptr_t __pyx_v_yaxis_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_split_2d", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_split_2d(__pyx_v_parent_team, __pyx_v_xrange, __pyx_v_xaxis_config, __pyx_v_xaxis_mask, __pyx_v_xaxis_team, __pyx_v_yaxis_config, __pyx_v_yaxis_mask, __pyx_v_yaxis_team, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_split_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":524
 * 
 * 
 * cpdef void team_destroy(int32_t team) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_destroy(<nvshmem_team_t>team)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_45team_destroy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_team_destroy(int32_t __pyx_v_team, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":525
 * 
 * cpdef void team_destroy(int32_t team) except*:
 *     nvshmem_team_destroy(<nvshmem_team_t>team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_destroy(((nvshmem_team_t)__pyx_v_team)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":524
 * 
 * 
 * cpdef void team_destroy(int32_t team) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_destroy(<nvshmem_team_t>team)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_45team_destroy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_44team_destroy, "team_destroy(int32_t team) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_45team_destroy = {"team_destroy", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_45team_destroy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_44team_destroy};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_45team_destroy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_destroy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 524, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_destroy", 0) < (0)) __PYX_ERR(0, 524, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_destroy", 1, 1, 1, i); __PYX_ERR(0, 524, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 524, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_destroy", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 524, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_44team_destroy(__pyx_self, __pyx_v_team);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_44team_destroy(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_destroy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_team_destroy(__pyx_v_team, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":528
 * 
 * 
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":529
 * 
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":530
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 530, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":529
 * 
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":531
 *     with nogil:
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":528
 * 
 * 
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_46bfloat16_alltoall_on_stream, "bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream = {"bfloat16_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_46bfloat16_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 528, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 528, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 528, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 528, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 528, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 528, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 528, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 528, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 528, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 528, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 528, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 528, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 528, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 528, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_46bfloat16_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_46bfloat16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":534
 * 
 * 
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":535
 * 
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":536
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 536, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":535
 * 
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":537
 *     with nogil:
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":534
 * 
 * 
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_48half_alltoall_on_stream, "half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream = {"half_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_48half_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 534, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 534, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 534, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 534, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 534, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 534, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 534, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 534, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 534, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 534, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 534, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 534, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 534, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 534, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_48half_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_48half_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":540
 * 
 * 
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":541
 * 
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":542
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 542, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":541
 * 
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":543
 *     with nogil:
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":540
 * 
 * 
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_50float_alltoall_on_stream, "float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream = {"float_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_50float_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 540, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 540, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 540, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 540, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 540, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 540, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 540, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 540, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 540, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 540, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 540, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 540, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 540, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 540, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_50float_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_50float_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":546
 * 
 * 
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":547
 * 
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":548
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 548, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":547
 * 
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":549
 *     with nogil:
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":546
 * 
 * 
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_52double_alltoall_on_stream, "double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream = {"double_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_52double_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 546, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 546, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 546, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 546, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 546, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 546, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 546, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 546, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 546, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_52double_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_52double_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":552
 * 
 * 
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":553
 * 
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":554
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 554, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":553
 * 
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":555
 *     with nogil:
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":552
 * 
 * 
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_54char_alltoall_on_stream, "char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream = {"char_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_54char_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 552, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 552, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 552, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 552, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 552, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 552, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 552, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 552, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 552, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_54char_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_54char_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":558
 * 
 * 
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":559
 * 
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":560
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 560, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":559
 * 
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":561
 *     with nogil:
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":558
 * 
 * 
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_56short_alltoall_on_stream, "short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream = {"short_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_56short_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 558, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 558, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 558, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 558, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 558, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 558, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 558, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 558, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 558, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 558, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 558, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 558, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 558, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 558, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_56short_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_56short_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":564
 * 
 * 
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":565
 * 
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":566
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 566, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":565
 * 
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":567
 *     with nogil:
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":564
 * 
 * 
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_58schar_alltoall_on_stream, "schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream = {"schar_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_58schar_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 564, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 564, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 564, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 564, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 564, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 564, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 564, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 564, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 564, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 564, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 564, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 564, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 564, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 564, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_58schar_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_58schar_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":570
 * 
 * 
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":571
 * 
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":572
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 572, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":571
 * 
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":573
 *     with nogil:
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":570
 * 
 * 
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_60int_alltoall_on_stream, "int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream = {"int_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_60int_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 570, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 570, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 570, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 570, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 570, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 570, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 570, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 570, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 570, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 570, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 570, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 570, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_60int_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_60int_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":576
 * 
 * 
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":577
 * 
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":578
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 578, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":577
 * 
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":579
 *     with nogil:
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":576
 * 
 * 
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_62long_alltoall_on_stream, "long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream = {"long_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_62long_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 576, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 576, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 576, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 576, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 576, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 576, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 576, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 576, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 576, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_62long_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_62long_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":582
 * 
 * 
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":583
 * 
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":584
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 584, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":583
 * 
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":585
 *     with nogil:
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":582
 * 
 * 
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_64longlong_alltoall_on_stream, "longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream = {"longlong_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_64longlong_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 582, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 582, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 582, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 582, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 582, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 582, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 582, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 582, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 582, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_64longlong_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_64longlong_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":588
 * 
 * 
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":589
 * 
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":590
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 590, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":589
 * 
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":591
 *     with nogil:
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":588
 * 
 * 
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_66int8_alltoall_on_stream, "int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream = {"int8_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_66int8_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 588, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 588, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 588, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 588, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 588, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 588, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 588, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 588, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 588, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_66int8_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_66int8_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":594
 * 
 * 
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":595
 * 
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":596
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 596, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":595
 * 
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":597
 *     with nogil:
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":594
 * 
 * 
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_68int16_alltoall_on_stream, "int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream = {"int16_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_68int16_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 594, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 594, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 594, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 594, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 594, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 594, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 594, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 594, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 594, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_68int16_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_68int16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":600
 * 
 * 
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":601
 * 
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":602
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 602, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":601
 * 
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":603
 *     with nogil:
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":600
 * 
 * 
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_70int32_alltoall_on_stream, "int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream = {"int32_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_70int32_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 600, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 600, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 600, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 600, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 600, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 600, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 600, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 600, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_70int32_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_70int32_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":606
 * 
 * 
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":607
 * 
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":608
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 608, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":607
 * 
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":609
 *     with nogil:
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":606
 * 
 * 
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_72int64_alltoall_on_stream, "int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream = {"int64_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_72int64_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 606, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 606, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 606, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 606, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 606, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 606, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 606, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 606, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 606, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 606, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 606, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 606, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 606, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 606, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_72int64_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_72int64_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":612
 * 
 * 
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":613
 * 
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":614
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 614, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":613
 * 
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":615
 *     with nogil:
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":612
 * 
 * 
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_74uint8_alltoall_on_stream, "uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream = {"uint8_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_74uint8_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 612, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 612, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 612, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 612, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 612, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 612, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 612, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 612, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 612, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 612, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 612, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 612, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 612, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 612, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_74uint8_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_74uint8_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":618
 * 
 * 
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":619
 * 
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":620
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 620, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":619
 * 
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":621
 *     with nogil:
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":618
 * 
 * 
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_76uint16_alltoall_on_stream, "uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream = {"uint16_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_76uint16_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 618, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 618, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 618, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 618, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 618, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 618, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 618, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 618, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 618, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 618, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 618, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 618, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 618, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 618, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_76uint16_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_76uint16_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":624
 * 
 * 
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":625
 * 
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":626
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 626, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":625
 * 
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":627
 *     with nogil:
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":624
 * 
 * 
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_78uint32_alltoall_on_stream, "uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream = {"uint32_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_78uint32_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 624, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 624, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 624, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 624, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 624, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 624, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 624, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 624, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 624, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_78uint32_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_78uint32_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":630
 * 
 * 
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":631
 * 
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":632
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 632, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":631
 * 
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":633
 *     with nogil:
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":630
 * 
 * 
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_80uint64_alltoall_on_stream, "uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream = {"uint64_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_80uint64_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 630, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 630, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 630, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 630, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 630, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 630, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 630, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 630, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 630, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 630, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 630, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 630, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 630, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 630, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_80uint64_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_80uint64_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":636
 * 
 * 
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_alltoall_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_alltoall_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":637
 * 
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":638
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_alltoall_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 638, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":637
 * 
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":639
 *     with nogil:
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":636
 * 
 * 
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_82size_alltoall_on_stream, "size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream = {"size_alltoall_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_82size_alltoall_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_alltoall_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 636, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_alltoall_on_stream", 0) < (0)) __PYX_ERR(0, 636, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_alltoall_on_stream", 1, 5, 5, i); __PYX_ERR(0, 636, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 636, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 636, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 636, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 636, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 636, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_alltoall_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 636, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_82size_alltoall_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_82size_alltoall_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_alltoall_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":642
 * 
 * 
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_85barrier_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_barrier_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":643
 * 
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":644
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_on_stream(((nvshmem_team_t)__pyx_v_team), ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 644, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":643
 * 
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":645
 *     with nogil:
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":642
 * 
 * 
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_85barrier_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_84barrier_on_stream, "barrier_on_stream(int32_t team, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_85barrier_on_stream = {"barrier_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_85barrier_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_84barrier_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_85barrier_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("barrier_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 642, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 642, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 642, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "barrier_on_stream", 0) < (0)) __PYX_ERR(0, 642, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("barrier_on_stream", 1, 2, 2, i); __PYX_ERR(0, 642, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 642, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 642, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 642, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 642, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("barrier_on_stream", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 642, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_84barrier_on_stream(__pyx_self, __pyx_v_team, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_84barrier_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_on_stream(__pyx_v_team, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":648
 * 
 * 
 * cpdef void barrier_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_barrier_all_on_stream(<Stream>stream)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all_on_stream(intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":649
 * 
 * cpdef void barrier_all_on_stream(intptr_t stream) except*:
 *     nvshmemx_barrier_all_on_stream(<Stream>stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_all_on_stream(((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":648
 * 
 * 
 * cpdef void barrier_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_barrier_all_on_stream(<Stream>stream)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_86barrier_all_on_stream, "barrier_all_on_stream(intptr_t stream) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream = {"barrier_all_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_86barrier_all_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("barrier_all_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 648, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 648, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "barrier_all_on_stream", 0) < (0)) __PYX_ERR(0, 648, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("barrier_all_on_stream", 1, 1, 1, i); __PYX_ERR(0, 648, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 648, __pyx_L3_error)
    }
    __pyx_v_stream = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("barrier_all_on_stream", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 648, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_86barrier_all_on_stream(__pyx_self, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_86barrier_all_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("barrier_all_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all_on_stream(__pyx_v_stream, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 648, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.barrier_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":652
 * 
 * 
 * cpdef int team_sync_on_stream(int32_t team, intptr_t stream) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_team_sync_on_stream(<nvshmem_team_t>team, <Stream>stream)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_team_sync_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":653
 * 
 * cpdef int team_sync_on_stream(int32_t team, intptr_t stream) except? 0:
 *     return nvshmemx_team_sync_on_stream(<nvshmem_team_t>team, <Stream>stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_sync_on_stream(((nvshmem_team_t)__pyx_v_team), ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":652
 * 
 * 
 * cpdef int team_sync_on_stream(int32_t team, intptr_t stream) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_team_sync_on_stream(<nvshmem_team_t>team, <Stream>stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_sync_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_88team_sync_on_stream, "team_sync_on_stream(int32_t team, intptr_t stream) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream = {"team_sync_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_88team_sync_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("team_sync_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 652, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 652, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 652, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "team_sync_on_stream", 0) < (0)) __PYX_ERR(0, 652, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("team_sync_on_stream", 1, 2, 2, i); __PYX_ERR(0, 652, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 652, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 652, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("team_sync_on_stream", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 652, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_sync_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_88team_sync_on_stream(__pyx_self, __pyx_v_team, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_88team_sync_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("team_sync_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_team_sync_on_stream(__pyx_v_team, __pyx_v_stream, 1); if (unlikely(__pyx_t_1 == ((int)0) && PyErr_Occurred())) __PYX_ERR(0, 652, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.team_sync_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":656
 * 
 * 
 * cpdef void sync_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_sync_all_on_stream(<Stream>stream)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_sync_all_on_stream(intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":657
 * 
 * cpdef void sync_all_on_stream(intptr_t stream) except*:
 *     nvshmemx_sync_all_on_stream(<Stream>stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_sync_all_on_stream(((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":656
 * 
 * 
 * cpdef void sync_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_sync_all_on_stream(<Stream>stream)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.sync_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_90sync_all_on_stream, "sync_all_on_stream(intptr_t stream) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream = {"sync_all_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_90sync_all_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sync_all_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 656, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 656, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "sync_all_on_stream", 0) < (0)) __PYX_ERR(0, 656, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("sync_all_on_stream", 1, 1, 1, i); __PYX_ERR(0, 656, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 656, __pyx_L3_error)
    }
    __pyx_v_stream = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sync_all_on_stream", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 656, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.sync_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_90sync_all_on_stream(__pyx_self, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_90sync_all_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sync_all_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_sync_all_on_stream(__pyx_v_stream, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.sync_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":660
 * 
 * 
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":661
 * 
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":662
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 662, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":661
 * 
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":663
 *     with nogil:
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":660
 * 
 * 
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_92bfloat16_broadcast_on_stream, "bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream = {"bfloat16_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_92bfloat16_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 660, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 660, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 660, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 660, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 660, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 660, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 660, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 660, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 660, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 660, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 660, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_92bfloat16_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_92bfloat16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":666
 * 
 * 
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":667
 * 
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":668
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 668, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":667
 * 
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":669
 *     with nogil:
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":666
 * 
 * 
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_94half_broadcast_on_stream, "half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream = {"half_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_94half_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 666, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 666, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 666, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 666, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 666, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 666, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 666, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 666, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 666, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 666, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 666, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 666, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_94half_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_94half_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":672
 * 
 * 
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":673
 * 
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":674
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 674, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":673
 * 
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":675
 *     with nogil:
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":672
 * 
 * 
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_96float_broadcast_on_stream, "float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream = {"float_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_96float_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 672, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 672, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 672, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 672, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 672, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 672, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 672, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 672, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 672, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 672, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 672, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_96float_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_96float_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":678
 * 
 * 
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":679
 * 
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":680
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 680, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":679
 * 
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":681
 *     with nogil:
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":678
 * 
 * 
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_98double_broadcast_on_stream, "double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream = {"double_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_98double_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 678, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 678, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 678, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 678, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 678, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_98double_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_98double_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":684
 * 
 * 
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":685
 * 
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":686
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 686, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":685
 * 
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":687
 *     with nogil:
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":684
 * 
 * 
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_100char_broadcast_on_stream, "char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream = {"char_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_100char_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 684, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 684, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 684, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 684, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 684, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 684, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 684, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 684, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 684, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 684, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 684, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_100char_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_100char_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":690
 * 
 * 
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":691
 * 
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":692
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 692, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":691
 * 
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":693
 *     with nogil:
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":690
 * 
 * 
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_102short_broadcast_on_stream, "short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream = {"short_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_102short_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 690, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 690, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 690, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 690, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 690, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 690, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 690, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 690, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 690, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 690, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_102short_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_102short_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":696
 * 
 * 
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":697
 * 
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":698
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 698, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":697
 * 
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":699
 *     with nogil:
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":696
 * 
 * 
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_104schar_broadcast_on_stream, "schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream = {"schar_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_104schar_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 696, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 696, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 696, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 696, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 696, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 696, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 696, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 696, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 696, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 696, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_104schar_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_104schar_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":702
 * 
 * 
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":703
 * 
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":704
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 704, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":703
 * 
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":705
 *     with nogil:
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":702
 * 
 * 
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_106int_broadcast_on_stream, "int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream = {"int_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_106int_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 702, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 702, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 702, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 702, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 702, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 702, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 702, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_106int_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_106int_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":708
 * 
 * 
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":709
 * 
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":710
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 710, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":709
 * 
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":711
 *     with nogil:
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":708
 * 
 * 
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_108long_broadcast_on_stream, "long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream = {"long_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_108long_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 708, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 708, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 708, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 708, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 708, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 708, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 708, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 708, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 708, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 708, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 708, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_108long_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_108long_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":714
 * 
 * 
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":715
 * 
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":716
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 716, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":715
 * 
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":717
 *     with nogil:
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":714
 * 
 * 
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_110longlong_broadcast_on_stream, "longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream = {"longlong_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_110longlong_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 714, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 714, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 714, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 714, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 714, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 714, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 714, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 714, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 714, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 714, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 714, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_110longlong_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_110longlong_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":720
 * 
 * 
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":721
 * 
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":722
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 722, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":721
 * 
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":723
 *     with nogil:
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":720
 * 
 * 
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_112int8_broadcast_on_stream, "int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream = {"int8_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_112int8_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 720, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 720, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 720, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 720, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 720, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 720, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 720, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 720, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 720, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 720, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 720, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_112int8_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_112int8_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":726
 * 
 * 
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":727
 * 
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":728
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 728, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":727
 * 
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":729
 *     with nogil:
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":726
 * 
 * 
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_114int16_broadcast_on_stream, "int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream = {"int16_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_114int16_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 726, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 726, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 726, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 726, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 726, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 726, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 726, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 726, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 726, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 726, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 726, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_114int16_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_114int16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":732
 * 
 * 
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":733
 * 
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":734
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 734, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":733
 * 
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":735
 *     with nogil:
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":732
 * 
 * 
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_116int32_broadcast_on_stream, "int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream = {"int32_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_116int32_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 732, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 732, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 732, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 732, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 732, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 732, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 732, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 732, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 732, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 732, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_116int32_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_116int32_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":738
 * 
 * 
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":739
 * 
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":740
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 740, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":739
 * 
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":741
 *     with nogil:
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":738
 * 
 * 
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_118int64_broadcast_on_stream, "int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream = {"int64_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_118int64_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 738, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 738, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 738, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 738, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 738, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 738, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 738, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 738, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 738, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 738, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 738, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_118int64_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_118int64_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":744
 * 
 * 
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":745
 * 
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":746
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 746, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":745
 * 
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":747
 *     with nogil:
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":744
 * 
 * 
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_120uint8_broadcast_on_stream, "uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream = {"uint8_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_120uint8_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 744, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 744, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 744, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 744, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 744, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 744, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 744, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_120uint8_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_120uint8_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":750
 * 
 * 
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":751
 * 
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":752
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 752, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":751
 * 
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":753
 *     with nogil:
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":750
 * 
 * 
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_122uint16_broadcast_on_stream, "uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream = {"uint16_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_122uint16_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 750, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 750, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 750, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 750, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 750, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 750, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 750, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 750, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 750, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 750, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 750, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_122uint16_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_122uint16_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":756
 * 
 * 
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":757
 * 
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":758
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 758, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":757
 * 
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":759
 *     with nogil:
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":756
 * 
 * 
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_124uint32_broadcast_on_stream, "uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream = {"uint32_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_124uint32_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 756, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 756, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 756, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 756, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 756, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 756, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 756, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 756, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 756, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 756, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 756, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_124uint32_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_124uint32_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":762
 * 
 * 
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":763
 * 
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":764
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 764, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":763
 * 
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":765
 *     with nogil:
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":762
 * 
 * 
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_126uint64_broadcast_on_stream, "uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream = {"uint64_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_126uint64_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 762, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 762, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 762, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 762, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 762, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 762, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 762, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 762, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 762, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 762, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 762, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_126uint64_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_126uint64_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":768
 * 
 * 
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_broadcast_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_broadcast_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":769
 * 
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":770
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_broadcast_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nelem, __pyx_v_pe_root, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 770, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":769
 * 
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":771
 *     with nogil:
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":768
 * 
 * 
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_128size_broadcast_on_stream, "size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream = {"size_broadcast_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_128size_broadcast_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  int __pyx_v_pe_root;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_broadcast_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_pe_root,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 768, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 768, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_broadcast_on_stream", 0) < (0)) __PYX_ERR(0, 768, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_broadcast_on_stream", 1, 6, 6, i); __PYX_ERR(0, 768, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 768, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 768, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 768, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 768, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 768, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 768, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
    __pyx_v_pe_root = __Pyx_PyLong_As_int(values[4]); if (unlikely((__pyx_v_pe_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[5]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_broadcast_on_stream", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 768, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_128size_broadcast_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_128size_broadcast_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_pe_root, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_broadcast_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_pe_root, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":774
 * 
 * 
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":775
 * 
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":776
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 776, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":775
 * 
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":777
 *     with nogil:
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":774
 * 
 * 
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_130bfloat16_fcollect_on_stream, "bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream = {"bfloat16_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_130bfloat16_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 774, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 774, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 774, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 774, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 774, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 774, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 774, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 774, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 774, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 774, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 774, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 774, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 774, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 774, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_130bfloat16_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_130bfloat16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":780
 * 
 * 
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":781
 * 
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":782
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 782, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":781
 * 
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":783
 *     with nogil:
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":780
 * 
 * 
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_132half_fcollect_on_stream, "half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream = {"half_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_132half_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 780, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 780, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 780, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 780, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 780, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 780, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 780, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 780, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 780, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_132half_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_132half_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":786
 * 
 * 
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":787
 * 
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":788
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 788, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":787
 * 
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":789
 *     with nogil:
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":786
 * 
 * 
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_134float_fcollect_on_stream, "float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream = {"float_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_134float_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 786, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 786, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 786, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 786, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 786, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 786, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 786, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 786, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 786, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 786, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 786, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 786, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 786, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 786, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_134float_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_134float_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":792
 * 
 * 
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":793
 * 
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":794
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 794, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":793
 * 
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":795
 *     with nogil:
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":792
 * 
 * 
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_136double_fcollect_on_stream, "double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream = {"double_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_136double_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 792, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 792, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 792, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 792, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 792, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 792, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 792, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 792, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 792, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 792, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 792, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 792, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 792, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 792, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_136double_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_136double_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":798
 * 
 * 
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":799
 * 
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":800
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 800, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":799
 * 
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":801
 *     with nogil:
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":798
 * 
 * 
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_138char_fcollect_on_stream, "char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream = {"char_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_138char_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 798, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 798, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 798, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 798, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 798, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 798, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 798, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 798, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 798, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 798, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 798, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 798, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 798, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 798, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_138char_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_138char_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":804
 * 
 * 
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":805
 * 
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":806
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 806, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":805
 * 
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":807
 *     with nogil:
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":804
 * 
 * 
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_140short_fcollect_on_stream, "short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream = {"short_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_140short_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 804, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 804, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 804, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 804, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 804, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 804, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 804, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_140short_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_140short_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":810
 * 
 * 
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":811
 * 
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":812
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 812, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":811
 * 
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":813
 *     with nogil:
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":810
 * 
 * 
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_142schar_fcollect_on_stream, "schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream = {"schar_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_142schar_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 810, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 810, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 810, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 810, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 810, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 810, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 810, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 810, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 810, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 810, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 810, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 810, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 810, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 810, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_142schar_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_142schar_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":816
 * 
 * 
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":817
 * 
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":818
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 818, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":817
 * 
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":819
 *     with nogil:
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":816
 * 
 * 
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_144int_fcollect_on_stream, "int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream = {"int_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_144int_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 816, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 816, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 816, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 816, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 816, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 816, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 816, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 816, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 816, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 816, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 816, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 816, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 816, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 816, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 816, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_144int_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_144int_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":822
 * 
 * 
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":823
 * 
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":824
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 824, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":823
 * 
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":825
 *     with nogil:
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":822
 * 
 * 
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_146long_fcollect_on_stream, "long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream = {"long_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_146long_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 822, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 822, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 822, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 822, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 822, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 822, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 822, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 822, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 822, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 822, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 822, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 822, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 822, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 822, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_146long_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_146long_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":828
 * 
 * 
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":829
 * 
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":830
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 830, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":829
 * 
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":831
 *     with nogil:
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":828
 * 
 * 
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_148longlong_fcollect_on_stream, "longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream = {"longlong_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_148longlong_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 828, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 828, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 828, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 828, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 828, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 828, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 828, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 828, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 828, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 828, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 828, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 828, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 828, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 828, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_148longlong_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_148longlong_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":834
 * 
 * 
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":835
 * 
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":836
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 836, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":835
 * 
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":837
 *     with nogil:
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":834
 * 
 * 
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_150int8_fcollect_on_stream, "int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream = {"int8_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_150int8_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 834, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 834, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 834, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 834, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 834, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 834, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 834, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 834, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 834, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 834, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 834, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 834, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 834, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 834, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_150int8_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_150int8_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":840
 * 
 * 
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":841
 * 
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":842
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 842, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":841
 * 
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":843
 *     with nogil:
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":840
 * 
 * 
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_152int16_fcollect_on_stream, "int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream = {"int16_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_152int16_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 840, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 840, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 840, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 840, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 840, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 840, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 840, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 840, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 840, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_152int16_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_152int16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":846
 * 
 * 
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":847
 * 
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":848
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 848, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":847
 * 
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":849
 *     with nogil:
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":846
 * 
 * 
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_154int32_fcollect_on_stream, "int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream = {"int32_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_154int32_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 846, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 846, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 846, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 846, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 846, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 846, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 846, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 846, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 846, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 846, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 846, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 846, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 846, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 846, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_154int32_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_154int32_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":852
 * 
 * 
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":853
 * 
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":854
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 854, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":853
 * 
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":855
 *     with nogil:
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":852
 * 
 * 
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_156int64_fcollect_on_stream, "int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream = {"int64_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_156int64_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 852, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 852, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 852, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 852, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 852, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 852, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 852, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 852, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 852, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 852, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 852, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 852, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 852, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 852, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_156int64_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_156int64_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":858
 * 
 * 
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":859
 * 
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":860
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 860, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":859
 * 
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":861
 *     with nogil:
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":858
 * 
 * 
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_158uint8_fcollect_on_stream, "uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream = {"uint8_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_158uint8_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 858, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 858, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 858, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 858, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 858, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 858, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 858, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 858, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 858, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 858, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 858, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 858, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 858, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 858, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_158uint8_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_158uint8_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":864
 * 
 * 
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":865
 * 
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":866
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 866, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":865
 * 
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":867
 *     with nogil:
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":864
 * 
 * 
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_160uint16_fcollect_on_stream, "uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream = {"uint16_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_160uint16_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 864, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 864, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 864, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 864, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 864, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 864, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 864, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 864, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 864, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 864, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 864, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 864, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 864, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 864, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_160uint16_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_160uint16_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":870
 * 
 * 
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":871
 * 
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":872
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 872, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":871
 * 
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":873
 *     with nogil:
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":870
 * 
 * 
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_162uint32_fcollect_on_stream, "uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream = {"uint32_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_162uint32_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 870, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 870, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 870, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 870, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 870, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 870, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 870, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 870, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 870, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 870, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 870, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 870, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 870, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 870, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_162uint32_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_162uint32_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":876
 * 
 * 
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":877
 * 
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":878
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 878, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":877
 * 
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":879
 *     with nogil:
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":876
 * 
 * 
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_164uint64_fcollect_on_stream, "uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream = {"uint64_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_164uint64_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 876, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 876, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 876, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 876, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 876, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 876, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 876, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 876, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 876, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 876, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 876, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 876, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 876, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 876, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_164uint64_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_164uint64_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":882
 * 
 * 
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_fcollect_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_fcollect_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":883
 * 
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":884
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_fcollect_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nelem, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 884, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":883
 * 
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":885
 *     with nogil:
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":882
 * 
 * 
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_166size_fcollect_on_stream, "size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream = {"size_fcollect_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_166size_fcollect_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nelem;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_fcollect_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nelem,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 882, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 882, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 882, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 882, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 882, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 882, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_fcollect_on_stream", 0) < (0)) __PYX_ERR(0, 882, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_fcollect_on_stream", 1, 5, 5, i); __PYX_ERR(0, 882, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 882, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 882, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 882, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 882, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 882, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L3_error)
    __pyx_v_nelem = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nelem == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_fcollect_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 882, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_166size_fcollect_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_166size_fcollect_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nelem, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_fcollect_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":888
 * 
 * 
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":889
 * 
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":890
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 890, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":889
 * 
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":891
 *     with nogil:
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":888
 * 
 * 
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_168int8_max_reduce_on_stream, "int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream = {"int8_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_168int8_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 888, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 888, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 888, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 888, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 888, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 888, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 888, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 888, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 888, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 888, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 888, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 888, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 888, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 888, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_168int8_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_168int8_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":894
 * 
 * 
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":895
 * 
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":896
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 896, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":895
 * 
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":897
 *     with nogil:
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":894
 * 
 * 
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_170int16_max_reduce_on_stream, "int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream = {"int16_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_170int16_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 894, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 894, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 894, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 894, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 894, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 894, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 894, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 894, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 894, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 894, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 894, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 894, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 894, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 894, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_170int16_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_170int16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":900
 * 
 * 
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":901
 * 
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":902
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 902, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":901
 * 
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":903
 *     with nogil:
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":900
 * 
 * 
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_172int32_max_reduce_on_stream, "int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream = {"int32_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_172int32_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 900, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 900, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 900, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 900, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 900, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 900, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 900, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 900, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 900, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 900, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 900, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 900, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 900, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 900, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_172int32_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_172int32_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":906
 * 
 * 
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":907
 * 
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":908
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 908, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":907
 * 
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":909
 *     with nogil:
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":906
 * 
 * 
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_174int64_max_reduce_on_stream, "int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream = {"int64_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_174int64_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 906, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 906, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 906, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 906, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 906, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 906, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 906, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 906, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 906, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 906, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 906, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 906, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 906, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 906, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_174int64_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_174int64_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":912
 * 
 * 
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":913
 * 
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":914
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 914, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":913
 * 
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":915
 *     with nogil:
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":912
 * 
 * 
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_176uint8_max_reduce_on_stream, "uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream = {"uint8_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_176uint8_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 912, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 912, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 912, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 912, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 912, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_176uint8_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_176uint8_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":918
 * 
 * 
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":919
 * 
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":920
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 920, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":919
 * 
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":921
 *     with nogil:
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":918
 * 
 * 
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_178uint16_max_reduce_on_stream, "uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream = {"uint16_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_178uint16_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 918, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 918, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 918, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 918, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 918, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 918, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 918, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 918, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 918, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 918, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 918, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 918, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 918, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 918, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 918, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 918, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 918, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 918, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 918, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_178uint16_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_178uint16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":924
 * 
 * 
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":925
 * 
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":926
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 926, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":925
 * 
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":927
 *     with nogil:
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":924
 * 
 * 
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_180uint32_max_reduce_on_stream, "uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream = {"uint32_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_180uint32_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 924, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 924, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 924, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 924, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 924, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 924, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 924, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 924, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 924, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 924, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 924, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 924, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 924, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 924, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 924, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_180uint32_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_180uint32_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":930
 * 
 * 
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":931
 * 
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":932
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 932, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":931
 * 
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":933
 *     with nogil:
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":930
 * 
 * 
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_182uint64_max_reduce_on_stream, "uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream = {"uint64_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_182uint64_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 930, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 930, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 930, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 930, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 930, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 930, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 930, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 930, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 930, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 930, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 930, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 930, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 930, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 930, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_182uint64_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_182uint64_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":936
 * 
 * 
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":937
 * 
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":938
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 938, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":937
 * 
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":939
 *     with nogil:
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":936
 * 
 * 
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_184size_max_reduce_on_stream, "size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream = {"size_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_184size_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 936, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 936, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 936, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 936, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 936, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 936, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 936, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 936, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 936, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 936, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 936, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_184size_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_184size_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":942
 * 
 * 
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":943
 * 
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":944
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 944, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":943
 * 
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":945
 *     with nogil:
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":942
 * 
 * 
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_186char_max_reduce_on_stream, "char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream = {"char_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_186char_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 942, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 942, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 942, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 942, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 942, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 942, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 942, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 942, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 942, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 942, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 942, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 942, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 942, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 942, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_186char_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_186char_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":948
 * 
 * 
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":949
 * 
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":950
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 950, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":949
 * 
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":951
 *     with nogil:
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":948
 * 
 * 
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_188schar_max_reduce_on_stream, "schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream = {"schar_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_188schar_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 948, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 948, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 948, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 948, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 948, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 948, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 948, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 948, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 948, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 948, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_188schar_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_188schar_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":954
 * 
 * 
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":955
 * 
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":956
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 956, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":955
 * 
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":957
 *     with nogil:
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":954
 * 
 * 
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_190short_max_reduce_on_stream, "short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream = {"short_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_190short_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 954, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 954, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 954, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 954, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 954, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 954, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 954, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 954, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 954, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 954, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_190short_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_190short_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":960
 * 
 * 
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":961
 * 
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":962
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 962, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":961
 * 
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":963
 *     with nogil:
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":960
 * 
 * 
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_192int_max_reduce_on_stream, "int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream = {"int_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_192int_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 960, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 960, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 960, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 960, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 960, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 960, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 960, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 960, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 960, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 960, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 960, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 960, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 960, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 960, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 960, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_192int_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_192int_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":966
 * 
 * 
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":967
 * 
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":968
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 968, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":967
 * 
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":969
 *     with nogil:
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":966
 * 
 * 
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_194long_max_reduce_on_stream, "long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream = {"long_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_194long_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 966, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 966, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 966, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 966, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 966, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 966, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 966, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 966, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 966, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 966, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_194long_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_194long_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":972
 * 
 * 
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":973
 * 
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":974
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 974, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":973
 * 
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":975
 *     with nogil:
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":972
 * 
 * 
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_196longlong_max_reduce_on_stream, "longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream = {"longlong_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_196longlong_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 972, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 972, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 972, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 972, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 972, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 972, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 972, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 972, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 972, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 972, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 972, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 972, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 972, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 972, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_196longlong_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_196longlong_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":978
 * 
 * 
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":979
 * 
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":980
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 980, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":979
 * 
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":981
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":978
 * 
 * 
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_198bfloat16_max_reduce_on_stream, "bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream = {"bfloat16_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_198bfloat16_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 978, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 978, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 978, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 978, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 978, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 978, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 978, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 978, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 978, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 978, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 978, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 978, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 978, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 978, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 978, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_198bfloat16_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_198bfloat16_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":984
 * 
 * 
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":985
 * 
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":986
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 986, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":985
 * 
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":987
 *     with nogil:
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":984
 * 
 * 
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_200half_max_reduce_on_stream, "half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream = {"half_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_200half_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 984, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 984, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 984, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 984, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 984, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 984, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 984, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 984, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 984, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 984, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 984, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 984, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 984, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 984, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 984, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 984, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 984, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 984, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 984, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_200half_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_200half_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":990
 * 
 * 
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":991
 * 
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":992
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 992, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":991
 * 
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":993
 *     with nogil:
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":990
 * 
 * 
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_202float_max_reduce_on_stream, "float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream = {"float_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_202float_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 990, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 990, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 990, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 990, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 990, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 990, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 990, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 990, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 990, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 990, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 990, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 990, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 990, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 990, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 990, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_202float_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_202float_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":996
 * 
 * 
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_max_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":997
 * 
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":998
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 998, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":997
 * 
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":999
 *     with nogil:
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":996
 * 
 * 
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_204double_max_reduce_on_stream, "double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream = {"double_max_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_204double_max_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_max_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 996, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 996, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 996, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 996, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 996, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 996, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_max_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 996, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_max_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 996, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 996, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 996, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 996, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 996, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 996, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_max_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 996, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_204double_max_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_204double_max_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_max_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1002
 * 
 * 
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1003
 * 
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1004
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1004, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1003
 * 
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1005
 *     with nogil:
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1002
 * 
 * 
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_206int8_min_reduce_on_stream, "int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream = {"int8_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_206int8_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1002, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1002, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1002, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1002, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1002, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1002, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1002, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1002, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1002, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1002, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1002, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1002, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1002, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1002, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1002, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_206int8_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_206int8_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1008
 * 
 * 
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1009
 * 
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1010
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1010, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1009
 * 
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1011
 *     with nogil:
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1008
 * 
 * 
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_208int16_min_reduce_on_stream, "int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream = {"int16_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_208int16_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1008, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1008, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1008, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1008, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1008, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1008, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1008, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1008, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1008, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1008, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1008, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1008, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1008, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1008, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_208int16_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_208int16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1014
 * 
 * 
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1015
 * 
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1016
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1016, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1015
 * 
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1017
 *     with nogil:
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1014
 * 
 * 
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_210int32_min_reduce_on_stream, "int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream = {"int32_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_210int32_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1014, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1014, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1014, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1014, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1014, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1014, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1014, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1014, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1014, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1014, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1014, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1014, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1014, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1014, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_210int32_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_210int32_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1020
 * 
 * 
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1021
 * 
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1022
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1022, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1021
 * 
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1023
 *     with nogil:
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1020
 * 
 * 
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_212int64_min_reduce_on_stream, "int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream = {"int64_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_212int64_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1020, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1020, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1020, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1020, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1020, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1020, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1020, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1020, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1020, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1020, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1020, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1020, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1020, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1020, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_212int64_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_212int64_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1026
 * 
 * 
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1027
 * 
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1028
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1028, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1027
 * 
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1029
 *     with nogil:
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1026
 * 
 * 
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_214uint8_min_reduce_on_stream, "uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream = {"uint8_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_214uint8_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1026, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1026, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1026, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1026, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1026, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1026, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1026, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1026, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1026, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1026, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1026, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1026, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1026, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1026, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_214uint8_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_214uint8_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1032
 * 
 * 
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1033
 * 
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1034
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1034, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1033
 * 
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1035
 *     with nogil:
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1032
 * 
 * 
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_216uint16_min_reduce_on_stream, "uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream = {"uint16_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_216uint16_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1032, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1032, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1032, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1032, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1032, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1032, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1032, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1032, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1032, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1032, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1032, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1032, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1032, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1032, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1032, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1032, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1032, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1032, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1032, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_216uint16_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_216uint16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1038
 * 
 * 
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1039
 * 
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1040
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1040, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1039
 * 
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1041
 *     with nogil:
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1038
 * 
 * 
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_218uint32_min_reduce_on_stream, "uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream = {"uint32_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_218uint32_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1038, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1038, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1038, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1038, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1038, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1038, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1038, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1038, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1038, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_218uint32_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_218uint32_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1044
 * 
 * 
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1045
 * 
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1046
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1046, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1045
 * 
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1047
 *     with nogil:
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1044
 * 
 * 
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_220uint64_min_reduce_on_stream, "uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream = {"uint64_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_220uint64_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1044, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1044, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1044, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1044, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1044, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1044, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1044, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1044, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1044, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_220uint64_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_220uint64_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1050
 * 
 * 
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1051
 * 
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1052
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1052, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1051
 * 
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1053
 *     with nogil:
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1050
 * 
 * 
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_222size_min_reduce_on_stream, "size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream = {"size_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_222size_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1050, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1050, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1050, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1050, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1050, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1050, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1050, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1050, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1050, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1050, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1050, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1050, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1050, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1050, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_222size_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_222size_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1056
 * 
 * 
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1057
 * 
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1058
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1058, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1057
 * 
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1059
 *     with nogil:
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1056
 * 
 * 
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_224char_min_reduce_on_stream, "char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream = {"char_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_224char_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1056, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1056, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1056, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1056, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1056, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1056, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1056, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1056, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1056, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1056, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1056, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1056, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1056, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1056, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_224char_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_224char_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1062
 * 
 * 
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1063
 * 
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1064
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1064, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1063
 * 
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1065
 *     with nogil:
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1062
 * 
 * 
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_226schar_min_reduce_on_stream, "schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream = {"schar_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_226schar_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1062, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1062, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1062, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1062, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1062, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1062, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1062, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1062, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1062, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1062, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1062, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1062, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1062, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1062, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1062, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1062, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1062, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1062, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1062, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_226schar_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_226schar_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1068
 * 
 * 
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1069
 * 
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1070
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1070, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1069
 * 
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1071
 *     with nogil:
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1068
 * 
 * 
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_228short_min_reduce_on_stream, "short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream = {"short_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_228short_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1068, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1068, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1068, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1068, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1068, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1068, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1068, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1068, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1068, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1068, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1068, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1068, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1068, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1068, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_228short_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_228short_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1074
 * 
 * 
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1075
 * 
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1076
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1076, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1075
 * 
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1077
 *     with nogil:
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1074
 * 
 * 
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_230int_min_reduce_on_stream, "int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream = {"int_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_230int_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1074, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1074, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1074, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1074, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1074, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1074, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1074, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1074, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1074, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1074, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1074, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1074, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1074, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1074, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1074, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1074, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1074, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1074, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1074, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_230int_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_230int_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1080
 * 
 * 
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1081
 * 
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1082
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1082, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1081
 * 
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1083
 *     with nogil:
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1080
 * 
 * 
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_232long_min_reduce_on_stream, "long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream = {"long_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_232long_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1080, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1080, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1080, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1080, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1080, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1080, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1080, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1080, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1080, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1080, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1080, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1080, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1080, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1080, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_232long_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_232long_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1086
 * 
 * 
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1087
 * 
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1088
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1088, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1087
 * 
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1089
 *     with nogil:
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1086
 * 
 * 
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_234longlong_min_reduce_on_stream, "longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream = {"longlong_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_234longlong_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1086, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1086, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1086, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1086, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1086, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1086, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1086, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1086, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1086, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1086, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_234longlong_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_234longlong_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1092
 * 
 * 
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1093
 * 
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1094
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1094, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1093
 * 
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1095
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1092
 * 
 * 
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_236bfloat16_min_reduce_on_stream, "bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream = {"bfloat16_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_236bfloat16_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1092, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1092, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1092, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1092, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1092, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1092, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1092, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1092, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1092, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1092, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1092, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1092, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1092, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1092, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_236bfloat16_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_236bfloat16_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1098
 * 
 * 
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1099
 * 
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1100
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1100, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1099
 * 
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1101
 *     with nogil:
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1098
 * 
 * 
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_238half_min_reduce_on_stream, "half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream = {"half_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_238half_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1098, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1098, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1098, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1098, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1098, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1098, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1098, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1098, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1098, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1098, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1098, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1098, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1098, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1098, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_238half_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_238half_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1104
 * 
 * 
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1105
 * 
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1106
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1106, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1105
 * 
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1107
 *     with nogil:
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1104
 * 
 * 
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_240float_min_reduce_on_stream, "float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream = {"float_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_240float_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1104, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1104, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1104, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1104, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1104, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1104, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1104, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1104, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_240float_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_240float_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1110
 * 
 * 
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_min_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1111
 * 
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1112
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1112, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1111
 * 
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1113
 *     with nogil:
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1110
 * 
 * 
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_242double_min_reduce_on_stream, "double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream = {"double_min_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_242double_min_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_min_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1110, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1110, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_min_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1110, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_min_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1110, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1110, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1110, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1110, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1110, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1110, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_min_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1110, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_242double_min_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_242double_min_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_min_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1116
 * 
 * 
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1117
 * 
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1118
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1118, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1117
 * 
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1119
 *     with nogil:
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1116
 * 
 * 
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_244int8_sum_reduce_on_stream, "int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream = {"int8_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_244int8_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1116, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1116, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1116, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1116, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1116, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1116, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1116, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1116, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1116, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1116, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1116, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1116, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1116, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1116, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_244int8_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_244int8_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1122
 * 
 * 
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1123
 * 
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1124
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1124, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1123
 * 
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1125
 *     with nogil:
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1122
 * 
 * 
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_246int16_sum_reduce_on_stream, "int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream = {"int16_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_246int16_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1122, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1122, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1122, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1122, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1122, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1122, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1122, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1122, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1122, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1122, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1122, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1122, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1122, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1122, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_246int16_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_246int16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1128
 * 
 * 
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1129
 * 
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1130
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1130, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1129
 * 
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1131
 *     with nogil:
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1128
 * 
 * 
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_248int32_sum_reduce_on_stream, "int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream = {"int32_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_248int32_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1128, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1128, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1128, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1128, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1128, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1128, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1128, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1128, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1128, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1128, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1128, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_248int32_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_248int32_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1134
 * 
 * 
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1135
 * 
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1136
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1136, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1135
 * 
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1137
 *     with nogil:
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1134
 * 
 * 
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_250int64_sum_reduce_on_stream, "int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream = {"int64_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_250int64_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1134, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1134, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1134, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1134, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1134, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1134, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1134, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1134, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1134, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1134, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1134, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1134, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1134, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1134, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_250int64_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_250int64_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1140
 * 
 * 
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1141
 * 
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1142
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1142, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1141
 * 
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1143
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1140
 * 
 * 
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_252uint8_sum_reduce_on_stream, "uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream = {"uint8_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_252uint8_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1140, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1140, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1140, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1140, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1140, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1140, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1140, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1140, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1140, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1140, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1140, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1140, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1140, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1140, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_252uint8_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_252uint8_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1146
 * 
 * 
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1147
 * 
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1148
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1148, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1147
 * 
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1149
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1146
 * 
 * 
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_254uint16_sum_reduce_on_stream, "uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream = {"uint16_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_254uint16_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1146, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1146, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1146, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1146, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1146, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1146, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1146, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1146, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1146, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1146, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1146, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1146, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1146, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1146, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1146, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1146, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1146, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1146, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1146, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_254uint16_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_254uint16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1152
 * 
 * 
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1153
 * 
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1154
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1154, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1153
 * 
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1155
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1152
 * 
 * 
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_256uint32_sum_reduce_on_stream, "uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream = {"uint32_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_256uint32_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1152, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1152, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1152, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1152, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1152, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1152, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1152, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1152, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1152, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1152, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1152, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1152, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1152, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1152, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_256uint32_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_256uint32_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1158
 * 
 * 
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1159
 * 
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1160
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1160, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1159
 * 
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1161
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1158
 * 
 * 
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_258uint64_sum_reduce_on_stream, "uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream = {"uint64_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_258uint64_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1158, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1158, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1158, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1158, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1158, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1158, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1158, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1158, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1158, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_258uint64_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_258uint64_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1164
 * 
 * 
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1165
 * 
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1166
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1166, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1165
 * 
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1167
 *     with nogil:
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1164
 * 
 * 
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_260size_sum_reduce_on_stream, "size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream = {"size_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_260size_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1164, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1164, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1164, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1164, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1164, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1164, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1164, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1164, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1164, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1164, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1164, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1164, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1164, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1164, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1164, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1164, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1164, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1164, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1164, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_260size_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_260size_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1170
 * 
 * 
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1171
 * 
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1172
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1172, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1171
 * 
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1173
 *     with nogil:
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1170
 * 
 * 
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_262char_sum_reduce_on_stream, "char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream = {"char_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_262char_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1170, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1170, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1170, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1170, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1170, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1170, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1170, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1170, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1170, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1170, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1170, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1170, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1170, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1170, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1170, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1170, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1170, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1170, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1170, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_262char_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_262char_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1176
 * 
 * 
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1177
 * 
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1178
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1178, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1177
 * 
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1179
 *     with nogil:
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1176
 * 
 * 
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_264schar_sum_reduce_on_stream, "schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream = {"schar_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_264schar_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1176, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1176, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1176, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1176, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1176, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1176, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1176, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1176, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1176, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1176, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1176, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_264schar_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_264schar_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1182
 * 
 * 
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1183
 * 
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1184
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1184, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1183
 * 
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1185
 *     with nogil:
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1182
 * 
 * 
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_266short_sum_reduce_on_stream, "short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream = {"short_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_266short_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1182, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1182, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1182, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1182, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1182, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1182, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1182, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1182, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1182, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1182, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1182, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1182, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1182, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1182, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_266short_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_266short_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1188
 * 
 * 
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1189
 * 
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1190
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1190, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1189
 * 
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1191
 *     with nogil:
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1188
 * 
 * 
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_268int_sum_reduce_on_stream, "int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream = {"int_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_268int_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1188, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1188, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1188, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1188, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1188, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1188, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1188, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1188, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1188, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1188, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_268int_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_268int_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1194
 * 
 * 
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1195
 * 
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1196
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1196, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1195
 * 
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1197
 *     with nogil:
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1194
 * 
 * 
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_270long_sum_reduce_on_stream, "long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream = {"long_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_270long_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1194, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1194, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1194, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1194, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1194, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1194, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1194, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1194, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1194, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1194, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_270long_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_270long_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1200
 * 
 * 
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1201
 * 
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1202
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1202, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1201
 * 
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1203
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1200
 * 
 * 
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_272longlong_sum_reduce_on_stream, "longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream = {"longlong_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_272longlong_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1200, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1200, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1200, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1200, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1200, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1200, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1200, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1200, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1200, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1200, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1200, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1200, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1200, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1200, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_272longlong_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_272longlong_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1206
 * 
 * 
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1207
 * 
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1208
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1208, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1207
 * 
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1209
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1206
 * 
 * 
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_274bfloat16_sum_reduce_on_stream, "bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream = {"bfloat16_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_274bfloat16_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1206, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1206, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1206, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1206, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1206, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1206, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1206, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1206, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1206, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_274bfloat16_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_274bfloat16_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1212
 * 
 * 
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1213
 * 
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1214
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1214, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1213
 * 
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1215
 *     with nogil:
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1212
 * 
 * 
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_276half_sum_reduce_on_stream, "half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream = {"half_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_276half_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1212, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1212, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1212, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1212, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1212, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1212, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1212, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1212, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1212, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1212, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1212, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1212, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1212, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1212, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_276half_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_276half_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1218
 * 
 * 
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1219
 * 
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1220
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1220, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1219
 * 
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1221
 *     with nogil:
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1218
 * 
 * 
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_278float_sum_reduce_on_stream, "float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream = {"float_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_278float_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1218, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1218, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1218, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1218, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1218, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1218, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1218, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1218, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1218, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1218, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1218, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1218, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1218, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1218, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_278float_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_278float_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1224
 * 
 * 
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reduce_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_sum_reduce_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1225
 * 
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1226
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reduce_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1226, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1225
 * 
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1227
 *     with nogil:
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1224
 * 
 * 
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_280double_sum_reduce_on_stream, "double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream = {"double_sum_reduce_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_280double_sum_reduce_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_sum_reduce_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1224, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1224, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1224, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1224, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1224, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1224, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_sum_reduce_on_stream", 0) < (0)) __PYX_ERR(0, 1224, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_sum_reduce_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1224, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1224, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1224, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1224, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1224, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1224, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_sum_reduce_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1224, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_280double_sum_reduce_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_280double_sum_reduce_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_sum_reduce_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1230
 * 
 * 
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1231
 * 
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1232
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1232, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1231
 * 
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1233
 *     with nogil:
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1230
 * 
 * 
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_282int8_max_reducescatter_on_stream, "int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream = {"int8_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_282int8_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1230, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1230, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1230, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1230, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1230, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1230, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1230, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1230, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1230, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_282int8_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_282int8_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1236
 * 
 * 
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1237
 * 
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1238
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1238, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1237
 * 
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1239
 *     with nogil:
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1236
 * 
 * 
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_284int16_max_reducescatter_on_stream, "int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream = {"int16_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_284int16_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1236, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1236, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1236, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1236, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1236, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1236, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1236, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1236, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1236, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_284int16_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_284int16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1242
 * 
 * 
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1243
 * 
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1244
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1244, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1243
 * 
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1245
 *     with nogil:
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1242
 * 
 * 
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_286int32_max_reducescatter_on_stream, "int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream = {"int32_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_286int32_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1242, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1242, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1242, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1242, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1242, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1242, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1242, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1242, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1242, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1242, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1242, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1242, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1242, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1242, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_286int32_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_286int32_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1248
 * 
 * 
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1249
 * 
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1250
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1250, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1249
 * 
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1251
 *     with nogil:
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1248
 * 
 * 
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_288int64_max_reducescatter_on_stream, "int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream = {"int64_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_288int64_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1248, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1248, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1248, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1248, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1248, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1248, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1248, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1248, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1248, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_288int64_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_288int64_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1254
 * 
 * 
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1255
 * 
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1256
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1256, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1255
 * 
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1257
 *     with nogil:
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1254
 * 
 * 
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_290uint8_max_reducescatter_on_stream, "uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream = {"uint8_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_290uint8_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1254, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1254, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1254, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1254, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1254, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1254, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1254, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1254, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1254, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1254, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1254, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1254, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1254, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1254, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_290uint8_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_290uint8_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1260
 * 
 * 
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1261
 * 
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1262
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1262, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1261
 * 
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1263
 *     with nogil:
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1260
 * 
 * 
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_292uint16_max_reducescatter_on_stream, "uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream = {"uint16_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_292uint16_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1260, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1260, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1260, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1260, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1260, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1260, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1260, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1260, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1260, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1260, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1260, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1260, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1260, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1260, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1260, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_292uint16_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_292uint16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1266
 * 
 * 
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1267
 * 
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1268
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1268, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1267
 * 
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1269
 *     with nogil:
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1266
 * 
 * 
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_294uint32_max_reducescatter_on_stream, "uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream = {"uint32_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_294uint32_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1266, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1266, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1266, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1266, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1266, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1266, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1266, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1266, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1266, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1266, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1266, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1266, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1266, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1266, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_294uint32_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_294uint32_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1272
 * 
 * 
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1273
 * 
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1274
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1274, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1273
 * 
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1275
 *     with nogil:
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1272
 * 
 * 
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_296uint64_max_reducescatter_on_stream, "uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream = {"uint64_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_296uint64_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1272, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1272, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1272, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1272, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1272, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1272, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1272, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1272, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1272, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1272, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1272, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1272, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1272, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1272, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_296uint64_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_296uint64_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1278
 * 
 * 
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1279
 * 
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1280
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1280, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1279
 * 
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1281
 *     with nogil:
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1278
 * 
 * 
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_298size_max_reducescatter_on_stream, "size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream = {"size_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_298size_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1278, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1278, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1278, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1278, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1278, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1278, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1278, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1278, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1278, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1278, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1278, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1278, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1278, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1278, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_298size_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_298size_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1284
 * 
 * 
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1285
 * 
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1286
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1286, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1285
 * 
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1287
 *     with nogil:
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1284
 * 
 * 
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_300char_max_reducescatter_on_stream, "char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream = {"char_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_300char_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1284, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1284, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1284, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1284, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1284, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1284, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1284, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1284, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1284, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1284, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1284, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1284, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1284, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1284, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_300char_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_300char_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1290
 * 
 * 
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1291
 * 
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1292
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1292, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1291
 * 
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1293
 *     with nogil:
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1290
 * 
 * 
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_302schar_max_reducescatter_on_stream, "schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream = {"schar_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_302schar_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1290, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1290, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1290, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1290, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1290, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1290, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1290, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1290, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1290, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1290, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1290, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1290, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1290, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1290, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_302schar_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_302schar_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1296
 * 
 * 
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1297
 * 
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1298
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1298, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1297
 * 
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1299
 *     with nogil:
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1296
 * 
 * 
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_304short_max_reducescatter_on_stream, "short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream = {"short_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_304short_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1296, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1296, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1296, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1296, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1296, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1296, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1296, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1296, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1296, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_304short_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_304short_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1302
 * 
 * 
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1303
 * 
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1304
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1304, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1303
 * 
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1305
 *     with nogil:
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1302
 * 
 * 
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_306int_max_reducescatter_on_stream, "int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream = {"int_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_306int_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1302, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1302, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1302, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1302, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1302, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1302, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1302, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1302, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1302, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1302, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1302, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1302, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1302, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1302, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_306int_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_306int_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1308
 * 
 * 
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1309
 * 
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1310
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1310, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1309
 * 
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1311
 *     with nogil:
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1308
 * 
 * 
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_308long_max_reducescatter_on_stream, "long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream = {"long_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_308long_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1308, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1308, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1308, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1308, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1308, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1308, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1308, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1308, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1308, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1308, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1308, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1308, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1308, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1308, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_308long_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_308long_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1314
 * 
 * 
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1315
 * 
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1316
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1316, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1315
 * 
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1317
 *     with nogil:
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1314
 * 
 * 
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_310longlong_max_reducescatter_on_stream, "longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream = {"longlong_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_310longlong_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1314, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1314, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1314, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1314, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1314, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1314, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1314, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1314, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1314, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_310longlong_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_310longlong_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1320
 * 
 * 
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1321
 * 
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1322
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1322, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1321
 * 
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1323
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1320
 * 
 * 
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_312bfloat16_max_reducescatter_on_stream, "bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream = {"bfloat16_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_312bfloat16_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1320, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1320, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1320, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1320, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1320, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1320, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1320, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1320, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1320, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1320, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_312bfloat16_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_312bfloat16_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1326
 * 
 * 
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1327
 * 
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1328
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1328, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1327
 * 
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1329
 *     with nogil:
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1326
 * 
 * 
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_314half_max_reducescatter_on_stream, "half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream = {"half_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_314half_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1326, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1326, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1326, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1326, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1326, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1326, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1326, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1326, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1326, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_314half_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_314half_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1332
 * 
 * 
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1333
 * 
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1334
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1334, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1333
 * 
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1335
 *     with nogil:
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1332
 * 
 * 
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_316float_max_reducescatter_on_stream, "float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream = {"float_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_316float_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1332, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1332, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1332, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1332, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1332, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1332, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1332, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1332, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1332, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1332, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1332, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1332, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1332, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1332, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_316float_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_316float_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1338
 * 
 * 
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_max_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1339
 * 
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1340
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1340, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1339
 * 
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1341
 *     with nogil:
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1338
 * 
 * 
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_318double_max_reducescatter_on_stream, "double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream = {"double_max_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_318double_max_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_max_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1338, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1338, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1338, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1338, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1338, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1338, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_max_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1338, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_max_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1338, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1338, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1338, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1338, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1338, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1338, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_max_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1338, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_318double_max_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_318double_max_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_max_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1344
 * 
 * 
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1345
 * 
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1346
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1346, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1345
 * 
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1347
 *     with nogil:
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1344
 * 
 * 
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_320int8_min_reducescatter_on_stream, "int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream = {"int8_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_320int8_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1344, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1344, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1344, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1344, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1344, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1344, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1344, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1344, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1344, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1344, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1344, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1344, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1344, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1344, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_320int8_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_320int8_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1350
 * 
 * 
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1351
 * 
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1352
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1352, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1351
 * 
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1353
 *     with nogil:
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1350
 * 
 * 
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_322int16_min_reducescatter_on_stream, "int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream = {"int16_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_322int16_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1350, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1350, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1350, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1350, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1350, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1350, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1350, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1350, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1350, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1350, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1350, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1350, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1350, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1350, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_322int16_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_322int16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1356
 * 
 * 
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1357
 * 
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1358
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1358, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1357
 * 
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1359
 *     with nogil:
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1356
 * 
 * 
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_324int32_min_reducescatter_on_stream, "int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream = {"int32_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_324int32_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1356, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1356, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1356, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1356, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1356, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1356, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1356, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1356, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1356, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_324int32_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_324int32_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1362
 * 
 * 
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1363
 * 
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1364
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1364, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1363
 * 
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1365
 *     with nogil:
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1362
 * 
 * 
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_326int64_min_reducescatter_on_stream, "int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream = {"int64_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_326int64_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1362, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1362, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1362, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1362, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1362, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1362, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1362, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1362, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1362, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1362, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1362, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1362, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1362, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1362, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_326int64_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_326int64_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1368
 * 
 * 
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1369
 * 
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1370
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1370, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1369
 * 
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1371
 *     with nogil:
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1368
 * 
 * 
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_328uint8_min_reducescatter_on_stream, "uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream = {"uint8_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_328uint8_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1368, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1368, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1368, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1368, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1368, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1368, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1368, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1368, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1368, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_328uint8_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_328uint8_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1374
 * 
 * 
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1375
 * 
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1376
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1376, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1375
 * 
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1377
 *     with nogil:
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1374
 * 
 * 
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_330uint16_min_reducescatter_on_stream, "uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream = {"uint16_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_330uint16_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1374, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1374, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1374, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1374, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1374, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1374, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1374, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1374, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1374, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_330uint16_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_330uint16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1380
 * 
 * 
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1381
 * 
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1382
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1382, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1381
 * 
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1383
 *     with nogil:
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1380
 * 
 * 
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_332uint32_min_reducescatter_on_stream, "uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream = {"uint32_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_332uint32_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1380, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1380, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1380, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1380, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1380, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_332uint32_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_332uint32_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1386
 * 
 * 
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1387
 * 
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1388
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1388, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1387
 * 
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1389
 *     with nogil:
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1386
 * 
 * 
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_334uint64_min_reducescatter_on_stream, "uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream = {"uint64_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_334uint64_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1386, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1386, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1386, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1386, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1386, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1386, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1386, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1386, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1386, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1386, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1386, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1386, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1386, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1386, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_334uint64_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_334uint64_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1392
 * 
 * 
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1393
 * 
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1394
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1394, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1393
 * 
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1395
 *     with nogil:
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1392
 * 
 * 
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_336size_min_reducescatter_on_stream, "size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream = {"size_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_336size_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1392, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1392, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1392, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1392, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1392, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1392, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1392, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1392, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1392, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1392, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1392, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1392, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1392, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1392, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_336size_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_336size_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1398
 * 
 * 
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1399
 * 
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1400
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1400, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1399
 * 
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1401
 *     with nogil:
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1398
 * 
 * 
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_338char_min_reducescatter_on_stream, "char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream = {"char_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_338char_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1398, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1398, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1398, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1398, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1398, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1398, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1398, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1398, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1398, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1398, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_338char_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_338char_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1404
 * 
 * 
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1405
 * 
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1406
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1406, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1405
 * 
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1407
 *     with nogil:
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1404
 * 
 * 
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_340schar_min_reducescatter_on_stream, "schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream = {"schar_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_340schar_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1404, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1404, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1404, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1404, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1404, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1404, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1404, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1404, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1404, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1404, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1404, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1404, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1404, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1404, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1404, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1404, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1404, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1404, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1404, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_340schar_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_340schar_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1410
 * 
 * 
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1411
 * 
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1412
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1412, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1411
 * 
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1413
 *     with nogil:
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1410
 * 
 * 
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_342short_min_reducescatter_on_stream, "short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream = {"short_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_342short_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1410, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1410, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1410, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1410, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1410, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1410, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1410, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1410, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1410, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1410, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1410, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1410, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1410, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1410, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_342short_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_342short_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1416
 * 
 * 
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1417
 * 
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1418
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1418, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1417
 * 
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1419
 *     with nogil:
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1416
 * 
 * 
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_344int_min_reducescatter_on_stream, "int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream = {"int_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_344int_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1416, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1416, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1416, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1416, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1416, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1416, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1416, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1416, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1416, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_344int_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_344int_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1422
 * 
 * 
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1423
 * 
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1424
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1424, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1423
 * 
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1425
 *     with nogil:
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1422
 * 
 * 
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_346long_min_reducescatter_on_stream, "long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream = {"long_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_346long_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1422, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1422, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1422, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1422, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1422, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1422, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1422, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1422, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1422, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1422, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1422, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1422, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1422, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1422, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1422, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1422, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1422, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1422, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1422, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_346long_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_346long_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1428
 * 
 * 
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1429
 * 
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1430
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1430, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1429
 * 
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1431
 *     with nogil:
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1428
 * 
 * 
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_348longlong_min_reducescatter_on_stream, "longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream = {"longlong_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_348longlong_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1428, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1428, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1428, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1428, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1428, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1428, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1428, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1428, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1428, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1428, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1428, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1428, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1428, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1428, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1428, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1428, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1428, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1428, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1428, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_348longlong_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_348longlong_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1434
 * 
 * 
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1435
 * 
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1436
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1436, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1435
 * 
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1437
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1434
 * 
 * 
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_350bfloat16_min_reducescatter_on_stream, "bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream = {"bfloat16_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_350bfloat16_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1434, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1434, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1434, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1434, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1434, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1434, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1434, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1434, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1434, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1434, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1434, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1434, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1434, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1434, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1434, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1434, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1434, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1434, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1434, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_350bfloat16_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_350bfloat16_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1440
 * 
 * 
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1441
 * 
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1442
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1442, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1441
 * 
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1443
 *     with nogil:
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1440
 * 
 * 
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_352half_min_reducescatter_on_stream, "half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream = {"half_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_352half_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1440, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1440, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1440, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1440, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1440, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1440, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1440, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1440, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1440, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1440, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1440, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1440, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1440, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1440, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_352half_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_352half_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1446
 * 
 * 
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1447
 * 
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1448
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1448, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1447
 * 
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1449
 *     with nogil:
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1446
 * 
 * 
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_354float_min_reducescatter_on_stream, "float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream = {"float_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_354float_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1446, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1446, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1446, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1446, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1446, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1446, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1446, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1446, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1446, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1446, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1446, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1446, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1446, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1446, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1446, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1446, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1446, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1446, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1446, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_354float_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_354float_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1452
 * 
 * 
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_min_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1453
 * 
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1454
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1454, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1453
 * 
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1455
 *     with nogil:
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1452
 * 
 * 
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_356double_min_reducescatter_on_stream, "double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream = {"double_min_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_356double_min_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_min_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1452, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1452, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_min_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1452, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_min_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1452, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1452, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1452, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1452, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1452, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1452, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_min_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1452, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_356double_min_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_356double_min_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_min_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1458
 * 
 * 
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1459
 * 
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1460
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int8_t *)__pyx_v_dest), ((int8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1460, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1459
 * 
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1461
 *     with nogil:
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1458
 * 
 * 
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_358int8_sum_reducescatter_on_stream, "int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream = {"int8_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_358int8_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int8_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1458, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1458, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1458, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1458, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1458, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1458, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int8_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1458, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int8_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1458, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1458, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1458, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1458, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1458, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1458, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int8_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1458, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_358int8_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_358int8_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int8_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1464
 * 
 * 
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1465
 * 
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1466
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int16_t *)__pyx_v_dest), ((int16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1466, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1465
 * 
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1467
 *     with nogil:
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1464
 * 
 * 
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_360int16_sum_reducescatter_on_stream, "int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream = {"int16_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_360int16_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int16_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1464, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int16_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1464, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int16_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1464, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1464, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1464, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1464, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1464, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1464, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int16_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1464, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_360int16_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_360int16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int16_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1470
 * 
 * 
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1471
 * 
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1472
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int32_t *)__pyx_v_dest), ((int32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1472, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1471
 * 
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1473
 *     with nogil:
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1470
 * 
 * 
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_362int32_sum_reducescatter_on_stream, "int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream = {"int32_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_362int32_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int32_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1470, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1470, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1470, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1470, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1470, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1470, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int32_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1470, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int32_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1470, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1470, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1470, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1470, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1470, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1470, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int32_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1470, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_362int32_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_362int32_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int32_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1476
 * 
 * 
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1477
 * 
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1478
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int64_t *)__pyx_v_dest), ((int64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1478, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1477
 * 
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1479
 *     with nogil:
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1476
 * 
 * 
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_364int64_sum_reducescatter_on_stream, "int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream = {"int64_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_364int64_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int64_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1476, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1476, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int64_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1476, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int64_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1476, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1476, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1476, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1476, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1476, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1476, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1476, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int64_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1476, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_364int64_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_364int64_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int64_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1482
 * 
 * 
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1483
 * 
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1484
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint8_t *)__pyx_v_dest), ((uint8_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1484, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1483
 * 
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1485
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1482
 * 
 * 
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_366uint8_sum_reducescatter_on_stream, "uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream = {"uint8_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_366uint8_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint8_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1482, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1482, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1482, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1482, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1482, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1482, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint8_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1482, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint8_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1482, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1482, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1482, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1482, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1482, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1482, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint8_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1482, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_366uint8_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_366uint8_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint8_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1488
 * 
 * 
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1489
 * 
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1490
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint16_t *)__pyx_v_dest), ((uint16_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1490, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1489
 * 
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1491
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1488
 * 
 * 
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_368uint16_sum_reducescatter_on_stream, "uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream = {"uint16_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_368uint16_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint16_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1488, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1488, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint16_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1488, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint16_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1488, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1488, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1488, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1488, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1488, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1488, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint16_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1488, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_368uint16_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_368uint16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint16_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1494
 * 
 * 
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1495
 * 
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1496
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint32_t *)__pyx_v_dest), ((uint32_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1496, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1495
 * 
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1497
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1494
 * 
 * 
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_370uint32_sum_reducescatter_on_stream, "uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream = {"uint32_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_370uint32_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint32_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1494, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint32_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1494, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint32_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1494, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1494, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1494, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1494, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1494, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1494, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint32_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1494, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_370uint32_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_370uint32_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint32_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1500
 * 
 * 
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1501
 * 
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1502
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((uint64_t *)__pyx_v_dest), ((uint64_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1502, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1501
 * 
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1503
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1500
 * 
 * 
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_372uint64_sum_reducescatter_on_stream, "uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream = {"uint64_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_372uint64_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("uint64_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1500, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1500, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "uint64_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1500, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("uint64_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1500, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1500, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1500, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1500, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1500, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1500, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("uint64_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1500, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_372uint64_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_372uint64_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uint64_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.uint64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1506
 * 
 * 
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1507
 * 
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1508
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((size_t *)__pyx_v_dest), ((size_t const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1508, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1507
 * 
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1509
 *     with nogil:
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1506
 * 
 * 
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_374size_sum_reducescatter_on_stream, "size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream = {"size_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_374size_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1506, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1506, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "size_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1506, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("size_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1506, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1506, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1506, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1506, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1506, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1506, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("size_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1506, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_374size_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_374size_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.size_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1512
 * 
 * 
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1513
 * 
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1514
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((char *)__pyx_v_dest), ((char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1514, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1513
 * 
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1515
 *     with nogil:
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1512
 * 
 * 
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_376char_sum_reducescatter_on_stream, "char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream = {"char_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_376char_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("char_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1512, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "char_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1512, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("char_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1512, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1512, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1512, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1512, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1512, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1512, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1512, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("char_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1512, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_376char_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_376char_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("char_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.char_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1518
 * 
 * 
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1519
 * 
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1520
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((signed char *)__pyx_v_dest), ((signed char const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1520, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1519
 * 
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1521
 *     with nogil:
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1518
 * 
 * 
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_378schar_sum_reducescatter_on_stream, "schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream = {"schar_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_378schar_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schar_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1518, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1518, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "schar_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1518, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("schar_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1518, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1518, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1518, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1518, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1518, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1518, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("schar_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1518, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_378schar_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_378schar_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("schar_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.schar_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1524
 * 
 * 
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1525
 * 
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1526
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((short *)__pyx_v_dest), ((short const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1526, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1525
 * 
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1527
 *     with nogil:
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1524
 * 
 * 
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_380short_sum_reducescatter_on_stream, "short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream = {"short_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_380short_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1524, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1524, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "short_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1524, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("short_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1524, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1524, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1524, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1524, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1524, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1524, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1524, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1524, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1524, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1524, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1524, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("short_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1524, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_380short_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_380short_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("short_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.short_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1530
 * 
 * 
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1531
 * 
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1532
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((int *)__pyx_v_dest), ((int const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1532, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1531
 * 
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1533
 *     with nogil:
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1530
 * 
 * 
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_382int_sum_reducescatter_on_stream, "int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream = {"int_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_382int_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("int_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1530, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1530, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1530, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1530, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1530, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1530, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "int_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1530, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("int_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1530, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1530, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1530, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1530, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1530, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1530, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("int_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1530, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_382int_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_382int_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("int_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.int_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1536
 * 
 * 
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1537
 * 
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1538
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((long *)__pyx_v_dest), ((long const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1538, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1537
 * 
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1539
 *     with nogil:
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1536
 * 
 * 
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_384long_sum_reducescatter_on_stream, "long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream = {"long_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_384long_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("long_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1536, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1536, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1536, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1536, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1536, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1536, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "long_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1536, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("long_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1536, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1536, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1536, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1536, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1536, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1536, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("long_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1536, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_384long_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_384long_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("long_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.long_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1542
 * 
 * 
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1543
 * 
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1544
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((PY_LONG_LONG *)__pyx_v_dest), ((PY_LONG_LONG const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1544, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1543
 * 
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1545
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1542
 * 
 * 
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_386longlong_sum_reducescatter_on_stream, "longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream = {"longlong_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_386longlong_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("longlong_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1542, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1542, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1542, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1542, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1542, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1542, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "longlong_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1542, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("longlong_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1542, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1542, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1542, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1542, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1542, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1542, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("longlong_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1542, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_386longlong_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_386longlong_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("longlong_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.longlong_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1548
 * 
 * 
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1549
 * 
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1550
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((__nv_bfloat16 *)__pyx_v_dest), ((__nv_bfloat16 const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1550, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1549
 * 
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1551
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1548
 * 
 * 
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_388bfloat16_sum_reducescatter_on_stream, "bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream = {"bfloat16_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_388bfloat16_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bfloat16_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1548, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1548, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1548, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1548, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1548, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1548, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bfloat16_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1548, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bfloat16_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1548, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1548, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1548, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1548, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1548, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1548, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bfloat16_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1548, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_388bfloat16_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_388bfloat16_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bfloat16_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.bfloat16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1554
 * 
 * 
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1555
 * 
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1556
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((half *)__pyx_v_dest), ((half const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1556, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1555
 * 
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1557
 *     with nogil:
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1554
 * 
 * 
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_390half_sum_reducescatter_on_stream, "half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream = {"half_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_390half_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("half_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1554, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1554, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1554, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1554, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1554, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1554, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "half_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1554, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("half_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1554, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1554, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1554, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1554, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1554, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1554, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("half_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1554, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_390half_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_390half_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("half_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.half_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1560
 * 
 * 
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1561
 * 
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1562
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((float *)__pyx_v_dest), ((float const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1562, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1561
 * 
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1563
 *     with nogil:
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1560
 * 
 * 
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_392float_sum_reducescatter_on_stream, "float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream = {"float_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_392float_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("float_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1560, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1560, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1560, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1560, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1560, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1560, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "float_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1560, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("float_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1560, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1560, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1560, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1560, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1560, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1560, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("float_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1560, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_392float_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_392float_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("float_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.float_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1566
 * 
 * 
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reducescatter_on_stream(int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_sum_reducescatter_on_stream", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1567
 * 
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1568
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reducescatter_on_stream(((nvshmem_team_t)__pyx_v_team), ((double *)__pyx_v_dest), ((double const *)__pyx_v_src), __pyx_v_nreduce, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_stream)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1568, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1567
 * 
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1569
 *     with nogil:
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1566
 * 
 * 
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_394double_sum_reducescatter_on_stream, "double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream = {"double_sum_reducescatter_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_394double_sum_reducescatter_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int32_t __pyx_v_team;
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_src;
  size_t __pyx_v_nreduce;
  intptr_t __pyx_v_stream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("double_sum_reducescatter_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_team,&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_nreduce,&__pyx_mstate_global->__pyx_n_u_stream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1566, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1566, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1566, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1566, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1566, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1566, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "double_sum_reducescatter_on_stream", 0) < (0)) __PYX_ERR(0, 1566, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("double_sum_reducescatter_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1566, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1566, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1566, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1566, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1566, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1566, __pyx_L3_error)
    }
    __pyx_v_team = __Pyx_PyLong_As_int32_t(values[0]); if (unlikely((__pyx_v_team == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1566, __pyx_L3_error)
    __pyx_v_dest = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1566, __pyx_L3_error)
    __pyx_v_src = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_src == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1566, __pyx_L3_error)
    __pyx_v_nreduce = __Pyx_PyLong_As_size_t(values[3]); if (unlikely((__pyx_v_nreduce == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1566, __pyx_L3_error)
    __pyx_v_stream = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_stream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1566, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("double_sum_reducescatter_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1566, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_394double_sum_reducescatter_on_stream(__pyx_self, __pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_394double_sum_reducescatter_on_stream(CYTHON_UNUSED PyObject *__pyx_self, int32_t __pyx_v_team, intptr_t __pyx_v_dest, intptr_t __pyx_v_src, size_t __pyx_v_nreduce, intptr_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("double_sum_reducescatter_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.double_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1572
 * 
 * 
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_init_attr(unsigned int __pyx_v_flags, intptr_t __pyx_v_attr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hostlib_init_attr", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1573
 * 
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1574
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):
 *     with nogil:
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_init_attr(__pyx_v_flags, ((nvshmemx_init_attr_t *)__pyx_v_attr)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1574, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1573
 * 
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1575
 *     with nogil:
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1572
 * 
 * 
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.hostlib_init_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_396hostlib_init_attr, "hostlib_init_attr(unsigned int flags, intptr_t attr)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr = {"hostlib_init_attr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_396hostlib_init_attr};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  unsigned int __pyx_v_flags;
  intptr_t __pyx_v_attr;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hostlib_init_attr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1572, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1572, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1572, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "hostlib_init_attr", 0) < (0)) __PYX_ERR(0, 1572, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("hostlib_init_attr", 1, 2, 2, i); __PYX_ERR(0, 1572, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1572, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1572, __pyx_L3_error)
    }
    __pyx_v_flags = __Pyx_PyLong_As_unsigned_int(values[0]); if (unlikely((__pyx_v_flags == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1572, __pyx_L3_error)
    __pyx_v_attr = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_attr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1572, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hostlib_init_attr", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1572, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.hostlib_init_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_396hostlib_init_attr(__pyx_self, __pyx_v_flags, __pyx_v_attr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_396hostlib_init_attr(CYTHON_UNUSED PyObject *__pyx_self, unsigned int __pyx_v_flags, intptr_t __pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hostlib_init_attr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_init_attr(__pyx_v_flags, __pyx_v_attr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.hostlib_init_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1578
 * 
 * 
 * cpdef void hostlib_finalize() except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_hostlib_finalize()
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_399hostlib_finalize(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_finalize(CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1579
 * 
 * cpdef void hostlib_finalize() except*:
 *     nvshmemx_hostlib_finalize()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_finalize(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1579, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1578
 * 
 * 
 * cpdef void hostlib_finalize() except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_hostlib_finalize()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.hostlib_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_399hostlib_finalize(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_398hostlib_finalize, "hostlib_finalize() -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_399hostlib_finalize = {"hostlib_finalize", (PyCFunction)__pyx_pw_7nvshmem_8bindings_7nvshmem_399hostlib_finalize, METH_NOARGS, __pyx_doc_7nvshmem_8bindings_7nvshmem_398hostlib_finalize};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_399hostlib_finalize(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hostlib_finalize (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_398hostlib_finalize(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_398hostlib_finalize(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hostlib_finalize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_finalize(1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1578, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.hostlib_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1582
 * 
 * 
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_uniqueid_args(int __pyx_v_myrank, int __pyx_v_nranks, intptr_t __pyx_v_uniqueid, intptr_t __pyx_v_attr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_attr_uniqueid_args", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1583
 * 
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1584
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):
 *     with nogil:
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_uniqueid_args(((int const )__pyx_v_myrank), ((int const )__pyx_v_nranks), ((nvshmemx_uniqueid_t const *)__pyx_v_uniqueid), ((nvshmemx_init_attr_t *)__pyx_v_attr)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1584, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1583
 * 
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1585
 *     with nogil:
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1582
 * 
 * 
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_uniqueid_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_400set_attr_uniqueid_args, "set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args = {"set_attr_uniqueid_args", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_400set_attr_uniqueid_args};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_myrank;
  int __pyx_v_nranks;
  intptr_t __pyx_v_uniqueid;
  intptr_t __pyx_v_attr;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_attr_uniqueid_args (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_myrank,&__pyx_mstate_global->__pyx_n_u_nranks,&__pyx_mstate_global->__pyx_n_u_uniqueid,&__pyx_mstate_global->__pyx_n_u_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1582, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1582, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_attr_uniqueid_args", 0) < (0)) __PYX_ERR(0, 1582, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_attr_uniqueid_args", 1, 4, 4, i); __PYX_ERR(0, 1582, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1582, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1582, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1582, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1582, __pyx_L3_error)
    }
    __pyx_v_myrank = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_myrank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L3_error)
    __pyx_v_nranks = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_nranks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L3_error)
    __pyx_v_uniqueid = PyLong_AsSsize_t(values[2]); if (unlikely((__pyx_v_uniqueid == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L3_error)
    __pyx_v_attr = PyLong_AsSsize_t(values[3]); if (unlikely((__pyx_v_attr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_attr_uniqueid_args", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1582, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_uniqueid_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_400set_attr_uniqueid_args(__pyx_self, __pyx_v_myrank, __pyx_v_nranks, __pyx_v_uniqueid, __pyx_v_attr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_400set_attr_uniqueid_args(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_myrank, int __pyx_v_nranks, intptr_t __pyx_v_uniqueid, intptr_t __pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_attr_uniqueid_args", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_uniqueid_args(__pyx_v_myrank, __pyx_v_nranks, __pyx_v_uniqueid, __pyx_v_attr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_uniqueid_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1588
 * 
 * 
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_mpi_comm_args(intptr_t __pyx_v_mpi_comm, intptr_t __pyx_v_nvshmem_attr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_attr_mpi_comm_args", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1589
 * 
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1590
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):
 *     with nogil:
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_mpi_comm_args(((void *)__pyx_v_mpi_comm), ((nvshmemx_init_attr_t *)__pyx_v_nvshmem_attr)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1590, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1589
 * 
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1591
 *     with nogil:
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1588
 * 
 * 
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_mpi_comm_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_402set_attr_mpi_comm_args, "set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args = {"set_attr_mpi_comm_args", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_402set_attr_mpi_comm_args};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_mpi_comm;
  intptr_t __pyx_v_nvshmem_attr;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_attr_mpi_comm_args (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mpi_comm,&__pyx_mstate_global->__pyx_n_u_nvshmem_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1588, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1588, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_attr_mpi_comm_args", 0) < (0)) __PYX_ERR(0, 1588, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_attr_mpi_comm_args", 1, 2, 2, i); __PYX_ERR(0, 1588, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1588, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1588, __pyx_L3_error)
    }
    __pyx_v_mpi_comm = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_mpi_comm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1588, __pyx_L3_error)
    __pyx_v_nvshmem_attr = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_nvshmem_attr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1588, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_attr_mpi_comm_args", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1588, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_mpi_comm_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_402set_attr_mpi_comm_args(__pyx_self, __pyx_v_mpi_comm, __pyx_v_nvshmem_attr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_402set_attr_mpi_comm_args(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_mpi_comm, intptr_t __pyx_v_nvshmem_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_attr_mpi_comm_args", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_mpi_comm_args(__pyx_v_mpi_comm, __pyx_v_nvshmem_attr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.set_attr_mpi_comm_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1594
 * 
 * 
 * cpdef get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_405get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem_get_uniqueid(intptr_t __pyx_v_uniqueid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_uniqueid", 0);

  /* "nvshmem/bindings/nvshmem.pyx":1595
 * 
 * cpdef get_uniqueid(intptr_t uniqueid):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
 *     check_status(status)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "nvshmem/bindings/nvshmem.pyx":1596
 * cpdef get_uniqueid(intptr_t uniqueid):
 *     with nogil:
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
*/
        __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_get_uniqueid(((nvshmemx_uniqueid_t *)__pyx_v_uniqueid)); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 1596, __pyx_L4_error)
        __pyx_v_status = __pyx_t_1;
      }

      /* "nvshmem/bindings/nvshmem.pyx":1595
 * 
 * cpdef get_uniqueid(intptr_t uniqueid):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
 *     check_status(status)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "nvshmem/bindings/nvshmem.pyx":1597
 *     with nogil:
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1594
 * 
 * 
 * cpdef get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_405get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_404get_uniqueid, "get_uniqueid(intptr_t uniqueid)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_405get_uniqueid = {"get_uniqueid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_405get_uniqueid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_404get_uniqueid};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_405get_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_uniqueid;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_uniqueid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_uniqueid,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1594, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1594, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_uniqueid", 0) < (0)) __PYX_ERR(0, 1594, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_uniqueid", 1, 1, 1, i); __PYX_ERR(0, 1594, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1594, __pyx_L3_error)
    }
    __pyx_v_uniqueid = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_uniqueid == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1594, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_uniqueid", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1594, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_404get_uniqueid(__pyx_self, __pyx_v_uniqueid);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_404get_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_uniqueid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_uniqueid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_get_uniqueid(__pyx_v_uniqueid, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1600
 * 
 * 
 * cpdef int cumodule_init(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_init(<void*>module)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_407cumodule_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_init(intptr_t __pyx_v_module, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1601
 * 
 * cpdef int cumodule_init(intptr_t module) except? -1:
 *     return nvshmemx_cumodule_init(<void*>module)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_init(((void *)__pyx_v_module)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1600
 * 
 * 
 * cpdef int cumodule_init(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_init(<void*>module)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_407cumodule_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_406cumodule_init, "cumodule_init(intptr_t module) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_407cumodule_init = {"cumodule_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_407cumodule_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_406cumodule_init};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_407cumodule_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_module;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cumodule_init (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_module,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1600, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1600, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cumodule_init", 0) < (0)) __PYX_ERR(0, 1600, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cumodule_init", 1, 1, 1, i); __PYX_ERR(0, 1600, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1600, __pyx_L3_error)
    }
    __pyx_v_module = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_module == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cumodule_init", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1600, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_406cumodule_init(__pyx_self, __pyx_v_module);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_406cumodule_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_module) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cumodule_init", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_init(__pyx_v_module, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1604
 * 
 * 
 * cpdef int cumodule_finalize(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_finalize(<void*>module)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_409cumodule_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_finalize(intptr_t __pyx_v_module, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1605
 * 
 * cpdef int cumodule_finalize(intptr_t module) except? -1:
 *     return nvshmemx_cumodule_finalize(<void*>module)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_finalize(((void *)__pyx_v_module)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1605, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1604
 * 
 * 
 * cpdef int cumodule_finalize(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_finalize(<void*>module)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_409cumodule_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_408cumodule_finalize, "cumodule_finalize(intptr_t module) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_409cumodule_finalize = {"cumodule_finalize", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_409cumodule_finalize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_408cumodule_finalize};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_409cumodule_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_module;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cumodule_finalize (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_module,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1604, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1604, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cumodule_finalize", 0) < (0)) __PYX_ERR(0, 1604, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cumodule_finalize", 1, 1, 1, i); __PYX_ERR(0, 1604, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1604, __pyx_L3_error)
    }
    __pyx_v_module = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_module == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1604, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cumodule_finalize", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1604, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_408cumodule_finalize(__pyx_self, __pyx_v_module);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_408cumodule_finalize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_module) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cumodule_finalize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_finalize(__pyx_v_module, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1604, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.cumodule_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1608
 * 
 * 
 * cpdef intptr_t buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_buffer_register_symmetric(<void*>buf_ptr, size, flags)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static intptr_t __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_register_symmetric(intptr_t __pyx_v_buf_ptr, size_t __pyx_v_size, int __pyx_v_flags, CYTHON_UNUSED int __pyx_skip_dispatch) {
  intptr_t __pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1609
 * 
 * cpdef intptr_t buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) except? 0:
 *     return <intptr_t>nvshmemx_buffer_register_symmetric(<void*>buf_ptr, size, flags)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_register_symmetric(((void *)__pyx_v_buf_ptr), __pyx_v_size, __pyx_v_flags); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1609, __pyx_L1_error)
  __pyx_r = ((intptr_t)__pyx_t_1);
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1608
 * 
 * 
 * cpdef intptr_t buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_buffer_register_symmetric(<void*>buf_ptr, size, flags)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_register_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_410buffer_register_symmetric, "buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) -> intptr_t");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric = {"buffer_register_symmetric", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_410buffer_register_symmetric};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_buf_ptr;
  size_t __pyx_v_size;
  int __pyx_v_flags;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer_register_symmetric (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_buf_ptr,&__pyx_mstate_global->__pyx_n_u_size,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1608, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1608, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1608, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1608, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "buffer_register_symmetric", 0) < (0)) __PYX_ERR(0, 1608, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("buffer_register_symmetric", 1, 3, 3, i); __PYX_ERR(0, 1608, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1608, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1608, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1608, __pyx_L3_error)
    }
    __pyx_v_buf_ptr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_buf_ptr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1608, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyLong_As_size_t(values[1]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1608, __pyx_L3_error)
    __pyx_v_flags = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1608, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("buffer_register_symmetric", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1608, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_register_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_410buffer_register_symmetric(__pyx_self, __pyx_v_buf_ptr, __pyx_v_size, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_410buffer_register_symmetric(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_buf_ptr, size_t __pyx_v_size, int __pyx_v_flags) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("buffer_register_symmetric", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_register_symmetric(__pyx_v_buf_ptr, __pyx_v_size, __pyx_v_flags, 1); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 1608, __pyx_L1_error)
  __pyx_t_2 = PyLong_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_register_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1612
 * 
 * 
 * cpdef int buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_buffer_unregister_symmetric(<void*>mmap_ptr, size)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_unregister_symmetric(intptr_t __pyx_v_mmap_ptr, size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1613
 * 
 * cpdef int buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) except? 0:
 *     return nvshmemx_buffer_unregister_symmetric(<void*>mmap_ptr, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_unregister_symmetric(((void *)__pyx_v_mmap_ptr), __pyx_v_size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1613, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1612
 * 
 * 
 * cpdef int buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_buffer_unregister_symmetric(<void*>mmap_ptr, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_unregister_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_412buffer_unregister_symmetric, "buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric = {"buffer_unregister_symmetric", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_412buffer_unregister_symmetric};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_mmap_ptr;
  size_t __pyx_v_size;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer_unregister_symmetric (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mmap_ptr,&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1612, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1612, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "buffer_unregister_symmetric", 0) < (0)) __PYX_ERR(0, 1612, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("buffer_unregister_symmetric", 1, 2, 2, i); __PYX_ERR(0, 1612, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1612, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1612, __pyx_L3_error)
    }
    __pyx_v_mmap_ptr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_mmap_ptr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1612, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyLong_As_size_t(values[1]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1612, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("buffer_unregister_symmetric", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1612, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_unregister_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_412buffer_unregister_symmetric(__pyx_self, __pyx_v_mmap_ptr, __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_412buffer_unregister_symmetric(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_mmap_ptr, size_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("buffer_unregister_symmetric", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_buffer_unregister_symmetric(__pyx_v_mmap_ptr, __pyx_v_size, 1); if (unlikely(__pyx_t_1 == ((int)0) && PyErr_Occurred())) __PYX_ERR(0, 1612, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.buffer_unregister_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1616
 * 
 * 
 * cpdef int culibrary_init(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_init(<void*>library)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_415culibrary_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_init(intptr_t __pyx_v_library, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1617
 * 
 * cpdef int culibrary_init(intptr_t library) except? -1:
 *     return nvshmemx_culibrary_init(<void*>library)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_init(((void *)__pyx_v_library)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1617, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1616
 * 
 * 
 * cpdef int culibrary_init(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_init(<void*>library)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_415culibrary_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_414culibrary_init, "culibrary_init(intptr_t library) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_415culibrary_init = {"culibrary_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_415culibrary_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_414culibrary_init};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_415culibrary_init(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_library;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("culibrary_init (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_library,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1616, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1616, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "culibrary_init", 0) < (0)) __PYX_ERR(0, 1616, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("culibrary_init", 1, 1, 1, i); __PYX_ERR(0, 1616, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1616, __pyx_L3_error)
    }
    __pyx_v_library = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_library == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1616, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("culibrary_init", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1616, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_414culibrary_init(__pyx_self, __pyx_v_library);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_414culibrary_init(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_library) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("culibrary_init", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_init(__pyx_v_library, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1616, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1620
 * 
 * 
 * cpdef int culibrary_finalize(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_finalize(<void*>library)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_417culibrary_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_finalize(intptr_t __pyx_v_library, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1621
 * 
 * cpdef int culibrary_finalize(intptr_t library) except? -1:
 *     return nvshmemx_culibrary_finalize(<void*>library)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_finalize(((void *)__pyx_v_library)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1621, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/nvshmem.pyx":1620
 * 
 * 
 * cpdef int culibrary_finalize(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_finalize(<void*>library)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_417culibrary_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_416culibrary_finalize, "culibrary_finalize(intptr_t library) -> int");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_417culibrary_finalize = {"culibrary_finalize", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_417culibrary_finalize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_416culibrary_finalize};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_417culibrary_finalize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_library;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("culibrary_finalize (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_library,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1620, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1620, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "culibrary_finalize", 0) < (0)) __PYX_ERR(0, 1620, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("culibrary_finalize", 1, 1, 1, i); __PYX_ERR(0, 1620, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1620, __pyx_L3_error)
    }
    __pyx_v_library = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_library == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1620, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("culibrary_finalize", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1620, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_416culibrary_finalize(__pyx_self, __pyx_v_library);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_416culibrary_finalize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_library) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("culibrary_finalize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_finalize(__pyx_v_library, 1); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1620, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.culibrary_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1624
 * 
 * 
 * cpdef void putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_419putmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_on_stream(intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1625
 * 
 * cpdef void putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:
 *     nvshmemx_putmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_on_stream(((void *)__pyx_v_dest), ((void const *)__pyx_v_source), __pyx_v_bytes, __pyx_v_pe, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstrm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1625, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1624
 * 
 * 
 * cpdef void putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_419putmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_418putmem_on_stream, "putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_419putmem_on_stream = {"putmem_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_419putmem_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_418putmem_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_419putmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_source;
  size_t __pyx_v_bytes;
  int __pyx_v_pe;
  intptr_t __pyx_v_cstrm;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("putmem_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_source,&__pyx_mstate_global->__pyx_n_u_bytes,&__pyx_mstate_global->__pyx_n_u_pe,&__pyx_mstate_global->__pyx_n_u_cstrm,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1624, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1624, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "putmem_on_stream", 0) < (0)) __PYX_ERR(0, 1624, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("putmem_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1624, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1624, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1624, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1624, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1624, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1624, __pyx_L3_error)
    }
    __pyx_v_dest = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L3_error)
    __pyx_v_source = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_source == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L3_error)
    __pyx_v_bytes = __Pyx_PyLong_As_size_t(values[2]); if (unlikely((__pyx_v_bytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L3_error)
    __pyx_v_pe = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L3_error)
    __pyx_v_cstrm = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_cstrm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("putmem_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1624, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_418putmem_on_stream(__pyx_self, __pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_418putmem_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("putmem_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1628
 * 
 * 
 * cpdef void putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_signal_on_stream(<void*>dest, <const void*>source, bytes, <uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_signal_on_stream(intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1629
 * 
 * cpdef void putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:
 *     nvshmemx_putmem_signal_on_stream(<void*>dest, <const void*>source, bytes, <uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_signal_on_stream(((void *)__pyx_v_dest), ((void const *)__pyx_v_source), __pyx_v_bytes, ((uint64_t *)__pyx_v_sig_addr), __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstrm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1628
 * 
 * 
 * cpdef void putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_signal_on_stream(<void*>dest, <const void*>source, bytes, <uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_signal_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_420putmem_signal_on_stream, "putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream = {"putmem_signal_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_420putmem_signal_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_source;
  size_t __pyx_v_bytes;
  intptr_t __pyx_v_sig_addr;
  uint64_t __pyx_v_signal;
  int __pyx_v_sig_op;
  int __pyx_v_pe;
  intptr_t __pyx_v_cstrm;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("putmem_signal_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_source,&__pyx_mstate_global->__pyx_n_u_bytes,&__pyx_mstate_global->__pyx_n_u_sig_addr,&__pyx_mstate_global->__pyx_n_u_signal,&__pyx_mstate_global->__pyx_n_u_sig_op,&__pyx_mstate_global->__pyx_n_u_pe,&__pyx_mstate_global->__pyx_n_u_cstrm,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1628, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1628, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "putmem_signal_on_stream", 0) < (0)) __PYX_ERR(0, 1628, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 8; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("putmem_signal_on_stream", 1, 8, 8, i); __PYX_ERR(0, 1628, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 8)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1628, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1628, __pyx_L3_error)
    }
    __pyx_v_dest = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_source = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_source == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_bytes = __Pyx_PyLong_As_size_t(values[2]); if (unlikely((__pyx_v_bytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_sig_addr = PyLong_AsSsize_t(values[3]); if (unlikely((__pyx_v_sig_addr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_signal = __Pyx_PyLong_As_uint64_t(values[4]); if (unlikely((__pyx_v_signal == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_sig_op = __Pyx_PyLong_As_int(values[5]); if (unlikely((__pyx_v_sig_op == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_pe = __Pyx_PyLong_As_int(values[6]); if (unlikely((__pyx_v_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_cstrm = PyLong_AsSsize_t(values[7]); if (unlikely((__pyx_v_cstrm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("putmem_signal_on_stream", 1, 8, 8, __pyx_nargs); __PYX_ERR(0, 1628, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_signal_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_420putmem_signal_on_stream(__pyx_self, __pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_420putmem_signal_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("putmem_signal_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_putmem_signal_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.putmem_signal_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1632
 * 
 * 
 * cpdef void getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_getmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_423getmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_getmem_on_stream(intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1633
 * 
 * cpdef void getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:
 *     nvshmemx_getmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_getmem_on_stream(((void *)__pyx_v_dest), ((void const *)__pyx_v_source), __pyx_v_bytes, __pyx_v_pe, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstrm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1633, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1632
 * 
 * 
 * cpdef void getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_getmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.getmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_423getmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_422getmem_on_stream, "getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_423getmem_on_stream = {"getmem_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_423getmem_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_422getmem_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_423getmem_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_dest;
  intptr_t __pyx_v_source;
  size_t __pyx_v_bytes;
  int __pyx_v_pe;
  intptr_t __pyx_v_cstrm;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getmem_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dest,&__pyx_mstate_global->__pyx_n_u_source,&__pyx_mstate_global->__pyx_n_u_bytes,&__pyx_mstate_global->__pyx_n_u_pe,&__pyx_mstate_global->__pyx_n_u_cstrm,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1632, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1632, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1632, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1632, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1632, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1632, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "getmem_on_stream", 0) < (0)) __PYX_ERR(0, 1632, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("getmem_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1632, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1632, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1632, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1632, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1632, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1632, __pyx_L3_error)
    }
    __pyx_v_dest = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_dest == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L3_error)
    __pyx_v_source = PyLong_AsSsize_t(values[1]); if (unlikely((__pyx_v_source == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L3_error)
    __pyx_v_bytes = __Pyx_PyLong_As_size_t(values[2]); if (unlikely((__pyx_v_bytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L3_error)
    __pyx_v_pe = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L3_error)
    __pyx_v_cstrm = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_cstrm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getmem_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1632, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.getmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_422getmem_on_stream(__pyx_self, __pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_422getmem_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_dest, intptr_t __pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, intptr_t __pyx_v_cstrm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getmem_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_getmem_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1632, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.getmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1636
 * 
 * 
 * cpdef void quiet_on_stream(intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_quiet_on_stream(<Stream>cstrm)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_425quiet_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_quiet_on_stream(intptr_t __pyx_v_cstrm, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1637
 * 
 * cpdef void quiet_on_stream(intptr_t cstrm) except*:
 *     nvshmemx_quiet_on_stream(<Stream>cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_quiet_on_stream(((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstrm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1637, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1636
 * 
 * 
 * cpdef void quiet_on_stream(intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_quiet_on_stream(<Stream>cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.quiet_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_425quiet_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_424quiet_on_stream, "quiet_on_stream(intptr_t cstrm) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_425quiet_on_stream = {"quiet_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_425quiet_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_424quiet_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_425quiet_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_cstrm;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("quiet_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_cstrm,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1636, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1636, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "quiet_on_stream", 0) < (0)) __PYX_ERR(0, 1636, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("quiet_on_stream", 1, 1, 1, i); __PYX_ERR(0, 1636, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1636, __pyx_L3_error)
    }
    __pyx_v_cstrm = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_cstrm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1636, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("quiet_on_stream", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1636, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.quiet_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_424quiet_on_stream(__pyx_self, __pyx_v_cstrm);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_424quiet_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_cstrm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("quiet_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_quiet_on_stream(__pyx_v_cstrm, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1636, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.quiet_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1640
 * 
 * 
 * cpdef void signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_op_on_stream(<uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_signal_op_on_stream(intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1641
 * 
 * cpdef void signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:
 *     nvshmemx_signal_op_on_stream(<uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_op_on_stream(((uint64_t *)__pyx_v_sig_addr), __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstrm)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1641, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1640
 * 
 * 
 * cpdef void signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_op_on_stream(<uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_op_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_426signal_op_on_stream, "signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream = {"signal_op_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_426signal_op_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_sig_addr;
  uint64_t __pyx_v_signal;
  int __pyx_v_sig_op;
  int __pyx_v_pe;
  intptr_t __pyx_v_cstrm;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("signal_op_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_sig_addr,&__pyx_mstate_global->__pyx_n_u_signal,&__pyx_mstate_global->__pyx_n_u_sig_op,&__pyx_mstate_global->__pyx_n_u_pe,&__pyx_mstate_global->__pyx_n_u_cstrm,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1640, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1640, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1640, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1640, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1640, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1640, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "signal_op_on_stream", 0) < (0)) __PYX_ERR(0, 1640, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("signal_op_on_stream", 1, 5, 5, i); __PYX_ERR(0, 1640, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1640, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1640, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1640, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1640, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1640, __pyx_L3_error)
    }
    __pyx_v_sig_addr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_sig_addr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L3_error)
    __pyx_v_signal = __Pyx_PyLong_As_uint64_t(values[1]); if (unlikely((__pyx_v_signal == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L3_error)
    __pyx_v_sig_op = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_sig_op == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L3_error)
    __pyx_v_pe = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_pe == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L3_error)
    __pyx_v_cstrm = PyLong_AsSsize_t(values[4]); if (unlikely((__pyx_v_cstrm == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("signal_op_on_stream", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1640, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_op_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_426signal_op_on_stream(__pyx_self, __pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_426signal_op_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, intptr_t __pyx_v_cstrm) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("signal_op_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_signal_op_on_stream(__pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1640, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_op_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nvshmem/bindings/nvshmem.pyx":1644
 * 
 * 
 * cpdef void signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_wait_until_on_stream(<uint64_t*>sig_addr, cmp, cmp_value, <Stream>cstream)
*/

static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7nvshmem_8bindings_7nvshmem_signal_wait_until_on_stream(intptr_t __pyx_v_sig_addr, int __pyx_v_cmp, uint64_t __pyx_v_cmp_value, intptr_t __pyx_v_cstream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1645
 * 
 * cpdef void signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) except*:
 *     nvshmemx_signal_wait_until_on_stream(<uint64_t*>sig_addr, cmp, cmp_value, <Stream>cstream)             # <<<<<<<<<<<<<<
*/
  __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_wait_until_on_stream(((uint64_t *)__pyx_v_sig_addr), __pyx_v_cmp, __pyx_v_cmp_value, ((__pyx_t_7nvshmem_8bindings_7nvshmem_Stream)__pyx_v_cstream)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L1_error)

  /* "nvshmem/bindings/nvshmem.pyx":1644
 * 
 * 
 * cpdef void signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_wait_until_on_stream(<uint64_t*>sig_addr, cmp, cmp_value, <Stream>cstream)
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_wait_until_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_428signal_wait_until_on_stream, "signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) -> void");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream = {"signal_wait_until_on_stream", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_428signal_wait_until_on_stream};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  intptr_t __pyx_v_sig_addr;
  int __pyx_v_cmp;
  uint64_t __pyx_v_cmp_value;
  intptr_t __pyx_v_cstream;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("signal_wait_until_on_stream (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_sig_addr,&__pyx_mstate_global->__pyx_n_u_cmp,&__pyx_mstate_global->__pyx_n_u_cmp_value,&__pyx_mstate_global->__pyx_n_u_cstream,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(0, 1644, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "signal_wait_until_on_stream", 0) < (0)) __PYX_ERR(0, 1644, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("signal_wait_until_on_stream", 1, 4, 4, i); __PYX_ERR(0, 1644, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1644, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1644, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1644, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1644, __pyx_L3_error)
    }
    __pyx_v_sig_addr = PyLong_AsSsize_t(values[0]); if (unlikely((__pyx_v_sig_addr == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_cmp = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_cmp == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_cmp_value = __Pyx_PyLong_As_uint64_t(values[2]); if (unlikely((__pyx_v_cmp_value == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_cstream = PyLong_AsSsize_t(values[3]); if (unlikely((__pyx_v_cstream == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("signal_wait_until_on_stream", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1644, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_wait_until_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_428signal_wait_until_on_stream(__pyx_self, __pyx_v_sig_addr, __pyx_v_cmp, __pyx_v_cmp_value, __pyx_v_cstream);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_428signal_wait_until_on_stream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_sig_addr, int __pyx_v_cmp, uint64_t __pyx_v_cmp_value, intptr_t __pyx_v_cstream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("signal_wait_until_on_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7nvshmem_8bindings_7nvshmem_signal_wait_until_on_stream(__pyx_v_sig_addr, __pyx_v_cmp, __pyx_v_cmp_value, __pyx_v_cstream, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.signal_wait_until_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_431__pyx_unpickle_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_430__pyx_unpickle_uniqueid, "__pyx_unpickle_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_431__pyx_unpickle_uniqueid = {"__pyx_unpickle_uniqueid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_431__pyx_unpickle_uniqueid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_430__pyx_unpickle_uniqueid};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_431__pyx_unpickle_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_uniqueid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_uniqueid", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_uniqueid", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_uniqueid", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_430__pyx_unpickle_uniqueid(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_430__pyx_unpickle_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_uniqueid", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = uniqueid.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_uniqueid__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_uniqueid__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_uniqueid__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_data);
  __Pyx_DECREF(__pyx_v___pyx_result->_data);
  __pyx_v___pyx_result->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_uniqueid__set_state(<uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_uniqueid__set_state(uniqueid __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_uniqueid__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_UniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_433__pyx_unpickle_UniqueId(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_432__pyx_unpickle_UniqueId, "__pyx_unpickle_UniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_433__pyx_unpickle_UniqueId = {"__pyx_unpickle_UniqueId", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_433__pyx_unpickle_UniqueId, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_432__pyx_unpickle_UniqueId};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_433__pyx_unpickle_UniqueId(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UniqueId (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_UniqueId", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UniqueId", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UniqueId", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_UniqueId", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_432__pyx_unpickle_UniqueId(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_432__pyx_unpickle_UniqueId(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UniqueId", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_UniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = UniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = UniqueId.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = UniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = UniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_UniqueId__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = UniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_UniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_UniqueId", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_UniqueId__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UniqueId__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._data);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._data);
  __pyx_v___pyx_result->__pyx_base._data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_UniqueId__set_state(<UniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UniqueId__set_state(UniqueId __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_UniqueId__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_team_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_435__pyx_unpickle_team_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_434__pyx_unpickle_team_uniqueid, "__pyx_unpickle_team_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_435__pyx_unpickle_team_uniqueid = {"__pyx_unpickle_team_uniqueid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_435__pyx_unpickle_team_uniqueid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_434__pyx_unpickle_team_uniqueid};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_435__pyx_unpickle_team_uniqueid(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_team_uniqueid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_team_uniqueid", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_team_uniqueid", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_team_uniqueid", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_team_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_434__pyx_unpickle_team_uniqueid(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_434__pyx_unpickle_team_uniqueid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_team_uniqueid", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_team_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = team_uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = team_uniqueid.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = team_uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = team_uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_team_uniqueid__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = team_uniqueid.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_team_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_team_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_team_uniqueid__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_team_uniqueid__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_data);
  __Pyx_DECREF(__pyx_v___pyx_result->_data);
  __pyx_v___pyx_result->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_team_uniqueid__set_state(<team_uniqueid> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_team_uniqueid__set_state(team_uniqueid __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_team_uniqueid__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_TeamUniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_437__pyx_unpickle_TeamUniqueId(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_436__pyx_unpickle_TeamUniqueId, "__pyx_unpickle_TeamUniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_437__pyx_unpickle_TeamUniqueId = {"__pyx_unpickle_TeamUniqueId", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_437__pyx_unpickle_TeamUniqueId, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_436__pyx_unpickle_TeamUniqueId};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_437__pyx_unpickle_TeamUniqueId(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamUniqueId (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_TeamUniqueId", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TeamUniqueId", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TeamUniqueId", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamUniqueId", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_436__pyx_unpickle_TeamUniqueId(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_436__pyx_unpickle_TeamUniqueId(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamUniqueId", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_TeamUniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = TeamUniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamUniqueId.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamUniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = TeamUniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamUniqueId__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamUniqueId.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_TeamUniqueId(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamUniqueId", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamUniqueId__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamUniqueId__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._data);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._data);
  __pyx_v___pyx_result->__pyx_base._data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_TeamUniqueId__set_state(<TeamUniqueId> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamUniqueId__set_state(TeamUniqueId __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamUniqueId__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InitAttr(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_439__pyx_unpickle_InitAttr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_438__pyx_unpickle_InitAttr, "__pyx_unpickle_InitAttr(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_439__pyx_unpickle_InitAttr = {"__pyx_unpickle_InitAttr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_439__pyx_unpickle_InitAttr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_438__pyx_unpickle_InitAttr};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_439__pyx_unpickle_InitAttr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InitAttr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_InitAttr", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InitAttr", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InitAttr", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_InitAttr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_438__pyx_unpickle_InitAttr(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_438__pyx_unpickle_InitAttr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InitAttr", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_InitAttr(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = InitAttr.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = InitAttr.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = InitAttr.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = InitAttr.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_InitAttr__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = InitAttr.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InitAttr(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_InitAttr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_InitAttr__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InitAttr__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_data);
  __Pyx_DECREF(__pyx_v___pyx_result->_data);
  __pyx_v___pyx_result->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_InitAttr__set_state(<InitAttr> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InitAttr__set_state(InitAttr __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_InitAttr__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_TeamConfig(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

/* Python wrapper */
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_441__pyx_unpickle_TeamConfig(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7nvshmem_8bindings_7nvshmem_440__pyx_unpickle_TeamConfig, "__pyx_unpickle_TeamConfig(__pyx_type, long __pyx_checksum, tuple __pyx_state)");
static PyMethodDef __pyx_mdef_7nvshmem_8bindings_7nvshmem_441__pyx_unpickle_TeamConfig = {"__pyx_unpickle_TeamConfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_441__pyx_unpickle_TeamConfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_440__pyx_unpickle_TeamConfig};
static PyObject *__pyx_pw_7nvshmem_8bindings_7nvshmem_441__pyx_unpickle_TeamConfig(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamConfig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len < 0)) __PYX_ERR(2, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_TeamConfig", 0) < (0)) __PYX_ERR(2, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TeamConfig", 1, 3, 3, i); __PYX_ERR(2, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(2, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(2, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TeamConfig", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamConfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(2, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_7nvshmem_8bindings_7nvshmem_440__pyx_unpickle_TeamConfig(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7nvshmem_8bindings_7nvshmem_440__pyx_unpickle_TeamConfig(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamConfig", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_TeamConfig(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')             # <<<<<<<<<<<<<<
 *     __pyx_result = TeamConfig.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, __pyx_k_data_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamConfig.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamConfig.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = TeamConfig.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(2, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamConfig__set_state(((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
 *     __pyx_result = TeamConfig.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_TeamConfig(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamConfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

static PyObject *__pyx_f_7nvshmem_8bindings_7nvshmem___pyx_unpickle_TeamConfig__set_state(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TeamConfig__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_data);
  __Pyx_DECREF(__pyx_v___pyx_result->_data);
  __pyx_v___pyx_result->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_TeamConfig__set_state(<TeamConfig> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TeamConfig__set_state(TeamConfig __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result._data = __pyx_state[0]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nvshmem.bindings.nvshmem.__pyx_unpickle_TeamConfig__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */

static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)o);
  p->_data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_data);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_data) {
    e = (*v)(p->_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid *)o;
  tmp = ((PyObject*)p->_data);
  p->_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_sq_item_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyLong_FromSsize_t(i); if(!x) return 0;
  #if CYTHON_USE_TYPE_SLOTS || (!CYTHON_USE_TYPE_SPECS && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  #else
  r = ((binaryfunc)PyType_GetSlot(Py_TYPE(o), Py_mp_subscript))(o, x);
  #endif
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_13__setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_tp_richcompare_7nvshmem_8bindings_7nvshmem_uniqueid(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_9__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_9__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid_ptr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3ptr_1__get__(o);
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_8uniqueid_version(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7version_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid__data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_uniqueid[] = {
  {"from_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_14from_data},
  {"from_ptr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_16from_ptr},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8uniqueid_20__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7nvshmem_8bindings_7nvshmem_uniqueid[] = {
  {"ptr", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid_ptr, 0, PyDoc_STR("Get the pointer address to the data as Python :py:`int`."), 0},
  {"version", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid_version, __pyx_setprop_7nvshmem_8bindings_7nvshmem_8uniqueid_version, PyDoc_STR("version (~_numpy.int32): "), 0},
  {"_data", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8uniqueid__data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_repr, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3__repr__},
  {Py_nb_int, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5__int__},
  {Py_sq_length, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__},
  {Py_sq_item, (void *)__pyx_sq_item_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_mp_length, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__},
  {Py_mp_subscript, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_11__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_doc, (void *)PyDoc_STR("uniqueid(size=1)\n\nEmpty-initialize an array of `nvshmemx_uniqueid_v1`.\n\nThe resulting object is of length `size` and of dtype `uniqueid_dtype`.\nIf default-constructed, the instance represents a single struct.\n\nArgs:\n    size (int): number of structs, default=1.\n\n\n.. seealso:: `nvshmemx_uniqueid_v1`")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_getset, (void *)__pyx_getsets_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_init, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid_spec = {
  "nvshmem.bindings.nvshmem.uniqueid",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_uniqueid = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_bool*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_5__int__, /*nb_int*/
  0, /*nb_long (reserved)*/
  0, /*nb_float*/
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  0, /*nb_matrix_multiply*/
  0, /*nb_inplace_matrix_multiply*/
};

static PySequenceMethods __pyx_tp_as_sequence_uniqueid = {
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_7nvshmem_8bindings_7nvshmem_uniqueid, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_uniqueid = {
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_7__len__, /*mp_length*/
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_11__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_7nvshmem_8bindings_7nvshmem_uniqueid, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""uniqueid", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_uniqueid), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3__repr__, /*tp_repr*/
  &__pyx_tp_as_number_uniqueid, /*tp_as_number*/
  &__pyx_tp_as_sequence_uniqueid, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_uniqueid, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("uniqueid(size=1)\n\nEmpty-initialize an array of `nvshmemx_uniqueid_v1`.\n\nThe resulting object is of length `size` and of dtype `uniqueid_dtype`.\nIf default-constructed, the instance represents a single struct.\n\nArgs:\n    size (int): number of structs, default=1.\n\n\n.. seealso:: `nvshmemx_uniqueid_v1`"), /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_clear*/
  __pyx_tp_richcompare_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_UniqueId[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8UniqueId_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_uniqueid},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_UniqueId},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId_spec = {
  "nvshmem.bindings.nvshmem.UniqueId",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId_slots,
};
#else

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""UniqueId", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_UniqueId), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_UniqueId, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8uniqueid_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_uniqueid, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)o);
  p->_data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid(PyObject *o) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_data);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_team_uniqueid(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_data) {
    e = (*v)(p->_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_team_uniqueid(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid *)o;
  tmp = ((PyObject*)p->_data);
  p->_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_ptr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3ptr_1__get__(o);
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_value(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5value_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid__data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_team_uniqueid[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_6__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid_8__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7nvshmem_8bindings_7nvshmem_team_uniqueid[] = {
  {"ptr", __pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_ptr, 0, PyDoc_STR("Get the pointer address to the data as Python :py:`int`."), 0},
  {"value", __pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_value, __pyx_setprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid_value, PyDoc_STR("team_uniqueid.value: int\n\nGet the value as Python int."), 0},
  {"_data", __pyx_getprop_7nvshmem_8bindings_7nvshmem_13team_uniqueid__data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_repr, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5__repr__},
  {Py_tp_str, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3__str__},
  {Py_tp_doc, (void *)PyDoc_STR("team_uniqueid(int value: int = 0)\n\nSimple wrapper for nvshmemx_team_uniqueid_t (typedef uint64_t).")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_getset, (void *)__pyx_getsets_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_init, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid_spec = {
  "nvshmem.bindings.nvshmem.team_uniqueid",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid_slots,
};
#else

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""team_uniqueid", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_team_uniqueid), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("team_uniqueid(int value: int = 0)\n\nSimple wrapper for nvshmemx_team_uniqueid_t (typedef uint64_t)."), /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_TeamUniqueId[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_TeamUniqueId},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId_spec = {
  "nvshmem.bindings.nvshmem.TeamUniqueId",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId_slots,
};
#else

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""TeamUniqueId", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamUniqueId), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_5__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_3__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_TeamUniqueId, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7nvshmem_8bindings_7nvshmem_13team_uniqueid_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_team_uniqueid, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_InitAttr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)o);
  p->_data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_InitAttr(PyObject *o) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_InitAttr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_data);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_InitAttr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_data) {
    e = (*v)(p->_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_InitAttr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr *)o;
  tmp = ((PyObject*)p->_data);
  p->_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_ptr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3ptr_1__get__(o);
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_8InitAttr_version(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_7version_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_mpi_comm(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_8InitAttr_mpi_comm(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_8mpi_comm_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr__data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_InitAttr[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_8InitAttr_4__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7nvshmem_8bindings_7nvshmem_InitAttr[] = {
  {"ptr", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_ptr, 0, PyDoc_STR("Get the pointer address to the data as Python :py:`int`."), 0},
  {"version", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_version, __pyx_setprop_7nvshmem_8bindings_7nvshmem_8InitAttr_version, PyDoc_STR("version (~_numpy.int32): "), 0},
  {"mpi_comm", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr_mpi_comm, __pyx_setprop_7nvshmem_8bindings_7nvshmem_8InitAttr_mpi_comm, PyDoc_STR("mpi_comm (~_numpy.intp): "), 0},
  {"_data", __pyx_getprop_7nvshmem_8bindings_7nvshmem_8InitAttr__data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_InitAttr},
  {Py_tp_doc, (void *)PyDoc_STR("InitAttr()")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_InitAttr},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_InitAttr},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_InitAttr},
  {Py_tp_getset, (void *)__pyx_getsets_7nvshmem_8bindings_7nvshmem_InitAttr},
  {Py_tp_init, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_InitAttr},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr_spec = {
  "nvshmem.bindings.nvshmem.InitAttr",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr_slots,
};
#else

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""InitAttr", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_InitAttr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("InitAttr()"), /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7nvshmem_8bindings_7nvshmem_8InitAttr_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_InitAttr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7nvshmem_8bindings_7nvshmem_TeamConfig(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)o);
  p->_data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_TeamConfig(PyObject *o) {
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_TeamConfig) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_data);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_TeamConfig(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_data) {
    e = (*v)(p->_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_TeamConfig(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *p = (struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig *)o;
  tmp = ((PyObject*)p->_data);
  p->_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_ptr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3ptr_1__get__(o);
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_version(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_7version_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_num_contexts(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_num_contexts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_12num_contexts_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_uniqueid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_1__get__(o);
}

static int __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_uniqueid(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_8uniqueid_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig__data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7nvshmem_8bindings_7nvshmem_TeamConfig[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7nvshmem_8bindings_7nvshmem_10TeamConfig_4__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7nvshmem_8bindings_7nvshmem_TeamConfig[] = {
  {"ptr", __pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_ptr, 0, PyDoc_STR("Get the pointer address to the data as Python :py:`int`."), 0},
  {"version", __pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_version, __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_version, PyDoc_STR("version (~_numpy.int32): "), 0},
  {"num_contexts", __pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_num_contexts, __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_num_contexts, PyDoc_STR("num_contexts (~_numpy.int32): "), 0},
  {"uniqueid", __pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_uniqueid, __pyx_setprop_7nvshmem_8bindings_7nvshmem_10TeamConfig_uniqueid, PyDoc_STR("uniqueid (~_numpy.uint64): "), 0},
  {"_data", __pyx_getprop_7nvshmem_8bindings_7nvshmem_10TeamConfig__data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {Py_tp_doc, (void *)PyDoc_STR("TeamConfig()")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {Py_tp_clear, (void *)__pyx_tp_clear_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {Py_tp_methods, (void *)__pyx_methods_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {Py_tp_getset, (void *)__pyx_getsets_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {Py_tp_init, (void *)__pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7nvshmem_8bindings_7nvshmem_TeamConfig},
  {0, 0},
};
static PyType_Spec __pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig_spec = {
  "nvshmem.bindings.nvshmem.TeamConfig",
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig_slots,
};
#else

static PyTypeObject __pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig = {
  PyVarObject_HEAD_INIT(0, 0)
  "nvshmem.bindings.nvshmem.""TeamConfig", /*tp_name*/
  sizeof(struct __pyx_obj_7nvshmem_8bindings_7nvshmem_TeamConfig), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("TeamConfig()"), /*tp_doc*/
  __pyx_tp_traverse_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_traverse*/
  __pyx_tp_clear_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7nvshmem_8bindings_7nvshmem_10TeamConfig_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7nvshmem_8bindings_7nvshmem_TeamConfig, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  {
    __pyx_t_1 = __Pyx_ApiExport_GetApiDict(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_export_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_int_int_intptr_t_intptr);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_export_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_export_name = __pyx_export_signature + 2040;
    void (*const __pyx_export_pointers[])(void) = {(void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_uniqueid_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_barrier, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_strided, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_split_2d, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_barrier_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_bfloat16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_char_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_double_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_float_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_half_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_int_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_long_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_longlong_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_schar_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_short_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_size_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_uint8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_set_attr_mpi_comm_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_init_attr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_init_status, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_translate_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_sync_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_culibrary_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_cumodule_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_buffer_unregister_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_mc_ptr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_ptr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_buffer_register_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_malloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_align, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_calloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_hostlib_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_destroy, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_team_get_config, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_barrier_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_free, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_quiet_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_sync_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_signal_wait_until_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_info_get_version, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_vendor_get_version_info, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_getmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_putmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_putmem_signal_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_7nvshmem_signal_op_on_stream, (void (*)(void)) NULL};
    void (*const *__pyx_export_pointer)(void) = __pyx_export_pointers;
    const char *__pyx_export_current_signature = __pyx_export_signature;
    while (*__pyx_export_pointer) {
      if (__Pyx_ExportFunction(__pyx_t_1, __pyx_export_name, *__pyx_export_pointer, __pyx_export_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_export_pointer;
      __pyx_export_name = strchr(__pyx_export_name, '\0') + 1;
      __pyx_export_signature = strchr(__pyx_export_signature, '\0') + 1;
      if (*__pyx_export_signature != '\0') __pyx_export_current_signature = __pyx_export_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid = &__pyx_type_7nvshmem_8bindings_7nvshmem_uniqueid;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_uniqueid, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId)) __PYX_ERR(0, 150, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId) < (0)) __PYX_ERR(0, 150, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId = &__pyx_type_7nvshmem_8bindings_7nvshmem_UniqueId;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId->tp_base = __pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId) < (0)) __PYX_ERR(0, 150, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_UniqueId, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId) < (0)) __PYX_ERR(0, 150, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId) < (0)) __PYX_ERR(0, 150, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid)) __PYX_ERR(0, 154, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid) < (0)) __PYX_ERR(0, 154, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid = &__pyx_type_7nvshmem_8bindings_7nvshmem_team_uniqueid;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid) < (0)) __PYX_ERR(0, 154, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 154, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__.doc = __pyx_doc_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7nvshmem_8bindings_7nvshmem_13team_uniqueid___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_team_uniqueid_2, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid) < (0)) __PYX_ERR(0, 154, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid) < (0)) __PYX_ERR(0, 154, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId)) __PYX_ERR(0, 187, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId) < (0)) __PYX_ERR(0, 187, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId = &__pyx_type_7nvshmem_8bindings_7nvshmem_TeamUniqueId;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId->tp_base = __pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId) < (0)) __PYX_ERR(0, 187, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_TeamUniqueId, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId) < (0)) __PYX_ERR(0, 187, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId) < (0)) __PYX_ERR(0, 187, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr)) __PYX_ERR(0, 201, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr) < (0)) __PYX_ERR(0, 201, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr = &__pyx_type_7nvshmem_8bindings_7nvshmem_InitAttr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr) < (0)) __PYX_ERR(0, 201, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_InitAttr, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr) < (0)) __PYX_ERR(0, 201, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr) < (0)) __PYX_ERR(0, 201, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig)) __PYX_ERR(0, 249, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig_spec, __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig = &__pyx_type_7nvshmem_8bindings_7nvshmem_TeamConfig;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig->tp_dictoffset && __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_TeamConfig, (PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig) < (0)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_8); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_2_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyLongObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyLongObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_8); if (!__pyx_mstate->__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_2_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyComplexObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_8(PyComplexObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_8); if (!__pyx_mstate->__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("nvshmem.bindings.cynvshmem"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_int_CUlibrary_int_CUmodule_int_i);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 4132;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_uniqueid_args, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_my_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_n_pes, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_sync_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_strided, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_2d, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_translate_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_get_uniqueid, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_get_uniqueid, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_init_attr, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_mpi_comm_args, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_unregister_symmetric, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_my_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_n_pes, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_init_status, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_all_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_quiet_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_sync_all_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_info_get_version, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_vendor_get_version_info, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_destroy, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_get_config, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_wait_until_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_op_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_free, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_getmem_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_signal_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier_all, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_mc_ptr, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_malloc, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_align, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_calloc, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_register_symmetric, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_ptr, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_8(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_nvshmem(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_nvshmem},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "nvshmem",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_nvshmem(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_nvshmem(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_nvshmem(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'nvshmem' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "nvshmem" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_nvshmem", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_nvshmem__bindings__nvshmem) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "nvshmem.bindings.nvshmem")) {
      if (unlikely((PyDict_SetItemString(modules, "nvshmem.bindings.nvshmem", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_export_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "nvshmem/bindings/nvshmem.pyx":18
 * )
 * 
 * from enum import IntEnum as _IntEnum             # <<<<<<<<<<<<<<
 * 
 * import numpy as _numpy
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 18, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IntEnum_2, __pyx_t_4) < (0)) __PYX_ERR(0, 18, __pyx_L1_error)
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":20
 * from enum import IntEnum as _IntEnum
 * 
 * import numpy as _numpy             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_numpy_2, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_numpy, __pyx_t_2) < (0)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":27
 * ###############################################################################
 * 
 * uniqueid_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("internal", _numpy.int8, (124,)),
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":28
 * 
 * uniqueid_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),             # <<<<<<<<<<<<<<
 *     ("internal", _numpy.int8, (124,)),
 *     ], align=True)
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_int32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_version, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":29
 * uniqueid_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),
 *     ("internal", _numpy.int8, (124,)),             # <<<<<<<<<<<<<<
 *     ], align=True)
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_int8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_Pack(3, __pyx_mstate_global->__pyx_n_u_internal, __pyx_t_8, __pyx_mstate_global->__pyx_tuple[0]); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":27
 * ###############################################################################
 * 
 * uniqueid_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("internal", _numpy.int8, (124,)),
*/
  __pyx_t_8 = __Pyx_PyList_Pack(2, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":30
 *     ("version", _numpy.int32, ),
 *     ("internal", _numpy.int8, (124,)),
 *     ], align=True)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_t_8};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_align, Py_True, __pyx_t_7, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uniqueid_dtype, __pyx_t_2) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":110
 *         self._data[key] = val
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_data(data):
 *         """Create an uniqueid instance wrapping the given NumPy array.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_15from_data, __Pyx_CYFUNCTION_STATICMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_uniqueid_from_data, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_data, __pyx_t_2) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = NULL;
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject*)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_staticmethod, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_data, __pyx_t_2) < (0)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":128
 *         return obj
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):
 *         """Create an uniqueid instance wrapping the given pointer.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_17from_ptr, __Pyx_CYFUNCTION_STATICMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_uniqueid_from_ptr, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_ptr, __pyx_t_2) < (0)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = NULL;
  __Pyx_GetNameInClass(__pyx_t_6, (PyObject*)__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_ptr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_staticmethod, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_from_ptr, __pyx_t_2) < (0)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_uniqueid___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_uniqueid__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8uniqueid_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_uniqueid___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_uniqueid, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8UniqueId_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_UniqueId___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UniqueId__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8UniqueId_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_UniqueId___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_UniqueId, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_13team_uniqueid_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_team_uniqueid___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_team_uniqueid, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_team_uniqueid__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_13team_uniqueid_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_team_uniqueid___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_team_uniqueid, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_TeamUniqueId___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamUniqueId, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamUniqueId__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_12TeamUniqueId_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_TeamUniqueId___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamUniqueId, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":194
 * # TODO: remove this once cybind supports nested structs.
 * 
 * init_attr_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("mpi_comm", _numpy.intp, ),
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":195
 * 
 * init_attr_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),             # <<<<<<<<<<<<<<
 *     ("mpi_comm", _numpy.intp, ),
 *     ("args", _numpy.int8, (sizeof(nvshmemx_init_args_t),)),  # opaque
*/
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_int32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_version, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":196
 * init_attr_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),
 *     ("mpi_comm", _numpy.intp, ),             # <<<<<<<<<<<<<<
 *     ("args", _numpy.int8, (sizeof(nvshmemx_init_args_t),)),  # opaque
 *     ], align=True)
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_intp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_mpi_comm, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":197
 *     ("version", _numpy.int32, ),
 *     ("mpi_comm", _numpy.intp, ),
 *     ("args", _numpy.int8, (sizeof(nvshmemx_init_args_t),)),  # opaque             # <<<<<<<<<<<<<<
 *     ], align=True)
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_int8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyLong_FromSize_t((sizeof(nvshmemx_init_args_t))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = PyTuple_Pack(1, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_Pack(3, __pyx_mstate_global->__pyx_n_u_args, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":194
 * # TODO: remove this once cybind supports nested structs.
 * 
 * init_attr_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("mpi_comm", _numpy.intp, ),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(3, __pyx_t_7, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":198
 *     ("mpi_comm", _numpy.intp, ),
 *     ("args", _numpy.int8, (sizeof(nvshmemx_init_args_t),)),  # opaque
 *     ], align=True)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, __pyx_t_11};
    __pyx_t_5 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_align, Py_True, __pyx_t_5, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 194, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_init_attr_dtype, __pyx_t_2) < (0)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8InitAttr_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_InitAttr___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InitAttr, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InitAttr__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_8InitAttr_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_InitAttr___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_InitAttr, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":241
 * # TODO: remove this once cybind supports nested structs.
 * 
 * team_config_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("num_contexts", _numpy.int32, ),
*/
  __pyx_t_8 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":242
 * 
 * team_config_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),             # <<<<<<<<<<<<<<
 *     ("num_contexts", _numpy.int32, ),
 *     ("uniqueid", _numpy.uint64, ),
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_int32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_version, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":243
 * team_config_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),
 *     ("num_contexts", _numpy.int32, ),             # <<<<<<<<<<<<<<
 *     ("uniqueid", _numpy.uint64, ),
 *     ("padding", _numpy.int8, (48,)),  # TEAM_CONFIG_V2_PADDING = 48
*/
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_int32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_num_contexts, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":244
 *     ("version", _numpy.int32, ),
 *     ("num_contexts", _numpy.int32, ),
 *     ("uniqueid", _numpy.uint64, ),             # <<<<<<<<<<<<<<
 *     ("padding", _numpy.int8, (48,)),  # TEAM_CONFIG_V2_PADDING = 48
 *     ], align=True)
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_uint64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_uniqueid, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":245
 *     ("num_contexts", _numpy.int32, ),
 *     ("uniqueid", _numpy.uint64, ),
 *     ("padding", _numpy.int8, (48,)),  # TEAM_CONFIG_V2_PADDING = 48             # <<<<<<<<<<<<<<
 *     ], align=True)
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_int8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_Pack(3, __pyx_mstate_global->__pyx_n_u_padding, __pyx_t_10, __pyx_mstate_global->__pyx_tuple[2]); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":241
 * # TODO: remove this once cybind supports nested structs.
 * 
 * team_config_dtype = _numpy.dtype([             # <<<<<<<<<<<<<<
 *     ("version", _numpy.int32, ),
 *     ("num_contexts", _numpy.int32, ),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(4, __pyx_t_5, __pyx_t_6, __pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":246
 *     ("uniqueid", _numpy.uint64, ),
 *     ("padding", _numpy.int8, (48,)),  # TEAM_CONFIG_V2_PADDING = 48
 *     ], align=True)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_8, __pyx_t_10};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_align, Py_True, __pyx_t_7, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 241, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_config_dtype, __pyx_t_2) < (0)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_10TeamConfig_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_TeamConfig___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TeamConfig, (type(self), 0xa75e18a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TeamConfig__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_10TeamConfig_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_TeamConfig___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7nvshmem_8bindings_7nvshmem_TeamConfig, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":297
 * ###############################################################################
 * 
 * class Signal_op(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_signal_op_t`."""
 *     SIGNAL_SET = NVSHMEM_SIGNAL_SET
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Signal_op, __pyx_mstate_global->__pyx_n_u_Signal_op, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_signal_op_t); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_2 != __pyx_t_11) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_11) < 0))) __PYX_ERR(0, 297, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":299
 * class Signal_op(_IntEnum):
 *     """See `nvshmemx_signal_op_t`."""
 *     SIGNAL_SET = NVSHMEM_SIGNAL_SET             # <<<<<<<<<<<<<<
 *     SIGNAL_ADD = NVSHMEM_SIGNAL_ADD
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_signal_op_t(NVSHMEM_SIGNAL_SET); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_SIGNAL_SET, __pyx_t_11) < (0)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":300
 *     """See `nvshmemx_signal_op_t`."""
 *     SIGNAL_SET = NVSHMEM_SIGNAL_SET
 *     SIGNAL_ADD = NVSHMEM_SIGNAL_ADD             # <<<<<<<<<<<<<<
 * 
 * class Tma_policy(_IntEnum):
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_signal_op_t(NVSHMEM_SIGNAL_ADD); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_SIGNAL_ADD, __pyx_t_11) < (0)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":297
 * ###############################################################################
 * 
 * class Signal_op(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_signal_op_t`."""
 *     SIGNAL_SET = NVSHMEM_SIGNAL_SET
*/
  __pyx_t_11 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_Signal_op, __pyx_t_2, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Signal_op, __pyx_t_11) < (0)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":302
 *     SIGNAL_ADD = NVSHMEM_SIGNAL_ADD
 * 
 * class Tma_policy(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_tma_policy_t`."""
 *     TMA_DISABLE = NVSHMEMX_TMA_DISABLE
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Tma_policy, __pyx_mstate_global->__pyx_n_u_Tma_policy, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_tma_policy_t); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__pyx_t_2 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_11, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 302, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":304
 * class Tma_policy(_IntEnum):
 *     """See `nvshmemx_tma_policy_t`."""
 *     TMA_DISABLE = NVSHMEMX_TMA_DISABLE             # <<<<<<<<<<<<<<
 *     TMA_ENABLE = NVSHMEMX_TMA_ENABLE
 *     TMA_FORCE = NVSHMEMX_TMA_FORCE
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_tma_policy_t(NVSHMEMX_TMA_DISABLE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_TMA_DISABLE, __pyx_t_7) < (0)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":305
 *     """See `nvshmemx_tma_policy_t`."""
 *     TMA_DISABLE = NVSHMEMX_TMA_DISABLE
 *     TMA_ENABLE = NVSHMEMX_TMA_ENABLE             # <<<<<<<<<<<<<<
 *     TMA_FORCE = NVSHMEMX_TMA_FORCE
 *     MAX = NVSHMEMX_TMA_POLICY_MAX
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_tma_policy_t(NVSHMEMX_TMA_ENABLE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_TMA_ENABLE, __pyx_t_7) < (0)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":306
 *     TMA_DISABLE = NVSHMEMX_TMA_DISABLE
 *     TMA_ENABLE = NVSHMEMX_TMA_ENABLE
 *     TMA_FORCE = NVSHMEMX_TMA_FORCE             # <<<<<<<<<<<<<<
 *     MAX = NVSHMEMX_TMA_POLICY_MAX
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_tma_policy_t(NVSHMEMX_TMA_FORCE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_TMA_FORCE, __pyx_t_7) < (0)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":307
 *     TMA_ENABLE = NVSHMEMX_TMA_ENABLE
 *     TMA_FORCE = NVSHMEMX_TMA_FORCE
 *     MAX = NVSHMEMX_TMA_POLICY_MAX             # <<<<<<<<<<<<<<
 * 
 * class Smem_amount(_IntEnum):
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_tma_policy_t(NVSHMEMX_TMA_POLICY_MAX); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_MAX, __pyx_t_7) < (0)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":302
 *     SIGNAL_ADD = NVSHMEM_SIGNAL_ADD
 * 
 * class Tma_policy(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_tma_policy_t`."""
 *     TMA_DISABLE = NVSHMEMX_TMA_DISABLE
*/
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Tma_policy, __pyx_t_2, __pyx_t_11, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_7);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Tma_policy, __pyx_t_7) < (0)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":309
 *     MAX = NVSHMEMX_TMA_POLICY_MAX
 * 
 * class Smem_amount(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_smem_amount_t`."""
 *     SMEM_RECOMMENDED = NVSHMEMX_SMEM_RECOMMENDED
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Smem_amount, __pyx_mstate_global->__pyx_n_u_Smem_amount, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_smem_amount_t); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_2 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 309, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":311
 * class Smem_amount(_IntEnum):
 *     """See `nvshmemx_smem_amount_t`."""
 *     SMEM_RECOMMENDED = NVSHMEMX_SMEM_RECOMMENDED             # <<<<<<<<<<<<<<
 *     SMEM_MINIMUM = NVSHMEMX_SMEM_MINIMUM
 *     SMEM_BARRIERS_ONLY = NVSHMEMX_SMEM_BARRIERS_ONLY
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_smem_amount_t(NVSHMEMX_SMEM_RECOMMENDED); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_SMEM_RECOMMENDED, __pyx_t_10) < (0)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":312
 *     """See `nvshmemx_smem_amount_t`."""
 *     SMEM_RECOMMENDED = NVSHMEMX_SMEM_RECOMMENDED
 *     SMEM_MINIMUM = NVSHMEMX_SMEM_MINIMUM             # <<<<<<<<<<<<<<
 *     SMEM_BARRIERS_ONLY = NVSHMEMX_SMEM_BARRIERS_ONLY
 *     MAX = NVSHMEMX_SMEM_AMOUNT_MAX
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_smem_amount_t(NVSHMEMX_SMEM_MINIMUM); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_SMEM_MINIMUM, __pyx_t_10) < (0)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":313
 *     SMEM_RECOMMENDED = NVSHMEMX_SMEM_RECOMMENDED
 *     SMEM_MINIMUM = NVSHMEMX_SMEM_MINIMUM
 *     SMEM_BARRIERS_ONLY = NVSHMEMX_SMEM_BARRIERS_ONLY             # <<<<<<<<<<<<<<
 *     MAX = NVSHMEMX_SMEM_AMOUNT_MAX
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_smem_amount_t(NVSHMEMX_SMEM_BARRIERS_ONLY); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_SMEM_BARRIERS_ONLY, __pyx_t_10) < (0)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":314
 *     SMEM_MINIMUM = NVSHMEMX_SMEM_MINIMUM
 *     SMEM_BARRIERS_ONLY = NVSHMEMX_SMEM_BARRIERS_ONLY
 *     MAX = NVSHMEMX_SMEM_AMOUNT_MAX             # <<<<<<<<<<<<<<
 * 
 * class Cmp_type(_IntEnum):
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_smem_amount_t(NVSHMEMX_SMEM_AMOUNT_MAX); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_MAX, __pyx_t_10) < (0)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":309
 *     MAX = NVSHMEMX_TMA_POLICY_MAX
 * 
 * class Smem_amount(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_smem_amount_t`."""
 *     SMEM_RECOMMENDED = NVSHMEMX_SMEM_RECOMMENDED
*/
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Smem_amount, __pyx_t_2, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_10);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Smem_amount, __pyx_t_10) < (0)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":316
 *     MAX = NVSHMEMX_SMEM_AMOUNT_MAX
 * 
 * class Cmp_type(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_cmp_type_t`."""
 *     CMP_EQ = NVSHMEM_CMP_EQ
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Cmp_type, __pyx_mstate_global->__pyx_n_u_Cmp_type, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_cmp_type_t); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_2 != __pyx_t_11) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_11) < 0))) __PYX_ERR(0, 316, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":318
 * class Cmp_type(_IntEnum):
 *     """See `nvshmemx_cmp_type_t`."""
 *     CMP_EQ = NVSHMEM_CMP_EQ             # <<<<<<<<<<<<<<
 *     CMP_NE = NVSHMEM_CMP_NE
 *     CMP_GT = NVSHMEM_CMP_GT
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_EQ); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_EQ, __pyx_t_11) < (0)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":319
 *     """See `nvshmemx_cmp_type_t`."""
 *     CMP_EQ = NVSHMEM_CMP_EQ
 *     CMP_NE = NVSHMEM_CMP_NE             # <<<<<<<<<<<<<<
 *     CMP_GT = NVSHMEM_CMP_GT
 *     CMP_LE = NVSHMEM_CMP_LE
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_NE); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_NE, __pyx_t_11) < (0)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":320
 *     CMP_EQ = NVSHMEM_CMP_EQ
 *     CMP_NE = NVSHMEM_CMP_NE
 *     CMP_GT = NVSHMEM_CMP_GT             # <<<<<<<<<<<<<<
 *     CMP_LE = NVSHMEM_CMP_LE
 *     CMP_LT = NVSHMEM_CMP_LT
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_GT); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_GT, __pyx_t_11) < (0)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":321
 *     CMP_NE = NVSHMEM_CMP_NE
 *     CMP_GT = NVSHMEM_CMP_GT
 *     CMP_LE = NVSHMEM_CMP_LE             # <<<<<<<<<<<<<<
 *     CMP_LT = NVSHMEM_CMP_LT
 *     CMP_GE = NVSHMEM_CMP_GE
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_LE); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_LE, __pyx_t_11) < (0)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":322
 *     CMP_GT = NVSHMEM_CMP_GT
 *     CMP_LE = NVSHMEM_CMP_LE
 *     CMP_LT = NVSHMEM_CMP_LT             # <<<<<<<<<<<<<<
 *     CMP_GE = NVSHMEM_CMP_GE
 *     CMP_SENTINEL = NVSHMEM_CMP_SENTINEL
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_LT); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_LT, __pyx_t_11) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":323
 *     CMP_LE = NVSHMEM_CMP_LE
 *     CMP_LT = NVSHMEM_CMP_LT
 *     CMP_GE = NVSHMEM_CMP_GE             # <<<<<<<<<<<<<<
 *     CMP_SENTINEL = NVSHMEM_CMP_SENTINEL
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_GE); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_GE, __pyx_t_11) < (0)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":324
 *     CMP_LT = NVSHMEM_CMP_LT
 *     CMP_GE = NVSHMEM_CMP_GE
 *     CMP_SENTINEL = NVSHMEM_CMP_SENTINEL             # <<<<<<<<<<<<<<
 * 
 * class Thread_support(_IntEnum):
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_cmp_type_t(NVSHMEM_CMP_SENTINEL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_CMP_SENTINEL, __pyx_t_11) < (0)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":316
 *     MAX = NVSHMEMX_SMEM_AMOUNT_MAX
 * 
 * class Cmp_type(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_cmp_type_t`."""
 *     CMP_EQ = NVSHMEM_CMP_EQ
*/
  __pyx_t_11 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_Cmp_type, __pyx_t_2, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Cmp_type, __pyx_t_11) < (0)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":326
 *     CMP_SENTINEL = NVSHMEM_CMP_SENTINEL
 * 
 * class Thread_support(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_thread_support_t`."""
 *     THREAD_SINGLE = NVSHMEM_THREAD_SINGLE
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Thread_support, __pyx_mstate_global->__pyx_n_u_Thread_support, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_thread_support_t); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__pyx_t_2 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_11, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":328
 * class Thread_support(_IntEnum):
 *     """See `nvshmemx_thread_support_t`."""
 *     THREAD_SINGLE = NVSHMEM_THREAD_SINGLE             # <<<<<<<<<<<<<<
 *     THREAD_FUNNELED = NVSHMEM_THREAD_FUNNELED
 *     THREAD_SERIALIZED = NVSHMEM_THREAD_SERIALIZED
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_thread_support_t(NVSHMEM_THREAD_SINGLE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_THREAD_SINGLE, __pyx_t_7) < (0)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":329
 *     """See `nvshmemx_thread_support_t`."""
 *     THREAD_SINGLE = NVSHMEM_THREAD_SINGLE
 *     THREAD_FUNNELED = NVSHMEM_THREAD_FUNNELED             # <<<<<<<<<<<<<<
 *     THREAD_SERIALIZED = NVSHMEM_THREAD_SERIALIZED
 *     THREAD_MULTIPLE = NVSHMEM_THREAD_MULTIPLE
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_thread_support_t(NVSHMEM_THREAD_FUNNELED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_THREAD_FUNNELED, __pyx_t_7) < (0)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":330
 *     THREAD_SINGLE = NVSHMEM_THREAD_SINGLE
 *     THREAD_FUNNELED = NVSHMEM_THREAD_FUNNELED
 *     THREAD_SERIALIZED = NVSHMEM_THREAD_SERIALIZED             # <<<<<<<<<<<<<<
 *     THREAD_MULTIPLE = NVSHMEM_THREAD_MULTIPLE
 *     THREAD_TYPE_SENTINEL = NVSHMEM_THREAD_TYPE_SENTINEL
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_thread_support_t(NVSHMEM_THREAD_SERIALIZED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_THREAD_SERIALIZED, __pyx_t_7) < (0)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":331
 *     THREAD_FUNNELED = NVSHMEM_THREAD_FUNNELED
 *     THREAD_SERIALIZED = NVSHMEM_THREAD_SERIALIZED
 *     THREAD_MULTIPLE = NVSHMEM_THREAD_MULTIPLE             # <<<<<<<<<<<<<<
 *     THREAD_TYPE_SENTINEL = NVSHMEM_THREAD_TYPE_SENTINEL
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_thread_support_t(NVSHMEM_THREAD_MULTIPLE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_THREAD_MULTIPLE, __pyx_t_7) < (0)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":332
 *     THREAD_SERIALIZED = NVSHMEM_THREAD_SERIALIZED
 *     THREAD_MULTIPLE = NVSHMEM_THREAD_MULTIPLE
 *     THREAD_TYPE_SENTINEL = NVSHMEM_THREAD_TYPE_SENTINEL             # <<<<<<<<<<<<<<
 * 
 * class Proxy_status(_IntEnum):
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_thread_support_t(NVSHMEM_THREAD_TYPE_SENTINEL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_THREAD_TYPE_SENTINEL, __pyx_t_7) < (0)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":326
 *     CMP_SENTINEL = NVSHMEM_CMP_SENTINEL
 * 
 * class Thread_support(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_thread_support_t`."""
 *     THREAD_SINGLE = NVSHMEM_THREAD_SINGLE
*/
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Thread_support, __pyx_t_2, __pyx_t_11, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_7);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Thread_support, __pyx_t_7) < (0)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":334
 *     THREAD_TYPE_SENTINEL = NVSHMEM_THREAD_TYPE_SENTINEL
 * 
 * class Proxy_status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_proxy_status_t`."""
 *     _PROXY_GLOBAL_EXIT_NOT_REQUESTED = PROXY_GLOBAL_EXIT_NOT_REQUESTED
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Proxy_status, __pyx_mstate_global->__pyx_n_u_Proxy_status, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_proxy_status_t); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_2 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 334, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":336
 * class Proxy_status(_IntEnum):
 *     """See `nvshmemx_proxy_status_t`."""
 *     _PROXY_GLOBAL_EXIT_NOT_REQUESTED = PROXY_GLOBAL_EXIT_NOT_REQUESTED             # <<<<<<<<<<<<<<
 *     _PROXY_GLOBAL_EXIT_INIT = PROXY_GLOBAL_EXIT_INIT
 *     _PROXY_GLOBAL_EXIT_REQUESTED = PROXY_GLOBAL_EXIT_REQUESTED
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_proxy_status_t(PROXY_GLOBAL_EXIT_NOT_REQUESTED); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PROXY_GLOBAL_EXIT_NOT_REQUESTED, __pyx_t_10) < (0)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":337
 *     """See `nvshmemx_proxy_status_t`."""
 *     _PROXY_GLOBAL_EXIT_NOT_REQUESTED = PROXY_GLOBAL_EXIT_NOT_REQUESTED
 *     _PROXY_GLOBAL_EXIT_INIT = PROXY_GLOBAL_EXIT_INIT             # <<<<<<<<<<<<<<
 *     _PROXY_GLOBAL_EXIT_REQUESTED = PROXY_GLOBAL_EXIT_REQUESTED
 *     _PROXY_GLOBAL_EXIT_FINISHED = PROXY_GLOBAL_EXIT_FINISHED
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_proxy_status_t(PROXY_GLOBAL_EXIT_INIT); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PROXY_GLOBAL_EXIT_INIT, __pyx_t_10) < (0)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":338
 *     _PROXY_GLOBAL_EXIT_NOT_REQUESTED = PROXY_GLOBAL_EXIT_NOT_REQUESTED
 *     _PROXY_GLOBAL_EXIT_INIT = PROXY_GLOBAL_EXIT_INIT
 *     _PROXY_GLOBAL_EXIT_REQUESTED = PROXY_GLOBAL_EXIT_REQUESTED             # <<<<<<<<<<<<<<
 *     _PROXY_GLOBAL_EXIT_FINISHED = PROXY_GLOBAL_EXIT_FINISHED
 *     _PROXY_GLOBAL_EXIT_MAX_STATE = PROXY_GLOBAL_EXIT_MAX_STATE
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_proxy_status_t(PROXY_GLOBAL_EXIT_REQUESTED); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PROXY_GLOBAL_EXIT_REQUESTED, __pyx_t_10) < (0)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":339
 *     _PROXY_GLOBAL_EXIT_INIT = PROXY_GLOBAL_EXIT_INIT
 *     _PROXY_GLOBAL_EXIT_REQUESTED = PROXY_GLOBAL_EXIT_REQUESTED
 *     _PROXY_GLOBAL_EXIT_FINISHED = PROXY_GLOBAL_EXIT_FINISHED             # <<<<<<<<<<<<<<
 *     _PROXY_GLOBAL_EXIT_MAX_STATE = PROXY_GLOBAL_EXIT_MAX_STATE
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_proxy_status_t(PROXY_GLOBAL_EXIT_FINISHED); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PROXY_GLOBAL_EXIT_FINISHED, __pyx_t_10) < (0)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":340
 *     _PROXY_GLOBAL_EXIT_REQUESTED = PROXY_GLOBAL_EXIT_REQUESTED
 *     _PROXY_GLOBAL_EXIT_FINISHED = PROXY_GLOBAL_EXIT_FINISHED
 *     _PROXY_GLOBAL_EXIT_MAX_STATE = PROXY_GLOBAL_EXIT_MAX_STATE             # <<<<<<<<<<<<<<
 * 
 * class Init_status(_IntEnum):
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmemx_proxy_status_t(PROXY_GLOBAL_EXIT_MAX_STATE); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PROXY_GLOBAL_EXIT_MAX_STATE, __pyx_t_10) < (0)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":334
 *     THREAD_TYPE_SENTINEL = NVSHMEM_THREAD_TYPE_SENTINEL
 * 
 * class Proxy_status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_proxy_status_t`."""
 *     _PROXY_GLOBAL_EXIT_NOT_REQUESTED = PROXY_GLOBAL_EXIT_NOT_REQUESTED
*/
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Proxy_status, __pyx_t_2, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_10);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Proxy_status, __pyx_t_10) < (0)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":342
 *     _PROXY_GLOBAL_EXIT_MAX_STATE = PROXY_GLOBAL_EXIT_MAX_STATE
 * 
 * class Init_status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_init_status_t`."""
 *     STATUS_NOT_INITIALIZED = NVSHMEM_STATUS_NOT_INITIALIZED
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Init_status, __pyx_mstate_global->__pyx_n_u_Init_status, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_init_status_t); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_2 != __pyx_t_11) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_11) < 0))) __PYX_ERR(0, 342, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":344
 * class Init_status(_IntEnum):
 *     """See `nvshmemx_init_status_t`."""
 *     STATUS_NOT_INITIALIZED = NVSHMEM_STATUS_NOT_INITIALIZED             # <<<<<<<<<<<<<<
 *     STATUS_IS_BOOTSTRAPPED = NVSHMEM_STATUS_IS_BOOTSTRAPPED
 *     STATUS_IS_INITIALIZED = NVSHMEM_STATUS_IS_INITIALIZED
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_NOT_INITIALIZED); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_NOT_INITIALIZED, __pyx_t_11) < (0)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":345
 *     """See `nvshmemx_init_status_t`."""
 *     STATUS_NOT_INITIALIZED = NVSHMEM_STATUS_NOT_INITIALIZED
 *     STATUS_IS_BOOTSTRAPPED = NVSHMEM_STATUS_IS_BOOTSTRAPPED             # <<<<<<<<<<<<<<
 *     STATUS_IS_INITIALIZED = NVSHMEM_STATUS_IS_INITIALIZED
 *     STATUS_LIMITED_MPG = NVSHMEM_STATUS_LIMITED_MPG
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_IS_BOOTSTRAPPED); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_IS_BOOTSTRAPPED, __pyx_t_11) < (0)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":346
 *     STATUS_NOT_INITIALIZED = NVSHMEM_STATUS_NOT_INITIALIZED
 *     STATUS_IS_BOOTSTRAPPED = NVSHMEM_STATUS_IS_BOOTSTRAPPED
 *     STATUS_IS_INITIALIZED = NVSHMEM_STATUS_IS_INITIALIZED             # <<<<<<<<<<<<<<
 *     STATUS_LIMITED_MPG = NVSHMEM_STATUS_LIMITED_MPG
 *     STATUS_FULL_MPG = NVSHMEM_STATUS_FULL_MPG
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_IS_INITIALIZED); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_IS_INITIALIZED, __pyx_t_11) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":347
 *     STATUS_IS_BOOTSTRAPPED = NVSHMEM_STATUS_IS_BOOTSTRAPPED
 *     STATUS_IS_INITIALIZED = NVSHMEM_STATUS_IS_INITIALIZED
 *     STATUS_LIMITED_MPG = NVSHMEM_STATUS_LIMITED_MPG             # <<<<<<<<<<<<<<
 *     STATUS_FULL_MPG = NVSHMEM_STATUS_FULL_MPG
 *     STATUS_INVALID = NVSHMEM_STATUS_INVALID
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_LIMITED_MPG); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_LIMITED_MPG, __pyx_t_11) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":348
 *     STATUS_IS_INITIALIZED = NVSHMEM_STATUS_IS_INITIALIZED
 *     STATUS_LIMITED_MPG = NVSHMEM_STATUS_LIMITED_MPG
 *     STATUS_FULL_MPG = NVSHMEM_STATUS_FULL_MPG             # <<<<<<<<<<<<<<
 *     STATUS_INVALID = NVSHMEM_STATUS_INVALID
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_FULL_MPG); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_FULL_MPG, __pyx_t_11) < (0)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":349
 *     STATUS_LIMITED_MPG = NVSHMEM_STATUS_LIMITED_MPG
 *     STATUS_FULL_MPG = NVSHMEM_STATUS_FULL_MPG
 *     STATUS_INVALID = NVSHMEM_STATUS_INVALID             # <<<<<<<<<<<<<<
 * 
 * class Qp_handle_index(_IntEnum):
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmemx_init_status_t(NVSHMEM_STATUS_INVALID); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_STATUS_INVALID, __pyx_t_11) < (0)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":342
 *     _PROXY_GLOBAL_EXIT_MAX_STATE = PROXY_GLOBAL_EXIT_MAX_STATE
 * 
 * class Init_status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_init_status_t`."""
 *     STATUS_NOT_INITIALIZED = NVSHMEM_STATUS_NOT_INITIALIZED
*/
  __pyx_t_11 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_Init_status, __pyx_t_2, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Init_status, __pyx_t_11) < (0)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":351
 *     STATUS_INVALID = NVSHMEM_STATUS_INVALID
 * 
 * class Qp_handle_index(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_qp_handle_index_t`."""
 *     QP_HOST = NVSHMEMX_QP_HOST
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Qp_handle_index, __pyx_mstate_global->__pyx_n_u_Qp_handle_index, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_qp_handle_index_t); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__pyx_t_2 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_11, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 351, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":353
 * class Qp_handle_index(_IntEnum):
 *     """See `nvshmemx_qp_handle_index_t`."""
 *     QP_HOST = NVSHMEMX_QP_HOST             # <<<<<<<<<<<<<<
 *     QP_DEFAULT = NVSHMEMX_QP_DEFAULT
 *     QP_ANY = NVSHMEMX_QP_ANY
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(NVSHMEMX_QP_HOST); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_QP_HOST, __pyx_t_7) < (0)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":354
 *     """See `nvshmemx_qp_handle_index_t`."""
 *     QP_HOST = NVSHMEMX_QP_HOST
 *     QP_DEFAULT = NVSHMEMX_QP_DEFAULT             # <<<<<<<<<<<<<<
 *     QP_ANY = NVSHMEMX_QP_ANY
 *     QP_ALL = NVSHMEMX_QP_ALL
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(NVSHMEMX_QP_DEFAULT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_QP_DEFAULT, __pyx_t_7) < (0)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":355
 *     QP_HOST = NVSHMEMX_QP_HOST
 *     QP_DEFAULT = NVSHMEMX_QP_DEFAULT
 *     QP_ANY = NVSHMEMX_QP_ANY             # <<<<<<<<<<<<<<
 *     QP_ALL = NVSHMEMX_QP_ALL
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(NVSHMEMX_QP_ANY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_QP_ANY, __pyx_t_7) < (0)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":356
 *     QP_DEFAULT = NVSHMEMX_QP_DEFAULT
 *     QP_ANY = NVSHMEMX_QP_ANY
 *     QP_ALL = NVSHMEMX_QP_ALL             # <<<<<<<<<<<<<<
 * 
 * class Pe_index(_IntEnum):
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(NVSHMEMX_QP_ALL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_QP_ALL, __pyx_t_7) < (0)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":351
 *     STATUS_INVALID = NVSHMEM_STATUS_INVALID
 * 
 * class Qp_handle_index(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_qp_handle_index_t`."""
 *     QP_HOST = NVSHMEMX_QP_HOST
*/
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Qp_handle_index, __pyx_t_2, __pyx_t_11, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_7);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Qp_handle_index, __pyx_t_7) < (0)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":358
 *     QP_ALL = NVSHMEMX_QP_ALL
 * 
 * class Pe_index(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmem_pe_index_t`."""
 *     PE_INVALID = NVSHMEM_PE_INVALID
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Pe_index, __pyx_mstate_global->__pyx_n_u_Pe_index, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmem_pe_index_t); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_2 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 358, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":360
 * class Pe_index(_IntEnum):
 *     """See `nvshmem_pe_index_t`."""
 *     PE_INVALID = NVSHMEM_PE_INVALID             # <<<<<<<<<<<<<<
 *     PE_ANY = NVSHMEMX_PE_ANY
 *     PE_ALL = NVSHMEMX_PE_ALL
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmem_pe_index_t(NVSHMEM_PE_INVALID); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PE_INVALID, __pyx_t_10) < (0)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":361
 *     """See `nvshmem_pe_index_t`."""
 *     PE_INVALID = NVSHMEM_PE_INVALID
 *     PE_ANY = NVSHMEMX_PE_ANY             # <<<<<<<<<<<<<<
 *     PE_ALL = NVSHMEMX_PE_ALL
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmem_pe_index_t(NVSHMEMX_PE_ANY); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PE_ANY, __pyx_t_10) < (0)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":362
 *     PE_INVALID = NVSHMEM_PE_INVALID
 *     PE_ANY = NVSHMEMX_PE_ANY
 *     PE_ALL = NVSHMEMX_PE_ALL             # <<<<<<<<<<<<<<
 * 
 * class Team_id(_IntEnum):
*/
  __pyx_t_10 = __Pyx_PyLong_From_nvshmem_pe_index_t(NVSHMEMX_PE_ALL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PE_ALL, __pyx_t_10) < (0)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":358
 *     QP_ALL = NVSHMEMX_QP_ALL
 * 
 * class Pe_index(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmem_pe_index_t`."""
 *     PE_INVALID = NVSHMEM_PE_INVALID
*/
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Pe_index, __pyx_t_2, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_10);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Pe_index, __pyx_t_10) < (0)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":364
 *     PE_ALL = NVSHMEMX_PE_ALL
 * 
 * class Team_id(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmem_team_id_t`."""
 *     TEAM_INVALID = NVSHMEM_TEAM_INVALID
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Team_id, __pyx_mstate_global->__pyx_n_u_Team_id, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmem_team_id_t); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_2 != __pyx_t_11) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_11) < 0))) __PYX_ERR(0, 364, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":366
 * class Team_id(_IntEnum):
 *     """See `nvshmem_team_id_t`."""
 *     TEAM_INVALID = NVSHMEM_TEAM_INVALID             # <<<<<<<<<<<<<<
 *     TEAM_WORLD = NVSHMEM_TEAM_WORLD
 *     TEAM_WORLD_INDEX = NVSHMEM_TEAM_WORLD_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_INVALID); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_INVALID, __pyx_t_11) < (0)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":367
 *     """See `nvshmem_team_id_t`."""
 *     TEAM_INVALID = NVSHMEM_TEAM_INVALID
 *     TEAM_WORLD = NVSHMEM_TEAM_WORLD             # <<<<<<<<<<<<<<
 *     TEAM_WORLD_INDEX = NVSHMEM_TEAM_WORLD_INDEX
 *     TEAM_SHARED = NVSHMEM_TEAM_SHARED
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_WORLD); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_WORLD, __pyx_t_11) < (0)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":368
 *     TEAM_INVALID = NVSHMEM_TEAM_INVALID
 *     TEAM_WORLD = NVSHMEM_TEAM_WORLD
 *     TEAM_WORLD_INDEX = NVSHMEM_TEAM_WORLD_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_SHARED = NVSHMEM_TEAM_SHARED
 *     TEAM_SHARED_INDEX = NVSHMEM_TEAM_SHARED_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_WORLD_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_WORLD_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":369
 *     TEAM_WORLD = NVSHMEM_TEAM_WORLD
 *     TEAM_WORLD_INDEX = NVSHMEM_TEAM_WORLD_INDEX
 *     TEAM_SHARED = NVSHMEM_TEAM_SHARED             # <<<<<<<<<<<<<<
 *     TEAM_SHARED_INDEX = NVSHMEM_TEAM_SHARED_INDEX
 *     TEAM_NODE = NVSHMEMX_TEAM_NODE
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_SHARED); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SHARED, __pyx_t_11) < (0)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":370
 *     TEAM_WORLD_INDEX = NVSHMEM_TEAM_WORLD_INDEX
 *     TEAM_SHARED = NVSHMEM_TEAM_SHARED
 *     TEAM_SHARED_INDEX = NVSHMEM_TEAM_SHARED_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_NODE = NVSHMEMX_TEAM_NODE
 *     TEAM_NODE_INDEX = NVSHMEM_TEAM_NODE_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_SHARED_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SHARED_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":371
 *     TEAM_SHARED = NVSHMEM_TEAM_SHARED
 *     TEAM_SHARED_INDEX = NVSHMEM_TEAM_SHARED_INDEX
 *     TEAM_NODE = NVSHMEMX_TEAM_NODE             # <<<<<<<<<<<<<<
 *     TEAM_NODE_INDEX = NVSHMEM_TEAM_NODE_INDEX
 *     TEAM_SAME_MYPE_NODE = NVSHMEMX_TEAM_SAME_MYPE_NODE
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEMX_TEAM_NODE); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_NODE, __pyx_t_11) < (0)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":372
 *     TEAM_SHARED_INDEX = NVSHMEM_TEAM_SHARED_INDEX
 *     TEAM_NODE = NVSHMEMX_TEAM_NODE
 *     TEAM_NODE_INDEX = NVSHMEM_TEAM_NODE_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_SAME_MYPE_NODE = NVSHMEMX_TEAM_SAME_MYPE_NODE
 *     TEAM_SAME_MYPE_NODE_INDEX = NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_NODE_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_NODE_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":373
 *     TEAM_NODE = NVSHMEMX_TEAM_NODE
 *     TEAM_NODE_INDEX = NVSHMEM_TEAM_NODE_INDEX
 *     TEAM_SAME_MYPE_NODE = NVSHMEMX_TEAM_SAME_MYPE_NODE             # <<<<<<<<<<<<<<
 *     TEAM_SAME_MYPE_NODE_INDEX = NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX
 *     TEAM_SAME_GPU = NVSHMEMI_TEAM_SAME_GPU
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEMX_TEAM_SAME_MYPE_NODE); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SAME_MYPE_NODE, __pyx_t_11) < (0)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":374
 *     TEAM_NODE_INDEX = NVSHMEM_TEAM_NODE_INDEX
 *     TEAM_SAME_MYPE_NODE = NVSHMEMX_TEAM_SAME_MYPE_NODE
 *     TEAM_SAME_MYPE_NODE_INDEX = NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_SAME_GPU = NVSHMEMI_TEAM_SAME_GPU
 *     TEAM_SAME_GPU_INDEX = NVSHMEM_TEAM_SAME_GPU_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SAME_MYPE_NODE_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":375
 *     TEAM_SAME_MYPE_NODE = NVSHMEMX_TEAM_SAME_MYPE_NODE
 *     TEAM_SAME_MYPE_NODE_INDEX = NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX
 *     TEAM_SAME_GPU = NVSHMEMI_TEAM_SAME_GPU             # <<<<<<<<<<<<<<
 *     TEAM_SAME_GPU_INDEX = NVSHMEM_TEAM_SAME_GPU_INDEX
 *     TEAM_GPU_LEADERS = NVSHMEMI_TEAM_GPU_LEADERS
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEMI_TEAM_SAME_GPU); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SAME_GPU, __pyx_t_11) < (0)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":376
 *     TEAM_SAME_MYPE_NODE_INDEX = NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX
 *     TEAM_SAME_GPU = NVSHMEMI_TEAM_SAME_GPU
 *     TEAM_SAME_GPU_INDEX = NVSHMEM_TEAM_SAME_GPU_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_GPU_LEADERS = NVSHMEMI_TEAM_GPU_LEADERS
 *     TEAM_GPU_LEADERS_INDEX = NVSHMEM_TEAM_GPU_LEADERS_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_SAME_GPU_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_SAME_GPU_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":377
 *     TEAM_SAME_GPU = NVSHMEMI_TEAM_SAME_GPU
 *     TEAM_SAME_GPU_INDEX = NVSHMEM_TEAM_SAME_GPU_INDEX
 *     TEAM_GPU_LEADERS = NVSHMEMI_TEAM_GPU_LEADERS             # <<<<<<<<<<<<<<
 *     TEAM_GPU_LEADERS_INDEX = NVSHMEM_TEAM_GPU_LEADERS_INDEX
 *     TEAM_MC_SHARED = NVSHMEM_TEAM_MC_SHARED
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEMI_TEAM_GPU_LEADERS); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_GPU_LEADERS, __pyx_t_11) < (0)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":378
 *     TEAM_SAME_GPU_INDEX = NVSHMEM_TEAM_SAME_GPU_INDEX
 *     TEAM_GPU_LEADERS = NVSHMEMI_TEAM_GPU_LEADERS
 *     TEAM_GPU_LEADERS_INDEX = NVSHMEM_TEAM_GPU_LEADERS_INDEX             # <<<<<<<<<<<<<<
 *     TEAM_MC_SHARED = NVSHMEM_TEAM_MC_SHARED
 *     TEAM_MC_SHARED_INDEX = NVSHMEM_TEAM_MC_SHARED_INDEX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_GPU_LEADERS_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_GPU_LEADERS_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":379
 *     TEAM_GPU_LEADERS = NVSHMEMI_TEAM_GPU_LEADERS
 *     TEAM_GPU_LEADERS_INDEX = NVSHMEM_TEAM_GPU_LEADERS_INDEX
 *     TEAM_MC_SHARED = NVSHMEM_TEAM_MC_SHARED             # <<<<<<<<<<<<<<
 *     TEAM_MC_SHARED_INDEX = NVSHMEM_TEAM_MC_SHARED_INDEX
 *     TEAMS_MIN = NVSHMEM_TEAMS_MIN
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_MC_SHARED); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_MC_SHARED, __pyx_t_11) < (0)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":380
 *     TEAM_GPU_LEADERS_INDEX = NVSHMEM_TEAM_GPU_LEADERS_INDEX
 *     TEAM_MC_SHARED = NVSHMEM_TEAM_MC_SHARED
 *     TEAM_MC_SHARED_INDEX = NVSHMEM_TEAM_MC_SHARED_INDEX             # <<<<<<<<<<<<<<
 *     TEAMS_MIN = NVSHMEM_TEAMS_MIN
 *     TEAM_INDEX_MAX = NVSHMEM_TEAM_INDEX_MAX
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_MC_SHARED_INDEX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_MC_SHARED_INDEX, __pyx_t_11) < (0)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":381
 *     TEAM_MC_SHARED = NVSHMEM_TEAM_MC_SHARED
 *     TEAM_MC_SHARED_INDEX = NVSHMEM_TEAM_MC_SHARED_INDEX
 *     TEAMS_MIN = NVSHMEM_TEAMS_MIN             # <<<<<<<<<<<<<<
 *     TEAM_INDEX_MAX = NVSHMEM_TEAM_INDEX_MAX
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAMS_MIN); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAMS_MIN, __pyx_t_11) < (0)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":382
 *     TEAM_MC_SHARED_INDEX = NVSHMEM_TEAM_MC_SHARED_INDEX
 *     TEAMS_MIN = NVSHMEM_TEAMS_MIN
 *     TEAM_INDEX_MAX = NVSHMEM_TEAM_INDEX_MAX             # <<<<<<<<<<<<<<
 * 
 * class Status(_IntEnum):
*/
  __pyx_t_11 = __Pyx_PyLong_From_nvshmem_team_id_t(NVSHMEM_TEAM_INDEX_MAX); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_TEAM_INDEX_MAX, __pyx_t_11) < (0)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":364
 *     PE_ALL = NVSHMEMX_PE_ALL
 * 
 * class Team_id(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmem_team_id_t`."""
 *     TEAM_INVALID = NVSHMEM_TEAM_INVALID
*/
  __pyx_t_11 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_Team_id, __pyx_t_2, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Team_id, __pyx_t_11) < (0)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":384
 *     TEAM_INDEX_MAX = NVSHMEM_TEAM_INDEX_MAX
 * 
 * class Status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_status`."""
 *     SUCCESS = NVSHMEMX_SUCCESS
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Status, __pyx_mstate_global->__pyx_n_u_Status, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_nvshmemx_status); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (__pyx_t_2 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_11, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 384, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":386
 * class Status(_IntEnum):
 *     """See `nvshmemx_status`."""
 *     SUCCESS = NVSHMEMX_SUCCESS             # <<<<<<<<<<<<<<
 *     ERROR_INVALID_VALUE = NVSHMEMX_ERROR_INVALID_VALUE
 *     ERROR_OUT_OF_MEMORY = NVSHMEMX_ERROR_OUT_OF_MEMORY
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_SUCCESS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_SUCCESS, __pyx_t_7) < (0)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":387
 *     """See `nvshmemx_status`."""
 *     SUCCESS = NVSHMEMX_SUCCESS
 *     ERROR_INVALID_VALUE = NVSHMEMX_ERROR_INVALID_VALUE             # <<<<<<<<<<<<<<
 *     ERROR_OUT_OF_MEMORY = NVSHMEMX_ERROR_OUT_OF_MEMORY
 *     ERROR_NOT_SUPPORTED = NVSHMEMX_ERROR_NOT_SUPPORTED
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_INVALID_VALUE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_INVALID_VALUE, __pyx_t_7) < (0)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":388
 *     SUCCESS = NVSHMEMX_SUCCESS
 *     ERROR_INVALID_VALUE = NVSHMEMX_ERROR_INVALID_VALUE
 *     ERROR_OUT_OF_MEMORY = NVSHMEMX_ERROR_OUT_OF_MEMORY             # <<<<<<<<<<<<<<
 *     ERROR_NOT_SUPPORTED = NVSHMEMX_ERROR_NOT_SUPPORTED
 *     ERROR_SYMMETRY = NVSHMEMX_ERROR_SYMMETRY
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_OUT_OF_MEMORY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_OUT_OF_MEMORY, __pyx_t_7) < (0)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":389
 *     ERROR_INVALID_VALUE = NVSHMEMX_ERROR_INVALID_VALUE
 *     ERROR_OUT_OF_MEMORY = NVSHMEMX_ERROR_OUT_OF_MEMORY
 *     ERROR_NOT_SUPPORTED = NVSHMEMX_ERROR_NOT_SUPPORTED             # <<<<<<<<<<<<<<
 *     ERROR_SYMMETRY = NVSHMEMX_ERROR_SYMMETRY
 *     ERROR_GPU_NOT_SELECTED = NVSHMEMX_ERROR_GPU_NOT_SELECTED
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_NOT_SUPPORTED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_NOT_SUPPORTED, __pyx_t_7) < (0)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":390
 *     ERROR_OUT_OF_MEMORY = NVSHMEMX_ERROR_OUT_OF_MEMORY
 *     ERROR_NOT_SUPPORTED = NVSHMEMX_ERROR_NOT_SUPPORTED
 *     ERROR_SYMMETRY = NVSHMEMX_ERROR_SYMMETRY             # <<<<<<<<<<<<<<
 *     ERROR_GPU_NOT_SELECTED = NVSHMEMX_ERROR_GPU_NOT_SELECTED
 *     ERROR_COLLECTIVE_LAUNCH_FAILED = NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_SYMMETRY); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_SYMMETRY, __pyx_t_7) < (0)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":391
 *     ERROR_NOT_SUPPORTED = NVSHMEMX_ERROR_NOT_SUPPORTED
 *     ERROR_SYMMETRY = NVSHMEMX_ERROR_SYMMETRY
 *     ERROR_GPU_NOT_SELECTED = NVSHMEMX_ERROR_GPU_NOT_SELECTED             # <<<<<<<<<<<<<<
 *     ERROR_COLLECTIVE_LAUNCH_FAILED = NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED
 *     ERROR_INTERNAL = NVSHMEMX_ERROR_INTERNAL
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_GPU_NOT_SELECTED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_GPU_NOT_SELECTED, __pyx_t_7) < (0)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":392
 *     ERROR_SYMMETRY = NVSHMEMX_ERROR_SYMMETRY
 *     ERROR_GPU_NOT_SELECTED = NVSHMEMX_ERROR_GPU_NOT_SELECTED
 *     ERROR_COLLECTIVE_LAUNCH_FAILED = NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED             # <<<<<<<<<<<<<<
 *     ERROR_INTERNAL = NVSHMEMX_ERROR_INTERNAL
 *     ERROR_SENTINEL = NVSHMEMX_ERROR_SENTINEL
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_COLLECTIVE_LAUNCH_FAILED, __pyx_t_7) < (0)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":393
 *     ERROR_GPU_NOT_SELECTED = NVSHMEMX_ERROR_GPU_NOT_SELECTED
 *     ERROR_COLLECTIVE_LAUNCH_FAILED = NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED
 *     ERROR_INTERNAL = NVSHMEMX_ERROR_INTERNAL             # <<<<<<<<<<<<<<
 *     ERROR_SENTINEL = NVSHMEMX_ERROR_SENTINEL
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_INTERNAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_INTERNAL, __pyx_t_7) < (0)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":394
 *     ERROR_COLLECTIVE_LAUNCH_FAILED = NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED
 *     ERROR_INTERNAL = NVSHMEMX_ERROR_INTERNAL
 *     ERROR_SENTINEL = NVSHMEMX_ERROR_SENTINEL             # <<<<<<<<<<<<<<
 * 
 * class Flags(_IntEnum):
*/
  __pyx_t_7 = __Pyx_PyLong_From_nvshmemx_status(NVSHMEMX_ERROR_SENTINEL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ERROR_SENTINEL, __pyx_t_7) < (0)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":384
 *     TEAM_INDEX_MAX = NVSHMEM_TEAM_INDEX_MAX
 * 
 * class Status(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `nvshmemx_status`."""
 *     SUCCESS = NVSHMEMX_SUCCESS
*/
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Status, __pyx_t_2, __pyx_t_11, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_7);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Status, __pyx_t_7) < (0)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":396
 *     ERROR_SENTINEL = NVSHMEMX_ERROR_SENTINEL
 * 
 * class Flags(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `flags`."""
 *     INIT_THREAD_PES = NVSHMEMX_INIT_THREAD_PES
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IntEnum_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Flags, __pyx_mstate_global->__pyx_n_u_Flags, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_kp_u_See_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_2 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 396, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":398
 * class Flags(_IntEnum):
 *     """See `flags`."""
 *     INIT_THREAD_PES = NVSHMEMX_INIT_THREAD_PES             # <<<<<<<<<<<<<<
 *     INIT_WITH_MPI_COMM = NVSHMEMX_INIT_WITH_MPI_COMM
 *     INIT_WITH_SHMEM = NVSHMEMX_INIT_WITH_SHMEM
*/
  __pyx_t_10 = __Pyx_PyLong_From_flags(NVSHMEMX_INIT_THREAD_PES); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INIT_THREAD_PES, __pyx_t_10) < (0)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":399
 *     """See `flags`."""
 *     INIT_THREAD_PES = NVSHMEMX_INIT_THREAD_PES
 *     INIT_WITH_MPI_COMM = NVSHMEMX_INIT_WITH_MPI_COMM             # <<<<<<<<<<<<<<
 *     INIT_WITH_SHMEM = NVSHMEMX_INIT_WITH_SHMEM
 *     INIT_WITH_UNIQUEID = NVSHMEMX_INIT_WITH_UNIQUEID
*/
  __pyx_t_10 = __Pyx_PyLong_From_flags(NVSHMEMX_INIT_WITH_MPI_COMM); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INIT_WITH_MPI_COMM, __pyx_t_10) < (0)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":400
 *     INIT_THREAD_PES = NVSHMEMX_INIT_THREAD_PES
 *     INIT_WITH_MPI_COMM = NVSHMEMX_INIT_WITH_MPI_COMM
 *     INIT_WITH_SHMEM = NVSHMEMX_INIT_WITH_SHMEM             # <<<<<<<<<<<<<<
 *     INIT_WITH_UNIQUEID = NVSHMEMX_INIT_WITH_UNIQUEID
 *     INIT_MAX = NVSHMEMX_INIT_MAX
*/
  __pyx_t_10 = __Pyx_PyLong_From_flags(NVSHMEMX_INIT_WITH_SHMEM); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INIT_WITH_SHMEM, __pyx_t_10) < (0)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":401
 *     INIT_WITH_MPI_COMM = NVSHMEMX_INIT_WITH_MPI_COMM
 *     INIT_WITH_SHMEM = NVSHMEMX_INIT_WITH_SHMEM
 *     INIT_WITH_UNIQUEID = NVSHMEMX_INIT_WITH_UNIQUEID             # <<<<<<<<<<<<<<
 *     INIT_MAX = NVSHMEMX_INIT_MAX
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_flags(NVSHMEMX_INIT_WITH_UNIQUEID); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INIT_WITH_UNIQUEID, __pyx_t_10) < (0)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":402
 *     INIT_WITH_SHMEM = NVSHMEMX_INIT_WITH_SHMEM
 *     INIT_WITH_UNIQUEID = NVSHMEMX_INIT_WITH_UNIQUEID
 *     INIT_MAX = NVSHMEMX_INIT_MAX             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_flags(NVSHMEMX_INIT_MAX); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_INIT_MAX, __pyx_t_10) < (0)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":396
 *     ERROR_SENTINEL = NVSHMEMX_ERROR_SENTINEL
 * 
 * class Flags(_IntEnum):             # <<<<<<<<<<<<<<
 *     """See `flags`."""
 *     INIT_THREAD_PES = NVSHMEMX_INIT_THREAD_PES
*/
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Flags, __pyx_t_2, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_10);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Flags, __pyx_t_10) < (0)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":409
 * ###############################################################################
 * 
 * class NVSHMEMError(Exception):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, status):
*/
  __pyx_t_2 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_11, __pyx_mstate_global->__pyx_n_u_NVSHMEMError, __pyx_mstate_global->__pyx_n_u_NVSHMEMError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, (PyObject *) NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_11 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 409, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":411
 * class NVSHMEMError(Exception):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         cdef str err = f"Status code {status}"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_NVSHMEMError___init, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_2) < (0)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":416
 *         super(NVSHMEMError, self).__init__(err)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_12NVSHMEMError_3__reduce__, 0, __pyx_mstate_global->__pyx_n_u_NVSHMEMError___reduce, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_reduce, __pyx_t_2) < (0)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":409
 * ###############################################################################
 * 
 * class NVSHMEMError(Exception):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, status):
*/
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_NVSHMEMError, __pyx_t_11, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NVSHMEMError, __pyx_t_2) < (0)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":420
 * 
 * 
 * @cython.profile(False)             # <<<<<<<<<<<<<<
 * cpdef inline check_status(int status):
 *     if status != 0:
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_1check_status, 0, __pyx_mstate_global->__pyx_n_u_check_status, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_check_status, __pyx_t_11) < (0)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":430
 * ###############################################################################
 * 
 * cpdef barrier(int32_t team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_barrier(<nvshmem_team_t>team)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_3barrier, 0, __pyx_mstate_global->__pyx_n_u_barrier, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_barrier, __pyx_t_11) < (0)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":436
 * 
 * 
 * cpdef void barrier_all() except*:             # <<<<<<<<<<<<<<
 *     nvshmem_barrier_all()
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_5barrier_all, 0, __pyx_mstate_global->__pyx_n_u_barrier_all, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_barrier_all, __pyx_t_11) < (0)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":440
 * 
 * 
 * cpdef int init_status() except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_init_status()
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_7init_status, 0, __pyx_mstate_global->__pyx_n_u_init_status, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_init_status, __pyx_t_11) < (0)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":444
 * 
 * 
 * cpdef int my_pe() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_my_pe()
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_9my_pe, 0, __pyx_mstate_global->__pyx_n_u_my_pe, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_my_pe, __pyx_t_11) < (0)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":448
 * 
 * 
 * cpdef int n_pes() except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_n_pes()
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_11n_pes, 0, __pyx_mstate_global->__pyx_n_u_n_pes, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_n_pes, __pyx_t_11) < (0)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":452
 * 
 * 
 * cpdef void info_get_version(intptr_t major, intptr_t minor) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_info_get_version(<int*>major, <int*>minor)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_13info_get_version, 0, __pyx_mstate_global->__pyx_n_u_info_get_version, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_info_get_version, __pyx_t_11) < (0)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":456
 * 
 * 
 * cpdef void vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_vendor_get_version_info(<int*>major, <int*>minor, <int*>patch)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_15vendor_get_version_info, 0, __pyx_mstate_global->__pyx_n_u_vendor_get_version_info, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_vendor_get_version_info, __pyx_t_11) < (0)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":460
 * 
 * 
 * cpdef intptr_t malloc(size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_malloc(size)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_17malloc, 0, __pyx_mstate_global->__pyx_n_u_malloc, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_malloc, __pyx_t_11) < (0)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":464
 * 
 * 
 * cpdef intptr_t calloc(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_calloc(count, size)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_19calloc, 0, __pyx_mstate_global->__pyx_n_u_calloc, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_calloc, __pyx_t_11) < (0)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":468
 * 
 * 
 * cpdef intptr_t align(size_t count, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_align(count, size)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_21align, 0, __pyx_mstate_global->__pyx_n_u_align, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_align, __pyx_t_11) < (0)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":472
 * 
 * 
 * cpdef void free(intptr_t ptr) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_free(<void*>ptr)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_23free, 0, __pyx_mstate_global->__pyx_n_u_free, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_free, __pyx_t_11) < (0)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":476
 * 
 * 
 * cpdef intptr_t ptr(intptr_t dest, int pe) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmem_ptr(<const void*>dest, pe)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_25ptr, 0, __pyx_mstate_global->__pyx_n_u_ptr, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ptr, __pyx_t_11) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":480
 * 
 * 
 * cpdef intptr_t mc_ptr(int32_t team, intptr_t ptr) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_mc_ptr(<nvshmem_team_t>team, <const void*>ptr)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_27mc_ptr, 0, __pyx_mstate_global->__pyx_n_u_mc_ptr, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mc_ptr, __pyx_t_11) < (0)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":484
 * 
 * 
 * cpdef int team_my_pe(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_my_pe(<nvshmem_team_t>team)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_29team_my_pe, 0, __pyx_mstate_global->__pyx_n_u_team_my_pe, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_my_pe, __pyx_t_11) < (0)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":488
 * 
 * 
 * cpdef int team_n_pes(int32_t team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_n_pes(<nvshmem_team_t>team)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_31team_n_pes, 0, __pyx_mstate_global->__pyx_n_u_team_n_pes, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_n_pes, __pyx_t_11) < (0)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":492
 * 
 * 
 * cpdef void team_get_config(int32_t team, intptr_t config) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_get_config(<nvshmem_team_t>team, <nvshmem_team_config_t*>config)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_33team_get_config, 0, __pyx_mstate_global->__pyx_n_u_team_get_config, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_get_config, __pyx_t_11) < (0)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":496
 * 
 * 
 * cpdef int team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmem_team_translate_pe(<nvshmem_team_t>src_team, src_pe, <nvshmem_team_t>dest_team)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_35team_translate_pe, 0, __pyx_mstate_global->__pyx_n_u_team_translate_pe, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_translate_pe, __pyx_t_11) < (0)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":500
 * 
 * 
 * cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_strided(<nvshmem_team_t>parent_team, pe_start, pe_stride, pe_size, <const nvshmem_team_config_t*>config, config_mask, <nvshmem_team_t*>new_team)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_37team_split_strided, 0, __pyx_mstate_global->__pyx_n_u_team_split_strided, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_split_strided, __pyx_t_11) < (0)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":506
 * 
 * 
 * cpdef team_get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_get_uniqueid(<nvshmemx_team_uniqueid_t*>uniqueid)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_39team_get_uniqueid, 0, __pyx_mstate_global->__pyx_n_u_team_get_uniqueid, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_get_uniqueid, __pyx_t_11) < (0)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":512
 * 
 * 
 * cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_team_init(<nvshmem_team_t*>team, <nvshmem_team_config_t*>config, config_mask, npes, pe_idx_in_team)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_41team_init, 0, __pyx_mstate_global->__pyx_n_u_team_init, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_init, __pyx_t_11) < (0)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":518
 * 
 * 
 * cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmem_team_split_2d(<nvshmem_team_t>parent_team, xrange, <const nvshmem_team_config_t*>xaxis_config, xaxis_mask, <nvshmem_team_t*>xaxis_team, <const nvshmem_team_config_t*>yaxis_config, yaxis_mask, <nvshmem_team_t*>yaxis_team)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_43team_split_2d, 0, __pyx_mstate_global->__pyx_n_u_team_split_2d, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_split_2d, __pyx_t_11) < (0)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":524
 * 
 * 
 * cpdef void team_destroy(int32_t team) except*:             # <<<<<<<<<<<<<<
 *     nvshmem_team_destroy(<nvshmem_team_t>team)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_45team_destroy, 0, __pyx_mstate_global->__pyx_n_u_team_destroy, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_destroy, __pyx_t_11) < (0)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":528
 * 
 * 
 * cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_alltoall_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_47bfloat16_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":534
 * 
 * 
 * cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_alltoall_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_49half_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":540
 * 
 * 
 * cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_alltoall_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_51float_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":546
 * 
 * 
 * cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_alltoall_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_53double_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":552
 * 
 * 
 * cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_alltoall_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_55char_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":558
 * 
 * 
 * cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_alltoall_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_57short_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":564
 * 
 * 
 * cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_alltoall_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_59schar_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":570
 * 
 * 
 * cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_alltoall_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_61int_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":576
 * 
 * 
 * cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_alltoall_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_63long_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":582
 * 
 * 
 * cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_alltoall_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_65longlong_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":588
 * 
 * 
 * cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_alltoall_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_67int8_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":594
 * 
 * 
 * cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_alltoall_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_69int16_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":600
 * 
 * 
 * cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_alltoall_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_71int32_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":606
 * 
 * 
 * cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_alltoall_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_73int64_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":612
 * 
 * 
 * cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_alltoall_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_75uint8_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":618
 * 
 * 
 * cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_alltoall_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_77uint16_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":624
 * 
 * 
 * cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_alltoall_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_79uint32_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":630
 * 
 * 
 * cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_alltoall_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_81uint64_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":636
 * 
 * 
 * cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_alltoall_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_83size_alltoall_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_alltoall_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[57])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_alltoall_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":642
 * 
 * 
 * cpdef barrier_on_stream(int32_t team, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_barrier_on_stream(<nvshmem_team_t>team, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_85barrier_on_stream, 0, __pyx_mstate_global->__pyx_n_u_barrier_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[58])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_barrier_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":648
 * 
 * 
 * cpdef void barrier_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_barrier_all_on_stream(<Stream>stream)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_87barrier_all_on_stream, 0, __pyx_mstate_global->__pyx_n_u_barrier_all_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[59])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_barrier_all_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":652
 * 
 * 
 * cpdef int team_sync_on_stream(int32_t team, intptr_t stream) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_team_sync_on_stream(<nvshmem_team_t>team, <Stream>stream)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_89team_sync_on_stream, 0, __pyx_mstate_global->__pyx_n_u_team_sync_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[60])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_team_sync_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":656
 * 
 * 
 * cpdef void sync_all_on_stream(intptr_t stream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_sync_all_on_stream(<Stream>stream)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_91sync_all_on_stream, 0, __pyx_mstate_global->__pyx_n_u_sync_all_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[61])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sync_all_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":660
 * 
 * 
 * cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_broadcast_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_93bfloat16_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[62])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":666
 * 
 * 
 * cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_broadcast_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_95half_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[63])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":672
 * 
 * 
 * cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_broadcast_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_97float_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[64])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":678
 * 
 * 
 * cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_broadcast_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_99double_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[65])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":684
 * 
 * 
 * cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_broadcast_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_101char_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[66])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":690
 * 
 * 
 * cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_broadcast_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_103short_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[67])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":696
 * 
 * 
 * cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_broadcast_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_105schar_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[68])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":702
 * 
 * 
 * cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_broadcast_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_107int_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[69])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":708
 * 
 * 
 * cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_broadcast_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_109long_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[70])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":714
 * 
 * 
 * cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_broadcast_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_111longlong_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[71])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":720
 * 
 * 
 * cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_broadcast_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_113int8_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[72])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":726
 * 
 * 
 * cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_broadcast_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_115int16_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[73])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":732
 * 
 * 
 * cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_broadcast_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_117int32_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[74])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":738
 * 
 * 
 * cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_broadcast_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_119int64_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[75])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":744
 * 
 * 
 * cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_broadcast_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_121uint8_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[76])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":750
 * 
 * 
 * cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_broadcast_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_123uint16_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[77])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":756
 * 
 * 
 * cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_broadcast_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_125uint32_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[78])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":762
 * 
 * 
 * cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_broadcast_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_127uint64_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[79])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":768
 * 
 * 
 * cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_broadcast_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, pe_root, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_129size_broadcast_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_broadcast_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[80])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_broadcast_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":774
 * 
 * 
 * cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_fcollect_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_131bfloat16_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[81])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":780
 * 
 * 
 * cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_fcollect_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_133half_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[82])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":786
 * 
 * 
 * cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_fcollect_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_135float_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[83])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":792
 * 
 * 
 * cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_fcollect_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_137double_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[84])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":798
 * 
 * 
 * cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_fcollect_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_139char_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[85])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":804
 * 
 * 
 * cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_fcollect_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_141short_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[86])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":810
 * 
 * 
 * cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_fcollect_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_143schar_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[87])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":816
 * 
 * 
 * cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_fcollect_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_145int_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[88])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":822
 * 
 * 
 * cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_fcollect_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_147long_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[89])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":828
 * 
 * 
 * cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_fcollect_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_149longlong_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[90])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":834
 * 
 * 
 * cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_fcollect_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_151int8_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[91])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":840
 * 
 * 
 * cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_fcollect_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_153int16_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[92])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":846
 * 
 * 
 * cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_fcollect_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_155int32_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[93])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":852
 * 
 * 
 * cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_fcollect_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_157int64_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[94])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":858
 * 
 * 
 * cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_fcollect_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_159uint8_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[95])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":864
 * 
 * 
 * cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_fcollect_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_161uint16_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[96])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":870
 * 
 * 
 * cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_fcollect_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_163uint32_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[97])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":876
 * 
 * 
 * cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_fcollect_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_165uint64_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[98])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":882
 * 
 * 
 * cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_fcollect_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nelem, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_167size_fcollect_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_fcollect_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[99])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_fcollect_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":888
 * 
 * 
 * cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_169int8_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[100])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":894
 * 
 * 
 * cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_171int16_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[101])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":900
 * 
 * 
 * cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_173int32_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[102])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":906
 * 
 * 
 * cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_175int64_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[103])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":912
 * 
 * 
 * cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_177uint8_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[104])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":918
 * 
 * 
 * cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_179uint16_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[105])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":924
 * 
 * 
 * cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_181uint32_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[106])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":930
 * 
 * 
 * cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_183uint64_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[107])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":936
 * 
 * 
 * cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_185size_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[108])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":942
 * 
 * 
 * cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_187char_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[109])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":948
 * 
 * 
 * cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_189schar_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[110])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":954
 * 
 * 
 * cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_191short_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[111])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":960
 * 
 * 
 * cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_193int_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[112])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":966
 * 
 * 
 * cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_195long_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[113])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":972
 * 
 * 
 * cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_197longlong_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[114])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":978
 * 
 * 
 * cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_199bfloat16_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[115])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":984
 * 
 * 
 * cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_201half_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[116])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":990
 * 
 * 
 * cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_203float_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[117])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":996
 * 
 * 
 * cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_205double_max_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_max_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[118])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_max_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1002
 * 
 * 
 * cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_207int8_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[119])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1008
 * 
 * 
 * cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_209int16_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[120])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1014
 * 
 * 
 * cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_211int32_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[121])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1020
 * 
 * 
 * cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_213int64_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[122])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1026
 * 
 * 
 * cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_215uint8_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[123])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1032
 * 
 * 
 * cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_217uint16_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[124])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1038
 * 
 * 
 * cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_219uint32_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[125])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1044
 * 
 * 
 * cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_221uint64_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[126])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1050
 * 
 * 
 * cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_223size_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[127])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1056
 * 
 * 
 * cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_225char_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[128])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1062
 * 
 * 
 * cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_227schar_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[129])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1068
 * 
 * 
 * cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_229short_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[130])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1074
 * 
 * 
 * cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_231int_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[131])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1080
 * 
 * 
 * cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_233long_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[132])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1086
 * 
 * 
 * cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_235longlong_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[133])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1092
 * 
 * 
 * cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_237bfloat16_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[134])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1098
 * 
 * 
 * cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_239half_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[135])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1104
 * 
 * 
 * cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_241float_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[136])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1110
 * 
 * 
 * cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_243double_min_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_min_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[137])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_min_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1116
 * 
 * 
 * cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reduce_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_245int8_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[138])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1122
 * 
 * 
 * cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reduce_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_247int16_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[139])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1128
 * 
 * 
 * cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reduce_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_249int32_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[140])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1134
 * 
 * 
 * cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reduce_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_251int64_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[141])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1140
 * 
 * 
 * cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reduce_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_253uint8_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[142])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1146
 * 
 * 
 * cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reduce_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_255uint16_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[143])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1152
 * 
 * 
 * cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reduce_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_257uint32_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[144])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1158
 * 
 * 
 * cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reduce_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_259uint64_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[145])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1164
 * 
 * 
 * cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reduce_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_261size_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[146])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1170
 * 
 * 
 * cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reduce_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_263char_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[147])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1176
 * 
 * 
 * cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reduce_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_265schar_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[148])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1182
 * 
 * 
 * cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reduce_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_267short_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[149])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1188
 * 
 * 
 * cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reduce_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_269int_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[150])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1194
 * 
 * 
 * cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reduce_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_271long_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[151])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1200
 * 
 * 
 * cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reduce_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_273longlong_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[152])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1206
 * 
 * 
 * cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reduce_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_275bfloat16_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[153])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1212
 * 
 * 
 * cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reduce_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_277half_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[154])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1218
 * 
 * 
 * cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reduce_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_279float_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[155])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1224
 * 
 * 
 * cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reduce_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_281double_sum_reduce_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_sum_reduce_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[156])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_sum_reduce_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1230
 * 
 * 
 * cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_max_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_283int8_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[157])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1236
 * 
 * 
 * cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_max_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_285int16_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[158])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1242
 * 
 * 
 * cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_max_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_287int32_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[159])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1248
 * 
 * 
 * cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_max_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_289int64_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[160])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1254
 * 
 * 
 * cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_291uint8_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[161])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1260
 * 
 * 
 * cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_293uint16_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_max_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[162])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_max_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1266
 * 
 * 
 * cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_295uint32_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_max_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[163])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_max_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1272
 * 
 * 
 * cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_max_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_297uint64_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_max_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[164])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_max_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1278
 * 
 * 
 * cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_max_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_299size_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[165])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1284
 * 
 * 
 * cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_max_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_301char_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[166])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1290
 * 
 * 
 * cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_max_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_303schar_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[167])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1296
 * 
 * 
 * cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_max_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_305short_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[168])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1302
 * 
 * 
 * cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_max_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_307int_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[169])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1308
 * 
 * 
 * cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_max_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_309long_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[170])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1314
 * 
 * 
 * cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_max_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_311longlong_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_max_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[171])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_max_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1320
 * 
 * 
 * cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_max_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_313bfloat16_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_max_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[172])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_max_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1326
 * 
 * 
 * cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_max_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_315half_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_max_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[173])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_max_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1332
 * 
 * 
 * cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_max_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_317float_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_max_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[174])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_max_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1338
 * 
 * 
 * cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_max_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_319double_max_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_max_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[175])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_max_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1344
 * 
 * 
 * cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_min_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_321int8_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[176])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1350
 * 
 * 
 * cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_min_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_323int16_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[177])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1356
 * 
 * 
 * cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_min_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_325int32_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[178])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1362
 * 
 * 
 * cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_min_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_327int64_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[179])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1368
 * 
 * 
 * cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_329uint8_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[180])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1374
 * 
 * 
 * cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_331uint16_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_min_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[181])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_min_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1380
 * 
 * 
 * cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_333uint32_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_min_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[182])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_min_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1386
 * 
 * 
 * cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_min_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_335uint64_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_min_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[183])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_min_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1392
 * 
 * 
 * cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_min_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_337size_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[184])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1398
 * 
 * 
 * cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_min_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_339char_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[185])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1404
 * 
 * 
 * cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_min_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_341schar_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[186])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1410
 * 
 * 
 * cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_min_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_343short_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[187])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1416
 * 
 * 
 * cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_min_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_345int_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[188])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1422
 * 
 * 
 * cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_min_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_347long_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[189])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1428
 * 
 * 
 * cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_min_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_349longlong_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_min_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[190])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_min_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1434
 * 
 * 
 * cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_min_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_351bfloat16_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_min_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[191])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_min_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1440
 * 
 * 
 * cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_min_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_353half_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_min_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[192])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_min_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1446
 * 
 * 
 * cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_min_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_355float_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_min_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[193])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_min_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1452
 * 
 * 
 * cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_min_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_357double_min_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_min_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[194])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_min_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1458
 * 
 * 
 * cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int8_t*>dest, <const int8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_359int8_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int8_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[195])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int8_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1464
 * 
 * 
 * cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int16_t*>dest, <const int16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_361int16_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int16_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[196])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int16_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1470
 * 
 * 
 * cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int32_t*>dest, <const int32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_363int32_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int32_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[197])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int32_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1476
 * 
 * 
 * cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int64_t*>dest, <const int64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_365int64_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int64_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[198])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int64_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1482
 * 
 * 
 * cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint8_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint8_t*>dest, <const uint8_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_367uint8_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint8_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[199])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint8_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1488
 * 
 * 
 * cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint16_t*>dest, <const uint16_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_369uint16_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint16_sum_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[200])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint16_sum_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1494
 * 
 * 
 * cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint32_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint32_t*>dest, <const uint32_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_371uint32_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint32_sum_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[201])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint32_sum_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1500
 * 
 * 
 * cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_uint64_sum_reducescatter_on_stream(<nvshmem_team_t>team, <uint64_t*>dest, <const uint64_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_373uint64_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_uint64_sum_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[202])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_uint64_sum_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1506
 * 
 * 
 * cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_size_sum_reducescatter_on_stream(<nvshmem_team_t>team, <size_t*>dest, <const size_t*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_375size_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_size_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[203])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_size_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1512
 * 
 * 
 * cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_char_sum_reducescatter_on_stream(<nvshmem_team_t>team, <char*>dest, <const char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_377char_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_char_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[204])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_char_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1518
 * 
 * 
 * cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_schar_sum_reducescatter_on_stream(<nvshmem_team_t>team, <signed char*>dest, <const signed char*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_379schar_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_schar_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[205])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_schar_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1524
 * 
 * 
 * cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_short_sum_reducescatter_on_stream(<nvshmem_team_t>team, <short*>dest, <const short*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_381short_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_short_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[206])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_short_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1530
 * 
 * 
 * cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_int_sum_reducescatter_on_stream(<nvshmem_team_t>team, <int*>dest, <const int*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_383int_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_int_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[207])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_int_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1536
 * 
 * 
 * cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_long_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long*>dest, <const long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_385long_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_long_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[208])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_long_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1542
 * 
 * 
 * cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_longlong_sum_reducescatter_on_stream(<nvshmem_team_t>team, <long long*>dest, <const long long*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_387longlong_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_longlong_sum_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[209])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_longlong_sum_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1548
 * 
 * 
 * cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_bfloat16_sum_reducescatter_on_stream(<nvshmem_team_t>team, <__nv_bfloat16*>dest, <const __nv_bfloat16*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_389bfloat16_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_bfloat16_sum_reducescatter_on_st, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[210])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_bfloat16_sum_reducescatter_on_st, __pyx_t_11) < (0)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1554
 * 
 * 
 * cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_half_sum_reducescatter_on_stream(<nvshmem_team_t>team, <half*>dest, <const half*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_391half_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_half_sum_reducescatter_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[211])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_half_sum_reducescatter_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1560
 * 
 * 
 * cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_float_sum_reducescatter_on_stream(<nvshmem_team_t>team, <float*>dest, <const float*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_393float_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_float_sum_reducescatter_on_strea, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[212])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_float_sum_reducescatter_on_strea, __pyx_t_11) < (0)) __PYX_ERR(0, 1560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1566
 * 
 * 
 * cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_double_sum_reducescatter_on_stream(<nvshmem_team_t>team, <double*>dest, <const double*>src, nreduce, <Stream>stream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_395double_sum_reducescatter_on_stream, 0, __pyx_mstate_global->__pyx_n_u_double_sum_reducescatter_on_stre, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[213])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_double_sum_reducescatter_on_stre, __pyx_t_11) < (0)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1572
 * 
 * 
 * cpdef hostlib_init_attr(unsigned int flags, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_hostlib_init_attr(flags, <nvshmemx_init_attr_t*>attr)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_397hostlib_init_attr, 0, __pyx_mstate_global->__pyx_n_u_hostlib_init_attr, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[214])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_hostlib_init_attr, __pyx_t_11) < (0)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1578
 * 
 * 
 * cpdef void hostlib_finalize() except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_hostlib_finalize()
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_399hostlib_finalize, 0, __pyx_mstate_global->__pyx_n_u_hostlib_finalize, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[215])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_hostlib_finalize, __pyx_t_11) < (0)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1582
 * 
 * 
 * cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_uniqueid_args(<const int>myrank, <const int>nranks, <const nvshmemx_uniqueid_t*>uniqueid, <nvshmemx_init_attr_t*>attr)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_401set_attr_uniqueid_args, 0, __pyx_mstate_global->__pyx_n_u_set_attr_uniqueid_args, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[216])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_set_attr_uniqueid_args, __pyx_t_11) < (0)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1588
 * 
 * 
 * cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_set_attr_mpi_comm_args(<void*>mpi_comm, <nvshmemx_init_attr_t*>nvshmem_attr)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_403set_attr_mpi_comm_args, 0, __pyx_mstate_global->__pyx_n_u_set_attr_mpi_comm_args, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[217])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_set_attr_mpi_comm_args, __pyx_t_11) < (0)) __PYX_ERR(0, 1588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1594
 * 
 * 
 * cpdef get_uniqueid(intptr_t uniqueid):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = nvshmemx_get_uniqueid(<nvshmemx_uniqueid_t*>uniqueid)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_405get_uniqueid, 0, __pyx_mstate_global->__pyx_n_u_get_uniqueid, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[218])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_uniqueid, __pyx_t_11) < (0)) __PYX_ERR(0, 1594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1600
 * 
 * 
 * cpdef int cumodule_init(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_init(<void*>module)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_407cumodule_init, 0, __pyx_mstate_global->__pyx_n_u_cumodule_init, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[219])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cumodule_init, __pyx_t_11) < (0)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1604
 * 
 * 
 * cpdef int cumodule_finalize(intptr_t module) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_cumodule_finalize(<void*>module)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_409cumodule_finalize, 0, __pyx_mstate_global->__pyx_n_u_cumodule_finalize, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[220])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cumodule_finalize, __pyx_t_11) < (0)) __PYX_ERR(0, 1604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1608
 * 
 * 
 * cpdef intptr_t buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) except? 0:             # <<<<<<<<<<<<<<
 *     return <intptr_t>nvshmemx_buffer_register_symmetric(<void*>buf_ptr, size, flags)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_411buffer_register_symmetric, 0, __pyx_mstate_global->__pyx_n_u_buffer_register_symmetric, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[221])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_buffer_register_symmetric, __pyx_t_11) < (0)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1612
 * 
 * 
 * cpdef int buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) except? 0:             # <<<<<<<<<<<<<<
 *     return nvshmemx_buffer_unregister_symmetric(<void*>mmap_ptr, size)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_413buffer_unregister_symmetric, 0, __pyx_mstate_global->__pyx_n_u_buffer_unregister_symmetric, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[222])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_buffer_unregister_symmetric, __pyx_t_11) < (0)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1616
 * 
 * 
 * cpdef int culibrary_init(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_init(<void*>library)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_415culibrary_init, 0, __pyx_mstate_global->__pyx_n_u_culibrary_init, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[223])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_culibrary_init, __pyx_t_11) < (0)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1620
 * 
 * 
 * cpdef int culibrary_finalize(intptr_t library) except? -1:             # <<<<<<<<<<<<<<
 *     return nvshmemx_culibrary_finalize(<void*>library)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_417culibrary_finalize, 0, __pyx_mstate_global->__pyx_n_u_culibrary_finalize, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[224])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_culibrary_finalize, __pyx_t_11) < (0)) __PYX_ERR(0, 1620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1624
 * 
 * 
 * cpdef void putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_419putmem_on_stream, 0, __pyx_mstate_global->__pyx_n_u_putmem_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[225])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_putmem_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1628
 * 
 * 
 * cpdef void putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_putmem_signal_on_stream(<void*>dest, <const void*>source, bytes, <uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_421putmem_signal_on_stream, 0, __pyx_mstate_global->__pyx_n_u_putmem_signal_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[226])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_putmem_signal_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1632
 * 
 * 
 * cpdef void getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_getmem_on_stream(<void*>dest, <const void*>source, bytes, pe, <Stream>cstrm)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_423getmem_on_stream, 0, __pyx_mstate_global->__pyx_n_u_getmem_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[227])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_getmem_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1636
 * 
 * 
 * cpdef void quiet_on_stream(intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_quiet_on_stream(<Stream>cstrm)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_425quiet_on_stream, 0, __pyx_mstate_global->__pyx_n_u_quiet_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[228])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_quiet_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1640
 * 
 * 
 * cpdef void signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_op_on_stream(<uint64_t*>sig_addr, signal, sig_op, pe, <Stream>cstrm)
 * 
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_427signal_op_on_stream, 0, __pyx_mstate_global->__pyx_n_u_signal_op_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[229])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_signal_op_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1644
 * 
 * 
 * cpdef void signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) except*:             # <<<<<<<<<<<<<<
 *     nvshmemx_signal_wait_until_on_stream(<uint64_t*>sig_addr, cmp, cmp_value, <Stream>cstream)
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_429signal_wait_until_on_stream, 0, __pyx_mstate_global->__pyx_n_u_signal_wait_until_on_stream, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[230])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_signal_wait_until_on_stream, __pyx_t_11) < (0)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_431__pyx_unpickle_uniqueid, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_uniqueid, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[231])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_uniqueid, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_433__pyx_unpickle_UniqueId, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_UniqueId, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[232])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_UniqueId, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_team_uniqueid(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_435__pyx_unpickle_team_uniqueid, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_team_uniqueid, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[233])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_team_uniqueid, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_437__pyx_unpickle_TeamUniqueId, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamUniqueId, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[234])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamUniqueId, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_InitAttr(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xa75e18a, 0xc05ad22, 0x7125153, b'_data')
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_439__pyx_unpickle_InitAttr, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InitAttr, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[235])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_InitAttr, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_11 = __Pyx_CyFunction_New(&__pyx_mdef_7nvshmem_8bindings_7nvshmem_441__pyx_unpickle_TeamConfig, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamConfig, NULL, __pyx_mstate_global->__pyx_n_u_nvshmem_bindings_nvshmem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[236])); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_11);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_TeamConfig, __pyx_t_11) < (0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "nvshmem/bindings/nvshmem.pyx":1
 * # Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.             # <<<<<<<<<<<<<<
 * # SPDX-License-Identifier: Apache-2.0
 * #
*/
  __pyx_t_11 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_11) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init nvshmem.bindings.nvshmem", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init nvshmem.bindings.nvshmem");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_id); if (!__pyx_builtin_id) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 414, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "nvshmem/bindings/nvshmem.pyx":29
 * uniqueid_dtype = _numpy.dtype([
 *     ("version", _numpy.int32, ),
 *     ("internal", _numpy.int8, (124,)),             # <<<<<<<<<<<<<<
 *     ], align=True)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_124); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "nvshmem/bindings/nvshmem.pyx":128
 *         return obj
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_ptr(intptr_t ptr, size_t size=1, bint readonly=False):
 *         """Create an uniqueid instance wrapping the given pointer.
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_int_1, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "nvshmem/bindings/nvshmem.pyx":245
 *     ("num_contexts", _numpy.int32, ),
 *     ("uniqueid", _numpy.uint64, ),
 *     ("padding", _numpy.int8, (48,)),  # TEAM_CONFIG_V2_PADDING = 48             # <<<<<<<<<<<<<<
 *     ], align=True)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_48); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<3; ++i) {
      #if PY_VERSION_HEX >= 0x030F0000
      PyUnstable_SetImmortal(table[i]);
      #elif CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      if ((PY_SSIZE_T_MAX <= _Py_IMMORTAL_REFCNT_LOCAL)) break;
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], ((Py_ssize_t)_Py_IMMORTAL_REFCNT_LOCAL + 1));
      }
      #else
      if ((PY_SSIZE_T_MAX < _Py_IMMORTAL_INITIAL_REFCNT)) break;
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 14; } index[] = {{1},{4},{179},{12},{25},{24},{26},{29},{30},{33},{27},{29},{22},{32},{28},{12},{1},{1},{1},{1},{8},{37},{21},{42},{7},{6},{2},{22},{102},{9},{9},{24},{28},{11},{24},{14},{20},{15},{16},{20},{6},{6},{6},{6},{6},{6},{12},{8},{30},{22},{14},{19},{19},{19},{14},{14},{5},{8},{15},{18},{15},{18},{8},{26},{28},{11},{7},{8},{3},{12},{21},{23},{6},{6},{10},{27},{23},{28},{32},{28},{8},{12},{20},{6},{6},{10},{7},{15},{10},{10},{18},{12},{16},{15},{14},{22},{21},{18},{22},{7},{9},{11},{6},{9},{16},{22},{14},{12},{14},{20},{9},{15},{13},{19},{19},{25},{11},{17},{10},{16},{15},{15},{17},{13},{20},{11},{10},{9},{10},{28},{30},{12},{30},{32},{7},{14},{10},{8},{26},{28},{5},{3},{12},{4},{18},{4},{7},{11},{21},{17},{27},{28},{27},{29},{36},{29},{36},{29},{36},{3},{7},{6},{25},{27},{5},{6},{23},{24},{23},{25},{32},{25},{32},{25},{32},{12},{17},{18},{3},{9},{6},{11},{5},{7},{5},{6},{18},{14},{17},{13},{4},{5},{4},{9},{8},{5},{7},{25},{26},{25},{27},{34},{27},{34},{27},{34},{5},{5},{4},{3},{4},{5},{24},{25},{24},{26},{33},{26},{33},{26},{33},{4},{9},{8},{8},{12},{16},{12},{23},{24},{23},{25},{32},{25},{32},{25},{32},{16},{17},{2},{16},{8},{15},{11},{24},{25},{24},{26},{33},{26},{33},{26},{33},{5},{24},{25},{24},{26},{33},{26},{33},{26},{33},{24},{25},{24},{26},{33},{26},{33},{26},{33},{4},{23},{24},{23},{25},{32},{25},{32},{25},{32},{22},{23},{22},{24},{31},{24},{31},{24},{31},{8},{4},{13},{5},{8},{7},{23},{24},{23},{25},{32},{25},{32},{25},{32},{27},{28},{27},{29},{36},{29},{36},{29},{36},{8},{5},{6},{6},{13},{5},{8},{6},{10},{8},{15},{5},{6},{5},{8},{7},{4},{5},{7},{8},{4},{6},{7},{12},{6},{5},{12},{24},{3},{7},{11},{5},{2},{14},{7},{7},{8},{9},{3},{11},{3},{16},{23},{12},{14},{12},{11},{10},{23},{25},{27},{23},{28},{23},{12},{15},{8},{8},{10},{17},{13},{24},{25},{24},{26},{33},{26},{33},{26},{33},{4},{22},{22},{12},{10},{12},{19},{24},{25},{24},{26},{33},{26},{33},{26},{33},{8},{6},{6},{19},{27},{4},{23},{24},{23},{25},{32},{25},{32},{25},{32},{6},{3},{6},{8},{5},{12},{6},{6},{5},{18},{4},{17},{12},{15},{17},{9},{10},{10},{13},{18},{19},{17},{13},{31},{33},{8},{25},{26},{25},{27},{34},{27},{34},{27},{34},{25},{26},{25},{27},{34},{27},{34},{27},{34},{6},{25},{26},{25},{27},{34},{27},{34},{27},{34},{24},{25},{24},{26},{33},{26},{33},{26},{33},{8},{26},{28},{14},{18},{17},{6},{12},{5},{6},{23},{7},{4},{12},{10},{10},{6},{12},{10},{10},{7653},{11654},{44},{30},{48},{24},{11},{27},{16},{12},{45},{47},{44},{40},{29},{17},{25},{54},{58},{58},{68},{17},{56},{45},{56},{45},{53},{44},{20},{45},{56},{45},{45},{56},{53},{60},{45},{56},{53},{60},{57},{17},{53},{53},{53},{58},{11},{65},{60},{17},{58},{57},{24},{11},{18},{17},{18},{46},{15},{19},{41},{11},{23},{15},{112},{103},{33},{54},{58},{18},{22},{20},{22},{54},{54},{19},{7},{24},{97},{97},{97},{97},{27},{11},{12},{89},{58},{66},{25},{11},{11},{56},{55},{55},{55},{7},{13},{11},{20}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (6068 bytes) */
const char* const cstring = "(\265/\375`\017~U\275\000J\227\350\037G\340\222\352\0010\314q\232\261\343Z\254\305Z\254\305l\236\347y\236\347i\306B\304\350\277/S\322\275\341\310\376?\334\244\3737\355\212\355\377sX\277\007\013/[%\354p\211\306\272i#\202`ff\206\337\335\233\340\004\264\003\307\001\304\001\001\002)\347\340*\304\350\272A\277G^w\214I\254\311\265\324\034\343\016\367\376\363\346\036\201F\240\n5\nU`\311E,\225F\247\022v\031G0\"x\302\240\017L\271\214\254\257\204\307Q8\n\305\244F\260T\212\211I\231J`\t\207LxT\231R\246\323I\005\2520hD\372V\244/L\201c8\032M!V\235D \2204\372B\330\345\023\252>\022\006W\243\221d\372\340\224)\225\002?$0\034\215!(\206 i\004E(\020\237J\037\374uH^\231\273<\255\271\037a\321\353\231g#$\"F\274\363\\w\272i5\265\023\242\267B\254]\303H\237\321n\242\036\252}7\271W;\236\356\253g\262s\327|\233\251\372\362\322\321\353\314\230\275\0253O\020]\373Z\213d\304o\263\227~M\311\327\326\372l\345n\226z\225\231{\365J\220\247\2269\317\345\366j't\r\275\354\351\226\347\373\212z~kt}\361\354e\237K\355\323\215\3132\253\350\220\244\360t\231\222^\2154\317T\305\031\365<U\325\265\3432k\016\363~\257)\232\325\024\237y\277\314~\302\260*\253\321R\256\366\325g\337\2617\267\211U\224\304\264\322\304h\326\336c\232.\267\211\265\212Q\215\302\372l\313\350\257|\272dxZ\213\035\335\317\034\3223\223\233\245'\362\332&\336\251\351\263O\367\316\"\236\327\335\333Z\365R\357\225ZooMk\236k\344u\272Vk\336\333\246?\315\265\345\271R\257\353\336\363I\266\326\356uO\352\031\232{\345\316\336\211\312\t\210\023x\3626\327\2364\271\357^\355\324>s\3352O]z\335\364v\n\363\265\217\3473\314{\332\247\241\336i\210\"\273\034q\252\275\226\352\222ku-5J\232*&9_S\323\233YFM=Oa\335\237kv\351Rw\327\251{\307\230\246I\315k\347N\265y\346U\246\317\030i\252{\365\347\3325\307\231\\U\254\271\212\261\232\252\232g\357\323\224kU\253*\317j\231\325NM\305\335\323\366\226\321\250>\365F\315r\243y\253x\307v\3465T\253\270\246M\356\267\306)GS\022\233*V3\232\275FI\024\345\351\014\374\332\351\211\273\211U\225\023\224\021$\266iv\256g\251k\233\245\351{\352:1\271W\027""\306h\312]Zs\3165\272n]\363*\252[+\231\226\330w]b}\256e\212\256\344\231\267\351\3025\236y\313<W\222DI2\326I\025%cdni2s\217\275~\341\030\2511\3170\222L\311\275j\024\255\365n\276\264\213{\025\273Vt\233d\247e\224\324\347\231\222.\3373Y\346\224\267R\217\342\3619=]\030k\223\363\334\342\336\347\256\267\351\327\224\247d'5\311]\353\271\277\270\306\370<\3638\226&.G\252\227\351\332\351\007\315z\004?)S(\004[>q\004\302_i\034E\341(\005\222N\243\022(\004\221R&\230\022X\372D\020\004\246\276O\2115\302$\261\364\201\243(\212\"\370\371\250F\245\021\251\305\010<\235\276\324(\216&S\013\026E$\222*\204E\025\010\004\205`IU\212!\205&,\246\300Q*\365\265`q\004E\020\350\013E#\0215\212&\227\221\244\372B \350\204r\351\224\247\351\020{\322\004\321j\247\223\010\246\300\361\033E\014\272\210!\026\210I_\213\211$\3021,S\352\024\023\222H\037\306<\264O#}\022\213Q\316\205\013\241|\324[\"\304MiCZ\r//\375w\374q\314OA\342\260`\301\371|V\177\345\303ii\331\314\324\324\257\236\020\211L\250\344V8\260\304\377\336\023\024\347t\332\340\270\301\372\035\325\3770\342\270\270\214\277\004\307\215\177\303F\215_\303\177\217~\224\004\344\200@\233\0324FDr!\034\350\344{^\022\337\353r/\264\233\326\005M\346\313\344\237!\210`|\0300\037\203\371M\326g\245\376\204\231\332\257\311\276l\306\237!\241\245FK\177J~r\365W\247\037!\207\202\373\220\370\256H\013\322\206\030\370!\357\010\016\3129O\216\013\027\2340\344\204B\033\330+\306\217!~\n\271\022'\305A\241V\340x\341\214F\234\357\333\320d^\374\027!\016\010nl\324h\361[\244~(GB\361=\267\025\031N4\342\207\\\023\013\177\004\027\272\2410\371)\354\303\\|\027\342\027W\376\352-\301\341{\332\r2\263?\333\177\277\376\253\365[\252?\313\r\270a\343\333X\377z\376S\177-\343\313\030\277+'\200\306\314\237\341\237\273\276\013\346\303\240>\201\234\007\\\313\210\277aY\330\n\335\377ZO\033\303v\202\306\200GM\367Cr/$\276(\344\r\271\036\367\203\007\310\353\t\371\2336\203\215\263\342\2508\252\315\215\256\333a\003\346\323r,p.\267\n-b~L\313/p\213\360\202\3306\266\005\0365\276\377\036\373\261\362\227\370c\361\257r3""\034\004\334\347\371\233W\004\310\007\371\250)@\240f\223\203\\\325\021x\343\022\222\022\332\252\202g\033c\356\305\\\256\244(4\221p\314\251\3043\361|W}V\265\353\272\256\323\232\223]\327u\335\356\272\256\353*]\327\225f:\033)\365^;\222n\226X5c\236g2Rb\252\235j\372\2653\247u\223\226i2\332\234\361\264\334z\326u_\265\272\256+>{\0366\204=i\022|\22042\255\327\375\334\276\341p7\256\014\305&\327\271\301pY\354\211\017\267\337\025\025V\305C\223\000\267B\304gSl5>Z\026[\003\217\232\024\352\307p+7\003\257\010\357\313\330\014\036:69d.\004\224;C\335Xb8,VL\177mR\361\363-\005\200\023@,\006\240E\013\000a\010\000\300\nN\253\305Y\255H\337\005g\343M\021\371.\311\204\277a%l\330\010,\321\247\335ZpX}9\030\334\207[\312\033\362\226n\347\363\344D\270\n7\221\027\344\035]\020\033\027\357\313E\271\016\\\222\020\035 9\021\034\215\023\367\203\202/Ka\275\016Y\210\215\201\003\026\264\301\320\314r\014hM\330\220\315c\023\321\034\260\2154\t\002t\037;69F\217\033\200\373\270\261\274X\334\314;\340\225\271!r.\364\337\263P\037%\307b\345\213%\354}q+x0\253\017\313\255\340\364\355\345\255\340n\334\021\034\354&\340\2462\362e\255\010\n\007v0|\024{\3015)qH$\0323%\226?\241\005\331M\203\242Ap\303\026\341\243&\333^h\016\200\260\244Mf\303\212\037\353aId/\204w\246\205\320\212\260 \354\314\005\002]\202\035_\016\231\216\257u\261(6\000\013\262\231\2401\320r\300PCcS3\032\005\267\331\233zKH\370%l\r\016P\370\024>jd7\025%_\005+\2629p?\327\245ua-|\324\240Jl\014\016\324\204\330:\315\t\315\001\002\335\202\035\337\r\305\007\220vdY\266\317\346a[\360Q\363\345\330\344Xn\035\376\327\342\333\211\356\243Z\026\347\213\371\261\234\014\216v\003\001\203\302\315<\274)\356\252\375Z\016\233\327\303\212\375X\013\036\036r/'\276\307\025\201\334Lt?\2732mn=\300\330\334T:\225\\\273\r\270=\370/\316\\\370\005\344\\T\276\354MA\271\257\026\325v8\275\310\261800\034\223\251\364g8Q\312\317/\312\213\342\206\270%\036!\247T\n\210>\250\224\030$Q\312)\250\030\215\014\006Q\004\000\223\301!h\230>\024\315\006-\246\016\004\304\004\347\301\201\243\341\300\002 \201`""\340@\360i\020\306 \r\305 \t \216@t\006\002 \206\000\226\206W\306\030\013+\271\370#X\303\330\321\023\324\217\230\306\2146\000\031\263\n\2069\250b\210H\t)\027HF\335\030pm\356\277a\202\267C\337\216'\207V\3512\364\273C7Pp\275\036\214\033\335\367\"`\267/\326O\021\336\3652U\207;\001\023Gw]\253x\320f\324\033*\211U(\216\260z\010\n`:A\021r\311}\211HM\253\034-/E\252\367\2627:\260\020\007\266\253nh\342\232\237\322\3371]\266\007iZ@R\242'\004Q\210\226\023\277\246\010\322\272Z\213\311\247sr\016\262\372\340Z\261q\n\244\270\024\242o(\273\302\253\224\206\201\275w.B\023\353(\032\275$\305\313\212/\234\363'\202KV\254A\334\302\036\t\233\001@cd]\000\013\344\262\203\274>\332yT\002\272f\200$ndh\276s\363\357\003\243:\002\346\310{\214A~n\023\2438\305\3151\231\254m4Ue\r\246\006\341\213|C\347\341M\321\000\320\245\207\031\340y\226\261c\324\351\261*\357g\216\330.1X\360>Yc\220\234\007\240\037\016'f\026?\213G\026:e\375\203\212\271\233J\2267\376\255dH\372\027C\257\177W \260\360\275##\322s\337\306\215,\222\352\014D\202sZ\000\254\024\374\223\034\312u\003MziT \355L3\345^]\007\300\341\305\025\311w\230\261N\271y\r\225,\370\375\211$\007r\377\304]H\200zw\260\206\217\316\213\316\334\353i\271\341\375\2604a\2532\243\274E\256.\315x=\006\202%\300Gn$R\342\263\010\333G\214\342\303\325\306n\014a\205h\207\255~\310~\235T&\253O\235\225\022\231\274\023K\272\276%UN\026v\365~\316\271f\211\025\22460\267\305>\017\245\017\345'=\344\250\177C\370Vu\014N\217\222\004\377\032\250\247\000'\010\367v\036\230\360\306jO\0371(6\263+\331\244\004\373\320\322\313\004\035\321f\021\224\342\300.b[\311\266\370\311\242@\240\2352Bj\305\374Z\033&\201c\033'9it\322\233\226\313\250l;\332J:\024\000T\306\"Y\340\2206P\003\357yH@*-\270\377\035\301q6\200e\375PP\341\372\306\251\241\006(\260?\024l(\r2@O(?\262\200 T\372\223\223\223_\271\216\322\227\332\023\263\036<\\\010\237g\276\022*\020t\267>\312\006i\302^\014\260\000\"b\202!\001\367\217a\354\255\010\371\222\250@(\004\207\271\236\351\305\330\332\246o\004H\026\005\332kUYf\306\025\235\211\256\273\252C""\233`E\214\200\246\202\037\274nx\253k\213\376\003\257\364vj%\353\212\335\371\307*~\337\361\201\251\213\2432\230\010\300\t\212\336\272D5\311\326R\333\332\025\341\353 \242\227\007\230^%\016/1y\222\251\212dJ\232F\214\247\322\346zLXB\n\264>Vn\323\270i\307Wq\01483\303\232\017|\324\336\013\203 M\304l\200s\300v9!\302A\225\253\0264\263\346\"V\226\035\240\033\026K\223\321\377\322\266?6\010\206\326vT\363\tk\320x\366\203\350L\313H]\035\255\r&V\3212\372\274\333\230t\314YdlU]\206\307|\340\274\261\342\026`\354\214K,\366cE\315\347\243\312C\355E\312XF\275M\255:XD\353\232\210C&\232\312]%\032\216\2255DR|kT\t\350@:\272\251bz\rE#U\352\004\262\034\305#!\231\324\345\325\020E+5\334\2407\365;\213\nb\201#\013T\307WL(5\031,\377F\225\324\353\221n\274Z[+\020\\\257\202\032\357R\271Z\332\312\321\245J{)'s\031\267\032p\032\312d\n\352Yh\240\333\312\352\323\2243\227*q\265\312\222\353\301\252\315o\363\201\225\330K\037D\016\257\247L\225j\270\360\331{\345j\330\246\022U\357\227\246taK\215$\345\347F\245m\023Xr\321\220R\224\271\007\312\344\026}\224Z\266*\033\274I\251%\t\346\321pB\024L\0165l(\3252\215P\330T43U\233s\025\"\032\nq\347cp\244:*>\\ 7\373'6h\210r\034\033jC\037l,\213\242K\035\250\035\336\247\270R?\253`\027\235S\270\220\\\330\2408=\343\223<\240}C<\0010g\252\003V\312\3402\245= y\325\225L\340$i\207\321\032\324rQ8h\270Gsw\242Z\304\023\365\260&%,\325\2303\341\210m&\036s\262s6\036*\354\220\021\221\021\364\242\266g \225\034B\325\333\t\340\377\264r5\321\355\310\254'.\355p\303=e\347\262\347venx\231\\\256\224\247\0331\307\241q2\345{\251\315\356L\247\322\216\016]\306\324Wt2\227(\247\201\307\370\306:\270ju\246W\300\320qXE\0059\241\224\345+\270>\003\233\365\021\252\241M\"\222\334J\246x\000J\210fM\020\341\nU\345\"\267\013\311b\030\221\370\216\315\272b\264\354zu,\220\"\325V\215E\036\255Z\225\224H.\373\020F\023\265\325\374\214\340\345\217]\252\025\354\314OO\323\276\365\344\350O\006Mu\3259\363x\024\254-\265\003\177\022\035\243\212r5\351\303<\257h\262\376\312\333\001\3554\313\300\232\213\344\353\234\365\014\266#%zK""7\377\3544\221\027]\243\020\240t'\333\356Y\240\264L\030%\r3\206\224\n?\310\370y8!\231\246%\357u\342'\254B!NwFz\367i\261@?\216\263\204\220\025\220\222Y5\351\006\322\236\010q\277\223'\234\225\030\350\300(\327\371xx\025\"|\362\233\030\032\365#\027D\244\261\223\2069\373\254\375\270\334B=\025\225\216#\217\210z\254\321\231\375l\215\221b\037L\372\320r:\243\000\241.\371\260\350[\034B\004\025`\223c\234\3709\342\2079@\241\212\340%+*766\361\025\006\231\345\263\021\033\251J\223\363\025od\235\330\215,;V\242\202\"\030\322\251Y\300\333z\214;\250\272\321:\210\226\353\276\204\253\033-\010_\317\305\345D=t\220\\\344\215_\206\320\252\\P\212\212i\220\023v\360\020UWg\204\216\212\370\205+&\025\366n\036\245\221\326\254|sk\\,\250ro\306\334\264\347=l\"W\336\267P\031\367\241\366\270g\3268\242/\213c$u\364\215\202\321\333\275\346\277+\0263\367\tJ\234%\307HnY6\301\211N\245oS#!i\361\274\360\374\255\321\307\340aMX\3223\264Z\021~\220\315\247\255b\2578\324\025\221\"\033b\360\023\203f\226O\334\330\370@\330\377\354b\347_\n\370]\213\347\023<\342o\226\014\343\304\006\241J\274\345\200[\346\332\024V\000R\214\364\022\207\341I\026_\353\335\276\355\341\374\274\237\224\254#\215&\002{H\347\374\0038\323\025\327\026\031\3146\263E\377\001\"\006)0.\262\004EZ\272\024\030R\033(A\241\221\3164\303Q\t\376\252\356\241\242\220u\351\014X|-\022\243S?5\020\240&a\250\235\220-\022\201\367\360p1\243a\261B\275\316]\361\236Cb\370\261Q\022\222\314Uf\010G\374W`C\303j\323/@\310;\272\262v\013%\377\310\325\034k\022\255\324^\320C\2411E\274\306;\256EO\024.\217,\263j\365X\354\246\032\257yG\351p\232e\021\n\000//.\024%7\245\332\342\217\232hB\277\225J\234+\200Q\276\363\222:\207\353\212\266i\n\200\232\273\217\031\277,(\213\tU\217\n\236\352-\345\261\367\270Q\321;Cl\357k\003\027]'/3\217\246{\202\375\3559\002\370dL\220~\000w\360\245\001\337(\272\016\360\243\260\275\240%l\3151q\017\250\360]\267\364\372\373x/\312\312`AS\206\177\363.\337!7q\3555\306\034t1\3452\036\371gL\335\323\236N\034j\212k\017\216\337\017\016REq\212\\\365.\372\203\307\325\316""\361>\340\031\241\252w\321+/\245\377\354U.\276\225t)\200j\022O\036\216\354\035@\314\3739~=\211\221\226\211\374\317>\037s{\257\364\217V\r[iI\321'\314n\277\nT\3457i\313\302/\316\316\220\317s4\272\320\206\231\312)Z\320=7\372\366\377F\235-\2439rN1\230\316D\023x\334\001\177s\357n|\314!\237\266M\014\325\0072v!=\243\014\202\t\001\361\212p\005/\016i\2235H|\226\"\361\023@N\204\t\360\340\334\314[\377\200\205R\332\226x\037\2037\310\310\201 Z\245\3317\024\225]\355\320.\3112S\227D\333\340\354\347\264\232]\233f\307\353\3658O\373!+\240\022#\321\027P\374\260\025\205\033\230\377\2425\035n\204.\257\005\345V\321\375SN\361_T\025M\316\262\267\254{\235\317\2735\364\307a\237\007>\344y^\355\254N$sA;\214\311t\260m\336@\013GCm\266\247 yV\036\354G\231\242\330\035`\036\215I\2157\351E>cF\365|\025kz\276\273\306\200\306+\345\nj\3249\303a\321\233=\252\352!>\370:\211&\362\3206\032\210\303G\331\007-Cd\363\306\025\306\372\340\226\3765\256\236r=\367\306`\330\2418\240\010A\005\264\243mJE hslq_\330Rv*\206\367\340\240\010\036\343\027\360Y\267:\034\026\223\014\376\211\324y\243K\"\2411v\2473\310\314q\324\220\311\006\356\214;N<yQ\362\226J\255<\366\220n+\324\234\362o):\221+\320\201a\201<\177\t\037}\200\005\017\352H\340\336u\316\035\t\016\354\213sL\232\336+\017\240\313\320\346\241%\260\320pc\345\0300E\313\306\326\360@b\252)\316\"W^I\244\373\375\014\216\330\350\362\267\241\344\"i\000\234J^6\232C7=4\033\335!\325\206\034Zl~3\313Y\007F\263\265\345\r\261\234+@\230\001\217\207\030\n\025\341#\273&p\227\336\205Omh\001%\336%\300NyU\016*/\222_\270i*\252\317P\330\310Om+\306EW\216\304\023\254\221\227\221e\026\362\366\225j\210\213\376\242\316W-\024\365\231{\331\014R\226\234\035\020\020\212\201\032\2632\214r\007\013.\031\3571\006D\312\254?I\345\201`\026\013\302\271\t\001\321\2639\260\030O\334b_k\213W\3247\263)K\347\322\243\333\233\326\213\211\304Z\346\261\3170Q\337\200\255\305d\375\265F\313\304\000\321A\332\377!\200\216\272,\244\001\354Q\317\262\267\227\211\253\241\027\251\315\030OQ\230\033\2708/\232l\335\037\302\261yc""\376&\220)8O\346\337\344\202\271YP\036\370\002\304\260%\2210\016\274\014\207\343\303\261j`\231N\037<\220T\231\264v\315\353>\010\316\331dA\304\265\310\000\033q\351\017e\245\354bw\3224=\215\270I\3718\203%\247Y\t\276~\351\237\216\242O\036\310\231\220\370\031\376\035x\230\225\325.\234Vd\253\250\244\014\201\242\214\367\241\007\355\013~\213{VO\332\021\307\313\206\260N`\355\211\256\262\314T\347\270G[yc\005\353\003m7\"1\233Gl\333x\272\214st\351\246=\340h\277S*\260\332J]Ba\261\253\2328\365\237\221\200\330_\2676\270\noO\302%\260T\202M\357\210'\264\370U\321\177\222\245\230'\221\026\0141-3\301E\220\006\000o\371Z\214\367\314.A\370\347\023\263G\306\330\257\316\317\346\000\025o\364\327k\374\325;\312x\365w\024\222\312\271O\3113Z\037\315\320F\222\004\341\312\241\353b]uXu\305\342*?g\326\307\312\327\346\210\277G\255\301c\354\367\365\372)\254\377\363\274\0008\251\217A!\227\354\217\307\334\201\364\325\350S\254\276[S\373\\+\2520\177\26280\272F\307B/~\305@+\000nP\233\303GE\363\262\241'\237\000\256n\321\224/\207\211c\334\022\266v\273\205\005y\362\264\374\330:#\026\350\244\036P\210\n';\0043\327d\277L&Q\357 \2473=n,\363H\307:\312\334\334\207W\372\207\336\tDW\271OP\206Hdao\013\0329\004\035\370\235T\366\036\236\356\366`@g-\211DB\214\030(\277\032\320\270%xH\364\346`\257\350\310`\336\003,N~\340\234\312\322\325(\000\270*\351\226\025\243\236\337\002\313\301\262Upr\253\357c\306\n\305\201f\207\211\031\t\306%VG\310\264Z\335\032\334;/H\324\344\271\205+T\351B=-?\304\377\337S\227\351\3123I\327\211L)\370\215+\314\325\302\272\221 \352n0\220-\252\276\271=\377\336\"\225\204\022,\244\3403\001\334\021\202\222t\204\017\367\314\237\370\203\210*\207\342\352\254x\313\337\377\013W|\215\365\217\252f\024\354{\221\205\221\030\2665sE\233\224\324W-\301\315\265\2012\225\314H \027\242|6K\r\207\352\301\"9\177(\241 \356\335\177\343\311\232\312HX\270\036\373#\013\247)\217O\270'\250X\004\336\210i!\\\314G `\021\311\037\3373\223\035\377\302\356\351\215\255a\227\223\236\275\377\223\272\207\335:\341\271\324\211B\007r\033\357""\342'\342\037\360\326\205\363\riB\034X\276\367\"\304\252/\330\256\340,sF\346\373s`C\014T\034D\031\332=\304v\336\3026l\037{ F\313A\333b\022=\272\305W\006\005\\\025\033\274\006\030\016=5O\331\317\257\016y\235\333\311\003U\256-\233\261u\304\330\243\323\312\0352'\232h9Wu$\337y\022\357\023\301\341XC\031\314D\352\"2\220\360\005\231\007\340\333\265\316\350\345\345\342\237\205\374\013.\373\377\270\200\336U6!\371}\366Q\305:\350\004\257X:\344\237\313\005o\334 %\257\254\371|\210H\035\030$\246\372=7\363t\3106Ozp\244V\344\307\320^$f\321\230)\367X\305\026\032\030\214W\317K\364~\203\356\221\246\315\377\314\247;\255\006\276\311,\315\331Bo\352~+\250\350\257/&J\002\361\270\304\3471#\010\245Oy\225\276\025\255\r\254\267\216\315\024\342\351\240\233\375\356\357\005\243\347l\265\324\355\301\3708\251\333C\222\233\206\010\276\350!G\2335\272q\211BW\324\307\252\002\343\343g\331f^/\225k\032\364de<\024ZW\242{Y\357Ry\033:\366\354\334\023\026b\267x\004'\324;\222{.i\272c\305,\023\305$E;\224/<\372\334\264Q\252\320\2060_Z=\204\347`#d\025*M\276\266\276\244k\371\024\370\203\314\345W\335h\016\211\216\357\362E\213e\313s\016=>\000,\023\026\350\233\330\026\021\177\215\030\371\013\200\022\300nGXy\230;\222:\230w\257\307\362A\n\210n\210\005\006:\227X\021\225\211\014+\223\337(\032\324\376";
    PyObject *data = __Pyx_DecompressString(cstring, 6068, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (5911 bytes) */
const char* const cstring = "BZh91AY&SY\264\235\217\025\000\010(\177\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\300@D@@@@@@@@@@\000@\000`\034\376>,\326\302A\003,\304\005\021\n\221\324U\005\340\005\360\037x[\214\226\373p\256\355\304\246\265\335\335>\017q\307v8R\225\350\313\\GP\266X\323\020\335\240\000\010\t\"\200\232L\021\223\000)\211\232&\214\232\003H\365O\3211O4&\231C\324\364\323$\333D4\300\246\324\3657\250\3016\206\222zc*\rF\231\212\224\312'\265&\365OP\001\223\021\220\014\200\323C@\310i\220\031\r4\000d4\304\032b\000\001\241\240\325S\377\323R\225 \0004\323A\246\200\000\310\320a22\r\003@\320\001\243@\003@\006 4\304\310i\240\321\372\251&\246%\036\223A\351\251\352\036\241\240\000\000\003@\000\000\000\000\000\000\000\000\000\000\205)\001\0014\232\233)\233I\222z\236\243i\032\006\206#A\265\r\006\200\000\000\014\203F\206@\031\0004h\t\022\023 @\t\240\231O&\224\3305S\362jOi5?T\331\014\230\247\250\375P\006@\323'\244h\001\24042\000\003A\340\nE\002T\302\237\370A\310H\312,p\302\203\020p\023\200:<\332z\200\3004\211I\261(\203 \276\376z\031\370\274\3179\017C\365~\235\004\363t\260\023\317\340\304<\375y\267W\034\305\024\320\nHHh\205\002\014L\257\212\317\006Ztx}\371\317+\235\360\"\261\0101\246\016B\013\242\020e+\2474!C\002\030+]\0023i\024hA\225 \376*\262\020[\tI\010\236T\261(\254J\353\201\005\264\027\261\215\264\265EB;\337\336\177\003\223L\364}\037H?\226\270\375\237\314\310\277\307\247z\177\363\324\364\324\013\342B\243#\201\340\213\345\013|\001\001\033 \365\353\366\227l\374?\020\356\020~Q\235\322A\316\361\027\177\277\340\237\341\304\361P\240\337\235\022\017\031@\377\206\304\351?h\374\277\247\227\321\233^\030|\376ZV1h\324\202\246-z\323\251\253\nK\215\367\360\363\362\357\323]\250\3141\3455\2130\252\326\032\002\374u\350\335\304\322\265\226q\354\343\331\267j\334\377G\272\030a\206\035\2607\031!!\201\220i$\030X2L\254\260\370H\370\016\207l\325\334\r\340p\025UR+\004\275\263f\010<\005\322\371D\013\351\240\304>\0244\262\006\202\310\244\004\314~@\214\354\0062H\245\311\002\211\240\026\222\346\2173""\001\264\203\026\006\213\265D\311\2014\002\205\242\3449\204\235\321\246\340\365\243\020\212\005\214p\213c}\254\203\nN\324\210\264E\320O:_\\\213\\[L\036\273k\273\031\321\031{\347\037\200\372\211\2666\233t\221E\002\320\204TEBM\020\225\333<\363\270l5\037\334\225\370$\273cY\031(R\254F\2760\234<\237\313\376\337\027\361\204\230P\000\270J\374\177\246\337#/\252T%c!\266\376VP5j\307)\336\321\251\315\242\265\254\353J\264U\315\242\254\275!\027\377\014\005X[T$\010\253\001\261\002\317\345\200_s\343\304y?\373\226\005\310W\332Y\321\304F2\2251\250\031N\004\205\205\2266\241(\265\252\222$\302\334{9:\372;;\375\275]jj\341\261v\215\272\235aPu\307\3651\202_rDI\314\302\374\226\035\0269\362\277B,[ww\267t\007\346\252\242b(\242\224\032*\273\277G\r-'\"\216Sv\343\3127\250M\266\316\367\251\232M\252\3017\340\021#\005\211Y\r\224U\036w\245\236\207O\313\373G\021\312\212\274\332\336\207\205\257\330\351\336\007>\363\234\331\306t\010\210\211\024\210\210\222\000#\2132\021\357b\2202b\"b\"\014$i\034`\232\253\033+L0\210\210\240H\221$\242\"&\031\206C\025\235\312*\330Q\237\336\325\311\242\\}\215\355\317\340\026b\305\274\016\356\037Vk\223\253C\326\232\246e\"\315\231'e\311\221b\342\240\213hI\026\241\203o\010\313t\021aP\222\257\317\303\004Z\361\347^\213\232\272<A\373\224C\246\334#\330:\000\241\255t_\343\373t_\014\004\033\003\"\344Z\210\200\373\277{\357=\376\000\314\004\025\010@\221}\213\374\342\326\\X6\006\365o\273y\3649\017U\346\204\222Xa\206gFf\323)\312\371\242\324)\004Q*1+!N(\225I\321\236^!\274\214\237\034\374lK\363\030\222\232\232\030\325\005\335\341:\037\272\217n\266W\330\007\331\007L:\312!\340\373P\322\004\210\010\217\273\260\021\355%7_\256\224\231A\227\223z\2627e\365\224\3524_\036\354\261\305$]\244\304o%\240\264\255\007a\024\265\306\301\351RT\257\330\371\271\372|\037g\253\274\333\315\266yN\203\275\352\013zE\250\322m\317\034T\215\304\300d\374\024\371\2344\331\223*\310pJ\010Q\325\361\365\236\322Y\213\216\201`p\026\364\007\347\313\\\341Mi\370\377X\241H\202\236'\321Wo\033\\\300prX\312\362V\364^e\311\037nPra""\325\240\231@\317\032f\212\001R\204\021\335\"\014\337\220\222\342\036\036?\336\206\034\241\312\031ma\323\354o\300<\207\263\352\016\344\245\344Y\254\314\307\336\014\333G8\321\270~9\207\266\032;\275?\340i\323\2403Z\016\224\031\363\213\245\256c#\177\274e\210b*\235\376~\260\214\317\034\267\351\002\252\235^\211\237\003\371\251E$]_\003\221?\266\244Tll\301\034\312\356\0202z\023OeC\"\255U\202\357\017\205\245\353\222C\330\\\232\030LK\227\007OW1\356\276\036\327\272L*\327\225\236z\"`^9Y\323\243\311\002\271\307\372-Z\265\030{\214\352L\371\340\373Wi\210iS\224F\355\236w9\257\"[m*\263]\323\023c`x\276\271\257X\377\232\356n&R\310\014\237\026i\006\301\321\240\210\230\022(\230x%\341\322S\202\316\0344\325B\342)\004\225\224\000=\334\354\343\005\272\340]\257<'B\216*F\321IR\013\023\352JS\274\024U\230\353\332\221\026\321d\230>u93\006\367\261\225+TGGq\324,c\0067b\255\\\312aZ1\321\007\007;D\337\300\270*\001\241\203H\037>\010\231\330\035\250\226\353\354\026\316\3603G]\010\200:\221I$j\243\035\240\200\037\253\035=\340\333\220\223\374u\035\035\350\313.N\231y\251w[Om\rw\353\272\244\001\237e\242\003\027)\374 \262\3741\273\355\002E\337G\333\252Z\372\375\337$\375\006\301\357z\23497\347\312\355\256\335\352\327}\360\352\213\226X9\216\362\016L\005~\362\371\260\315\220\230K\353\322\257\3530{\373\261\360\207^\354\316\324\212\246J\322\301U\232\214\021\356\205\227Z\252p`\256hUi\315\277g\247\234[\010-:\234\215\375\307\217\271\017\335Xi\0368\242\364\274\274\341\206\036\016UK\3578\365!\266\305\301\030\247\201\316\006\023\024\233,\234\305\272\327K\217z~Z+\203\245\301\2241x\267\024H\271B\225\003}-9\021\213)\014\374*$^\202}K'z56^=\202\021BN\256\325\326\004daT.\333f\355\316\375|\324\351\035\221\355s\316\276K\272\021\221\211\214\313\256\247+\202\351\336V%IH\325\356?f\227\230]\332\031\004\261\334\362<N\272\364\353y\027\335\r\335l\265i\335:N\3722u\217D\357\240\301\341Z\327/\035[\316zo\263\226\033X\335\316%5Rh\373F\233\274\322D\321\321\234^\365\177\177\007X#\365o/'\276\310\314\026D\300\210\351n\200\360&F\000\000<""\n\372z\302X\3661s\345;LY\220\313c\247\025\245\266\337\304\357\267\252\177\007\344}ZbR\326\233\226\3350\341\244\221$\222ph\271IS\016\032If\031\264E\357x\207ak\333Tt\200\275\2009f\314\265\323\237:\255u\334\0270\223t\236\033l\224I1SNIt\303\206\222Y\014\327\265\206q}`2\022-}\200\213\310\032\202^\364\321S\200\000\350\330\032\364i\217/\354a\256\350\317d\366l\331LJZ\323\264\355kS\016\032I\022I'\006\213\224\2250\341\244\226\265^\3778\210\215}\200\321\251\2048z\026Km(\333\003`\313\213CG!'\030Hqq\200\356Qq3X\260\265\202\270\260\004cW!\323fz\272'F\215\025\220\246\345\313n\230p\322H\222I84\\\244\251\207\r$\260h\271IS\024\334\271m\323\016\032I\022I'\006\213\224\2250\341\244\226|\364\323\324\234\262\312\265\0057.[t\303\206\222D\222I\301\242\345%L8i%\203E\312J\230\246\345\313n\230p\322H\222I84\\\244\251\207\rV\265\275U<\223hvknp\022\247<c&\333\2119I\267\306%796\025\024\006\306\212\003Z(\005E\000\250\240\025\024\002\242\200TP\n\212\003\320\231\251\226\342\005\255P4\271\230\201kT\013Z\240\n\240Z\325\002\326\250\026\265@\265\252\005\255P\005P9\251\231\226\342\005\255P\025\265@\262\203T\013Z\240Z\325\002\326\250\031+\206 \025\024\006V\212\005\242(\005E\000\250\240\025\024\014\271\003& \013\024\001b\200,P\005\212\001\252\314\231\200,\250\004X\240,\250\004X\240\013\024\001b\201\243\204p\001b\200i\2313\000n\252\314\231\200\013\024\005\226M\266JD\244\333rF\215jm\3326\242\316'\300\307m\305^\334\3512\204\025*]\335\355\334\221\257 \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\200\000\000\000\000\000\000\000\000\000\000\000\000\000\000.\320\273@\013\260\000\273\30338\3358-p\222\031\231\001\251\201\322@0s9\036/73\020\355kZ\300\001{V\234\243\031\347N\2355\260\024\334\271m\323\016\032I\022I'\006\213\224\2250\341\244\226\r`\272J\251\323r\345\267L8i$\222I#\203E\312N\233\206\222ZB\026\315\257\266\256\005\240H\201\231d2\0202\022\206\240 \202\001\006\316\252\250\314\314\343\030\304\314\314\253Z\302\256\004\240\241\224- 7&\275Xd\225\225d\230\211YLd\214K\304""\312\344\230b\356\262R\\\254e\022]R%%\004,-=\344\266fj\234U*6\301\260B\331YtSf\353P\225\226\022r\213\332@T(\022w\344\245*\024\"\0055%Y@\270/7\207\310e@\0338u6\220\327_o\255N\211\357\004\210\322l\275\352\351\246\323\251\002_\026\354w\335s\325\273\t\316>\233q\251\345\252\370\234\006\202E\310\031\001@\322\024\224\255\005\257\325\321b\013\346{=\235\240\345\000\0201\204\023Y(\203\rC4\251\375hXcA$\266\027%TO\201\274\026\353\300\275[pU\002\\6\030e\0041\tb\225@\013<\022/XE\032\301!\010\310\036\030\320\036\315\274Az\335\365\026\365\361\274\216\341\3431\343\236I\356M\023f\253;Y\357u\375\004\256?\211\312]\336\336\3261\277\307%7ji\321&w\230\000\000ww|\215\037\177-\"\360L\222ZJ|\223\300\231\002On~$4\035\356\301,\215\005\177\203\306p\224Gewp\267\017\017}t\272\350.\212\"o7\363\")\315\002B/\366\000\r@c\215\346\247)A\233\224\240\223\224\240\223\224\240\223\224\240\223\224\240\223\224\240\223\224\240\223qq\257\321\253VzE\024\225{G\252\350\357\374\375\\\020%4\262h\305\251\200&r\335\267N\252\211n\311B\004\233%zA\372,\005\273\035\321\027D_\021+\020%\267\344\352\370}\257\303\370\267\273\302\030\000\301\203\0376`y\004U5Q/?T\300\351\002\233\r\236\017\215\321\346z\273\321^|\0258DN\345\202'\004FP]\215\000\007\270\231*\302\240o\230\337\030\210\200\000\006\231\243g^\377e\253N\233\254\252I*\246u&\215\233\276^s\234\251>G\230\305LL\356\354\343h\347\027\240\335\343H\t\310\207xr\036J\372SK5\270\030\016I\207\262UA\303\266U\22131\2661\273\250\3061\2467\"Cmk\333\334\323Hu\025\021Q\025\021Z\324D\330\352\340\235f\262\242)E\3147\202\212x\241\262\013\031\\\2251B\204\240i\205\031TfU\252\024-\271\306q\177\321j\201)\305\346\3031l\303+R\311\000\221\352\004\352\351OU\326\225\005~ \225m>b5_\\\251\261F\020\002F2P\010B\303M\364\332\250\262\344\331\331+\363\321\3258\311d\004\241\002\200_`\303=M3N\036%\334\234\234\266\271U\354\226\204\204\"Y23\236m\315\251m\277\312\304\323\301@,\345U\022\006)C\002V\271\342V\365\255\237\035/\364\364\371\377{\211\337Lg\213\232HF\226(\214\320%a'J\232\002\004\026\312\323Q\335\205\264""\222\314\222\273\275\207wwwww\210\210\210\210\210\300XVt\246~\177\023\325u\275g\213\304\036\347\037\207\r\275&\264&\206\t\303@4\014\234\2277\030\363B]l\331\273h\342\217)\246\344L\227\000d\341\266\346\306OW\320\326h\363\362_\335\247\276\000k\010\3247\306\"\"\"\000\000\344\323D\252\252U\313\275\331*J\223}&$\221UJ\030\343g\306\216M\344\272y%\252|6\347K\230\213\351+BqM\245t\206\"\220e\202\006b\036iKPe9\210\366_\007\264=8\241\334c\341K\222\350\243%v1\317\312 K\022\2505@t\027\324\222\346\2766B\204\224\231\251\254\r\024\374uQUT\025UUUUUC\325<J\206\232A\256\204\010d\201\246\316\201\256\323L6\332i\206\226\232dE\261\026\304[\021lE\2614\204\000\314\263\275\026\315\005\261\026\304[\021lE\261\026\304[\023H@\016bRjGM4\210\266\"\330\213b-\210\266\"\330\213b-\2114\204\200\022L\313<B-\320[\021lE\261\026\304[\021lH\2654\204\000\271\226\333m\000\0006\357}\225M3b]\325U\313\271w.]\335\311c\006\372\215\231\013[\366\300\017d\020l\272\247\32140\2376\210\000;\205a\032S\022k\225A\237\013\264Nw\n@$4\327O\324\273\230J\327\352$\312uS\344\341\353\340e\345`\004\013V\000\342\t\036`\360c\342\3545\333F\323A\364\264\310\r-{\250\003%\310h\003$\3007\371\021\2749]} \353\365\2021\t\344'\00445 \270\362p*\335A\274=\243\235\203[\367\270\263\2642\233$\000\326\202\300\001\310*\360\217~\017K\255\337\336-\245\245\330\221\021\021\021\021\021\021\021\021\026\352+KK@\000Y\310j\227\2638\032\352J\254\244\312q\033\335\206\366\367\017\242\366*J\212s\350\340p\222Y\217\026f\230\350\301\240\351\3211\322\002\230\232e\302\320Ag@\021\200\002\023\331u9\216k\263\277\036xAv\251k\352h&\310\037\030<\035\270\206\361\211\221\0208\216[\203~\"\274A$\201&$\216-\001\2442\342.3\327\321Kb*\037.D70\254Pr\345\204\202c\001\264\241\202\014\221\004\201\010R\333DK\315\255k5\025\0354\302CiT\263\366\034\304\tM$\220\003\035\n\244i\232\010)\320\364lK\0353\252&\311\227\253\260\235\342\264mw\224@@\301B\256\351G\036\350\r\300\270>(JRk5\306n\347-\267\013\305\3104Y^\000\250^$\0131\203`\200F%\320\220\234\234\024\205\t\002\226?\222A\216\233\310\013k""\370m\272\254\017u\311\014\206\030xUH\246{\250ST\240jl\200\002\212\024\241\030\302p@ \211F\035\236\237O?\323\277a\315\363\216\276\373\236\311\246\201) K\244I-+\242\003\306\316\314\346\202\210\364\211\007\211Q\002W\203\014J\2448'\274U@/\037Q\332\tN\024T\342\304\007^\355\273\344\307Z\372]\2651\n\264\204\330\244\261}<\275\353\322K\273a\207\203n\203}\344\014k\216\251\322z4LP\003\243\271b\025\230\324\324\271W)\324\000\"2\354\212\272\221\341\237\003B\020\323\216\002M\352\202\030\363\260LR\274X\\a\312^H\321\331\005\314\314@\237k\014\2358DK\221\224\234\332\004L\364x\311\034\200\201\321\321\036\377\275\354\221\357\311x~\254\302s\231\341\267g\203\247\334\373\001\344\264\262\016\013\235\217\250\033\010\216b\023j\005 r\035\030\007\035\020\024\300a\300\026n\317\273\353\3665<\240l\347[J\354\035{\203\277:\372\321A\016\016~\354y\342\224D\343\224D\322\\\342C\037^\035\236\032j\232\013\002\200L\332\333\326\364l@\226Aj\252\234\271\362\204\313\244\312{\002\337\252\020\001*\000\2063%8\014\017\273\243\235O.\263a\346\036C\031\023\346\224<%\n\212\201)\022\242\223)B\t\304E\020\220\216/r\345\352\000\000\0063\340\33665M\315:w-\225\311\0275\363\362T\225&rc+\227\244\000\000\023\215\304\230\331/-\314\344\335\234\r\255UY\315\023cN\370\310\000\000\031J\350n\245\r\210Lq7\311\000\0030Z\345m\267L\000\230\273\250\252\215y\336\000\232'4\t;\267\r\245{\224\220i\203\001r\212`\000,Z\004\226\346$\032\030J\345.VF\232\010\022\305\222\241\tk\331\013D\303\307cW\252\350\237\003\2408\272\252\241R \344S\027\013\227\200\3213A%\217\022\010\001\014C&L\000\241%\251\320)H\013\232%(\r$\300\325\335m\331r\232\366\000\322%$B0%\020\245$@R$\300D\301R\320U4PD\025@\301\"QN\244\312h\252h\252h\252h\252h\252h\252h\252h\252i\022B\274\031\326\357\222\253c\310\302\2677q\021\024W\246\232s\230\3468eX\305\336 .\320\027h\013\264\005\332\014h\333UUUUUSQTUUQUY\206UUUfeUUUUUUU\013\273\273\273\273\273\274\344\231\347w\234\005\332\013\275\014\345h\253\225\250E\301\014\241\213\004\331\351m\245$\200\255K\247u\004\252lb\331s\216\267\034\221\017""\203\216H\213\234`\342\216!\202\024[\3709iEV\024ye\230h\327\232B\021\230\245\244\321U.TeD\2264\242\333\274\225\330\340*\312\215\266\027\233P]\334},\020{\316J\211\211$\231\024\235\341g\000(H\336\t\r\010H\333\220\241\024\255\322r\205\r\231\360\270F\315HHGX4\n\355\371\246\265\337\004\271\316\320\360q\335\245\371\240\204\222\262\000H\337\256J\251\267\234\222\356J\253\234\032\221W\303n\334\275`\000\000\332\223\030\300\000\000\006$\306;\323 \000\000nh\227u#,\226X\000\002\314\006(0\027\035\033c\320\344\330\201.\222(&w\366b\036\323X3\233\031\311\0314\263h4\226P\267hR\350\021\\ \262&\271E\312,\343\223Sh\317<\016}\362\320\305\215f!TZ\033\031 @\203$\274\241\332y\374\275\335+\324\026\3527\232#a\251<\326y\346cnM\236(6\230|\3467\241\n\032\006\301\007\026!\016\364At\300\002\000*5\020\344(Cy]\367o`\303\031\210T\340\025\312\371S\343\256\330\273\367*\"\201\243~\254\020\002\312r\001T\026i\016-Z\262d\333B\024\273\347e\001c\236\235\247\267\240J\332\232B8\256M\364\327CV\345\332\010\025\202\375\374\230\314\273\367\363\346\214\346\222A6PX\266J\321\004JXW\242\376T\255,\250\224\317\204\363\2523e<\201\304\363\020\310T9\372\375_\274\344\267\364\301\247\235\03414\016.\373\014\305\266\"aqY\"\026\024si\300\016\3127SG\307\206X^7\331B\361\007\350\300\231\036\317PU\n\317K\\\343\214\306\310!P\020*A\276z^\320\311`\354\206\326\020`\372\367\013\241h?\212\245\256\352=\225\375\3423\033\277\261\221\3310\330\351\207#\304\026\272\341\207\247\374\2340\327\275\344^r\337\366\370@\336\301\213\205\340n\020\327\233=|E\344\356\364\273p\277\023u\241c\301{D\371k\265\374\001\331\272\333s\036\026\364\227\276\347v\350\217\344.r\375/w\250\333\201\346\014\013\245gX}\317Y\370\2304dE\347\030\267\265\276\034\320\265\262_a=\326k&1kU\337\201\253k@\266x\306\357q\331\230\340^.x\271\303`\335l\253\351\203\315\242c#W\001\251\366\237\260\033\255\371\017\251,\006\373\024Wv@\272SZ\336\254EcS\252\240\302\254\331\371\271\335\003D\372\024\246\245\005\325\r\367$+4\374\377qWoc\345\263\217\243\373\276\261\251\361\316\340\355\355&\315\361""\213c\037\247\324I\377-[\034\177\250M\\\236\027\303\350\254o\034\346_\325\356g\321\030\315\320F\323ek1K\214\302\241\266uw\210\267\315\366\2562\240[3\377\017\335\355\353=\277u\326\247\207O\220.W#\231\254\320\273i\240\356\000\350\005\251\311\346{\346\226xC\006!<\207\346\205<S,\024E\2024\247\"`\230\234K!~E!\031\271\237{6\365\254\336O\2447\235\277\226\364\336\237\342\302\346\235B\357\337\\v\347\177\232\273Y:x\341\247)\275m,\357\213\253\275n\017\031\317\347\337\364\337\2009\347q\216\202#GIv\205\256\277\3623\317Xw\227(:\n\360}am\037\272\332\014ab\206\320sE\264\217b\371\321\345Ka/\244\364\0131.z\311\351.\260lu\027\200\376\241t9M\255\004\0204\034\306\346E\006\\\021E8\374J\317\344\213\005\337f\177~o[\345{\374\277\257\252f\327\022\215\230\302\233A\314\327\375=\017\221\243_\355\003@n\330\231\267l\260\t\372\304\nH\300.\244\244\025\020):\311\322\263C)>\022\220hFy\t\003i\r\237:;\362\3445#\235o\337\316\034\021\3005\206\223\377\305\334\221N\024$-'c\305@";
    PyObject *data = __Pyx_DecompressString(cstring, 5911, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (6629 bytes) */
const char* const cstring = "x\332\315=I{\033\307\261EZ\211\345\310\317\211\222<;\211\263@\357%\326b\2311mYB,;\016D\202\024e\020\\\000R\224lg<\004\206$L`\000b\246)B\336t\344\221G\034y\304\021G\036y\344\261\217<\352'\370'\244\272g\301,\275\014E\275\357{\376\254\231\236\332\246\272z\231\252\352\202\364I\271m[\345\266k\345\334-\323\315M\365\334\255\266\235k8\271\272\325l\254[]\323\265\232\275\234\343v\0335\327\3522\";\267X\\|\357V\376V\316\264\353\271\256\365\215Us\235\234C\326kM\323q,'\327\336\310\255\223F\323m\3309\267\327\261\234\211\334\334F\256\327&9\333\262\3529\267\235\353 ]\224\301\335\262\354\234c\271\254\221\273j\332v\3335\335F\3336\220\275ao^\315\325\033]|Ic\327b\3343f\323\261&*\226\225\373z\243in:_{m{\327\331jY-\243c\031\r\273n\355\031n\002\341Zf\313h\324\223\360=\243\326\352\2607YiL\303n\270\206\203\312\020'\215\354t\333{=)v\247c\240\251\352M\2116{\206\323\330\264\315\246\321\356\010PL]\263\325&\266+@\362\027&\241\356V\3272\353\206C:\235vW\300\345\266L\243\323n6j=\216\3432r\265v\335\312\375\363\372\304gf\275n\240\321\255\272\351\2329\263\273IZ\226\355\346Z\304qs\353V\316\314\225Ik\261\227\263\353f\267k\366|\"l\205\024\223\323\002 N\203:3k\216\330\215\035b\241\345\371c\275\341\230\353M\313\262\331u\263\306\255\303\346[\233\270|\346`\257\353N\303v\257]\027i\262\336s-\347\275fc\033\345\257\263\231\307x\234\306S\324a\"Wm\3476\375Y\324i\243\010\234\257\330\263\256\345\3609\211\023\227+x3G\034+7\321\301)\355xZ\324\033\256\325\342Br\255\206\3232\335\332\026\316J\234\363\004\345s\270o\307\277\257\243\2728#\235\277\373\200\211No/P\004W\017\312\364tEc2}m\322l\346\256\275\177\235O\275\300\014\327\342O\273f\223X\237^s\2738^\033]s\223u\370\372Dh\263\002S\331\010\237#/\233\232_4\212K\354:[\344\327*\273\226x\273\304\333e\336\256\024\313\325\271r\2614\345O\363\342\362\362\302\2621\265PB\322\352\334j\321(\025V\312S\367\215\231\302\\\2518\355ag\027W\214\362B\025y\031Q\000\235+W\213\313\345B)xZ-\224\346\246\r\274\256\024=\020gYY\\\\X\016y\026V\252\306\302\2141_\234_X~\344\201\002\205\374\247G\363\363""\305\352\362\243\031\266\226\347\312sUc\276\260\306\357\325\373\313\305\302\264\261X\254\360\307\207s\325\373\306\374\342\034\252>\077\077\202T\356\243\354\321\343Jyni\24587=\207K\267\340\272\335\340>a\030]\253Nj\226Q\343[\234aD0\270\367\260e\025\307\371+m\316v\2136i\031\376\035\225+\257\362w\026\273\335v7\332FA|\3030\022@\377\275\330\021\243P*\261k\371\021^\003\373-./\254=2fK\013\367\n%\243\270\206\035\231\301\356T\356\027E(\336\3214\030\2652*\325B\265(\300\261AY.\242Q*U\241\310\020\267\350\357U\213\321\215\315X\354\355\341\237i\334\376\215\262\265\347.[\033K\213\254\037\354Z~\204\327\351\342La\245T\305\326\375\205Ju)\276\357U\346fq\302\030\205\351i\277U)V+h\033\343^ayy\256\270\\1\026\312\245G\034\302\006q~e\236\267\227\213l\220\213\345\351\3424\353\326J\305\230Y)\225p\364g\375G\337x\301S\305\270\267\260P\255T\227\013\213\213\305\010\224\231k\016\t\037\207\300\322\334\374\034\263\303H\024\263O\224nej\252X\251T\202\355\2712\332\214\275m\263Z,\314W\230\266\254\301WJ\t\247)v%\371\214R\247\213k\034\312[l\224\374'\256<o\317O\341\014.,\027\023O\021\336\362\302t1lD\340\225\302|\221\275.\366\220\304\317\263\231\026J\210\203\242\264#\025R\357\177\270\260\\\232\036\265|\214\2676gV\312l!O\373\217\3638\017\346\026KE\377\261R\\\366\215\032\000\346\312\263!\266\312\364\010\366\202\352|\301\230\236\253\024\356!\032\233\305r\320\232YX\236*Vq\307\234j\333\033\215\315Q+\275\240c\270\324\222f\330\025\276\213\316\325\243m\261\234\010V(\t\035\211j\354\223[\r?\260\001\253\374\005*\341f\223O\274&\266<?\0107\016\374\016:\246\323\263k\215\366D\255\335\305/e\303\266\034\023\367\256u\374\2425\254\340\206\314\315H\323@y\370\021Cm\003\340\010\260\321l\233\356\344mF\346\266\343\264\001j\275\3336\3535\323q\005\270\215Z\273\331\304\357\220\000\3252\367\202\376\252\220N\r\365\027\252\324j\330\n\001!R.\300\301\335Z*`\204L\013 \033\370\277\201\037q\274mX\376\025\2517\033\016#tz\255\226\305\034a\037Al\001\212\371'54h\273V\3332\273\002\373r\260\300\266\034\236\266+\007\213l\232@\244z\343\341\005\266L ""\304\214\"\033&\020\002F\253\266\035~7\270so\240?\306\234+\366\210s\026\277\t\206\3335k\326\272Y\333F\267\233y\336\334\377\251\361e\353]\261S\316v\215\355\2675_.\336Z5\036J\324\010\006%]\263\33336\032\270J\3205\033A\330\267\267FZ\355:\301\217\317\010\355\003\030\2269\251\006\273\324-\307e\177\274\230\300\250\263\217[p\253\267k\354B\320/\024\214\236\217\020\214\237\217I\217\240\217\020\215a\n\225\262i@!\030\307\024J\306,\032\313\024*\315\314\014n\265:n\317B\267\307\352vY\260\305\003.\276\224\004\266\361\340\002\323x\210\264e<\270\3100ILJ;\237@`\226$F\302*2J\022\223fE'}\243\333n\361I\304\033\270Y\030\306\006\261q\316\340T\017\275zl3\267!d\344Xo\243G\367\250\271!\260\036\007\013\214\307\341i\333q\260\310t\tD\252\017\036^`\270\004B\314(2[\002\221fl;..\322pM\006\317\334Yf_\261\006\306T\033mn\300]\253\353\260\310\337\367\244C\022/v\214\304\343\030\336\t?_\036\\`G\017\2216\244\007\027Y2\211Iu\314'\020\3302\211\221\260\212\254\231\304\210X?\374\200_\304\335G\270\270\373\210\020v\037\341\222\356\3070bEd\335\217a$\254\222\356\3070\"\326\333\267\304=G\270\270\347\210\020\366\034\341\222\236\3070\022\035$=\217a$\254\222\236\3070\"\326<\373#\356|^\322\367\274\270\353yY\317\363\232\216\347e\375\316k\272\235\227\365:\257\351\264\270\277\222\356\212{+\353\254\246\257\262\256jz*\353\250\246\237V\227\355\217\266\3331\032\216\021\372\371^n\312OP\371\276N\263mo\n\254\302\301\002\263px\332.\034,2L\002\221\322\325\303\013L\223@\210\031E\306I \204\214\212N\253:\256\350\274\322\000Y\214\2404D\026c(\r\2423\212\201\352\241x\274~\323\356\266x\314\321\252y\036\t\006#\246\347\2003\r\020\3332;\014\343;\303Fx\3574p\246\265\230\250n\333\260l\374\026[\214\247gt\254V\257k\332\3336\266\020b\233-$\3673\300v\275\321\262\255&s\353m\353\211wa\236\264\215\2446c\302+W\231\345\315\320\251w\255=\3271\360\241\323\363.~F\236}\330\203Dj\220Y\235\360\001\355\365o:f\235\201:f\327\262=W\275\303R\263\035\213\245\370\353{<\230`@\313\350\266\333.\336\330\na7\327\354\362G\346YX\235v\3070:]\213\2111\230\r:$\356\233""\371\317A>~d\335No\317\250\231h!\303o\263\370\206\215\007\177\352Z\016i\272^\333\367\354X\223\037$\360\026\261;\215\3326Z9H4&\300\243t\205\000\021\244\010\022(\t8\226QN\340F\340\035b6\275q\334!\r+\262\026X\036\243m7{]\253\306\307w\224\262L\245.B\200\265g\030\216$\304ud1\256#\tr\035i\224\353\350\302\\G\032\347:\272@\327\221F\272\216.\324u\254\346\206c\371\256i\270\210X\236&\204\206vgP\236\350\361\027\021\266\352\326\206\311'O\230\3761\004\251 g\213\035\345\010\314\313\341\"\363r\204\300\274\034.4o\002\223\356\250G 2o\002#a\025\2327\201I\263\342\232`'7\354\336\356x+st^\026%c\240'f\203E^n#j#\334\n\330\037\221\371\030Xd=\006\027\030\217\201\205\266\213#\004\235x*\016\334\023\0101\243\320nqD\232\261M\272\010\356\326\360\177\266m\343\225\355\014|V\261K\243\206\337\205\255v\335\213\240|\036\322\301\220\253\207\001l\314H\256\377\307\360s2<\374\342\000\2267\351\266{\274\315\"6\217 |\014\246\275w\332\212;\037o\360\017\no\361\357\to9\235&\216\333\007\365\310\203\267c\373\020\246SL\037\226:\262\235&[\"\276\260\330\313\202\207t\2765\211N\2554F\202S\301 \262\250\222H\303J\"\213+\211<\260$\332\310\222\310CK\242\215-\211<\270$\332\350\222\310\"K\"\r-\211,\266$\362\340\222h\243K\"\017/\2116\276$\362\000\223h#L\302#1\"\0134\2114\322$\262P\223\310cM\242\r6\211<\332$\332p\223\310\343M\242\r8\211$\332$\262p\223H\342M\"\r8\211.\342$\322\220\223\350bN\"\r:\211.\352\0146\013\371\236\242\332N\342\005\017!e\2309\214C\320%%\035\004[\304\261BY<+\316/\316\256e\327\333\335Xv\214\245\313\374\366n\303z\262g\356\361\210\221m\304^\233\245\320\275\026\333\373\366p\337\334\264z\021\252^H\325\013\251\026{\013^m\301\215kh\236\233\271\340\302\002\212d+\347\373\275\333\215\216Qo8\334/\277\016q\t\270\310^\2008q\361_\311B\241\363\250\240\221\365\222\336rn6\277\305\277\363\302\021\020\313M\377\367\002o:\373K\376_\374\027\357\251\266\0232\362\263\r\235|\346\214.Z9\304f\376\253U\317<\314\014qM1>\001^\263\360\322t\374\242\341J1e\3256\303\274J\220F&\245T4'=\263B#\256\021\361\0319\222kS\307y\206\316\004\244Z\026""\330m7\352\252\251\020\3425\203\232\246\323\0311\344\3100\254\002\332\2339\337\353x\341\267\235\213\345\305\231\317\274%g\022\345Cb6\311\376\216\344+^P\2148c\001~\315\010\244\3033\210\205o\220*.\001U\t\t\310rC ?Z\007\351\3112\310NMA\032\260\2014\216\001\251c\0172\207\027$\007\021 \313H\203*]\r\322\304\031Hs> Kg\200<f\005y,\007\362\350\006\244n?(\312\211@Q3\004\312\242!\310R5\004\312\262!\310R7\004\312\302!\310R9\004\222D(H\362\235 \315w\202.\337\t\322|'\350\362\235 \315w\202.\337\t\322\202\030\220\326\275\200\242\360\005\364\225/\240(}\001}\355\013(\212_@_\375\002\262\"\027\220\025\263\200\274\232\005\264\345, \257g\001mA\013\310+Z@[\322\002\222r\024\220\324\235\200\264\360\004t\225' -=\001]\355\tH\213O@W}\002\262\234\036\3102w \317\334\2016s\007\362\314\035h3w \317\334\2016s\007\262\314\035\310\362s \317\317\2016?\007\362\374\034h\363s \317\317\2016?\007\262\314\034\310\362o \317\277\2016\377\006\362\374\033h\363o \317\277\2016\377\006\222\374\033H\322l M\263\201.\315\006\3224\033\350\322l M\263\201.\315\006\342\352\016\020\327q\200\254\220\0034\225\034 +\345\000M-\007\310\2129@S\315\001\222b\005\220T#\200\264\024\001tu\010 -B\000]\005\002H\313\017@W{\000\212\212\014P\224]\200\262\356\002\262\024^\200\262\362\002\262\224^\200\262\366\002\262\024_\200\354\360\033d\207\334 ?\345\006\35517\310\317\271A{\320\r\362\223n\320\036u\203\354\030\032d\307\315 ?o\006\355\2013\310O\234A{\344\014\3623g\320\036:\203\344\274\030$\007\303 =\031\006\335\3210H\317\206Aw8\014\322\323a\320\035\017\203\364x\023\244\247\230\2408\306\004\3759&(\0162A\177\222\t\212\243L\320\237e\202\3640\023\244g\226\2408\264\004\375\251%(\216-A\177n\t\212\203K\320\237\\\202\364\320\022\244g\223\2408\234\004\375\351$(\216'A\177>\t\212\003J\320\237P\202\354\210\022dG\221 ?\213\004\355a$\310O#A{\034\t\362\363H\320\036HB\264\266\002R\325\026 .F\202\260\n\003RE\367\020)\246\007^\237\001\2746\003F\005\0330\252\330\200T\001\006\010\2524 \375\373\034\210\377@\007R\277\320\201\330Ot@\361\333*\360\252\033\201\375""\221\3766\013\274BH\340\277\243\003\357\227X\020\371Q\034$\177\215\000\321\252\026H\224\265\200\360\347t\300~\007\002\211\":H\027\321\200\242@\t\222?y\000\3119/$\177K\002\311\002F\220T0\202\240d\212\037\265L\255\370C\022\034\006M\255xC\0209\034\312\241\001\0347\347\345\215\203v\370\327\t\204\251\342\000w#\202\034\375h\003\341\376AP\354/a\210\222\033\321*#\216\210\237\245\t\371\303\323\2568\370fn\361\221QZ(\317\362\013\223\025{\0165\r2\3605R7+^\261\247\033\036p\276\004\271\\\367\270p\241T\303\260w\215 \277\310\004\304\001\347\320\367\254\222\263j\314\234E\306\307\357\347\320O-'\2636\361\367\ti\274T\037\223\356\267\316\241\265^VV\315\371@0F\257q\016\235t\222\262j\304\322h\214\217\337\317\241\217ZNVm\030\360Ft\357y!]TR\316\240I\364\"\336\265B\351\336\336\225\334\354T\242_H\340\377\265\036\nc\240\307\035\030\2257\3177<Zig\030&v\206\355\313\342\315\363i\246\225v\006\315\330\241\254/\2137\317\247\231V\332\0314\313\207\242\362\347\326K#+\253Vl\352\006S\370<\032\251\345d\325\206g\r8#o\234C\037\235\244\314\032yEC\301\0274\372x\036\355\316 5\273\246\214~\304y>\375t\262\262jE\"\333\016y\t\273X6yg\321.\330z\310K\330\311\262\311;\213v\301\366C^\302n\226M\336Y\264\313\217\204\345\317\257\233N\232F\263\275\304\017\263R\237\340=\021.^\027\250\214\252x\005\320\215\214D\236\322\021\340\365\260F+i\225Q}\231\357>\335\270.\200%PI\017B\004\225\206}\0019\211\177\340\242EMq%\323\344\242\362'\021\217g\217\370\323\315\034\277g\031\340L|/\252W \376F\312n\321\367\\\017\210\202!\215=\216F:\200\307g\200\037VGq\241\376\0343\372\213\021\323\331\034\001\216\247m\"\360d~'\215\212sH\352\317\342k(\312b\004\245i1\033\371)\254(\310\313\\\205\202TGVi\"A\3520M$L#\252\310\322\211=\001\265(\275\250\"\313\"T\230vT\221e\021*\252\005\013\251TEai\"\225\271\325eb*2E'\324\205c*\262,B\325\346\316TS\226\246V\232[v\034\031'P\231Y~<)#Q(-?\256\224\221\350\204\251M\252=\306\214S\252gn\252\2425\276\031%\222\336!R^$\227$Q\r\203\252lNN\244\350\260\252\220NN\244\027\250\036\220\0145vIZ\345\240H\013\362\022\024*\323*J\364""\2444\n\265\025E{R\032\2558\265U\365\365|\tR\245Me\265\177q\002\225E\345\325\2002\022\205\306\362\372@\031\211N\230\332\230\332\322\3018\245\322\224\222J\246\030^eHim\223\204B\241\254\264\332IB\241\021\245\266\241\256\020*F\250\260\240!\370\235\201\000\365A\002\032;\221\214\276K\343wH\353\035\022\024\332!\311\352n(\352 \2444Zq\332\241\311\352b\310\177*\020%\021\227R$(46\225\225VHi\324Z\313J-\2444ZqZ\233jJ0\022\244:\233\212\2535\022\024\032\233\312\2527\2444j\255e\325\034R\032\2558\255M5U\036\tR\235M\363:\223\346u\026\315\353\r\232\317l\317\274\336\234\371\314\326\314\353\215\231\317l\313\274\332\224\231\202em\240\234!H\316\036 g\010\216\263\007\306\031\202\342\354\001\261\332\224\322\"\312\004\205\312\230\212\262J)\215BiE\241\245\224F+NmQ}\rf\202TmSm\310\353hc^'C\320\353d\217z\235\014a\257\223=\356u2\004\276N\366\310\327\321\207\276\262\032\3308\201\322\242\322\252X\031\211Jci\235\254\214D'LcL]\tm\234RiJ\242\367B\211\336\r%Y\374Pr\006G\224d\361D\311\031\\Q\222\305\027%gpFI\006o\224\350\335Q\242\367GI\026\207\224\234\301#%Y\\Rr\006\237\224dqJ\311\031\274R\222\301-%z\277\224\350\035S\222\3053%gpMI\026\337\224\234\3019%Y\274Sr\006\367\224d\360O\211\326A%Z\017\225dpQIv\037\225dpRIv/\225dpSIv?\225\350\035\325t=t\210\022C\323\325\320\351\343\250x%\365(9\254(P\016R\023\361\263\250\3041T\264\356:\225r\226L\211d\255\361H\333t\321q\360\332T\201q\310$\2534\216eV\202\212\350\0300R\031\235\374\207\314\204E\316\202\177\355,\255(\257\246\216\016X\274\272y\364\017\256\021\r&U\000\23580\344\365\337\2519\220>\250\364\353\314\303\321\364\212\310\303\277\265\226W\226\007O\265\0302\234\036\212\311\201\242\237\215\235\276\366\213\323\213\227\351\345\367\207ct\262p\374\346\361\356\311*\255\030\324\250\323\3726\335v\250\343\236^\370\325\301\330\301\345\030\355\360\255\243_\320{\217\350\243\307B\354\333G\367\217\077\077\331\243\017\221\377\373\237\000~\030[\031\307\333\312\370\352x\202\376\340/\207c\364\312\344\360\362\010~\341\335\301\344`fxe$\222\001f\351\207\017N\256\370T\024\020tz\341""\322\376\307\007\377>\234<,Px\273\317\237?\355\217\215\330P*\375\340\336\361;'\257\321\245\207\364\341\006\335\300\036ui\367)}\372mJ\355\220\370\347'\033\264\3725\375\332\242V\2236\t%\337\323\357\177\220\223\277z\262KW\326\351\372&\335\264\251\275Kw\237\244h\377\367p\222\376\365\243\341*\275[\241\225/\350\027[t\013I\333):\354\023\275\371\311\321\314\361\265\260\243\314\034\3305\372\307\233\203\031\372\001\032 A={\274\223\022s\367h\222~R:\231\241\313\217\351c\354\010\276\013\207\341;\264\377\367cE6\014\305\361\231q\005\327\227\364K\354\017\332\351\t\322\356\215\025\030Ka|\216\335\346\306\037\2508\361ehd4\357\217H\373\257\200e\211\335\226\306\227\307e\035\236?\371\003\255\374\233\376{\023\351\266\306f\030\371\314x\231\335\312\343_\260\333\027\343Mvk\216\267\306\303\t\360\301\020mt\211^\272rx\371\360\352 \"\024\021\364\303\251\343\333'o\321\305U\272Z\2435\264w\233\266\367PB/\350\315\275\2642!\337\357\370l\301\341\377\206~\263Cwz\264\367TN\374{\272\274F\327p\360q\256d\260r\310\207=\306\325\323\240\r\034\035\331\034\013\211\377L\037~E\277B]pzy\313\3117\322\254\342\ro\262\211iR\023G\004{\377\204>\331K\330\016\027\317\027\207c\207\177\032l\014\013#\376\017\207\223\364\326\364\361\314\311Un\006\334\006PI\331\242\t\211\257q3\204+\354[\324\356\273\261i\246\344\364xq\\\316w\235\233\001\007\250E[.u\277\243\337}\237\215\030-\361\003\375\341G9\361\rZ\305\351\217C\210\235GboF\336g\032\335\307\211)\347{\227\256\340\022\305\321Ds\353\246L\3104A\327p\376\262%C\237\374\200\324?\216\371oZd\267E\\\002)\336[\303\002\375\250x\274z\222\347\275\362f\220l&\204\304\377\340\275\n\247\215\367\256Y\366\222Y\354\234\234\357c\336+\034\234\016\355\364\220\372\351\330=\306to|J\301t\227\256\342^\200\306\306\216yLS\214ij\374sv\373|\274\244\340\375'}\204\037\021\\\021\236\r\375\267=`\267\007\310\037NEd\340\313\370\177\016\257\034\336\031D\266\267\217\206K\364\366\314\361\306\311\034W\335\033H\335\200\204L\237\363\025\223m\347\013\231J|\024\321Bh\331\254L\264\344m\227{t\317[\231>\313<\273\315\343.\026""\331\310o\014\256\014n\017#Bn\017Mzg\366x\227\316\243\262\270Q\341V\343\215\250\277\274\027\330ma\374!\273=\034_Kk\021\0108\371\222\233;\272C\370\223\302\027\2618\2628\362p\213\377\355p\351ps\2604\022vg\270C\363\367O~N\313h\005\\\367\331\367\361\200\363\244\311?8\272O\007\005\034\320/\371\016D\377\033y\217^=\352!\263\311l\304\346A\036?\013\220?\032\363\366\250\257\274\231\021@\350\245w\016\315\303'\003\223\302?\216.\217|\200\331\301\344H!\304\320\217\037\234\274\303\277\0008@\222=\343\302\037\373\205\376*J\177g\310\334\206\013\177=,\034\256\016\356\014\337A\215v\217\227\002\320C\364F\336;\332;~r\262M\037\3410y~\021.u\317<\205\361@\324\332\341\022\205\217\217\374\315\325@\277\344o\370\331?\372\364d\354Y\341\364\342\257\016.\034\024\016\326\372\325\303\313\370\370\356'G\205\237~\005\257\275\276\177k\3379\270r\372\372\257\017n\240\210\245\323\213\177\354\337?\274?(\014\260\371\333\203\331\376g\364\306gG;\247\027\377@\377\300\334\244\327\177s\360\021\375\335G\303\345\341\356\021R\374r\177\027E.\367\367\230IN\377\214\352\236^\274\264\377\217\203\351\376\317\372K\375\215\303\302s$\331yV\370\351\rx\355O#\301\227\360\255O\373\227\373W\017\337<|:|s\270\303\336\177\275?\331/x\270\335\203J\177,\242\023\203=9\250\365/\307`\351\327\234^|}\377\301\201yz\361\255\203\037p~a\343\277\366\227\366\355~\361\360\3678\371##\204\275\247\237\3163_\323[?\336\324\325\355\322#.o\212\342B}\212\264\337\006\237\232\022\273\225p\361\2053\355\346\300\333_\376\324_\242\177\236`\263\246x\274\304!W\017w\350\265\374\321\317\217\266\216\371\314\2736\030\033\274=|pd\035\337=\331\242\253\017\031,\207\337\307\267\007\017\206[HTg\037\271\312H\027\034gz\367\363\223\333ti\215\357\033\3316\327\021\227\347\234\341\246\206\337=\361G\235B\371\304[\253\177\037\354\014_\035>=\346\223\026\273\342\337\372\2338]?=\272x\374\352\261\313\276#\325gc?]\204\237\275\261_=\300\225\361\372\376\354\301\344\301L\377&\016\303\205W\237\355\356?<(\366\177\313F\356\3278\337p\204v\236\343\305\355\337a\376\023\243\300I\366K\372\313w""\007\267\370\333\276=\276zr\371y\034r\345e\274\340\372\340\267\203\245\301\346\360\013T{\347y\0342\3662^\360\336\240:\274<\274z\364\366\361\354\311\344\3638\244\3602^01\250\017\257\014\357\034\275{\374\360\244\360<\016\301\035\000\367\234\207|\330\376\352\355\000\217\217\366\350,\372Rkl\330\330\310\276\305\243\224\013\277\353\217\321\337_\217\372\256\270\265\321\253w\206\337\037\337\247%\374\342y\237;\377\213\262\312n\253\343\353\354\266>\376\257W\330Ry\245\314n\345W\036\263\333\343W\266\331m\373\225\346+\251i\367\317#\223~\266\340{\204\350k\204_\032\255\2372\342D\217x\233y\341\301\354\366\2770>\353\032\273\255\215?JK\340\235\372\0247\253\321\276\313\266\360]\376\355\341\313\356\315\301\316\263\261\347\027\376\322\337\241\271\367\207\227\216\356\036o\323\245\345\323\013\2779\370\364\360\"\342\271\351\277;\270\323\3071\301\245@\337\377\327\361\033|e\354OJ\370\362\375<\213\013#|\354\343{wh\037\027Tl\037\263-\334\233\013>\033\356\037\203\257\216&N\306Tl\237\340\366\2733\030\213\260\341N3\370\361\310>\t\336vz\001\347\022\005\334O\370\304@\031\014\2067\326?&\365\325gO\366k8%/\276\261\337\354_\351O\376\007\260\007\210\013";
    PyObject *data = __Pyx_DecompressString(cstring, 6629, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (32527 bytes) */
const char* const bytes = "<NoneNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.See `flags`.See `nvshmem_pe_index_t`.See `nvshmem_team_id_t`.See `nvshmemx_cmp_type_t`.See `nvshmemx_init_status_t`.See `nvshmemx_proxy_status_t`.See `nvshmemx_qp_handle_index_t`.See `nvshmemx_signal_op_t`.See `nvshmemx_smem_amount_t`.See `nvshmemx_status`.See `nvshmemx_thread_support_t`.See `nvshmemx_tma_policy_t`.Status code >).?add_notedata argument must be a NumPy ndarraydata array must be 1Ddata array must be of dtype uniqueid_dtypedisableenablegcindex is out of boundsint() argument must be a bytes-like object of size 1. To get the pointer address of an array, use .ptrisenableditemsize  mismatches struct size nvshmem/bindings/nvshmem.pyx object at ptr must not be null (0)team_uniqueid(team_uniqueid(value=(tree fragment).uniqueid_Array_.uniqueid object at CMP_EQCMP_GECMP_GTCMP_LECMP_LTCMP_NECMP_SENTINELCmp_typeERROR_COLLECTIVE_LAUNCH_FAILEDERROR_GPU_NOT_SELECTEDERROR_INTERNALERROR_INVALID_VALUEERROR_NOT_SUPPORTEDERROR_OUT_OF_MEMORYERROR_SENTINELERROR_SYMMETRYFlagsINIT_MAXINIT_THREAD_PESINIT_WITH_MPI_COMMINIT_WITH_SHMEMINIT_WITH_UNIQUEIDInitAttrInitAttr.__reduce_cython__InitAttr.__setstate_cython__Init_statusIntEnum_IntEnumMAXNVSHMEMErrorNVSHMEMError.__init__NVSHMEMError.__reduce__PE_ALLPE_ANYPE_INVALID_PROXY_GLOBAL_EXIT_FINISHED_PROXY_GLOBAL_EXIT_INIT_PROXY_GLOBAL_EXIT_MAX_STATE_PROXY_GLOBAL_EXIT_NOT_REQUESTED_PROXY_GLOBAL_EXIT_REQUESTEDPe_indexProxy_status__Pyx_PyDict_NextRefQP_ALLQP_ANYQP_DEFAULTQP_HOSTQp_handle_indexSIGNAL_ADDSIGNAL_SETSMEM_BARRIERS_ONLYSMEM_MINIMUMSMEM_RECOMMENDEDSTATUS_FULL_MPGSTATUS_INVALIDSTATUS_IS_BOOTSTRAPPEDSTATUS_IS_INITIALIZEDSTATUS_LIMITED_MPGSTATUS_NOT_INITIALIZEDSUCCESSSignal_opSmem_amountStatusTEAMS_MINTEAM_GPU_LEADERSTEAM_GPU_LEADERS_INDEXTEAM_INDEX_MAXTEAM_INVALIDTEAM_MC_SHAREDTEAM_MC_SHARED_INDEXTEAM_NODETEAM_NODE_INDEXTEAM_SAME_GPUTEAM_SAME_GPU_""INDEXTEAM_SAME_MYPE_NODETEAM_SAME_MYPE_NODE_INDEXTEAM_SHAREDTEAM_SHARED_INDEXTEAM_WORLDTEAM_WORLD_INDEXTHREAD_FUNNELEDTHREAD_MULTIPLETHREAD_SERIALIZEDTHREAD_SINGLETHREAD_TYPE_SENTINELTMA_DISABLETMA_ENABLETMA_FORCETeamConfigTeamConfig.__reduce_cython__TeamConfig.__setstate_cython__TeamUniqueIdTeamUniqueId.__reduce_cython__TeamUniqueId.__setstate_cython__Team_idThread_supportTma_policyUniqueIdUniqueId.__reduce_cython__UniqueId.__setstate_cython__alignall__annotate__argsasyncio.coroutinesattrbarrierbarrier_allbarrier_all_on_streambarrier_on_streambfloat16_alltoall_on_streambfloat16_broadcast_on_streambfloat16_fcollect_on_streambfloat16_max_reduce_on_streambfloat16_max_reducescatter_on_streambfloat16_min_reduce_on_streambfloat16_min_reducescatter_on_streambfloat16_sum_reduce_on_streambfloat16_sum_reducescatter_on_streambufbuf_ptrbufferbuffer_register_symmetricbuffer_unregister_symmetricbytescallocchar_alltoall_on_streamchar_broadcast_on_streamchar_fcollect_on_streamchar_max_reduce_on_streamchar_max_reducescatter_on_streamchar_min_reduce_on_streamchar_min_reducescatter_on_streamchar_sum_reduce_on_streamchar_sum_reducescatter_on_streamcheck_status__class_getitem__cline_in_tracebackcmpcmp_valueconfigconfig_maskcountcstreamcstrmctypesculibrary_finalizeculibrary_initcumodule_finalizecumodule_initdata_datadestdest_team__dict___dict__doc__double_alltoall_on_streamdouble_broadcast_on_streamdouble_fcollect_on_streamdouble_max_reduce_on_streamdouble_max_reducescatter_on_streamdouble_min_reduce_on_streamdouble_min_reducescatter_on_streamdouble_sum_reduce_on_streamdouble_sum_reducescatter_on_streamdtypeemptyenumerrflagflagsfloat_alltoall_on_streamfloat_broadcast_on_streamfloat_fcollect_on_streamfloat_max_reduce_on_streamfloat_max_reducescatter_on_streamfloat_min_reduce_on_streamfloat_min_reducescatter_on_streamfloat_sum_reduce_on_streamfloat_sum_reducescatter_on_streamfreefrom_datafrom_ptr__func__get_uniqueidgetmem_on_stream__getstate__half_alltoall_on_streamhalf_broadcast_on_strea""mhalf_fcollect_on_streamhalf_max_reduce_on_streamhalf_max_reducescatter_on_streamhalf_min_reduce_on_streamhalf_min_reducescatter_on_streamhalf_sum_reduce_on_streamhalf_sum_reducescatter_on_streamhostlib_finalizehostlib_init_attridinfo_get_version__init__init_attr_dtypeinit_statusint16_alltoall_on_streamint16_broadcast_on_streamint16_fcollect_on_streamint16_max_reduce_on_streamint16_max_reducescatter_on_streamint16_min_reduce_on_streamint16_min_reducescatter_on_streamint16_sum_reduce_on_streamint16_sum_reducescatter_on_streamint32int32_alltoall_on_streamint32_broadcast_on_streamint32_fcollect_on_streamint32_max_reduce_on_streamint32_max_reducescatter_on_streamint32_min_reduce_on_streamint32_min_reducescatter_on_streamint32_sum_reduce_on_streamint32_sum_reducescatter_on_streamint64_alltoall_on_streamint64_broadcast_on_streamint64_fcollect_on_streamint64_max_reduce_on_streamint64_max_reducescatter_on_streamint64_min_reduce_on_streamint64_min_reducescatter_on_streamint64_sum_reduce_on_streamint64_sum_reducescatter_on_streamint8int8_alltoall_on_streamint8_broadcast_on_streamint8_fcollect_on_streamint8_max_reduce_on_streamint8_max_reducescatter_on_streamint8_min_reduce_on_streamint8_min_reducescatter_on_streamint8_sum_reduce_on_streamint8_sum_reducescatter_on_streamint_alltoall_on_streamint_broadcast_on_streamint_fcollect_on_streamint_max_reduce_on_streamint_max_reducescatter_on_streamint_min_reduce_on_streamint_min_reducescatter_on_streamint_sum_reduce_on_streamint_sum_reducescatter_on_streaminternalintp_is_coroutineitemsitemsizelibrarylong_alltoall_on_streamlong_broadcast_on_streamlong_fcollect_on_streamlong_max_reduce_on_streamlong_max_reducescatter_on_streamlong_min_reduce_on_streamlong_min_reducescatter_on_streamlong_sum_reduce_on_streamlong_sum_reducescatter_on_streamlonglong_alltoall_on_streamlonglong_broadcast_on_streamlonglong_fcollect_on_streamlonglong_max_reduce_on_streamlonglong_max_reducescatter_on_streamlonglong_min_reduce_on_streamlonglong_min_reducescatter""_on_streamlonglong_sum_reduce_on_streamlonglong_sum_reducescatter_on_stream__main__majormallocmc_ptr__metaclass__minormmap_ptrmodule__module__mpi_comm__mro_entries__my_pemyrankn_pes__name__ndarrayndimnelem__new__new_teamnpesnranksnreducenum_contexts_numpynumpynvshmem_attrnvshmem.bindings.nvshmemobjpaddingparent_teampatchpepe_idx_in_teampe_rootpe_sizepe_startpe_stridepop__prepare__ptrputmem_on_streamputmem_signal_on_stream__pyx_capi____pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_InitAttr__pyx_unpickle_TeamConfig__pyx_unpickle_TeamUniqueId__pyx_unpickle_UniqueId__pyx_unpickle_team_uniqueid__pyx_unpickle_uniqueid__qualname__quiet_on_streamreadonlyrecarray__reduce____reduce_cython____reduce_ex__schar_alltoall_on_streamschar_broadcast_on_streamschar_fcollect_on_streamschar_max_reduce_on_streamschar_max_reducescatter_on_streamschar_min_reduce_on_streamschar_min_reducescatter_on_streamschar_sum_reduce_on_streamschar_sum_reducescatter_on_streamselfset_attr_mpi_comm_argsset_attr_uniqueid_args__set_name__setdefault__setstate____setstate_cython__short_alltoall_on_streamshort_broadcast_on_streamshort_fcollect_on_streamshort_max_reduce_on_streamshort_max_reducescatter_on_streamshort_min_reduce_on_streamshort_min_reducescatter_on_streamshort_sum_reduce_on_streamshort_sum_reducescatter_on_streamsig_addrsig_opsignalsignal_op_on_streamsignal_wait_until_on_streamsizesize_alltoall_on_streamsize_broadcast_on_streamsize_fcollect_on_streamsize_max_reduce_on_streamsize_max_reducescatter_on_streamsize_min_reduce_on_streamsize_min_reducescatter_on_streamsize_sum_reduce_on_streamsize_sum_reducescatter_on_streamsourcesrcsrc_pesrc_teamstatestaticmethodstatusstreamsupersync_all_on_streamteamteam_config_dtypeteam_destroyteam_get_configteam_get_uniqueidteam_initteam_my_peteam_n_pesteam_split_2dteam_split_stridedteam_sync_on_streamteam_translate_peteam_uniqueidteam_uniqueid.__reduce_cython__team_uniqueid.__setstate_cython____test__uint16_alltoall_on_streamuint16_broadcast_on_stream""uint16_fcollect_on_streamuint16_max_reduce_on_streamuint16_max_reducescatter_on_streamuint16_min_reduce_on_streamuint16_min_reducescatter_on_streamuint16_sum_reduce_on_streamuint16_sum_reducescatter_on_streamuint32_alltoall_on_streamuint32_broadcast_on_streamuint32_fcollect_on_streamuint32_max_reduce_on_streamuint32_max_reducescatter_on_streamuint32_min_reduce_on_streamuint32_min_reducescatter_on_streamuint32_sum_reduce_on_streamuint32_sum_reducescatter_on_streamuint64uint64_alltoall_on_streamuint64_broadcast_on_streamuint64_fcollect_on_streamuint64_max_reduce_on_streamuint64_max_reducescatter_on_streamuint64_min_reduce_on_streamuint64_min_reducescatter_on_streamuint64_sum_reduce_on_streamuint64_sum_reducescatter_on_streamuint8_alltoall_on_streamuint8_broadcast_on_streamuint8_fcollect_on_streamuint8_max_reduce_on_streamuint8_max_reducescatter_on_streamuint8_min_reduce_on_streamuint8_min_reducescatter_on_streamuint8_sum_reduce_on_streamuint8_sum_reducescatter_on_streamuniqueiduniqueid.__reduce_cython__uniqueid.__setstate_cython__uniqueid_dtypeuniqueid.from_datauniqueid.from_ptrupdateuse_setstatevaluevaluesvendor_get_version_infoversionviewxaxis_configxaxis_maskxaxis_teamxrangeyaxis_configyaxis_maskyaxis_teamPyObject *(int, int, intptr_t, intptr_t, int __pyx_skip_dispatch)\000PyObject *(int32_t, int __pyx_skip_dispatch)\000PyObject *(int32_t, int, int, int, intptr_t, long, intptr_t, int __pyx_skip_dispatch)\000PyObject *(int32_t, int, intptr_t, long, intptr_t, intptr_t, long, intptr_t, int __pyx_skip_dispatch)\000PyObject *(int32_t, intptr_t, int __pyx_skip_dispatch)\000PyObject *(int32_t, intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch)\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000PyObject *(int32_t, intptr_t, intptr_t, size_t, intptr_t, int __pyx_skip_dispatch)\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000""\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000PyObject *(intptr_t, int __pyx_skip_dispatch)\000\000PyObject *(intptr_t, intptr_t, int __pyx_skip_dispatch)\000PyObject *(intptr_t, intptr_t, long, int, int, int __pyx_skip_dispatch)\000PyObject *(unsigned int, intptr_t, int __pyx_skip_dispatch)\000int (int __pyx_skip_dispatch)\000\000\000int (int32_t, int __pyx_skip_dispatch)\000\000int (int32_t, int, int32_t, int __pyx_skip_dispatch)\000int (int32_t, intptr_t, int __pyx_skip_dispatch)\000int (intptr_t, int __pyx_skip_dispatch)\000\000\000\000int (intptr_t, size_t, int __pyx_skip_dispatch)\000intptr_t (int32_t, intptr_t, int __pyx_skip_dispatch)\000intptr_t (intptr_t, int, int __pyx_skip_dispatch)\000intptr_t (intptr_t, size_t, int, int __pyx_skip_dispatch)\000intptr_t (size_t, int __pyx_skip_dispatch)\000intptr_t (size_t, size_t, int __pyx_skip_dispatch)\000\000void (int __pyx_skip_dispatch)\000\000void (int32_t, int __pyx_skip_dispatch)\000void (int32_t, intptr_t, int __pyx_skip_dispatch)\000void (intptr_t, int __pyx_skip_dispatch)\000\000\000\000void (intptr_t, int, uint64_t, intptr_t, int __pyx_skip_dispatch)\000void (intptr_t, intptr_t, int __pyx_skip_dispatch)\000void (intptr_t, intptr_t, intptr_t, int __pyx_skip_dispatch)\000void (intptr_t, intptr_t, size_t, int, intptr_t, int __pyx_skip_dispatch)\000\000void (intptr_t, intptr_t, size_t, intptr_t, uint64_t, int, int, intptr_t, int __pyx_skip_dispatch)\000void (intptr_t, uint64_t, int, int, intptr_t, int __pyx_skip_dispatch)\000set_attr_uniqueid_args\000barrier\000team_split_strided\000team_split_2d\000barrier_on_s""tream\000bfloat16_broadcast_on_stream\000char_broadcast_on_stream\000double_broadcast_on_stream\000float_broadcast_on_stream\000half_broadcast_on_stream\000int16_broadcast_on_stream\000int32_broadcast_on_stream\000int64_broadcast_on_stream\000int8_broadcast_on_stream\000int_broadcast_on_stream\000long_broadcast_on_stream\000longlong_broadcast_on_stream\000schar_broadcast_on_stream\000short_broadcast_on_stream\000size_broadcast_on_stream\000uint16_broadcast_on_stream\000uint32_broadcast_on_stream\000uint64_broadcast_on_stream\000uint8_broadcast_on_stream\000bfloat16_alltoall_on_stream\000bfloat16_fcollect_on_stream\000bfloat16_max_reduce_on_stream\000bfloat16_max_reducescatter_on_stream\000bfloat16_min_reduce_on_stream\000bfloat16_min_reducescatter_on_stream\000bfloat16_sum_reduce_on_stream\000bfloat16_sum_reducescatter_on_stream\000char_alltoall_on_stream\000char_fcollect_on_stream\000char_max_reduce_on_stream\000char_max_reducescatter_on_stream\000char_min_reduce_on_stream\000char_min_reducescatter_on_stream\000char_sum_reduce_on_stream\000char_sum_reducescatter_on_stream\000double_alltoall_on_stream\000double_fcollect_on_stream\000double_max_reduce_on_stream\000double_max_reducescatter_on_stream\000double_min_reduce_on_stream\000double_min_reducescatter_on_stream\000double_sum_reduce_on_stream\000double_sum_reducescatter_on_stream\000float_alltoall_on_stream\000float_fcollect_on_stream\000float_max_reduce_on_stream\000float_max_reducescatter_on_stream\000float_min_reduce_on_stream\000float_min_reducescatter_on_stream\000float_sum_reduce_on_stream\000float_sum_reducescatter_on_stream\000half_alltoall_on_stream\000half_fcollect_on_stream\000half_max_reduce_on_stream\000half_max_reducescatter_on_stream\000half_min_reduce_on_stream\000half_min_reducescatter_on_stream\000half_sum_reduce_on_stream\000half_sum_reducescatter_on_stream\000int16_alltoall_on_stream\000int16_fcollect_on_stream\000int16_max_reduce_on_stream\000int16_max_reducescatter_on_stream\000int16_min_red""uce_on_stream\000int16_min_reducescatter_on_stream\000int16_sum_reduce_on_stream\000int16_sum_reducescatter_on_stream\000int32_alltoall_on_stream\000int32_fcollect_on_stream\000int32_max_reduce_on_stream\000int32_max_reducescatter_on_stream\000int32_min_reduce_on_stream\000int32_min_reducescatter_on_stream\000int32_sum_reduce_on_stream\000int32_sum_reducescatter_on_stream\000int64_alltoall_on_stream\000int64_fcollect_on_stream\000int64_max_reduce_on_stream\000int64_max_reducescatter_on_stream\000int64_min_reduce_on_stream\000int64_min_reducescatter_on_stream\000int64_sum_reduce_on_stream\000int64_sum_reducescatter_on_stream\000int8_alltoall_on_stream\000int8_fcollect_on_stream\000int8_max_reduce_on_stream\000int8_max_reducescatter_on_stream\000int8_min_reduce_on_stream\000int8_min_reducescatter_on_stream\000int8_sum_reduce_on_stream\000int8_sum_reducescatter_on_stream\000int_alltoall_on_stream\000int_fcollect_on_stream\000int_max_reduce_on_stream\000int_max_reducescatter_on_stream\000int_min_reduce_on_stream\000int_min_reducescatter_on_stream\000int_sum_reduce_on_stream\000int_sum_reducescatter_on_stream\000long_alltoall_on_stream\000long_fcollect_on_stream\000long_max_reduce_on_stream\000long_max_reducescatter_on_stream\000long_min_reduce_on_stream\000long_min_reducescatter_on_stream\000long_sum_reduce_on_stream\000long_sum_reducescatter_on_stream\000longlong_alltoall_on_stream\000longlong_fcollect_on_stream\000longlong_max_reduce_on_stream\000longlong_max_reducescatter_on_stream\000longlong_min_reduce_on_stream\000longlong_min_reducescatter_on_stream\000longlong_sum_reduce_on_stream\000longlong_sum_reducescatter_on_stream\000schar_alltoall_on_stream\000schar_fcollect_on_stream\000schar_max_reduce_on_stream\000schar_max_reducescatter_on_stream\000schar_min_reduce_on_stream\000schar_min_reducescatter_on_stream\000schar_sum_reduce_on_stream\000schar_sum_reducescatter_on_stream\000short_alltoall_on_stream\000short_fcollect_on_stream\000short_max_reduce_on_stream\000sh""ort_max_reducescatter_on_stream\000short_min_reduce_on_stream\000short_min_reducescatter_on_stream\000short_sum_reduce_on_stream\000short_sum_reducescatter_on_stream\000size_alltoall_on_stream\000size_fcollect_on_stream\000size_max_reduce_on_stream\000size_max_reducescatter_on_stream\000size_min_reduce_on_stream\000size_min_reducescatter_on_stream\000size_sum_reduce_on_stream\000size_sum_reducescatter_on_stream\000uint16_alltoall_on_stream\000uint16_fcollect_on_stream\000uint16_max_reduce_on_stream\000uint16_max_reducescatter_on_stream\000uint16_min_reduce_on_stream\000uint16_min_reducescatter_on_stream\000uint16_sum_reduce_on_stream\000uint16_sum_reducescatter_on_stream\000uint32_alltoall_on_stream\000uint32_fcollect_on_stream\000uint32_max_reduce_on_stream\000uint32_max_reducescatter_on_stream\000uint32_min_reduce_on_stream\000uint32_min_reducescatter_on_stream\000uint32_sum_reduce_on_stream\000uint32_sum_reducescatter_on_stream\000uint64_alltoall_on_stream\000uint64_fcollect_on_stream\000uint64_max_reduce_on_stream\000uint64_max_reducescatter_on_stream\000uint64_min_reduce_on_stream\000uint64_min_reducescatter_on_stream\000uint64_sum_reduce_on_stream\000uint64_sum_reducescatter_on_stream\000uint8_alltoall_on_stream\000uint8_fcollect_on_stream\000uint8_max_reduce_on_stream\000uint8_max_reducescatter_on_stream\000uint8_min_reduce_on_stream\000uint8_min_reducescatter_on_stream\000uint8_sum_reduce_on_stream\000uint8_sum_reducescatter_on_stream\000get_uniqueid\000team_get_uniqueid\000set_attr_mpi_comm_args\000team_init\000hostlib_init_attr\000init_status\000my_pe\000n_pes\000team_my_pe\000team_n_pes\000team_translate_pe\000team_sync_on_stream\000culibrary_finalize\000culibrary_init\000cumodule_finalize\000cumodule_init\000buffer_unregister_symmetric\000mc_ptr\000ptr\000buffer_register_symmetric\000malloc\000align\000calloc\000barrier_all\000hostlib_finalize\000team_destroy\000team_get_config\000barrier_all_on_stream\000free\000quiet_on_stream\000sync_all_on_stream""\000signal_wait_until_on_stream\000info_get_version\000vendor_get_version_info\000getmem_on_stream\000putmem_on_stream\000putmem_signal_on_stream\000signal_op_on_streamint (CUlibrary)\000\000int (CUmodule)\000\000int (int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *)\000int (nvshmem_team_t *, nvshmem_team_config_t *, long, int, int)\000int (nvshmem_team_t)\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, char *, char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, cudaStream_t)\000\000int (nvshmem_team_t, double *, double const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, float *, float const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, half *, half const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int *, int const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_t)\000int (nvshmem_team_t, int""16_t *, int16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, long *, long const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, short *, short const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint8_t *, u""int8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t)\000int (nvshmemx_team_uniqueid_t *)\000int (nvshmemx_uniqueid_t *)\000int (unsigned int, nvshmemx_init_attr_t *)\000int (void *, nvshmemx_init_attr_t *)\000int (void *, size_t)\000int (void)\000\000\000void (cudaStream_t)\000\000\000void (int *, int *)\000void (int *, int *, int *)\000void (nvshmem_team_t)\000void (nvshmem_team_t, nvshmem_team_config_t *)\000void (uint64_t *, int, uint64_t, cudaStream_t)\000void (uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void *)\000void (void *, void const *, size_t, int, cudaStream_t)\000\000void (void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void)\000\000void *(nvshmem_team_t, void const *)\000void *(size_t)\000void *(size_t, size_t)\000\000void *(void *, size_t, int)\000void *(void const *, int)\000nvshmemx_culibrary_finalize\000nvshmemx_culibrary_init\000nvshmemx_cumodule_finalize\000nvshmemx_cumodule_init\000nvshmemx_set_attr_uniqueid_args\000nvshmemx_team_init\000nvshmem_barrier\000nvshmem_team_my_pe\000nvshmem_team_n_pes\000nvshmemx_longlong_alltoall_on_stream\000nvshmemx_longlong_fcollect_on_stream\000nvshmemx_longlong_max_reduce_on_stream\000nvshmemx_longlong_max_reducescatter_on_stream\000nvshmemx_longlong_min_reduce_on_stream\000nvshmemx_longlong_min_reducescatter_on_stream\000nvshmemx_longlong_sum_reduce_on_stream\000nvshmemx_longlong_sum_reducescatter_on_stream\000nvshmemx_longlong_broadcast_on_stream\000nvshmemx_bfloat16_alltoall_on_stream\000nvshmemx_bfloat16_fcollect_on_stream\000nvshmemx_bfloat16_max_reduce_on_stream\000nvshmemx_bfloat16_max_reducescatter_on_stream\000nvshmemx_bfloat16_min_reduce_on_stream\000nvshmemx_bfloat16_min_reducescatter_on_stream\000nvshmemx_bfloat16_sum_reduce_on_stream\000nvshmemx_bfloat16_sum_reducescatter_on_stream\000nvshmemx_bfloat16_broadcast_on_stream\000nvshmemx_char_alltoall_on_stream""\000nvshmemx_char_fcollect_on_stream\000nvshmemx_char_max_reduce_on_stream\000nvshmemx_char_max_reducescatter_on_stream\000nvshmemx_char_min_reduce_on_stream\000nvshmemx_char_min_reducescatter_on_stream\000nvshmemx_char_sum_reduce_on_stream\000nvshmemx_char_sum_reducescatter_on_stream\000nvshmemx_char_broadcast_on_stream\000nvshmemx_barrier_on_stream\000nvshmemx_team_sync_on_stream\000nvshmemx_double_alltoall_on_stream\000nvshmemx_double_fcollect_on_stream\000nvshmemx_double_max_reduce_on_stream\000nvshmemx_double_max_reducescatter_on_stream\000nvshmemx_double_min_reduce_on_stream\000nvshmemx_double_min_reducescatter_on_stream\000nvshmemx_double_sum_reduce_on_stream\000nvshmemx_double_sum_reducescatter_on_stream\000nvshmemx_double_broadcast_on_stream\000nvshmemx_float_alltoall_on_stream\000nvshmemx_float_fcollect_on_stream\000nvshmemx_float_max_reduce_on_stream\000nvshmemx_float_max_reducescatter_on_stream\000nvshmemx_float_min_reduce_on_stream\000nvshmemx_float_min_reducescatter_on_stream\000nvshmemx_float_sum_reduce_on_stream\000nvshmemx_float_sum_reducescatter_on_stream\000nvshmemx_float_broadcast_on_stream\000nvshmemx_half_alltoall_on_stream\000nvshmemx_half_fcollect_on_stream\000nvshmemx_half_max_reduce_on_stream\000nvshmemx_half_max_reducescatter_on_stream\000nvshmemx_half_min_reduce_on_stream\000nvshmemx_half_min_reducescatter_on_stream\000nvshmemx_half_sum_reduce_on_stream\000nvshmemx_half_sum_reducescatter_on_stream\000nvshmemx_half_broadcast_on_stream\000nvshmemx_int_alltoall_on_stream\000nvshmemx_int_fcollect_on_stream\000nvshmemx_int_max_reduce_on_stream\000nvshmemx_int_max_reducescatter_on_stream\000nvshmemx_int_min_reduce_on_stream\000nvshmemx_int_min_reducescatter_on_stream\000nvshmemx_int_sum_reduce_on_stream\000nvshmemx_int_sum_reducescatter_on_stream\000nvshmemx_int_broadcast_on_stream\000nvshmem_team_split_strided\000nvshmem_team_split_2d\000nvshmem_team_translate_pe\000nvshmemx_int16_alltoall_on_stream\000nvshmemx_int16_fcollect_on_stream\000nvsh""memx_int16_max_reduce_on_stream\000nvshmemx_int16_max_reducescatter_on_stream\000nvshmemx_int16_min_reduce_on_stream\000nvshmemx_int16_min_reducescatter_on_stream\000nvshmemx_int16_sum_reduce_on_stream\000nvshmemx_int16_sum_reducescatter_on_stream\000nvshmemx_int16_broadcast_on_stream\000nvshmemx_int32_alltoall_on_stream\000nvshmemx_int32_fcollect_on_stream\000nvshmemx_int32_max_reduce_on_stream\000nvshmemx_int32_max_reducescatter_on_stream\000nvshmemx_int32_min_reduce_on_stream\000nvshmemx_int32_min_reducescatter_on_stream\000nvshmemx_int32_sum_reduce_on_stream\000nvshmemx_int32_sum_reducescatter_on_stream\000nvshmemx_int32_broadcast_on_stream\000nvshmemx_int64_alltoall_on_stream\000nvshmemx_int64_fcollect_on_stream\000nvshmemx_int64_max_reduce_on_stream\000nvshmemx_int64_max_reducescatter_on_stream\000nvshmemx_int64_min_reduce_on_stream\000nvshmemx_int64_min_reducescatter_on_stream\000nvshmemx_int64_sum_reduce_on_stream\000nvshmemx_int64_sum_reducescatter_on_stream\000nvshmemx_int64_broadcast_on_stream\000nvshmemx_int8_alltoall_on_stream\000nvshmemx_int8_fcollect_on_stream\000nvshmemx_int8_max_reduce_on_stream\000nvshmemx_int8_max_reducescatter_on_stream\000nvshmemx_int8_min_reduce_on_stream\000nvshmemx_int8_min_reducescatter_on_stream\000nvshmemx_int8_sum_reduce_on_stream\000nvshmemx_int8_sum_reducescatter_on_stream\000nvshmemx_int8_broadcast_on_stream\000nvshmemx_long_alltoall_on_stream\000nvshmemx_long_fcollect_on_stream\000nvshmemx_long_max_reduce_on_stream\000nvshmemx_long_max_reducescatter_on_stream\000nvshmemx_long_min_reduce_on_stream\000nvshmemx_long_min_reducescatter_on_stream\000nvshmemx_long_sum_reduce_on_stream\000nvshmemx_long_sum_reducescatter_on_stream\000nvshmemx_long_broadcast_on_stream\000nvshmemx_short_alltoall_on_stream\000nvshmemx_short_fcollect_on_stream\000nvshmemx_short_max_reduce_on_stream\000nvshmemx_short_max_reducescatter_on_stream\000nvshmemx_short_min_reduce_on_stream\000nvshmemx_short_min_reducescatter_on_stream\000nvshmemx_short_su""m_reduce_on_stream\000nvshmemx_short_sum_reducescatter_on_stream\000nvshmemx_short_broadcast_on_stream\000nvshmemx_schar_alltoall_on_stream\000nvshmemx_schar_fcollect_on_stream\000nvshmemx_schar_max_reduce_on_stream\000nvshmemx_schar_max_reducescatter_on_stream\000nvshmemx_schar_min_reduce_on_stream\000nvshmemx_schar_min_reducescatter_on_stream\000nvshmemx_schar_sum_reduce_on_stream\000nvshmemx_schar_sum_reducescatter_on_stream\000nvshmemx_schar_broadcast_on_stream\000nvshmemx_size_alltoall_on_stream\000nvshmemx_size_fcollect_on_stream\000nvshmemx_size_max_reduce_on_stream\000nvshmemx_size_max_reducescatter_on_stream\000nvshmemx_size_min_reduce_on_stream\000nvshmemx_size_min_reducescatter_on_stream\000nvshmemx_size_sum_reduce_on_stream\000nvshmemx_size_sum_reducescatter_on_stream\000nvshmemx_size_broadcast_on_stream\000nvshmemx_uint16_alltoall_on_stream\000nvshmemx_uint16_fcollect_on_stream\000nvshmemx_uint16_max_reduce_on_stream\000nvshmemx_uint16_max_reducescatter_on_stream\000nvshmemx_uint16_min_reduce_on_stream\000nvshmemx_uint16_min_reducescatter_on_stream\000nvshmemx_uint16_sum_reduce_on_stream\000nvshmemx_uint16_sum_reducescatter_on_stream\000nvshmemx_uint16_broadcast_on_stream\000nvshmemx_uint32_alltoall_on_stream\000nvshmemx_uint32_fcollect_on_stream\000nvshmemx_uint32_max_reduce_on_stream\000nvshmemx_uint32_max_reducescatter_on_stream\000nvshmemx_uint32_min_reduce_on_stream\000nvshmemx_uint32_min_reducescatter_on_stream\000nvshmemx_uint32_sum_reduce_on_stream\000nvshmemx_uint32_sum_reducescatter_on_stream\000nvshmemx_uint32_broadcast_on_stream\000nvshmemx_uint64_alltoall_on_stream\000nvshmemx_uint64_fcollect_on_stream\000nvshmemx_uint64_max_reduce_on_stream\000nvshmemx_uint64_max_reducescatter_on_stream\000nvshmemx_uint64_min_reduce_on_stream\000nvshmemx_uint64_min_reducescatter_on_stream\000nvshmemx_uint64_sum_reduce_on_stream\000nvshmemx_uint64_sum_reducescatter_on_stream\000nvshmemx_uint64_broadcast_on_stream\000nvshmemx_uint8_alltoall_on_stream\000nvsh""memx_uint8_fcollect_on_stream\000nvshmemx_uint8_max_reduce_on_stream\000nvshmemx_uint8_max_reducescatter_on_stream\000nvshmemx_uint8_min_reduce_on_stream\000nvshmemx_uint8_min_reducescatter_on_stream\000nvshmemx_uint8_sum_reduce_on_stream\000nvshmemx_uint8_sum_reducescatter_on_stream\000nvshmemx_uint8_broadcast_on_stream\000nvshmemx_team_get_uniqueid\000nvshmemx_get_uniqueid\000nvshmemx_hostlib_init_attr\000nvshmemx_set_attr_mpi_comm_args\000nvshmemx_buffer_unregister_symmetric\000nvshmem_my_pe\000nvshmem_n_pes\000nvshmemx_init_status\000nvshmemx_barrier_all_on_stream\000nvshmemx_quiet_on_stream\000nvshmemx_sync_all_on_stream\000nvshmem_info_get_version\000nvshmemx_vendor_get_version_info\000nvshmem_team_destroy\000nvshmem_team_get_config\000nvshmemx_signal_wait_until_on_stream\000nvshmemx_signal_op_on_stream\000nvshmem_free\000nvshmemx_getmem_on_stream\000nvshmemx_putmem_on_stream\000nvshmemx_putmem_signal_on_stream\000nvshmem_barrier_all\000nvshmemx_hostlib_finalize\000nvshmemx_mc_ptr\000nvshmem_malloc\000nvshmem_align\000nvshmem_calloc\000nvshmemx_buffer_register_symmetric\000nvshmem_ptr\200\001\330\t\n\330\010\021\320\0210\260\001\3201A\300\026\300v\310V\320S_\320_d\320dk\320ks\320st\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0210\260\001\260\027\270\n\320BY\320YZ\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0210\260\001\260\033\270H\300K\310x\320Ws\320s}\360\000\000~\001U\002\360\000\000U\002V\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\220\037\240\001\320!1\260\021\330\004\020\220\001\220\021\200\001\330\004+\2501\250F\260!\200\001\330\t\n\330\010\021\320\021+\2501\250G\3203J\310!\330\004\020\220\001\220\021\320\000+\2501\330\004\013\210:\220^\2401\240A\320\000\033\2301\330\004\013\210=\230\001\200\001\330\t\n\330\010\021\320\0211\260\021\3202B\300&\310\t\320QW\320Wf\320fk\320kr\320rz\320z{\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0211\260\021\3202B\300&\310\006\310f\320T`\320`e\320el\320lu""\320u}\320}~\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0211\260\021\3202B\300&\310\007\310v\320Ub\320bg\320gn\320nv\320vw\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021#\2401\320$5\260V\320;S\320S[\320[h\320hn\320no\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021+\2501\320,<\270F\300(\310!\330\004\020\220\001\220\021\200\001\330\004\033\2301\320\034,\250F\3202J\310!\200\001\330\t\n\330\010\021\320\021+\2501\320,G\300q\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021;\2701\320<L\310F\320RZ\320Z`\320`n\320ns\320s|\360\000\000}\001E\002\360\000\000E\002F\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021;\2701\320<L\310F\320R\\\320\\b\320br\320rw\360\000\000x\001A\002\360\000\000A\002I\002\360\000\000I\002J\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021;\2701\320<L\310F\320R`\320`f\320fz\320z\177\360\000\000@\002I\002\360\000\000I\002Q\002\360\000\000Q\002R\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021+\2501\320,<\270M\310\032\320S^\320^g\360\000\000h\001F\002\360\000\000F\002N\002\360\000\000N\002[\002\360\000\000[\002l\002\360\000\000l\002m\002\330\004\020\220\001\220\021\320\0002\260!\330\004\013\320\013!\240\021\240'\250\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\027\320PV\320Vc\320ch\320ho\320ox\360\000\000y\001A\002\360\000\000A\002B\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\030\320QW\320We\320ej\320jq\320qy\320yz\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\031\320RX\320Xg\320gl\320ls\320s|\360\000\000}\001E\002\360\000\000E\002F\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\032\320SY\320Yi\320in\320nu\320u}\320}~\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\036\320W]\320]q\320qv\320v}\360\000\000~\001F\002\360\000\000F\002G\002\330\004\020\220\001\220""\021\200\001\330\t\n\330\010\021\320\0212\260!\3203C\3006\310\026\310v\320Ua\320af\320fo\320ow\320wx\330\004\020\220\001\220\021\320\0002\260!\330\004\013\210:\220[\240\001\240\035\250f\260A\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310'\320QW\320Wd\320di\320ir\320rz\320z{\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310(\320RX\320Xf\320fk\320kr\320r{\360\000\000|\001D\002\360\000\000D\002E\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310)\320SY\320Yh\320hm\320mt\320t|\320|}\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310)\320SY\320Yh\320hm\320mv\320v~\320~\177\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310*\320TZ\320Zj\320jo\320ov\320v\177\360\000\000@\002H\002\360\000\000H\002I\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310+\320U[\320[l\320lq\320qx\360\000\000y\001A\002\360\000\000A\002B\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0213\2601\3204D\300F\310.\320X^\320^r\320rw\320w~\360\000\000\177\001H\002\360\000\000H\002P\002\360\000\000P\002Q\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0214\260A\3205E\300V\3108\320SY\320Yg\320gl\320lu\320u}\320}~\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0214\260A\3205E\300V\3109\320TZ\320Zi\320in\320nu\320u~\360\000\000\177\001G\002\360\000\000G\002H\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0214\260A\3205E\300V\310:\320U[\320[k\320kp\320py\360\000\000z\001B\002\360\000\000B\002C\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0214\260A\3205E\300V\310;\320V\\\320\\m\320mr\320ry\360\000\000z\001C\002\360\000\000C\002K\002\360\000\000K\002L\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0214\260A\3205E\300V\310>\320Y_\320_s\320sx\360\000\000y\001B\002\360\000\000B\002J\002\360\000\000J\002K\002\330\004\020\220\001\220""\021\320\0004\260A\330\004\013\320\013\"\240!\2407\250!\200\001\330\t\n\330\010\021\320\0215\260Q\3206F\300f\310I\320U[\320[j\320jo\320ox\360\000\000y\001A\002\360\000\000A\002B\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0215\260Q\3206F\300f\310K\320W]\320]n\320ns\320s|\360\000\000}\001E\002\360\000\000E\002F\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0215\260Q\3206F\300f\310L\320X^\320^p\320pu\320u|\360\000\000}\001E\002\360\000\000E\002F\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0215\260Q\3206F\300f\320L\\\320\\b\320bx\320x}\360\000\000~\001E\002\360\000\000E\002M\002\360\000\000M\002N\002\330\004\020\220\001\220\021\200\001\330\004*\250!\2506\260\021\200\001\330\t\n\330\010\021\320\0216\260a\3207G\300v\320M]\320]c\320cy\320y~\360\000\000\177\001F\002\360\000\000F\002O\002\360\000\000O\002W\002\360\000\000W\002X\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0216\260a\3207G\300v\310\\\320Y_\320_q\320qv\320v}\360\000\000~\001G\002\360\000\000G\002O\002\360\000\000O\002P\002\330\004\020\220\001\220\021\320\0006\260a\330\004\013\320\013%\240Q\240g\250Q\200\001\330\t\n\330\010\021\320\0217\260q\3208H\310\006\320N^\320^d\320dz\320z\177\360\000\000@\002I\002\360\000\000I\002Q\002\360\000\000Q\002R\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\0217\260q\3208H\310\006\310l\320Z`\320`r\320rw\360\000\000x\001A\002\360\000\000A\002I\002\360\000\000I\002J\002\330\004\020\220\001\220\021\320\000[\320[\\\330\004\013\210:\320\0257\260q\270\007\270y\310\006\310a\200\001\330\004\"\240!\2408\2501\320\0008\270\001\330\004\013\210:\220]\240!\2407\250!\320\0008\270\001\330\004\013\320\013&\240a\240w\250a\320\0009\270\021\330\004\013\210:\220^\2401\240G\2501\200\001\330\t\n\330\010\021\320\0219\270\021\320:J\310&\320PV\320V\\\320\\h\320hm\320mv\320v~\320~\177\330\004\020\220\001\220\021\200\001\330\004\034\230A\230V\2407\250&\260\001\200\001\330\004$\240A\240V\2507\260&\270\007\270v\300Q""\200\001\330\004$\240A\240W\250F\260-\270x\300w\310k\320Yc\320ck\320ks\320sw\320w\177\360\000\000@\002A\002\200\001\330\004\034\230A\230X\240Q\320\000:\270!\330\004\013\210:\220_\240A\320%5\260V\270=\310\001\200A\330\010\020\220\004\220A\220X\230T\240\021\200A\330+<\270A\360\020\000\t\014\2104\210s\220!\330\014\022\220*\230A\230Q\330\010\034\230H\240H\250A\250Q\330\010\024\220G\230?\320*?\270q\330\010\032\320\0321\260\021\330\014\023\2205\320\0305\260R\260v\270Q\330\010\017\210v\220X\230R\230x\240w\250a\330\036$\240A\330\010\013\2109\220D\230\005\230Q\230f\240A\340\010\017\210q\200A\360\016\000\t\035\230H\240H\250A\250Q\330\010\013\2104\210z\230\021\230'\240\026\240z\260\026\260q\330\014\022\220)\2301\230A\330\010\013\2104\210v\220S\230\001\330\014\022\220*\230A\230Q\330\010\013\2104\210w\220c\230\021\330\014\022\220*\230A\230Q\330\010\013\2109\220D\230\005\230Q\230f\240A\340\010\017\210q\200A\330\010\014\210J\220a\330\010\027\220~\240Q\240a\330\010\r\210Q\210n\230E\240\031\250!\2501\200\001\330\t\n\330\010\021\320\021<\270A\320=M\310V\320S\\\320\\b\320bq\320qv\320v\177\360\000\000@\002H\002\360\000\000H\002I\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021<\270A\320=M\310V\320S^\320^d\320du\320uz\360\000\000{\001D\002\360\000\000D\002L\002\360\000\000L\002M\002\330\004\020\220\001\220\021\320\000,\250A\330\004\013\320\013\035\230Q\320\036.\250a\320\000E\300Q\330\004\013\320\013'\240q\320(8\270\006\270h\300a\200\001\330\004(\250\001\250\033\260J\270e\300;\310h\320VW\200\001\330\004 \240\001\240\033\250J\260h\270h\300d\310(\320RS\200\001\330\t\n\330\010\021\320\021:\270!\320;K\3106\320QX\320X^\320^k\320kp\320py\360\000\000z\001B\002\360\000\000B\002C\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021:\270!\320;K\3106\320QZ\320Z`\320`o\320ot\320t}\360\000\000~\001F\002\360\000\000F\002G\002\330\004\020\220\001\220\021\320\000N\310a\330\004\013\320\013/\250q\260\007\260z\300\021\200\001\330\004\035\230Q\200\001\330\004\035\230Q\230g""\240V\250=\270\010\300\007\300t\3108\320ST\200\001\360\010\000\005\016\210T\220\021\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2307\240'\250\021\330\004\007\200q\330\010\017\320\017+\2504\250q\260\007\260{\300'\310\021\340\010\017\320\017+\2504\250q\260\007\260{\300!\200\001\360\010\000\005\016\210T\220\021\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2307\240'\250\021\330\004\007\200q\330\010\017\320\017)\250\024\250Q\250g\260[\300\007\300q\340\010\017\320\017)\250\024\250Q\250g\260[\300\001\200\001\360\010\000\005\016\210T\220\021\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2307\240'\250\021\330\004\007\200q\330\010\017\320\017-\250T\260\021\260'\270\033\300G\3101\340\010\017\320\017-\250T\260\021\260'\270\033\300A\200\001\360\010\000\005\016\210T\220\021\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\2307\240'\250\021\330\004\007\200q\330\010\017\320\017.\250d\260!\2607\270+\300W\310A\340\010\017\320\017.\250d\260!\2607\270+\300Q\320\000V\320VW\330\004\013\320\013$\240A\320%5\260Z\270x\320GW\320WX\200\001\330\004\020\220\001\220\027\230\001\200\001\330\004\030\230\001\320\031)\250\021\200\001\330\t\n\330\010\021\320\021&\240a\320'7\260}\300H\320Lj\320jx\360\000\000y\001E\002\360\000\000E\002V\002\360\000\000V\002b\002\360\000\000b\002@\003\360\000\000@\003N\003\360\000\000N\003Z\003\360\000\000Z\003k\003\360\000\000k\003l\003\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021>\270a\320?O\310v\320Ua\320ag\320gy\320y~\360\000\000\177\001H\002\360\000\000H\002P\002\360\000\000P\002Q\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021>\270a\320?O\310v\320Ue\320ek\360\000\000l\001B\002\360\000\000B\002G\002\360""\000\000G\002P\002\360\000\000P\002X\002\360\000\000X\002Y\002\330\004\020\220\001\220\021\200\001\330\t\n\330\010\021\320\021&\240a\320'=\270Q\330\004\020\220\001\220\021\200\001\330\004&\240a\240v\250Q\200\001\330\004(\250\001\250\026\250q\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220=\240\010\250\001\250\021\330\004\007\200|\2207\230!\330\010/\250q\3200@\300\016\310a\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\2208\2308\2401\240A\330\004\007\200|\2207\230!\330\010*\250!\250;\260n\300A\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220:\230X\240Q\240a\330\004\007\200|\2207\230!\330\010,\250A\250]\270.\310\001\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220<\230x\240q\250\001\330\004\007\200|\2207\230!\330\010.\250a\250\177\270n\310A\330\004\013\2101\200\001\330\004\027\220q\320\000 \240\001\330\004\013\320\013\037\230q\200\001\330\004\037\230q\240\010\250\001\200\001\340\004\007\200w\210c\220\021\330\010\016\210l\230!\2301";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 499; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 40) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 499; i < 598; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 598; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 499;
      for (Py_ssize_t i=0; i<99; ++i) {
        #if PY_VERSION_HEX >= 0x030F0000
        PyUnstable_SetImmortal(table[i]);
        #elif CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        if ((PY_SSIZE_T_MAX <= _Py_IMMORTAL_REFCNT_LOCAL)) break;
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], ((Py_ssize_t)_Py_IMMORTAL_REFCNT_LOCAL + 1));
        }
        #else
        if ((PY_SSIZE_T_MAX < _Py_IMMORTAL_INITIAL_REFCNT)) break;
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1,48,124};
    int32_t const cint_constants_4[] = {175497610L};
    for (int i = 0; i < 5; i++) {
      numbertab[i] = PyLong_FromLong((i < 4 ? cint_constants_1[i - 0] : cint_constants_4[i - 4]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<5; ++i) {
      #if PY_VERSION_HEX >= 0x030F0000
      PyUnstable_SetImmortal(table[i]);
      #elif CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      if ((PY_SSIZE_T_MAX <= _Py_IMMORTAL_REFCNT_LOCAL)) break;
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], ((Py_ssize_t)_Py_IMMORTAL_REFCNT_LOCAL + 1));
      }
      #else
      if ((PY_SSIZE_T_MAX < _Py_IMMORTAL_INITIAL_REFCNT)) break;
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 4;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 110};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_obj};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_from_data, __pyx_mstate->__pyx_kp_b_iso88591_A_HHAQ_4z_z_q_1A_4vS_AQ_4wc_AQ_9, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 128};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_ptr, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_readonly, __pyx_mstate->__pyx_n_u_obj, __pyx_mstate->__pyx_n_u_flag, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_data};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_from_ptr, __pyx_mstate->__pyx_kp_b_iso88591_A_A_4s_AQ_HHAQ_G_q_1_5_5RvQ_vXRx, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_Qg_q_Qg, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_avQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_Qg_q_Qg, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_avQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_d_7_WA_d_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_1F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_T_G1_T_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_Qg_q_Qg, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_avQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_G1F_a_vWE_Q_q_t7_q_4q_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 411};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_status, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_Ja_Qa_QnE_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 416};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_reduce, __pyx_mstate->__pyx_kp_b_iso88591_A_AXT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 420};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_check_status, __pyx_mstate->__pyx_kp_b_iso88591_wc_l_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 430};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_barrier, __pyx_mstate->__pyx_kp_b_iso88591_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 436};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_barrier_all, __pyx_mstate->__pyx_kp_b_iso88591_q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 440};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_init_status, __pyx_mstate->__pyx_kp_b_iso88591_q_3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 444};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_my_pe, __pyx_mstate->__pyx_kp_b_iso88591_1_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 448};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_n_pes, __pyx_mstate->__pyx_kp_b_iso88591_1_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 452};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_major, __pyx_mstate->__pyx_n_u_minor};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_info_get_version, __pyx_mstate->__pyx_kp_b_iso88591_AV7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 456};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_major, __pyx_mstate->__pyx_n_u_minor, __pyx_mstate->__pyx_n_u_patch};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_vendor_get_version_info, __pyx_mstate->__pyx_kp_b_iso88591_AV7_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 460};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_malloc, __pyx_mstate->__pyx_kp_b_iso88591_1_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 464};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_count, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_calloc, __pyx_mstate->__pyx_kp_b_iso88591_9_1G1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 468};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_count, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_align, __pyx_mstate->__pyx_kp_b_iso88591_8_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 472};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_ptr};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_free, __pyx_mstate->__pyx_kp_b_iso88591__6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 476};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_pe};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_ptr, __pyx_mstate->__pyx_kp_b_iso88591_2_fA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 480};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_ptr};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_mc_ptr, __pyx_mstate->__pyx_kp_b_iso88591_A_5V, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 484};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_my_pe, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 488};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_n_pes, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 492};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_config};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_get_config, __pyx_mstate->__pyx_kp_b_iso88591_1_F2J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 496};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_src_team, __pyx_mstate->__pyx_n_u_src_pe, __pyx_mstate->__pyx_n_u_dest_team};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_translate_pe, __pyx_mstate->__pyx_kp_b_iso88591_VVW_A_5ZxGWWX, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 500};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_parent_team, __pyx_mstate->__pyx_n_u_pe_start, __pyx_mstate->__pyx_n_u_pe_stride, __pyx_mstate->__pyx_n_u_pe_size, __pyx_mstate->__pyx_n_u_config, __pyx_mstate->__pyx_n_u_config_mask, __pyx_mstate->__pyx_n_u_new_team};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_split_strided, __pyx_mstate->__pyx_kp_b_iso88591_1_M_S_g_h_F_F_N_N_l_l_m, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 506};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_uniqueid};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_get_uniqueid, __pyx_mstate->__pyx_kp_b_iso88591_1_Gq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 512};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_config, __pyx_mstate->__pyx_n_u_config_mask, __pyx_mstate->__pyx_n_u_npes, __pyx_mstate->__pyx_n_u_pe_idx_in_team};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_init, __pyx_mstate->__pyx_kp_b_iso88591_1_5V_SS_hhnno, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {8, 0, 0, 8, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 518};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_parent_team, __pyx_mstate->__pyx_n_u_xrange, __pyx_mstate->__pyx_n_u_xaxis_config, __pyx_mstate->__pyx_n_u_xaxis_mask, __pyx_mstate->__pyx_n_u_xaxis_team, __pyx_mstate->__pyx_n_u_yaxis_config, __pyx_mstate->__pyx_n_u_yaxis_mask, __pyx_mstate->__pyx_n_u_yaxis_team};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_split_2d, __pyx_mstate->__pyx_kp_b_iso88591_a_7_HLjjx_y_E_E_V_V_b_b_N_N_Z_Z, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 524};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_destroy, __pyx_mstate->__pyx_kp_b_iso88591__7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 528};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfL_bbxx_E_E_M_M_N, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 534};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 540};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_QWWeejjqqyyz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 546};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmtt, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 552};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 558};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_QWWeejjqqyyz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 564};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_W_qqvv_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 570};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_0_1A_vVS__ddkksst, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 576};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 582};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfLX_ppuu_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 588};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_QWWffkkrrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 594};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 600};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 606};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 612};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 618};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 624};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 630};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 636};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[57] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_alltoall_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_QWWffkkrrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[57])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 642};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[58] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_barrier_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[58])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 648};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[59] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_barrier_all_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_81, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[59])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 652};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[60] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_team_sync_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_EQ_q_8_ha, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[60])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 656};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[61] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_sync_all_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_q_4, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[61])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 660};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[62] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_6a7GvM_ccyy_F_F_O_O_W_W_X, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[62])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 666};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[63] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_PVVcchhoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[63])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 672};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[64] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_RXXffkkrr_D_D_E, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[64])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 678};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[65] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV9TZZiinnuu_G_G_H, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[65])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 684};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[66] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_PVVcchhoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[66])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 690};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[67] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_RXXffkkrr_D_D_E, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[67])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 696};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[68] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_X_rrww_H_H_P_P_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[68])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 702};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[69] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_fT_eelluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[69])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 708};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[70] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_PVVcchhoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[70])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 714};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[71] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_6a7Gv_Y__qqvv_G_G_O_O_P, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[71])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 720};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[72] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_RXXggllss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[72])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 726};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[73] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_TZZjjoovv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[73])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 732};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[74] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_TZZjjoovv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[74])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 738};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[75] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_TZZjjoovv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[75])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 744};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[76] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_TZZjjoovv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[76])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 750};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[77] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_V_mmrry_z_C_C_K_K_L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[77])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 756};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[78] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_V_mmrry_z_C_C_K_K_L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[78])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 762};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[79] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_V_mmrry_z_C_C_K_K_L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[79])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 768};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_pe_root, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[80] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_broadcast_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_RXXggllss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[80])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 774};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[81] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfL_bbxx_E_E_M_M_N, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[81])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 780};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[82] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[82])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 786};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[83] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_QWWeejjqqyyz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[83])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 792};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[84] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmtt, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[84])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 798};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[85] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[85])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 804};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[86] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_QWWeejjqqyyz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[86])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 810};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[87] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_W_qqvv_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[87])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 816};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[88] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_0_1A_vVS__ddkksst, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[88])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 822};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[89] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_vUbbggnnvvw, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[89])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 828};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[90] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfLX_ppuu_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[90])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 834};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[91] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_QWWffkkrrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[91])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 840};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[92] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[92])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 846};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[93] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[93])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 852};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[94] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[94])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 858};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[95] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_SYYiinnuu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[95])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 864};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[96] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[96])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 870};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[97] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[97])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 876};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[98] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_U_llqqx_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[98])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 882};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nelem, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[99] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_fcollect_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_1_2B_QWWffkkrrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[99])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 888};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[100] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[100])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 894};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[101] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[101])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 900};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[102] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[102])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 906};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[103] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[103])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 912};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[104] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[104])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 918};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[105] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[105])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 924};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[106] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[106])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 930};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[107] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[107])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 936};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[108] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[108])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 942};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[109] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[109])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 948};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[110] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_Y__ssx_y_B_B_J_J_K, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[110])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 954};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[111] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[111])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 960};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[112] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_vUaaffoowwx, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[112])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 966};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[113] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[113])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 972};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[114] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_lZ_rrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[114])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 978};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[115] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_N_ddzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[115])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 984};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[116] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[116])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 990};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[117] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[117])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 996};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[118] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_max_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfIU_jjoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[118])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1002};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[119] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[119])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1008};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[120] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[120])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1014};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[121] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[121])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1020};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[122] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[122])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1026};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[123] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[123])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1032};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[124] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[124])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1038};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[125] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[125])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1044};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[126] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[126])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1050};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[127] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[127])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1056};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[128] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[128])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1062};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[129] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_Y__ssx_y_B_B_J_J_K, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[129])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1068};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[130] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[130])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1074};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[131] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_vUaaffoowwx, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[131])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1080};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[132] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[132])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1086};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[133] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_lZ_rrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[133])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1092};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[134] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_N_ddzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[134])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1098};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[135] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[135])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1104};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[136] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[136])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1110};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[137] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_min_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfIU_jjoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[137])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1116};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[138] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[138])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1122};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[139] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[139])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1128};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[140] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[140])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1134};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[141] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[141])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1140};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[142] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_U_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[142])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1146};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[143] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[143])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1152};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[144] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[144])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1158};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[145] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfKW_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[145])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1164};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[146] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_SYYhhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[146])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1170};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[147] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[147])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1176};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[148] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV_Y__ssx_y_B_B_J_J_K, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[148])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1182};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[149] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[149])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1188};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[150] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_2_3C6_vUaaffoowwx, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[150])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1194};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[151] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[151])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1200};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[152] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_lZ_rrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[152])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1206};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[153] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_7q8H_N_ddzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[153])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1212};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[154] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_314DF_QWWddiirrzz, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[154])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1218};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[155] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_4A5EV8SYYgglluu, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[155])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1224};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[156] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_sum_reduce_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_5Q6FfIU_jjoox_y_A_A_B, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[156])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1230};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[157] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[157])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1236};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[158] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[158])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1242};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[159] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[159])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1248};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[160] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[160])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1254};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[161] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[161])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1260};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[162] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_max_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[162])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1266};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[163] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_max_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[163])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1272};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[164] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_max_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[164])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1278};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[165] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[165])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1284};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[166] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[166])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1290};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[167] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_ffzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[167])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1296};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[168] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[168])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1302};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[169] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_9_J_PVV_hhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[169])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1308};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[170] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[170])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1314};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[171] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_max_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUaaggyy_H_H_P_P_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[171])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1320};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[172] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_max_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUeek_l_B_B_G_G_P_P_X_X_Y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[172])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1326};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[173] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_max_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[173])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1332};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[174] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_max_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[174])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1338};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[175] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_max_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_bbqqvv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[175])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1344};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[176] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[176])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1350};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[177] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[177])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1356};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[178] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[178])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1362};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[179] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[179])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1368};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[180] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[180])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1374};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[181] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_min_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[181])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1380};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[182] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_min_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[182])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1386};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[183] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_min_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[183])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1392};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[184] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[184])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1398};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[185] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[185])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1404};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[186] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_ffzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[186])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1410};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[187] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[187])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1416};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[188] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_9_J_PVV_hhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[188])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1422};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[189] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[189])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1428};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[190] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_min_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUaaggyy_H_H_P_P_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[190])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1434};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[191] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_min_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUeek_l_B_B_G_G_P_P_X_X_Y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[191])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1440};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[192] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_min_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[192])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1446};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[193] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_min_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[193])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1452};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[194] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_min_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_bbqqvv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[194])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1458};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[195] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int8_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[195])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1464};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[196] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int16_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[196])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1470};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[197] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int32_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[197])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1476};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[198] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int64_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[198])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1482};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[199] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint8_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_bbrrw_x_A_A_I_I_J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[199])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1488};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[200] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint16_sum_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[200])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1494};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[201] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint32_sum_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[201])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1500};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[202] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_uint64_sum_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_dduuz_D_D_L_L_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[202])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1506};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[203] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_size_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QZZ_oott_F_F_G, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[203])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1512};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[204] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_char_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[204])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1518};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[205] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_schar_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFR_ffzz_I_I_Q_Q_R, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[205])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1524};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[206] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_short_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[206])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1530};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[207] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_int_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_9_J_PVV_hhmmvv, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[207])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1536};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[208] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_long_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[208])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1542};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[209] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_longlong_sum_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUaaggyy_H_H_P_P_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[209])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1548};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[210] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_bfloat16_sum_reducescatter_on_st, __pyx_mstate->__pyx_kp_b_iso88591_a_OvUeek_l_B_B_G_G_P_P_X_X_Y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[210])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1554};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[211] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_half_sum_reducescatter_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_K6QXX_kkppy_z_B_B_C, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[211])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1560};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[212] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_float_sum_reducescatter_on_strea, __pyx_mstate->__pyx_kp_b_iso88591_1_LFRZZ_nnss_E_E_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[212])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1566};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_team, __pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_nreduce, __pyx_mstate->__pyx_n_u_stream};
    __pyx_mstate_global->__pyx_codeobj_tab[213] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_double_sum_reducescatter_on_stre, __pyx_mstate->__pyx_kp_b_iso88591_A_MVS_bbqqvv_H_H_I, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[213])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1572};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[214] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_hostlib_init_attr, __pyx_mstate->__pyx_kp_b_iso88591_1G3J, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[214])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 0, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1578};
    PyObject* const varnames[] = {0};
    __pyx_mstate_global->__pyx_codeobj_tab[215] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_hostlib_finalize, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[215])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1582};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_myrank, __pyx_mstate->__pyx_n_u_nranks, __pyx_mstate->__pyx_n_u_uniqueid, __pyx_mstate->__pyx_n_u_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[216] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_set_attr_uniqueid_args, __pyx_mstate->__pyx_kp_b_iso88591_0_HKxWss_U_U_V, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[216])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1588};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_mpi_comm, __pyx_mstate->__pyx_n_u_nvshmem_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[217] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_set_attr_mpi_comm_args, __pyx_mstate->__pyx_kp_b_iso88591_0_BYYZ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[217])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1594};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_uniqueid};
    __pyx_mstate_global->__pyx_codeobj_tab[218] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_get_uniqueid, __pyx_mstate->__pyx_kp_b_iso88591_a_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[218])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1600};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_module};
    __pyx_mstate_global->__pyx_codeobj_tab[219] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_cumodule_init, __pyx_mstate->__pyx_kp_b_iso88591_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[219])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1604};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_module};
    __pyx_mstate_global->__pyx_codeobj_tab[220] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_cumodule_finalize, __pyx_mstate->__pyx_kp_b_iso88591_6a_QgQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[220])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1608};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_buf_ptr, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_flags};
    __pyx_mstate_global->__pyx_codeobj_tab[221] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_buffer_register_symmetric, __pyx_mstate->__pyx_kp_b_iso88591_7q_y_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[221])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1612};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_mmap_ptr, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[222] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_buffer_unregister_symmetric, __pyx_mstate->__pyx_kp_b_iso88591_Na_q_z, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[222])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1616};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_library};
    __pyx_mstate_global->__pyx_codeobj_tab[223] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_culibrary_init, __pyx_mstate->__pyx_kp_b_iso88591_4A_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[223])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1620};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_library};
    __pyx_mstate_global->__pyx_codeobj_tab[224] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_culibrary_finalize, __pyx_mstate->__pyx_kp_b_iso88591_8_awa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[224])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1624};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_source, __pyx_mstate->__pyx_n_u_bytes, __pyx_mstate->__pyx_n_u_pe, __pyx_mstate->__pyx_n_u_cstrm};
    __pyx_mstate_global->__pyx_codeobj_tab[225] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_putmem_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_QgV_t8ST, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[225])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {8, 0, 0, 8, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1628};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_source, __pyx_mstate->__pyx_n_u_bytes, __pyx_mstate->__pyx_n_u_sig_addr, __pyx_mstate->__pyx_n_u_signal, __pyx_mstate->__pyx_n_u_sig_op, __pyx_mstate->__pyx_n_u_pe, __pyx_mstate->__pyx_n_u_cstrm};
    __pyx_mstate_global->__pyx_codeobj_tab[226] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_putmem_signal_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_AWF_xwkYcckkssww_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[226])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1632};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_dest, __pyx_mstate->__pyx_n_u_source, __pyx_mstate->__pyx_n_u_bytes, __pyx_mstate->__pyx_n_u_pe, __pyx_mstate->__pyx_n_u_cstrm};
    __pyx_mstate_global->__pyx_codeobj_tab[227] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_getmem_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_QgV_t8ST, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[227])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1636};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_cstrm};
    __pyx_mstate_global->__pyx_codeobj_tab[228] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_quiet_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_AXQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[228])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1640};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_sig_addr, __pyx_mstate->__pyx_n_u_signal, __pyx_mstate->__pyx_n_u_sig_op, __pyx_mstate->__pyx_n_u_pe, __pyx_mstate->__pyx_n_u_cstrm};
    __pyx_mstate_global->__pyx_codeobj_tab[229] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_signal_op_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_Jhhd_RS, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[229])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1644};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_sig_addr, __pyx_mstate->__pyx_n_u_cmp, __pyx_mstate->__pyx_n_u_cmp_value, __pyx_mstate->__pyx_n_u_cstream};
    __pyx_mstate_global->__pyx_codeobj_tab[230] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_nvshmem_bindings_nvshmem_pyx, __pyx_mstate->__pyx_n_u_signal_wait_until_on_stream, __pyx_mstate->__pyx_kp_b_iso88591_Je_hVW, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[230])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[231] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_uniqueid, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_881A_7_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[231])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[232] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_UniqueId, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_881A_7_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[232])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[233] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_team_uniqueid, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[233])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[234] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_TeamUniqueId, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[234])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[235] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_InitAttr, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_881A_7_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[235])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[236] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_pyx_unpickle_TeamConfig, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_XQa_7_A_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[236])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        Py_INCREF(src[i]);
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    Py_INCREF(key);
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (size_t) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_size_t(size_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(size_t)*3+2];
    char *dpos, *end = digits + sizeof(size_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    size_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (size_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (size_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (size_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        Py_INCREF(values[i]);
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_mstate_global->__pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyLong_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyLong_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* pybuiltin_invalid (used by pyint_simplify) */
static void __Pyx_PyBuiltin_Invalid(PyObject *obj, const char *type_name, const char *argname) {
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    if (argname) {
        PyErr_Format(PyExc_TypeError,
            "Argument '%.200s' has incorrect type (expected %.200s, got " __Pyx_FMT_TYPENAME ")",
            argname, type_name, obj_type_name
        );
    } else {
        PyErr_Format(PyExc_TypeError,
            "Expected %.200s, got " __Pyx_FMT_TYPENAME,
            type_name, obj_type_name
        );
    }
    __Pyx_DECREF_TypeName(obj_type_name);
}

/* pyint_simplify */
static CYTHON_INLINE int __Pyx_PyInt_FromNumber(PyObject **number_var, const char *argname, int accept_none) {
    PyObject *number = *number_var;
    if (likely((accept_none && number == Py_None) || PyLong_CheckExact(number))) {
        return 0;
    }
    PyObject *int_object;
    if (likely(PyNumber_Check(number))) {
        int_object = PyNumber_Long(number);
        if (unlikely(!int_object)) goto bad;
    } else {
        __Pyx_PyBuiltin_Invalid(number, "int", argname);
        goto bad;
    }
    *number_var = int_object;
    Py_DECREF(number);
    return 0;
bad:
    *number_var = NULL;
    Py_DECREF(number);
    return -1;
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared))) {
            Py_INCREF(v);
            return PyList_SetItem(o, n, v);
        } else if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old;
            Py_INCREF(v);
            old = PyList_GET_ITEM(o, n);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 0;
        }
    } else
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_SetItem(o, i, v);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_SetItemInt_Generic(o, PyLong_FromSsize_t(i), v);
}

/* ErrOccurredWithGIL */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void) {
  int err;
  PyGILState_STATE _save = PyGILState_Ensure();
  err = !!PyErr_Occurred();
  PyGILState_Release(_save);
  return err;
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* FunctionExport */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig) {
    PyObject *cobj;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(api_dict, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(cobj);
    return -1;
}

/* GetApiDict */
static PyObject *__Pyx_ApiExport_GetApiDict(void) {
    PyObject *d;
    if (__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_capi, &d) == -1)
        return NULL;
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_pyx_capi, d) < 0)
            goto bad;
    }
    return d;
bad:
    Py_XDECREF(d);
    return NULL;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_8
#define __PYX_HAVE_RT_ImportType_3_2_8
static PyTypeObject *__Pyx_ImportType_3_2_8(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_8 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_8 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_8 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_8
#define __PYX_HAVE_RT_ImportFromPxd_3_2_8
static int __Pyx_ImportFromPxd_3_2_8(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_8
#define __PYX_HAVE_RT_ImportFunction_3_2_8
static int __Pyx_ImportFunction_3_2_8(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_8(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__4);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int __Pyx_CyFunction_set_annotate_in_dict_if_exists(PyObject *op_in, PyObject *value);
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    if (unlikely(__Pyx_CyFunction_set_annotate_in_dict_if_exists((PyObject*) op, Py_None) < 0)) return -1;
    return 0;
}
static int
__Pyx_CyFunction_get_dict_if_exists(PyObject *op_in, PyObject **dict) {
    /* Return 1 if the function dict exists, 0 otherwise.  This cannot fail:
     * _PyObject_GetDictPtr() may clear errors internally, but never reports them. */
#if CYTHON_COMPILING_IN_PYPY
    *dict = PyObject_GenericGetDict(op_in, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030C0000
    *dict = ((__pyx_CyFunctionObject*) op_in)->func_dict;
#else
    PyObject **dictptr = _PyObject_GetDictPtr(op_in);
    *dict = likely(dictptr) ? *dictptr : NULL;
#endif
    return *dict ? 1 : 0;
}
static int
__Pyx_CyFunction_get_annotate_from_dict_if_exists(PyObject *op_in, PyObject **annotate) {
    PyObject *dict;
    int dict_found;
    *annotate = NULL;
    dict_found = __Pyx_CyFunction_get_dict_if_exists(op_in, &dict);
    if (!dict_found) return 0;
    return __Pyx_PyDict_GetItemRef(dict, __pyx_mstate_global->__pyx_n_u_annotate, annotate);
}
static int
__Pyx_CyFunction_set_annotate_in_dict_if_exists(PyObject *op_in, PyObject *value) {
    PyObject *dict;
    int dict_found;
    dict_found = __Pyx_CyFunction_get_dict_if_exists(op_in, &dict);
    if (!dict_found) return 0;
    return PyDict_SetItem(dict, __pyx_mstate_global->__pyx_n_u_annotate, value);
}
static int
__Pyx_CyFunction_set_annotate_in_dict(PyObject *op_in, PyObject *value) {
    PyObject *dict;
    int result;
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = __Pyx_CyFunction_get_dict((__pyx_CyFunctionObject*) op_in, NULL);
#else
    dict = PyObject_GenericGetDict(op_in, NULL);
#endif
    if (unlikely(!dict)) return -1;
    result = PyDict_SetItem(dict, __pyx_mstate_global->__pyx_n_u_annotate, value);
    Py_DECREF(dict);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(PyObject *op_in, void *context) {
    PyObject *annotate = NULL;
    PyObject *result = NULL;
    __pyx_CyFunctionObject *op = (__pyx_CyFunctionObject*) op_in;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op_in);
    result = __Pyx_XNewRef(op->func_annotations);
    __Pyx_END_CRITICAL_SECTION();
    if (result) return result;
    if (unlikely(__Pyx_CyFunction_get_annotate_from_dict_if_exists(op_in, &annotate) < 0)) return NULL;
    if (!annotate || annotate == Py_None) {
        Py_XDECREF(annotate);
        __Pyx_BEGIN_CRITICAL_SECTION(op_in);
        result = __Pyx_CyFunction_get_annotations_locked(op);
        __Pyx_END_CRITICAL_SECTION();
        return result;
    }
    PyObject *format = PyLong_FromLong(1L);  // annotationlib.Format.VALUE
    if (likely(format)) {
        result = __Pyx_PyObject_CallOneArg(annotate, format);
        Py_DECREF(format);
    }
    Py_DECREF(annotate);
    if (unlikely(!result)) return NULL;
    if (unlikely(!PyDict_Check(result))) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ must return a dict");
        Py_DECREF(result);
        return NULL;
    }
    __Pyx_BEGIN_CRITICAL_SECTION(op_in);
    __Pyx_Py_XDECREF_SET(op->func_annotations, __Pyx_NewRef(result));
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *__Pyx_CyFunction_annotate_impl(PyObject *self, PyObject *args) {
    CYTHON_UNUSED_VAR(args);
    if (unlikely(!self)) {
        PyErr_SetString(PyExc_SystemError, "cython __annotate__ called without 'self' argument");
    }
    Py_XINCREF(self);
    return self;
}
static PyMethodDef __Pyx_CyFunction_annotate_method = {
    "__annotate__",
    (PyCFunction)(void (*)(void))__Pyx_CyFunction_annotate_impl,
    METH_VARARGS,
    "Placeholder __annotate__ function to allow 'functools.wraps' to work "
    "on Cython functions."
};
static PyObject *
__Pyx_CyFunction_get_annotate(PyObject *op_in, void *context) {
    PyObject *annotate = NULL;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(__Pyx_CyFunction_get_annotate_from_dict_if_exists(op_in, &annotate) < 0)) return NULL;
    if (annotate) return annotate;
    PyObject *annotations = __Pyx_CyFunction_get_annotations(op_in, NULL);
    if (unlikely(!annotations)) return NULL;
    PyObject *method = PyCFunction_New(
        &__Pyx_CyFunction_annotate_method,
        annotations);
    Py_DECREF(annotations);
    return method;
}
static int
__Pyx_CyFunction_set_annotate(PyObject *op_in, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ cannot be deleted");
        return -1;
    }
    if (unlikely(value != Py_None && !PyCallable_Check(value))) {
        PyErr_SetString(PyExc_TypeError, "__annotate__ must be callable or None");
        return -1;
    }
    if (value != Py_None) {
        __pyx_CyFunctionObject *op = (__pyx_CyFunctionObject*) op_in;
        __Pyx_BEGIN_CRITICAL_SECTION(op_in);
        Py_CLEAR(op->func_annotations);
        __Pyx_END_CRITICAL_SECTION();
    }
    return __Pyx_CyFunction_set_annotate_in_dict(op_in, value);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"__annotate__", (getter)__Pyx_CyFunction_get_annotate, (setter)__Pyx_CyFunction_set_annotate, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_annotations);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_mstate_global->__pyx_n_u_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base = NULL, *meth, *new_base, *result, *new_bases = NULL;
#if CYTHON_ASSUME_SAFE_SIZE
    size_bases = PyTuple_GET_SIZE(bases);
#else
    size_bases = PyTuple_Size(bases);
    if (size_bases < 0) return NULL;
#endif
    for (i = 0; i < size_bases; i++) {
#if CYTHON_AVOID_BORROWED_REFS
        Py_CLEAR(base);
#endif
#if CYTHON_ASSUME_SAFE_MACROS
        base = PyTuple_GET_ITEM(bases, i);
#else
        base = PyTuple_GetItem(bases, i);
        if (!base) goto error;
#endif
#if CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(base);
#endif
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_mstate_global->__pyx_n_u_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                PyObject *base_from_list;
#if CYTHON_ASSUME_SAFE_MACROS
                base_from_list = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base_from_list);
                Py_INCREF(base_from_list);
#else
                base_from_list = PyTuple_GetItem(bases, j);
                if (!base_from_list) goto error;
                Py_INCREF(base_from_list);
                if (PyList_SetItem(new_bases, j, base_from_list) < 0) goto error;
#endif
            }
        }
#if CYTHON_ASSUME_SAFE_SIZE
        j = PyList_GET_SIZE(new_bases);
#else
        j = PyList_Size(new_bases);
        if (j < 0) goto error;
#endif
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return result;
error:
    Py_XDECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_SIZE
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
        metaclass = &PyType_Type;
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args (used by Py3ClassCreate) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectLookupSpecial (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_mstate_global->__pyx_n_u_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_module_2, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_mstate_global->__pyx_n_u_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_signal_op_t(nvshmemx_signal_op_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_signal_op_t neg_one = (nvshmemx_signal_op_t) -1, const_zero = (nvshmemx_signal_op_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_signal_op_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_signal_op_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_signal_op_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_signal_op_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_signal_op_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_signal_op_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_signal_op_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_tma_policy_t(nvshmemx_tma_policy_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_tma_policy_t neg_one = (nvshmemx_tma_policy_t) -1, const_zero = (nvshmemx_tma_policy_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_tma_policy_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_tma_policy_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_tma_policy_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_tma_policy_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_tma_policy_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_tma_policy_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_tma_policy_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_smem_amount_t(nvshmemx_smem_amount_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_smem_amount_t neg_one = (nvshmemx_smem_amount_t) -1, const_zero = (nvshmemx_smem_amount_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_smem_amount_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_smem_amount_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_smem_amount_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_smem_amount_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_smem_amount_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_smem_amount_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_smem_amount_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_cmp_type_t(nvshmemx_cmp_type_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_cmp_type_t neg_one = (nvshmemx_cmp_type_t) -1, const_zero = (nvshmemx_cmp_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_cmp_type_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_cmp_type_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_cmp_type_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_cmp_type_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_cmp_type_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_cmp_type_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_cmp_type_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_thread_support_t(nvshmemx_thread_support_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_thread_support_t neg_one = (nvshmemx_thread_support_t) -1, const_zero = (nvshmemx_thread_support_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_thread_support_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_thread_support_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_thread_support_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_thread_support_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_thread_support_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_thread_support_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_thread_support_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_proxy_status_t(nvshmemx_proxy_status_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_proxy_status_t neg_one = (nvshmemx_proxy_status_t) -1, const_zero = (nvshmemx_proxy_status_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_proxy_status_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_proxy_status_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_proxy_status_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_proxy_status_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_proxy_status_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_proxy_status_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_proxy_status_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_init_status_t(nvshmemx_init_status_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_init_status_t neg_one = (nvshmemx_init_status_t) -1, const_zero = (nvshmemx_init_status_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_init_status_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_init_status_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_init_status_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_init_status_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_init_status_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_init_status_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_init_status_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_qp_handle_index_t(nvshmemx_qp_handle_index_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_qp_handle_index_t neg_one = (nvshmemx_qp_handle_index_t) -1, const_zero = (nvshmemx_qp_handle_index_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_qp_handle_index_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_qp_handle_index_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_qp_handle_index_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_qp_handle_index_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_qp_handle_index_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_qp_handle_index_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_qp_handle_index_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmem_pe_index_t(nvshmem_pe_index_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmem_pe_index_t neg_one = (nvshmem_pe_index_t) -1, const_zero = (nvshmem_pe_index_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmem_pe_index_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmem_pe_index_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmem_pe_index_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmem_pe_index_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmem_pe_index_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmem_pe_index_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmem_pe_index_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmem_team_id_t(nvshmem_team_id_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmem_team_id_t neg_one = (nvshmem_team_id_t) -1, const_zero = (nvshmem_team_id_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmem_team_id_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmem_team_id_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmem_team_id_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmem_team_id_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmem_team_id_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmem_team_id_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmem_team_id_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_nvshmemx_status(nvshmemx_status value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const nvshmemx_status neg_one = (nvshmemx_status) -1, const_zero = (nvshmemx_status) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(nvshmemx_status) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_status) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(nvshmemx_status) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(nvshmemx_status) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(nvshmemx_status) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(nvshmemx_status),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(nvshmemx_status));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_flags(flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const flags neg_one = (flags) -1, const_zero = (flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyLong_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 2 * PyLong_SHIFT)) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 3 * PyLong_SHIFT)) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 4 * PyLong_SHIFT)) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int32_t) 1) << (sizeof(int32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyLong_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyLong_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module_2);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__5);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
