/* Generated by Cython 3.2.8 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "language": "c++",
        "name": "nvshmem.bindings.cynvshmem",
        "sources": [
            "nvshmem/bindings/cynvshmem.pyx"
        ]
    },
    "module_name": "nvshmem.bindings.cynvshmem"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_8"
#define CYTHON_HEX_VERSION 0x030208F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__nvshmem__bindings__cynvshmem
#define __PYX_HAVE_API__nvshmem__bindings__cynvshmem
/* Early includes */
#include <stdint.h>
#include "cuda_runtime_api.h"
#include "cuda_fp16.h"
#include "cuda_bf16.h"
#include "limits.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "nvshmem/bindings/cynvshmem.pyx",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* Atomics.proto (used by CodeObjectCache) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* #### Code section: numeric_typedefs ### */

/* "nvshmem/bindings/cynvshmem.pxd":143
 * 
 * # Longs
 * ctypedef long longlong             # <<<<<<<<<<<<<<
 * ctypedef signed char schar
 * 
*/
typedef long __pyx_t_7nvshmem_8bindings_9cynvshmem_longlong;

/* "nvshmem/bindings/cynvshmem.pxd":144
 * # Longs
 * ctypedef long longlong
 * ctypedef signed char schar             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef signed char __pyx_t_7nvshmem_8bindings_9cynvshmem_schar;

/* "nvshmem/bindings/cynvshmem.pxd":151
 *     int version
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'             # <<<<<<<<<<<<<<
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
*/
typedef int nvshmemx_qp_handle_t;

/* "nvshmem/bindings/cynvshmem.pxd":152
 *     char internal[124]
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
*/
typedef uint64_t nvshmemx_team_uniqueid_t;

/* "nvshmem/bindings/cynvshmem.pxd":153
 * ctypedef int nvshmemx_qp_handle_t 'nvshmemx_qp_handle_t'
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
*/
typedef int32_t nvshmem_team_t;

/* "nvshmem/bindings/cynvshmem.pxd":165
 *     nvshmemx_team_uniqueid_t uniqueid
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
*/
typedef nvshmem_team_t nvshmemx_team_t;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct nvshmemx_uniqueid_v1;
typedef struct nvshmemx_uniqueid_v1 nvshmemx_uniqueid_v1;
struct nvshmemx_uniqueid_args_v1;
typedef struct nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_v1;
struct nvshmem_team_config_v2;
typedef struct nvshmem_team_config_v2 nvshmem_team_config_v2;
struct nvshmemx_init_args_v2;
typedef struct nvshmemx_init_args_v2 nvshmemx_init_args_v2;
struct nvshmemx_init_attr_v2;
typedef struct nvshmemx_init_attr_v2 nvshmemx_init_attr_v2;

/* "nvshmem/bindings/cynvshmem.pxd":21
 * 
 * # enums
 * ctypedef enum nvshmemx_signal_op_t "nvshmemx_signal_op_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_SIGNAL_SET "NVSHMEM_SIGNAL_SET" = 9
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
*/
enum nvshmemx_signal_op_t {
  NVSHMEM_SIGNAL_SET = 9,
  NVSHMEM_SIGNAL_ADD = 10
};
typedef enum nvshmemx_signal_op_t nvshmemx_signal_op_t;

/* "nvshmem/bindings/cynvshmem.pxd":25
 *     NVSHMEM_SIGNAL_ADD "NVSHMEM_SIGNAL_ADD" = 10
 * 
 * ctypedef enum nvshmemx_tma_policy_t "nvshmemx_tma_policy_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_TMA_DISABLE "NVSHMEMX_TMA_DISABLE" = 0
 *     NVSHMEMX_TMA_ENABLE "NVSHMEMX_TMA_ENABLE" = 1
*/
enum nvshmemx_tma_policy_t {
  NVSHMEMX_TMA_DISABLE = 0,
  NVSHMEMX_TMA_ENABLE = 1,
  NVSHMEMX_TMA_FORCE = 2,
  NVSHMEMX_TMA_POLICY_MAX = 0x7FFF
};
typedef enum nvshmemx_tma_policy_t nvshmemx_tma_policy_t;

/* "nvshmem/bindings/cynvshmem.pxd":31
 *     NVSHMEMX_TMA_POLICY_MAX "NVSHMEMX_TMA_POLICY_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_smem_amount_t "nvshmemx_smem_amount_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SMEM_RECOMMENDED "NVSHMEMX_SMEM_RECOMMENDED" = 0
 *     NVSHMEMX_SMEM_MINIMUM "NVSHMEMX_SMEM_MINIMUM" = 1
*/
enum nvshmemx_smem_amount_t {
  NVSHMEMX_SMEM_RECOMMENDED = 0,
  NVSHMEMX_SMEM_MINIMUM = 1,
  NVSHMEMX_SMEM_BARRIERS_ONLY = 2,
  NVSHMEMX_SMEM_AMOUNT_MAX = 0x7FFF
};
typedef enum nvshmemx_smem_amount_t nvshmemx_smem_amount_t;

/* "nvshmem/bindings/cynvshmem.pxd":37
 *     NVSHMEMX_SMEM_AMOUNT_MAX "NVSHMEMX_SMEM_AMOUNT_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_cmp_type_t "nvshmemx_cmp_type_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_CMP_EQ "NVSHMEM_CMP_EQ" = 0
 *     NVSHMEM_CMP_NE "NVSHMEM_CMP_NE"
*/
enum nvshmemx_cmp_type_t {
  NVSHMEM_CMP_EQ = 0,
  NVSHMEM_CMP_NE,
  NVSHMEM_CMP_GT,
  NVSHMEM_CMP_LE,
  NVSHMEM_CMP_LT,
  NVSHMEM_CMP_GE,
  NVSHMEM_CMP_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_cmp_type_t nvshmemx_cmp_type_t;

/* "nvshmem/bindings/cynvshmem.pxd":46
 *     NVSHMEM_CMP_SENTINEL "NVSHMEM_CMP_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_thread_support_t "nvshmemx_thread_support_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_THREAD_SINGLE "NVSHMEM_THREAD_SINGLE" = 0
 *     NVSHMEM_THREAD_FUNNELED "NVSHMEM_THREAD_FUNNELED"
*/
enum nvshmemx_thread_support_t {
  NVSHMEM_THREAD_SINGLE = 0,
  NVSHMEM_THREAD_FUNNELED,
  NVSHMEM_THREAD_SERIALIZED,
  NVSHMEM_THREAD_MULTIPLE,
  NVSHMEM_THREAD_TYPE_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_thread_support_t nvshmemx_thread_support_t;

/* "nvshmem/bindings/cynvshmem.pxd":53
 *     NVSHMEM_THREAD_TYPE_SENTINEL "NVSHMEM_THREAD_TYPE_SENTINEL" = 32767
 * 
 * ctypedef enum nvshmemx_proxy_status_t "nvshmemx_proxy_status_t":             # <<<<<<<<<<<<<<
 *     PROXY_GLOBAL_EXIT_NOT_REQUESTED "PROXY_GLOBAL_EXIT_NOT_REQUESTED" = 0
 *     PROXY_GLOBAL_EXIT_INIT "PROXY_GLOBAL_EXIT_INIT"
*/
enum nvshmemx_proxy_status_t {
  PROXY_GLOBAL_EXIT_NOT_REQUESTED = 0,
  PROXY_GLOBAL_EXIT_INIT,
  PROXY_GLOBAL_EXIT_REQUESTED,
  PROXY_GLOBAL_EXIT_FINISHED,
  PROXY_GLOBAL_EXIT_MAX_STATE = 0x7FFF
};
typedef enum nvshmemx_proxy_status_t nvshmemx_proxy_status_t;

/* "nvshmem/bindings/cynvshmem.pxd":60
 *     PROXY_GLOBAL_EXIT_MAX_STATE "PROXY_GLOBAL_EXIT_MAX_STATE" = 32767
 * 
 * ctypedef enum nvshmemx_init_status_t "nvshmemx_init_status_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_STATUS_NOT_INITIALIZED "NVSHMEM_STATUS_NOT_INITIALIZED" = 0
 *     NVSHMEM_STATUS_IS_BOOTSTRAPPED "NVSHMEM_STATUS_IS_BOOTSTRAPPED"
*/
enum nvshmemx_init_status_t {
  NVSHMEM_STATUS_NOT_INITIALIZED = 0,
  NVSHMEM_STATUS_IS_BOOTSTRAPPED,
  NVSHMEM_STATUS_IS_INITIALIZED,
  NVSHMEM_STATUS_LIMITED_MPG,
  NVSHMEM_STATUS_FULL_MPG,
  NVSHMEM_STATUS_INVALID = 0x7FFF
};
typedef enum nvshmemx_init_status_t nvshmemx_init_status_t;

/* "nvshmem/bindings/cynvshmem.pxd":68
 *     NVSHMEM_STATUS_INVALID "NVSHMEM_STATUS_INVALID" = 32767
 * 
 * ctypedef enum nvshmemx_qp_handle_index_t "nvshmemx_qp_handle_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_QP_HOST "NVSHMEMX_QP_HOST" = 0
 *     NVSHMEMX_QP_DEFAULT "NVSHMEMX_QP_DEFAULT" = 1
*/
enum nvshmemx_qp_handle_index_t {
  NVSHMEMX_QP_HOST = 0,
  NVSHMEMX_QP_DEFAULT = 1,
  NVSHMEMX_QP_ANY = 0x7FFF,
  NVSHMEMX_QP_ALL = 0x7FFF
};
typedef enum nvshmemx_qp_handle_index_t nvshmemx_qp_handle_index_t;

/* "nvshmem/bindings/cynvshmem.pxd":74
 *     NVSHMEMX_QP_ALL "NVSHMEMX_QP_ALL" = 32767
 * 
 * ctypedef enum nvshmem_pe_index_t "nvshmem_pe_index_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
*/
enum nvshmem_pe_index_t {

  /* "nvshmem/bindings/cynvshmem.pxd":77
 *     NVSHMEM_PE_INVALID "NVSHMEM_PE_INVALID" = -(1)
 *     NVSHMEMX_PE_ANY "NVSHMEMX_PE_ANY" = (1 << 15)
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)             # <<<<<<<<<<<<<<
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":
*/
  NVSHMEM_PE_INVALID = -1L,
  NVSHMEMX_PE_ANY = (1 << 15),
  NVSHMEMX_PE_ALL = (1 << 15)
};
typedef enum nvshmem_pe_index_t nvshmem_pe_index_t;

/* "nvshmem/bindings/cynvshmem.pxd":79
 *     NVSHMEMX_PE_ALL "NVSHMEMX_PE_ALL" = (1 << 15)
 * 
 * ctypedef enum nvshmem_team_id_t "nvshmem_team_id_t":             # <<<<<<<<<<<<<<
 *     NVSHMEM_TEAM_INVALID "NVSHMEM_TEAM_INVALID" = -(1)
 *     NVSHMEM_TEAM_WORLD "NVSHMEM_TEAM_WORLD" = 0
*/
enum nvshmem_team_id_t {
  NVSHMEM_TEAM_INVALID = -1L,
  NVSHMEM_TEAM_WORLD = 0,
  NVSHMEM_TEAM_WORLD_INDEX = 0,
  NVSHMEM_TEAM_SHARED = 1,
  NVSHMEM_TEAM_SHARED_INDEX = 1,
  NVSHMEMX_TEAM_NODE = 2,
  NVSHMEM_TEAM_NODE_INDEX = 2,
  NVSHMEMX_TEAM_SAME_MYPE_NODE = 3,
  NVSHMEM_TEAM_SAME_MYPE_NODE_INDEX = 3,
  NVSHMEMI_TEAM_SAME_GPU = 4,
  NVSHMEM_TEAM_SAME_GPU_INDEX = 4,
  NVSHMEMI_TEAM_GPU_LEADERS = 5,
  NVSHMEM_TEAM_GPU_LEADERS_INDEX = 5,
  NVSHMEM_TEAM_MC_SHARED = 6,
  NVSHMEM_TEAM_MC_SHARED_INDEX = 6,
  NVSHMEM_TEAMS_MIN = 7,
  NVSHMEM_TEAM_INDEX_MAX = 0x7FFF
};
typedef enum nvshmem_team_id_t nvshmem_team_id_t;

/* "nvshmem/bindings/cynvshmem.pxd":98
 *     NVSHMEM_TEAM_INDEX_MAX "NVSHMEM_TEAM_INDEX_MAX" = 32767
 * 
 * ctypedef enum nvshmemx_status "nvshmemx_status":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_SUCCESS "NVSHMEMX_SUCCESS"
 *     NVSHMEMX_ERROR_INVALID_VALUE "NVSHMEMX_ERROR_INVALID_VALUE"
*/
enum nvshmemx_status {
  NVSHMEMX_SUCCESS,
  NVSHMEMX_ERROR_INVALID_VALUE,
  NVSHMEMX_ERROR_OUT_OF_MEMORY,
  NVSHMEMX_ERROR_NOT_SUPPORTED,
  NVSHMEMX_ERROR_SYMMETRY,
  NVSHMEMX_ERROR_GPU_NOT_SELECTED,
  NVSHMEMX_ERROR_COLLECTIVE_LAUNCH_FAILED,
  NVSHMEMX_ERROR_INTERNAL,
  NVSHMEMX_ERROR_SENTINEL = 0x7FFF
};
typedef enum nvshmemx_status nvshmemx_status;

/* "nvshmem/bindings/cynvshmem.pxd":109
 *     NVSHMEMX_ERROR_SENTINEL "NVSHMEMX_ERROR_SENTINEL" = 32767
 * 
 * ctypedef enum flags "flags":             # <<<<<<<<<<<<<<
 *     NVSHMEMX_INIT_THREAD_PES "NVSHMEMX_INIT_THREAD_PES" = 1
 *     NVSHMEMX_INIT_WITH_MPI_COMM "NVSHMEMX_INIT_WITH_MPI_COMM" = (1 << 1)
*/
enum flags {

  /* "nvshmem/bindings/cynvshmem.pxd":114
 *     NVSHMEMX_INIT_WITH_SHMEM "NVSHMEMX_INIT_WITH_SHMEM" = (1 << 2)
 *     NVSHMEMX_INIT_WITH_UNIQUEID "NVSHMEMX_INIT_WITH_UNIQUEID" = (1 << 3)
 *     NVSHMEMX_INIT_MAX "NVSHMEMX_INIT_MAX" = (1 << 31)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  NVSHMEMX_INIT_THREAD_PES = 1,
  NVSHMEMX_INIT_WITH_MPI_COMM = (1 << 1),
  NVSHMEMX_INIT_WITH_SHMEM = (1 << 2),
  NVSHMEMX_INIT_WITH_UNIQUEID = (1 << 3),
  NVSHMEMX_INIT_MAX = (1 << 31)
};
typedef enum flags flags;

/* "nvshmem/bindings/cynvshmem.pxd":122
 *     ctypedef CUstream_st* cudaStream_t
 * 
 * ctypedef void* CUmodule 'CUmodule'             # <<<<<<<<<<<<<<
 * ctypedef void* CUlibrary 'CUlibrary'
 * 
*/
typedef void *CUmodule;

/* "nvshmem/bindings/cynvshmem.pxd":123
 * 
 * ctypedef void* CUmodule 'CUmodule'
 * ctypedef void* CUlibrary 'CUlibrary'             # <<<<<<<<<<<<<<
 * 
 * # Types for NVSHMEM Collectives
*/
typedef void *CUlibrary;

/* "nvshmem/bindings/cynvshmem.pxd":148
 * 
 * 
 * ctypedef struct nvshmemx_uniqueid_v1 'nvshmemx_uniqueid_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     char internal[124]
*/
struct nvshmemx_uniqueid_v1 {
  int version;
  char internal[124];
};

/* "nvshmem/bindings/cynvshmem.pxd":154
 * ctypedef uint64_t nvshmemx_team_uniqueid_t 'nvshmemx_team_uniqueid_t'
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':
 *     int version
*/
typedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t;

/* "nvshmem/bindings/cynvshmem.pxd":155
 * ctypedef int32_t nvshmem_team_t 'nvshmem_team_t'
 * ctypedef nvshmemx_uniqueid_v1 nvshmemx_uniqueid_t 'nvshmemx_uniqueid_t'
 * ctypedef struct nvshmemx_uniqueid_args_v1 'nvshmemx_uniqueid_args_v1':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_v1* id
*/
struct nvshmemx_uniqueid_args_v1 {
  int version;
  nvshmemx_uniqueid_v1 *id;
  int myrank;
  int nranks;
};

/* "nvshmem/bindings/cynvshmem.pxd":160
 *     int myrank
 *     int nranks
 * ctypedef struct nvshmem_team_config_v2 'nvshmem_team_config_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     int num_contexts
*/
struct nvshmem_team_config_v2 {
  int version;
  int num_contexts;
  nvshmemx_team_uniqueid_t uniqueid;
  char padding[48];
};

/* "nvshmem/bindings/cynvshmem.pxd":166
 *     char padding[48]
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'             # <<<<<<<<<<<<<<
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
*/
typedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t;

/* "nvshmem/bindings/cynvshmem.pxd":167
 * ctypedef nvshmem_team_t nvshmemx_team_t 'nvshmemx_team_t'
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':
 *     int version
*/
typedef nvshmem_team_config_v2 nvshmem_team_config_t;

/* "nvshmem/bindings/cynvshmem.pxd":168
 * ctypedef nvshmemx_uniqueid_args_v1 nvshmemx_uniqueid_args_t 'nvshmemx_uniqueid_args_t'
 * ctypedef nvshmem_team_config_v2 nvshmem_team_config_t 'nvshmem_team_config_t'
 * ctypedef struct nvshmemx_init_args_v2 'nvshmemx_init_args_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     nvshmemx_uniqueid_args_t uid_args
*/
struct nvshmemx_init_args_v2 {
  int version;
  nvshmemx_uniqueid_args_t uid_args;
  int cuda_device_id;
  char content[92];
};

/* "nvshmem/bindings/cynvshmem.pxd":173
 *     int cuda_device_id
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'             # <<<<<<<<<<<<<<
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':
 *     int version
*/
typedef nvshmemx_init_args_v2 nvshmemx_init_args_t;

/* "nvshmem/bindings/cynvshmem.pxd":174
 *     char content[92]
 * ctypedef nvshmemx_init_args_v2 nvshmemx_init_args_t 'nvshmemx_init_args_t'
 * ctypedef struct nvshmemx_init_attr_v2 'nvshmemx_init_attr_v2':             # <<<<<<<<<<<<<<
 *     int version
 *     void* mpi_comm
*/
struct nvshmemx_init_attr_v2 {
  int version;
  void *mpi_comm;
  nvshmemx_init_args_t args;
};

/* "nvshmem/bindings/cynvshmem.pxd":178
 *     void* mpi_comm
 *     nvshmemx_init_args_t args
 * ctypedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t 'nvshmemx_init_attr_t'             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef nvshmemx_init_attr_v2 nvshmemx_init_attr_t;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* ErrOccurredWithGIL.proto */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig);

/* GetApiDict.proto */
static PyObject *__Pyx_ApiExport_GetApiDict(void);

/* IncludeStringH.proto */
#include <string.h>

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_8(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* dict_setdefault.proto (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* PyDictVersioning.proto (used by CLineInTraceback) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by CLineInTraceback) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* GCCDiagnostics.proto (used by CIntToPy) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectVectorCallKwBuilder) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectVectorCallKwBuilder.proto (used by CIntToPy) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* AddModuleRef.proto */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."


/* Module declarations from "libc.stdint" */

/* Module declarations from "nvshmem.bindings._internal" */

/* Module declarations from "nvshmem.bindings._internal.nvshmem" */
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier)(nvshmem_team_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier_all)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_init_status)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_my_pe)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_n_pes)(void); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_info_get_version)(int *, int *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_vendor_get_version_info)(int *, int *, int *); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_malloc)(size_t); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_calloc)(size_t, size_t); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_align)(size_t, size_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_free)(void *); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_ptr)(void const *, int); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_mc_ptr)(nvshmem_team_t, void const *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_my_pe)(nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_n_pes)(nvshmem_team_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_get_config)(nvshmem_team_t, nvshmem_team_config_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_translate_pe)(nvshmem_team_t, int, nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_strided)(nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_get_uniqueid)(nvshmemx_team_uniqueid_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_init)(nvshmem_team_t *, nvshmem_team_config_t *, long, int, int); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_2d)(nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_destroy)(nvshmem_team_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_alltoall_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_alltoall_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_alltoall_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_alltoall_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_alltoall_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_alltoall_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_alltoall_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_alltoall_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_alltoall_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_alltoall_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_alltoall_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_alltoall_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_alltoall_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_alltoall_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_alltoall_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_alltoall_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_alltoall_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_alltoall_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_alltoall_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_on_stream)(nvshmem_team_t, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_all_on_stream)(cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_sync_on_stream)(nvshmem_team_t, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_sync_all_on_stream)(cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_broadcast_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_broadcast_on_stream)(nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_broadcast_on_stream)(nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_broadcast_on_stream)(nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_broadcast_on_stream)(nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_broadcast_on_stream)(nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_broadcast_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_broadcast_on_stream)(nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_broadcast_on_stream)(nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_broadcast_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_broadcast_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_broadcast_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_broadcast_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_broadcast_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_broadcast_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_broadcast_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_broadcast_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_broadcast_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_broadcast_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_fcollect_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_fcollect_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_fcollect_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_fcollect_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_fcollect_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_fcollect_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_fcollect_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_fcollect_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_fcollect_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_fcollect_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_fcollect_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_fcollect_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_fcollect_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_fcollect_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_fcollect_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_fcollect_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_fcollect_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_fcollect_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_fcollect_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reduce_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reduce_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reduce_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reduce_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reduce_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reduce_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reduce_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reduce_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reduce_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reduce_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reduce_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reduce_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reduce_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reduce_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reduce_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reduce_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reduce_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reduce_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reduce_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reducescatter_on_stream)(nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reducescatter_on_stream)(nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reducescatter_on_stream)(nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reducescatter_on_stream)(nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reducescatter_on_stream)(nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reducescatter_on_stream)(nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reducescatter_on_stream)(nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reducescatter_on_stream)(nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reducescatter_on_stream)(nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reducescatter_on_stream)(nvshmem_team_t, char *, char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reducescatter_on_stream)(nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reducescatter_on_stream)(nvshmem_team_t, short *, short const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reducescatter_on_stream)(nvshmem_team_t, int *, int const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reducescatter_on_stream)(nvshmem_team_t, long *, long const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reducescatter_on_stream)(nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reducescatter_on_stream)(nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reducescatter_on_stream)(nvshmem_team_t, half *, half const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reducescatter_on_stream)(nvshmem_team_t, float *, float const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reducescatter_on_stream)(nvshmem_team_t, double *, double const *, size_t, cudaStream_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_init_attr)(unsigned int, nvshmemx_init_attr_t *); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_finalize)(void); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_uniqueid_args)(int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_mpi_comm_args)(void *, nvshmemx_init_attr_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_get_uniqueid)(nvshmemx_uniqueid_t *); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_init)(CUmodule); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_finalize)(CUmodule); /*proto*/
static void *(*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_register_symmetric)(void *, size_t, int); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_unregister_symmetric)(void *, size_t); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_init)(CUlibrary); /*proto*/
static int (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_finalize)(CUlibrary); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_on_stream)(void *, void const *, size_t, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_signal_on_stream)(void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_getmem_on_stream)(void *, void const *, size_t, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_quiet_on_stream)(cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_op_on_stream)(uint64_t *, uint64_t, int, int, cudaStream_t); /*proto*/
static void (*__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_wait_until_on_stream)(uint64_t *, int, uint64_t, cudaStream_t); /*proto*/

/* Module declarations from "nvshmem.bindings.cynvshmem" */
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "nvshmem.bindings.cynvshmem"
extern int __pyx_module_is_main_nvshmem__bindings__cynvshmem;
int __pyx_module_is_main_nvshmem__bindings__cynvshmem = 0;

/* Implementation of "nvshmem.bindings.cynvshmem" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyObject *__pyx_string_tab[11];
/* #### Code section: module_state_contents ### */
/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[1]
#define __pyx_n_u_main __pyx_string_tab[2]
#define __pyx_n_u_module __pyx_string_tab[3]
#define __pyx_n_u_name __pyx_string_tab[4]
#define __pyx_n_u_pyx_capi __pyx_string_tab[5]
#define __pyx_n_u_qualname __pyx_string_tab[6]
#define __pyx_n_u_setdefault __pyx_string_tab[7]
#define __pyx_n_u_test __pyx_string_tab[8]
#define __pyx_kp_b_int_CUlibrary_int_CUmodule_int_i __pyx_string_tab[9]
#define __pyx_kp_b_int_CUlibrary_int_CUmodule_int_i_2 __pyx_string_tab[10]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  for (int i=0; i<11; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  for (int i=0; i<11; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "nvshmem/bindings/cynvshmem.pyx":16
 * ###############################################################################
 * 
 * cdef int nvshmem_barrier(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_barrier(team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":17
 * 
 * cdef int nvshmem_barrier(nvshmem_team_t team) except* nogil:
 *     return _nvshmem._nvshmem_barrier(team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":16
 * ###############################################################################
 * 
 * cdef int nvshmem_barrier(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_barrier(team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_barrier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":20
 * 
 * 
 * cdef void nvshmem_barrier_all() except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_barrier_all()
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier_all(void) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":21
 * 
 * cdef void nvshmem_barrier_all() except* nogil:
 *     _nvshmem._nvshmem_barrier_all()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier_all(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 21, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":20
 * 
 * 
 * cdef void nvshmem_barrier_all() except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_barrier_all()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_barrier_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":24
 * 
 * 
 * cdef int nvshmemx_init_status() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_init_status()
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_init_status(void) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":25
 * 
 * cdef int nvshmemx_init_status() except* nogil:
 *     return _nvshmem._nvshmemx_init_status()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_init_status(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":24
 * 
 * 
 * cdef int nvshmemx_init_status() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_init_status()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_init_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":28
 * 
 * 
 * cdef int nvshmem_my_pe() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_my_pe()
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_my_pe(void) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":29
 * 
 * cdef int nvshmem_my_pe() except* nogil:
 *     return _nvshmem._nvshmem_my_pe()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_my_pe(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":28
 * 
 * 
 * cdef int nvshmem_my_pe() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_my_pe()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":32
 * 
 * 
 * cdef int nvshmem_n_pes() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_n_pes()
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_n_pes(void) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":33
 * 
 * cdef int nvshmem_n_pes() except* nogil:
 *     return _nvshmem._nvshmem_n_pes()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_n_pes(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 33, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":32
 * 
 * 
 * cdef int nvshmem_n_pes() except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_n_pes()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":36
 * 
 * 
 * cdef void nvshmem_info_get_version(int* major, int* minor) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_info_get_version(major, minor)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_info_get_version(int *__pyx_v_major, int *__pyx_v_minor) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":37
 * 
 * cdef void nvshmem_info_get_version(int* major, int* minor) except* nogil:
 *     _nvshmem._nvshmem_info_get_version(major, minor)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_info_get_version(__pyx_v_major, __pyx_v_minor); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 37, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":36
 * 
 * 
 * cdef void nvshmem_info_get_version(int* major, int* minor) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_info_get_version(major, minor)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_info_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":40
 * 
 * 
 * cdef void nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_vendor_get_version_info(major, minor, patch)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_vendor_get_version_info(int *__pyx_v_major, int *__pyx_v_minor, int *__pyx_v_patch) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":41
 * 
 * cdef void nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:
 *     _nvshmem._nvshmemx_vendor_get_version_info(major, minor, patch)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_vendor_get_version_info(__pyx_v_major, __pyx_v_minor, __pyx_v_patch); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 41, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":40
 * 
 * 
 * cdef void nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_vendor_get_version_info(major, minor, patch)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_vendor_get_version_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":44
 * 
 * 
 * cdef void* nvshmem_malloc(size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_malloc(size)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_malloc(size_t __pyx_v_size) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":45
 * 
 * cdef void* nvshmem_malloc(size_t size) except* nogil:
 *     return _nvshmem._nvshmem_malloc(size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_malloc(__pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 45, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":44
 * 
 * 
 * cdef void* nvshmem_malloc(size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_malloc(size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":48
 * 
 * 
 * cdef void* nvshmem_calloc(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_calloc(count, size)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_calloc(size_t __pyx_v_count, size_t __pyx_v_size) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":49
 * 
 * cdef void* nvshmem_calloc(size_t count, size_t size) except* nogil:
 *     return _nvshmem._nvshmem_calloc(count, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_calloc(__pyx_v_count, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":48
 * 
 * 
 * cdef void* nvshmem_calloc(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_calloc(count, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":52
 * 
 * 
 * cdef void* nvshmem_align(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_align(count, size)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_align(size_t __pyx_v_count, size_t __pyx_v_size) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":53
 * 
 * cdef void* nvshmem_align(size_t count, size_t size) except* nogil:
 *     return _nvshmem._nvshmem_align(count, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_align(__pyx_v_count, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":52
 * 
 * 
 * cdef void* nvshmem_align(size_t count, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_align(count, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":56
 * 
 * 
 * cdef void nvshmem_free(void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_free(ptr)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_free(void *__pyx_v_ptr) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":57
 * 
 * cdef void nvshmem_free(void* ptr) except* nogil:
 *     _nvshmem._nvshmem_free(ptr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_free(__pyx_v_ptr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 57, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":56
 * 
 * 
 * cdef void nvshmem_free(void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_free(ptr)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_free", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":60
 * 
 * 
 * cdef void* nvshmem_ptr(const void* dest, int pe) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_ptr(dest, pe)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_ptr(void const *__pyx_v_dest, int __pyx_v_pe) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":61
 * 
 * cdef void* nvshmem_ptr(const void* dest, int pe) except* nogil:
 *     return _nvshmem._nvshmem_ptr(dest, pe)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_ptr(__pyx_v_dest, __pyx_v_pe); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":60
 * 
 * 
 * cdef void* nvshmem_ptr(const void* dest, int pe) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_ptr(dest, pe)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":64
 * 
 * 
 * cdef void* nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_mc_ptr(team, ptr)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_mc_ptr(nvshmem_team_t __pyx_v_team, void const *__pyx_v_ptr) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":65
 * 
 * cdef void* nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:
 *     return _nvshmem._nvshmemx_mc_ptr(team, ptr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_mc_ptr(__pyx_v_team, __pyx_v_ptr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":64
 * 
 * 
 * cdef void* nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_mc_ptr(team, ptr)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_mc_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":68
 * 
 * 
 * cdef int nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_my_pe(team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_my_pe(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":69
 * 
 * cdef int nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:
 *     return _nvshmem._nvshmem_team_my_pe(team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_my_pe(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":68
 * 
 * 
 * cdef int nvshmem_team_my_pe(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_my_pe(team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_my_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":72
 * 
 * 
 * cdef int nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_n_pes(team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_n_pes(nvshmem_team_t __pyx_v_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":73
 * 
 * cdef int nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:
 *     return _nvshmem._nvshmem_team_n_pes(team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_n_pes(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":72
 * 
 * 
 * cdef int nvshmem_team_n_pes(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_n_pes(team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_n_pes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":76
 * 
 * 
 * cdef void nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_team_get_config(team, config)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_get_config(nvshmem_team_t __pyx_v_team, nvshmem_team_config_t *__pyx_v_config) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":77
 * 
 * cdef void nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:
 *     _nvshmem._nvshmem_team_get_config(team, config)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_get_config(__pyx_v_team, __pyx_v_config); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 77, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":76
 * 
 * 
 * cdef void nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_team_get_config(team, config)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_get_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":80
 * 
 * 
 * cdef int nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_translate_pe(src_team, src_pe, dest_team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_translate_pe(nvshmem_team_t __pyx_v_src_team, int __pyx_v_src_pe, nvshmem_team_t __pyx_v_dest_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":81
 * 
 * cdef int nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:
 *     return _nvshmem._nvshmem_team_translate_pe(src_team, src_pe, dest_team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_translate_pe(__pyx_v_src_team, __pyx_v_src_pe, __pyx_v_dest_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":80
 * 
 * 
 * cdef int nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_translate_pe(src_team, src_pe, dest_team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_translate_pe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":84
 * 
 * 
 * cdef int nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_split_strided(parent_team, PE_start, PE_stride, PE_size, config, config_mask, new_team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_strided(nvshmem_team_t __pyx_v_parent_team, int __pyx_v_PE_start, int __pyx_v_PE_stride, int __pyx_v_PE_size, nvshmem_team_config_t const *__pyx_v_config, long __pyx_v_config_mask, nvshmem_team_t *__pyx_v_new_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":85
 * 
 * cdef int nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:
 *     return _nvshmem._nvshmem_team_split_strided(parent_team, PE_start, PE_stride, PE_size, config, config_mask, new_team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_strided(__pyx_v_parent_team, __pyx_v_PE_start, __pyx_v_PE_stride, __pyx_v_PE_size, __pyx_v_config, __pyx_v_config_mask, __pyx_v_new_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":84
 * 
 * 
 * cdef int nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_split_strided(parent_team, PE_start, PE_stride, PE_size, config, config_mask, new_team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_split_strided", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":88
 * 
 * 
 * cdef int nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_get_uniqueid(uniqueid)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t *__pyx_v_uniqueid) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":89
 * 
 * cdef int nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:
 *     return _nvshmem._nvshmemx_team_get_uniqueid(uniqueid)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_get_uniqueid(__pyx_v_uniqueid); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 89, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":88
 * 
 * 
 * cdef int nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_get_uniqueid(uniqueid)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_team_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":92
 * 
 * 
 * cdef int nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_init(team, config, config_mask, npes, pe_idx_in_team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_init(nvshmem_team_t *__pyx_v_team, nvshmem_team_config_t *__pyx_v_config, long __pyx_v_config_mask, int __pyx_v_npes, int __pyx_v_pe_idx_in_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":93
 * 
 * cdef int nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:
 *     return _nvshmem._nvshmemx_team_init(team, config, config_mask, npes, pe_idx_in_team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_init(__pyx_v_team, __pyx_v_config, __pyx_v_config_mask, __pyx_v_npes, __pyx_v_pe_idx_in_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":92
 * 
 * 
 * cdef int nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_init(team, config, config_mask, npes, pe_idx_in_team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_team_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":96
 * 
 * 
 * cdef int nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_split_2d(parent_team, xrange, xaxis_config, xaxis_mask, xaxis_team, yaxis_config, yaxis_mask, yaxis_team)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_2d(nvshmem_team_t __pyx_v_parent_team, int __pyx_v_xrange, nvshmem_team_config_t const *__pyx_v_xaxis_config, long __pyx_v_xaxis_mask, nvshmem_team_t *__pyx_v_xaxis_team, nvshmem_team_config_t const *__pyx_v_yaxis_config, long __pyx_v_yaxis_mask, nvshmem_team_t *__pyx_v_yaxis_team) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":97
 * 
 * cdef int nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:
 *     return _nvshmem._nvshmem_team_split_2d(parent_team, xrange, xaxis_config, xaxis_mask, xaxis_team, yaxis_config, yaxis_mask, yaxis_team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_2d(__pyx_v_parent_team, __pyx_v_xrange, __pyx_v_xaxis_config, __pyx_v_xaxis_mask, __pyx_v_xaxis_team, __pyx_v_yaxis_config, __pyx_v_yaxis_mask, __pyx_v_yaxis_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":96
 * 
 * 
 * cdef int nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmem_team_split_2d(parent_team, xrange, xaxis_config, xaxis_mask, xaxis_team, yaxis_config, yaxis_mask, yaxis_team)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_split_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":100
 * 
 * 
 * cdef void nvshmem_team_destroy(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_team_destroy(team)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_destroy(nvshmem_team_t __pyx_v_team) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":101
 * 
 * cdef void nvshmem_team_destroy(nvshmem_team_t team) except* nogil:
 *     _nvshmem._nvshmem_team_destroy(team)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_destroy(__pyx_v_team); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":100
 * 
 * 
 * cdef void nvshmem_team_destroy(nvshmem_team_t team) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmem_team_destroy(team)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmem_team_destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":104
 * 
 * 
 * cdef int nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":105
 * 
 * cdef int nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":104
 * 
 * 
 * cdef int nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":108
 * 
 * 
 * cdef int nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_alltoall_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":109
 * 
 * cdef int nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":108
 * 
 * 
 * cdef int nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":112
 * 
 * 
 * cdef int nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_alltoall_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":113
 * 
 * cdef int nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":112
 * 
 * 
 * cdef int nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":116
 * 
 * 
 * cdef int nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_alltoall_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":117
 * 
 * cdef int nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 117, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":116
 * 
 * 
 * cdef int nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":120
 * 
 * 
 * cdef int nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_alltoall_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":121
 * 
 * cdef int nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":120
 * 
 * 
 * cdef int nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":124
 * 
 * 
 * cdef int nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_alltoall_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":125
 * 
 * cdef int nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 125, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":124
 * 
 * 
 * cdef int nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":128
 * 
 * 
 * cdef int nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_alltoall_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":129
 * 
 * cdef int nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":128
 * 
 * 
 * cdef int nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":132
 * 
 * 
 * cdef int nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":133
 * 
 * cdef int nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":132
 * 
 * 
 * cdef int nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":136
 * 
 * 
 * cdef int nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_alltoall_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":137
 * 
 * cdef int nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":136
 * 
 * 
 * cdef int nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":140
 * 
 * 
 * cdef int nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":141
 * 
 * cdef int nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":140
 * 
 * 
 * cdef int nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":144
 * 
 * 
 * cdef int nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":145
 * 
 * cdef int nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":144
 * 
 * 
 * cdef int nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":148
 * 
 * 
 * cdef int nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":149
 * 
 * cdef int nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":148
 * 
 * 
 * cdef int nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":152
 * 
 * 
 * cdef int nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":153
 * 
 * cdef int nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":152
 * 
 * 
 * cdef int nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":156
 * 
 * 
 * cdef int nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_alltoall_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":157
 * 
 * cdef int nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":156
 * 
 * 
 * cdef int nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":160
 * 
 * 
 * cdef int nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":161
 * 
 * cdef int nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":160
 * 
 * 
 * cdef int nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":164
 * 
 * 
 * cdef int nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":165
 * 
 * cdef int nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":164
 * 
 * 
 * cdef int nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":168
 * 
 * 
 * cdef int nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":169
 * 
 * cdef int nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 169, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":168
 * 
 * 
 * cdef int nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":172
 * 
 * 
 * cdef int nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":173
 * 
 * cdef int nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":172
 * 
 * 
 * cdef int nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":176
 * 
 * 
 * cdef int nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_alltoall_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":177
 * 
 * cdef int nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_alltoall_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_alltoall_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":176
 * 
 * 
 * cdef int nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_alltoall_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_alltoall_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":180
 * 
 * 
 * cdef int nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_barrier_on_stream(team, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_on_stream(nvshmem_team_t __pyx_v_team, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":181
 * 
 * cdef int nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_barrier_on_stream(team, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_on_stream(__pyx_v_team, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":180
 * 
 * 
 * cdef int nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_barrier_on_stream(team, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_barrier_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":184
 * 
 * 
 * cdef void nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_barrier_all_on_stream(stream)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_all_on_stream(cudaStream_t __pyx_v_stream) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":185
 * 
 * cdef void nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:
 *     _nvshmem._nvshmemx_barrier_all_on_stream(stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_all_on_stream(__pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 185, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":184
 * 
 * 
 * cdef void nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_barrier_all_on_stream(stream)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_barrier_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":188
 * 
 * 
 * cdef int nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_sync_on_stream(team, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_sync_on_stream(nvshmem_team_t __pyx_v_team, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":189
 * 
 * cdef int nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_team_sync_on_stream(team, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_sync_on_stream(__pyx_v_team, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":188
 * 
 * 
 * cdef int nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_team_sync_on_stream(team, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_team_sync_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":192
 * 
 * 
 * cdef void nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_sync_all_on_stream(stream)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_sync_all_on_stream(cudaStream_t __pyx_v_stream) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":193
 * 
 * cdef void nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:
 *     _nvshmem._nvshmemx_sync_all_on_stream(stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_sync_all_on_stream(__pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 193, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":192
 * 
 * 
 * cdef void nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_sync_all_on_stream(stream)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_sync_all_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":196
 * 
 * 
 * cdef int nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":197
 * 
 * cdef int nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 197, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":196
 * 
 * 
 * cdef int nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":200
 * 
 * 
 * cdef int nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_broadcast_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":201
 * 
 * cdef int nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":200
 * 
 * 
 * cdef int nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":204
 * 
 * 
 * cdef int nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_broadcast_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":205
 * 
 * cdef int nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 205, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":204
 * 
 * 
 * cdef int nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":208
 * 
 * 
 * cdef int nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_broadcast_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":209
 * 
 * cdef int nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":208
 * 
 * 
 * cdef int nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":212
 * 
 * 
 * cdef int nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_broadcast_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":213
 * 
 * cdef int nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 213, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":212
 * 
 * 
 * cdef int nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":216
 * 
 * 
 * cdef int nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_broadcast_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":217
 * 
 * cdef int nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":216
 * 
 * 
 * cdef int nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":220
 * 
 * 
 * cdef int nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_broadcast_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":221
 * 
 * cdef int nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 221, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":220
 * 
 * 
 * cdef int nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":224
 * 
 * 
 * cdef int nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":225
 * 
 * cdef int nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 225, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":224
 * 
 * 
 * cdef int nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":228
 * 
 * 
 * cdef int nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_broadcast_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":229
 * 
 * cdef int nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 229, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":228
 * 
 * 
 * cdef int nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":232
 * 
 * 
 * cdef int nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":233
 * 
 * cdef int nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":232
 * 
 * 
 * cdef int nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":236
 * 
 * 
 * cdef int nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":237
 * 
 * cdef int nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":236
 * 
 * 
 * cdef int nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":240
 * 
 * 
 * cdef int nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":241
 * 
 * cdef int nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":240
 * 
 * 
 * cdef int nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":244
 * 
 * 
 * cdef int nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":245
 * 
 * cdef int nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":244
 * 
 * 
 * cdef int nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":248
 * 
 * 
 * cdef int nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_broadcast_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":249
 * 
 * cdef int nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":248
 * 
 * 
 * cdef int nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":252
 * 
 * 
 * cdef int nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":253
 * 
 * cdef int nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 253, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":252
 * 
 * 
 * cdef int nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":256
 * 
 * 
 * cdef int nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":257
 * 
 * cdef int nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":256
 * 
 * 
 * cdef int nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":260
 * 
 * 
 * cdef int nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":261
 * 
 * cdef int nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 261, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":260
 * 
 * 
 * cdef int nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":264
 * 
 * 
 * cdef int nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":265
 * 
 * cdef int nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 265, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":264
 * 
 * 
 * cdef int nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":268
 * 
 * 
 * cdef int nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_broadcast_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, int __pyx_v_PE_root, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":269
 * 
 * cdef int nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_broadcast_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_PE_root, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 269, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":268
 * 
 * 
 * cdef int nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_broadcast_on_stream(team, dest, src, nelem, PE_root, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_broadcast_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":272
 * 
 * 
 * cdef int nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":273
 * 
 * cdef int nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 273, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":272
 * 
 * 
 * cdef int nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":276
 * 
 * 
 * cdef int nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_fcollect_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":277
 * 
 * cdef int nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":276
 * 
 * 
 * cdef int nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":280
 * 
 * 
 * cdef int nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_fcollect_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":281
 * 
 * cdef int nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":280
 * 
 * 
 * cdef int nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":284
 * 
 * 
 * cdef int nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_fcollect_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":285
 * 
 * cdef int nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 285, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":284
 * 
 * 
 * cdef int nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":288
 * 
 * 
 * cdef int nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_fcollect_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":289
 * 
 * cdef int nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 289, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":288
 * 
 * 
 * cdef int nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":292
 * 
 * 
 * cdef int nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_fcollect_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":293
 * 
 * cdef int nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":292
 * 
 * 
 * cdef int nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":296
 * 
 * 
 * cdef int nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_fcollect_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":297
 * 
 * cdef int nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 297, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":296
 * 
 * 
 * cdef int nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":300
 * 
 * 
 * cdef int nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":301
 * 
 * cdef int nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 301, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":300
 * 
 * 
 * cdef int nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":304
 * 
 * 
 * cdef int nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_fcollect_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":305
 * 
 * cdef int nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":304
 * 
 * 
 * cdef int nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":308
 * 
 * 
 * cdef int nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":309
 * 
 * cdef int nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 309, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":308
 * 
 * 
 * cdef int nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":312
 * 
 * 
 * cdef int nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":313
 * 
 * cdef int nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":312
 * 
 * 
 * cdef int nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":316
 * 
 * 
 * cdef int nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":317
 * 
 * cdef int nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":316
 * 
 * 
 * cdef int nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":320
 * 
 * 
 * cdef int nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":321
 * 
 * cdef int nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 321, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":320
 * 
 * 
 * cdef int nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":324
 * 
 * 
 * cdef int nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_fcollect_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":325
 * 
 * cdef int nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 325, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":324
 * 
 * 
 * cdef int nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":328
 * 
 * 
 * cdef int nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":329
 * 
 * cdef int nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 329, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":328
 * 
 * 
 * cdef int nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":332
 * 
 * 
 * cdef int nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":333
 * 
 * cdef int nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":332
 * 
 * 
 * cdef int nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":336
 * 
 * 
 * cdef int nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":337
 * 
 * cdef int nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 337, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":336
 * 
 * 
 * cdef int nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":340
 * 
 * 
 * cdef int nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":341
 * 
 * cdef int nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 341, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":340
 * 
 * 
 * cdef int nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":344
 * 
 * 
 * cdef int nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_fcollect_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nelem, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":345
 * 
 * cdef int nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_fcollect_on_stream(team, dest, src, nelem, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_fcollect_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nelem, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":344
 * 
 * 
 * cdef int nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_fcollect_on_stream(team, dest, src, nelem, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_fcollect_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":348
 * 
 * 
 * cdef int nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":349
 * 
 * cdef int nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 349, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":348
 * 
 * 
 * cdef int nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":352
 * 
 * 
 * cdef int nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":353
 * 
 * cdef int nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":352
 * 
 * 
 * cdef int nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":356
 * 
 * 
 * cdef int nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":357
 * 
 * cdef int nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":356
 * 
 * 
 * cdef int nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":360
 * 
 * 
 * cdef int nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":361
 * 
 * cdef int nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 361, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":360
 * 
 * 
 * cdef int nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":364
 * 
 * 
 * cdef int nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":365
 * 
 * cdef int nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 365, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":364
 * 
 * 
 * cdef int nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":368
 * 
 * 
 * cdef int nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":369
 * 
 * cdef int nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 369, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":368
 * 
 * 
 * cdef int nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":372
 * 
 * 
 * cdef int nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":373
 * 
 * cdef int nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 373, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":372
 * 
 * 
 * cdef int nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":376
 * 
 * 
 * cdef int nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":377
 * 
 * cdef int nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":376
 * 
 * 
 * cdef int nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":380
 * 
 * 
 * cdef int nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":381
 * 
 * cdef int nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":380
 * 
 * 
 * cdef int nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":384
 * 
 * 
 * cdef int nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":385
 * 
 * cdef int nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 385, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":384
 * 
 * 
 * cdef int nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":388
 * 
 * 
 * cdef int nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":389
 * 
 * cdef int nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 389, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":388
 * 
 * 
 * cdef int nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":392
 * 
 * 
 * cdef int nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":393
 * 
 * cdef int nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":392
 * 
 * 
 * cdef int nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":396
 * 
 * 
 * cdef int nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":397
 * 
 * cdef int nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 397, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":396
 * 
 * 
 * cdef int nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":400
 * 
 * 
 * cdef int nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":401
 * 
 * cdef int nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 401, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":400
 * 
 * 
 * cdef int nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":404
 * 
 * 
 * cdef int nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":405
 * 
 * cdef int nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":404
 * 
 * 
 * cdef int nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":408
 * 
 * 
 * cdef int nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":409
 * 
 * cdef int nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 409, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":408
 * 
 * 
 * cdef int nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":412
 * 
 * 
 * cdef int nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":413
 * 
 * cdef int nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 413, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":412
 * 
 * 
 * cdef int nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":416
 * 
 * 
 * cdef int nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":417
 * 
 * cdef int nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 417, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":416
 * 
 * 
 * cdef int nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":420
 * 
 * 
 * cdef int nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":421
 * 
 * cdef int nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_max_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 421, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":420
 * 
 * 
 * cdef int nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_max_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_max_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":424
 * 
 * 
 * cdef int nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":425
 * 
 * cdef int nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 425, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":424
 * 
 * 
 * cdef int nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":428
 * 
 * 
 * cdef int nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":429
 * 
 * cdef int nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 429, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":428
 * 
 * 
 * cdef int nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":432
 * 
 * 
 * cdef int nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":433
 * 
 * cdef int nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 433, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":432
 * 
 * 
 * cdef int nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":436
 * 
 * 
 * cdef int nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":437
 * 
 * cdef int nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 437, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":436
 * 
 * 
 * cdef int nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":440
 * 
 * 
 * cdef int nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":441
 * 
 * cdef int nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 441, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":440
 * 
 * 
 * cdef int nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":444
 * 
 * 
 * cdef int nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":445
 * 
 * cdef int nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":444
 * 
 * 
 * cdef int nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":448
 * 
 * 
 * cdef int nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":449
 * 
 * cdef int nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":448
 * 
 * 
 * cdef int nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":452
 * 
 * 
 * cdef int nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":453
 * 
 * cdef int nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 453, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":452
 * 
 * 
 * cdef int nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":456
 * 
 * 
 * cdef int nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":457
 * 
 * cdef int nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":456
 * 
 * 
 * cdef int nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":460
 * 
 * 
 * cdef int nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":461
 * 
 * cdef int nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 461, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":460
 * 
 * 
 * cdef int nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":464
 * 
 * 
 * cdef int nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":465
 * 
 * cdef int nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":464
 * 
 * 
 * cdef int nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":468
 * 
 * 
 * cdef int nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":469
 * 
 * cdef int nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 469, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":468
 * 
 * 
 * cdef int nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":472
 * 
 * 
 * cdef int nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":473
 * 
 * cdef int nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":472
 * 
 * 
 * cdef int nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":476
 * 
 * 
 * cdef int nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":477
 * 
 * cdef int nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":476
 * 
 * 
 * cdef int nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":480
 * 
 * 
 * cdef int nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":481
 * 
 * cdef int nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 481, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":480
 * 
 * 
 * cdef int nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":484
 * 
 * 
 * cdef int nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":485
 * 
 * cdef int nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 485, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":484
 * 
 * 
 * cdef int nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":488
 * 
 * 
 * cdef int nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":489
 * 
 * cdef int nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 489, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":488
 * 
 * 
 * cdef int nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":492
 * 
 * 
 * cdef int nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":493
 * 
 * cdef int nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 493, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":492
 * 
 * 
 * cdef int nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":496
 * 
 * 
 * cdef int nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":497
 * 
 * cdef int nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_min_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":496
 * 
 * 
 * cdef int nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_min_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_min_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":500
 * 
 * 
 * cdef int nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":501
 * 
 * cdef int nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":500
 * 
 * 
 * cdef int nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":504
 * 
 * 
 * cdef int nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":505
 * 
 * cdef int nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 505, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":504
 * 
 * 
 * cdef int nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":508
 * 
 * 
 * cdef int nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":509
 * 
 * cdef int nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 509, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":508
 * 
 * 
 * cdef int nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":512
 * 
 * 
 * cdef int nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":513
 * 
 * cdef int nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 513, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":512
 * 
 * 
 * cdef int nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":516
 * 
 * 
 * cdef int nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":517
 * 
 * cdef int nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 517, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":516
 * 
 * 
 * cdef int nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":520
 * 
 * 
 * cdef int nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":521
 * 
 * cdef int nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 521, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":520
 * 
 * 
 * cdef int nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":524
 * 
 * 
 * cdef int nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":525
 * 
 * cdef int nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 525, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":524
 * 
 * 
 * cdef int nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":528
 * 
 * 
 * cdef int nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":529
 * 
 * cdef int nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 529, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":528
 * 
 * 
 * cdef int nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":532
 * 
 * 
 * cdef int nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":533
 * 
 * cdef int nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 533, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":532
 * 
 * 
 * cdef int nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":536
 * 
 * 
 * cdef int nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":537
 * 
 * cdef int nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 537, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":536
 * 
 * 
 * cdef int nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":540
 * 
 * 
 * cdef int nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":541
 * 
 * cdef int nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 541, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":540
 * 
 * 
 * cdef int nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":544
 * 
 * 
 * cdef int nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":545
 * 
 * cdef int nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":544
 * 
 * 
 * cdef int nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":548
 * 
 * 
 * cdef int nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":549
 * 
 * cdef int nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":548
 * 
 * 
 * cdef int nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":552
 * 
 * 
 * cdef int nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":553
 * 
 * cdef int nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 553, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":552
 * 
 * 
 * cdef int nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":556
 * 
 * 
 * cdef int nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":557
 * 
 * cdef int nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 557, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":556
 * 
 * 
 * cdef int nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":560
 * 
 * 
 * cdef int nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":561
 * 
 * cdef int nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 561, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":560
 * 
 * 
 * cdef int nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":564
 * 
 * 
 * cdef int nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":565
 * 
 * cdef int nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 565, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":564
 * 
 * 
 * cdef int nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":568
 * 
 * 
 * cdef int nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":569
 * 
 * cdef int nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 569, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":568
 * 
 * 
 * cdef int nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":572
 * 
 * 
 * cdef int nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":573
 * 
 * cdef int nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_sum_reduce_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reduce_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":572
 * 
 * 
 * cdef int nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_sum_reduce_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_sum_reduce_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":576
 * 
 * 
 * cdef int nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":577
 * 
 * cdef int nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 577, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":576
 * 
 * 
 * cdef int nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":580
 * 
 * 
 * cdef int nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":581
 * 
 * cdef int nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":580
 * 
 * 
 * cdef int nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":584
 * 
 * 
 * cdef int nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":585
 * 
 * cdef int nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 585, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":584
 * 
 * 
 * cdef int nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":588
 * 
 * 
 * cdef int nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":589
 * 
 * cdef int nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 589, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":588
 * 
 * 
 * cdef int nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":592
 * 
 * 
 * cdef int nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":593
 * 
 * cdef int nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 593, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":592
 * 
 * 
 * cdef int nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":596
 * 
 * 
 * cdef int nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":597
 * 
 * cdef int nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 597, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":596
 * 
 * 
 * cdef int nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":600
 * 
 * 
 * cdef int nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":601
 * 
 * cdef int nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 601, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":600
 * 
 * 
 * cdef int nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":604
 * 
 * 
 * cdef int nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":605
 * 
 * cdef int nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":604
 * 
 * 
 * cdef int nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":608
 * 
 * 
 * cdef int nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":609
 * 
 * cdef int nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 609, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":608
 * 
 * 
 * cdef int nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":612
 * 
 * 
 * cdef int nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":613
 * 
 * cdef int nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 613, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":612
 * 
 * 
 * cdef int nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":616
 * 
 * 
 * cdef int nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":617
 * 
 * cdef int nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 617, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":616
 * 
 * 
 * cdef int nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":620
 * 
 * 
 * cdef int nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":621
 * 
 * cdef int nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 621, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":620
 * 
 * 
 * cdef int nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":624
 * 
 * 
 * cdef int nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":625
 * 
 * cdef int nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 625, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":624
 * 
 * 
 * cdef int nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":628
 * 
 * 
 * cdef int nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":629
 * 
 * cdef int nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 629, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":628
 * 
 * 
 * cdef int nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":632
 * 
 * 
 * cdef int nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":633
 * 
 * cdef int nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 633, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":632
 * 
 * 
 * cdef int nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":636
 * 
 * 
 * cdef int nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":637
 * 
 * cdef int nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 637, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":636
 * 
 * 
 * cdef int nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":640
 * 
 * 
 * cdef int nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":641
 * 
 * cdef int nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 641, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":640
 * 
 * 
 * cdef int nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":644
 * 
 * 
 * cdef int nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":645
 * 
 * cdef int nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 645, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":644
 * 
 * 
 * cdef int nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":648
 * 
 * 
 * cdef int nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":649
 * 
 * cdef int nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_max_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 649, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":648
 * 
 * 
 * cdef int nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_max_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_max_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":652
 * 
 * 
 * cdef int nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":653
 * 
 * cdef int nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 653, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":652
 * 
 * 
 * cdef int nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":656
 * 
 * 
 * cdef int nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":657
 * 
 * cdef int nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 657, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":656
 * 
 * 
 * cdef int nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":660
 * 
 * 
 * cdef int nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":661
 * 
 * cdef int nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":660
 * 
 * 
 * cdef int nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":664
 * 
 * 
 * cdef int nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":665
 * 
 * cdef int nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 665, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":664
 * 
 * 
 * cdef int nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":668
 * 
 * 
 * cdef int nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":669
 * 
 * cdef int nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":668
 * 
 * 
 * cdef int nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":672
 * 
 * 
 * cdef int nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":673
 * 
 * cdef int nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 673, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":672
 * 
 * 
 * cdef int nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":676
 * 
 * 
 * cdef int nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":677
 * 
 * cdef int nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 677, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":676
 * 
 * 
 * cdef int nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":680
 * 
 * 
 * cdef int nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":681
 * 
 * cdef int nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":680
 * 
 * 
 * cdef int nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":684
 * 
 * 
 * cdef int nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":685
 * 
 * cdef int nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 685, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":684
 * 
 * 
 * cdef int nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":688
 * 
 * 
 * cdef int nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":689
 * 
 * cdef int nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 689, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":688
 * 
 * 
 * cdef int nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":692
 * 
 * 
 * cdef int nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":693
 * 
 * cdef int nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 693, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":692
 * 
 * 
 * cdef int nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":696
 * 
 * 
 * cdef int nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":697
 * 
 * cdef int nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":696
 * 
 * 
 * cdef int nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":700
 * 
 * 
 * cdef int nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":701
 * 
 * cdef int nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 701, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":700
 * 
 * 
 * cdef int nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":704
 * 
 * 
 * cdef int nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":705
 * 
 * cdef int nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 705, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":704
 * 
 * 
 * cdef int nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":708
 * 
 * 
 * cdef int nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":709
 * 
 * cdef int nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 709, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":708
 * 
 * 
 * cdef int nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":712
 * 
 * 
 * cdef int nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":713
 * 
 * cdef int nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 713, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":712
 * 
 * 
 * cdef int nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":716
 * 
 * 
 * cdef int nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":717
 * 
 * cdef int nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 717, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":716
 * 
 * 
 * cdef int nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":720
 * 
 * 
 * cdef int nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":721
 * 
 * cdef int nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 721, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":720
 * 
 * 
 * cdef int nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":724
 * 
 * 
 * cdef int nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":725
 * 
 * cdef int nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_min_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 725, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":724
 * 
 * 
 * cdef int nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_min_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_min_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":728
 * 
 * 
 * cdef int nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int8_t *__pyx_v_dest, int8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":729
 * 
 * cdef int nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 729, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":728
 * 
 * 
 * cdef int nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":732
 * 
 * 
 * cdef int nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int16_t *__pyx_v_dest, int16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":733
 * 
 * cdef int nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 733, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":732
 * 
 * 
 * cdef int nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":736
 * 
 * 
 * cdef int nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int32_t *__pyx_v_dest, int32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":737
 * 
 * cdef int nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 737, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":736
 * 
 * 
 * cdef int nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":740
 * 
 * 
 * cdef int nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int64_t *__pyx_v_dest, int64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":741
 * 
 * cdef int nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":740
 * 
 * 
 * cdef int nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":744
 * 
 * 
 * cdef int nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint8_t *__pyx_v_dest, uint8_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":745
 * 
 * cdef int nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 745, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":744
 * 
 * 
 * cdef int nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint8_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint8_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":748
 * 
 * 
 * cdef int nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint16_t *__pyx_v_dest, uint16_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":749
 * 
 * cdef int nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 749, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":748
 * 
 * 
 * cdef int nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":752
 * 
 * 
 * cdef int nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint32_t *__pyx_v_dest, uint32_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":753
 * 
 * cdef int nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 753, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":752
 * 
 * 
 * cdef int nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint32_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint32_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":756
 * 
 * 
 * cdef int nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, uint64_t *__pyx_v_dest, uint64_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":757
 * 
 * cdef int nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_uint64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 757, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":756
 * 
 * 
 * cdef int nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_uint64_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_uint64_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":760
 * 
 * 
 * cdef int nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, size_t *__pyx_v_dest, size_t const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":761
 * 
 * cdef int nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_size_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 761, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":760
 * 
 * 
 * cdef int nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_size_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_size_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":764
 * 
 * 
 * cdef int nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, char *__pyx_v_dest, char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":765
 * 
 * cdef int nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_char_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 765, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":764
 * 
 * 
 * cdef int nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_char_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_char_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":768
 * 
 * 
 * cdef int nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, signed char *__pyx_v_dest, signed char const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":769
 * 
 * cdef int nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_schar_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 769, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":768
 * 
 * 
 * cdef int nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_schar_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_schar_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":772
 * 
 * 
 * cdef int nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, short *__pyx_v_dest, short const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":773
 * 
 * cdef int nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_short_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":772
 * 
 * 
 * cdef int nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_short_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_short_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":776
 * 
 * 
 * cdef int nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, int *__pyx_v_dest, int const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":777
 * 
 * cdef int nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_int_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 777, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":776
 * 
 * 
 * cdef int nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_int_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_int_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":780
 * 
 * 
 * cdef int nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, long *__pyx_v_dest, long const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":781
 * 
 * cdef int nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_long_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 781, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":780
 * 
 * 
 * cdef int nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_long_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_long_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":784
 * 
 * 
 * cdef int nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, PY_LONG_LONG *__pyx_v_dest, PY_LONG_LONG const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":785
 * 
 * cdef int nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_longlong_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":784
 * 
 * 
 * cdef int nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_longlong_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_longlong_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":788
 * 
 * 
 * cdef int nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, __nv_bfloat16 *__pyx_v_dest, __nv_bfloat16 const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":789
 * 
 * cdef int nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_bfloat16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 789, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":788
 * 
 * 
 * cdef int nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_bfloat16_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_bfloat16_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":792
 * 
 * 
 * cdef int nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, half *__pyx_v_dest, half const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":793
 * 
 * cdef int nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_half_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 793, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":792
 * 
 * 
 * cdef int nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_half_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_half_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":796
 * 
 * 
 * cdef int nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, float *__pyx_v_dest, float const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":797
 * 
 * cdef int nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_float_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 797, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":796
 * 
 * 
 * cdef int nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_float_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_float_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":800
 * 
 * 
 * cdef int nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t __pyx_v_team, double *__pyx_v_dest, double const *__pyx_v_src, size_t __pyx_v_nreduce, cudaStream_t __pyx_v_stream) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":801
 * 
 * cdef int nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:
 *     return _nvshmem._nvshmemx_double_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reducescatter_on_stream(__pyx_v_team, __pyx_v_dest, __pyx_v_src, __pyx_v_nreduce, __pyx_v_stream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 801, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":800
 * 
 * 
 * cdef int nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_double_sum_reducescatter_on_stream(team, dest, src, nreduce, stream)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_double_sum_reducescatter_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":804
 * 
 * 
 * cdef int nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_hostlib_init_attr(flags, attr)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_init_attr(unsigned int __pyx_v_flags, nvshmemx_init_attr_t *__pyx_v_attr) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":805
 * 
 * cdef int nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:
 *     return _nvshmem._nvshmemx_hostlib_init_attr(flags, attr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_init_attr(__pyx_v_flags, __pyx_v_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 805, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":804
 * 
 * 
 * cdef int nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_hostlib_init_attr(flags, attr)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_hostlib_init_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":808
 * 
 * 
 * cdef void nvshmemx_hostlib_finalize() except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_hostlib_finalize()
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_finalize(void) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":809
 * 
 * cdef void nvshmemx_hostlib_finalize() except* nogil:
 *     _nvshmem._nvshmemx_hostlib_finalize()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_finalize(); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 809, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":808
 * 
 * 
 * cdef void nvshmemx_hostlib_finalize() except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_hostlib_finalize()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_hostlib_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":812
 * 
 * 
 * cdef int nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_set_attr_uniqueid_args(myrank, nranks, uniqueid, attr)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_uniqueid_args(int const __pyx_v_myrank, int const __pyx_v_nranks, nvshmemx_uniqueid_t const *__pyx_v_uniqueid, nvshmemx_init_attr_t *__pyx_v_attr) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":813
 * 
 * cdef int nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:
 *     return _nvshmem._nvshmemx_set_attr_uniqueid_args(myrank, nranks, uniqueid, attr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_uniqueid_args(__pyx_v_myrank, __pyx_v_nranks, __pyx_v_uniqueid, __pyx_v_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 813, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":812
 * 
 * 
 * cdef int nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_set_attr_uniqueid_args(myrank, nranks, uniqueid, attr)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_set_attr_uniqueid_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":816
 * 
 * 
 * cdef int nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_set_attr_mpi_comm_args(mpi_comm, nvshmem_attr)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_mpi_comm_args(void *__pyx_v_mpi_comm, nvshmemx_init_attr_t *__pyx_v_nvshmem_attr) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":817
 * 
 * cdef int nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:
 *     return _nvshmem._nvshmemx_set_attr_mpi_comm_args(mpi_comm, nvshmem_attr)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_mpi_comm_args(__pyx_v_mpi_comm, __pyx_v_nvshmem_attr); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 817, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":816
 * 
 * 
 * cdef int nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_set_attr_mpi_comm_args(mpi_comm, nvshmem_attr)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_set_attr_mpi_comm_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":820
 * 
 * 
 * cdef int nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_get_uniqueid(uniqueid)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_get_uniqueid(nvshmemx_uniqueid_t *__pyx_v_uniqueid) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":821
 * 
 * cdef int nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:
 *     return _nvshmem._nvshmemx_get_uniqueid(uniqueid)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_get_uniqueid(__pyx_v_uniqueid); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 821, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":820
 * 
 * 
 * cdef int nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_get_uniqueid(uniqueid)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_get_uniqueid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":824
 * 
 * 
 * cdef int nvshmemx_cumodule_init(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_cumodule_init(module)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_init(CUmodule __pyx_v_module) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":825
 * 
 * cdef int nvshmemx_cumodule_init(CUmodule module) except* nogil:
 *     return _nvshmem._nvshmemx_cumodule_init(module)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_init(__pyx_v_module); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 825, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":824
 * 
 * 
 * cdef int nvshmemx_cumodule_init(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_cumodule_init(module)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_cumodule_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":828
 * 
 * 
 * cdef int nvshmemx_cumodule_finalize(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_cumodule_finalize(module)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_finalize(CUmodule __pyx_v_module) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":829
 * 
 * cdef int nvshmemx_cumodule_finalize(CUmodule module) except* nogil:
 *     return _nvshmem._nvshmemx_cumodule_finalize(module)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_finalize(__pyx_v_module); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 829, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":828
 * 
 * 
 * cdef int nvshmemx_cumodule_finalize(CUmodule module) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_cumodule_finalize(module)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_cumodule_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":832
 * 
 * 
 * cdef void* nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_buffer_register_symmetric(buf_ptr, size, flags)
 * 
*/

static void *__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_register_symmetric(void *__pyx_v_buf_ptr, size_t __pyx_v_size, int __pyx_v_flags) {
  void *__pyx_r;
  void *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":833
 * 
 * cdef void* nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:
 *     return _nvshmem._nvshmemx_buffer_register_symmetric(buf_ptr, size, flags)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_register_symmetric(__pyx_v_buf_ptr, __pyx_v_size, __pyx_v_flags); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 833, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":832
 * 
 * 
 * cdef void* nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_buffer_register_symmetric(buf_ptr, size, flags)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_buffer_register_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":836
 * 
 * 
 * cdef int nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_buffer_unregister_symmetric(mmap_ptr, size)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_unregister_symmetric(void *__pyx_v_mmap_ptr, size_t __pyx_v_size) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":837
 * 
 * cdef int nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:
 *     return _nvshmem._nvshmemx_buffer_unregister_symmetric(mmap_ptr, size)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_unregister_symmetric(__pyx_v_mmap_ptr, __pyx_v_size); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 837, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":836
 * 
 * 
 * cdef int nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_buffer_unregister_symmetric(mmap_ptr, size)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_buffer_unregister_symmetric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":840
 * 
 * 
 * cdef int nvshmemx_culibrary_init(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_culibrary_init(library)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_init(CUlibrary __pyx_v_library) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":841
 * 
 * cdef int nvshmemx_culibrary_init(CUlibrary library) except* nogil:
 *     return _nvshmem._nvshmemx_culibrary_init(library)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_init(__pyx_v_library); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 841, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":840
 * 
 * 
 * cdef int nvshmemx_culibrary_init(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_culibrary_init(library)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_culibrary_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":844
 * 
 * 
 * cdef int nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_culibrary_finalize(library)
 * 
*/

static int __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_finalize(CUlibrary __pyx_v_library) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":845
 * 
 * cdef int nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:
 *     return _nvshmem._nvshmemx_culibrary_finalize(library)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_finalize(__pyx_v_library); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 845, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "nvshmem/bindings/cynvshmem.pyx":844
 * 
 * 
 * cdef int nvshmemx_culibrary_finalize(CUlibrary library) except* nogil:             # <<<<<<<<<<<<<<
 *     return _nvshmem._nvshmemx_culibrary_finalize(library)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_culibrary_finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
  return __pyx_r;
}

/* "nvshmem/bindings/cynvshmem.pyx":848
 * 
 * 
 * cdef void nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_putmem_on_stream(dest, source, bytes, pe, cstrm)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":849
 * 
 * cdef void nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:
 *     _nvshmem._nvshmemx_putmem_on_stream(dest, source, bytes, pe, cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 849, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":848
 * 
 * 
 * cdef void nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_putmem_on_stream(dest, source, bytes, pe, cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_putmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":852
 * 
 * 
 * cdef void nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_putmem_signal_on_stream(dest, source, bytes, sig_addr, signal, sig_op, pe, cstrm)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_signal_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, uint64_t *__pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":853
 * 
 * cdef void nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:
 *     _nvshmem._nvshmemx_putmem_signal_on_stream(dest, source, bytes, sig_addr, signal, sig_op, pe, cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_signal_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 853, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":852
 * 
 * 
 * cdef void nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_putmem_signal_on_stream(dest, source, bytes, sig_addr, signal, sig_op, pe, cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_putmem_signal_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":856
 * 
 * 
 * cdef void nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_getmem_on_stream(dest, source, bytes, pe, cstrm)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_getmem_on_stream(void *__pyx_v_dest, void const *__pyx_v_source, size_t __pyx_v_bytes, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":857
 * 
 * cdef void nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:
 *     _nvshmem._nvshmemx_getmem_on_stream(dest, source, bytes, pe, cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_getmem_on_stream(__pyx_v_dest, __pyx_v_source, __pyx_v_bytes, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 857, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":856
 * 
 * 
 * cdef void nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_getmem_on_stream(dest, source, bytes, pe, cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_getmem_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":860
 * 
 * 
 * cdef void nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_quiet_on_stream(cstrm)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_quiet_on_stream(cudaStream_t __pyx_v_cstrm) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":861
 * 
 * cdef void nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:
 *     _nvshmem._nvshmemx_quiet_on_stream(cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_quiet_on_stream(__pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 861, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":860
 * 
 * 
 * cdef void nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_quiet_on_stream(cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_quiet_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":864
 * 
 * 
 * cdef void nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_signal_op_on_stream(sig_addr, signal, sig_op, pe, cstrm)
 * 
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_op_on_stream(uint64_t *__pyx_v_sig_addr, uint64_t __pyx_v_signal, int __pyx_v_sig_op, int __pyx_v_pe, cudaStream_t __pyx_v_cstrm) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":865
 * 
 * cdef void nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:
 *     _nvshmem._nvshmemx_signal_op_on_stream(sig_addr, signal, sig_op, pe, cstrm)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_op_on_stream(__pyx_v_sig_addr, __pyx_v_signal, __pyx_v_sig_op, __pyx_v_pe, __pyx_v_cstrm); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 865, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":864
 * 
 * 
 * cdef void nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_signal_op_on_stream(sig_addr, signal, sig_op, pe, cstrm)
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_signal_op_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}

/* "nvshmem/bindings/cynvshmem.pyx":868
 * 
 * 
 * cdef void nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_signal_wait_until_on_stream(sig_addr, cmp, cmp_value, cstream)
*/

static void __pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_wait_until_on_stream(uint64_t *__pyx_v_sig_addr, int __pyx_v_cmp, uint64_t __pyx_v_cmp_value, cudaStream_t __pyx_v_cstream) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyGILState_STATE __pyx_gilstate_save;

  /* "nvshmem/bindings/cynvshmem.pyx":869
 * 
 * cdef void nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:
 *     _nvshmem._nvshmemx_signal_wait_until_on_stream(sig_addr, cmp, cmp_value, cstream)             # <<<<<<<<<<<<<<
*/
  __pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_wait_until_on_stream(__pyx_v_sig_addr, __pyx_v_cmp, __pyx_v_cmp_value, __pyx_v_cstream); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 869, __pyx_L1_error)

  /* "nvshmem/bindings/cynvshmem.pyx":868
 * 
 * 
 * cdef void nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil:             # <<<<<<<<<<<<<<
 *     _nvshmem._nvshmemx_signal_wait_until_on_stream(sig_addr, cmp, cmp_value, cstream)
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  __Pyx_AddTraceback("nvshmem.bindings.cynvshmem.nvshmemx_signal_wait_until_on_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  __pyx_L0:;
}
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  {
    __pyx_t_1 = __Pyx_ApiExport_GetApiDict(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_export_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_int_CUlibrary_int_CUmodule_int_i);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_export_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_export_name = __pyx_export_signature + 4132;
    void (*const __pyx_export_pointers[])(void) = {(void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_culibrary_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_cumodule_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_uniqueid_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_init, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_longlong_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_bfloat16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_char_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_sync_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_double_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_float_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_half_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_strided, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_split_2d, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_translate_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_int8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_long_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_short_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_schar_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_size_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint16_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint32_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint64_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_alltoall_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_fcollect_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_max_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_min_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reduce_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_sum_reducescatter_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_uint8_broadcast_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_team_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_get_uniqueid, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_init_attr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_set_attr_mpi_comm_args, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_unregister_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_my_pe, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_n_pes, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_init_status, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_barrier_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_quiet_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_sync_all_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_info_get_version, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_vendor_get_version_info, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_destroy, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_team_get_config, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_wait_until_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_signal_op_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_free, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_getmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_putmem_signal_on_stream, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_barrier_all, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_hostlib_finalize, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_mc_ptr, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_malloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_align, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_calloc, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmemx_buffer_register_symmetric, (void (*)(void))&__pyx_f_7nvshmem_8bindings_9cynvshmem_nvshmem_ptr, (void (*)(void)) NULL};
    void (*const *__pyx_export_pointer)(void) = __pyx_export_pointers;
    const char *__pyx_export_current_signature = __pyx_export_signature;
    while (*__pyx_export_pointer) {
      if (__Pyx_ExportFunction(__pyx_t_1, __pyx_export_name, *__pyx_export_pointer, __pyx_export_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_export_pointer;
      __pyx_export_name = strchr(__pyx_export_name, '\0') + 1;
      __pyx_export_signature = strchr(__pyx_export_signature, '\0') + 1;
      if (*__pyx_export_signature != '\0') __pyx_export_current_signature = __pyx_export_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("nvshmem.bindings._internal.nvshmem"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_int_CUlibrary_int_CUmodule_int_i_2);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 4132;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_culibrary_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_cumodule_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_uniqueid_args, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_init, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_my_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_n_pes, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_longlong_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_bfloat16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_char_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_sync_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_double_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_float_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_half_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_strided, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_split_2d, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_translate_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int32_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int64_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_int8_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_long_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_short_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_schar_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_size_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint16_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint32_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint64_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_alltoall_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_fcollect_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_max_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_min_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reduce_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_sum_reducescatter_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_uint8_broadcast_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_team_get_uniqueid, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_get_uniqueid, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_init_attr, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_set_attr_mpi_comm_args, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_unregister_symmetric, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_my_pe, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_n_pes, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_init_status, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_barrier_all_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_quiet_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_sync_all_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_info_get_version, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_vendor_get_version_info, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_destroy, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_team_get_config, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_wait_until_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_signal_op_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_free, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_getmem_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_putmem_signal_on_stream, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_barrier_all, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_hostlib_finalize, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_mc_ptr, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_malloc, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_align, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_calloc, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmemx_buffer_register_symmetric, (void (**)(void))&__pyx_f_7nvshmem_8bindings_9_internal_7nvshmem__nvshmem_ptr, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_8(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cynvshmem(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cynvshmem},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "cynvshmem",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_cynvshmem(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cynvshmem(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cynvshmem(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cynvshmem' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "cynvshmem" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_cynvshmem", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_nvshmem__bindings__cynvshmem) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "nvshmem.bindings.cynvshmem")) {
      if (unlikely((PyDict_SetItemString(modules, "nvshmem.bindings.cynvshmem", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_export_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  (void)__Pyx_modinit_type_import_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "nvshmem/bindings/cynvshmem.pyx":1
 * # Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.             # <<<<<<<<<<<<<<
 * # SPDX-License-Identifier: Apache-2.0
 * #
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init nvshmem.bindings.cynvshmem", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init nvshmem.bindings.cynvshmem");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 14; } index[] = {{1},{18},{8},{10},{8},{12},{12},{10},{8},{11654},{11868}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (1856 bytes) */
const char* const cstring = "(\265/\375`9[\2659\000\326\033M\035@ss\217\210\204eKB\206\2505\331M\n\321\356f\357\001L\321ug9\346(zkC\000K\000C\000k5Z\226\3433M\033\005^\330\0318\r\331)\000@\240y\371\240e\222E\300i\336\277\241\366\337,f\244\253\025\262\214\226\377\337)\3253\232\251%_Y\\\356\347\315\366WTB\252\362\365y\257\274L)\002,y\243G%x\002\217\263\326]8f\342R\226\3669\nL\034\020\005\003\202\3608\021K\312\343(\203\242[\217\363\222\242\036\007A\350t,\202~y\035\311\025\225n\026>\016\303\361h|\002\221\3301i\243\267\024\374u\347{y\237\007\rZ\326\320QZ\245\270\331\204N\327N\0069_\027\226\256\355e\023\257b\r\204\251V\272\007\265\010y\237v\247\202\252\342\322\367L\266V\312\230S\021<\273\273f,\317:]\323\236\356{\t\035\000\245\227\n\003W\377\377\377\205\377\377\025K\356\365t\325\342\244\372\242\311mob\316\222%\215{\ru=\267\230\335\302\247\027\027\351F\013\214\235\r\325\276\326\3264\360O\033\256\244\217\343\237FQ\204Zs\352;\005\203\217\250\362\256\021IRPP\3122\006B\251\010\304\2211I\322P=B\030\014\022\002!\010C!(\204 \010\303 \013\210\200\010\210\300\220\022\010\005&\026(\303v_\356\204\227S\362\"E\022\323p\271kp\313\271h97\345uE6<v\363\030XX\343\324\312)\031\203\252\234O*'\010\211M\271\005\203\224[Qn\271\214\031(w\244\376\344\334N\316\007\251\311\255,\315\344\332\023]\231\270\027+\271SA[\333h;\323Mp\344\324n\344b\224\"\267j$\3751\344z\002\203\231\004\261~T\360\203\027\017? c\366\377\2635\321\3538\356\274\350!\010Z\200X\311G\006\234b\\\331\216\265]\256\222<\360\225\374\310\300)j+;\355\205\241e|\262\236.\254\303\"X\351\273`\333\232'\350\207cW\237L\316jL\212\3705\334\325t=\277\013G\257]\202\205c\033\246\002\310\247\2547\217\206j\326\2553\270\301Q\004;K\n\257\226T\227/\202;\327\013z-|\314Z\0236\201c<J!\334s\"w&\365\201/e8c\336\303UIi\251\343Z\024[\316\367\240C\367\222\"\277g\354\224$\034\215\026Yd\n\310\317I6\330\024@$0\"\370Q5\261\235\003\016\271\t\274s@\324\270R\220\010D\321 \354\026r[P\314\220u\273\027\322\356c\352v\216\223\014'\240 \006\275\320s\323U\203YT\274\241\n$\241/\267\277R\353\204,<\3511d\334""\264\327\361l]\\\364b\241\204\333\355`2\235 \206G=\202\3676M?\242e\306jg0\364\315m?\252\303I\237(\326\231b\263\320\335w*\227v\253\334a0a[\360\265\035\374\335\265\255\253\014\321;\337\026\265\303=\025\342b\t\222u\201\363\353Q\004-XA\036\303\262\214^{=\215\230\234\255\205)\r\346N\271\026\320f\353\327s\224:\216\312ub\024,\327\034\230tV\223\255\307.\331\210\350\221m#d;\303\234\205*\310\360\216\255\272\203\374\213\261ug\025\236G\300\027\202b\013\306\020Hlo\2362\005B\231\2102\r`\035\225`\246\n<\2766@\022MyVF\343\007\264\302p#\350(\305\257u9\3265\230\020\264\034\274\254\305!\356!\366$\333f\3260t\334\006\311jJ\220\244\256\300\014\365o\236\354\204I\321)*\340\337\n\3141GC\203\375\255\323\202\002)\354L\374\356QQ\253<j\223=\272\245\375\245\244Vbv\246\213/\t\310ajc\244ufu\257\236`\"+\022$S\203*^:>\336u\033\257)\201H\345\014\250X]L\215\260O\001\275s\016\031\372\333\024\321\313\216P\231\312\210\237%\302\364\031b\322GR\211`/\r\342b9ut\260\372\025^F\237\250\360\221\032&|!Q\216;\365\323-\224$\224\215Ji\003\263V\323\250\374r\325\020N$\261\001^p\204JR\"\355,\017\346\013\350\222&\366\316ru\314\221*N\322>?\325pUpyN\024\336\240f\211~\312\006\376\365\245\317\211m\352\2544;\023J\311H^0g\237\223\237\227P(!o^\nitkb\026\3253T\021\022A\254\330\024\317\000\365I\223\236;=6\253\342\222\331U\207Y\235u\n\225\244D\334Y\036\314\013\370\222yE\026\010j\224\3343\246=\211%\251\2227x\t\364\263,G\243\244\226t\367\220\262\375\366\016I\244\276\203\372n\"\277+t\213l\020:R\321Y\305I\313\342\300\256\304R\244e\372\303\226\005T7N\374d\n\025\ta\223\245\222\346\267\026\323\251\374C\325\020\n\253\266A\0350Wf\r\2245\231\224'l\221NT\235\317v1X\257/\324\257\206E\250\017&\276,Pv\246\022uH\367\004=\350&#\274)\216\311#\346\226!)\001p\334- @\362\030\273\235\312\270o\365\020\262\371\274\200F\021\247f\315.\255\032\223(7\342\276&\371N\313ZV\351\307oP\226\320\343\250\223\234^\342\340\323(^}\341k\225\t\n\363\312\036\371\241\251'\216Y\3248\233\350\311Bx\200\013\323\000\313\231\361\260\002\364\215P\204\024O\303\356\233=\252\014""\017{\026m\271\252\277\034\026\234\023 o/r\347\346\017\025\026$@\265\002o7\212\246\264\366m\367}\236\345%\274\336\017\332\227\227.Y<\320\301\265T\356H\022s\320\315h\233\0039c3\001\345P\375\204\273\354\335&\331\033\311\3365c\0100\260\356\n'B\210h\036V>\210v\366\246\027\354\031&\241\360\223\264k\014\004\225\r\370\034\341q\370z\220M\262j\034\016\264\004V\322K%*\001\003\323\034\000\265J\007[\024\345\303\3378E\004\344\340\236\334\324\325=)\310\360\333O\n$\341\370\367\003\374!\016\231\314\321A\373\313]\364\000q\206\303\307\304#\270\256\314\306\267v\237Q8\364\"=kS\364tW\3564\201L\270\020~\214\343_\344\333\022\245\302_c\375\370R\365\314\267'\312\206\177\221\2103\362;}[a\224F\3236\332\240\363\304\353\340\323kU\205Fq\242\036\332\t$\200\200\240r.\242\r\245\275\356\005\236\021\377e\037\336r\2014\274F\003.\t\344\301E\360\232\271\276\2140_\354\235 !o\375\225\n\032\023.\210q\221g2\352\277'\373\263e\356\275\221\356\351\316\332\030\251m\2612B_\334\222\265\022\272\2633C\224M-\205\343\314\000\363~K\376F:T)\254\303\350\030\314\255gez\226\004d\024\001Ub\265\006\202\002\277\272>\376\315\210H25R\345\371D\247\330\264J\005t\007\0348b\344\241=|q\341\371\320\250Z!\031\221nY\n\032\313\034\001\225\233D\332";
    PyObject *data = __Pyx_DecompressString(cstring, 1856, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (1744 bytes) */
const char* const cstring = "BZh91AY&SYY\313\336\365\000\005\366_\200D\000@t=@\210\205\312 \277\357\377\360`\014\377\000\007\243\n.\003\301\316\001\351@@\211\020\004\000\2350\006\251\264\307\245F\251<(\320\030\232\000\000\n\247\376*T\320\000\000\000\000\000\032\217\365TI\246\215\014\021\202`C\t\246\232\004\247\352\2254\325\032\217P\311\247\352\230\230\002i\210b\014\001\246\206\215\030\214\200h\000\320\300\250\241\023@\232F\322\n4bi\265\031\032zM$*B\245HR\252\225*C\372~\377\347\365\257\333_\225\377_}k\374o\177\346\021\313\030\207]\272\\\241\2761\017\333;\353Q\023\024b\223\215\360\217\265\221\265\210v\356t*\245P\237\243\267\033\357\\\365\322\372zk]\367\275\260\216X\304:\355\322\345\r\361\210zg}j\"b\214Rq\276\021\351dmb\035\273\275\016\251*\252\225U!\310\350\225*\211*\202\225T\245RR*\277\003\352u;\032>\305I;\225'%6)\205HUI%)\362#\005UT\300\341\303g\r8Tp\255*9994\344\331\311Q\311ZTrS\200&\024D\236DO\222\224\244\252U4#J\366s`\312\315\224)Q(\262\225T\252\252lBJ`\245)J\211\024\245)QJap\265$p\2510\245)JQJR\261$\226\330\266K\245\265\264\245%*\221\024\245)R\225JR\251RC\352D\360\301\023\260\037A\200*AH\235\276\207\342\033\207\344\037\220u\037\250\221\371\237\340nt{\217\232K)/_j\373_\337I)\317z\352\274q\214v\342\253\272~\037\037\225w'\014\0037V\344\ni\303\000\343\277\030\022\332\022\376\245\374Ial\226x\327\032\230\326\273;(ym\262\312\271\213\266\350ym\275\322\322\306\333%\321p\262\266YIaal\225\356\353S\276\374V\223\206\001\233\253r\0054\341\200x\327\342\":\241\332-\277gV\004\333\313i\3138\30696\036\365$\222\354\355\306\376Ns\313\227*\346\2340\014\335[\220)\247\014\0037V\344\ni\303\000\315\325\271\002\232p\3004\244\263ym\326\351\303\000\315\325\271\002\232p\3003un@\246\2340\014\335[\220)\247\014\016[\272\016\243\364H\211\366\377\267\254\337\030\353\233n1\234g6\337\2265\255f\332\263@,\320\005P\0134\002\315\000\263@,\320\0134\001T\006\2663\234\256`\026h\n\226\200p\204\320\0134\002\315\000\263@ZL\350\312\346\001f\200\322\3479\200Y\240\026h\002\250\005\232\001f\200Y\240\026h\005\232\000\252\003N3\234\256`\026h""\n\226\200p\204\320\0134\002\315\000\263@X\337\302`\340\3065q\253\215\375\032%\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000.\304\245'\215e$\266VR\222\262^\227\026\313e\266\313\013\235\274\365\236\235:WT\341\200f\352\334\201M8`\031\272\267 SN\030\006n\255\310\024\323\206\001iiiB\302^\204\301\321\273\203\203\007\006\234\216f\346\346\343st\341\263f\315\233\264\323M2\313,\264\323M4\344\2226Q\025ERT\205eC\010\214\030\021\205%R9\224U\021+\277|\262!\330\354\313&\031aOWa#^\336R\034\276\356i\374\"ua,\330\226\311i-\223okr\220\373\372\010CwF\334\0217:\r&\222\034\310\235\230BM\211\335\347\261!\364;l\203\272&S\010\216xU^\346\355O\t\017\335\t9!'\347P\223\360\016\274\000\000\000\000\034\177\275\366eqfu\202vHT\207\331Bdg\351\350\3060\2220(\252N\202G4\206\0367z8\030C\221\t<I\370?\226wR\335\337&e\372\315\000\027\205\232\000\007\222\302N\214\346\333m\220O(\211\031!R\025H\252%T\244\221\255$<\267F\311\t1O\004\341$A\370\274'8\221l$\264\236\312VN\307T$\3277\271\375\234(\305s;\242x\370$;\035\317\004\312|G\315\023\010d\256\254\021*0NFP\312\037\003q\262\036\010\235\320!\203\326vF\3473\330IF\021\331\334\2338\017\203\317\251!\327t\337\004q\177\024\177\n\252\252\222I$\251$\222I$\222I$\222J\317\322\274\322J\222\254\256;\300\000\000\000\r\245$\226\200\000\000\000\r\245$\226\200\000\000\000\014\332T\222U\240\000\000\000\003iI%\266Wc\342O%N\212\212\356\251\242$\370\341\232\036\3261C\0260i\344s\225)I\244\206\221>\007\335\261\227\036\346\020\236\351\013\"=\210\234\036\003\307\213V\325\266\333@\000\001ewqfW6Yf]\256\322\315\254JI\200\026qY\224\266\262\254Y\231\230\255\240\212\376\250z\025T\210\033\232$:\2773\303\337\241\263\242D%;r\267\332\262`\220\302\"E\312T\337\2773\223[#r`\251\007\004\300S\273 \231%\231gK\273+\270\000\000\374\354\267@\r\256\322Ie\231t\350\000\000\000\000\000\000\000\000\000\000\000\000\000\000t\263t\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000:[\240\000\000\000""\000\000\000\013m\266\333m\270e\234\372\2374\207\301!\350$\365\223\352$\362#\261\315\021#\004\331 \240\230}^RJ\211\315\341\244z\364|\314\217GRa\337\332\255\344\027\006\t\302\206YH\223\002\230\025g\315*\004\233\020\223\325\364\305\273\232N\302\356\220\254\216G\010\230O>\372\\\\-BN$\217a\020y\030\022\216\256\246P\311=\324\"\253L\224\023~i\014\275\016\246\t\0224\244FSdD\302JW\253s(\303$T\245s0`\"\221J4\323\003J\214\262\300\312\243\005\teDZ\212\244J\"\251B\2522R\305,\261R\263\022\316\231[\021\021\021\021\021\021\021\021\026Y\231u\272\366\343\216E\311vVR\262\313m\026E\226\000\000\000\016s\261\004\000@u\333@`\202\000 7m\003\232\347\236wwwww\232\342W+\263\204\201\341\t4c\223\006\336-\235\022\037NV\332\216StD\217\270\221\310\303j\346\003\371\364D\371\220\223\300\221\222y+\242\037\001P\242\241\361`\252L!F\362nq\314\220\301\262M\325\335\324\365\031O\2206'ER\232\021\006\2374O\255HI\324\220\251\020\360\367L\035\307_tI?\370\273\222)\302\204\202\316^\367\250";
    PyObject *data = __Pyx_DecompressString(cstring, 1744, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2181 bytes) */
const char* const cstring = "x\332\355\232O\217\3336\024\304\363Q\366\230,rIR\004\276\365\320C/E[\240\350\241\247\201,\323^\241\372\343H\342v\267\237\276\"e\313\242L\2777\214\2677\005\210W\026G\343\347\241,?\312\277\037\363\262\250\r\212\032}\233\345f\233\345\177\003U6<\037\376U\315\316\226\306m\325Y\345\377\036__\220g\307\302m\177\263Y9\356\357L\2773\373\314\226=\320\233nx,\352\376\341\375O\177\226\305\266\315\332\327\017\357\336\235v\214\216\347\347\356!o\352\256\177\370\3700\337\256\237\273\247\312T/\260u\361\315\232b\207\363\330\343l\260\250\213\036Y\337\267\303\350\343\207\321\36148T\221U\230\313\307=\203\311\2768\214\003eS\037\374\313\372\207\350\361C\235\261\335\037\037~\377\013\277\374\366\353\317\376\301y\005\317\247J\273\342_\343\324\271\335e\177\364\355d\351\377\275\201\257\257=4\217\272\016\323\367\214\355\276l\262\376\323Wg\020\356\270\243\336Tg\266\342\374)k\335q\376\357\035\365\311>t5\341\353E5\273\306nK\343\334O[wT\255{\261\225\373\211p\007\216\033w\324\2449\261\025=e\345\336\035\347\377\336Q\217\354\303V\343v>\316\257=\337U\213\344\222P\311\374!~\325\232\334\307k\327\362b'Y\177\227\341\377]\207\020\306\247\2578\207\3527\357\233\036\325-a\232\276|\236\274\374\346}\225\251n\t\225}\375a\362\362\233\367U\246\272%T\266\231\2546w\327\245x\261U\271S\367|\n\337S\221\354\303V\323=5\355x\240\337\270\243\036\315\211\256\2508\324f7}\203\316\237\336S]\202+_\251\323_\216\274\257>\315\213\255\312\316.;\366\r\256b\234_Ju\347K\217}\203+\031\347\227R\335\371\362c\337\340j\306\371\245T\267\271\230m\356\257MsS*{\031\315f\213\245\307\245 2f\353\323\247o\376\275\034_U=7\305N]z\235Ec\321\263\235\356\375\373\301\367\313T\306\275\263\366i\360\272\336\267\030Zv\020\261\2757\227}g\271\r\277\340.\247\3052\347k\371Eyi\327b\307\214y\204\317>>\370\277\314\004S\307}o]g\373\307\253\334\346\257\363\341,:Oi\360\3642\323\347\375\341\031pZV\317\307\246\372\375\310t6\345\366t\237\000\373\242\316\312\341\350\330\230;\347\346\373O7'b\207\234\206\302#:s:e\247\317B\326\036\272w\341gh~\010\266Y\333\026\246}\027dT\275""\342h\302]\365\260gf\344:\001\367\037YY\366\315\360\200\246F\347g!\"\332\347MY\232\274\027EU\366\202\326\354lnHY\227\017o\325\264\262\272\250)\323I\306\230v\266bL/2\306t\3336\331.\317\272hH\347\273\rb\334\223H\212{\022\311qGd\302\233\270\250\305\270#2\306T\216;\"cL\305\270]\257&F\355\005R\314^ G\274\220\010E\217J1\332\205D3\223#]H43\371\314\035\257/\261!\177]\351^\353<68\336\025\022'\341$\221\246\341$\221'\342J$\274\341\263V\234\214+\221n(O\310\225H7\024'\305\177\010\304lG\205\024\355\250\220\223]j\204\262OR1\327\245F\265\223S]jT;1Sw\223N\214\324\013\244D\275@\016t!\021*\036\225b\234\013\211f&\207\271\220hfb\224C\347$&\351\306\245 \335\270\234c\250\020\212\365B1\305P\241X\311\031\206\n\305JH\360te=\226\303\002f\030)vf\027\033\372\274\330\333\267Y\335\225Yof\315\237\177-\245\357\030\025\312\204\250\035\307R#\277}\265\327XjT;uj\330\026c\224j'\370\227\317Z\246\203B\311tP\250\231\006\032\271j'\3252\r4\252\235\232i\240Q\355\264L\207\225\241\222\351\240P2\035\024j\246\201F\256\332I\265L\003\215j\247f\032hT;-\323\215\026\351FKt\243\007\272\241\363\334\350qn\35047z\230\033:\313\215\034%\265XV\027\312\304\"\231_ \023\213c~aL,\212\371\005\261\034\245\277\351/f9*\2440G\205\234\346R#\024}\222\212y.5\252\235\234\350R\243\332\311\231\252K\336N]\363v\304\242\267\343W\275\035\261\354\355\370uoG,|;~\345\333\351K_\177\213P\214\324\t\304D\235@\t4\224H\025{\245\034g(\321\314\2240C\211f&Fi\365.\324\352m\250e\372P\233\320\210Z\246\023\265\t\255\250ezQ\233\320\214Z\242\033\265z;j\365~\3242\r\251M\350H-\323\222\332\204\236\3242M\251M\350J-\321\226Z\275/\265zcj\231\316\324&\264\246\226\351MmBsj\231\356\324&\264\247\226\350O\255\332\240Z\265C\265D\213j\371\036\325\022M\252\345\273TK\264\251\226\357S\255\336\250\372[\001\007\323O?!]\206\342{\237\232\256/\213\355\345\367\322\310\317Q\325\261@\336T\325\342\347\250\255\335\357\215\373\261\2525\207\242s5w\257Ue\372\266\310\247[\023\341oQ\213\237\241\374kv}\326\333\356\372\226\363\215S\342\233-L\374\333\326\335\212\216\0364\274\316\276\361o\377""\331\264]\321\324\227\203\236M\275k\332\371\230\027\207wVvf\360k^\303\235\356\220\361G\333\3717\341\241\316J\374\223\025.\351\276\270\3216xQs\214\024\272o\215\t&\314\355\214x\034\2552r~\221\253W\230\245{}\016\\\377PY\3458^N\211\3413T\226\315er\007\361a\n\023y08\235\036\302\3111X\257\024\366Ja\257\024\366Ja\257\024\366Ja\257\024\366Ja\257\024\366Ja\257\024\366Ja\257\024\366Ja\257\024\366Ja\273\021H\0306nq\330\020@l\334 \261\241\241\330\210\260\330X\302\330\210\320\330\210\340\330\240xlP@6H\"\033iH6H&\033iP6H*\033iX68.\033\024\230\r\212\314\006\211f#\215\315\006\tg#\215\316\006\211g#\215\317\006\007hC%\264\241\"\332 \030m\360\2206\010J\033<\246\r\202\323\006\017jC'\265!\240\332\020Ym\020\2606\010Z\033\024\256\215\024^\033\024\260\215\024b\033\024\262\215\024f\033\014\264\r\235\332\206\216m\203\341\266\221\000n\203!\267\221\200n\203a\267\221\000o\203\240\267\241\342\333P\371m\020\0007x\202\033\004\302\r\236\341\006\001q\203\247\270\241c\334\3208nh 7t\222\0334\312\r\235\345\006\rsC\247\271A\343\334Pyn\010@7\342D7n#\335\320\231n\350P7\030\252\033\tX7\030\256\033\t`7\030\262\033\th7\010\266\033:\334\r\235\356\006\203w#\201\357\006\003x#\201\360\006\203x#\201\361\006\001yC\247\274\241c\336`8o$\200\336`Ho$\240\336`Xo$\300\336 ho\250\2707T\336\033\004\360\r\236\370\006\201|\203g\276A@\337\340\251o\350\3307T\356\033*\370\r\202\374\006\217~\203`\277\301\303\337 \350o\360\3707t\376\033:\000\016\235\000\007\203\200#\201\001\007\003\201#\201\002\007\203\201#\201\003\007\001\202C'\301\241\243\340`Xp$\300\340`hp$\340\340`xp$\000\341 \210p\250H8T&\034\004\024\016\236\n\007\201\205\203\347\302A\200\341\340\311p\350h8\0106\034\004\034\016\212\016G\n\036\016\212\017G\n \016\212\020G\n\"\016\206\021\007\001\211\203\240\304Aa\342H\341\304A\201\342H!\305A\241\342Ha\305\301\300\342 hq\020\2708(^\034)\3008(b\034)\3108(f\034)\3208\030j\034:6\016\235\033\007\003\216#\201\034\007\203\216#\201\035\007\003\217#\201\036\007\201\217C\340\307\021\007\310!\020\344\320\020rP\0149B\210\034!E\216(F""\016\205#\307M\220\034\022I\216\233(9T\226\034Q\230\034\267hrP89D\236\034\001P\216\333D9n#\345P\231r\304\240r\334\246\312\261\304\312\261\340\312\021\202\345X\220\345 \320r\314\330\362\377\000 =\275N";
    PyObject *data = __Pyx_DecompressString(cstring, 2181, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (23609 bytes) */
const char* const bytes = "?cline_in_traceback__main____module____name____pyx_capi____qualname__setdefault__test__int (CUlibrary)\000\000int (CUmodule)\000\000int (int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *)\000int (nvshmem_team_t *, nvshmem_team_config_t *, long, int, int)\000int (nvshmem_team_t)\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, char *, char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, cudaStream_t)\000\000int (nvshmem_team_t, double *, double const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, double *, double const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, float *, float const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, half *, half const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int *, int const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_t)\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int"" (nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, long *, long const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, short *, short const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem""_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t)\000int (nvshmemx_team_uniqueid_t *)\000int (nvshmemx_uniqueid_t *)\000int (unsigned int, nvshmemx_init_attr_t *)\000int (void *, nvshmemx_init_attr_t *)\000int (void *, size_t)\000int (void)\000\000\000void (cudaStream_t)\000\000\000void (int *, int *)\000void (int *, int *, int *)\000void (nvshmem_team_t)\000void (nvshmem_team_t, nvshmem_team_config_t *)\000void (uint64_t *, int, uint64_t, cudaStream_t)\000void (uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void *)\000void (void *, void const *, size_t, int, cudaStream_t)\000\000void (void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void)\000\000void *(nvshmem_team_t, void const *)\000void *(size_t)\000void *(size_t, size_t)\000\000void *(void *, size_t, int)\000void *(void const *, int)\000nvshmemx_culibrary_finalize\000nvshmemx_culibrary_init\000nvshmemx_cumodule_finalize\000nvshmemx_cumodule_init\000nvshmemx_set_attr_uniqueid_args\000nvshmemx_team_init\000nvshmem_barrier\000nvshmem_team_my_pe\000nvshmem_team_n_pes\000nvshmemx_longlong_alltoall_on_stream\000nvshmemx_longlong_fcollect_on_stream\000nvshmemx_longlong_max_reduce_on_stream\000nvshmemx_longlong_max_reducescatter_on_stream\000nvshmemx_longlong_min_reduce_on_stream\000nvshmemx_longlong_min_reducescatter_on_stream\000nvshmemx_longlong_sum_reduce_on_stream\000nvshmemx_longlong_sum_reducescatter_on_stream\000nvshmemx_longlong_broadcast_on_stream\000nvshmemx_bfloat16_alltoall_on_stream\000nvshmemx_bfloat16_fcollect_on_stream\000nvshmemx_bfloat16_max_reduce_on_stream\000nvshmemx_bfloat16_max_reducescatter_on_stream\000nvshmemx_bfloat16_min_reduce_on_stream\000nvshmemx_bfloat16_min_reducescatter_on_stream\000nvshmemx_bfloat16_sum_reduce_on_stream\000nvshmemx_bfloat16_sum_reducescatter_on_stream\000nvshmemx_bfloat16_broadcast_on_stream\000nvshmemx_char_alltoall_on_stream\000nvshmemx_char_fcollect_on_stream\000nvshmemx_char_max_reduce_on_stream\000nvshmem""x_char_max_reducescatter_on_stream\000nvshmemx_char_min_reduce_on_stream\000nvshmemx_char_min_reducescatter_on_stream\000nvshmemx_char_sum_reduce_on_stream\000nvshmemx_char_sum_reducescatter_on_stream\000nvshmemx_char_broadcast_on_stream\000nvshmemx_barrier_on_stream\000nvshmemx_team_sync_on_stream\000nvshmemx_double_alltoall_on_stream\000nvshmemx_double_fcollect_on_stream\000nvshmemx_double_max_reduce_on_stream\000nvshmemx_double_max_reducescatter_on_stream\000nvshmemx_double_min_reduce_on_stream\000nvshmemx_double_min_reducescatter_on_stream\000nvshmemx_double_sum_reduce_on_stream\000nvshmemx_double_sum_reducescatter_on_stream\000nvshmemx_double_broadcast_on_stream\000nvshmemx_float_alltoall_on_stream\000nvshmemx_float_fcollect_on_stream\000nvshmemx_float_max_reduce_on_stream\000nvshmemx_float_max_reducescatter_on_stream\000nvshmemx_float_min_reduce_on_stream\000nvshmemx_float_min_reducescatter_on_stream\000nvshmemx_float_sum_reduce_on_stream\000nvshmemx_float_sum_reducescatter_on_stream\000nvshmemx_float_broadcast_on_stream\000nvshmemx_half_alltoall_on_stream\000nvshmemx_half_fcollect_on_stream\000nvshmemx_half_max_reduce_on_stream\000nvshmemx_half_max_reducescatter_on_stream\000nvshmemx_half_min_reduce_on_stream\000nvshmemx_half_min_reducescatter_on_stream\000nvshmemx_half_sum_reduce_on_stream\000nvshmemx_half_sum_reducescatter_on_stream\000nvshmemx_half_broadcast_on_stream\000nvshmemx_int_alltoall_on_stream\000nvshmemx_int_fcollect_on_stream\000nvshmemx_int_max_reduce_on_stream\000nvshmemx_int_max_reducescatter_on_stream\000nvshmemx_int_min_reduce_on_stream\000nvshmemx_int_min_reducescatter_on_stream\000nvshmemx_int_sum_reduce_on_stream\000nvshmemx_int_sum_reducescatter_on_stream\000nvshmemx_int_broadcast_on_stream\000nvshmem_team_split_strided\000nvshmem_team_split_2d\000nvshmem_team_translate_pe\000nvshmemx_int16_alltoall_on_stream\000nvshmemx_int16_fcollect_on_stream\000nvshmemx_int16_max_reduce_on_stream\000nvshmemx_int16_max_reducescatter_on_stream\000nvsh""memx_int16_min_reduce_on_stream\000nvshmemx_int16_min_reducescatter_on_stream\000nvshmemx_int16_sum_reduce_on_stream\000nvshmemx_int16_sum_reducescatter_on_stream\000nvshmemx_int16_broadcast_on_stream\000nvshmemx_int32_alltoall_on_stream\000nvshmemx_int32_fcollect_on_stream\000nvshmemx_int32_max_reduce_on_stream\000nvshmemx_int32_max_reducescatter_on_stream\000nvshmemx_int32_min_reduce_on_stream\000nvshmemx_int32_min_reducescatter_on_stream\000nvshmemx_int32_sum_reduce_on_stream\000nvshmemx_int32_sum_reducescatter_on_stream\000nvshmemx_int32_broadcast_on_stream\000nvshmemx_int64_alltoall_on_stream\000nvshmemx_int64_fcollect_on_stream\000nvshmemx_int64_max_reduce_on_stream\000nvshmemx_int64_max_reducescatter_on_stream\000nvshmemx_int64_min_reduce_on_stream\000nvshmemx_int64_min_reducescatter_on_stream\000nvshmemx_int64_sum_reduce_on_stream\000nvshmemx_int64_sum_reducescatter_on_stream\000nvshmemx_int64_broadcast_on_stream\000nvshmemx_int8_alltoall_on_stream\000nvshmemx_int8_fcollect_on_stream\000nvshmemx_int8_max_reduce_on_stream\000nvshmemx_int8_max_reducescatter_on_stream\000nvshmemx_int8_min_reduce_on_stream\000nvshmemx_int8_min_reducescatter_on_stream\000nvshmemx_int8_sum_reduce_on_stream\000nvshmemx_int8_sum_reducescatter_on_stream\000nvshmemx_int8_broadcast_on_stream\000nvshmemx_long_alltoall_on_stream\000nvshmemx_long_fcollect_on_stream\000nvshmemx_long_max_reduce_on_stream\000nvshmemx_long_max_reducescatter_on_stream\000nvshmemx_long_min_reduce_on_stream\000nvshmemx_long_min_reducescatter_on_stream\000nvshmemx_long_sum_reduce_on_stream\000nvshmemx_long_sum_reducescatter_on_stream\000nvshmemx_long_broadcast_on_stream\000nvshmemx_short_alltoall_on_stream\000nvshmemx_short_fcollect_on_stream\000nvshmemx_short_max_reduce_on_stream\000nvshmemx_short_max_reducescatter_on_stream\000nvshmemx_short_min_reduce_on_stream\000nvshmemx_short_min_reducescatter_on_stream\000nvshmemx_short_sum_reduce_on_stream\000nvshmemx_short_sum_reducescatter_on_stream\000nvshmemx_short_br""oadcast_on_stream\000nvshmemx_schar_alltoall_on_stream\000nvshmemx_schar_fcollect_on_stream\000nvshmemx_schar_max_reduce_on_stream\000nvshmemx_schar_max_reducescatter_on_stream\000nvshmemx_schar_min_reduce_on_stream\000nvshmemx_schar_min_reducescatter_on_stream\000nvshmemx_schar_sum_reduce_on_stream\000nvshmemx_schar_sum_reducescatter_on_stream\000nvshmemx_schar_broadcast_on_stream\000nvshmemx_size_alltoall_on_stream\000nvshmemx_size_fcollect_on_stream\000nvshmemx_size_max_reduce_on_stream\000nvshmemx_size_max_reducescatter_on_stream\000nvshmemx_size_min_reduce_on_stream\000nvshmemx_size_min_reducescatter_on_stream\000nvshmemx_size_sum_reduce_on_stream\000nvshmemx_size_sum_reducescatter_on_stream\000nvshmemx_size_broadcast_on_stream\000nvshmemx_uint16_alltoall_on_stream\000nvshmemx_uint16_fcollect_on_stream\000nvshmemx_uint16_max_reduce_on_stream\000nvshmemx_uint16_max_reducescatter_on_stream\000nvshmemx_uint16_min_reduce_on_stream\000nvshmemx_uint16_min_reducescatter_on_stream\000nvshmemx_uint16_sum_reduce_on_stream\000nvshmemx_uint16_sum_reducescatter_on_stream\000nvshmemx_uint16_broadcast_on_stream\000nvshmemx_uint32_alltoall_on_stream\000nvshmemx_uint32_fcollect_on_stream\000nvshmemx_uint32_max_reduce_on_stream\000nvshmemx_uint32_max_reducescatter_on_stream\000nvshmemx_uint32_min_reduce_on_stream\000nvshmemx_uint32_min_reducescatter_on_stream\000nvshmemx_uint32_sum_reduce_on_stream\000nvshmemx_uint32_sum_reducescatter_on_stream\000nvshmemx_uint32_broadcast_on_stream\000nvshmemx_uint64_alltoall_on_stream\000nvshmemx_uint64_fcollect_on_stream\000nvshmemx_uint64_max_reduce_on_stream\000nvshmemx_uint64_max_reducescatter_on_stream\000nvshmemx_uint64_min_reduce_on_stream\000nvshmemx_uint64_min_reducescatter_on_stream\000nvshmemx_uint64_sum_reduce_on_stream\000nvshmemx_uint64_sum_reducescatter_on_stream\000nvshmemx_uint64_broadcast_on_stream\000nvshmemx_uint8_alltoall_on_stream\000nvshmemx_uint8_fcollect_on_stream\000nvshmemx_uint8_max_reduce_on_stream\000nvshmemx_uint""8_max_reducescatter_on_stream\000nvshmemx_uint8_min_reduce_on_stream\000nvshmemx_uint8_min_reducescatter_on_stream\000nvshmemx_uint8_sum_reduce_on_stream\000nvshmemx_uint8_sum_reducescatter_on_stream\000nvshmemx_uint8_broadcast_on_stream\000nvshmemx_team_get_uniqueid\000nvshmemx_get_uniqueid\000nvshmemx_hostlib_init_attr\000nvshmemx_set_attr_mpi_comm_args\000nvshmemx_buffer_unregister_symmetric\000nvshmem_my_pe\000nvshmem_n_pes\000nvshmemx_init_status\000nvshmemx_barrier_all_on_stream\000nvshmemx_quiet_on_stream\000nvshmemx_sync_all_on_stream\000nvshmem_info_get_version\000nvshmemx_vendor_get_version_info\000nvshmem_team_destroy\000nvshmem_team_get_config\000nvshmemx_signal_wait_until_on_stream\000nvshmemx_signal_op_on_stream\000nvshmem_free\000nvshmemx_getmem_on_stream\000nvshmemx_putmem_on_stream\000nvshmemx_putmem_signal_on_stream\000nvshmem_barrier_all\000nvshmemx_hostlib_finalize\000nvshmemx_mc_ptr\000nvshmem_malloc\000nvshmem_align\000nvshmem_calloc\000nvshmemx_buffer_register_symmetric\000nvshmem_ptrint (CUlibrary)\000\000int (CUmodule)\000\000int (int const , int const , nvshmemx_uniqueid_t const *, nvshmemx_init_attr_t *)\000int (nvshmem_team_t *, nvshmem_team_config_t *, long, int, int)\000int (nvshmem_team_t)\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, PY_LONG_LONG *, PY_LONG_LONG const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, __nv_bfloat16 *, __nv_bfloat16 const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, char *, char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, char *, char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, cudaStream_t)\000\000int (nvshmem_team_t, double *, double const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, double *, d""ouble const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, float *, float const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, float *, float const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, half *, half const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, half *, half const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int *, int const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int *, int const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int, int, int, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_config_t const *, long, nvshmem_team_t *, nvshmem_team_config_t const *, long, nvshmem_team_t *)\000int (nvshmem_team_t, int, nvshmem_team_t)\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int16_t *, int16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int32_t *, int32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int64_t *, int64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, int8_t *, int8_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, long *, long const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, long *, long const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, short *, short const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, short *, short const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, signed char *, signed char const *, size_t, cudaStream_t)\000\000\000\000\000\000\000""\000int (nvshmem_team_t, signed char *, signed char const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, size_t *, size_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, size_t *, size_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint16_t *, uint16_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint32_t *, uint32_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint64_t *, uint64_t const *, size_t, int, cudaStream_t)\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, cudaStream_t)\000\000\000\000\000\000\000\000int (nvshmem_team_t, uint8_t *, uint8_t const *, size_t, int, cudaStream_t)\000int (nvshmemx_team_uniqueid_t *)\000int (nvshmemx_uniqueid_t *)\000int (unsigned int, nvshmemx_init_attr_t *)\000int (void *, nvshmemx_init_attr_t *)\000int (void *, size_t)\000int (void)\000\000\000void (cudaStream_t)\000\000\000void (int *, int *)\000void (int *, int *, int *)\000void (nvshmem_team_t)\000void (nvshmem_team_t, nvshmem_team_config_t *)\000void (uint64_t *, int, uint64_t, cudaStream_t)\000void (uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void *)\000void (void *, void const *, size_t, int, cudaStream_t)\000\000void (void *, void const *, size_t, uint64_t *, uint64_t, int, int, cudaStream_t)\000void (void)\000\000void *(nvshmem_team_t, void const *)\000void *(size_t)\000void *(size_t, size_t)\000\000void *(void *, size_t, int)\000void *(void const *, int)\000_nvshmemx_culibrary_finalize\000_nvshmemx_culibrary_init\000_nvshmemx_cumodule_finalize\000_nvshmemx_cumodule_init\000_nvshmemx_set_attr_uniqueid_args\000_nvshmemx_team_init\000_nvshmem_barrier\000_nvsh""mem_team_my_pe\000_nvshmem_team_n_pes\000_nvshmemx_longlong_alltoall_on_stream\000_nvshmemx_longlong_fcollect_on_stream\000_nvshmemx_longlong_max_reduce_on_stream\000_nvshmemx_longlong_max_reducescatter_on_stream\000_nvshmemx_longlong_min_reduce_on_stream\000_nvshmemx_longlong_min_reducescatter_on_stream\000_nvshmemx_longlong_sum_reduce_on_stream\000_nvshmemx_longlong_sum_reducescatter_on_stream\000_nvshmemx_longlong_broadcast_on_stream\000_nvshmemx_bfloat16_alltoall_on_stream\000_nvshmemx_bfloat16_fcollect_on_stream\000_nvshmemx_bfloat16_max_reduce_on_stream\000_nvshmemx_bfloat16_max_reducescatter_on_stream\000_nvshmemx_bfloat16_min_reduce_on_stream\000_nvshmemx_bfloat16_min_reducescatter_on_stream\000_nvshmemx_bfloat16_sum_reduce_on_stream\000_nvshmemx_bfloat16_sum_reducescatter_on_stream\000_nvshmemx_bfloat16_broadcast_on_stream\000_nvshmemx_char_alltoall_on_stream\000_nvshmemx_char_fcollect_on_stream\000_nvshmemx_char_max_reduce_on_stream\000_nvshmemx_char_max_reducescatter_on_stream\000_nvshmemx_char_min_reduce_on_stream\000_nvshmemx_char_min_reducescatter_on_stream\000_nvshmemx_char_sum_reduce_on_stream\000_nvshmemx_char_sum_reducescatter_on_stream\000_nvshmemx_char_broadcast_on_stream\000_nvshmemx_barrier_on_stream\000_nvshmemx_team_sync_on_stream\000_nvshmemx_double_alltoall_on_stream\000_nvshmemx_double_fcollect_on_stream\000_nvshmemx_double_max_reduce_on_stream\000_nvshmemx_double_max_reducescatter_on_stream\000_nvshmemx_double_min_reduce_on_stream\000_nvshmemx_double_min_reducescatter_on_stream\000_nvshmemx_double_sum_reduce_on_stream\000_nvshmemx_double_sum_reducescatter_on_stream\000_nvshmemx_double_broadcast_on_stream\000_nvshmemx_float_alltoall_on_stream\000_nvshmemx_float_fcollect_on_stream\000_nvshmemx_float_max_reduce_on_stream\000_nvshmemx_float_max_reducescatter_on_stream\000_nvshmemx_float_min_reduce_on_stream\000_nvshmemx_float_min_reducescatter_on_stream\000_nvshmemx_float_sum_reduce_on_stream\000_nvshmemx_float_sum_reducescatter_on_stream""\000_nvshmemx_float_broadcast_on_stream\000_nvshmemx_half_alltoall_on_stream\000_nvshmemx_half_fcollect_on_stream\000_nvshmemx_half_max_reduce_on_stream\000_nvshmemx_half_max_reducescatter_on_stream\000_nvshmemx_half_min_reduce_on_stream\000_nvshmemx_half_min_reducescatter_on_stream\000_nvshmemx_half_sum_reduce_on_stream\000_nvshmemx_half_sum_reducescatter_on_stream\000_nvshmemx_half_broadcast_on_stream\000_nvshmemx_int_alltoall_on_stream\000_nvshmemx_int_fcollect_on_stream\000_nvshmemx_int_max_reduce_on_stream\000_nvshmemx_int_max_reducescatter_on_stream\000_nvshmemx_int_min_reduce_on_stream\000_nvshmemx_int_min_reducescatter_on_stream\000_nvshmemx_int_sum_reduce_on_stream\000_nvshmemx_int_sum_reducescatter_on_stream\000_nvshmemx_int_broadcast_on_stream\000_nvshmem_team_split_strided\000_nvshmem_team_split_2d\000_nvshmem_team_translate_pe\000_nvshmemx_int16_alltoall_on_stream\000_nvshmemx_int16_fcollect_on_stream\000_nvshmemx_int16_max_reduce_on_stream\000_nvshmemx_int16_max_reducescatter_on_stream\000_nvshmemx_int16_min_reduce_on_stream\000_nvshmemx_int16_min_reducescatter_on_stream\000_nvshmemx_int16_sum_reduce_on_stream\000_nvshmemx_int16_sum_reducescatter_on_stream\000_nvshmemx_int16_broadcast_on_stream\000_nvshmemx_int32_alltoall_on_stream\000_nvshmemx_int32_fcollect_on_stream\000_nvshmemx_int32_max_reduce_on_stream\000_nvshmemx_int32_max_reducescatter_on_stream\000_nvshmemx_int32_min_reduce_on_stream\000_nvshmemx_int32_min_reducescatter_on_stream\000_nvshmemx_int32_sum_reduce_on_stream\000_nvshmemx_int32_sum_reducescatter_on_stream\000_nvshmemx_int32_broadcast_on_stream\000_nvshmemx_int64_alltoall_on_stream\000_nvshmemx_int64_fcollect_on_stream\000_nvshmemx_int64_max_reduce_on_stream\000_nvshmemx_int64_max_reducescatter_on_stream\000_nvshmemx_int64_min_reduce_on_stream\000_nvshmemx_int64_min_reducescatter_on_stream\000_nvshmemx_int64_sum_reduce_on_stream\000_nvshmemx_int64_sum_reducescatter_on_stream\000_nvshmemx_int64_broadcast_on_stream\000_nvshmemx_int8_al""ltoall_on_stream\000_nvshmemx_int8_fcollect_on_stream\000_nvshmemx_int8_max_reduce_on_stream\000_nvshmemx_int8_max_reducescatter_on_stream\000_nvshmemx_int8_min_reduce_on_stream\000_nvshmemx_int8_min_reducescatter_on_stream\000_nvshmemx_int8_sum_reduce_on_stream\000_nvshmemx_int8_sum_reducescatter_on_stream\000_nvshmemx_int8_broadcast_on_stream\000_nvshmemx_long_alltoall_on_stream\000_nvshmemx_long_fcollect_on_stream\000_nvshmemx_long_max_reduce_on_stream\000_nvshmemx_long_max_reducescatter_on_stream\000_nvshmemx_long_min_reduce_on_stream\000_nvshmemx_long_min_reducescatter_on_stream\000_nvshmemx_long_sum_reduce_on_stream\000_nvshmemx_long_sum_reducescatter_on_stream\000_nvshmemx_long_broadcast_on_stream\000_nvshmemx_short_alltoall_on_stream\000_nvshmemx_short_fcollect_on_stream\000_nvshmemx_short_max_reduce_on_stream\000_nvshmemx_short_max_reducescatter_on_stream\000_nvshmemx_short_min_reduce_on_stream\000_nvshmemx_short_min_reducescatter_on_stream\000_nvshmemx_short_sum_reduce_on_stream\000_nvshmemx_short_sum_reducescatter_on_stream\000_nvshmemx_short_broadcast_on_stream\000_nvshmemx_schar_alltoall_on_stream\000_nvshmemx_schar_fcollect_on_stream\000_nvshmemx_schar_max_reduce_on_stream\000_nvshmemx_schar_max_reducescatter_on_stream\000_nvshmemx_schar_min_reduce_on_stream\000_nvshmemx_schar_min_reducescatter_on_stream\000_nvshmemx_schar_sum_reduce_on_stream\000_nvshmemx_schar_sum_reducescatter_on_stream\000_nvshmemx_schar_broadcast_on_stream\000_nvshmemx_size_alltoall_on_stream\000_nvshmemx_size_fcollect_on_stream\000_nvshmemx_size_max_reduce_on_stream\000_nvshmemx_size_max_reducescatter_on_stream\000_nvshmemx_size_min_reduce_on_stream\000_nvshmemx_size_min_reducescatter_on_stream\000_nvshmemx_size_sum_reduce_on_stream\000_nvshmemx_size_sum_reducescatter_on_stream\000_nvshmemx_size_broadcast_on_stream\000_nvshmemx_uint16_alltoall_on_stream\000_nvshmemx_uint16_fcollect_on_stream\000_nvshmemx_uint16_max_reduce_on_stream\000_nvshmemx_uint16_max_reducescatter_on_stream""\000_nvshmemx_uint16_min_reduce_on_stream\000_nvshmemx_uint16_min_reducescatter_on_stream\000_nvshmemx_uint16_sum_reduce_on_stream\000_nvshmemx_uint16_sum_reducescatter_on_stream\000_nvshmemx_uint16_broadcast_on_stream\000_nvshmemx_uint32_alltoall_on_stream\000_nvshmemx_uint32_fcollect_on_stream\000_nvshmemx_uint32_max_reduce_on_stream\000_nvshmemx_uint32_max_reducescatter_on_stream\000_nvshmemx_uint32_min_reduce_on_stream\000_nvshmemx_uint32_min_reducescatter_on_stream\000_nvshmemx_uint32_sum_reduce_on_stream\000_nvshmemx_uint32_sum_reducescatter_on_stream\000_nvshmemx_uint32_broadcast_on_stream\000_nvshmemx_uint64_alltoall_on_stream\000_nvshmemx_uint64_fcollect_on_stream\000_nvshmemx_uint64_max_reduce_on_stream\000_nvshmemx_uint64_max_reducescatter_on_stream\000_nvshmemx_uint64_min_reduce_on_stream\000_nvshmemx_uint64_min_reducescatter_on_stream\000_nvshmemx_uint64_sum_reduce_on_stream\000_nvshmemx_uint64_sum_reducescatter_on_stream\000_nvshmemx_uint64_broadcast_on_stream\000_nvshmemx_uint8_alltoall_on_stream\000_nvshmemx_uint8_fcollect_on_stream\000_nvshmemx_uint8_max_reduce_on_stream\000_nvshmemx_uint8_max_reducescatter_on_stream\000_nvshmemx_uint8_min_reduce_on_stream\000_nvshmemx_uint8_min_reducescatter_on_stream\000_nvshmemx_uint8_sum_reduce_on_stream\000_nvshmemx_uint8_sum_reducescatter_on_stream\000_nvshmemx_uint8_broadcast_on_stream\000_nvshmemx_team_get_uniqueid\000_nvshmemx_get_uniqueid\000_nvshmemx_hostlib_init_attr\000_nvshmemx_set_attr_mpi_comm_args\000_nvshmemx_buffer_unregister_symmetric\000_nvshmem_my_pe\000_nvshmem_n_pes\000_nvshmemx_init_status\000_nvshmemx_barrier_all_on_stream\000_nvshmemx_quiet_on_stream\000_nvshmemx_sync_all_on_stream\000_nvshmem_info_get_version\000_nvshmemx_vendor_get_version_info\000_nvshmem_team_destroy\000_nvshmem_team_get_config\000_nvshmemx_signal_wait_until_on_stream\000_nvshmemx_signal_op_on_stream\000_nvshmem_free\000_nvshmemx_getmem_on_stream\000_nvshmemx_putmem_on_stream\000_nvshmemx_putmem_signal_on_stream\000_nv""shmem_barrier_all\000_nvshmemx_hostlib_finalize\000_nvshmemx_mc_ptr\000_nvshmem_malloc\000_nvshmem_align\000_nvshmem_calloc\000_nvshmemx_buffer_register_symmetric\000_nvshmem_ptr";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 9; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 1) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 9; i < 11; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 11; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 9;
      for (Py_ssize_t i=0; i<2; ++i) {
        #if PY_VERSION_HEX >= 0x030F0000
        PyUnstable_SetImmortal(table[i]);
        #elif CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        if ((PY_SSIZE_T_MAX <= _Py_IMMORTAL_REFCNT_LOCAL)) break;
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], ((Py_ssize_t)_Py_IMMORTAL_REFCNT_LOCAL + 1));
        }
        #else
        if ((PY_SSIZE_T_MAX < _Py_IMMORTAL_INITIAL_REFCNT)) break;
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */

static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* ErrOccurredWithGIL */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void) {
  int err;
  PyGILState_STATE _save = PyGILState_Ensure();
  err = !!PyErr_Occurred();
  PyGILState_Release(_save);
  return err;
}

/* FunctionExport */
static int __Pyx_ExportFunction(PyObject *api_dict, const char *name, void (*f)(void), const char *sig) {
    PyObject *cobj;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(api_dict, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(cobj);
    return -1;
}

/* GetApiDict */
static PyObject *__Pyx_ApiExport_GetApiDict(void) {
    PyObject *d;
    if (__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_capi, &d) == -1)
        return NULL;
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_pyx_capi, d) < 0)
            goto bad;
    }
    return d;
bad:
    Py_XDECREF(d);
    return NULL;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_8
#define __PYX_HAVE_RT_ImportFromPxd_3_2_8
static int __Pyx_ImportFromPxd_3_2_8(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_8
#define __PYX_HAVE_RT_ImportFunction_3_2_8
static int __Pyx_ImportFunction_3_2_8(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_8(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* dict_setdefault (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* PyDictVersioning (used by CLineInTraceback) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by CLineInTraceback) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u_);
    }
    goto done;
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectVectorCallKwBuilder) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectVectorCallKwBuilder (used by CIntToPy) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    Py_INCREF(key);
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPyVerify (used by CIntFromPy) */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* AddModuleRef */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
